Package: ravetools
Type: Package
Title: Signal and Image Processing Toolbox for Analyzing Intracranial
        Electroencephalography Data
Version: 0.2.6
Language: en-US
Authors@R: c(
    person("Zhengjia", "Wang", email = "dipterix.wang@gmail.com", role = c("aut", "cre")),
    person("John", "Magnotti", email = "john.magnotti@pennmedicine.upenn.edu", role = c("aut")),
    person("Michael", "Beauchamp", email = "michael.beauchamp@pennmedicine.upenn.edu", role = c("aut")),
    person("Trustees of the University of Pennsylvania", role = c("cph"), 
            comment = "All files in this package unless explicitly stated in the file or listed in the 'Copyright' section below."),
    person("Karim", "Rahim", email = "karim.rahim@queensu.ca", role = c("cph", "ctb"), 
            comment = "Contributed to src/ffts.h and stc/ffts.cpp"),
    person("Thomas", "Possidente", role = c("cph", "ctb"), 
            comment = "Contributed to R/multitaper.R"),
    person("Michael", "Prerau", role = c("cph", "ctb"), 
            comment = "Contributed to R/multitaper.R"),
    person("Marcus", "Geelnard", role = c("ctb", "cph"), 
            comment = "TinyThread library, tinythreadpp.bitsnbites.eu, located at inst/include/tthread/"),
    person("Stefan", "Schlager", role = c("ctb", "cph"), 
            comment = "R-vcg interface, located at src/vcgCommon.h"),
    person("Visual Computing Lab, ISTI", role = c("ctb", "cph"), 
            comment = "Copyright holder of vcglib, located at src/vcglib/")
    )
Description: Implemented fast and memory-efficient Notch-filter, 
    Welch-periodogram, discrete wavelet spectrogram for minutes of 
    high-resolution signals, fast 3D convolution, image registration,
    3D mesh manipulation; providing fundamental toolbox for intracranial 
    Electroencephalography (iEEG) pipelines. 
    Documentation and examples about 'RAVE' project are provided at 
    <https://rave.wiki>, and the paper by John F. Magnotti, 
    Zhengjia Wang, Michael S. Beauchamp (2020) 
    <doi:10.1016/j.neuroimage.2020.117341>; see 'citation("ravetools")' for 
    details.
BugReports: https://github.com/dipterix/ravetools/issues
URL: https://rave.wiki, https://dipterix.org/ravetools/,
        https://github.com/dipterix/ravetools
License: GPL (>= 2)
Encoding: UTF-8
Depends: R (>= 4.0.0)
SystemRequirements: fftw3 (libfftw3-dev (deb), or fftw-devel (rpm)),
        pkg-config
Copyright: Karim Rahim (author of R package 'fftwtools', licensed under
        'GPL-2' or later) is the original author of 'src/ffts.h' and
        'src/ffts.cpp'. Prerau's Lab wrote the original
        'R/multitaper.R', licensed under 'MIT'. Marcus Geelnard wrote
        the source code of 'TinyThread' library ('MIT' license) located
        at 'inst/include/tthread'. Stefan Schlager wrote the original
        code that converts R objects to 'vcg' (see 'src/vcgCommon.h',
        licensed under 'GPL-2' or later). Visual Computing Lab is the
        copyright holder of 'vcglib' source code (see 'src/vcglib',
        licensed under GPL-2 or later).
Imports: graphics, stats, filearray (>= 0.1.3), Rcpp, waveslim (>=
        1.8.2), pracma, digest (>= 0.6.29), splines, RNiftyReg (>=
        2.7.1), R6 (>= 2.5.1), gsignal (>= 0.3.5)
LinkingTo: Rcpp, RcppEigen
Suggests: fftwtools, bit64, grDevices, microbenchmark,
        freesurferformats, testthat, vctrs
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-31 00:30:12 UTC; dipterix
Author: Zhengjia Wang [aut, cre],
  John Magnotti [aut],
  Michael Beauchamp [aut],
  Trustees of the University of Pennsylvania [cph] (All files in this
    package unless explicitly stated in the file or listed in the
    'Copyright' section below.),
  Karim Rahim [cph, ctb] (Contributed to src/ffts.h and stc/ffts.cpp),
  Thomas Possidente [cph, ctb] (Contributed to R/multitaper.R),
  Michael Prerau [cph, ctb] (Contributed to R/multitaper.R),
  Marcus Geelnard [ctb, cph] (TinyThread library,
    tinythreadpp.bitsnbites.eu, located at inst/include/tthread/),
  Stefan Schlager [ctb, cph] (R-vcg interface, located at
    src/vcgCommon.h),
  Visual Computing Lab, ISTI [ctb, cph] (Copyright holder of vcglib,
    located at src/vcglib/)
Maintainer: Zhengjia Wang <dipterix.wang@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-31 05:10:27 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-31 08:25:58 UTC; unix
Archs: ravetools.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("==",Matrix4)
S3method("==",Quaternion)
S3method("==",Vector3)
S3method("[",Matrix4)
S3method("[",Quaternion)
S3method("[",Vector3)
S3method(as.matrix,Matrix4)
S3method(as.matrix,Vector3)
S3method(baseline_array,array)
S3method(collapse,array)
S3method(dim,Matrix4)
S3method(dim,Vector3)
S3method(format,"ravetools-design_filter")
S3method(format,"ravetools-freqz2")
S3method(format,"ravetools-printable")
S3method(format,"ravetools-test_filter")
S3method(format,"summary.ravetools-freqz2")
S3method(plot,"ravetools-freqz2")
S3method(plot,"ravetools-pwelch")
S3method(plot,"ravetools-wavelet-kernels")
S3method(plot,ravetools_crp)
S3method(plot,ravetools_curve)
S3method(print,"ravetools-multitaper-config")
S3method(print,"ravetools-printable")
S3method(print,"ravetools-pwelch")
S3method(print,"ravetools-wavelet-kernels")
S3method(print,ravetools_curve)
S3method(pwelch,default)
S3method(pwelch,matrix)
S3method(shift_array,array)
S3method(summary,"ravetools-freqz2")
export(as_matrix4)
export(as_quaternion)
export(as_vector3)
export(band_pass1)
export(band_pass2)
export(baseline_array)
export(blackman)
export(blackmanharris)
export(blackmannuttall)
export(bohmanwin)
export(buttap)
export(butter)
export(butter_max_order)
export(buttord)
export(carla)
export(catmull_rom_3d)
export(cheb1ap)
export(cheb1ord)
export(cheb2ap)
export(cheb2ord)
export(cheby1)
export(cheby2)
export(check_filter)
export(collapse)
export(convolve_image)
export(convolve_signal)
export(convolve_volume)
export(crp)
export(decimate)
export(design_filter)
export(design_filter_fir)
export(design_filter_iir)
export(detect_threads)
export(detrend)
export(diagnose_channel)
export(diagnose_filter)
export(dijkstras_surface_distance)
export(ellip)
export(ellipap)
export(ellipord)
export(ensure_mesh3d)
export(fast_cov)
export(fast_median)
export(fast_mvmedian)
export(fast_mvquantile)
export(fast_quantile)
export(fftw_c2c)
export(fftw_c2c_2d)
export(fftw_c2c_3d)
export(fftw_c2r)
export(fftw_r2c)
export(fftw_r2c_2d)
export(fftw_r2c_3d)
export(fill_surface)
export(filter_signal)
export(filtfilt)
export(find_peaks)
export(fir1)
export(firls)
export(flattopwin)
export(freqz2)
export(gammatone_fast)
export(grow_volume)
export(hamming)
export(hanning)
export(hilbert)
export(internal_rave_function)
export(is_not_cran)
export(matlab_palette)
export(mesh_from_volume)
export(morlet_wavelet)
export(multitaper)
export(multitaper_config)
export(mv_pwelch)
export(mvfftw_c2c)
export(mvfftw_c2r)
export(mvfftw_r2c)
export(naive_nmf)
export(new_matrix4)
export(new_quaternion)
export(new_vector3)
export(notch_filter)
export(plane_geometry)
export(plot_mesh_dotcloud)
export(plot_mesh_polygon)
export(plot_signals)
export(project_plane)
export(pwelch)
export(ravetools_threads)
export(raw_to_float)
export(raw_to_int16)
export(raw_to_int32)
export(raw_to_int64)
export(raw_to_int8)
export(raw_to_string)
export(raw_to_uint16)
export(raw_to_uint32)
export(raw_to_uint8)
export(rcond_filter_ar)
export(register_volume)
export(resample)
export(resample_3d_volume)
export(rgl_call)
export(rgl_plot_normals)
export(rgl_view)
export(shift_array)
export(stimpulse_align)
export(stimpulse_extract)
export(stimpulse_find)
export(stimpulse_interpolate)
export(surface_path)
export(vcg_isosurface)
export(vcg_kdtree_nearest)
export(vcg_mesh_volume)
export(vcg_raycaster)
export(vcg_smooth_explicit)
export(vcg_smooth_implicit)
export(vcg_sphere)
export(vcg_subdivision)
export(vcg_subset_vertex)
export(vcg_uniform_remesh)
export(vcg_update_normals)
export(wavelet_cycles_suggest)
export(wavelet_kernels)
importFrom(R6,R6Class)
importFrom(R6,is.R6)
importFrom(Rcpp,sourceCpp)
importFrom(graphics,image)
importFrom(gsignal,buttap)
importFrom(gsignal,butter)
importFrom(gsignal,buttord)
importFrom(gsignal,cheb1ap)
importFrom(gsignal,cheb1ord)
importFrom(gsignal,cheb2ap)
importFrom(gsignal,cheb2ord)
importFrom(gsignal,cheby1)
importFrom(gsignal,cheby2)
importFrom(gsignal,ellip)
importFrom(gsignal,ellipap)
importFrom(gsignal,ellipord)
importFrom(gsignal,hilbert)
importFrom(gsignal,resample)
importFrom(stats,approx)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,sd)
useDynLib(ravetools, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# ravetools

<!-- badges: start -->

[![R-CMD-check](https://github.com/dipterix/ravetools/workflows/R-CMD-check/badge.svg)](https://github.com/dipterix/ravetools/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/ravetools)](https://CRAN.R-project.org/package=ravetools)
[![r-universe](https://dipterix.r-universe.dev/badges/ravetools)](https://dipterix.r-universe.dev/)
<!-- badges: end -->

The goal of `ravetools` is to provide memory-efficient signal & image
processing toolbox for `intracranial Electroencephalography`.
Highlighted features include:

- [`Notch filter`](#) (remove electrical line frequencies)
- [`Welch Periodogram`](#) (averaged power over frequencies)
- [`Wavelet`](#) (frequency-time decomposition)
- 2D, 3D image convolution via `FFT`
- `CT/MRI` to `MRI` image alignment

## Installation

The package is available on `CRAN`. To install the compiled version,
simply run:

``` r
install.packages("ravetools")
```

Installing the package from source requires installation of proper
compilers and some `C` libraries; see [this
document](https://github.com/dipterix/ravetools/blob/master/installation.md)
for details.

## `iEEG` preprocess pipeline

This is a basic example which shows you how to preprocess an `iEEG`
signal. The goal here is to:

- Plot diagnostic graphs to inspect channels
- Apply Notch filters to remove electrical line noise
- Frequency-time decomposition and show the power densities

<small>\* Channel referencing is not included</small>

### 1. Generate toy examples:

``` r
library(ravetools)

# Generate 20 second data at 2000 Hz
time <- seq(0, 20, by = 1 / 2000)
signal <- sin( 120 * pi * time) + 
  sin(time * 20*pi) + 
  exp(-time^2) * 
  cos(time * 10*pi) +
  rnorm(length(time))

diagnose_channel(signal, srate = 2000)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/toy-data-1.png?raw=true" width="100%">

### 2. Apply `Notch` filters and inspect `Periodograms`

``` r
## ------- Notch filter --------
signal2 <- notch_filter(signal, sample_rate = 2000)

diagnose_channel(signal, signal2, srate = 2000,
                 name = c("Raw", "Filtered"))
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/notch-filter-1.png?raw=true" width="100%">

### 3. Frequency-time decomposition

Current version of `ravetools` provides two approaches: `Wavelet` and
`Multi-taper`. `Wavelet` uses the
[`Morlet wavelet`](https://en.wikipedia.org/wiki/Morlet_wavelet) and
obtains both amplitude and phase data, while `Multi-taper` does not
generate phase data. However, the amplitude obtained from `Multi-taper`
is smoother than `Wavelet`.

#### Using `Wavelet`:

``` r
## ---------- Wavelet -----------
coef <- morlet_wavelet(
  signal2, freqs = seq(1, 100, by = 1), 
  srate = 2000, wave_num = c(2, 15))
amplitude <- 20 * log10(Mod(coef[]))

# For each frequency, decimate to 100 Hz
downsample_amp <- apply(amplitude, 2, decimate, q = 20)
downsample_time <- decimate(time, q = 20)

par(mfrow = c(1,1))
image(
  z = downsample_amp,
  x = downsample_time,
  y = seq(1, 100, by = 1),
  xlab = "Time (s)",
  ylab = "Frequency (Hz)",
  main = "Amplitude (dB)",
  sub = "Wavelet at 2000 Hz, then down-sampled to 100 Hz", 
  col = matlab_palette()
)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/wavelet-1.png?raw=true" width="100%">

#### `Multi-taper`

Alternatively you can use `Multi-tapers` to obtain amplitude data. The
algorithm is modified from source code
[here](https://github.com/preraulab/multitaper_toolbox). Please credit
them as well if you adopt this approach.

``` r
## ---------- Multitaper -----------
res <- multitaper(
  data = signal2,
  fs = 2000,
  frequency_range = c(1, 100),
  time_bandwidth = 1.5,
  window_params = c(2, 0.01),
  nfft = 100
)

par(mfrow = c(1,1))
image(
  x = res$time,
  y = res$frequency,
  z = 10 * log10(res$spec),
  xlab = "Time (s)",
  ylab = 'Frequency (Hz)',
  col = matlab_palette(),
  main = "Amplitude (dB)"
)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/multitaper-1.png?raw=true" width="100%">

#### Image alignment

`ravetools` provides imaging co-registration via `NiftyReg`
(`doi.org/10.1117/1.JMI.1.2.024003`). You can align `CT` to `MRI`, or
`MRI` (T2) to `MRI` (T1). The method can be body `rigid`, `affine`, or
`non-linear`.

``` r
source <- system.file("extdata", "epi_t2.nii.gz", package="RNiftyReg")
target <- system.file("extdata", "flash_t1.nii.gz", package="RNiftyReg")
aligned <- register_volume(source, target, verbose = FALSE)

source_img <- aligned$source[[1]]
target_img <- aligned$target
aligned_img <- aligned$image

par(mfrow = c(2, 2), mar = c(0.1, 0.1, 3.1, 0.1))

pal <- grDevices::grey.colors(256, alpha = 1)
image(source_img[,,30], asp = 1, axes = FALSE,
      col = pal, main = "Source image")
image(target_img[,,64], asp = 1, axes = FALSE,
      col = pal, main = "Target image")
image(aligned_img[,,64], asp = 1, axes = FALSE,
      col = pal, main = "Aligned image")

# bucket fill and calculate differences
aligned_img[is.nan(aligned_img) | aligned_img <= 1] <- 1
target_img[is.nan(target_img) | aligned_img <= 1] <- 1
diff <- abs(aligned_img / target_img - 1)
image(diff[,,64], asp = 1, axes = FALSE,
      col = pal, main = "Percentage Difference")
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/coregistration-1.png?raw=true" width="80%">

## References

#### To cite ravetools in publications use, please cite the `RAVE` paper from `Beauchamp's lab`

    Magnotti, JF, and Wang, Z, and Beauchamp, MS. RAVE: comprehensive
      open-source software for reproducible analysis and visualization of
      intracranial EEG data. NeuroImage, 223, p.117341.

The `multitaper` function (MIT License) uses the script derived from
`Prerau's lab`. The `TinyParallel` script is derived from `RcppParallel`
package (GPL License) with `TBB` features removed (only use
`tinythreads`). The `register_volume` function uses `NiftyReg` (BSD
License) developed by `CMIC` at University College London, UK (its R
implementation is released under GPL license).

    [1] Magnotti, JF, and Wang, Z, and Beauchamp, MS. RAVE: comprehensive
        open-source software for reproducible analysis and visualization of
        intracranial EEG data. NeuroImage, 223, p.117341.
    [2] Prerau, Michael J, and Brown, Ritchie E, and Bianchi, Matt T, and
        Ellenbogen, Jeffrey M, and Purdon, Patrick L. Sleep Neurophysiological
        Dynamics Through the Lens of Multitaper Spectral Analysis. Physiology,
        December 7, 2016, 60-92.
    [3] Modat, M., Cash, D.M., Daga, P., Winston, G.P., Duncan, J.S. and 
        Ourselin, S., 2014. Global image registration using a symmetric 
        block-matching approach. Journal of medical imaging, 1(2), pp.024003-024003.
    [4] JJ Allaire, Romain Francois, Kevin Ushey, Gregory Vandenbrouck, Marcus
        Geelnard and Intel (2022). RcppParallel: Parallel Programming Tools for
        'Rcpp'. R package version 5.1.5.
        https://CRAN.R-project.org/package=RcppParallel

<!-- README.md is generated from README.Rmd. Please edit that file -->

# ravetools

<!-- badges: start -->

[![R-CMD-check](https://github.com/dipterix/ravetools/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/dipterix/ravetools/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/ravetools)](https://CRAN.R-project.org/package=ravetools)
[![r-universe](https://dipterix.r-universe.dev/badges/ravetools)](https://dipterix.r-universe.dev/)
<!-- badges: end -->

`ravetools` is the signal- and image-processing toolbox behind `RAVE`, a
software suite for analyzing `intracranial Electroencephalography`
(`iEEG`) data.

It is deliberately a mathematical toolbox rather than an analysis
pipeline. It supplies the fast, memory-efficient primitives - filters,
spectral decomposition, image registration, surface geometry - and
leaves the user-facing pipelines to the other `RAVE` packages built on
top of it. Every routine is a plain function operating on plain R
objects.

Highlights:

- Memory-efficient `Notch` filter, `Welch` periodogram, and `Morlet`
  wavelet that scale to hours of high-resolution recording
- Single-pulse electrical stimulation analysis: `CARLA` re-referencing,
  canonical response parameterization, and basis profile curves
- `FIR` and `IIR` filter design driven by a specification instead of a
  filter order, with stability diagnostics
- Self-contained 3D image registration (`rigid`, `affine`, and `SyN`)
  that depends on no external registration library
- Surface reconstruction, re-meshing, inflation, curvature, and
  collision detection
- 3D mesh rendering in base R, with no `rgl` dependency
- Parallel `C++` array primitives and direct access to `FFTW3`

## What is inside

| Part | Contents |
|:---|:---|
| [1. Signal processing](#1-signal-processing) | Channel diagnosis and time-frequency decomposition; brain stimulation tools; filter design |
| [2. Image processing](#2-image-processing) | 3D volume registration and resampling; surface reconstruction and rendering; low-level geometry |
| [3. Utilities and helpers](#3-utilities-and-helpers) | `FFTW3` bindings, parallel array operations, geometry classes, binary readers |

The complete function reference lives at
[`dipterix.org/ravetools`](https://dipterix.org/ravetools/reference/).

## Installation

The package is on `CRAN`. To install the compiled version, simply run:

``` r
install.packages("ravetools")
```

Development builds are on `r-universe`:

``` r
install.packages("ravetools", repos = "https://dipterix.r-universe.dev")
```

Installing from source needs a `C++` compiler, the
[`FFTW3`](https://www.fftw.org/) library, and `pkg-config` - nothing
else. See [this
document](https://github.com/dipterix/ravetools/blob/master/installation.md)
for platform-by-platform instructions.

------------------------------------------------------------------------

## 1. Signal processing

### 1.1 Data analysis and diagnosis

The starting point for a channel is usually to look at it, clean the
line noise, and decompose it in time and frequency.

Generate a toy channel and inspect it. `diagnose_channel` combines the
trace, the `Welch` periodogram, and the histogram into one figure:

``` r
library(ravetools)
set.seed(2023)

# Generate 20 second data at 2000 Hz
time <- seq(0, 20, by = 1 / 2000)
signal <- sin( 120 * pi * time) +
  sin(time * 20*pi) +
  exp(-time^2) *
  cos(time * 10*pi) +
  rnorm(length(time))

diagnose_channel(signal, srate = 2000)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/toy-data-1.png?raw=true" width="100%">

`notch_filter` removes the electrical line frequency and its harmonics.
Passing two signals to `diagnose_channel` compares them side by side:

``` r
## ------- Notch filter --------
signal2 <- notch_filter(signal, sample_rate = 2000)

diagnose_channel(signal, signal2, srate = 2000,
                 name = c("Raw", "Filtered"))
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/notch-filter-1.png?raw=true" width="100%">

There are two ways to obtain a time-frequency decomposition.
`morlet_wavelet` uses the
[`Morlet wavelet`](https://en.wikipedia.org/wiki/Morlet_wavelet) and
returns both amplitude and phase; `multitaper` returns no phase, but its
amplitude is smoother.

Using `morlet_wavelet`:

``` r
## ---------- Wavelet -----------
coef <- morlet_wavelet(
  signal2, freqs = seq(1, 100, by = 1),
  srate = 2000, wave_num = c(2, 15))
amplitude <- 20 * log10(Mod(coef[]))

# For each frequency, decimate to 100 Hz
downsample_amp <- apply(amplitude, 2, decimate, q = 20)
downsample_time <- decimate(time, q = 20)

par(mfrow = c(1,1))
image(
  z = downsample_amp,
  x = downsample_time,
  y = seq(1, 100, by = 1),
  xlab = "Time (s)",
  ylab = "Frequency (Hz)",
  main = "Amplitude (dB)",
  sub = "Wavelet at 2000 Hz, then down-sampled to 100 Hz",
  col = matlab_palette()
)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/wavelet-1.png?raw=true" width="100%">

For very long recordings, `morlet_wavelet(segment_length = ...)`
processes the signal in overlapping segments, cutting peak memory and
`FFT` cost without changing the result on the signal interior.

Using `multitaper`. The algorithm is modified from source code
[here](https://github.com/preraulab/multitaper_toolbox); please credit
them as well if you adopt this approach.

``` r
## ---------- Multitaper -----------
res <- multitaper(
  data = signal2,
  fs = 2000,
  frequency_range = c(1, 100),
  time_bandwidth = 5,
  window_params = c(2, 0.05)
)

par(mfrow = c(1,1))
image(
  x = res$time,
  y = res$frequency,
  z = 10 * log10(res$spec),
  xlab = "Time (s)",
  ylab = 'Frequency (Hz)',
  col = matlab_palette(),
  main = "Amplitude (dB)"
)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/multitaper-1.png?raw=true" width="100%">

Also in this group: `pwelch` and `mv_pwelch` for power spectral density,
`plot_signals` for drawing many traces on one canvas, `baseline_array`
for baseline correction across array margins (seven methods, including
decibel and z-score), plus `decimate`, `detrend`, `find_peaks`,
`gammatone_fast`, `wavelet_kernels`, and `wavelet_cycles_suggest`.

### 1.2 Brain stimulation tools

Single-pulse electrical stimulation produces cortico-cortical evoked
potentials (`CCEPs`). `ravetools` implements three published methods for
analyzing them. Each is a faithful re-implementation, so please cite the
original work if you use it.

**`carla`** selects the subset of channels that makes the cleanest
common average reference, by choosing the subset size with the least
anti-correlation against the remaining channels. From Huang and
colleagues (2024),
[`doi:10.1016/j.jneumeth.2024.110153`](https://doi.org/10.1016/j.jneumeth.2024.110153);
the `virtual_reference = TRUE` variant follows Huang and colleagues
(2025),
[`doi:10.1016/j.jneumeth.2025.110461`](https://doi.org/10.1016/j.jneumeth.2025.110461).

**`crp`** parameterizes single-trial evoked responses: it estimates the
response duration, extracts the canonical response shape, and reports
per-trial weights, signal-to-noise, and explained variance. From Miller
and colleagues (2023),
[`doi:10.1371/journal.pcbi.1011105`](https://doi.org/10.1371/journal.pcbi.1011105).

**`bpc`** identifies a small set of basis profile curves shared across
stimulation sites, then assigns each site to the curve that best
explains its trials. From Miller and colleagues (2021),
[`doi:10.1371/journal.pcbi.1008710`](https://doi.org/10.1371/journal.pcbi.1008710).

A basic run through the first two:

``` r
set.seed(42)

# A toy CCEP epoch: 16 channels x 30 trials, only channels 1-4 respond
sample_rate <- 1000
time <- seq(-0.1, 0.4 - 1 / sample_rate, by = 1 / sample_rate)
nchan <- 16
ntrial <- 30
evoked <- ifelse(time >= 0, 80 * exp(-time / 0.05) * sin(2 * pi * 12 * time), 0)

x <- array(rnorm(length(time) * ntrial * nchan, sd = 5),
           dim = c(length(time), ntrial, nchan))
for (ch in 1:4) {
  for (k in seq_len(ntrial)) {
    x[, k, ch] <- x[, k, ch] + evoked * runif(1, 0.6, 1.4)
  }
}

# 1. Which channels make the cleanest common average reference?
#    `x` is time x trials x channels, cropped to the responsive window.
fit <- carla(x[time > 0 & time <= 0.3, , , drop = FALSE])
fit$n_optimum   #> 12
fit$channels    #> 5 6 7 8 9 10 11 12 13 14 15 16   (channels 1-4 excluded)

# 2. Re-reference against that subset
reference <- apply(x[, , fit$channels, drop = FALSE], c(1, 2), mean)
x_ref <- sweep(x, c(1, 2), reference, "-")

# 3. Parameterize the evoked response on a responsive channel
res <- crp(x_ref[, , 1], time = time, t_start = 0.015, t_end = 0.3)
res$tau_R       #> 0.119   (response duration, in seconds)

plot(res)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/ccep-1.png?raw=true" width="100%">

`carla` also accepts a `filearray` in place of an in-memory array, so
large datasets can stay on disk.

For continuous stimulation rather than single pulses, `stimpulse_find`,
`stimpulse_align`, `stimpulse_extract`, and `stimpulse_interpolate`
locate the pulse train, align the pulses, and interpolate the artifact
away. The bundled `stimulation_signal` dataset is a one-second recording
at 30 kHz to try them on.

### 1.3 Low-level filters and designs

Most filtering work starts from a specification - a pass band of 8 to 12
Hz, fully attenuated beyond 6 and 14 Hz - rather than from a filter
order. `design_filter` takes that specification and, given `data`,
designs and applies the filter in a single call:

``` r
filtered <- design_filter(
  data = signal, sample_rate = 2000,
  high_pass_freq = 8, low_pass_freq = 12
)
```

`design_filter_fir` and `design_filter_iir` return the filter itself, so
it can be inspected before being committed to. `diagnose_filter` draws
the pass band, the stop band, the phase response, and the effect on a
sample signal in one figure:

``` r
set.seed(1)
sample_rate <- 200
time <- seq(0, 10, by = 1 / sample_rate)

# A mixture of 2 Hz, 10 Hz, and 60 Hz
x <- sin(time * 4 * pi) + sin(time * 20 * pi) +
  2 * sin(time * 120 * pi) + rnorm(length(time), sd = 0.4)

# Design an alpha-band filter from the specification rather than from an order
filter <- design_filter_fir(
  sample_rate = sample_rate,
  high_pass_freq = 8, high_pass_trans_freq = 2,
  low_pass_freq = 12, low_pass_trans_freq = 2
)
filter          #> <FIR filter>   Method: kaiser   Order: 223

# Inspect the filter against the signal it will be applied to
diagnose_filter(filter$b, filter$a, fs = sample_rate, sample = x)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/filter-design-1.png?raw=true" width="100%">

`IIR` filters can be numerically unstable at high orders, so
`design_filter` checks the order it derives against `rcond_filter_ar`,
the reciprocal condition number of the autoregressive coefficients, and
`check_filter` reports the achieved magnitude at any frequency. The
filter printed above carries both.

Also in this group: `filter_signal` and `filtfilt` for applying filters
(`filtfilt` accepts a `gsignal` second-order-section object), `freqz2`
for the frequency response, `band_pass1` and `band_pass2`, the `fir1`
and `firls` designers, and the window functions `hamming`, `hanning`,
`blackman`, `blackmanharris`, `blackmannuttall`, `bohmanwin`, and
`flattopwin`.

The classical analog prototype designers - `butter`, `cheby1`, `cheby2`,
`ellip`, and their order-selection companions - are re-exported from
`gsignal` for convenience. They belong to that package, not to this one.

------------------------------------------------------------------------

## 2. Image processing

### 2.1 3D volume processing and visualization

`register_volume3d` aligns two 3D volumes: `CT` to `MRI`, `MRI` to
`MRI`, or a subject to a template. It is implemented from scratch in
`RcppEigen` and mirrors the behavior of `ANTs` `antsRegistration` - a
multi-resolution, physical-shift-scaled gradient-descent optimizer
driving a similarity metric, working entirely in anatomical `RAS` space.
It depends on no external registration library.

Each volume carries its own `vox2ras` matrix, so volumes differing in
sampling, orientation, or field of view are handled correctly. The
transform can be `rigid` (6 degrees of freedom), `affine` (12), or `syn`
(an `affine` followed by a symmetric `diffeomorphic` deformation).

The example below is self-contained: it displaces a known volume by a
known transform, then recovers it.

``` r
data("left_hippocampus_mask")

# Blur the binary mask into a smooth intensity volume, and place it in RAS
gauss3 <- function(n = 5, sigma = 1.2) {
  g <- exp(-(seq(-(n - 1) / 2, (n - 1) / 2)^2) / (2 * sigma^2))
  array(outer(outer(g, g), g), dim = c(n, n, n)) / sum(g)^3
}
target <- convolve_volume(left_hippocampus_mask, gauss3())
vox2ras <- diag(4)
vox2ras[1:3, 4] <- -dim(target) / 2

# A known displacement: 8 degrees about the z axis, plus a translation
theta <- 8 * pi / 180
known <- diag(4)
known[1:2, 1:2] <- c(cos(theta), sin(theta), -sin(theta), cos(theta))
known[1:3, 4] <- c(2.5, -3, 1.5)

# Resample the target through the inverse to obtain a misaligned source
source <- apply_transform3d(target, vox2ras, solve(known))

# Recover the displacement
aligned <- register_volume3d(
  source, target,
  source_vox2ras = vox2ras, target_vox2ras = vox2ras,
  type = "syn", metric = "cc", verbose = FALSE
)
round(aligned$transform[1:3, 4], 2)   #> 2.51 -3.04 1.49   (known: 2.5 -3 1.5)

slice <- 22
zlim <- range(target)
pal <- grDevices::grey.colors(256, start = 0, end = 1)
oldpar <- par(mfrow = c(1, 4), mar = c(0.5, 0.5, 2.5, 0.5))
image(target[, , slice], asp = 1, axes = FALSE, col = pal, zlim = zlim,
      main = "Target (fixed)")
image(source[, , slice], asp = 1, axes = FALSE, col = pal, zlim = zlim,
      main = "Source (moving)")
image(aligned$image[, , slice], asp = 1, axes = FALSE, col = pal, zlim = zlim,
      main = "Aligned")
image(abs(aligned$image - target)[, , slice], asp = 1, axes = FALSE, col = pal,
      zlim = zlim, main = "Absolute difference")
par(oldpar)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/registration-1.png?raw=true" width="100%">

The result carries the 4x4 `RAS`-to-`RAS` `transform`, the warped
`image`, and, for `syn`, the `forward_field` and `inverse_field`
deformations. `apply_transform3d` replays any of them onto a new volume.

`register_volume3d` accepts more than one modality: pass a `list` of
co-registered volumes with per-channel `weights` and `metric` to let a
`T1`, a `T2`, and an atlas jointly drive the deformable stage. Masks
(`source_mask`, `target_mask`) restrict where the metric is evaluated,
and corresponding landmarks (`source_points`, `target_points`) add a
surface term that recovers the cortical folding an intensity metric
would otherwise blur over.

Results are compatible with `ANTs`: `write_ants_transform` writes the
linear part as an `ITK` `affine` `.mat` file and `write_ants_warp`
writes the deformation as an `ANTs` 5-D warp `NIfTI`, both readable by
`antsApplyTransforms`. `save_registration` and `load_registration`
round-trip an entire result. No `HDF5` is involved anywhere along this
path.

A note on history: `register_volume` predates all of the above and wraps
`NiftyReg` (`Modat` and colleagues, 2014). It still works and is kept
for existing code, but `register_volume3d` is the recommended path going
forward.

Also in this group: `resample_3d_volume` for re-sampling onto a new
grid, `convolve_volume` and `convolve_image` for `FFT`-based 2D and 3D
convolution, `grow_volume` for dilating a mask, and `fill_surface` for
filling a volume from a surface envelope.

### 2.2 Surface processing and visualization

A volume becomes a surface through `vcg_isosurface` or
`mesh_from_volume`. From there the mesh can be repaired, smoothed,
re-meshed, inflated, and measured. Every mesh object returned by
`ravetools` carries the `ravetools_mesh3d` class, so `plot()` renders it
directly - in base R graphics, with no `rgl` dependency.

``` r
data("left_hippocampus_mask")

# Volume -> mesh, repaired and smoothed
mesh <- vcg_isosurface(left_hippocampus_mask)
mesh <- vcg_fix_defects(mesh, verbose = FALSE)
mesh <- mris_smooth(mesh, niterations = 20L)
mesh$vb[1:3, ] <- mesh$vb[1:3, ] - rowMeans(mesh$vb[1:3, ])

# Per-vertex curvature, mapped onto a diverging color ramp centered at zero
curv <- mris_curvature(mesh)
lim <- stats::quantile(abs(curv$mean), 0.95)
col <- color_ramp_continuous(
  curv$mean, clim = c(-lim, lim),
  cmap = c("#2c5f8a", "#f2f2f2", "#a33a3a")
)

# Inflate the surface while preserving total surface area
inflated <- mris_inflate(mesh, n_averages = 4L, niterations = 8L,
                         scale_brain = FALSE, verbose = FALSE)

oldpar <- par(mfrow = c(1, 2), mar = c(0.1, 0.1, 2.1, 0.1))
plot(mesh, col = col, eye = c(0, 100, 30), up = c(0, 0, 1),
     main = "Mean curvature")
plot(inflated$mesh, col = col, eye = c(0, 100, 30), up = c(0, 0, 1),
     main = "The same, inflated")
par(oldpar)
```

<img src="https://github.com/dipterix/ravetools/blob/master/adhoc/README-figures/surface-1.png?raw=true" width="100%">

`plot_mesh_polygon` draws flat-shaded `Lambert`-lit triangles and
`plot_mesh_dotcloud` draws an orthographic rim-lit dot cloud. Both
support per-vertex colors, per-mesh alpha, camera-facing clipping, and
arbitrary clipping planes; `plot()` dispatches to whichever suits the
mesh. If `rgl` is available, `rgl_view`, `rgl_call`, and
`rgl_plot_normals` drive it instead.

Two further surface operations are worth naming:
`dijkstras_surface_distance` and `surface_path` find geodesic shortest
paths across a mesh, and `vcg_detect_collision` reports exact collisions
and proximity within a tolerance between any two of a point cloud, a
chain of connected line segments (diffusion streamlines, for instance),
and a triangular mesh.

`ensure_mesh3d` coerces the common surface formats - `mesh3d`,
`ieegio_surface`, `fs.surface`, and `surf.asc` - into a canonical mesh,
so surfaces from `FreeSurfer` and friends drop straight into any of the
above.

### 2.3 High-performance low-level geometry

The `vcg_*` and `mris_*` families are the building blocks the two
sections above are assembled from. They are usable directly:

| Group | Functions |
|:---|:---|
| Construct | `vcg_isosurface`, `vcg_sphere`, `plane_geometry`, `mesh_from_volume` |
| Repair and re-mesh | `vcg_fix_defects`, `vcg_count_edge_defects`, `vcg_uniform_remesh`, `vcg_subdivision`, `vcg_subdivide_max_edge_length`, `mris_remesh` |
| Smooth and deform | `vcg_smooth_explicit`, `vcg_smooth_implicit`, `mris_smooth`, `mris_inflate`, `mris_sphere` |
| Measure | `vcg_mesh_volume`, `vcg_average_edge_length`, `vcg_max_edge_length`, `mris_curvature` |
| Query | `vcg_kdtree_nearest`, `vcg_raycaster`, `vcg_detect_collision` |
| Subset and extract | `vcg_subset_vertex`, `vcg_mesh_patch`, `mris_make_surfaces` |

The `vcg_*` functions are built on
[`vcglib`](https://github.com/cnr-isti-vclab/vcglib) from the Visual
Computing Lab, `ISTI`. `mris_remesh` implements the isotropic re-meshing
algorithm of `Botsch` and `Kobbelt` (2003), and the `mris_inflate` and
`mris_sphere` pair follows the cortical inflation and spherical mapping
of `Fischl` and colleagues (1999).

------------------------------------------------------------------------

## 3. Utilities and helpers

Below the two domains sit the primitives everything else is built from.
They are exported because they are useful on their own.

| Group | Functions |
|:---|:---|
| `FFTW3` | `fftw_r2c`, `fftw_c2r`, `fftw_c2c` and their `2D`/`3D` variants; the batched `mvfftw_r2c`, `mvfftw_c2r`, `mvfftw_c2c`; `convolve_signal`, `convolve_image`, `convolve_volume` |
| Parallel array operations | `collapse`, `shift_array`, `baseline_array`, `fast_cov`, `fast_median`, `fast_quantile`, `fast_mvmedian`, `fast_mvquantile` |
| Threading | `ravetools_threads`, `detect_threads` |
| Geometry and math | `new_vector3`, `new_matrix4`, `new_quaternion` (in-place classes), `apply_transform3d`, `catmull_rom_3d`, `project_plane`, `naive_nmf` |
| Binary and color | `raw_to_int8` through `raw_to_int64`, `raw_to_float`, `raw_to_string`, `matlab_palette`, `color_ramp_continuous` |

The heavy loops are `C++` parallelized with `TinyThread`.
`detect_threads` reports how many threads are available and
`ravetools_threads` sets how many are used:

``` r
detect_threads()
ravetools_threads(n_threads = 4)
```

Where a computation would not fit in memory, the entry points accept
on-disk `filearray` inputs instead - `carla` is one - and
`morlet_wavelet(segment_length = ...)` streams long recordings segment
by segment.

The `fftw_*` wrappers are thin bindings over `FFTW3` for callers who
need maximum throughput and are willing to manage the layout themselves.
The `C++` headers are exposed in `inst/include`, so other packages can
link against them directly.

------------------------------------------------------------------------

## Citation

`ravetools` is part of `RAVE`. To cite it in publications, please cite
the `RAVE` paper from `Beauchamp's lab`:

    Magnotti, JF, and Wang, Z, and Beauchamp, MS. RAVE: comprehensive
      open-source software for reproducible analysis and visualization of
      intracranial EEG data. NeuroImage, 223, p.117341.

Several functions re-implement published methods. If you use one of
them, please cite the original work as well; `citation("ravetools")`
lists them all.

## References

    [1] Magnotti, JF, and Wang, Z, and Beauchamp, MS (2020). RAVE: comprehensive
        open-source software for reproducible analysis and visualization of
        intracranial EEG data. NeuroImage, 223, p.117341.
    [2] Prerau, MJ, and Brown, RE, and Bianchi, MT, and Ellenbogen, JM, and
        Purdon, PL (2016). Sleep Neurophysiological Dynamics Through the Lens of
        Multitaper Spectral Analysis. Physiology, December 7, 2016, 60-92.
    [3] Avants, BB, and Epstein, CL, and Grossman, M, and Gee, JC (2008).
        Symmetric diffeomorphic image registration with cross-correlation:
        evaluating automated labeling of elderly and neurodegenerative brain.
        Medical Image Analysis, 12(1), 26-41.
    [4] Fischl, B, and Sereno, MI, and Dale, AM (1999). Cortical surface-based
        analysis II: inflation, flattening, and a surface-based coordinate
        system. NeuroImage, 9(2), 195-207.
    [5] Huang, H, and Ojeda Valencia, G, and Gregg, NM, and Osman, GM, and
        Montoya, MN, and Worrell, GA, and Miller, KJ, and Hermes, D (2024).
        CARLA: Adjusted common average referencing for cortico-cortical evoked
        potential data. Journal of Neuroscience Methods, 407, 110153.
    [6] Huang, H, and Adkinson, JA, and Jensen, MA, and Hasen, M, and Danstrom,
        IA, and Bijanki, KR, and Gregg, NM, and Miller, KJ, and Sheth, SA, and
        Hermes, D, and Bartoli, E (2025). Proper reference selection and
        re-referencing to mitigate bias in single pulse electrical stimulation
        data. Journal of Neuroscience Methods, 419, 110461.
    [7] Miller, KJ, and Mueller, K-R, and Ojeda Valencia, G, and Huang, H, and
        Gregg, NM, and Worrell, GA, and Hermes, D (2023). Canonical response
        parameterization: Quantifying the structure of responses to single-pulse
        intracranial electrical brain stimulation. PLOS Computational Biology,
        19(5), e1011105.
    [8] Miller, KJ, and Mueller, K-R, and Hermes, D (2021). Basis profile curve
        identification to understand electrical stimulation effects in human
        brain networks. PLOS Computational Biology, 17(9), e1008710.
    [9] Botsch, M, and Kobbelt, L (2003). A remeshing approach to multiresolution
        modeling. Proceedings of Shape Modelling International, 49-58.
    [10] Modat, M, and Cash, DM, and Daga, P, and Winston, GP, and Duncan, JS,
        and Ourselin, S (2014). Global image registration using a symmetric
        block-matching approach. Journal of Medical Imaging, 1(2), 024003.
    [11] Allaire, JJ, and Francois, R, and Ushey, K, and Vandenbrouck, G, and
        Geelnard, M, and Intel (2022). RcppParallel: Parallel Programming Tools
        for 'Rcpp'. R package version 5.1.5.
        https://CRAN.R-project.org/package=RcppParallel

### Attributions

The `multitaper` function (`MIT` license) derives from a script by
`Prerau's lab`. The `TinyParallel` code derives from the `RcppParallel`
package (`GPL` license) with the `TBB` features removed, using only
`tinythreads`. The `vcg_*` functions build on `vcglib`, copyright the
Visual Computing Lab, `ISTI`, with the R interface layer derived from
work by `Stefan Schlager`. `register_volume` wraps `NiftyReg`, developed
by `CMIC` at University College London.

The **_Visualization and Computer Graphics Library_** (VCGlib for short) is an open source, portable, and templated library written in C++, with no external dependencies, for manipulation, processing, cleaning, and simplifying triangle meshes.

[![BuildExamples](https://github.com/cnr-isti-vclab/vcglib/actions/workflows/BuildExamples.yml/badge.svg)](https://github.com/cnr-isti-vclab/vcglib/actions/workflows/BuildExamples.yml)

The library, composed by more than 100k lines of code, is released under the GPL license, and it is the base of most of the software tools of the [Visual Computing Lab](http://vcg.isti.cnr.it) of the Italian National Research Council Institute - ISTI, like [MeshLab](http://www.meshlab.net/), [Metro](http://vcg.isti.cnr.it/vcglib/metro.html) and many others.

The VCG library is tailored to mostly manage triangular meshes: The library is fairly large and offers many state-of-the-art capabilities for processing meshes, such as:

- high quality quadric-error edge-collapse based simplfication
- efficient spatial query structures (uniform grids, hashed grids, kdtree, etc)
- advanced smoothing and fairing algorithms
- computation of curvature
- optimization of texture coordinates
- Hausdorff distance computation
- geodesic paths
- mesh repairing capabilities
- isosurface extraction and advancing front meshing algorithms
- Poisson Disk sampling and other tools to sample point distributions over meshes
- subdivision surfaces

## Notable Applications

A number of applications have been developed using the VCGlib:

- MeshLab: the renowed open source mesh processing software
- Metro, the tool for measuring differences between meshes
- The first high quality out-of-core mesh simplifier that was used by the Stanford Digital Michelangelo project to process their huge 3D scanned models.

## Contacts

For any info about licensing (portions of) the library please contact us:
Paolo Cignoni (p.cignoni@isti.cnr.it)
Visual Computing Lab of the Italian National Research Council - ISTI

In case of bugs please report them [here](https://github.com/cnr-isti-vclab/vcglib/issues).

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 3.2,
  dpi = 96,
  out.width = "100%"
)
library(ravetools)
oldpar <- graphics::par(no.readonly = TRUE)

## ----primitives---------------------------------------------------------------
sphere <- vcg_sphere(sub_division = 4)
sphere

plane <- plane_geometry(width = 2.6, height = 2.6, shape = c(40, 40))
plane

## ----tilt---------------------------------------------------------------------
tilt <- new_matrix4()$make_rotation_y(25 * pi / 180)$to_array()
plane$vb <- (tilt %*% rbind(plane$vb[1:3, ], 1))[1:3, ] + c(0, 0, 0.35)

## ----normals------------------------------------------------------------------
plane <- vcg_update_normals(plane, weight = "area")
dim(plane$normals)

## ----collision----------------------------------------------------------------
cut <- vcg_detect_collision(sphere, plane, radius = 0.02)
cut$summary$y$unit_type
sum(cut$hit_unit)
head(cut$representation)

## ----band---------------------------------------------------------------------
band <- plane
band$it <- plane$it[, which(cut$hit_unit), drop = FALSE]
band

## ----radius-------------------------------------------------------------------
vapply(c(0.02, 0.1, 0.3), function(r) {
  sum(vcg_detect_collision(sphere, plane, radius = r)$hit_unit)
}, 0L)

## ----streamlines--------------------------------------------------------------
streamlines <- rbind(
  cbind(seq(-3, 3, by = 0.5), 0, 0),
  c(NA, NA, NA),
  cbind(seq(-3, 3, by = 0.5), 5, 0)
)
res <- vcg_detect_collision(sphere, streamlines,
                            mode_y = "segments", radius = 0.1)
res$hit_unit
res$representation

## ----test-level---------------------------------------------------------------
vcg_detect_collision(sphere, streamlines, mode_y = "segments",
                     radius = 0.1, test_level = "whole")$collide

## ----interior-----------------------------------------------------------------
center <- rbind(c(0, 0, 0))
vcg_detect_collision(sphere, center, radius = 0.1)$collide
vcg_detect_collision(sphere, center, radius = 0.1,
                     include_interior = TRUE)$collide

## ----kdtree-------------------------------------------------------------------
kd <- vcg_kdtree_nearest(target = sphere, query = plane, k = 1)
str(kd)
range(kd$distance)

## ----kdtree-k-----------------------------------------------------------------
kd3 <- vcg_kdtree_nearest(target = sphere, query = plane, k = 3)
head(kd3$index)

## ----raycast------------------------------------------------------------------
above <- plane$vb
above[3, ] <- above[3, ] + 1.8
rays <- vcg_raycaster(sphere, ray_origin = above, ray_direction = c(0, 0, -1))
str(rays[c("has_intersection", "distance", "face_index")])
sum(rays$has_intersection)

## ----pierce-------------------------------------------------------------------
pierce <- structure(
  list(vb = rays$intersection[, rays$has_intersection, drop = FALSE]),
  class = "mesh3d"
)

## ----query-figure, fig.height = 3.6-------------------------------------------
col <- color_ramp_continuous(kd$distance[, 1],
                             cmap = c("#f2f2f2", "#7fa8c9", "#2c5f8a"))

graphics::par(mfrow = c(1, 2), mar = c(0.1, 0.1, 2.1, 0.1), cex.main = 0.95)
plot_mesh_polygon(
  list(sphere, plane, band),
  col = list("gray55", col, "#a33a3a"), alpha = c(0.5, 0.92, 1),
  eye = c(3.5, -4, 2.2), up = c(0, 0, 1), zoom = 1.15,
  shadow_color = "white", ambient_intensity = 0.55,
  main = "Distance; cut faces in red"
)
plot_mesh_polygon(
  list(sphere, pierce),
  col = list("gray72", "#a33a3a"), cex = 0.03,
  eye = c(3.5, -4, 2.2), up = c(0, 0, 1), zoom = 0.68,
  shadow_color = "white", ambient_intensity = 0.55,
  main = "Ray entry points"
)

## ----isosurface---------------------------------------------------------------
data("left_hippocampus_mask", package = "ravetools")
dim(left_hippocampus_mask)

raw_mesh <- vcg_isosurface(left_hippocampus_mask)
raw_mesh

## ----mesh-from-volume---------------------------------------------------------
smoothed <- mesh_from_volume(
  left_hippocampus_mask, output_format = "rgl", threshold = 0.5,
  remesh = TRUE, remesh_voxel_size = 1, smooth = TRUE, verbose = FALSE
)
smoothed

## ----defects------------------------------------------------------------------
vcg_count_edge_defects(raw_mesh)

## ----measure------------------------------------------------------------------
vcg_mesh_volume(raw_mesh)
vcg_average_edge_length(raw_mesh)
vcg_max_edge_length(raw_mesh)

## ----curvature-fields---------------------------------------------------------
curv <- mris_curvature(raw_mesh)
str(curv)

## ----fix----------------------------------------------------------------------
mesh <- vcg_fix_defects(raw_mesh, verbose = FALSE)
info <- attr(mesh, "info")
info[c("boundary_edges_before", "boundary_edges_after",
       "holes_filled", "is_closed_manifold")]

# the same call that warned above, now on a closed surface
vcg_mesh_volume(mesh)

## ----center-------------------------------------------------------------------
mesh$vb[1:3, ] <- mesh$vb[1:3, ] - rowMeans(mesh$vb[1:3, ])

## ----remesh-------------------------------------------------------------------
uniform  <- vcg_uniform_remesh(mesh, voxel_size = 1, verbose = FALSE)
split    <- vcg_subdivision(mesh, method = "edge")
capped   <- vcg_subdivide_max_edge_length(mesh, max_edge_len = 0.8)
isotropic <- mris_remesh(mesh, target_edge_length = 1.5, verbose = FALSE)

data.frame(
  method = c("input", "vcg_uniform_remesh", "vcg_subdivision",
             "vcg_subdivide_max_edge_length", "mris_remesh"),
  vertices = c(ncol(mesh$vb), ncol(uniform$vb), ncol(split$vb),
               ncol(capped$vb), ncol(isotropic$vb)),
  avg_edge = round(vapply(list(mesh, uniform, split, capped, isotropic),
                          vcg_average_edge_length, 0), 3),
  max_edge = round(vapply(list(mesh, uniform, split, capped, isotropic),
                          vcg_max_edge_length, 0), 3)
)

## ----smooth-------------------------------------------------------------------
taubin <- vcg_smooth_explicit(mesh, type = "taubin", iteration = 10)
implicit <- vcg_smooth_implicit(mesh, lambda = 0.2, degree = 2)
fs_style <- mris_smooth(mesh, niterations = 20L)

vapply(list(mesh, taubin, implicit, fs_style), vcg_mesh_volume, 0)

## ----inflate------------------------------------------------------------------
inflated <- mris_inflate(fs_style, n_averages = 4L, niterations = 8L,
                         scale_brain = FALSE, verbose = FALSE)
names(inflated)
range(inflated$sulc)

## ----sphere-map---------------------------------------------------------------
spherical <- mris_sphere(fs_style, target_radius = 100, verbose = FALSE)
radius <- sqrt(colSums(spherical$vb[1:3, ]^2))
c(min = min(radius), max = max(radius), cv = stats::sd(radius) / mean(radius))

## ----inflate-figure, fig.height = 3.2-----------------------------------------
curv <- mris_curvature(fs_style)
lim <- stats::quantile(abs(curv$mean), 0.95)
curv_col <- color_ramp_continuous(
  curv$mean, clim = c(-lim, lim),
  cmap = c("#2c5f8a", "#f2f2f2", "#a33a3a")
)

graphics::par(mfrow = c(1, 2), mar = c(0.1, 0.1, 2.1, 0.1))
plot_mesh_polygon(fs_style, col = curv_col, eye = c(0, 100, 30),
                  up = c(0, 0, 1), zoom = 1.1, main = "Mean curvature")
plot_mesh_polygon(inflated$mesh, col = curv_col, eye = c(0, 100, 30),
                  up = c(0, 0, 1), zoom = 1.1, main = "The same, inflated")

## ----subset-------------------------------------------------------------------
selector <- mesh$vb[1, ] > 0
half <- vcg_subset_vertex(mesh, selector)
c(input = ncol(mesh$vb), kept = ncol(half$vb))

## ----patch--------------------------------------------------------------------
target <- vcg_uniform_remesh(vcg_sphere(), verbose = FALSE)
patches <- vcg_mesh_patch(target, waypoints = diag(1, 3))
vapply(patches, function(p) ncol(p$it), 0L)

## ----patch-figure, fig.height = 3.4-------------------------------------------
graphics::par(mar = c(0.1, 0.1, 0.1, 0.1))
plot_mesh_polygon(patches, col = list("#a33a3a", "gray70"),
                  alpha = c(1, 0.55), eye = c(10, 10, 10), zoom = 1.2,
                  shadow_color = "white", ambient_intensity = 0.55)

## ----dijkstra-----------------------------------------------------------------
dist <- dijkstras_surface_distance(
  positions = t(mesh$vb[1:3, ]),
  faces = t(mesh$it),
  start_node = 1,
  face_index_start = 1
)
path <- surface_path(dist, target_node = ncol(mesh$vb))
c(vertices_on_path = length(path$path), length = max(path$distance))

## ----clipping, fig.height = 3.4-----------------------------------------------
bbox <- apply(fs_style$vb[1:3, ], 1L, range)
set.seed(1)
candidates <- cbind(
  stats::runif(600, bbox[1, 1], bbox[2, 1]),
  stats::runif(600, bbox[1, 2], bbox[2, 2]),
  stats::runif(600, bbox[1, 3], bbox[2, 3])
)
inside <- vcg_detect_collision(fs_style, candidates,
                               include_interior = TRUE)$hit_unit %in% TRUE
probes <- structure(
  list(vb = t(candidates[inside, , drop = FALSE])),
  class = "mesh3d"
)
sum(inside)

eye <- c(0, 100, 30)
graphics::par(mfrow = c(1, 3), mar = c(0.1, 0.1, 2.1, 0.1), cex.main = 0.95)

plot_mesh_polygon(fs_style, col = "steelblue", eye = eye, up = c(0, 0, 1),
                  zoom = 1.1, main = "Whole surface")

plot_mesh_polygon(
  list(fs_style, probes),
  col = list("steelblue", "#a33a3a"), cex = 1.1,
  eye = eye, up = c(0, 0, 1), zoom = 1.1,
  clipping_plane = c(0, 1, 0, 0, -1),
  clipping_plane_enabled = c(TRUE, FALSE),
  main = "Cut open, probes exempt"
)

# `plot_mesh_dotcloud` has no `main`; add the title afterwards
plot_mesh_dotcloud(fs_style, col = "steelblue", eye = eye, up = c(0, 0, 1),
                   zoom = 1.1, cex = 0.45)
graphics::title(main = "plot_mesh_dotcloud")

## ----rgl, eval = FALSE--------------------------------------------------------
# rgl_view({
#   rgl_call("shade3d", mesh, col = "steelblue")
#   rgl_call("wire3d", mesh, col = "black")
# })

## ----cleanup, include = FALSE-------------------------------------------------
graphics::par(oldpar)

#' @title 'Morlet' wavelet transform (Discrete)
#' @name wavelet
#' @description Transform analog voltage signals with 'Morlet'
#' wavelets: complex wavelet kernels with \eqn{\pi/2} phase
#' differences.
#' @param data numerical vector such as analog voltage signals
#' @param freqs frequency in which \code{data} will be projected on
#' @param srate sample rate, number of time points per second
#' @param wave_num desired number of cycles in wavelet kernels to
#' balance the precision in time and amplitude (control the
#' smoothness); positive integers are strongly suggested
#' @param precision the precision of computation; choices are
#' \code{'float'} (default) and \code{'double'}.
#' @param trend choices are \code{'constant'}: center the signal at zero;
#' \code{'linear'}: remove the linear trend; \code{'none'} do nothing
#' @param ... further passed to \code{\link{detrend}};
#' @returns \code{wavelet_kernels} returns wavelet kernels to be
#' used for wavelet function; \code{morlet_wavelet} returns a file-based array
#' if \code{precision} is \code{'float'}, or a list of real and imaginary
#' arrays if \code{precision} is \code{'double'}
#' @param frequency_range frequency range to calculate, default is 2 to 200
#' @param cycle_range number of cycles corresponding to \code{frequency_range}.
#' For default frequency range (2 - 200), the default \code{cycle_range} is
#' 3 to 20. That is, 3 wavelet kernel cycles at 2 Hertz, and 20 cycles at 200
#' Hertz.
#' @param signature signature to calculate kernel path to save, internally used
#' @param segment_length optional positive integer; when provided, long signals
#' are processed in overlapping segments of this length (in samples) using
#' batched \code{\link{mvfftw_c2c}} convolutions instead of a single full-length
#' FFT. This dramatically reduces peak memory and FFT cost for long recordings
#' (e.g. multi-hour). Must be strictly greater than the longest wavelet kernel
#' length (the kernel at the lowest frequency); otherwise an error is raised.
#' Default is \code{NULL}, which uses the legacy single-shot path. When
#' \code{segment_length >= length(data)} the function silently falls back to
#' the legacy path. Results match the legacy path on the interior of the
#' signal up to floating-point error; the first/last
#' \code{ceiling(max_kernel_len/2)} samples may differ near the global
#' boundaries (both implementations have boundary artifacts there).
#'
#' @examples
#'
#' \donttest{
#'
#' # generate sine waves
#' time <- seq(0, 3, by = 0.01)
#' x <- sin(time * 20*pi) + exp(-time^2) * cos(time * 10*pi)
#'
#' plot(time, x, type = 'l')
#'
#' # freq from 1 - 15 Hz; wavelet using float precision
#' freq <- seq(1, 15, 0.2)
#' coef <- morlet_wavelet(x, freq, 100, c(2,3))
#'
#' # to get coefficients in complex number from 1-10 time points
#' coef[1:10, ]
#'
#' # power
#' power <- Mod(coef[])^2
#'
#' # Power peaks at 5Hz and 10Hz at early stages
#' # After 1.0 second, 5Hz component fade away
#' image(power, x = time, y = freq, ylab = "frequency")
#'
#' # wavelet using double precision
#' coef2 <- morlet_wavelet(x, freq, 100, c(2,3), precision = "double")
#' power2 <- (coef2$real[])^2 + (coef2$imag[])^2
#'
#' image(power2, x = time, y = freq, ylab = "frequency")
#'
#' # The maximum relative change of power with different precisions
#' max(abs(power/power2 - 1))
#'
#' # display kernels
#' freq <- seq(1, 15, 1)
#' kern <- wavelet_kernels(freq, 100, c(2,3))
#' print(kern)
#'
#' plot(kern)
#'
#' }
#'
NULL

#' @rdname wavelet
#' @export
wavelet_kernels <- function(freqs, srate, wave_num) {
  # calculate wavelet cycles for each frequencies
  if (length(wave_num) != length(freqs)) {
    # calculate wavelet cycles for each frequencies
    ratio <- (log(max(wave_num)) - log(min(wave_num))) / (log(max(freqs)) - log(min(freqs)))
    wavelet_cycles <- round(exp((log(freqs) - log(min(freqs))) * ratio + log(min(wave_num))))
  } else {
    wavelet_cycles <- wave_num
  }
  f_l <- length(freqs)


  # wavelet window calc - each columns of final wave is a wavelet kernel (after fft)
  # sts = wavelet_cycles / (2 * pi * freqs)
  # wavelet_wins = cbind(-3 * sts, 3 * sts)

  fft_waves <- lapply(seq_len(f_l), function(ii) {
    fq <- freqs[ii]
    cycles <- wavelet_cycles[ii]
    # standard error
    st <- cycles / (2 * pi * fq)

    # calculate window size
    wavelet_win <- seq(-3 * st, 3 * st, by = 1 / srate)

    # half of window length
    w_l_half <- (length(wavelet_win) - 1) / 2

    # wavelet 1: calc sinus in complex domain
    tmp_sine <- exp((0 + 1i) * 2 * pi * fq / srate * (-w_l_half:w_l_half))

    # Gaussian normalization part
    A <- 1 / sqrt(st * sqrt(pi))

    # wavelet 2: calc gaussian wrappers
    tmp_gaus_win <- A * exp(-wavelet_win^2 / (2 * (cycles / (2 * pi * fq))^2))

    # wave kernel
    tmp_wavelet <- tmp_sine * tmp_gaus_win

    tmp_wavelet
  })

  structure(list(
    kernels = fft_waves,
    wavelet_cycles = wavelet_cycles,
    sample_rate = srate,
    frequencies = freqs
  ), class = "ravetools-wavelet-kernels")
}

#' @export
`print.ravetools-wavelet-kernels` <- function(x, plot = FALSE, ...) {
  cat("Discrete wavelet kernels\n")
  cat("  number of kernels/frequencies:", length(x$kernels), "\n")
  cat(sprintf("  frequency range: %.2f Hz - %.2f Hz\n", min(x$frequencies), max(x$frequencies)))
  cat(sprintf("  number of cycles: %.2f - %.2f\n", min(x$wavelet_cycles), max(x$wavelet_cycles)))
  invisible(x)
}

#' @export
`plot.ravetools-wavelet-kernels` <- function(
  x, cex = 1.2, cex.lab = cex * 1.2, cex.main = cex * 1.33,
  cex.axis = cex, mai = c(0.8, 0.5, 0.4, 0.1), ...) {

  fft_waves <- x$kernels
  srate <- x$sample_rate
  freqs <- x$frequencies
  wavelet_cycles <- x$wavelet_cycles
  max_l <- as.integer(max(sapply(fft_waves, length)) + 0.1 * srate)
  s <- sapply(fft_waves, function(s) {
    l <- (max_l - length(s))
    pre <- floor(l / 2)
    post <- ceiling(l / 2)
    c(rep(NA, pre), s, rep(NA, post))
  })
  s_re <- Re(s)
  s_im <- Im(s)
  ind <- exp(seq(log(min(freqs)), log(max(freqs)), length.out = 10))
  ind <- sapply(ind, function(i) {
    which.min(abs(i - freqs))[[1]]
  })
  ind <- unique(sort(ind))
  gap <- seq_along(ind) * 1.8 * max(abs(s_re), na.rm = TRUE)
  tmp_re <- t(t(s_re[, ind]) + gap)
  tmp_im <- t(t(s_im[, ind]) + gap)
  tmp <- rbind(tmp_re)
  x_all <- seq_len(max_l) / srate
  x_re <- x_all
  x_im <- x_re + max(x_all)
  # grid::grid.newpage()
  lay <- rbind(c(1, 1), c(2, 3))

  graphics::layout(mat = lay)

  old_mai <- graphics::par("mai")
  graphics::par(mai = mai)
  on.exit({
    graphics::par(mai = old_mai)
  }, add = TRUE)

  # old_mar <- par('mar')
  # on.exit({
  #   par(mar = old_mar)
  # })
  # par(mar = c(5.1, 4.1, 4.1, 2.1) * (cex / 2 + 0.5))
  graphics::matplot(y = tmp_re, x = x_re, type = "l", col = "red",
                    xlim = c(0, max(x_im)), ylim = c(min(tmp_re, na.rm = TRUE), max(gap) + 1.5 * min(gap)),
                    lty = 1, cex.lab = cex.lab, cex.main = cex.main, xlab = "Wavelet Length (seconds)", cex.axis = cex.axis,
                    ylab = "Frequency (Hz)", main = "Wavelet Kernels (Real & Imaginary)", yaxt = "n", xaxt = "n")

  graphics::matlines(y = tmp_im, x = x_im, type = "l", col = "red", lty = 1)

  n_halftickers <- 7
  x_actual <- c(x_re, x_im)
  x_label <- c(x_all, x_all) - mean(x_all)
  xind <- seq(1, length(x_re), length.out = n_halftickers)
  xind <- c(xind, xind + length(x_re))
  xind <- as.integer(xind[-n_halftickers])
  x_label <- x_label[xind]
  x_label[n_halftickers] <- abs(x_label[n_halftickers])
  x_label <- sprintf("%.2f", x_label)
  x_label[n_halftickers] <- paste0("\u00B1", x_label[n_halftickers])
  x_text <- stats::median(x_actual)

  graphics::axis(1, at = x_actual[xind], x_label, cex.axis = cex.axis)
  graphics::axis(2, at = gap, freqs[ind], cex.axis = cex.axis, las = 1)
  graphics::abline(h = gap, col = "grey80", lty = 2)
  leading_mod <- sapply(ind, function(ii) {
    x <- s[, ii]
    cycles <- wavelet_cycles[ii]
    x <- x[!is.na(x)] #Mod(x[1]) / max(Mod(x)) * 100  #= 1.111%
    c(length(x) / srate, cycles)
  })
  graphics::text(x = x_text, y = gap, "|", cex = cex)
  graphics::text(x = x_text, y = gap, sprintf("%.3f", leading_mod[1, ]), cex = cex,  pos = 2)
  graphics::text(x = x_text, y = gap, sprintf("%.2f", leading_mod[2, ]), cex = cex,  pos = 4)
  y_mini_title <- min(gap) + max(gap)
  graphics::text(x = x_text, y = y_mini_title, "|", cex = cex.lab)
  graphics::text(x = x_text, y = y_mini_title, "Wave Length", cex = cex.lab, pos = 2)
  graphics::text(x = x_text, y = y_mini_title, "# of Cycles", cex = cex.lab, pos = 4)

  # plot freq over wavelength and wave cycles
  wave_len <- sapply(fft_waves, length) / srate
  graphics::plot(freqs, wave_len, type = "l", ylab = "Wavelet Length (seconds)",
                 xlab = "Frequency (Hz)", main = "Wavelet Length | Frequency",
                 las = 1, cex.lab = cex.lab, cex.main = cex.main, cex.axis = cex.axis, col = "grey80")
  graphics::points(freqs, wave_len, col = "red", pch = 4)

  graphics::plot(freqs, wavelet_cycles, type = "l", ylab = "Wavelet Cycle",
                 xlab = "Frequency (Hz)", main = "Wavelet Cycle | Frequency",
                 las = 1, cex.lab = cex.lab, cex.main = cex.main, cex.axis = cex.axis, col = "grey80")
  graphics::points(freqs, wavelet_cycles, col = "red", pch = 4)

  return(invisible())
}


# Build a segmentation plan to chunk a long signal into overlapping segments.
#
# The overlap on each side equals ceil(max_kernel_len / 2) so the interior
# rows of each segment (i.e. samples at least `overlap` rows from both segment
# ends) are NOT corrupted by that segment's circular-FFT wraparound and
# therefore match the legacy single-shot result up to floating-point error.
wavelet_segment_plan <- function(data_length, segment_length, max_kernel_len) {
  data_length <- as.integer(data_length)
  segment_length <- as.integer(segment_length)
  max_kernel_len <- as.integer(max_kernel_len)
  if (segment_length <= max_kernel_len) {
    stop(sprintf(
      "`segment_length` (%d) must be greater than the longest wavelet kernel length (%d). Increase `segment_length`, raise the lowest frequency, or reduce the wavelet cycle count.",
      segment_length, max_kernel_len
    ))
  }
  overlap <- as.integer(ceiling(max_kernel_len / 2))
  step <- segment_length - 2L * overlap
  if (step < 1L) {
    stop(sprintf(
      "`segment_length` (%d) is too small relative to the wavelet kernel length (%d).",
      segment_length, max_kernel_len
    ))
  }
  n_seg <- as.integer(ceiling(data_length / step))
  out_starts <- (seq_len(n_seg) - 1L) * step + 1L
  out_ends <- pmin(out_starts + step - 1L, data_length)
  in_starts <- out_starts - overlap
  in_ends <- in_starts + segment_length - 1L
  local_starts <- rep(overlap + 1L, n_seg)
  local_ends <- overlap + (out_ends - out_starts + 1L)
  list(
    n_seg = n_seg,
    overlap = overlap,
    step = step,
    segment_length = segment_length,
    data_length = data_length,
    out_starts = out_starts,
    out_ends = out_ends,
    in_starts = in_starts,
    in_ends = in_ends,
    local_starts = local_starts,
    local_ends = local_ends
  )
}

# Build a segment_length x n_seg real matrix from a 1D signal using a plan.
# Out-of-range input rows are zero-padded.
wavelet_build_segment_matrix <- function(data, plan) {
  d_l <- length(data)

  if (storage.mode(data) != "double") {
    storage.mode(data) <- "double"
  }

  segment_length <- plan$segment_length
  seg_mat <- vapply(seq_len(plan$n_seg), function(k) {
    in_s <- plan$in_starts[k]
    in_e <- plan$in_ends[k]
    src_s <- max(in_s, 1L)
    src_e <- min(in_e, d_l)
    data_seg <- data[src_s:src_e]
    if (length(data_seg) == segment_length) {
      return(data_seg)
    } else {
      re <- double(segment_length)
      if (src_e >= src_s) {
        dst_s <- as.integer(src_s - in_s + 1L)
        dst_e <- as.integer(src_e - in_s + 1L)
        re[dst_s:dst_e] <- data_seg
      }
      return(re)
    }
  }, FUN.VALUE = double(segment_length))
  seg_mat
}

# Returns a normalized integer segment_length when segmentation should run;
# otherwise NA_integer_ (legacy single-shot path).
wavelet_normalize_segment_length <- function(segment_length, data_length) {
  if (is.null(segment_length)) { return(NA_integer_) }
  if (length(segment_length) != 1L) {
    stop("`segment_length` must be a single value or NULL")
  }
  if (is.na(segment_length)) { return(NA_integer_) }
  segment_length <- suppressWarnings(as.integer(segment_length))
  if (is.na(segment_length) || segment_length <= 0L) { return(NA_integer_) }
  if (segment_length >= data_length) { return(NA_integer_) }
  segment_length
}


## Precision-specific implementations live in:
##   R/wavelet-float.R   (morlet_wavelet_float, _segmented, kernels)
##   R/wavelet-double.R  (morlet_wavelet_double, _segmented, kernels)


#' @rdname wavelet
#' @export
morlet_wavelet <- function(data, freqs, srate, wave_num, precision = c("float", "double"),
                           trend = c("constant", "linear", "none"),
                           signature = NULL, segment_length = NULL, ...) {
  precision <- match.arg(precision)
  if (precision == "float") {
    re <- morlet_wavelet_float(data = data, freqs = freqs, srate = srate,
                               wave_num = wave_num, trend = trend,
                               signature = signature,
                               segment_length = segment_length, ...)
  } else {
    re <- morlet_wavelet_double(data = data, freqs = freqs, srate = srate,
                               wave_num = wave_num, trend = trend,
                               signature = signature,
                               segment_length = segment_length, ...)
  }
  return(re)
}

#' @rdname wavelet
#' @export
wavelet_cycles_suggest <- function(
    freqs,
    frequency_range = c(2, 200),
    cycle_range = c(3, 20)
) {
  v1 <- log(cycle_range[[1]])
  v2 <- log(cycle_range[[2]])
  cycle <- (v2 - v1) / (log(frequency_range[[2]]) - log(frequency_range[[1]])) *
    (log(freqs) - log(frequency_range[[1]])) + v1
  cycle <- round(exp(cycle))
  cycle[cycle <= 0] <- 1

  data.frame(
    Frequency = freqs,
    Cycles = cycle
  )
}


# x <- rnorm(10000)
# y2 <- morlet_wavelet(x, freqs = 2:200, srate = 2000, wave_num = c(2,20), demean = TRUE)
# y1 <- rave::wavelet(x, freqs = 2:200, srate = 2000, wave_num = c(2,20), demean = TRUE)

require(testthat)

# Helpers --------------------------------------------------------------------

make_signal <- function(n, srate = 200, seed = 1) {
  set.seed(seed)
  t <- seq_len(n) / srate
  sin(2 * pi * 10 * t) +
    0.5 * cos(2 * pi * 5 * t) +
    0.25 * sin(2 * pi * 25 * t) +
    0.1 * rnorm(n)
}

# Read coefficient power from either precision return type.
as_matrix2 <- function(x, ncol_) {
  if (is.null(dim(x))) matrix(x, ncol = ncol_) else x
}
coef_power <- function(re) {
  if (inherits(re, "FileArray")) {
    z <- re[]
    z <- as_matrix2(z, dim(re)[2])
    Mod(z) ^ 2
  } else {
    r <- as_matrix2(re$real[], dim(re$real)[2])
    i <- as_matrix2(re$imag[], dim(re$imag)[2])
    r ^ 2 + i ^ 2
  }
}

coef_complex <- function(re) {
  if (inherits(re, "FileArray")) {
    z <- re[]
    as_matrix2(z, dim(re)[2])
  } else {
    r <- as_matrix2(re$real[], dim(re$real)[2])
    i <- as_matrix2(re$imag[], dim(re$imag)[2])
    r + 1i * i
  }
}

interior_compare <- function(a, b, max_kernel_len) {
  pad <- as.integer(ceiling(max_kernel_len / 2))
  rows <- (pad + 1):(nrow(a) - pad)
  list(a = a[rows, , drop = FALSE], b = b[rows, , drop = FALSE])
}

with_freshtemp <- function(expr) {
  # Force unique kernel cache + output paths so we don't pick up a stale cache.
  old <- options(ravetools.debug = FALSE)
  on.exit(options(old), add = TRUE)
  expr
}

# Plan helper ----------------------------------------------------------------

test_that("wavelet_segment_plan covers the full data range", {
  plan <- ravetools:::wavelet_segment_plan(
    data_length = 1000L, segment_length = 200L, max_kernel_len = 50L
  )
  expect_identical(plan$out_starts[1], 1L)
  expect_identical(tail(plan$out_ends, 1L), 1000L)
  # consecutive segments are contiguous
  diffs <- plan$out_starts[-1] - plan$out_ends[-plan$n_seg]
  expect_true(all(diffs == 1L))
  # local rows have correct length
  lens <- plan$local_ends - plan$local_starts + 1L
  expect_identical(lens, plan$out_ends - plan$out_starts + 1L)
})

test_that("wavelet_segment_plan errors when segment_length <= max_kernel_len", {
  expect_error(
    ravetools:::wavelet_segment_plan(1000L, 50L, 50L),
    "segment_length"
  )
  expect_error(
    ravetools:::wavelet_segment_plan(1000L, 49L, 50L),
    "segment_length"
  )
})

# Numerical equivalence: float ----------------------------------------------

test_that("morlet_wavelet float: segmented matches legacy on the interior", {
  with_freshtemp({
    srate <- 200
    n <- 6000
    x <- make_signal(n, srate)
    freqs <- c(4, 8, 16, 32)

    leg <- ravetools::morlet_wavelet(x, freqs = freqs, srate = srate,
                                     wave_num = c(3, 7))
    seg <- ravetools::morlet_wavelet(x, freqs = freqs, srate = srate,
                                     wave_num = c(3, 7),
                                     segment_length = 1024L)

    # discover max kernel length to choose the comparison interior
    kern <- ravetools:::wavelet_kernels(freqs, srate, c(3, 7))
    max_kl <- max(vapply(kern$kernels, length, integer(1L)))

    pa <- coef_power(leg)
    pb <- coef_power(seg)
    cmp <- interior_compare(pa, pb, max_kl)
    rel <- max(abs(cmp$a - cmp$b) / pmax(abs(cmp$a), 1e-12))
    expect_lt(rel, 1e-4)
  })
})

test_that("morlet_wavelet float: multiple segment_length values agree", {
  with_freshtemp({
    srate <- 200
    n <- 5000
    x <- make_signal(n, srate, seed = 7)
    freqs <- c(5, 12, 30)
    kern <- ravetools:::wavelet_kernels(freqs, srate, c(3, 8))
    max_kl <- max(vapply(kern$kernels, length, integer(1L)))

    leg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 8))
    pa <- coef_power(leg)

    for (L in c(512L, 1024L, 2048L)) {
      seg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 8),
                                       segment_length = L)
      pb <- coef_power(seg)
      cmp <- interior_compare(pa, pb, max_kl)
      rel <- max(abs(cmp$a - cmp$b) / pmax(abs(cmp$a), 1e-12))
      expect_lt(rel, 1e-4)
    }
  })
})

# Numerical equivalence: double ---------------------------------------------

test_that("morlet_wavelet double: segmented matches legacy on the interior", {
  with_freshtemp({
    srate <- 200
    n <- 5000
    x <- make_signal(n, srate, seed = 3)
    freqs <- c(4, 10, 20)
    kern <- ravetools:::wavelet_kernels(freqs, srate, c(3, 7))
    max_kl <- max(vapply(kern$kernels, length, integer(1L)))

    leg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7),
                                     precision = "double")
    seg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7),
                                     precision = "double",
                                     segment_length = 1024L)

    za <- coef_complex(leg)
    zb <- coef_complex(seg)
    cmp <- interior_compare(za, zb, max_kl)
    rel <- max(Mod(cmp$a - cmp$b) / pmax(Mod(cmp$a), 1e-12))
    expect_lt(rel, 1e-9)
  })
})

# Trend modes ---------------------------------------------------------------

test_that("morlet_wavelet: segmented matches legacy across trend modes", {
  with_freshtemp({
    srate <- 200
    n <- 4000
    x <- make_signal(n, srate, seed = 11) + seq_len(n) * 0.001  # add a slope
    freqs <- c(6, 12)
    kern <- ravetools:::wavelet_kernels(freqs, srate, c(3, 7))
    max_kl <- max(vapply(kern$kernels, length, integer(1L)))

    for (tr in c("constant", "linear", "none")) {
      leg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7), trend = tr)
      seg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7), trend = tr,
                                       segment_length = 1024L)
      pa <- coef_power(leg)
      pb <- coef_power(seg)
      cmp <- interior_compare(pa, pb, max_kl)
      rel <- max(abs(cmp$a - cmp$b) / pmax(abs(cmp$a), 1e-12))
      expect_lt(rel, 1e-4)
    }
  })
})

# Edge cases ----------------------------------------------------------------

test_that("morlet_wavelet: segment_length <= max_kernel_len errors", {
  with_freshtemp({
    srate <- 200
    x <- rnorm(2000)
    # freqs with a low component -> long kernel
    expect_error(
      ravetools::morlet_wavelet(x, freqs = c(2, 10), srate = srate,
                                wave_num = c(3, 7),
                                segment_length = 16L),
      "segment_length"
    )
  })
})

test_that("morlet_wavelet: segment_length >= length(data) silently falls back", {
  with_freshtemp({
    srate <- 200
    n <- 1000
    x <- make_signal(n, srate, seed = 2)
    freqs <- c(8, 16)
    leg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7))
    fallback <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7),
                                          segment_length = n)
    fallback2 <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7),
                                           segment_length = n + 100L)
    expect_equal(coef_power(leg), coef_power(fallback))
    expect_equal(coef_power(leg), coef_power(fallback2))
  })
})

test_that("morlet_wavelet: segment_length NULL/NA both use legacy path", {
  with_freshtemp({
    srate <- 200
    n <- 1500
    x <- make_signal(n, srate, seed = 5)
    freqs <- c(8, 16)
    leg <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7))
    null_path <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7),
                                           segment_length = NULL)
    na_path <- ravetools::morlet_wavelet(x, freqs, srate, c(3, 7),
                                         segment_length = NA)
    expect_equal(coef_power(leg), coef_power(null_path))
    expect_equal(coef_power(leg), coef_power(na_path))
  })
})

test_that("morlet_wavelet: single frequency works", {
  with_freshtemp({
    srate <- 200
    n <- 3000
    x <- make_signal(n, srate, seed = 9)
    freqs <- 10
    kern <- ravetools:::wavelet_kernels(freqs, srate, 5)
    max_kl <- length(kern$kernels[[1]])
    leg <- ravetools::morlet_wavelet(x, freqs, srate, 5)
    seg <- ravetools::morlet_wavelet(x, freqs, srate, 5,
                                     segment_length = 512L)
    pa <- coef_power(leg)
    pb <- coef_power(seg)
    cmp <- interior_compare(pa, pb, max_kl)
    rel <- max(abs(cmp$a - cmp$b) / pmax(abs(cmp$a), 1e-12))
    expect_lt(rel, 1e-4)
  })
})

test_that("morlet_wavelet: segment_length must be a single value", {
  expect_error(
    ravetools::morlet_wavelet(rnorm(1000), 10, 200, 5,
                              segment_length = c(100L, 200L)),
    "single value"
  )
})

if (system.file(package = "spelling") != "") {
  spelling <- asNamespace("spelling")
  tryCatch({
    spelling$spell_check_test(vignettes = TRUE, error = FALSE,
                              skip_on_cran = TRUE)
  }, error = function(...) {})
}

#' @title Create a two-dimensional plane in three dimensional space
#' @param width,height width and height of the plane, must not be \code{NA}
#' @param shape length of two to indicate the number of vertices along width
#' and height, default is only \code{c(2, 2)} (2 vertices each side, hence
#' one grid)
#' @returns A triangular mesh of class \code{'mesh3d'}
#' @examples
#'
#' plane <- plane_geometry(5, 10, c(12, 22))
#'
#' if(FALSE) {
#'
#'   rgl_view({
#'
#'     rgl_call("shade3d", plane, col = 3)
#'     rgl_call("wire3d", plane, col = 1)
#'
#'   })
#'
#' }
#'
#' @export
plane_geometry <- function(width = 1, height = 1, shape = c(2, 2)) {

  stopifnot2(isTRUE(is.finite(width)), msg = "plane_geometry: `width` must be a valid number")
  stopifnot2(isTRUE(is.finite(height)), msg = "plane_geometry: `height` must be a valid number")

  width_half <- width / 2
  height_half <- height / 2

  grid_size <- shape - 1

  grid_size <- floor(grid_size)
  if (length(grid_size) == 1) {
    grid_size <- c(grid_size, grid_size)
  } else {
    grid_size <- grid_size[c(1, 2)]
  }

  grid_x1 <- grid_size[[1]] + 1
  grid_y1 <- grid_size[[2]] + 1

  segment_size <- c(width, height) / grid_size

  ix <- seq(0, grid_size[[1]])
  vertices <- sapply(seq(0, grid_size[[2]]), function(iy) {
    y <- iy * segment_size[[2]] - height_half

    x <- ix * segment_size[[1]] - width_half

    rbind(x, -y, 0)
  }, simplify = TRUE, USE.NAMES = FALSE)

  vertices <- matrix(as.vector(vertices), nrow = 3L, byrow = FALSE)

  ix <- seq_len(grid_size[[1]])
  indices <- sapply(seq_len(grid_size[[2]]), function(iy) {
    a <- ix + grid_x1 * (iy - 1)
    b <- ix + grid_x1 * iy
    c <- ix + 1 + grid_x1 * iy
    d <- ix + 1 + grid_x1 * (iy - 1)
    rbind(a, b, d, b, c, d)
  })
  indices <- matrix(as.vector(indices), nrow = 3L, byrow = FALSE)

  # return rgl-style surface
  structure(list(vb = vertices, it = indices), class = "mesh3d")
}

#' @title Project plane to a surface
#' @description
#' Project a two-dimensional plane (such as \code{'ECoG'} grid) to a
#' three-dimensional surface while preserving the order
#' @param target target surface to be projected to, must be object that can
#' be converted to \code{'mesh3d'} (\code{'rgl'} surface), for example,
#' \code{'fs.surface'} (from \code{'freesurferformat'} package) or
#' \code{'ieegio_surface'} from \code{'ieegio'} package.
#' @param width,height width and height of the plane in world space (for
#' \code{'ECoG'} grid, the unit is millimeter)
#' @param shape vector of two integers: the first element is the number of
#' vertices (or electrode contacts) along \code{'width'} direction; the second
#' element is the number of vertices along \code{'height'} direction. The total
#' number of vertices of the plane will be \code{prod(shape)}. Notice
#' @param initial_positions a \code{shape[[1]] x shape[[2]] x 3} array or a
#' \code{n x 3} matrix, where \code{n} is \code{prod(shape)}, the number of
#' vertices indicating the initial vertex positions of the plane
#' @param translate_first whether to translate the plane first if the
#' plane center is far from the surface; default is \code{FALSE}; set to
#' \code{TRUE} for a warm start
#' @param diagnostic whether to plot diagnostic figures showing the morphing
#' progress.
#' @param n_iters number of iterations; default is five
#' @returns The projected vertex locations, same order as \code{initial_positions}.
#'
#' @examples
#'
#'
#' # Construct target surface
#'
#' sphere <- vcg_sphere()
#' target <- structure(
#'   class = "mesh3d",
#'   list(
#'     vb = cbind(
#'       sphere$vb[1:3, ] - c(0.8, 0, 0),
#'       sphere$vb[1:3, ] + c(0.8, 0, 0)
#'     ),
#'     it = cbind(
#'       sphere$it[1:3, ],
#'       sphere$it[1:3, ] + ncol(sphere$vb)
#'     )
#'   )
#' )
#' n_surfverts <- ncol(target$vb)
#'
#' plane <- plane_geometry(width = 3, height = 3, shape = c(30, 30))
#' plane$vb <- plane$vb[1:3, , drop = FALSE] + c(0, 0, 2)
#' n_contacts <- ncol(plane$vb)
#'
#' # First plot
#' x <- t(cbind(target$vb, plane$vb))
#' colnames(x) <- c('x', 'y', 'z')
#' graphics::pairs(
#'   x = x, asp = 1,
#'   col = c(
#'     rep("black", n_surfverts),
#'     rep("green", n_contacts)
#'   ),
#'   pch = c(
#'     rep(46, n_surfverts),
#'     rep(20, n_contacts)
#'   )
#' )
#'
#' projected <- project_plane(
#'   target = target, width = 3, height = 3, shape = c(30, 30),
#'   initial_positions = t(plane$vb),
#'   translate_first = TRUE, diagnostic = FALSE
#' )
#'
#' y <- rbind(x, projected)
#' graphics::pairs(
#'   x = y, asp = 1,
#'   col = c(
#'     rep("black", ncol(target$vb)),
#'     rep("green", n_contacts),
#'     rep("red", n_contacts)
#'   ),
#'   pch = c(
#'     rep(46, n_surfverts),
#'     rep(1, n_contacts),
#'     rep(20, n_contacts)
#'   )
#' )
#'
#'
#' @export
project_plane <- function(
    target, width, height, shape, initial_positions,
    translate_first = TRUE, diagnostic = FALSE, n_iters = 5) {

  target <- ensure_mesh3d(target)
  plane <- plane_geometry(width = width, height = height, shape = shape)

  # set initial electrode locations
  n_contacts <- ncol(plane$vb)
  if (n_contacts == 1) {
    initial_positions <- matrix(initial_positions, nrow = 1, ncol = 3)
  } else {
    if (!is.matrix(initial_positions) && is.array(initial_positions)) {
      dm <- dim(initial_positions)
      dim(initial_positions) <- c(dm[[1]] * dm[[2]], dm[[3]])
    }
    stopifnot2(ncol(initial_positions) == 3, msg = "project_plane: `initial_positions` must be a matrix with 3 columns")
    stopifnot2(nrow(initial_positions) == n_contacts, msg = "project_plane: number of `initial_positions` (row count) must equal to `shape[1] x shape[2]`")
    initial_positions <- as.matrix(initial_positions)
  }

  # keep a copy of model positions
  plane$vb[1:3, ] <- t(initial_positions)
  plane_orig <- plane
  # update normals
  plane <- vcg_update_normals(plane)

  # register utility functions
  diagnose_plot <- function() {
    if (!diagnostic) { return() }

    x <- t(cbind(plane_orig$vb[1:3, , drop = FALSE], plane$vb[1:3, , drop = FALSE]))
    colnames(x) <- c("x", "y", "z")
    col <- c(rep("green", n_contacts), rep("red", n_contacts))
    pch <- c(rep(1, n_contacts), rep(20, n_contacts))

    graphics::pairs(x, col = col, pch = pch,
                    asp = 1,
                    main = sprintf("Diagnostic plot (green=init,red=current)\ndim: x=%.1f y=%.1f, shape: %dx%d",
                                   width, height, shape[[1]], shape[[2]]))
  }

  nearest_nodes <- function(v_ind, radius) {
    dist <- sqrt(colSums((plane_orig$vb - plane_orig$vb[, v_ind])^2))
    which(dist <= radius)
  }

  # first projection - project the whole grid center to the surface
  if ( translate_first ) {

    # compute the distances from contacts to the surface
    search_results <- vcg_kdtree_nearest(target, query = plane, k = 1)
    if (mean(search_results$distance) > 1) {

      # translate is needed when all contacts are atleast 1mm away from the surface

      average_normals <- rowMeans(plane$normals[1:3, , drop = FALSE])
      # project the plane along
      raycaster <- vcg_raycaster(
        target,
        ray_origin = plane$vb[1:3, , drop = FALSE],
        ray_direction = average_normals,
        both_sides = TRUE
      )

      # Find within contacts with ray-casting intersection
      idx_has_intersection <- which(raycaster$has_intersection)

      if (length(idx_has_intersection)) {

        contact_idx <- floor((shape[[2]] - 1) / 2) * shape[[1]] + round(shape[[1]] / 2)
        if (!contact_idx %in% idx_has_intersection) {
          idx2 <- which.min(abs(raycaster$distance)[idx_has_intersection])
          contact_idx <- idx_has_intersection[idx2][[1]]
        }
        dist <- raycaster$distance[contact_idx]
        proj_dir <- raycaster$ray_direction[, contact_idx, drop = TRUE] * dist

        # move the plane along the direction
        plane$vb[1:3, ] <- plane$vb[1:3, ] + proj_dir

      }

    }

    diagnose_plot()

  }

  # Total number of iterations are n_contacts x 5
  # 1, 2. Hebbian learning (random contacts)
  # 3, 4. iterate all contacts with large radius
  # 5. refine with small radius

  index_sampler <- c(sample(n_contacts), sample(n_contacts), sample(n_contacts), sample(n_contacts))
  lambda_fun <- function(iter) {
    # 0.2, 0.1, 0.2, 0.2, 0.2
    if (iter <= n_contacts) { return(0.2) }
    if (iter <= 2 * n_contacts) { return(0.1) }
    if (iter <= 3 * n_contacts) { return(0.2) }
    if (iter <= 4 * n_contacts) { return(0.2) }
    if (iter <= 5 * n_contacts) { return(0.2) }
    return(0.1)
  }

  plane_diag <- sqrt(sum(c(width, height)^2))
  contact_diag <- sqrt(sum(c(width / (shape[[1]] + 1), height / (shape[[1]] + 1))^2)) + 1e-5
  radius_ratio <- function(iter) {
    if (iter <= min(n_contacts, 100)) { return( 0.5 ) }
    if (iter <= n_contacts) { return(0.3) }
    if (iter <= 2 * n_contacts) { return(0.2) }
    if (iter <= 3 * n_contacts) { return(0.2) }
    if (iter <= 4 * n_contacts) { return(0.3) }
    if (iter <= 5 * n_contacts) { return(0.1) }
    return(0.05)
  }
  radius_fun <- function(iter) {
    max(radius_ratio(iter) * plane_diag, contact_diag)
  }

  # iterate
  for (iter in seq_len(n_contacts * n_iters)) {

    lambda <- lambda_fun(iter)
    radius <- radius_fun(iter)

    # for every grand iteration, update the projection twice
    if (iter %% floor(n_contacts / 2) == 1) {
      search_results <- vcg_kdtree_nearest(target, query = plane, k = 1)
    }

    # project direction (to nearest location on mesh)
    proj_dir <- target$vb[1:3, search_results$index, drop = FALSE] - plane$vb[1:3, , drop = FALSE]
    proj_dist <- sqrt(colSums(proj_dir^2))
    proj_dist[proj_dist <= 0] <- 1
    proj_dir <- t(t(proj_dir) / proj_dist)

    # dot product to electrode normal - we want to project electrodes,
    # but we do not want to bend the electrode too much
    dot_prod <- abs(colSums(proj_dir * plane$normals[1:3, ]))

    if (iter <= n_contacts) {
      # These are easy ones: the contacts are close to the mesh
      # or the projection aligns with the normals
      if (stats::runif(1) > 0.85) {
        # have prob to project distanced contacts
        min_idx <- sample(
          c(which(search_results$distance >= stats::quantile(search_results$distance, 0.9))),
          size = 1
        )
      } else {
        min_idx <- sample(
          c(which(dot_prod > 0.95), which.min(search_results$distance)),
          size = 1
        )
      }
    } else {
      min_idx <- index_sampler[iter - n_contacts]
    }

    neib <- nearest_nodes(v_ind = min_idx, radius = radius)
    surface_points <- target$vb[1:3, search_results$index[neib], drop = FALSE]
    update_vec <- rowMeans(surface_points - plane$vb[1:3, neib, drop = FALSE])

    if (dot_prod[min_idx] <= 0.86) {
      # the projection direction and plane normals have at least 30-degree dis-agreement
      # use the plane normal instead
      update_vec2 <- plane$normals[1:3, min_idx] * sqrt(sum(update_vec^2))
      if (sum(update_vec2 * update_vec) <= 0) {
        update_vec2 <- -update_vec2
      }
      update_vec <- update_vec2
    }

    # update_nodes <- nearest_nodes(min_idx, radius = radius)
    update_nodes <- neib

    # update plane
    plane$vb[1:3, update_nodes] <- plane$vb[1:3, update_nodes] + update_vec * lambda

    # smooth
    plane <- vcg_update_normals(plane, weight = "area")
    if (iter %% n_contacts == 0) {
      plane <- vcg_smooth_explicit(plane, type = "taubin", iteration = 2)
      plane <- vcg_update_normals(plane, weight = "area")
    }

    if ( diagnostic && iter %% 100 == 0) {
      qdist <- stats::quantile(proj_dist, c(0.5, 1))
      dmsg <- sprintf("\r%d, current_dist=%.2f, media_dist=%.2f, max_dist=%.2f",
                      iter, proj_dist[min_idx], qdist[[1]], qdist[[2]])
      diagnose_plot()
      message(dmsg, appendLF = FALSE)
    }

  }

  # Finally, ray-caster
  # determine the hemisphere
  plane_center <- rowMeans(plane$vb)[1:3]
  target_center <- rowMeans(target$vb)[1:3]
  dir <- plane_center - target_center
  vec3 <- new_vector3(dir[1], dir[2], dir[3])$normalize()

  plane_nromals <- new_vector3(
    plane$normals[1, ],
    plane$normals[2, ],
    plane$normals[3, ]
  )
  dot <- plane_nromals$dot(vec3)
  normals <- plane$normals[1:3, , drop = FALSE] * ifelse(dot >= 0, 1, -1)

  raycaster <- vcg_raycaster(
    target, plane$vb[1:3, , drop = FALSE], ray_direction = normals, both_sides = FALSE)
  search_results <- vcg_kdtree_nearest(target, query = plane, k = 1)

  raycaster_valid <- raycaster$has_intersection &
    abs(raycaster$distance) < search_results$distance[, 1] * 2

  raycaster_invalid <- raycaster$has_intersection & !raycaster_valid
  # raycaster$distance[raycaster_invalid]
  # search_results$distance[raycaster_invalid, 1]
  # plane$normals[,1]

  if (any(raycaster_invalid)) {

    # ray-casting not working
    # change ray directions so the direction points towards
    # mean of normals and shortest point
    idx <- search_results$index[raycaster_invalid]
    plane_nromals$from_array(
      target$vb[1:3, idx, drop = FALSE] - plane$vb[1:3, raycaster_invalid, drop = FALSE])
    plane_nromals$normalize()
    plane_nromals$from_array(plane_nromals[] + normals[, raycaster_invalid])$normalize()

    raycaster2 <- vcg_raycaster(
      target, plane$vb[1:3, raycaster_invalid, drop = FALSE],
      ray_direction = plane_nromals[], both_sides = FALSE)

    raycaster2_valid <- raycaster2$has_intersection &
      abs(raycaster2$distance) < search_results$distance[raycaster_invalid, 1] * 2

    if (any(raycaster2_valid)) {
      idx <- which(raycaster_invalid)[raycaster2_valid]
      raycaster$intersection[, idx] <- raycaster2$intersection[, raycaster2_valid]
      raycaster_valid[idx] <- TRUE
    }
  }

  # plane_proj <- plane
  plane$vb[, raycaster_valid] <- raycaster$intersection[, raycaster_valid]
  plane <- vcg_update_normals(plane, weight = "area")
  plane <- vcg_smooth_explicit(plane, type = "taubin", iteration = 2)

  if (diagnostic && system.file(package = "ieegio") != "") {
    # for RAVE users only
    ieegio <- asNamespace("ieegio")
    merged <- merge(ieegio$as_ieegio_surface(plane),
                    ieegio$as_ieegio_surface(target),
                    merge_type = "geometry", verbose = FALSE)

    pal <- grDevices::colorRampPalette(
      c("#053061", "#2166ac", "#4393c3", "#92c5de", "#b4b4b4",
        "#f4a582", "#d6604d", "#b2182b", "#67001f"))(255)
    print(plot(
      merge(
        merged,
        ieegio$as_ieegio_surface(
          measurements = data.frame(
            distance_to_surface = c(
              search_results$distance[, 1], rep(0, ncol(target$vb))
            ),
            has_intersection = c(
              raycaster$has_intersection * 3 + 1, rep(0, ncol(target$vb))
            )
          )
        ),
        verbose = FALSE
      ),
      col = c("#AAAAAAAA", paste0(pal, "AA")),
      name = c("measurements", "distance_to_surface")
    ))
  }

  t(plane$vb[1:3, , drop = FALSE])

}

Matrix4 <- R6::R6Class(
  classname = "Matrix4",
  portable = TRUE,
  cloneable = FALSE,
  lock_objects = FALSE,
  private = list(
    .extern_ptr = NULL
  ),
  active = list(
    is_matrix4 = function() { TRUE },
    pointer = function() { private$.extern_ptr },
    elements = function() {
      self$to_array()
    }
  ),
  public = list(
    initialize = function() {
      private$.extern_ptr <- Matrix4__new()
    },
    format = function(...) {
      m <- self$to_array()
      c(
        "<Matrix4>",
        utils::capture.output({
          print(m)
        })
      )
    },
    clone2 = function(...) {
      Matrix4$new()$copy(self)
    },
    set = function(n11, ..., byrow = TRUE) {
      if (is.matrix(n11)) {
        data <- n11
      } else {
        data <- matrix(c(n11, ...), nrow = 4, byrow = byrow)
      }
      if (!is.numeric(data)) {
        data <- as.numeric(data)
      }
      if (length(data) != 16) {
        stop("Matrix4$set() requires 16 elements")
      }
      Matrix4__from_array(private$.extern_ptr, data)
      self
    },
    to_array = function() {
      Matrix4__to_array(private$.extern_ptr)
    },
    identity = function() {
      Matrix4__identity(private$.extern_ptr)
      self
    },
    copy = function(m) {
      stopifnot(R6::is.R6(m) && isTRUE(m$is_matrix4))
      Matrix4__copy(private$.extern_ptr, m$pointer)
      self
    },
    copy_position = function(m) {
      m <- as_matrix4(m)
      Matrix4__copy_position(private$.extern_ptr, m$pointer)
      self
    },
    extract_basis = function(x_axis = NULL, y_axis = NULL, z_axis = NULL) {
      if (is.null(x_axis)) {
        x_axis <- new_vector3()
      } else {
        x_axis <- as_vector3(x_axis)
      }
      if (is.null(y_axis)) {
        y_axis <- new_vector3()
      } else {
        y_axis <- as_vector3(y_axis)
      }
      if (is.null(z_axis)) {
        z_axis <- new_vector3()
      } else {
        z_axis <- as_vector3(z_axis)
      }
      Matrix4__extract_basis(private$.extern_ptr, x_axis$pointer, y_axis$pointer, z_axis$pointer)
      self
    },
    make_basis = function(x_axis = NULL, y_axis = NULL, z_axis = NULL) {
      if (is.null(x_axis)) {
        x_axis <- new_vector3()
      } else {
        x_axis <- as_vector3(x_axis)
      }
      if (is.null(y_axis)) {
        y_axis <- new_vector3()
      } else {
        y_axis <- as_vector3(y_axis)
      }
      if (is.null(z_axis)) {
        z_axis <- new_vector3()
      } else {
        z_axis <- as_vector3(z_axis)
      }
      Matrix4__make_basis(private$.extern_ptr, x_axis$pointer, y_axis$pointer, z_axis$pointer)
      self
    },
    extract_rotation = function(m) {
      m <- as_matrix4(m)
      Matrix4__extract_rotation(private$.extern_ptr, m$pointer)
      self
    },
    look_at = function(eye, target, up) {
      eye <- as_vector3(eye)
      target <- as_vector3(target)
      up <- as_vector3(up)
      Matrix4__look_at(private$.extern_ptr, eye$pointer, target$pointer, up$pointer)
      self
    },
    multiply_matrices = function(a, b) {
      a <- as_matrix4(a)
      b <- as_matrix4(b)
      Matrix4__multiply_matrices(private$.extern_ptr, a$pointer, b$pointer)
      self
    },
    multiply = function(m) {
      m <- as_matrix4(m)
      Matrix4__multiply_matrices(private$.extern_ptr, private$.extern_ptr, m$pointer)
      self
    },
    premultiply = function(m) {
      m <- as_matrix4(m)
      Matrix4__multiply_matrices(private$.extern_ptr, m$pointer, private$.extern_ptr)
      self
    },
    multiply_scalar = function(scalar) {
      if (!is.numeric(scalar)) {
        scalar <- as.numeric(scalar[[1]])
      } else {
        scalar <- scalar[[1]]
      }
      Matrix4__multiply_scalar(private$.extern_ptr, scalar)
      self
    },
    determinant = function() {
      Matrix4__determinant(private$.extern_ptr)
    },
    transpose = function() {
      Matrix4__transpose(private$.extern_ptr)
      self
    },
    set_position = function(v, ...) {
      if (R6::is.R6(v) && isTRUE(v$is_vector3)) {
        v <- as.numeric(v[, 1])
      } else {
        v <- as.numeric(c(v, ...))[1:3]
      }
      Matrix4__set_position(private$.extern_ptr, v[[1]], v[[2]], v[[3]])
      self
    },
    invert = function() {
      Matrix4__invert(private$.extern_ptr)
      self
    },
    scale = function(v, ...) {
      if (!R6::is.R6(v) || !isTRUE(v$is_vector3)) {
        v <- as_vector3(c(v, ...))
      }
      Matrix4__scale(private$.extern_ptr, v$pointer)
      self
    },
    get_max_scale_on_axis = function() {
      Matrix4__get_max_scale_on_axis(private$.extern_ptr)
    },
    make_translation = function(x, y = NULL, z = NULL) {
      if (R6::is.R6(x) && isTRUE(x$is_vector3)) {
        x <- as.numeric(x[, 1])
      } else {
        x <- as.numeric(c(x, y, z))[1:3]
      }
      Matrix4__make_translation(private$.extern_ptr, x[[1]], x[[2]], x[[3]])
      self
    },
    make_rotation_x = function(theta) {
      theta <- as.double(theta)[[1]]
      Matrix4__make_rotation_x(private$.extern_ptr, theta)
      self
    },
    make_rotation_y = function(theta) {
      theta <- as.double(theta)[[1]]
      Matrix4__make_rotation_y(private$.extern_ptr, theta)
      self
    },
    make_rotation_z = function(theta) {
      theta <- as.double(theta)[[1]]
      Matrix4__make_rotation_z(private$.extern_ptr, theta)
      self
    },
    make_rotation_axis = function(axis, angle) {
      axis <- as_vector3(axis)
      angle <- as.double(angle)[[1]]
      Matrix4__make_rotation_axis(private$.extern_ptr, axis$pointer, angle)
      self
    },
    make_scale = function(x, y = NULL, z = NULL) {
      if (R6::is.R6(x) && isTRUE(x$is_vector3)) {
        x <- as.numeric(x[, 1])
      } else {
        x <- as.numeric(c(x, y, z))[1:3]
      }
      Matrix4__make_scale(private$.extern_ptr, x[[1]], x[[2]], x[[3]])
      self
    },
    make_shear = function(xy, xz, yx, yz, zx, zy) {
      xy <- as.double(xy)[[1]]
      xz <- as.double(xz)[[1]]
      yx <- as.double(yx)[[1]]
      yz <- as.double(yz)[[1]]
      zx <- as.double(zx)[[1]]
      zy <- as.double(zy)[[1]]
      Matrix4__make_shear(private$.extern_ptr, xy, xz, yx, yz, zx, zy)
      self
    },
    make_perpective = function(left, right, top, bottom, near, far) {
      left <- as.double(left)[[1]]
      right <- as.double(right)[[1]]
      top <- as.double(top)[[1]]
      bottom <- as.double(bottom)[[1]]
      near <- as.double(near)[[1]]
      far <- as.double(far)[[1]]
      Matrix4__make_perspective(private$.extern_ptr, left, right, top, bottom, near, far)
      self
    },
    make_orthographic = function(left, right, top, bottom, near, far) {
      left <- as.double(left)[[1]]
      right <- as.double(right)[[1]]
      top <- as.double(top)[[1]]
      bottom <- as.double(bottom)[[1]]
      near <- as.double(near)[[1]]
      far <- as.double(far)[[1]]
      Matrix4__make_orthographic(private$.extern_ptr, left, right, top, bottom, near, far)
      self
    }
  )
)

#' Create a \code{Matrix4} instance for \code{'Affine'} transform
#' @param m a matrix or a vector to be converted to the \code{Matrix4} instance;
#' \code{m} must be one of the followings: for matrices, the dimension must be
#' \code{4x4}, \code{3x4} (the last row will be \code{0 0 0 1}), or
#' \code{3x3} (linear transform); for vectors, the length must be
#' \code{16}, \code{12} (will append \code{0 0 0 1} internally),
#' \code{3} (translation), or \code{1} (scale).
#' @returns A \code{Matrix4} instance
#' @seealso \code{\link{new_vector3}}, \code{\link{new_quaternion}}
#' @export
new_matrix4 <- function() {
  Matrix4$new()
}

#' @export
as.matrix.Matrix4 <- function(x, ...) {
  if (R6::is.R6(x) && isTRUE(x$is_matrix4)) {
    return(x$to_array())
  }
  NextMethod("as.matrix")
}

#' @export
`[.Matrix4` <- function(x, i, ..., drop = TRUE) {
  if (missing(i)) {
    x$to_array()[, ..., drop = drop]
  } else {
    x$to_array()[i, ..., drop = drop]
  }
}

#' @export
dim.Matrix4 <- function(x) {
  if (R6::is.R6(x) && isTRUE(x$is_matrix4)) {
    return(c(4L, 4L))
  }
  NextMethod("dim")
}

#' @export
`==.Matrix4` <- function(e1, e2) {
  e2 == e1$to_array()
}

#' @rdname new_matrix4
#' @export
as_matrix4 <- function(m) {
  if (R6::is.R6(m) && isTRUE(m$is_matrix4)) { return(m) }
  if (is.matrix(m)) {
    nc <- ncol(m)
    nr <- nrow(m)
    if (!nc %in% c(3, 4)) {
      stop("`as_matrix4`: when input `m` is a matrix, its column numbers must be 3 or 4")
    }
    if (!nr %in% c(3, 4)) {
      stop("`as_matrix4`: when input `m` is a matrix, its row numbers must be 3 or 4")
    }
    if ( nc == 3 && nr == 3 ) {
      m <- cbind(rbind(m, c(0, 0, 0)), c(0, 0, 0, 1))
    } else if ( nc == 3 ) {
      m <- cbind(m, c(0, 0, 0, 1))
    } else if ( nr == 3 ) {
      m <- rbind(m, c(0, 0, 0, 1))
    }
    return(new_matrix4()$set(m))
  }
  m <- as.numeric(m)
  if (length(m) == 12) {
    m <- c(m, 0, 0, 0, 1)
  } else if (length(m) == 3) {
    m <- matrix(c(
      1, 0, 0, m[[1]],
      0, 1, 0, m[[2]],
      0, 0, 1, m[[3]],
      0, 0, 0, 1
    ), nrow = 4, byrow = TRUE)
    return(new_matrix4()$set(m))
  }
  new_matrix4()$set(m)
}

# ANTs / ITK-compatible transform I/O ----------------------------------------
#
# Helpers to read and write the transforms produced by `register_volume3d` in
# the file formats used by 'ANTs' / 'ITK', so they interoperate with
# `antsApplyTransforms`, 'ITK-SNAP', 'FreeSurfer', etc.:
#
#   * the linear part  -> an 'ITK' affine `.mat` (a MATLAB level-4 binary file,
#     LPS convention), hand-written here so the package gains no dependency.
#   * the deformation  -> an 'ANTs' 5-D warp 'NIfTI' (`.nii.gz`, LPS vectors),
#     via the (suggested) 'freesurferformats' package.
#
# Conventions: 'ravetools' works in RAS while 'ITK'/'ANTs' work in LPS. The two
# are related by the symmetric conjugation  M_lps = F %*% M_ras %*% F  with
# F = diag(c(-1, -1, 1, 1)) (F is its own inverse), and for displacement vectors
# by negating the x and y components.


# Convert a 4x4 transform between \verb{RAS} and \verb{LPS}
ras_lps_conjugate <- function(m) {
  m <- as.matrix(m)
  if (nrow(m) == 3L && ncol(m) == 4L) {
    m <- rbind(m, c(0, 0, 0, 1))
  }
  if (!all(dim(m) == c(4L, 4L))) {
    stop("`ras_lps_conjugate`: input must be a 4x4 (or 3x4) matrix.")
  }
  f <- diag(c(-1, -1, 1, 1))
  f %*% m %*% f
}


# ---- MATLAB level-4 (v4) primitives, specialized to 'ITK' transforms --------
# 'ITK' / 'ANTs' write transform `.mat` files in the *level-4* MAT format (not
# level-5): there is no 128-byte file header; each variable is a 20-byte header
# (MOPT type, mrows, ncols, imagf, name length incl. null) followed by the
# null-terminated name and the column-major real data. MOPT = M*1000 + O*100 +
# P*10 + T, where M is the byte order (0 little, 1 big), P the precision (0
# double, 1 single), T the matrix type (0 full). 'ANTs' stores
# `AffineTransform_double_3_3` (12 x 1) and `fixed` (3 x 1), both `double`.

# Is `t` a plausible level-4 MOPT value (M in 0:4, O == 0, P in 0:5, T in 0:2)?
mat4_valid_mopt <- function(t) {
  if (is.na(t) ||
      t < 0L) {
    return(FALSE)
  }
  m <- t %/% 1000L
  rem <- t %% 1000L
  o <- rem %/% 100L
  p <- (rem %/% 10L) %% 10L
  tt <- t %% 10L

  m %in% 0:4 &&
    o == 0L &&
    p %in% 0:5 &&
    tt %in% 0:2
}

# Detect file byte order from the first variable's MOPT word.
mat4_endian <- function(raw) {
  t_le <- readBin(raw[1:4], "integer", n = 1L, size = 4L, endian = "little")
  if (mat4_valid_mopt(t_le)) return("little")
  t_be <- readBin(raw[1:4], "integer", n = 1L, size = 4L, endian = "big")
  if (mat4_valid_mopt(t_be)) return("big")
  "little"
}

# Serialize one real double variable in level-4 format (little-endian).
mat4_write_var <- function(name, values) {
  name_raw <- c(charToRaw(name), as.raw(0L))               # null-terminated
  header <- writeBin(
    as.integer(c(0L, length(values), 1L, 0L, length(name_raw))),
    raw(), size = 4L, endian = "little")                   # MOPT=0, m, n=1, imagf=0, namlen
  c(header, name_raw, writeBin(as.double(values), raw(), size = 8L, endian = "little"))
}

# Parse an entire level-4 buffer into a named list of numeric variables.
mat4_parse <- function(raw) {
  n <- length(raw)
  if (n < 20L)
    stop("not a valid MATLAB level-4 file (too short).")
  endian <- mat4_endian(raw)
  cur <- 1L
  vars <- list()
  while (cur + 19L <= n) {
    hdr <- readBin(raw[cur:(cur + 19L)],
                   "integer",
                   n = 5L,
                   size = 4L,
                   endian = endian)
    type <- hdr[1L]
    mrows <- hdr[2L]
    ncols <- hdr[3L]
    imagf <- hdr[4L]
    namlen <- hdr[5L]
    if (!mat4_valid_mopt(type) ||
        namlen < 1L || mrows < 0L || ncols < 0L) {
      break
    }
    # precision code
    p <- (type %/% 10L) %% 10L
    # drop trailing null
    name <- rawToChar(raw[(cur + 20L):(cur + 20L + namlen - 2L)])
    dstart <- cur + 20L + namlen
    nval <- mrows * ncols

    # double vs single
    sz <- if (p == 0L) {
      8L
    } else {
      4L
    }
    bytes <- nval * sz
    vals <- readBin(raw[dstart:(dstart + bytes - 1L)],
                    "double",
                    n = nval,
                    size = sz,
                    endian = endian)
    vars[[name]] <- vals
    # skip imaginary part if any
    cur <- dstart + bytes * (1L + (imagf == 1L))
  }
  vars
}


#' @title Read and write an \verb{'ITK'}/\verb{'ANTs'} \verb{affine} transform
#' @description
#' \code{write_ants_transform} stores a \eqn{4\times 4} \verb{RAS}-to-\verb{RAS}
#' \verb{affine} (such as the \code{transform} returned by
#' \code{\link{register_volume3d}}) as an \pkg{'ITK'} \code{.mat} file (a
#' \code{'MATLAB'} level-5 binary, \verb{LPS} convention) compatible with
#' \code{antsApplyTransforms}. \code{read_ants_transform} reads such a file back
#' into a \eqn{4\times 4} \verb{RAS} matrix. The reader folds a non-zero
#' \verb{ITK} center of rotation into the translation, so transforms written by
#' \pkg{'ANTs'} are read correctly.
#' @param transform a \eqn{4\times 4} (or \eqn{3\times 4}) \verb{RAS} \verb{affine}
#' @param file path to the \code{.mat} file
#' @returns \code{write_ants_transform} returns \code{file} invisibly;
#' \code{read_ants_transform} returns a \eqn{4\times 4} \verb{RAS} matrix.
#' @seealso \code{\link{register_volume3d}}, \code{\link{save_registration}}
#' @examples
#' tf <- tempfile(fileext = ".mat")
#' m <- diag(4); m[1:3, 4] <- c(3, -2, 1)
#' write_ants_transform(m, tf)
#' read_ants_transform(tf)
#' @export
write_ants_transform <- function(transform, file) {
  m_ras <- as.matrix(transform)
  if (nrow(m_ras) == 3L && ncol(m_ras) == 4L) m_ras <- rbind(m_ras, c(0, 0, 0, 1))
  if (!all(dim(m_ras) == c(4L, 4L))) {
    stop("`write_ants_transform`: `transform` must be a 4x4 (or 3x4) matrix.")
  }
  m_lps <- ras_lps_conjugate(m_ras)
  a <- m_lps[1:3, 1:3]
  tr <- m_lps[1:3, 4]
  # 'ITK' stores the 3x3 matrix row-major, then the translation; center = 0.
  params <- c(as.vector(t(a)), tr)
  out <- c(mat4_write_var("AffineTransform_double_3_3", params),
           mat4_write_var("fixed", c(0, 0, 0)))
  writeBin(out, file)
  invisible(file)
}

#' @rdname write_ants_transform
#' @export
read_ants_transform <- function(file) {
  raw <- readBin(file, "raw", n = file.size(file))
  vars <- mat4_parse(raw)
  anm <- grep("^AffineTransform_(double|float)_[0-9]+_[0-9]+$", names(vars), value = TRUE)
  if (!length(anm)) {
    stop("`read_ants_transform`: no 'AffineTransform_*' variable found in '", file, "'.")
  }
  p <- vars[[anm[1]]]
  if (length(p) < 12L) {
    stop("`read_ants_transform`: '", anm[1], "' has fewer than 12 parameters.")
  }
  a <- matrix(p[1:9], 3L, 3L, byrow = TRUE)
  tr <- p[10:12]
  ctr <- if (!is.null(vars[["fixed"]])) vars[["fixed"]][1:3] else c(0, 0, 0)
  # 'ITK': offset = translation + (I - A) %*% center.
  w <- tr + (diag(3) - a) %*% ctr
  m_lps <- rbind(cbind(a, as.vector(w)), c(0, 0, 0, 1))
  ras_lps_conjugate(m_lps)
}


# ---- hidden-affine codec for the warp 'NIfTI' header ------------------------
# The 4x4 RAS affine is encoded (lossily, opt-in on read) into two standard
# header text fields so a single warp `.nii.gz` can recover it as a last resort.
#   descrip  (80 chars): the RAS 3x3, column-major, 9 x "%+08.5f" comma-joined.
#   aux_file (24 chars): the RAS translation, 3 x "%+07.2f" comma-joined (23),
#                        then a 1-char direction flag "F" (forward) / "I"
#                        (inverse). 'ANTs' ignores both fields.

encode_affine_header <- function(affine, direction = c("forward", "inverse")) {
  direction <- match.arg(direction)
  flag <- if (direction == "forward") "F" else "I"
  m <- as.matrix(affine)
  rot <- sprintf("%+08.5f", as.vector(m[1:3, 1:3]))    # column-major, 9 values
  tr <- sprintf("%+07.2f", m[1:3, 4])
  if (any(nchar(rot) != 8L) || any(nchar(tr) != 7L)) {
    warning("ravetools: affine values exceed the fixed-width header encoding; ",
            "not embedding the affine in the NIfTI header (the .mat keeps full precision).")
    return(list(descrip = "", aux_file = ""))
  }
  list(descrip = paste(rot, collapse = ","),
       aux_file = paste0(paste(tr, collapse = ","), flag))
}

decode_affine_header <- function(descrip, aux_file) {
  descrip <- descrip %||% ""
  aux_file <- aux_file %||% ""
  if (!nzchar(descrip) || !nzchar(aux_file)) return(NULL)
  rot <- tryCatch(scan(text = descrip, sep = ",", quiet = TRUE), error = function(e) numeric(0))
  if (length(rot) != 9L || anyNA(rot)) return(NULL)
  last <- substr(aux_file, nchar(aux_file), nchar(aux_file))
  if (last %in% c("F", "I")) {
    flag <- last
    trtext <- substr(aux_file, 1L, nchar(aux_file) - 1L)
  } else {
    flag <- "F"
    trtext <- aux_file
  }
  tr <- tryCatch(scan(text = trtext, sep = ",", quiet = TRUE), error = function(e) numeric(0))
  if (length(tr) != 3L || anyNA(tr)) return(NULL)
  m <- diag(4)
  m[1:3, 1:3] <- matrix(rot, 3L, 3L)                   # column-major
  m[1:3, 4] <- tr
  list(transform = m, direction = if (identical(flag, "I")) "inverse" else "forward")
}

# Construct a complete NIfTI-1 header list for a 5-D warp field, using only
# exported 'freesurferformats' helpers (no ':::').
build_warp_header <- function(data_dim, vox2ras, descrip = "", aux_file = "") {
  v <- as.matrix(vox2ras)
  sp <- sqrt(colSums(v[1:3, 1:3]^2))                   # voxel spacing
  list(
    endian = "little", sizeof_hdr = 348L,
    dim = freesurferformats::nifti.datadim.to.dimfield(data_dim),
    uses_freesurfer_hack = FALSE,
    intent_p1 = 0, intent_p2 = 0, intent_p3 = 0,
    intent_code = 1007L,                               # NIFTI_INTENT_VECTOR
    datatype = 16L, bitpix = 32L,                      # float32
    slice_start = 0L,
    pix_dim = c(1, sp[1], sp[2], sp[3], 0, 0, 0, 0),
    vox_offset = 352, scl_slope = 0, scl_inter = 0,
    slice_end = 0L, slice_code = 0L,
    xyzt_units = 2L,                                   # NIFTI_UNITS_MM
    cal_max = 0, cal_min = 0, slice_duration = 0, toffset = 0,
    glmax = 0L, glmin = 0L,
    descrip = descrip, aux_file = aux_file,
    qform_code = 0L, sform_code = 1L,                  # use sform (SCANNER_ANAT)
    quatern_b = 0, quatern_c = 0, quatern_d = 0,
    qoffset_x = 0, qoffset_y = 0, qoffset_z = 0,
    srow_x = v[1, ], srow_y = v[2, ], srow_z = v[3, ],
    intent_name = "", magic = "n+1"
  )
}


#' @title Read and write an \verb{'ANTs'} deformation (warp) field
#' @description
#' \code{write_ants_warp} stores a dense displacement field (such as the
#' \code{forward_field} / \code{inverse_field} from
#' \code{\link{register_volume3d}}) as an \pkg{'ANTs'} 5-D warp \code{'NIfTI'}
#' (\code{intent_code} \code{1007}, \verb{LPS} vectors) readable by
#' \code{antsApplyTransforms}; \code{read_ants_warp} reads it back. Requires the
#' suggested \pkg{freesurferformats} package.
#'
#' The field is stored on disk with displacement vectors in \verb{LPS} (the x
#' and y components are negated relative to the \verb{RAS} field) and an
#' \code{sform} equal to \code{vox2ras}, so the file describes the fixed grid's
#' physical space exactly as \pkg{'ANTs'} expects. Optionally the \verb{RAS}
#' \verb{affine} can be hidden in the header text fields
#' (\code{descrip}/\code{aux_file}); \pkg{'ANTs'} ignores those, and recovery on
#' read is opt-in.
#' @param field a \code{(nx, ny, nz, 3)} array of \verb{RAS} displacements; the
#' \code{vox2ras} attribute is used when \code{vox2ras} is not supplied
#' @param file path to the warp \code{.nii} / \code{.nii.gz} file
#' @param vox2ras the fixed-grid \eqn{4\times 4} voxel-to-\verb{RAS} matrix
#' @param affine optional \eqn{4\times 4} \verb{RAS} \verb{affine} to embed (with
#' reduced precision) in the header; \code{NULL} (default) embeds nothing
#' @param direction \code{"forward"} or \code{"inverse"}; recorded as the
#' header direction flag when \code{affine} is embedded
#' @param recover_affine logical; if \code{TRUE}, attempt to decode a hidden
#' \verb{affine} from the header and attach it as the \code{"transform"} attribute
#' (with a \code{"direction"} attribute). Default \code{FALSE}
#' @returns \code{write_ants_warp} returns \code{file} invisibly;
#' \code{read_ants_warp} returns the \code{(nx, ny, nz, 3)} \verb{RAS} field with
#' a \code{"vox2ras"} attribute (and, if recovered, \code{"transform"} /
#' \code{"direction"} attributes).
#' @seealso \code{\link{register_volume3d}}, \code{\link{save_registration}}
#' @export
write_ants_warp <- function(field, file, vox2ras = attr(field, "vox2ras"),
                            affine = NULL, direction = c("forward", "inverse")) {
  if (!requireNamespace("freesurferformats", quietly = TRUE)) {
    stop("`write_ants_warp` requires the 'freesurferformats' package; please install it.")
  }
  direction <- match.arg(direction)
  field <- as.array(field)
  storage.mode(field) <- "double"
  d <- dim(field)
  if (length(d) != 4L || d[4] != 3L) {
    stop("`write_ants_warp`: `field` must be a 4D array with dim (nx, ny, nz, 3).")
  }
  if (is.null(vox2ras)) {
    stop("`write_ants_warp`: `vox2ras` is required (pass it, or set a 'vox2ras' attribute on `field`).")
  }
  v <- as.matrix(vox2ras)
  # RAS -> LPS: negate x and y displacement components.
  field[, , , 1] <- -field[, , , 1]
  field[, , , 2] <- -field[, , , 2]
  # 'ANTs' warps are 5-D: (nx, ny, nz, 1, 3).
  arr <- array(0.0, c(d[1], d[2], d[3], 1L, 3L))
  arr[, , , 1, ] <- field
  txt <- if (is.null(affine)) list(descrip = "", aux_file = "") else
    encode_affine_header(affine, direction)
  h <- build_warp_header(dim(arr), v, txt$descrip, txt$aux_file)
  freesurferformats::write.nifti1(file, arr, niiheader = h)
  invisible(file)
}

#' @rdname write_ants_warp
#' @export
read_ants_warp <- function(file, recover_affine = FALSE) {
  if (!requireNamespace("freesurferformats", quietly = TRUE)) {
    stop("`read_ants_warp` requires the 'freesurferformats' package; please install it.")
  }
  h <- freesurferformats::read.nifti1.header(file)
  arr <- freesurferformats::read.nifti1.data(file, drop_empty_dims = FALSE, header = h)
  storage.mode(arr) <- "double"
  d <- dim(arr)
  if (length(d) == 5L && d[4] == 1L && d[5] == 3L) {
    dim(arr) <- d[c(1L, 2L, 3L, 5L)]
  } else if (!(length(d) == 4L && d[4] == 3L)) {
    stop("`read_ants_warp`: unexpected warp dimensions (", paste(d, collapse = " x "),
         "); expected (nx, ny, nz, 1, 3).")
  }
  # LPS -> RAS: negate x and y displacement components.
  arr[, , , 1] <- -arr[, , , 1]
  arr[, , , 2] <- -arr[, , , 2]
  attr(arr, "vox2ras") <- rbind(h$srow_x, h$srow_y, h$srow_z, c(0, 0, 0, 1))
  if (isTRUE(recover_affine)) {
    dec <- decode_affine_header(h$descrip, h$aux_file)
    if (!is.null(dec)) {
      attr(arr, "transform") <- dec$transform
      attr(arr, "direction") <- dec$direction
    }
  }
  arr
}


# ---- DCF manifest helpers ---------------------------------------------------

mat_to_str <- function(m) {
  m <- as.matrix(m)
  paste(apply(m, 1L, function(r) paste(formatC(r, format = "g", digits = 12L), collapse = " ")),
        collapse = "; ")
}

str_to_mat <- function(s) {
  rows <- strsplit(s, ";", fixed = TRUE)[[1]]
  do.call(rbind, lapply(rows, function(r) as.numeric(strsplit(trimws(r), "\\s+")[[1]])))
}


#' @title Save or load a registration result in \verb{'ANTs'}-compatible files
#' @description
#' \code{save_registration} writes a \code{\link{register_volume3d}} result to
#' disk as \pkg{'ANTs'}-style files: an \verb{'ITK'} \verb{affine} \code{.mat}, the
#' forward/inverse warp \code{'NIfTI'}(s) when present, and a small
#' \verb{'DCF'} manifest (the same key-value format as \code{'DESCRIPTION'})
#' recording what each file is plus the registration parameters.
#' \code{load_registration} reads any of those back: a \code{.mat} returns a
#' \eqn{4\times 4} \verb{RAS} matrix, a warp \code{.nii}/\code{.nii.gz} returns
#' the field, and a \code{.dcf} manifest reassembles the full object.
#' @param x a \code{ravetools_register_volume3d} object (or, for the default
#' method, a \eqn{4\times 4} matrix)
#' @param path output directory, or the manifest (\code{.dcf}) path
#' @param prefix file-name prefix; defaults to \code{"registration"}
#' @param compress logical; write warp fields as \code{.nii.gz} (default) or
#' \code{.nii}
#' @param file a \code{.mat}, \code{.nii}/\code{.nii.gz}, or \code{.dcf} path
#' @param recover_affine_from_header logical; only used as a last resort when
#' loading. If \code{TRUE} and the \verb{affine} \code{.mat} is missing (or a warp
#' file is loaded directly), the \verb{affine} is recovered from the warp
#' header's hidden encoding. Off by default because that encoding is
#' reduced-precision
#' @param ... passed to methods
#' @returns \code{save_registration} returns the manifest path invisibly.
#' \code{load_registration} returns a \eqn{4\times 4} matrix, a field array
#' (with a \code{"vox2ras"} attribute), or a \code{ravetools_register_volume3d}
#' object, depending on the file type.
#' @seealso \code{\link{register_volume3d}}, \code{\link{write_ants_transform}},
#' \code{\link{write_ants_warp}}
#' @export
save_registration <- function(x, path, ...) {
  UseMethod("save_registration")
}

#' @rdname save_registration
#' @export
save_registration.ravetools_register_volume3d <- function(
    x, path, prefix = "registration", compress = TRUE, ...) {

  if (grepl("\\.dcf$", path, ignore.case = TRUE)) {
    manifest <- path
    dir <- dirname(path)
    prefix <- sub("\\.dcf$", "", basename(path), ignore.case = TRUE)
  } else {
    dir <- path
    manifest <- file.path(dir, paste0(prefix, ".dcf"))
  }
  if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)

  ext <- if (isTRUE(compress)) ".nii.gz" else ".nii"
  geom <- x$geometry

  affine_file <- paste0(prefix, "0GenericAffine.mat")
  write_ants_transform(x$transform, file.path(dir, affine_file))

  rec <- list(
    Generator = "ravetools",
    ObjectClass = "ravetools_register_volume3d",
    TransformType = x$type %||% "",
    Metric = paste(x$metric, collapse = ","),
    AffineTransform = affine_file)

  if (!is.null(x$forward_field)) {
    fwd <- paste0(prefix, "1Warp", ext)
    write_ants_warp(
      x$forward_field, file.path(dir, fwd),
      vox2ras = attr(x$forward_field, "vox2ras") %||% geom$target_vox2ras,
      affine = x$transform, direction = "forward")
    rec$ForwardField <- fwd
  }
  if (!is.null(x$inverse_field)) {
    inv <- paste0(prefix, "1InverseWarp", ext)
    write_ants_warp(
      x$inverse_field, file.path(dir, inv),
      vox2ras = attr(x$inverse_field, "vox2ras") %||% geom$target_vox2ras,
      affine = solve(x$transform), direction = "inverse")
    rec$InverseField <- inv
  }
  if (!is.null(geom)) {
    if (!is.null(geom$source_vox2ras)) rec$SourceVox2RAS <- mat_to_str(geom$source_vox2ras)
    if (!is.null(geom$target_vox2ras)) rec$TargetVox2RAS <- mat_to_str(geom$target_vox2ras)
    if (!is.null(geom$target_dim))     rec$TargetDim <- paste(geom$target_dim, collapse = " ")
  }

  write.dcf(as.data.frame(rec, stringsAsFactors = FALSE, check.names = FALSE), manifest)
  invisible(manifest)
}

#' @rdname save_registration
#' @export
save_registration.default <- function(x, path, prefix = "registration", ...) {
  if (is.matrix(x) || (is.numeric(x) && length(x) %in% c(12L, 16L))) {
    file <- if (grepl("\\.mat$", path, ignore.case = TRUE)) path else
      file.path(path, paste0(prefix, "0GenericAffine.mat"))
    if (!dir.exists(dirname(file))) dir.create(dirname(file), recursive = TRUE)
    write_ants_transform(matrix(x, 4L, 4L), file)
    return(invisible(file))
  }
  stop("`save_registration`: don't know how to save an object of class '",
       paste(class(x), collapse = "/"), "'.")
}

#' @rdname save_registration
#' @export
load_registration <- function(file, recover_affine_from_header = FALSE) {
  lf <- tolower(file)
  if (grepl("\\.mat$", lf)) {
    return(read_ants_transform(file))
  }
  if (grepl("\\.nii(\\.gz)?$", lf)) {
    return(read_ants_warp(file, recover_affine = recover_affine_from_header))
  }
  if (grepl("\\.dcf$", lf)) {
    return(load_registration_manifest(file, recover_affine_from_header))
  }
  stop("`load_registration`: unsupported file '", file,
       "'. Expected a .mat, .nii/.nii.gz, or .dcf manifest.")
}

load_registration_manifest <- function(file, recover_affine_from_header) {
  d <- read.dcf(file)
  rec <- as.list(d[1L, ])
  dir <- dirname(file)

  res <- list(
    type = rec$TransformType,
    metric = if (!is.null(rec$Metric)) strsplit(rec$Metric, ",", fixed = TRUE)[[1]] else NULL)

  af <- if (!is.null(rec$AffineTransform)) file.path(dir, rec$AffineTransform) else NULL
  if (!is.null(af) && file.exists(af)) {
    res$transform <- read_ants_transform(af)
  } else if (isTRUE(recover_affine_from_header) && !is.null(rec$ForwardField)) {
    w <- read_ants_warp(file.path(dir, rec$ForwardField), recover_affine = TRUE)
    res$transform <- attr(w, "transform")
  }

  geom <- list()
  if (!is.null(rec$SourceVox2RAS)) geom$source_vox2ras <- str_to_mat(rec$SourceVox2RAS)
  if (!is.null(rec$TargetVox2RAS)) geom$target_vox2ras <- str_to_mat(rec$TargetVox2RAS)
  if (!is.null(rec$TargetDim)) geom$target_dim <- as.integer(strsplit(trimws(rec$TargetDim), "\\s+")[[1]])
  if (length(geom)) res$geometry <- geom

  if (!is.null(rec$ForwardField)) {
    res$forward_field <- read_ants_warp(file.path(dir, rec$ForwardField))
  }
  if (!is.null(rec$InverseField)) {
    res$inverse_field <- read_ants_warp(file.path(dir, rec$InverseField))
  }

  class(res) <- "ravetools_register_volume3d"
  res
}

#' @export
print.ravetools_register_volume3d <- function(x, ...) {
  cat("<ravetools registration>\n")
  cat("  type:   ", x$type %||% "?", "\n")
  cat("  metric: ", paste(x$metric, collapse = ", "), "\n")
  has <- c(
    if (!is.null(x$transform)) "affine",
    if (!is.null(x$forward_field)) "forward_field",
    if (!is.null(x$inverse_field)) "inverse_field")
  cat("  parts:  ", paste(has, collapse = ", "), "\n")
  if (!is.null(x$geometry$target_dim)) {
    cat("  grid:   ", paste(x$geometry$target_dim, collapse = " x "), "\n")
  }
  invisible(x)
}

# ANTs / ITK-compatible transform I/O

rand_affine <- function(seed = 1) {
  set.seed(seed)
  # a proper rotation (QR of a random matrix) plus an anisotropic scale and shift
  q <- qr.Q(qr(matrix(rnorm(9), 3, 3)))
  if (det(q) < 0)
    q[, 1] <- -q[, 1]
  a <- q %*% diag(c(1.05, 0.97, 1.1))
  m <- diag(4)
  m[1:3, 1:3] <- a
  m[1:3, 4] <- c(12.5, -7.25, 3.0)
  m
}

call_rpyANTs <- function(.f, ...) {
  if (system.file(package = "rpyANTs") == "") {
    stop("`rpyANTs` is not installed")
  }
  if (.f == "py_to_r") {
    ns <- asNamespace("reticulate")
  } else {
    ns <- asNamespace("rpyANTs")
  }
  f <- ns[[.f]]
  f(...)
}

test_that("ras_lps_conjugate is an involution with the expected sign pattern", {
  m <- rand_affine(7)
  expect_equal(ras_lps_conjugate(ras_lps_conjugate(m)), m, tolerance = 1e-12)
  l <- ras_lps_conjugate(m)
  # F = diag(-1,-1,1,1): negate M[1,3], M[1,4], M[2,3], M[2,4], M[3,1], M[3,2]
  expect_equal(l[1, 3], -m[1, 3])
  expect_equal(l[1, 4], -m[1, 4])
  expect_equal(l[2, 3], -m[2, 3])
  expect_equal(l[2, 4], -m[2, 4])
  expect_equal(l[3, 1], -m[3, 1])
  expect_equal(l[3, 2], -m[3, 2])
  # unaffected entries
  expect_equal(l[1, 1], m[1, 1])
  expect_equal(l[3, 3], m[3, 3])
  # 3x4 input is accepted and squared up
  expect_equal(ras_lps_conjugate(m[1:3, ]), l)
})

test_that("affine .mat round-trips at full double precision", {
  m <- rand_affine(2)
  f <- tempfile(fileext = ".mat")
  write_ants_transform(m, f)
  # ITK level-4 .mat: no 128-byte header, two double variables
  expect_equal(unname(file.size(f)), 193)
  expect_equal(read_ants_transform(f), m, tolerance = 1e-12)
})

test_that("warp field round-trips through an ANTs 5-D NIfTI", {
  skip_if_not_installed("freesurferformats")
  set.seed(3)
  nd <- c(7L, 6L, 5L)
  field <- array(rnorm(prod(nd) * 3, sd = 2), c(nd, 3))
  v2r <- matrix(c(0.9, 0, 0, -3, 0, 1.1, 0, -2, 0, 0, 1.2, -1, 0, 0, 0, 1), 4, 4, byrow = TRUE)

  f <- tempfile(fileext = ".nii.gz")
  write_ants_warp(field, f, vox2ras = v2r)
  g <- read_ants_warp(f)
  expect_equal(dim(g), c(nd, 3L))
  expect_equal(as.vector(g), as.vector(field), tolerance = 1e-5)         # float32
  expect_equal(attr(g, "vox2ras"), v2r, tolerance = 1e-5)

  # on disk: 5-D vector image with the vector intent code and our sform
  h <- freesurferformats::read.nifti1.header(f)
  expect_equal(h$dim[1], 5L)
  expect_equal(h$dim[c(2, 3, 4, 5, 6)], c(nd, 1L, 3L))
  expect_equal(h$intent_code, 1007L)
  expect_equal(h$sform_code, 1L)
  expect_equal(h$srow_x, v2r[1, ], tolerance = 1e-5)

  # vox2ras can also travel as an attribute on the field
  attr(field, "vox2ras") <- v2r
  f2 <- tempfile(fileext = ".nii.gz")
  write_ants_warp(field, f2)
  expect_equal(attr(read_ants_warp(f2), "vox2ras"), v2r, tolerance = 1e-5)
})

test_that("the affine can be hidden in / recovered from the warp header (opt-in)", {
  skip_if_not_installed("freesurferformats")
  set.seed(4)
  nd <- c(6L, 5L, 4L)
  field <- array(rnorm(prod(nd) * 3), c(nd, 3))
  v2r <- diag(4)
  v2r[1:3, 4] <- c(-3, -2.5, -2)
  m <- rand_affine(5)

  f <- tempfile(fileext = ".nii.gz")
  write_ants_warp(field,
                  f,
                  vox2ras = v2r,
                  affine = m,
                  direction = "inverse")

  # off by default
  g0 <- read_ants_warp(f)
  expect_null(attr(g0, "transform"))

  # opt-in recovery; lossy precision (5 dp rotation, 2 dp translation)
  g <- read_ants_warp(f, recover_affine = TRUE)
  expect_false(is.null(attr(g, "transform")))
  expect_equal(attr(g, "direction"), "inverse")
  expect_equal(attr(g, "transform")[1:3, 1:3], m[1:3, 1:3], tolerance = 1e-4)
  expect_equal(attr(g, "transform")[1:3, 4], m[1:3, 4], tolerance = 1e-2)

  # values that overflow the fixed-width encoding are skipped with a warning
  big <- diag(4)
  big[1, 1] <- 42
  f2 <- tempfile(fileext = ".nii.gz")
  expect_warning(write_ants_warp(field, f2, vox2ras = v2r, affine = big))
  expect_null(attr(read_ants_warp(f2, recover_affine = TRUE), "transform"))
})

test_that("save_registration / load_registration round-trip via the manifest", {
  skip_if_not_installed("freesurferformats")
  set.seed(6)
  nd <- c(8L, 7L, 6L)
  m <- rand_affine(6)
  v2r <- diag(4)
  v2r[1:3, 4] <- c(-4, -3.5, -3)
  fwd <- array(rnorm(prod(nd) * 3), c(nd, 3))
  attr(fwd, "vox2ras") <- v2r
  inv <- array(rnorm(prod(nd) * 3), c(nd, 3))
  attr(inv, "vox2ras") <- v2r
  res <- structure(
    list(
      transform = m,
      forward_field = fwd,
      inverse_field = inv,
      type = "syn",
      metric = "mattes",
      geometry = list(
        source_vox2ras = v2r,
        target_vox2ras = v2r,
        target_dim = nd
      )
    ),
    class = "ravetools_register_volume3d"
  )

  d <- tempfile()
  dir.create(d)
  man <- save_registration(res, d)
  expect_true(file.exists(man))
  expect_setequal(
    list.files(d),
    c(
      "registration.dcf",
      "registration0GenericAffine.mat",
      "registration1Warp.nii.gz",
      "registration1InverseWarp.nii.gz"
    )
  )

  # load from a different working directory: paths resolve relative to manifest
  old <- setwd(tempdir())
  on.exit(setwd(old), add = TRUE)
  lr <- load_registration(man)
  setwd(old)

  expect_s3_class(lr, "ravetools_register_volume3d")
  expect_equal(lr$transform, m, tolerance = 1e-12)            # .mat is full precision
  expect_equal(as.vector(lr$forward_field), as.vector(fwd), tolerance = 1e-5)
  expect_equal(as.vector(lr$inverse_field), as.vector(inv), tolerance = 1e-5)
  expect_equal(lr$geometry$target_dim, nd)
  expect_equal(lr$geometry$source_vox2ras, v2r, tolerance = 1e-5)
  expect_equal(lr$type, "syn")

  # single-file dispatch
  expect_equal(load_registration(file.path(d, "registration0GenericAffine.mat")), m, tolerance = 1e-12)
  g <- load_registration(file.path(d, "registration1InverseWarp.nii.gz"),
                         recover_affine_from_header = TRUE)
  expect_equal(dim(g), c(nd, 3L))
  expect_equal(attr(g, "direction"), "inverse")
})

# ---- cross-validation against a real ANTs install --------------------------
# All rpyANTs calls should be

ants_ready <- function() {
  isTRUE(tryCatch({
    call_rpyANTs("ants_available")
  }, error = function(e) {
    FALSE
  }))
}

test_that("our affine .mat is read correctly by ANTs (and vice versa)", {
  skip_on_cran()
  skip_if_not_installed("rpyANTs")
  skip_if_not(ants_ready())
  ants <- call_rpyANTs("load_ants")
  m <- rand_affine(11)

  # forward: write ours, read with ANTs. as_ANTsTransform(path)[] is the LPS
  # (ITK-native) matrix, so conjugate back to RAS to compare.
  f <- tempfile(fileext = ".mat")
  write_ants_transform(m, f)
  m_lps <- call_rpyANTs("as_ANTsTransform", f)[]
  expect_equal(ras_lps_conjugate(m_lps), m, tolerance = 1e-5)

  # reverse: ANTs writes the LPS matrix, we read it back as RAS.
  tx <- call_rpyANTs("as_ANTsTransform", ras_lps_conjugate(m))
  f2 <- tempfile(fileext = ".mat")
  ants$write_transform(tx, f2)
  expect_equal(read_ants_transform(f2), m, tolerance = 1e-5)
})

test_that("our warp NIfTI is read correctly by ANTs", {
  skip_on_cran()
  skip_if_not_installed("rpyANTs")
  skip_if_not_installed("freesurferformats")
  skip_if_not(ants_ready())
  ants <- call_rpyANTs("load_ants")

  set.seed(12)
  nd <- c(7L, 6L, 5L)
  field <- array(rnorm(prod(nd) * 3, sd = 1.5), c(nd, 3))
  v2r <- matrix(c(0.9, 0, 0, -3, 0, 1.1, 0, -2, 0, 0, 1.2, -1, 0, 0, 0, 1), 4, 4, byrow = TRUE)
  f <- tempfile(fileext = ".nii.gz")
  write_ants_warp(field, f, vox2ras = v2r)

  img <- ants$image_read(f)

  expect_equal(as.integer(call_rpyANTs("py_to_r", img$components)), 3L)
  expect_equal(unlist(call_rpyANTs("py_to_r", img$shape)), nd, tolerance = 1e-6)
  expect_equal(unlist(call_rpyANTs("py_to_r", img$spacing)), c(0.9, 1.1, 1.2), tolerance = 1e-5)

  # ANTs stores vectors in LPS: x,y components negated relative to our RAS field
  arr <- call_rpyANTs("py_to_r", img$numpy())
  expect_equal(dim(arr), c(nd, 3L))
  lps <- field
  lps[, , , 1] <- -lps[, , , 1]
  lps[, , , 2] <- -lps[, , , 2]
  expect_equal(as.vector(arr), as.vector(lps), tolerance = 1e-5)
})

#' @title Shift array by index
#' @description Re-arrange arrays in parallel
#' @param x array, must have at least matrix
#' @param along_margin which index is to be shifted
#' @param unit_margin which dimension decides \code{shift_amount}
#' @param shift_amount shift amount along \code{along_margin}
#' @returns An array with same dimensions as the input \code{x}, but with
#' index shifted. The missing elements will be filled with \code{NA}.
#' @details
#' A simple use-case for this function is to think of a matrix where each row
#' is a signal and columns stand for time. The objective is to align (time-lock)
#' each signal according to certain events. For each signal, we want to shift
#' the time points by certain amount.
#'
#' In this case, the shift amount is defined by \code{shift_amount}, whose
#' length equals to number of signals. \code{along_margin=2} as we want to shift
#' time points (column, the second dimension) for each signal. \code{unit_margin=1}
#' because the shift amount is depend on the signal number.
#'
#' @examples
#'
#' # Set ncores = 2 to comply to CRAN policy. Please don't run this line
#' ravetools_threads(n_threads = 2L)
#'
#'
#'
#' x <- matrix(1:10, nrow = 2, byrow = TRUE)
#' z <- shift_array(x, 2, 1, c(1,2))
#'
#' y <- NA * x
#' y[1,1:4] = x[1,2:5]
#' y[2,1:3] = x[2,3:5]
#'
#' # Check if z ang y are the same
#' z - y
#'
#' # array case
#' # x is Trial x Frequency x Time
#' x <- array(1:27, c(3,3,3))
#'
#' # Shift time for each trial, amount is 1, -1, 0
#' shift_amount <- c(1,-1,0)
#' z <- shift_array(x, 3, 1, shift_amount)
#'
#' oldpar <- par(mfrow = c(3, 2), mai = c(0.8, 0.6, 0.4, 0.1))
#' for( ii in 1:3 ) {
#'   image(t(x[ii, ,]), ylab = 'Frequency', xlab = 'Time',
#'         main = paste('Trial', ii))
#'   image(t(z[ii, ,]), ylab = 'Frequency', xlab = 'Time',
#'         main = paste('Shifted amount:', shift_amount[ii]))
#' }
#' par(oldpar)
#'
#' @export
shift_array <- function(x, along_margin, unit_margin, shift_amount) {
  UseMethod("shift_array")
}

#' @export
shift_array.array <- function(x, along_margin, unit_margin, shift_amount) {

  along_margin <- as.integer(along_margin)[[1]]
  unit_margin <- as.integer(unit_margin)[[1]]

  if (is.na(along_margin) || is.na(unit_margin)) {
    stop("`shift_array`: `along_margin` and `unit_margin` cannot contain NA")
  }
  xdim <- dim(x)

  if (!is.integer(xdim)) {
    xdim_int <- as.integer(xdim)
    if (!all(xdim_int == xdim)) {
      stop("Array `x` has large dimension (>= 2^31 elements) that cannot be indexed by integers.")
    }
    dim(x) <- xdim_int
  }

  re <- shiftArray(x, along_margin, unit_margin, shift_amount)
  if (inherits(re, "ravetools_error")) {
    stop(re)
  }
  re

}

test_that("Shift array with different types", {
  x0 <- as.integer(1:10)
  x <- matrix(x0, nrow = 2, byrow = TRUE)
  z <- shift_array(x, 2, 1, c(1, 2))
  y <- NA * x
  y[1, 1:4] <- x[1, 2:5]
  y[2, 1:3] <- x[2, 3:5]
  expect_identical(y, z)

  x0 <- as.double(1:10)
  x <- matrix(x0, nrow = 2, byrow = TRUE)
  z <- shift_array(x, 2, 1, c(-2, 1))
  y <- NA * x
  y[1, 3:5] <- x[1, 1:3]
  y[2, 1:4] <- x[2, 2:5]
  expect_identical(y, z)

  x0 <- as.complex(1:10)
  x <- matrix(x0, nrow = 2, byrow = TRUE)
  z <- shift_array(x, 2, 1, c(-2, 1))
  y <- NA * x
  y[1, 3:5] <- x[1, 1:3]
  y[2, 1:4] <- x[2, 2:5]
  expect_identical(y, z)

  x0 <- sample(c(TRUE, FALSE), 10, replace = TRUE)
  x <- matrix(x0, nrow = 2, byrow = TRUE)
  z <- shift_array(x, 2, 1, c(1, 2))
  y <- x
  y[] <- NA
  y[1, 1:4] <- x[1, 2:5]
  y[2, 1:3] <- x[2, 3:5]
  expect_identical(y, z)

})


test_that("Shift array with different arrays", {
  dim <- c(4, 15, 10)
  x <- array(rnorm(prod(dim)), dim)
  shift_amount <- c(2, 10, -3, NA)
  z <- shift_array(
    x,
    along_margin = 3,
    unit_margin = 1,
    shift_amount = shift_amount
  )

  y <- x * NA
  for (ii in 1:3) {
    idx <- shift_amount[ii] + 1:10
    sel <- idx >= 1 & idx <= 10
    y[ii, , sel] <- x[ii, , idx[sel]]
  }
  y[4, , ] <- NA
  expect_identical(y, z)


  shift_amount <- sample(c(-5:5, NA, NA), size = 15, replace = TRUE)
  z <- shift_array(
    x,
    along_margin = 1,
    unit_margin = 2,
    shift_amount = shift_amount
  )

  y <- x * NA
  for (ii in 1:15) {
    idx <- shift_amount[ii] + 1:4
    if (!is.na(shift_amount[ii])) {
      sel <- idx >= 1 & idx <= 4
      y[sel, ii, ] <- x[idx[sel], ii, ]
    }
  }
  expect_identical(y, z)
})

# Resolve one side of `vcg_detect_collision` into the layout the C++ core
# expects: a 3 x n coordinate matrix plus, in mesh mode, a 0-based 3 x m face
# matrix. Returns the resolved mode alongside, since "auto" is decided here.
resolve_collision_input <- function(obj, mode, arg) {
  is_matrix_like <- is.matrix(obj) || (is.array(obj) && length(dim(obj)) == 2L)

  if (mode == "auto" && !is_matrix_like) {
    mesh <- ensure_mesh3d(obj)
    if (length(mesh$it)) {
      mode <- "mesh"
    }
  }

  if (mode == "auto") {
    # Check points or segments
    mode <- "points"
  }

  if (mode == "mesh") {
    if (is_matrix_like) {
      stop(sprintf(
        "`vcg_detect_collision`: `%s` is a matrix, which carries no faces; `mode_%s = \"mesh\"` needs a `mesh3d` object. Use \"points\" or \"segments\" instead.",
        arg, arg))
    }
    surface <- meshintegrity(mesh = ensure_mesh3d(obj), facecheck = TRUE)
    return(list(
      mode = mode,
      vb = surface$vb[1:3, , drop = FALSE],
      it = surface$it - 1L
    ))
  }

  vb <- as_point_cloud_matrix(obj, "vcg_detect_collision")

  # Normalize partial NA / non-finite rows to whole NA columns so the C++ side
  # only ever sees a column that is either wholly usable or a separator.
  incomplete <- !is.finite(vb[1L, ]) | !is.finite(vb[2L, ]) | !is.finite(vb[3L, ])
  if (any(incomplete)) {
    vb[, incomplete] <- NA_real_
  }

  list(mode = mode, vb = vb, it = NULL)
}


#' @title Detect collisions between two geometries
#' @description
#' Reports, for every element of \code{y}, whether it comes within
#' \code{radius} of \code{x}, together with the exact minimum distance. Each
#' side may be a point cloud, a set of connected line segments (for example
#' diffusion streamlines or an electrode shaft), or a triangular mesh, so the
#' function covers all nine pairings with one call.
#'
#' The test is exact: it produces neither false negatives nor false positives,
#' and \code{distance} is the true minimum distance rather than a
#' vertex-sampled approximation.
#'
#' @param x the geometry that gets indexed; a matrix with \code{n} rows and 2
#' or 3 columns, or a surface (see \code{\link{ensure_mesh3d}})
#' @param y the geometry that gets queried; results are aligned to this side.
#' Same accepted types as \code{x}
#' @param mode_x,mode_y how to interpret \code{x} and \code{y}: \code{'points'}
#' treats each row as an isolated point, \code{'segments'} chains consecutive
#' rows together (\code{p1-p2}, \code{p2-p3}, ...), and \code{'mesh'} uses
#' the triangular faces. The default \code{'auto'} resolves to \code{'mesh'}
#' for a surface that has faces and \code{'points'} otherwise, so a matrix of
#' connected line segments must ask for \code{'segments'} explicitly
#' @param radius distance tolerance; a collision is reported when the minimum
#' distance is at most \code{radius}. Defaults to \code{0}, meaning literal
#' contact; a positive tolerance is usually what is wanted
#' @param early_stop whether to stop as soon as a collision is found, scanning
#' independently within each group of \code{y} delimited by \code{NA} rows. A
#' matrix carrying no \code{NA} rows is a single group, so a point cloud stops
#' after one hit overall rather than per row; likewise \code{mode_y = 'mesh'}
#' stops at the first colliding face. Skipped elements report \code{hit = NA},
#' meaning \emph{not evaluated} rather than \emph{no collision}, and the
#' collision reported is the first in row order, not the closest. Use
#' \code{FALSE} (the default) whenever the per-element profile or a minimum
#' distance is wanted
#' @param include_interior whether geometry lying strictly inside a closed
#' \code{x} surface counts as a collision even when it never comes within
#' \code{radius} of the surface itself; requires \code{mode_x = 'mesh'}.
#' Default is \code{FALSE}
#'
#' @returns A list of three vectors, each aligned one-to-one with the rows of
#' \code{y} (or with the faces of \code{y} when \code{mode_y = 'mesh'}):
#' \describe{
#'   \item{\code{hit}}{logical. \code{NA} marks an element with nothing to
#'     test: a separator row, or the final vertex of each group when
#'     \code{mode_y = 'segments'}. When \code{early_stop = TRUE}, \code{NA}
#'     also marks elements the scan never reached}
#'   \item{\code{distance}}{the exact minimum distance where \code{hit} is
#'     \code{TRUE}, and \code{NA} everywhere else. Distances beyond
#'     \code{radius} are never computed, so a \code{FALSE} element carries no
#'     distance}
#'   \item{\code{index}}{1-based index of the closest element of \code{x}
#'     within \code{radius}: a row when \code{mode_x} is \code{'points'}, the
#'     row where the segment starts when \code{'segments'}, or a face column
#'     when \code{'mesh'}. \code{NA} where there is no collision}
#' }
#'
#' @details
#' \code{x} is indexed once into a uniform spatial grid and \code{y} is
#' streamed against it, so put the larger or repeatedly-reused geometry in
#' \code{x} and the geometry you want per-element answers about in \code{y}.
#'
#' Several chains share one matrix and are delimited by rows of \code{NA}; any
#' row that is not fully finite is treated as a separator. This lets a whole
#' bundle of streamlines be passed as a single matrix:
#'
#' \preformatted{
#'   rbind(
#'     c(0, 0, 0), c(1, 0, 0), c(2, 0, 0),   # first streamline
#'     c(NA, NA, NA),                        # separator
#'     c(0, 5, 0), c(1, 5, 0)                # second streamline
#'   )
#' }
#'
#' Those separators also define the unit that \code{early_stop} works on: it
#' stops once per group, so a bundle of streamlines yields at most one hit per
#' streamline.
#'
#' When \code{mode_y = 'segments'}, row \code{i} of the result describes the
#' segment running from row \code{i} to row \code{i + 1}.
#'
#' Both geometries must already share one coordinate space. Surface
#' coordinates, volume \verb{IJK} indices, and scanner coordinates are all
#' different spaces, and no transform is applied here.
#'
#' \code{include_interior} relies on ray casting and therefore needs \code{x}
#' to be watertight with coherently oriented faces; see
#' \code{\link{vcg_fix_defects}} to repair a surface that is not.
#'
#' Multi-threading follows \code{\link{ravetools_threads}}. The interior test
#' always runs single-threaded.
#'
#' @inheritSection ensure_mesh3d Coercing Surface Inputs
#'
#' @examples
#'
#' library(ravetools)
#'
#' # A spherical region of interest
#' roi <- vcg_sphere()
#'
#' # Two streamlines in one matrix, separated by an NA row: the first passes
#' # through the sphere, the second stays well clear of it
#' streamlines <- rbind(
#'   cbind(seq(-3, 3, by = 0.5), 0, 0),
#'   c(NA, NA, NA),
#'   cbind(seq(-3, 3, by = 0.5), 5, 0)
#' )
#'
#' result <- vcg_detect_collision(
#'   x = roi,
#'   y = streamlines,
#'   mode_y = "segments",
#'   radius = 0.1
#' )
#'
#' # Which segments touch the sphere, and how close do they get?
#' data.frame(
#'   hit = result$hit,
#'   distance = round(result$distance, 4)
#' )
#'
#' # A point cloud against the same region, asking only for proximity
#' points <- rbind(c(0, 0, 0), c(1.05, 0, 0), c(10, 10, 10))
#' vcg_detect_collision(roi, points, radius = 0.1)$hit
#'
#' # The centre of the sphere is far from its surface, so it only counts as a
#' # collision when the interior is included
#' vcg_detect_collision(roi, rbind(c(0, 0, 0)), radius = 0.1)$hit
#' vcg_detect_collision(roi, rbind(c(0, 0, 0)), radius = 0.1,
#'                      include_interior = TRUE)$hit
#'
#' @export
vcg_detect_collision <- function(
    x, y,
    mode_x = c("auto", "points", "segments", "mesh"),
    mode_y = c("auto", "points", "segments", "mesh"),
    radius = 0,
    early_stop = FALSE,
    include_interior = FALSE) {

  mode_x <- match.arg(mode_x)
  mode_y <- match.arg(mode_y)

  radius <- as.double(radius)
  stopifnot2(
    length(radius) == 1L && !is.na(radius) && is.finite(radius) && radius >= 0,
    msg = "`vcg_detect_collision`: `radius` must be a single finite non-negative number."
  )

  early_stop <- isTRUE(as.logical(early_stop))
  include_interior <- isTRUE(as.logical(include_interior))

  side_x <- resolve_collision_input(x, mode_x, "x")
  side_y <- resolve_collision_input(y, mode_y, "y")

  if (include_interior && side_x$mode != "mesh") {
    stop("`vcg_detect_collision`: `include_interior = TRUE` needs `x` to be a closed surface, so `mode_x` must resolve to \"mesh\" (it resolved to \"", side_x$mode, "\").")
  }

  mode_code <- c(points = 0L, segments = 1L, mesh = 2L)

  vcgDetectCollision(
    x_ = side_x$vb,
    x_it_ = side_x$it,
    y_ = side_y$vb,
    y_it_ = side_y$it,
    mode_x = mode_code[[side_x$mode]],
    mode_y = mode_code[[side_y$mode]],
    radius = radius,
    early_stop = early_stop,
    include_interior = include_interior
  )
}

#' Calculate 'Welch Periodogram'
#' @description \code{pwelch} is for single signal trace only; \code{mv_pwelch}
#' is for multiple traces. Currently \code{mv_pwelch} is experimental and
#' should not be called directly.
#' @param x numerical vector or a row-major vector, signals.
#' If \code{x} is a matrix, then each row is a channel. For \code{plot}
#' function, \code{x} is the instance returned by \code{pwelch} function.
#' @param fs sample rate, average number of time points per second
#' @param window window length in time points, default size is \code{64}
#' @param window_family function generator for generating filter windows,
#' default is \code{\link{hamming}}. This can be any window function listed in
#' the filter window family, or any window generator function from package
#' \code{gsignal}. Default is \code{\link{hamming}}. For 'iEEG' users, both
#' \code{hamming} and \code{\link{blackmanharris}} are offered by 'EEG-lab';
#' while \code{blackmanharris} offers better attenuation than Hamming windows,
#' it also has lower spectral resolution. \code{hamming} has a 42.5 dB side-lobe
#' attenuation. This may mask spectral content below this value (relative
#' to the peak spectral content). Choosing different windows enables
#' you to make trade-off between resolution (e.g., using a rectangular
#' window) and side-lobe attenuation (e.g., using a \code{\link{hanning}}
#' window)
#' @param nfft number of points in window function; default is automatically
#' determined from input data and window, scaled up to the nearest power of 2
#' @param noverlap overlap between two adjacent windows, measured in time
#' points; default is half of the \code{window}
#' @param log indicates which axis should be \code{log10}-transformed, used by the plot function. For \code{'x'} axis, it's \code{log10}-transform; for \code{'y'} axis, it's \code{10log10}-transform (decibel unit). Choices are \code{"xy"}, \code{"x"}, \code{"y"}, and \code{""}.
#' @param plot integer, whether to plot the result or not; choices are \code{0}, no plot; \code{1} plot on a new canvas; \code{2} add to existing canvas
#' @param add logical, whether the plot should be added to existing canvas
#' @param ... will be passed to \code{plot.pwelch} or ignored
#' @param col,xlim,ylim,main,type,cex,las,xlab,ylab,lty,lwd,xaxs,yaxs,mar,mgp,tck parameters passed to \code{\link[graphics]{plot.default}}
#' @param se logical or a positive number indicating whether to plot standard
#' error of mean; default is false. If provided with a number, then a multiple
#' of standard error will be drawn. This option is only available when power
#' is in log-scale (decibel unit)
#' @param col.se,alpha.se controls the color and opacity of the standard error
#' @param xticks ticks to show on frequency axis
#' @param xline,yline controls how close the axis labels to the corresponding axes
#' @param grid whether to draw rectangular grid lines to the plot; only
#' respected when \code{add=FALSE}; default is true
#' @param margin the margin in which \code{pwelch} should be applied to
#' @returns A list with class \code{'ravetools-pwelch'} that contains the
#' following items:
#' \describe{
#' \item{\code{freq}}{frequencies used to calculate the 'periodogram'}
#' \item{\code{spec}}{resulting spectral power for each frequency}
#' \item{\code{window}}{window function(in numerical vector) used}
#' \item{\code{noverlap}}{number of overlapping time-points between two adjacent windows}
#' \item{\code{nfft}}{number of basis functions}
#' \item{\code{fs}}{sample rate}
#' \item{\code{x_len}}{input signal length}
#' \item{\code{method}}{a character string \code{'Welch'}}
#' }
#' @examples
#'
#' x <- rnorm(1000)
#' pwel <- pwelch(x, 100)
#' pwel
#'
#' plot(pwel, log = "xy")
#'
#' @export
pwelch <- function(
    x, fs, window = 64, noverlap = window / 2, nfft = "auto", window_family = hamming,
    col = "black", xlim = NULL, ylim = NULL, main = "Welch periodogram",
    plot = 0, log = c("xy", "", "x", "y"), ...
) {
  UseMethod("pwelch")
}

#' @export
pwelch.matrix <- function(
    x, fs, window = 64, noverlap = window / 2, nfft = "auto", window_family = hamming,
    col = "black", xlim = NULL, ylim = NULL, main = "Welch periodogram",
    plot = 0, log = c("xy", "", "x", "y"), margin = 1L, ...) {

  if (margin != 2) {
    # column-major
    x <- t(x)
  }
  x_len <- nrow(x)
  noverlap <- ceiling(noverlap)

  if (identical(nfft, "auto") || nfft <= 1) {
    nfft <- 256
    nfft <- max(min(nfft, x_len), window)
    nfft <- 2^ceiling(log2(nfft))
  } else {
    nfft <- max(min(nfft, x_len), window)
  }
  nfft <- ceiling(nfft)

  window <- window_family(window)

  # window_norm = norm(window, '2')
  window_len <- length(window)

  # normalization <- mean(window^2)

  step <- max(floor(window_len - noverlap + 0.99), 1)

  ## Average the slices
  offset <- seq(1, max(x_len - window_len + 1, 1), by = step)

  N <- length(offset)

  # re-slice data
  nchannels <- ncol(x)
  re <- sapply(offset, function(idx_start) {
    x[idx_start - 1 + seq_len(window_len), , drop = FALSE]
  })
  dim(re) <- c(window_len, length(re) / window_len)

  re <- apply(re, 2, function(slice) {
    a <- detrend_naive(slice)
    postpad(a$Y * window, nfft)
  })

  re <- Mod(mvfftw_r2c(re)) ^ 2

  NN <- ceiling((nfft + 1) / 2)
  re <- re[seq_len(NN), , drop = FALSE] / window_len

  # calculate mean
  dim(re) <- c(NN, nchannels, N)
  re <- collapse(re, keep = c(2L, 1L), average = TRUE)
  # re <- apply(re, 1, rowMeans) # nchannels x NN
  # if(!is.matrix(re)) {
  #   dim(re) <- c(nchannels, NN)
  # }

  freq <- seq(0, fs / 2, length.out = NN)

  res <- structure(list(
    freq = freq,
    spec = re,
    nchannels = nchannels,
    df = N - 1,
    window = window,
    noverlap = noverlap,
    nfft = nfft,
    fs = fs,
    x_len = length(x),
    method = "Welch"
  ), class = c("pwelch-multi", "ravetools-pwelch", "pwelch"))

  if ( plot ) {
    if (!is.null(log)) {
      log <- match.arg(log)
    }
    plot(res, col = col, xlim = xlim, ylim = ylim, main = main,
         add = plot >= 2, log = log, ...)
    return(invisible(res))
  }
  return(res)
}

#' @export
pwelch.default <- function(
  x, fs, window = 64, noverlap = window / 2, nfft = "auto", window_family = hamming,
  col = "black", xlim = NULL, ylim = NULL, main = "Welch periodogram",
  plot = 0, log = c("xy", "", "x", "y"), ...) {


  # list2env(list(window = 64, noverlap = 8, nfft = 256,
  #               col = "black", xlim = NULL, ylim = NULL, main = "Welch periodogram",
  #               plot = TRUE, log = "xy", spec_func = stats::spectrum, cex = 1), .GlobalEnv)

  x <- as.vector(x)
  x_len <- length(x)
  noverlap <- ceiling(noverlap)

  if (identical(nfft, "auto") || nfft <= 1) {
    nfft <- 256
    nfft <- max(min(nfft, x_len), window)
    nfft <- 2^ceiling(log2(nfft))
  } else {
    nfft <- max(min(nfft, x_len), window)
  }
  nfft <- ceiling(nfft)


  window <- window_family(window)

  # window_norm = norm(window, '2')
  window_len <- length(window)

  # normalization <- mean(window^2)

  step <- max(floor(window_len - noverlap + 0.99), 1)

  ## Average the slices
  offset <- seq(1, max(x_len - window_len + 1, 1), by = step)

  N <- length(offset)

  re <- sapply(seq_len(N), function(i) {
    slice <- x[offset[i] - 1 + seq_len(window_len)]
    slice[is.na(slice)] <- 0
    a <- detrend_naive(slice)
    postpad(a$Y * window, nfft)
    # # HermConj = 0 : without the "Hermitian" redundancy
    # a = fftwtools::fftw_r2c(postpad(a$Y * window, nfft), HermConj = 0)
    # Mod(a)^2
  })

  re <- Mod(mvfftw_r2c(re))^2

  NN <- ceiling((nfft + 1) / 2)

  re <- re[seq_len(NN), , drop = FALSE] / window_len

  spec <- rowMeans(re)

  # decibel unit so we can calculate sterr of mean
  re_db <- 10 * log10(re)
  spec_db <- rowMeans(re_db)
  if (N > 1) {
    spec_db_se <- apply(re_db, 1, stats::sd) / sqrt(N - 1)
  } else {
    spec_db_se <- rep(N, nrow(re))
  }

  freq <- seq(0, fs / 2, length.out = NN)

  res <- structure(list(
    freq = freq,
    spec = spec,
    spec_db = spec_db,
    spec_db_se = spec_db_se,
    nchannels = 1,
    df = N - 1,
    window = window,
    noverlap = noverlap,
    nfft = nfft,
    fs = fs,
    x_len = length(x),
    method = "Welch"
  ), class = c("ravetools-pwelch", "pwelch"))

  if ( plot ) {
    if (!is.null(log)) {
      log <- match.arg(log)
    }
    plot(res, col = col, xlim = xlim, ylim = ylim, main = main,
         add = plot >= 2, log = log, ...)
    return(invisible(res))
  }
  return(res)
}

#' @rdname pwelch
#' @export
`print.ravetools-pwelch` <- function(x, ...) {
  cat(paste0(
    "Welch Periodogram:\n",
    sprintf("  # channels: %.0f\n", x$nchannels),
    sprintf("  time points: %d\n", x$x_len),
    sprintf("  sample rate: %.2f\n", x$fs),
    sprintf("  window size: %d\n", length(x$window)),
    sprintf("  window overlaps: %d\n", x$noverlap),
    sprintf("  filter count: %d\n", x$nfft)
  ))
  invisible(x)
}

#' @rdname pwelch
#' @export
`plot.ravetools-pwelch` <- function(
    x, log = c("xy", "x", "y", ""), se = FALSE, xticks, type = "l", add = FALSE,
    col = graphics::par("fg"), col.se = "orange", alpha.se = 0.5, lty = 1, lwd = 1,
    cex = 1, las = 1, main = "Welch periodogram", xlab, ylab,
    xlim = NULL, ylim = NULL, xaxs = "i", yaxs = "i",
    xline = 1.2 * cex, yline = 2.0 * cex,
    mar = c(2.6, 3.8, 2.1, 0.6) * (0.5 + cex / 2), mgp = cex * c(2, 0.5, 0),
    tck = -0.02 * cex, grid = TRUE, ...) {
  # DIPSAUS DEBUG START
  # x <- mv_pwelch(rbind(rep(0, 2000), rep(0, 2000)), 1, 100)
  # x <- mv_pwelch(rbind(rnorm(2000), rnorm(2000)), 1, 100)
  # log = "xy"; se = FALSE; type = 'l'; add = FALSE
  # col = graphics::par("fg"); col.se = "orange"; alpha.se = 0.5; lty = 1; lwd = 1
  # cex = 1; las = 1; main = 'Welch periodogram'
  # xlim = NULL; ylim = NULL; xaxs ="i"; yaxs = "i"
  # xline = 1.2 * cex; yline = 2.0 * cex
  # mar = c(2.6, 3.8, 2.1, 0.6) * (0.5 + cex / 2); mgp = cex * c(2, 0.5, 0);
  # tck = -0.02 * cex; grid = TRUE

  if (!is.null(log) && !identical(log, "")) {
    log <- match.arg(log)
  } else {
    log <- ""
  }
  freq <- x$freq

  if (!length(xlim)) {
    xlim <- range(freq)
  } else {
    if (x$fs > 500 && xlim[[1]] == 0 && xlim[[2]] < 2.5) {
      warning("`plot.pwelch`: `xlim` should be the frequency range in native values not log-range. Please check your plot function")
      xlim <- 10^(xlim)
    }
  }

  if (!missing(xticks)) {
    xticks <- xticks[xticks >= min(xlim) & xticks <= max(xlim)]
    xlabel <- c(xticks, xlim)
  } else {
    xlabel <- pretty(xlim)
  }

  if ( x$df < 1 || log %in% c("x", "") || !length(x$spec_db_se) ) {
    se <- FALSE
  }
  spec <- x$spec
  if ( se ) {
    spec_lb <- 10 * log10(spec) - x$spec_db_se * as.numeric(se)
    spec_ub <- 10 * log10(spec) + x$spec_db_se * as.numeric(se)
  } else {
    spec_lb <- 0
    spec_ub <- 0
  }

  switch(
    log,
    "xy" = {
      xlab %?<-% "Log10(Frequency)"
      ylab %?<-% "Power (dB)"

      xat <- xlabel
      xat[xat <= 0 ] <- min(freq[freq > 0])
      xat <- log10(xat)
      freq <- log10(freq)
      spec <- 10 * log10(spec)

      xlim <- range(xat, na.rm = TRUE)
      if (xlim[1] < min(freq, na.rm = TRUE)) {
        xlim[1] <- min(freq, na.rm = TRUE)
      }
    },
    "x" = {
      xlab %?<-% "Log10(Frequency)"
      ylab %?<-% "Power"
      xat <- xlabel
      xat[xat <= 0 ] <- min(freq[freq > 0])
      xat <- log10(xat)
      freq <- log10(freq)
      spec <- x$spec

      xlim <- range(xat, na.rm = TRUE)
      if (xlim[1] < min(freq, na.rm = TRUE)) {
        xlim[1] <- min(freq, na.rm = TRUE)
      }
      se <- FALSE
    },
    "y" = {
      xlab %?<-% "Frequency"
      ylab %?<-% "Power (dB)"
      spec <- 10 * log10(spec)
      xat <- xlabel
    },
    {
      xlab %?<-% "Frequency"
      ylab %?<-% "Power"
      spec <- x$spec
      xat <- xlabel
      se <- FALSE
    }
  )
  if (!length(ylim)) {
    spec_ <- spec[is.finite(spec)]
    if (!length(spec_)) {
      ylim <- c(-100, 0)
      spec_range <- c(-100, 0)
    } else {
      ylim <- range(pretty(spec), na.rm = TRUE)
      spec_range <- range(spec, na.rm = TRUE)
    }
  } else {
    spec_range <- range(spec, na.rm = TRUE)
  }

  if (!is.matrix(spec)) {
    spec <- matrix(spec, nrow = x$nchannels)
  }

  cex_params <- graphics::par("mgp", "mar", "mai", "cex.main", "cex.lab", "cex.axis", "cex.sub")

  if (!add) {

    graphics::par(mgp = mgp, mar = mar)
    on.exit({
      do.call(graphics::par, cex_params)
    })

    graphics::plot(
      range(freq, na.rm = TRUE), spec_range, type = "n", xlab = "", ylab = "",
      xlim = xlim, ylim = ylim, main = main, las = las,
      axes = FALSE, xaxs = xaxs, yaxs = yaxs,
      cex = cex, cex.main = cex_params$cex.main * cex,
      cex.lab = cex_params$cex.lab * cex,
      cex.axis = cex_params$cex.axis * cex, ...)
    graphics::axis(1, at = xat, labels = xlabel, tck = tck,
                   cex = cex, cex.main = cex_params$cex.main * cex,
                   cex.lab = cex_params$cex.lab * cex,
                   cex.axis = cex_params$cex.axis * cex)
    graphics::axis(2, at = pretty(ylim), las = 1, tck = tck,
                   cex = cex, cex.main = cex_params$cex.main * cex,
                   cex.lab = cex_params$cex.lab * cex,
                   cex.axis = cex_params$cex.axis * cex)

    graphics::mtext(side = 2, text = ylab, line = yline,
                    cex = cex_params$cex.lab * cex)
    graphics::mtext(side = 1, text = xlab, line = xline,
                    cex = cex_params$cex.lab * cex)


    if (grid) {
      graphics::grid()
    }

    if (!isFALSE(se)) {
      graphics::polygon(c(freq, rev(freq)), c(spec_lb, rev(spec_ub)),
              density = NA, border = NA,
              col = grDevices::adjustcolor(col.se, alpha.f = alpha.se),
              cex = cex, cex.main = cex_params$cex.main * cex,
              cex.lab = cex_params$cex.lab * cex,
              cex.axis = cex_params$cex.axis * cex, )
    }
    # graphics::lines(x = freq, y = spec, col = col, lty = lty, lwd = lwd, type = type)

  }

  spec_t <- t(spec)
  sel <- is.finite(freq) & (rowSums(!is.finite(spec_t)) == 0)
  if (any(sel)) {
    graphics::matpoints(freq[sel], spec_t[sel, , drop = FALSE],  type = type, col = col, lty = lty, lwd = lwd,
                        cex = cex, cex.main = cex_params$cex.main * cex,
                        cex.lab = cex_params$cex.lab * cex,
                        cex.axis = cex_params$cex.axis * cex, ...)

  }

  invisible(list(
    xlim = xlim,
    ylim = ylim
  ))
}



#' @rdname pwelch
#' @export
mv_pwelch <- function(x, margin, fs, window = 64, noverlap = window / 2,
                      nfft = "auto", window_family = hamming) {
  if (margin != 2L) {
    x <- t(x)
  }
  noverlap <- ceiling(noverlap)
  xlen <- length(x) / dim(x)[[2]]

  if (identical(nfft, "auto") || nfft <= 1) {
    nfft <- 256
    nfft <- max(min(nfft, xlen), window)
    nfft <- 2^ceiling(log2(nfft))
  } else {
    nfft <- max(min(nfft, xlen), window)
  }
  nfft <- ceiling(nfft)
  window <- window_family(window)
  window_len <- length(window)

  if ( noverlap >= window_len ) {
    noverlap <- window_len - 1
  } else if ( noverlap < 0 ) {
    noverlap <- 0
  }
  if ( xlen > window_len ) {
    start <- seq(1, xlen - window_len + 1, by = window_len - noverlap)
    slices <- sapply(start, function(si) {
      x[seq(si, si + window_len - 1), , drop = FALSE]
    })
    dim(slices) <- c(window_len, length(slices) / window_len)
    slices <- apply(slices, 2L, function(s) {
      a <- detrend_naive(s)
      postpad(a$Y * window, nfft)
    })
  } else {
    slices <- apply(x, 2L, function(s) {
      a <- detrend_naive(s)
      postpad(a$Y * window, nfft)
    })
  }

  re <- Mod(mvfftw_r2c(slices))^2
  NN <- ceiling((nfft + 1) / 2)
  spec <- rowMeans(re) / window_len
  spec <- spec[seq_len(NN)]
  freq <- seq(0, fs / 2, length.out = NN)

  re <- structure(list(
    freq = freq,
    spec = spec,
    window = window,
    noverlap = NA,
    nfft = nfft,
    fs = fs,
    x_len = xlen,
    method = "Welch",
    nchannels = 1L
  ), class = c("ravetools-pwelch", "pwelch"))

  re
}

#' Apply 'Notch' filter
#' @details Mainly used to remove electrical line frequencies
#' at 60, 120, and 180 \code{Hz}.
#' @param s numerical vector if \code{domain=1} (voltage
#' signals), or complex vector if \code{domain=0}
#' @param sample_rate sample rate
#' @param lb filter lower bound of the frequencies to remove
#' @param ub filter upper bound of the frequencies to remove;
#' shares the same length as \code{lb}
#' @param domain \code{1} if the input signal is in the
#' time domain, \code{0} if it is in the frequency domain
#' @returns filtered signal in time domain (real numerical
#' vector)
#' @examples
#'
#'
#' time <- seq(0, 3, 0.005)
#' s <- sin(120 * pi * time) + rnorm(length(time))
#'
#' # Welch periodogram shows a peak at 60Hz
#' pwelch(s, 200, plot = 1, log = "y")
#'
#' # notch filter to remove 60Hz
#' s1 <- notch_filter(s, 200, lb = 59, ub = 61)
#' pwelch(s1, 200, plot = 2, log = "y", col = "red")
#'
#'
#' @export
notch_filter <- function(
  s, sample_rate, lb = c(59, 118, 178),
  ub = c(61, 122, 182), domain = 1) {

  stopifnot2(length(lb) == length(ub), msg = "Inconsistent length: `lb` vs `ub`")

  max_freq <- sample_rate / 2
  n <- length(s)
  df <- 2 * max_freq / length(s)
  centre_freq <- (lb + ub) / 2
  filter_width <- (-lb + ub)

  x <- seq(0, max_freq, by = df)
  get_kernel <- function(ii) {
    lb <- lb[ii]
    ub <- ub[ii]
    centre_freq <- centre_freq[ii]
    filter_width <- filter_width[ii]
    gauss <- exp(-(centre_freq - x)^2 * 10)
    cnt_gauss <- round(centre_freq / df)
    flat_padd <- 0 # flat padding at the max value of the gaussian
    padd_left <- floor(flat_padd / 2)
    padd_right <- ceiling(flat_padd / 2)

    gauss_left <- gauss[(padd_left + 1):cnt_gauss]
    gauss_right <- gauss[-((padd_left + 1):cnt_gauss)]

    our_wind <- 1 - c(gauss_left, rep(0, flat_padd), gauss_right)
    n_r <- length(our_wind)
    if (n %% 2 == 0) {
      n_r <- n_r - 1
      our_wind <- our_wind[c(1:n_r, n_r:1)]
    } else {
      our_wind <- our_wind[c(1:n_r, (n_r - 1):1)]
    }
    our_wind
  }

  kernel <- get_kernel(1)
  for (ii in seq_along(lb)) {
    if (ii > 1) {
      kernel <- kernel * get_kernel(ii)
    }
  }

  if (domain == 1) {
    s <- fftw_r2c(s) / n
  }

  fftw_c2r(s * kernel)

}

#' @title Fill a volume cube based on water-tight surface
#' @author Zhengjia Wang
#' @description Create a cube volume (\code{256} 'voxels' on each margin), fill
#' in the 'voxels' that are inside of the surface.
#'
#' @param surface a surface mesh; accepted classes include \code{'mesh3d'}
#' (from \code{rgl}), \code{'fs.surface'} (from \code{freesurferformats}),
#' \code{'ieegio_surface'} (from \pkg{ieegio}), or a bare list containing a
#' \code{vb} vertex matrix; see \code{\link{ensure_mesh3d}} for details and
#' for how \code{'ieegio_surface'} inputs are coerced
#' @param inflate amount of 'voxels' to inflate on the final result; must be
#' a non-negative integer. A zero \code{inflate} value means the resulting
#' volume is tightly close to the surface
#' @param close_radius radius (in 'voxels' along the finest 'IJK' axis; see
#' \code{IJK2RAS}) of the morphological closing (dilate-then-erode) operation
#' used to connect the embedded surface into a water-tight shell \emph{and} to
#' remove small topological handles/tunnels (for example a hole drilled
#' through an otherwise solid structure) before the surface is filled; default
#' is \code{NULL}, which derives a radius from the surface geometry (capped at
#' \code{15} 'voxels'). Handles/tunnels wider than roughly twice this radius
#' (in physical, \verb{RAS} space) will not be closed, so when the input
#' surface is known to contain larger topological defects, set this to a
#' larger number explicitly. Larger radii cost more time and memory (the
#' closing operates on a dense \code{resolution^3} volume) and - because the
#' operation is performed in a fixed-size volume - must leave enough margin
#' that the dilated shell does not reach the volume boundary; radii that are
#' too large for the working cube will be capped with a warning
#' @param resolution number of 'voxels' along each margin of the working
#' cube volume that the surface is embedded into; default is \code{256}.
#' Larger values increase the spatial precision of the fill and the closing
#' operation (at the cost of memory and computation time, which scale with
#' \code{resolution^3}); the value must leave enough room for the surface
#' (transformed into 'IJK' space via \code{IJK2RAS}) plus the closing margin
#' to fit inside the cube
#' @param IJK2RAS volume 'IJK' (zero-indexed coordinate index) to
#' \code{'tkrRAS'} transform, default is automatically determined
#' leave it `NULL` if you don't know how to set it. When the linear part of
#' \code{IJK2RAS} encodes voxel sizes that differ across axes (\code{pixdim},
#' i.e. the three 'IJK' axes do not advance by the same physical distance per
#' 'voxel'), the dilate/erode operations underlying \code{close_radius} and
#' \code{inflate} automatically use a different number of 'voxels' along each
#' axis so that the resulting structuring element stays isotropic in physical
#' (\verb{RAS}) space
#' @param preview whether to preview the results; default is false
#' @param preview_frame integer from 1 to 256 the depth frame used to generate
#' preview.
#'
#' @details This function creates a volume (256 on each margin) and fill in
#' the volume from a surface mesh. The surface vertex points will be embedded
#' into the volume first. These points may not be connected together, hence
#' for each 'voxel', a cube patch will be applied to grow the volume. Then,
#' the volume will be bucket-filled from a corner, forming a negated mask of
#' "outside-of-surface" area. The inverted bucket-filled volume is then shrunk
#' so the mask boundary tightly fits the surface
#'
#'
#' @returns A list containing the filled volume and parameters used to generate
#' the volume
#'
#' @inheritSection ensure_mesh3d Coercing Surface Inputs
#'
#' @seealso \code{\link{ensure_mesh3d}}
#'
#' @examples
#'
#' \donttest{
#'
#' # takes > 5s to run example
#'
#' # Generate a sphere
#' surface <- vcg_sphere()
#' surface$vb[1:3, ] <- surface$vb[1:3, ] * 50
#'
#' fill_surface(surface, preview = TRUE)
#'
#' }
#'
#' @name fill_surface
NULL

#' @title Coerce a surface object to a \code{'mesh3d'} mesh
#' @description
#' Internal helper used throughout \pkg{ravetools} to accept a variety of
#' surface representations and return a list of class \code{'mesh3d'} (the
#' format used by the \code{rgl} package). Most mesh-consuming functions in
#' this package call \code{ensure_mesh3d} on their surface arguments, so the
#' coercion rules described here apply to all of them.
#'
#' @param surface a surface object. One of the following:
#' \describe{
#'   \item{\code{'mesh3d'}}{returned unchanged.}
#'   \item{\code{'fs.surface'}}{a surface in the format produced by
#'     \code{freesurferformats::read.fs.surface}. Vertices and faces are
#'     copied into a \code{'mesh3d'} list; zero-indexed faces are bumped
#'     by 1.}
#'   \item{\code{'ieegio_surface'}}{a surface object produced by
#'     \pkg{ieegio} (e.g. from \code{read_surface} or
#'     \code{volume_to_surface}). See the section below for important
#'     behavior changes.}
#'   \item{a bare \code{list}}{must contain a numeric \code{vb} matrix with
#'     3 or 4 rows; the list is reclassified as \code{'mesh3d'} with no
#'     other modification.}
#' }
#' Any other input triggers an error.
#'
#' @returns An object of class \code{'mesh3d'} with at least the \code{vb}
#' (vertex) component and, when face information is available, an \code{it}
#' (triangle index) component.
#'
#' @section Coercing Surface Inputs:
#' The surface objects are converted to \code{'mesh3d'} object before
#' applying further calculations.
#'
#' When \code{surface} is a surface \pkg{ieegio} object, the returned
#' \code{mesh3d$vb} contains vertices that have been left-multiplied by
#' \code{surface$geometry$transforms[[1]]} (the first transform stored in the
#' geometry, typically the \code{ScannerAnat} or voxel-to-world transform).
#'
#' \strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
#' returned the raw \code{surface$geometry$vertices} without applying any
#' transform, so downstream code often multiplied by
#' \code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
#' working in world space. Such code will now \emph{double}
#' apply the transform and produce incorrect coordinates. If you previously
#' applied a transform from \code{surface$geometry$transforms} by hand after
#' calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
#' remove that manual step.
#'
#' Surfaces with an empty or missing \code{geometry$transforms} list (for
#' example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
#' which stores an identity transform) are unaffected.
#'
#' If \code{geometry$transforms} contains multiple transforms targeting
#' different coordinate spaces, only the first one is used. Callers that need
#' a specific target space should select and apply that transform themselves
#' before calling \pkg{ravetools} mesh functions.
#'
#' @examples
#'
#' # mesh3d input is returned unchanged in shape
#' sphere <- vcg_sphere()
#' m <- ensure_mesh3d(sphere)
#' identical(m$vb, sphere$vb)
#'
#' # A bare list with a `vb` slot is reclassified to mesh3d
#' bare <- list(vb = rbind(matrix(rnorm(30), nrow = 3), 1))
#' m <- ensure_mesh3d(bare)
#' inherits(m, "mesh3d")
#'
#' @export
ensure_mesh3d <- function(surface) {
  # check if surface is read from `freesurferformats`
  if (inherits(surface, "fs.surface")) {
    # surface <- freesurferformats::fs.surface.to.tmesh3d( surface )
    face <- t(surface$faces)
    if ( min(face, na.rm = TRUE) == 0 ) {
      face <- face + 1
    }
    surface <- structure(list(
      vb = rbind(t(surface$vertices), 1),
      it = face), class = c("ravetools_mesh3d", "mesh3d"))
  } else if (inherits(surface, "ieegio_surface")) {
    if (!inherits(surface, "ieegio_surface_contains_geometry")) {
      stop("Surface object has no geometry")
    }
    if (isTRUE(surface$sparse) || !length(surface$sparse_node_index)) {
      vertices <- surface$geometry$vertices
    } else {
      node_index <- surface$sparse_node_index
      vertices <- array(0.0, c(3, max(node_index)))
      vertices[1:3, ] <- surface$geometry$vertices[1:3, ]
      vertices <- rbind(vertices, 1)
    }

    # Apply transforms
    if (length(surface$geometry$transforms)) {
      transform <- surface$geometry$transforms[[1]]
      vertices <- transform %*% vertices
    }

    face <- surface$geometry$faces
    if (length(face)) {
      face_start <- surface$geometry$face_start
      if (length(face_start) == 1 && !isTRUE(face_start == 1)) {
        face <- face - face_start + 1L
      }
      storage.mode(face) <- "integer"

      surface <- structure(list(
        vb = vertices,
        it = face), class = c("ravetools_mesh3d", "mesh3d"))
    } else {
      surface <- structure(list(
        vb = vertices), class = c("ravetools_mesh3d", "mesh3d"))
    }
  }
  if (!inherits(surface, "mesh3d")) {
    # check if surface has names
    if (
      is.list(surface) &&
      "vb" %in% names(surface) &&
      is.matrix(surface$vb) &&
      nrow(surface$vb) %in% c(3L, 4L)
    ) {
      surface <- structure(surface, class = c("ravetools_mesh3d", "mesh3d"))
    } else {
      stop("ravetools:::ensure_mesh3d: `surface` must be a mesh3d object (package rgl) or a fs.surface object (package freesurferformats)")
    }
  }
  if (!inherits(surface, "ravetools_mesh3d")) {
    class(surface) <- c("ravetools_mesh3d", class(surface))
  }
  return(surface)
}

#' @rdname fill_surface
#' @export
fill_surface <- function(surface, inflate = 0, close_radius = NULL,
                         resolution = 256L, IJK2RAS = NULL,
                         preview = FALSE, preview_frame = 128) {
  # DIPSAUS DEBUG START
  # surface <- freesurferformats::read.fs.surface('~/Dropbox (PENN Neurotrauma)/RAVE/Samples/raw/PAV010/rave-imaging/fs/surf/rh.pial')
  # IJK2RAS <- NULL
  # verbose <- TRUE
  # preview <- TRUE
  # output_format <- "rgl"
  # inflate <- 3

  inflate <- as.integer(inflate)
  if (inflate < 0) { inflate <- 0 }

  # setup
  preview2D <- function(expr) {
    expr <- substitute(expr)
    if (interactive() && preview) {
      eval(expr)
    }
  }

  # Initialize the parameters
  resolution <- as.integer(resolution)
  if (length(resolution) != 1L || is.na(resolution) || resolution < 16L) {
    stop("`fill_surface`: `resolution` must be a single integer no less than 16")
  }
  if (length(IJK2RAS) != 16L) {
    IJK2RAS <- matrix(
      nrow = 4, byrow = TRUE,
      c(-1, 0, 0, resolution / 2,
        0, 0, 1, -resolution / 2,
        0, -1, 0, resolution / 2,
        0, 0, 0, 1))
  }

  # Voxel size ('pixdim') along each 'IJK' axis (in the units of `IJK2RAS`,
  # typically millimeters): the magnitude of each column of the linear part of
  # `IJK2RAS` is the physical distance spanned by one 'voxel' step along that
  # axis. The default `IJK2RAS` above is an axis-permuted identity (isotropic,
  # pixdim = c(1, 1, 1)); user-supplied transforms (e.g. from non-conformed
  # volumes) may be anisotropic.
  pixdim <- sqrt(colSums(IJK2RAS[1:3, 1:3, drop = FALSE] ^ 2))

  # Convert an isotropic radius in physical ('RAS') space - expressed as a
  # 'voxel' count along the finest (smallest-pixdim) axis - into a per-axis
  # 'voxel' count, so `grow_volume`'s box structuring element ends up
  # isotropic in physical space even when `pixdim` is anisotropic: an axis
  # with larger 'voxels' needs proportionally fewer of them to span the same
  # physical distance.
  to_voxel_radius <- function(radius) {
    as.integer(round(radius * min(pixdim) / pixdim))
  }

  # Make sure the surface is a `mesh3d` object
  surface_orig <- ensure_mesh3d(surface)

  # Transform surface to IJK voxel space so can be fitted into a volume
  # Remember IJK starts from 0
  surface <- surface_orig
  surface$vb <- solve(IJK2RAS) %*% surface$vb

  # Largest per-axis radius the dilated shell can use without reaching the
  # volume boundary (the bucket-fill corner must remain outside the dilated
  # shell)
  vertex_range <- apply(surface$vb, 1, range)
  safe_radius_per_axis <- floor(pmin(
    vertex_range[1, 1:3],
    resolution - vertex_range[2, 1:3]
  ) - 2)
  safe_radius <- min(safe_radius_per_axis)

  if (length(close_radius)) {
    close_radius <- as.numeric(close_radius)
    if (length(close_radius) != 1L || is.na(close_radius) || close_radius < 0) {
      stop("`fill_surface`: `close_radius` must be a single non-negative number")
    }
  } else {
    # Previous default behavior: derive a radius from the surface geometry,
    # capped at 15 'voxels' (along the finest axis)
    close_radius <- min(safe_radius, 15)
  }

  close_radius_voxels <- to_voxel_radius(close_radius)
  if (any(close_radius_voxels > safe_radius_per_axis)) {
    close_radius_voxels <- pmin(close_radius_voxels, safe_radius_per_axis)
    warning(sprintf(
      paste("`fill_surface`: requested `close_radius` (%g) is too large for",
            "this surface's position in the working %d^3 volume along at",
            "least one axis and has been capped at c(%s) 'voxels' (per I/J/K",
            "axis) to avoid the dilated shell reaching the volume boundary;",
            "topological handles/tunnels wider than roughly twice this",
            "radius (in physical space) may not be closed"),
      close_radius, resolution, paste(close_radius_voxels, collapse = ", ")
    ))
  }

  # Creating a volume
  volume <- array(0L, dim = rep(resolution, 3))

  # embed the surface in volume space
  surface_index <- round(surface$vb[c(1, 2, 3), ])
  surface_index <- surface_index[1, ] + surface_index[2, ] * resolution +
    surface_index[3, ] * (resolution^2) + 1
  volume[surface_index] <- 1L

  # Morphological closing (dilate by `close_radius`, then erode back): connects
  # the embedded surface points into a water-tight shell *and* bridges/closes
  # any topological handles or tunnels no wider than ~2x `close_radius` (in
  # physical space). The structuring element is `close_radius_voxels`, a
  # per-axis 'voxel' count that compensates for anisotropic `pixdim`.
  volume_grew <- grow_volume(
    volume,
    x = close_radius_voxels[[1]],
    y = close_radius_voxels[[2]],
    z = close_radius_voxels[[3]]
  )

  # bucket-fill from the corner, bucketFillVolume is in-place
  volume_filled <- bucketFillVolume(volume_grew + 0L, x = 1, y = 1, z = 1, fill = 1L)

  # Fill the voxels within the surface and shrink back by `close_radius`,
  # leaving a net `inflate` (also pixdim-corrected) relative to the surface
  shrink_amount_voxels <- close_radius_voxels - to_voxel_radius(inflate)
  volume2 <- 1L - grow_volume(
    volume_filled - volume_grew,
    x = shrink_amount_voxels[[1]],
    y = shrink_amount_voxels[[2]],
    z = shrink_amount_voxels[[3]]
  )

  # preview
  preview2D({
    oldPar <- graphics::par(c("mfrow", "mar"))
    graphics::par(mfrow = c(1, 3),
                  mar = c(3.1, 0.1, 3.1, 0.1))
    on.exit({ do.call(graphics::par, oldPar) })

    frame <- as.integer(preview_frame)
    if (is.na(frame) || frame <= 1 || frame > resolution) {
      frame <- ceiling(resolution / 2)
    }

    image(volume[, , frame], axes = FALSE, main = "1. Initial surface embed", asp = 1)

    image(volume_grew[, , frame] + volume[, , frame], axes =  FALSE, main = "2. Grow voxels", asp = 1)

    image(volume2[, , frame] + volume[, , frame], axes = FALSE,  main = "3. Bucket-fill+shrink", asp = 1)

  })

  # free up memory ~ 500MB
  # rm(volume, volume_shrunk, volume_filled, volume_grew)

  return(list(
    volume = volume2,
    IJK2RAS = IJK2RAS,
    resolution = resolution,
    fill = close_radius,
    close_radius = close_radius,
    close_radius_voxels = close_radius_voxels,
    pixdim = pixdim,
    inflate = inflate
  ))

}

# wrap phase angles
unwrap <- function(x, tol = pi) {
  if (is.vector(x)) {
    x <- as.matrix(x, ncol = 1)
    vec <- TRUE
  } else {
    vec <- FALSE
  }
  nr <- nrow(x)
  nc <- ncol(x)
  if (!is.numeric(x)) {
    stop("x must be a numeric matrix or vector")
  }
  tol <- abs(tol)
  y <- x
  if (nr > 1) {
    rng <- 2 * pi
    for (col in seq_len(nc)) {
      valid <- is.finite(x[, col])
      d <- diff(x[valid, col])
      p <- round(abs(d) / rng) * rng * (((d > tol) > 0) - ((d < -tol) > 0))
      r <- cumsum(p)
      y[valid, col] <- x[valid, col] - c(0, r)
    }
  }
  if (vec) {
    y <- as.vector(y)
  }
  y
}

#' Frequency response of digital filter
#' @description
#' Compute the z-plane frequency response of an \code{ARMA} model.
#' @param b the moving-average coefficients of an \code{ARMA} model
#' @param a the auto-regressive coefficients of an \code{ARMA} filter; default
#' is \code{1}
#' @param fs sampling frequency in \code{Hz}
#' @param n number of points at which to evaluate the frequency response;
#' default is \code{512}
#' @param whole whether to evaluate beyond \code{Nyquist} frequency; default
#' is false
#' @param ... ignored
#' @returns A list of frequencies and corresponding responses in complex vector
#' @export
freqz2 <- function(b, a = 1, fs = 2 * pi, n = 512, whole = FALSE, ...) {
  freqz_internal <- asNamespace("gsignal")[["freqz.default"]]
  hw <- freqz_internal(unclass(b), a = unclass(a), fs = fs, n = n, whole = whole, ...)

  hw$f <- hw$w
  hw$srate <- fs
  class(hw) <- c("ravetools-freqz2", "ravetools-printable")
  hw
}

#' @export
`summary.ravetools-freqz2` <- function(object, ..., cutoffs = c(-3, -6, -12)) {
  nm <- deparse(substitute(object))
  h <- object$h
  w <- object$w
  rw <- c(0, object$srate / 2)
  mag <- 20 * log10(abs(h))
  phase <- unwrap(Arg(h))
  rp <- range(phase, na.rm = TRUE)

  isfinite <- is.finite(mag)
  if (any(isfinite)) {
    idx <- which.max(mag[isfinite])
    idx <- which(isfinite)[[idx[[1]]]]
    peak <- list(which = idx, freq = object$w[[idx]], mag = mag[[idx]])
  } else {
    peak <- list(which = NA, freq = NA, mag = NA)
  }

  cutoffs <- as.numeric(cutoffs)
  cutoffs <- cutoffs[is.finite(cutoffs)]

  cutoffs <- lapply(cutoffs, function(cutoff) {
    # cutoff <- -3
    idx <- which(diff(as.integer(isfinite & mag > cutoff)) != 0)
    if (!length(idx)) { return(NULL) }

    m1 <- mag[idx]
    m2 <- mag[idx + 1]

    a <- ifelse(m1 == m2, 1, (cutoff - m2) / (m1 - m2))
    freq <- a * w[idx] + (1 - a) * w[idx + 1]


    list(
      cutoff = cutoff,
      frequency = freq
    )
  })

  cutoffs <- cutoffs[!vapply(cutoffs, is.null, FALSE)]

  structure(
    list(
      name = nm,
      unit = object$u,
      frequency_range = rw,
      phase_range = rp,
      peak = peak,
      cutoffs = cutoffs
    ),
    class = c("summary.ravetools-freqz2", "ravetools-printable", "list")
  )
}

#' @export
`format.ravetools-freqz2` <- function(x, ..., name = deparse(substitute(x))) {
  force(name)
  res <- summary(x, ...)
  res$name <- name
  format(res)
}

#' @export
`format.summary.ravetools-freqz2` <- function(x, ...) {

  s <- c(
    sprintf("<RAVE filter freqz summary> %s", paste(x$name, collapse = "")),
    sprintf("Frequency ranges: %.5g to %.5g %s", x$frequency_range[[1]], x$frequency_range[[2]], x$unit)
  )
  if (!is.na(x$peak[[1]])) {
    s <- c(
      s,
      sprintf("Peak magnitude: %.2f dB at frequency %.5g %s",
              x$peak$mag, x$peak$freq, x$unit)
    )
  }

  s <- c(s, unlist(lapply(x$cutoffs, function(item) {
    # item <- x$cutoffs[[1]]
    if (length(item$frequency) > 4) {
      re <- c(sprintf("%.4g %s", item$frequency[1:4], x$unit), "...")
    } else {
      re <- sprintf("%.4g %s", item$frequency, x$unit)
    }

    sprintf("  %4.4g dB cutoff at %s", item$cutoff, paste(re, collapse = ", "))
  })))

  rpd <- x$phase_range * 360 / (2 * pi)
  s <- c(
    s,
    sprintf("Phase ranging from %.4g to %.4g rad (%.3f to %.3f degrees)",
            x$phase_range[[1]], x$phase_range[[2]], rpd[[1]], rpd[[2]])
  )
  s
}

#' @export
`plot.ravetools-freqz2` <- function(
    x, ..., cutoffs = c(-3, -6, -12), vlines = NULL,
    ylim = NULL, xlim = NULL, color_palette = NULL,
    add = FALSE, which = c(1, 2, 3), draw_legend = TRUE
) {

  # DIPSAUS DEBUG START
  # list2env(list(fs = 500, n = 6000, whole = FALSE), .GlobalEnv)
  # b <- ravetools::fir1(160, c(1 / 250, 70 / 250), "pass")$b
  # a <- 1
  # xlim = NULL
  # vlines <- NULL
  # ylim <- NULL
  # cutoffs = c(-3, -6, -12)
  # color_palette <- NULL
  # which = c(1,2,3)
  # add = FALSE
  # x <- freqz2(b = b, a = a, fs = fs, n = n, whole = whole)
  # draw_legend <- TRUE


  which <- which[which %in% c(1, 2, 3)]
  if (!length(which)) {
    which <- c(1, 2, 3)
  }

  mag <- 20 * log10(abs(x$h))
  argh <- Arg(x$h)

  isfinite <- is.finite(mag)
  if (any(isfinite)) {
    idx <- which.max(mag[isfinite])
    idx <- which(isfinite)[[idx[[1]]]]
    maxmag <- mag[[idx]]
  } else {
    idx <- NULL
    maxmag <- 1
  }

  phase <- unwrap(argh)

  w <- x$w
  if (!length(cutoffs)) { cutoffs <- -3 }
  smry <- `summary.ravetools-freqz2`(x, cutoffs = cutoffs)
  smry_str <- format(smry)

  if (length(smry$cutoffs)) {
    if (!length(color_palette)) {
      color_palette <- grDevices::adjustcolor(seq_along(smry$cutoffs) + 1)
    }
    if (length(color_palette) < length(smry$cutoffs)) {
      color_palette <- c(grDevices::adjustcolor(color_palette), grDevices::adjustcolor(seq_along(smry$cutoffs) + 1))
    }
  }


  legend_text <- NULL
  legend_col <- NULL
  legend_lty <- NULL
  if (length(vlines)) {
    legend_text <- "Filter frequency"
    legend_col <- graphics::par("fg")
    legend_lty <- 3
  }

  if (length(smry$cutoffs)) {
    cutoffs_drawn <- vapply(smry$cutoffs, `[[`, 0, "cutoff")
    legend_text <- c(legend_text, sprintf("Cutoff %.2f dB", cutoffs_drawn))
    legend_col <- c(legend_col, color_palette[seq_along(cutoffs_drawn)])
    legend_lty <- c(legend_lty, rep(2, length(cutoffs_drawn)))
    has_legend <- TRUE
  } else {
    has_legend <- FALSE
    draw_legend <- FALSE
  }

  if (!add) {
    if ( !isFALSE(draw_legend) ) {
      graphics::layout(
        mat = matrix(c(seq_along(which) + 1, 1), ncol = 1),
        heights = c(rep(1, length(which)), graphics::lcm(3))
      )
    } else {
      graphics::layout(
        mat = matrix(c(seq_along(which)), ncol = 1),
        heights = c(rep(1, length(which)), graphics::lcm(3))
      )
    }
    oldpar <- graphics::par(c("mfrow", "mar"))
    # mfrow = c(length(which), 1), mar = c(4, 4, 1.5, 1))
    on.exit({
      do.call(graphics::par, oldpar)
    })
  }

  # Legend and summary text
  if ( !isFALSE(draw_legend) ) {
    if (!identical(draw_legend, "add")) {
      graphics::par(mar = c(0.1, 0.1, 0.1, 0.1))
      plot(c(0, 1), c(0, 1), type = "n", axes = FALSE, xlab = "", ylab = "", main = "")
    }

    graphics::legend(x = 0, y = 0.5, xjust = 0, yjust = 0.5,
                     smry_str[-1],
                     bty = "n")

    if ( has_legend ) {
      graphics::legend(x = 0.65, y = 0.5, xjust = 0, yjust = 0.5,
                       legend_text,
                       col = legend_col,
                       lty = legend_lty,
                       bty = "n")
    }

  }

  graphics::par(mar = c(2.1, 3.6, 1.5, 1))

  if (!length(xlim)) {
    xlim <- range(pretty(w))
  }
  if (!length(ylim)) {
    ylim <- range(pretty(mag))
  }

  if (1 %in% which) {
    ylim2 <- c(min(cutoffs) - 1, maxmag)
    graphics::plot(range(pretty(w)), ylim2, ...,
                   xlim = xlim, type = "n", xlab = "", ylab = "", main = "", axes = FALSE)
    graphics::lines(w, mag)
    graphics::axis(side = 1, at = pretty(xlim))
    graphics::axis(side = 2, at = pretty(ylim2), las = 1)
    lapply(seq_along(smry$cutoffs), function(ii) {
      # item <- smry$cutoffs[[1]]
      item <- smry$cutoffs[[ii]]
      if (length(item$frequency) > 4) {
        item$frequency <- item$frequency[1:4]
      }
      graphics::abline(h = item$cutoff, v = item$frequency, lty = 2, col = color_palette[[ii]])
    })
    if (length(vlines)) {
      graphics::abline(v = vlines, lty = 3)
    }
    graphics::title("Pass band (dB)")
  }

  if (2 %in% which) {
    graphics::plot(range(pretty(w)), range(mag, na.rm = TRUE), ylim = ylim, ...,
                   type = "n", xlab = "", ylab = "", main = "", axes = FALSE)
    graphics::lines(w, mag)
    graphics::axis(side = 1, at = pretty(w))
    graphics::axis(side = 2, at = pretty(ylim), las = 1)

    lapply(seq_along(smry$cutoffs), function(ii) {
      # item <- smry$cutoffs[[1]]
      item <- smry$cutoffs[[ii]]
      if (length(item$frequency) > 4) {
        item$frequency <- item$frequency[1:4]
      }
      graphics::abline(h = item$cutoff, v = item$frequency, lty = 2, col = color_palette[[ii]])
    })
    if (length(vlines)) {
      graphics::abline(v = vlines, lty = 3)
    }
    graphics::title("Stop band (dB)")
  }

  if (3 %in% which) {
    d <- phase * 360 / (2 * pi)
    xlim <- range(xlim, na.rm = TRUE)
    if (length(xlim)) {
      sel <- w >= xlim[[1]] & w < xlim[[2]]
      ylim <- range(d[sel], na.rm = TRUE)
    }
    xlim <- range(pretty(xlim))
    ylim <- range(pretty(ylim))

    graphics::plot(xlim, ylim, ...,
                   type = "n", xlab = "", ylab = "", main = "", axes = FALSE)
    graphics::axis(side = 1, at = pretty(xlim))
    graphics::axis(side = 2, at = pretty(ylim), las = 1)
    graphics::lines(w, d)
    lapply(seq_along(smry$cutoffs), function(ii) {
      # item <- smry$cutoffs[[1]]
      item <- smry$cutoffs[[ii]]
      if (length(item$frequency) > 4) {
        item$frequency <- item$frequency[1:4]
      }
      graphics::abline(v = item$frequency, lty = 2, col = color_palette[[ii]])
    })
    if (length(vlines)) {
      graphics::abline(v = vlines, lty = 3)
    }
    graphics::title("Phase (degrees)")
    graphics::mtext("Frequency", side = 1, outer = FALSE, line = 1.6, cex = 0.8)
  }
  invisible(smry)
}

sample_signal <- function(n) {
  a0 <- c(0, 5, 30, 1, 0, -30, 110, -100, 0, 14, 25, 2, 0, 0, 0)
  d0 <- c(0, 2, 6, 10, 13, 14, 16, 18, 19, 28, 30, 34, 36, 39, 44)
  a <- a0 / max(a0)
  d <- round(d0 * n / d0[15])
  d[15] <- n

  x <- list()
  for (i in seq_len(14)) {
    m <- seq(d[i], d[i + 1] - 1)
    slope <- (a[i + 1] - a[i]) / (d[i + 1] - d[i])
    x[[length(x) + 1]] <- a[i] + slope * (m - d[i])
  }
  unlist(x)
}

#' Diagnose digital filter
#' @description
#' Generate frequency response plot with sample-data simulation
#' @param b the moving-average coefficients of an \code{ARMA} model
#' @param a the auto-regressive coefficients of an \code{ARMA} filter; default
#' is \code{1}
#' @param fs sampling frequency in \code{Hz}
#' @param n number of points at which to evaluate the frequency response;
#' default is \code{512}
#' @param whole whether to evaluate beyond \code{Nyquist} frequency; default
#' is false
#' @param sample sample signal of length \code{n} for simulation
#' @param vlines additional vertical lines (frequencies) to plot
#' @param xlim frequency limit of frequency response plot; default is
#' \code{"auto"}, can be \code{"full"} or a numeric of length 2
#' @param cutoffs cutoff decibel powers to draw on the frequency plot, also used
#' to calculate the frequency limit when \code{xlim} is \code{"auto"}
#' @returns Nothing
#'
#' @examples
#'
#'
#'
#' library(ravetools)
#'
#' # sample rate
#' srate <- 500
#'
#' # signal length
#' npts <- 1000
#'
#' # band-pass
#' bpass <- c(1, 50)
#'
#' # Nyquist
#' fn <- srate / 2
#' w <- bpass / fn
#'
#' # ---- FIR filter ------------------------------------------------
#' order <- 160
#'
#' # FIR1 is MA filter, a = 1
#' filter <- fir1(order, w, "pass")
#'
#' diagnose_filter(
#'   b = filter$b, a = filter$a, n = npts,
#'   fs = srate, vlines = bpass
#' )
#'
#' # ---- Butter filter --------------------------------------------
#' filter <- butter(3, w, "pass")
#'
#' diagnose_filter(
#'   b = filter$b, a = filter$a, n = npts,
#'   fs = srate, vlines = bpass
#' )
#'
#'
#'
#' @export
diagnose_filter <- function(b, a, fs, n = 512, whole = FALSE,
                            sample = stats::rnorm(n, mean = sample_signal(n), sd = 0.2),
                            vlines = NULL, xlim = "auto",
                            cutoffs = c(-3, -6, -12)) {

  # DIPSAUS DEBUG START
  # list2env(list(fs = 500, n = 6000, whole = FALSE), .GlobalEnv)
  # b <- ravetools::fir1(160, c(1 / 250, 70 / 250), "pass")$b
  # a <- 1
  # sample = sample_signal(n)
  # xlim = "auto"
  # vlines <- c(0, 1, 70)
  # cutoffs = c(-3, -6, -12)

  hw <- freqz2(b = b, a = a, fs = fs, n = n, whole = whole)
  force(sample)
  time <- seq_along(sample) / fs
  # res1 <- filter_signal(b = b, a = a, x = sample)[[1]]
  res2 <- filtfilt(unclass(b), a = a, x = sample)

  h <- hw$h
  w <- hw$w
  mag <- 20 * log10(abs(h))

  cutoffs <- cutoffs[is.finite(cutoffs)]
  if (length(xlim) != 2) {
    if (identical(xlim, "auto")) {
      xlim <- NULL
      if (length(cutoffs)) {
        rg <- range(c(w[mag >= min(cutoffs)], vlines), na.rm = TRUE)
        if (all(is.finite(rg))) {
          rg[[1]] <- floor(rg[[1]])
          rg[[2]] <- ceiling(rg[[2]])
          xlim <- rg
        }
      }
    } else {
      xlim <- NULL
    }
  }

  oldpar <- graphics::par(c("mfrow", "mar", "mgp", "tck", "xaxs", "yaxs", "cex.lab"))
  on.exit({
    do.call(graphics::par, oldpar)
  })

  graphics::layout(
    mat = matrix(
      byrow = FALSE,
      ncol = 2,
      c(
        2, 3, 4, 1,
        5, 6, 7, 1
      )
    ), widths = c(3, 2), heights = c(1, 1, 1, graphics::lcm(3))
  )

  cex <- 1

  # mar = c(2.6, 3.8, 2.1, 0.6) * (0.5 + cex / 2), mgp = cex * c(2, 0.5, 0), tck = -0.02 * cex, xaxs =
  # "i", yaxs = "i",
  graphics::par(
    mar = c(0.1, 3.6, 0.1, 1),
    mgp = cex * c(2, 0.5, 0),
    tck = -0.02 * cex,
    xaxs = "i", yaxs = "i",
    cex.lab = 0.8
  )
  plot(c(0, 1.1), c(0, 1), type = "n", axes = FALSE, xlab = "", ylab = "", main = "")

  graphics::legend(x = 0.9, y = 0.5, xjust = 0, yjust = 0.5, lty = 1, bty = "n",
         legend = c("Sample", "Filtered"), col = c("gray", "orange"))
  plot(hw, xlim = xlim, add = TRUE, vlines = vlines, cutoffs = cutoffs, draw_legend = "add")

  graphics::par(
    mar = c(2.1, 3.6, 1.5, 1),
    mgp = cex * c(2, 0.5, 0),
    tck = -0.02 * cex,
    xaxs = "i",
    yaxs = "i",
    cex.lab = 0.8
  )
  ylim <- range(pretty(c(sample, sample)))
  plot(range(pretty(time)), ylim, type = "n", axes = FALSE, xlab = "", ylab = "", main = "")
  graphics::title("Simulation")

  graphics::lines(time, sample, col = "gray")
  graphics::lines(time, res2, col = "orange")

  graphics::axis(side = 1, at = pretty(time))
  graphics::axis(side = 2, at = pretty(ylim), las = 1)

  graphics::mtext(text = "Time", side = 1, line = 1.6, cex = 0.8)

  # lines(time, sample, lwd = 2)

  pn <- pwelch(sample, fs = fs, window = fs * 2, noverlap = fs)

  # p0 <- pwelch(sample, fs = fs, window = fs * 2, noverlap = fs)

  pr <- pwelch(res2, fs = fs, window = fs * 2, noverlap = fs)

  specs <- pn$spec
  specs <- specs[is.finite(specs) & specs > 0]
  if (length(specs) <= 1) {
    plim <- c(-100, 0)
  } else {
    plim <- 10 * log10( range(specs) )
    plim[[1]] <- plim[[1]] + min(-3, cutoffs, na.rm = TRUE)
  }

  for (plog in c("y", "xy")) {
    plot(pn, add = FALSE, col = "gray", log = plog, mar = c(2.1, 3.6, 1.5, 1), yline = 1.6, grid = FALSE, ylim = plim, ylab = "")
    # plot(p0, add = TRUE, col = "black", log = plog)
    plot(pr, add = TRUE, col = "orange", log = plog)

    if (length(vlines)) {
      if ( grepl("x", plog) ) {
        log_vlines <- log10(vlines[vlines > 0])
        if (0 %in% vlines) {
          log_vlines <- c(log_vlines, 0)
        }
        graphics::abline(v = log_vlines, lty = 3)
      } else {
        graphics::abline(v = vlines, lty = 3)
      }
    }
  }

  invisible()

}

detrend_naive <- function(x, y) {
  if (missing(y)) {
    y <- x
    x <- seq_along(y)
  } else {
    stopifnot2(length(x) == length(y), msg = "x and y must have the same length.")
  }
  n <- length(y)
  b <- (y[n] - y[1]) / (x[n] - x[1])
  a <- y[1] - b * x[1]
  list(Y = y - (a + b * x), a = a, b = b)
}

# x is either a vector or a column-major matrix
postpad <- function(x, n) {
  if (is.matrix(x)) {
    x_len <- nrow(x)
    if (n > x_len) {
      nc <- ncol(x)
      return(rbind(x, array(0, dim = c(n - x_len, nc))))
    } else {
      return(x[seq_len(n), , drop = FALSE])
    }
  } else {
    x_len <- length(x)
    if (n > x_len) {
      return(c(x, rep(0, n - x_len)))
    } else {
      return(x[seq_len(n)])
    }
  }
}

#' Show channel signals with diagnostic plots
#' @description The diagnostic plots include 'Welch Periodogram'
#' (\code{\link{pwelch}}) and histogram (\code{\link[graphics]{hist}})
#' @param s1 the main signal to draw
#' @param s2 the comparing signal to draw; usually \code{s1} after some filters;
#' must be in the same sampling rate with \code{s1}; can be \code{NULL}
#' @param sc decimated \code{s1} to show if \code{srate} is too high; will
#' be automatically generated if \code{NULL}
#' @param srate sampling rate
#' @param name name of \code{s1}, or a vector of two names of \code{s1} and
#' \code{s2} if \code{s2} is provided
#' @param try_compress whether try to compress (decimate) \code{s1} if
#' \code{srate} is too high for performance concerns
#' @param max_freq the maximum frequency to display in 'Welch Periodograms'
#' @param plim the y-axis limit to draw in 'Welch Periodograms'
#' @param window,noverlap see \code{\link{pwelch}}
#' @param cex,lwd,mar,cex.lab,mgp,xaxs,yaxs,tck,... graphical parameters; see
#' \code{\link[graphics]{par}}
#' @param xline,yline distance of axis labels towards ticks
#' @param nclass number of classes to show in histogram
#' (\code{\link[graphics]{hist}})
#' @param main the title of the signal plot
#' @param col colors of \code{s1} and \code{s2}
#' @param which \code{NULL} or integer from 1 to 4; if \code{NULL}, all plots
#' will be displayed; otherwise only the subplot will be displayed
#' @param start_time the starting time of channel (will only be used to draw
#' signals)
#' @param boundary a red boundary to show in channel plot; default is
#' to be automatically determined by \code{std}
#' @param std the standard deviation of the channel signals used to determine
#' \code{boundary}; default is plus-minus 3 standard deviation
#' @returns A list of boundary and y-axis limit used to draw the channel
#' @examples
#' library(ravetools)
#'
#' # Generate 20 second data at 2000 Hz
#' time <- seq(0, 20, by = 1 / 2000)
#' signal <- sin( 120 * pi * time) +
#'   sin(time * 20*pi) +
#'   exp(-time^2) *
#'   cos(time * 10*pi) +
#'   rnorm(length(time))
#'
#' signal2 <- notch_filter(signal, 2000)
#'
#' diagnose_channel(signal, signal2, srate = 2000,
#'                  name = c("Raw", "Filtered"), cex = 1)
#'
#' @export
diagnose_channel <- function(
  s1, s2 = NULL, sc = NULL, srate, name = "", try_compress = TRUE,
  max_freq = 300, window = ceiling(srate * 2), noverlap = window / 2, std = 3,
  which = NULL, main = "Channel Inspection", col = c("black", "red"),
  cex = 1.2, cex.lab = 1,
  lwd = 0.5, plim = NULL, nclass = 100, start_time = 0, boundary = NULL,
  mar = c(3.1, 4.1, 2.1, 0.8) * (0.25 + cex * 0.75) + 0.1,
  # mai = c(0.41, 0.51, 0.4, 0.1) * (0.5 + cex / 2),
  mgp = cex * c(2, 0.5, 0),
  xaxs = "i", yaxs = "i", xline = 1.66 * cex, yline = 2.66 * cex,
  tck = -0.005 * (3 + cex),
  ...) {

  # is sc not specified, and srate is too high, compress s1
  if (try_compress && (is.null(sc) || (srate > 200 && length(s1) / srate > 300))) {
    sratec <- 100
    sc <- s1[round(seq(1, length(s1), by = srate / sratec))]
  } else {
    sc %?<-% s1
    sratec <- srate / length(s1) * length(sc)
  }
  max_freq <- min(max_freq, floor(srate / 2))
  xlim <- c(1, max_freq)

  # Calculate boundary to draw
  if (is.null(boundary)) {
    boundary <- std * stats::sd(s1)
  }
  ylim <- max(abs(s1), boundary)

  # Grid layout

  par_opt <- graphics::par(c("mai", "mar", "mgp", "cex.main", "cex.lab", "cex.axis", "cex.sub"))

  if ( length(which) == 0 || prod(graphics::par("mfrow")) > 1 ) {
    par_opt$cex.lab <- cex.lab * 0.7
  } else {
    par_opt$cex.lab <- cex.lab
  }


  graphics::par(mar = mar, mgp = mgp, cex.lab = par_opt$cex.lab)
  on.exit({graphics::par(par_opt)}, add = TRUE)

  if (length(which) == 0) {
    # grid::grid.newpage()
    lay <- rbind(c(1, 1, 1), c(2, 3, 4))
    graphics::layout(mat = lay)
  }

  # First plot: plot sc directly with col[1]
  if (length(which) == 0 || 1 %in% which) {
    graphics::plot(
      start_time + (seq_along(sc) / sratec), sc, xlab = "", ylab = "",
      main = main, lwd = lwd,
      type = "l", ylim = c(-ylim - 1, ylim + 1), yaxt = "n", col = col[1],
      xaxs = xaxs, yaxs = yaxs, tck = tck, ...,
      cex = cex, cex.main = par_opt$cex.main * cex,
      cex.lab = par_opt$cex.lab * cex,
      cex.axis = par_opt$cex.axis * cex
    )
    graphics::abline(h = c(-1, 1) * boundary, col = "red")
    ticks <- c(-ylim, -boundary, 0, boundary, ylim)
    graphics::axis(2, at = ticks, labels = round(ticks), las = 1, tck = tck,
                   cex = cex, cex.main = par_opt$cex.main * cex,
                   cex.lab = par_opt$cex.lab * cex,
                   cex.axis = par_opt$cex.axis * cex)
    # graphics::title(main = main, cex.main = cex * graphics::par('cex.main'))
    graphics::mtext(side = 2, text = "Voltage", line = yline,
                    cex = par_opt$cex.lab * cex)
    graphics::mtext(side = 1, text = "Time (seconds)", line = xline,
                    cex = par_opt$cex.lab * cex)
  }

  # plot 2, 3 too slow, need to be faster - pwelch periodogram
  if (length(which) == 0 || any(c(2, 3) %in% which)) {
    ps1 <- pwelch(s1, fs = srate, window = window, noverlap = noverlap, plot = FALSE)
    has_s2 <- !is.null(s2)
    if (has_s2) {
      ps2 <- pwelch(s2, fs = srate, window = window, noverlap = noverlap, plot = FALSE)
    } else {
      ps2 <- NULL
    }

    if ( length(which) == 0 || 2 %in% which ) {

      if (has_s2) {
        plot(ps2, add = FALSE, col = col[2], cex = cex, ylim = plim,
             log = "y", xlim = xlim, xaxs = xaxs, yaxs = yaxs, mar = mar,
             xline = xline, yline = yline, main = "Welch Periodogram",
             mgp = mgp, tck = tck)
        plot(ps1, add = TRUE, col = col[1], cex = cex, ylim = plim,
             log = "y", xlim = xlim, xaxs = xaxs, yaxs = yaxs, mar = mar,
             xline = xline, yline = yline, main = "Welch Periodogram",
             mgp = mgp, tck = tck)
        if (length(name) && any(nzchar(name))) {
          graphics::legend("topright", legend = name, col = col, lty = 1, bty = "n", cex = cex)
        }
      } else {
        plot(ps1, add = FALSE, col = col[1], cex = cex, ylim = plim,
             log = "y", xlim = xlim, xaxs = xaxs, yaxs = yaxs, mar = mar,
             xline = xline, yline = yline, main = "Welch Periodogram",
             mgp = mgp, tck = tck)
        if (length(name) && nzchar(name[[1]])) {
          graphics::legend("topright", legend = name[[1]], col = col[[1]], lty = 1, bty = "n", cex = cex)
        }
      }

    }

    if ( length(which) == 0 || 3 %in% which ) {

      if (has_s2) {
        plot(ps2, add = FALSE, col = col[2], cex = cex, ylim = plim,
             log = "xy", xlim = xlim, xaxs = xaxs, yaxs = yaxs, mar = mar,
             xline = xline, yline = yline, main = "Welch Periodogram",
             mgp = mgp, tck = tck)
        plot(ps1, add = TRUE, col = col[1], cex = cex, ylim = plim,
             log = "xy", xlim = xlim, xaxs = xaxs, yaxs = yaxs, mar = mar,
             xline = xline, yline = yline, main = "Welch Periodogram",
             mgp = mgp, tck = tck)
        if (length(name) && any(nzchar(name))) {
          graphics::legend("topright", legend = name, col = col, lty = 1, bty = "n", cex = cex)
        }
      } else {
        plot(ps1, add = FALSE, col = col[1], cex = cex, ylim = plim,
             log = "xy", xlim = xlim, xaxs = xaxs, yaxs = yaxs, mar = mar,
             xline = xline, yline = yline, main = "Welch Periodogram",
             mgp = mgp, tck = tck)
        if (length(name) && nzchar(name[[1]])) {
          graphics::legend("topright", legend = name[[1]], col = col[[1]], lty = 1, bty = "n", cex = cex)
        }
      }

    }

  }


  if (length(which) == 0 || 4 %in% which) {
    # Plot 4:
    graphics::hist(s1, nclass = nclass, probability = FALSE,
                   xlab = "", ylab = "", main = paste0("Histogram ", name[[1]]),
                   tck = tck, las = 1,
                   cex = cex, cex.main = par_opt$cex.main * cex,
                   cex.lab = par_opt$cex.lab * cex,
                   cex.axis = par_opt$cex.axis * cex)
    graphics::mtext(side = 2, text = "Frequency", line = yline,
                    cex = par_opt$cex.lab * cex)
    graphics::mtext(side = 1, text = "Signal Voltage Histogram", line = xline,
                    cex = par_opt$cex.lab * cex)
  }

  return(invisible(list(
    ylim = ylim,
    boundary = boundary
  )))
}

#' @title Smooth a surface mesh
#' @description
#' Smooths a closed triangular surface mesh by repeatedly replacing each
#' vertex with the average of itself and its immediate neighbors
#' (Laplacian/neighbor-averaging smoothing), optionally rescaling the surface
#' back to its original area afterwards.
#'
#' @details
#' For each of \code{npasses} passes, the algorithm:
#' \enumerate{
#'   \item Replaces each vertex position by the include-self mean of itself
#'     and its 1-ring (directly-connected) neighbors, repeated
#'     \code{niterations} times.
#'   \item Recomputes vertex normals and total surface area.
#'   \item Optionally rescales the surface back to its original area (only if
#'     \code{rescale = TRUE}).
#' }
#'
#' @param mesh triangular mesh of class \code{'mesh3d'}.  Must be closed,
#'   manifold, and genus-0 (watertight, no boundary or non-manifold edges,
#'   single connected component) - the same hard precondition as
#'   \code{\link{mris_inflate}} (and for the same reason: an open or
#'   defective mesh introduces a persistent directional bias into the
#'   1-ring averaging).  \code{mris_smooth} will raise an error if such
#'   defects are detected.
#' @param niterations number of include-self 1-ring averaging rounds applied
#'   to the vertex positions in each pass.  Default \code{10}.
#' @param npasses number of outer smoothing passes.  Default \code{1}.
#' @param rescale logical; whether to rescale the surface back to its original
#'   area after each pass.  Default \code{FALSE}.
#' @param verbose logical; print per-pass progress.  Default \code{FALSE}.
#'
#' @returns The smoothed surface as a \code{'mesh3d'} object with \code{vb},
#'   \code{it}, and \code{normals}.
#'
#' @examples
#'
#' if (is_not_cran()) {
#'
#' sphere <- vcg_sphere(sub_division = 4L)
#'
#' # roughen the sphere slightly so smoothing has something to do
#' sphere$vb[1, ] <- sphere$vb[1, ] * (1 + 0.05 * rnorm(ncol(sphere$vb)))
#'
#' # Fix defects
#' sphere <- vcg_fix_defects(sphere)
#'
#' smoothed <- mris_smooth(sphere, niterations = 5L, verbose = TRUE)
#'
#' plot_mesh_polygon(
#'   list(sphere, smoothed),
#'   alpha = c(0.3, 0.5),
#'   col = list("gray", "red"),
#'   main = "Gray: original; red: smoothed"
#' )
#'
#' }
#'
#' @export
mris_smooth <- function(
    mesh,
    niterations = 10L,
    npasses     = 1L,
    rescale     = FALSE,
    verbose     = FALSE
) {
  mesh <- meshintegrity(mesh, facecheck = TRUE)

  vb <- mesh$vb[1:3, , drop = FALSE]
  storage.mode(vb) <- "double"

  it <- mesh$it
  storage.mode(it) <- "integer"

  tmp <- mrisSmooth(
    vb_         = vb,
    it_         = it,
    niterations = as.integer(niterations)[[1L]],
    npasses     = as.integer(npasses)[[1L]],
    rescale     = as.logical(rescale)[[1L]],
    verbose     = as.logical(verbose)[[1L]]
  )

  structure(
    list(
      vb      = rbind(tmp$vb, 1),
      it      = it,
      normals = rbind(tmp$normals, 1)
    ),
    class = c("ravetools_mesh3d", "mesh3d")
  )
}

# core util functions in signal filter

clamp <- function(x, min = NA, max = NA) {
  if (!is.na(min)) {
    x[x < min] <- min
  }
  if (!is.na(max)) {
    x[x > max] <- max
  }
  x
}

# ---- Shape/Window functions --------------------------------------------------
sineint <- function(x) {
  neg <- x < 0
  x[neg] <- -x[neg]

  re <- Im(pracma::expint(1i * x)) + pi / 2
  re[neg] <- -re[neg]
  re[x == 0] <- 0
  re
}

polyval <- function(coef, z)
{
  lz <- length(z)
  if (!lz)
    return(numeric(0))
  n <- length(coef)
  if (!n) {
    z[] <- 0
    return(z)
  }
  if (!(mode(coef) == "numeric") && !(mode(coef) == "complex"))
    stop("Argument 'coef' must be a real or complex vector.")
  d_z <- dim(z)
  dim(z) <- lz
  y <- outer(z, (n - 1):0, "^") %*% coef
  dim(y) <- d_z
  return(y)
}

#' @title Filter window functions
#' @name filter-window
#' @param n number of time-points in window
#' @returns A numeric vector of window with length \code{n}
#' @examples
#'
#' hanning(10)
#' hamming(11)
#' blackmanharris(21)
#'
NULL

#' @rdname filter-window
#' @export
hanning <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    c <- 1
  } else {
    c <- 0.5 - 0.5 * cos(2 * pi * seq(0, n - 1) / (n - 1))
  }
  c
}

#' @rdname filter-window
#' @export
hamming <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    c <- 1
  } else {
    n <- n - 1
    c <- 0.54 - 0.46 * cos(2 * pi * (0:n) / n)
  }
  c
}

#' @rdname filter-window
#' @export
blackman <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    c <- 1
  } else {
    x <- seq.int(0, n - 1) * (2 * pi / (n - 1))
    c <- 0.42 - 0.5 * cos(x) + 0.08 * cos(x * 2)
  }
  c
}


#' @rdname filter-window
#' @export
blackmannuttall <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    c <- 1
  } else {
    x <- seq.int(0, n - 1) * (2 * pi / (n - 1))
    c <- 0.3635819 - 0.4891775 * cos(x) + 0.1365995 * cos(x * 2) - 0.0106411 * cos(x * 3)
  }
  c
}

#' @rdname filter-window
#' @export
blackmanharris <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    c <- 1
  } else {
    x <- seq.int(0, n - 1) * (2 * pi / (n - 1))
    c <- 0.35875 - 0.48829 * cos(x) + 0.14128 * cos(x * 2) - 0.01168 * cos(x * 3)
  }
  c
}

#' @rdname filter-window
#' @export
flattopwin <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    c <- 1
  } else {
    x <- seq.int(0, n - 1) * (2 * pi / (n - 1))
    c <- 0.21557895 - 0.41663158 * cos(x) + 0.277263158 * cos(x * 2) - 0.083578947 * cos(x * 3) + 0.006947368 * cos(x * 4)
  }
  c
}

#' @rdname filter-window
#' @export
bohmanwin <- function(n) {
  if (!(n == round(n) && n > 0)) {
    stop("n has to be an integer > 0")
  }

  if (n == 1) {
    w <- 1
  } else {
    N <- n - 1
    k <- seq(-N / 2, N / 2)
    w <- (1 - 2 * abs(k) / N) * cos(2 * pi * abs(k) / N) + (1 / pi) *
      sin(2 * pi * abs(k) / N)
    w[1] <- w[length(w)] <- 0
  }
  w
}

# ---- fft transform -----------------------------------------------------------
fft <- function(x, inverse = FALSE) {
  if (inverse) {
    fftw_c2r(x)
  } else {
    fftw_r2c(x)
  }
}
ifft <- function(x) {
  fftw_c2r(x) / length(x)
}



# ---- Estimation, validation,  ------------------------------------------------

filtmag_db <- function(b, a, f) {
  # Find filter's magnitude response in decibels at given frequency.
  nb <- length(b)
  na <- length(a)

  sapply(f, function(fi) {
    top <- sum(exp(-1i * seq(0, nb - 1) * pi * fi) * b)
    bot <- sum(exp(-1i * seq(0, na - 1) * pi * fi) * a)

    H <- 20 * log10(abs(top / bot))

    H
  })

}

#' Check filter parameters
#' @description
#' Internally (not exported) used to check the validity of filter inputs.
#' The function only checks if the inputs are ordered correctly, instead of
#' checking the feasibility of the filter.
#'
#' @param type filter type
#' @param w_peak frequency correspond to peak amplitude within 0 to 1, where 1
#' stands for 'Nyquist' frequency. This is the frequency range that users want
#' to "keep". For low-pass and high-pass filters \code{w_peak} is a single
#' number. For band-pass or band-stop filters, \code{w_peak} should be a
#' numeric vector of length 2, with first element being the high-pass frequency
#' and second element low-pass frequency. For band-pass filter, \code{w_peak}
#' is increasing; for band-stop filter, \code{w_peak} is decreasing.
#' @param r_peak expected decibel power (loss) of the peak power
#' @param w_pass pass-band frequency or frequency range, see to \code{w_peak}
#' for format
#' @param r_pass expected band-pass ripple
#' @param w_stop stop-band frequency or frequency range, see to \code{w_peak}
#' for format
#' @param r_stop expected stop-band attenuation
#'
#'
#' @noRd
check_filter_params <- function(
    type = c("low", "high", "pass", "stop"),
    w_peak, w_pass, w_stop,
    r_peak = 0.1, r_pass = 3, r_stop = 60
) {
  type <- match.arg(type)
  switch(
    type,
    "low" = {
      w_stop %?<-% 1
      stopifnot2(
        length(w_peak) == 1 && length(w_pass) == 1 && length(w_stop) == 1,
        msg = "Low-pass frequency cutoff length must be 1"
      )
      stopifnot2(
        0 < w_peak && w_peak <= w_pass && w_pass <= w_stop && w_stop <= 1,
        msg = "Low-pass frequency/nyquist must be within [0,1]. 0 < peak power frequency <= passband frequency <= stopband frequency <= 1"
      )
    },
    "high" = {
      w_stop %?<-% 0
      stopifnot2(
        length(w_peak) == 1 && length(w_pass) == 1 && length(w_stop) == 1,
        msg = "High-pass frequency cutoff length must be 1"
      )
      stopifnot2(
        0 <= w_stop && w_stop <= w_pass && w_pass <= w_peak && w_peak < 1,
        msg = "High-pass frequency/nyquist must be within [0,1]. 0 < stopband frequency <= passband frequency <= peak power frequency <= 1"
      )
    },
    "pass" = {
      w_stop %?<-% c(0, 1)
      stopifnot2(
        length(w_peak) == 2 && length(w_pass) == 2 && length(w_stop) == 2,
        msg = "Band-pass frequency cutoff length must be 2"
      )
      stopifnot2(
        0 <= w_stop[[1]] && w_stop[[1]] <= w_pass[[1]] && w_pass[[1]] <= w_peak[[1]] &&
          w_peak[[1]] <= w_peak[[2]] && w_peak[[2]] <= w_pass[[2]] && w_pass[[2]] <= w_stop[[2]] &&
          w_stop[[2]] <= 1,
        msg = "Band-pass frequency/nyquist must be within [0,1] in non-decreasing order: 0, w_stop[1], w_pass[1], w_peak[1], w_peak[2], w_pass[2], w_stop[2], 1"
      )
    },
    "stop" = {
      w_stop %?<-% c(1, 0)
      stopifnot2(
        length(w_peak) == 2 && length(w_pass) == 2 && length(w_stop) == 2,
        msg = "Band-stop frequency cutoff length must be 2"
      )
      stopifnot2(
        1 >= w_peak[[1]] && w_peak[[1]] >= w_pass[[1]] && w_pass[[1]] >= w_stop[[1]] &&
          w_stop[[1]] >= w_stop[[2]] && w_stop[[2]] >= w_pass[[2]] && w_pass[[2]] >= w_peak[[2]] &&
          w_peak[[2]] >= 0,
        msg = "Band-stop frequency/nyquist must be within [0,1] in non-decreasing order: 0, w_peak[2], w_pass[2], w_stop[2], w_stop[1], w_pass[1], w_peak[1], 1"
      )
    }
  )

  r_peak <- abs(r_peak)
  r_pass <- abs(r_pass)
  r_stop <- abs(r_stop)

  stopifnot2(
    0 <= r_peak && r_peak < r_pass && r_pass < r_stop,
    msg = "Absolute ripple (dB) must be in increasing order: 0 <= r_peak < r_pass < r_stop"
  )

  list(
    type = type,
    w_peak = w_peak,
    w_pass = w_pass,
    w_stop = w_stop,
    r_peak = r_peak,
    r_pass = r_pass,
    r_stop = r_stop
  )
}


#' Computer reciprocal condition number of an 'Arma' filter
#' @description
#' Test whether the filter is numerically stable for \code{\link{filtfilt}}.
#'
#' @param a auto-regression coefficient, numerical vector; the first element
#' must not be zero
#' @returns Reciprocal condition number of matrix \code{z1}, used in
#' \code{\link{filtfilt}}. If the number is less than
#' \code{.Machine$double.eps}, then \code{\link{filtfilt}} will fail.
#' @examples
#'
#'
#' # Butterworth filter with low-pass at 0.1 Hz (order = 4)
#' filter <- butter(4, 0.1, "low")
#'
#' # TRUE
#' rcond_filter_ar(filter$a) > .Machine$double.eps
#'
#' diagnose_filter(filter$b, filter$a, 500)
#'
#' # Bad filter (order is too high)
#' filter <- butter(50, 0.1, "low")
#'
#' rcond_filter_ar(filter$a) > .Machine$double.eps
#'
#' # filtfilt needs to inverse a singular matrix
#' diagnose_filter(filter$b, filter$a, 500)
#'
#' @seealso \code{\link{check_filter}}
#' @export
rcond_filter_ar <- function(a) {
  z1_r <- length(a)
  if (z1_r == 1) { return(1) }

  if (a[[1]] == 0) { stop("ARMA filter AR[1] must not be zero") }
  a <- a / a[[1]]

  if (z1_r == 2) { return( 1 + a[-1] ) }
  z1 <- diag(1, z1_r - 1) - cbind( -a[-1], rbind(diag(1, z1_r - 2), 0) )
  rcon <- abs(rcond(z1))
  rcon
}

#' @title Check 'Arma' filter
#' @param b moving average (\code{MA}) polynomial coefficients.
#' @param a auto-regressive (\code{AR}) polynomial coefficients.
#' @param w normalized frequency, ranging from 0 to 1, where 1 is 'Nyquist'
#' @param r_expected attenuation in decibel of each \code{w}
#' @param fs sample rate, used to infer the frequencies and formatting print
#' message, not used in calculation; leave it blank by default
#' @returns A list of power estimation and the reciprocal condition number
#' of the \code{AR} coefficients.
#' @examples
#'
#'
#' # create a butterworth filter with -3dB (half-power) at [1, 5] Hz
#' # and -60dB stop-band attenuation at [0.5, 6] Hz
#'
#' sample_rate <- 20
#' nyquist <- sample_rate / 2
#'
#' specs <- buttord(
#'   Wp = c(1, 5) / nyquist,
#'   Ws = c(0.5, 6) / nyquist,
#'   Rp = 3,
#'   Rs = 60
#' )
#' filter <- butter(specs)
#'
#' # filter quality is poor because the AR-coefficients
#' # creates singular matrix with unstable inverse,
#' # this will cause `filtfilt` to fail
#' check_filter(
#'   b = filter$b, a = filter$a,
#'
#'   # frequencies (normalized) where power is evaluated
#'   w = c(1, 5, 0.5, 6) / nyquist,
#'
#'   # expected power
#'   r_expected = c(3, 3, 60, 60)
#'
#' )
#'
#'
#' @export
check_filter <- function(b, a, w = NULL, r_expected = NULL, fs = NULL) {

  # r_expected <- abs(r_expected)

  if (length(w)) {
    attenuation <- data.frame(
      frequency = w,
      expected_db = r_expected,
      filtmag_db = filtmag_db(b, a, w)
    )
    attenuation$diff <- attenuation$filtmag_db - attenuation$expected_db
  } else {
    attenuation <- data.frame()
  }

  # atten_passed <- abs(attenuation$diff) < tol
  # if(!all(atten_passed)) {
  #   warn <- c(warn, " * Some frequencies might have unexpected attenuation")
  # }

  rcon <- rcond_filter_ar(a)

  structure(
    list(
      b = b,
      a = a,
      fs = fs,
      attenuation = attenuation,
      reciprocal_condition = rcon
    ),
    class = c("ravetools-test_filter", "ravetools-printable")
  )
}

#' @export
`format.ravetools-test_filter` <- function(x, ...) {
  rcon <- x$reciprocal_condition
  rcond_passed <- rcon > .Machine$double.eps
  attenuation <- x$attenuation
  sample_rate <- x$fs
  if (length(sample_rate) == 1 && is.numeric(sample_rate) && !is.na(sample_rate)) {
    nyquist <- sample_rate / 2
    unit <- "Hz"
  } else {
    nyquist <- 1
    unit <- "xPi rad/s"
  }

  if (!isTRUE(rcon > .Machine$double.eps)) {
    warn <- c("", "WARNING: ", " * Unstable autoregressive (AR) polynomial coefficients")
  } else {
    warn <- NULL
  }
  c(
    "<RAVE filter quality test>",
    "Attenuation: ",
    sprintf("  Freq=%.2g %s, mag=%.4g dB (expected=%.4g dB)",
            attenuation$frequency * nyquist, unit, attenuation$filtmag_db, attenuation$expected_db),
    sprintf("Reciprocal condition number: %.2g %s .Machine$double.eps",
            rcon, ifelse(rcond_passed, ">", "<")),
    warn
  )
}

deparse_svec <- function(nums, connect = "-", concatenate = TRUE, collapse = ",", max_lag = 1) {
  nums <- nums[is.finite(nums)]
  if (length(nums) == 0) {
    return("")
  }
  alag <- seq_len(max(1, max_lag))
  nums <- sort(unique(nums))
  lg <- c(NA, nums)[seq_along(nums)]
  ind <- nums - lg
  ind[1] <- 0
  ind2 <- c(ind[-1], -1)
  re <- apply(cbind(nums[!ind %in% alag], nums[!ind2 %in% alag]), 1, function(x) {
    if (x[1] == x[2]) {
      as.character(x[1])
    } else {
      paste(x, collapse = connect)
    }
  })
  if (concatenate) {
    re <- paste(re, collapse = collapse)
  }
  re
}


parse_svec <- function(text, sep = ",", connect = "-:|", sort = FALSE, unique = TRUE) {
  connect <- unique(unlist(strsplit(connect, "")))
  connect[connect %in% c("|", ":", "~")] <- paste0("\\", connect[connect %in% c("|", ":", "~")])
  if ("-" %in% connect) {
    connect <- c(connect[connect != "-"], "-")
  }
  connect <- paste(connect, collapse = "")

  if (length(text) != 1) {
    text <- paste(text, collapse = sep)
  }


  if (length(text) == 0 || !nzchar(trimws(text))) {
    return(NULL)
  }

  if (is.numeric(text)) {
    if (unique) {
      text <- unique(text)
    }
    if (sort) {
      text <- sort(text)
    }
    return(text)
  }
  s <- unlist(strsplit(text, sep, perl = TRUE))
  s <- trimws(s)
  s <- s[s != ""]

  s <- s[grepl(sprintf("^[0-9\\ %s]+$", connect), s)]

  re <- NULL
  for (ss in s) {
    if (grepl(sprintf("[%s]", connect), ss)) {
      ss <- unlist(strsplit(ss,  sprintf("[%s]", connect), perl = TRUE))
      ss <- trimws(ss)
      ss <- ss[grepl("^[0-9]+$", ss)]
      ss <- as.numeric(ss)
      ss <- ss[!is.na(ss)]
      if (length(ss) >= 2) {
        re <- c(re, (ss[1]:ss[2]))
      }
    } else {
      re <- c(re, as.numeric(ss))
    }
  }

  if (unique) {
    re <- unique(re)
  }

  if (sort) {
    re <- sort(re)
  }

  return(re)
}

#' @title Calculate massive covariance matrix in parallel
#'
#' @description Speed up covariance calculation for large matrices. The
#' default behavior is the same as \code{\link{cov}} (\code{'pearson'},
#' no \code{NA} handling).
#'
#' @param x a numeric vector, matrix or data frame; a matrix is highly
#' recommended to maximize the performance
#' @param y NULL (default) or a vector, matrix or data frame with compatible
#' dimensions to x; the default is equivalent to \code{y = x}
#' @param col_x integers indicating the subset indices (columns) of \code{x} to
#' calculate the covariance, or \code{NULL} to include all the columns; default
#' is \code{NULL}
#' @param col_y integers indicating the subset indices (columns) of \code{y} to
#' calculate the covariance, or \code{NULL} to include all the columns; default
#' is \code{NULL}
#' @param df a scalar indicating the degrees of freedom; default is
#' \code{nrow(x)-1}
#'
#' @returns A covariance matrix of \code{x} and \code{y}. Note that there is no
#' \code{NA} handling. Any missing values will lead to \code{NA} in the
#' resulting covariance matrices.
#'
#' @examples
#'
#' # Set ncores = 2 to comply to CRAN policy. Please don't run this line
#' ravetools_threads(n_threads = 2L)
#'
#' x <- matrix(rnorm(400), nrow = 100)
#'
#' # Call `cov(x)` to compare
#' fast_cov(x)
#'
#' # Calculate covariance of subsets
#' fast_cov(x, col_x = 1, col_y = 1:2)
#'
#' \donttest{
#'
#' # Speed comparison, better to use multiple cores (4, 8, or more)
#' # to show the differences.
#'
#' ravetools_threads(n_threads = -1)
#' x <- matrix(rnorm(100000), nrow = 1000)
#' microbenchmark::microbenchmark(
#'   fast_cov = {
#'     fast_cov(x, col_x = 1:50, col_y = 51:100)
#'   },
#'   cov = {
#'     cov(x[,1:50], x[,51:100])
#'   },
#'   unit = 'ms', times = 10
#' )
#'
#' }
#'
#'
#' @export
fast_cov <- function(x, y = NULL, col_x = NULL, col_y = NULL, df = NA) {
  if (!is.matrix(x)) {
    x <- as.matrix(x)
  }
  stopifnot(is.matrix(x) && (is.numeric(x) || is.logical(x)))
  if (is.null(y)) {
    y <- x
  } else {
    if (!is.matrix(y)) {
      y <- as.matrix(y)
    }
    stopifnot(is.matrix(x) && (is.numeric(x) || is.logical(x)))
  }
  if (!is.null(col_x)) {
    col_x <- as.integer(col_x)
  }
  if (!is.null(col_y)) {
    col_y <- as.integer(col_y)
  }
  if (!is.finite(df) || df <= 0) {
    df <- nrow(x) - 1
  }
  re <- fastcov(x1 = x, x2 = y, col1 = col_x, col2 = col_y, df = df)
  if (inherits(re, "ravetools_error")) {
    stop(re)
  }
  re
}

test_that("fast_cov", {
  a <- matrix(rnorm(1000), nrow = 100)
  b <- matrix(1:5000, nrow = 100)
  eps <- .Machine$double.eps^0.5

  # default
  y <- cov(a)
  z <- fast_cov(a)
  expect_equal(z, y, tolerance = eps)

  y <- cov(a, b)
  z <- fast_cov(a, b)
  expect_equal(z, y, tolerance = eps)

  # Selected columns
  col1 <- sample(c(NA, -3:13), 20, replace = TRUE)
  col1_ <- col1
  col1[is.na(col1) | col1 < 1 | col1 > 10] <- NA

  y <- cov(a[, col1], b)
  z <- fast_cov(a, b, col_x = col1_)
  expect_equal(z, y, tolerance = eps)

  col2 <- sample(c(NA, -3:60), 100, replace = TRUE)
  col2_ <- col2
  col2[is.na(col2) | col2 < 1 | col2 > 50] <- NA

  y <- cov(a, b[, col2])
  z <- fast_cov(a, b, col_y = col2_)
  expect_equal(z, y, tolerance = eps)

  y <- cov(a[, col1], b[, col2])
  z <- fast_cov(a, b, col_x = col1_, col_y = col2_)
  expect_equal(z, y, tolerance = eps)

})

#' @title Isotropic re-triangulation of a surface mesh
#' @description
#' Re-triangulates a closed surface mesh so that all edges approach a uniform
#' target length, improving triangle regularity and quality without changing
#' the surface topology or shape. The output has a different vertex and face
#' count than the input but closely tracks the original surface geometry.
#'
#' @details
#' Each iteration applies three steps following the procedure described in the
#' \strong{References}:
#' \enumerate{
#'   \item \strong{Edge split}: every edge longer than
#'     \eqn{4/3 \times \code{target\_edge\_length}} is split at its midpoint.
#'     One-edge splits produce 2 sub-triangles; two-edge splits produce 3;
#'     three-edge splits produce 4 (the standard 1-to-4 uniform refinement).
#'   \item \strong{Edge collapse}: every edge shorter than
#'     \eqn{4/5 \times \code{target\_edge\_length}} is collapsed to its
#'     midpoint. A collapse is skipped if it would flip any surrounding
#'     face normal (manifold-safety check).
#'   \item \strong{Tangential smoothing}: \code{n_smooth} passes of
#'     uniform-weight 1-ring \verb{Laplacian} averaging, with each
#'     displacement projected onto the vertex tangent plane (the normal
#'     component is removed) before application. This improves vertex
#'     regularity while keeping vertices near the original surface.
#' }
#' After all iterations, vertex normals are refreshed.
#'
#' Unlike \code{\link{vcg_uniform_remesh}}, which uses volumetric resampling
#' and may change topology, this function operates entirely on the mesh
#' surface and preserves the input genus and manifold structure.
#'
#' @param mesh triangular mesh of class \code{'mesh3d'} (or coercible via
#' \code{\link{ensure_mesh3d}}). The mesh \strong{must} be closed, manifold,
#' and genus-0 (watertight, no boundary or non-manifold edges, single
#' connected component); \code{mris_remesh} raises an error if such defects
#' are detected
#' @param target_edge_length numeric; desired uniform edge length in the same
#' units as \code{mesh$vb}. Default \code{NULL} uses the current average edge
#' length (quality improvement without size change)
#' @param niterations integer; number of split/collapse/smooth iterations.
#' Default \code{5}
#' @param n_smooth integer; number of tangential-smooth passes per iteration.
#' Default \code{2}
#' @param damping numeric; fraction of the tangential displacement applied at
#' each smooth pass (between 0 and 1). Default \code{0.99}
#' @param verbose logical; print per-iteration vertex and face counts.
#' Default \code{FALSE}
#'
#' @returns A \code{'mesh3d'} object with \code{vb}, \code{it}, and
#' \code{normals}. Vertex and face counts differ from the input; the surface
#' geometry is preserved.
#'
#' @references
#' A \verb{remeshing} approach to \verb{multiresolution} modeling.
#' \emph{Proceedings of Shape \verb{Modelling} International}, 49-58 (2003).
#'
#' @examples
#'
#' sphere <- vcg_sphere()
#' sphere
#'
#' vcg_average_edge_length(sphere)
#' plot(sphere)
#'
#' remeshed <- mris_remesh(
#'   sphere,
#'   target_edge_length = 0.3
#' )
#'
#' plot(remeshed)
#'
#'
#'
#' @export
mris_remesh <- function(
    mesh,
    target_edge_length = NULL,
    niterations        = 5L,
    n_smooth           = 2L,
    damping            = 0.99,
    verbose            = FALSE
) {
  mesh <- meshintegrity(mesh, facecheck = TRUE)

  vb <- mesh$vb[1:3, , drop = FALSE]
  storage.mode(vb) <- "double"

  it <- mesh$it
  storage.mode(it) <- "integer"

  if (is.null(target_edge_length)) {
    target_edge_length <- vcg_average_edge_length(mesh)
  }

  tmp <- mrisRemesh(
    vb_            = vb,
    it_            = it,
    target_edge_len = as.double(target_edge_length)[[1L]],
    niterations    = as.integer(niterations)[[1L]],
    n_smooth       = as.integer(n_smooth)[[1L]],
    damping        = as.double(damping)[[1L]],
    verbose        = as.logical(verbose)[[1L]]
  )

  structure(
    list(
      vb      = rbind(tmp$vb, 1),
      it      = tmp$it,
      normals = rbind(tmp$normals, 1)
    ),
    class = c("ravetools_mesh3d", "mesh3d")
  )
}

#' @name dijkstras-path
#' @title Calculate distances along a surface
#' @description
#' Calculate surface distances of graph or mesh using 'Dijkstra' method.
#' @param start_node integer, row index of \code{positions} on where to start
#' calculating the distances. This integer must be 1-indexed and cannot exceed
#' the total number of \code{positions} rows
#' @param positions numeric matrix with no \code{NA} values. The number of row
#' is the total count of nodes (vertices), and the number of columns represent
#' the node dimension. Each row represents a node.
#' @param faces integer matrix with each row containing indices of nodes. For
#' graphs, \code{faces} is a matrix with two columns defining the connecting
#' edges; for '3D' mesh, \code{faces} is a three-column matrix defining the
#' face index of mesh triangles.
#' @param face_index_start integer, the start of the nodes in \code{faces};
#' please specify this input explicitly if the first node is not contained
#' in \code{faces}.
#' Default is \code{NA} (determined by the minimal number in \code{faces}).
#' The reason to set this input is because some programs use \code{1} to
#' represent the first node, some start from \code{0}.
#' @param max_search_distance numeric, maximum distance to iterate;
#' default is \code{NA},
#' that is to iterate and search the whole mesh
#' @param ... reserved for backward compatibility
#' @param x distance calculation results returned by
#' \code{dijkstras_surface_distance} function
#' @param target_node the target node number to reach (from the starting node);
#' \code{target_node} is always 1-indexed.
#' @returns \code{dijkstras_surface_distance} returns a list distance
#' table with the meta configurations. \code{surface_path} returns a data frame
#' of the node ID (from \code{start_node} to \code{target_node}) and cumulative
#' distance along the shortest path.
#'
#' @examples
#'
#' # ---- Toy example --------------------
#'
#' # Position is 2D, total 6 points
#' positions <- matrix(runif(6 * 2), ncol = 2)
#'
#' # edges defines connected nodes
#' edges <- matrix(ncol = 2, byrow = TRUE, data = c(
#'   1,2,
#'   2,3,
#'   1,3,
#'   2,4,
#'   3,4,
#'   2,5,
#'   4,5,
#'   2,5,
#'   4,6,
#'   5,6
#' ))
#'
#' # calculate distances
#' ret <- dijkstras_surface_distance(
#'   start_node = 1,
#'   positions = positions,
#'   faces = edges,
#'   face_index_start = 1
#' )
#'
#' # get shortest path from the first node to the last
#' path <- surface_path(ret, target_node = 6)
#'
#' # plot the results
#' from_node <- path$path[-nrow(path)]
#' to_node <- path$path[-1]
#' plot(positions, pch = 16, axes = FALSE,
#'      xlab = "X", ylab = "Y", main = "Dijkstra's shortest path")
#' segments(
#'   x0 = positions[edges[,1],1], y0 = positions[edges[,1],2],
#'   x1 = positions[edges[,2],1], y1 = positions[edges[,2],2]
#' )
#'
#' points(positions[path$path,], col = "steelblue", pch = 16)
#' arrows(
#'   x0 = positions[from_node,1], y0 = positions[from_node,2],
#'   x1 = positions[to_node,1], y1 = positions[to_node,2],
#'   col = "steelblue", lwd = 2, length = 0.1, lty = 2
#' )
#'
#' points(positions[1,,drop=FALSE], pch = 16, col = "orangered")
#' points(positions[6,,drop=FALSE], pch = 16, col = "purple3")
#'
#' # ---- Example with mesh ------------------------------------
#'
#' \dontrun{
#'
#'   # Please install the down-stream package `threeBrain`
#'   # and call library(threeBrain)
#'   # the following code set up the files
#'
#'   read.fs.surface <- internal_rave_function(
#'     "read.fs.surface", "threeBrain")
#'   default_template_directory <- internal_rave_function(
#'     "default_template_directory", "threeBrain")
#'   surface_path <- file.path(default_template_directory(),
#'                             "N27", "surf", "lh.pial")
#'   if (!file.exists(surface_path)) {
#'     internal_rave_function(
#'       "download_N27", "threeBrain")()
#'   }
#'
#'   # Example starts from here --->
#'   # Load the mesh
#'   mesh <- read.fs.surface(surface_path)
#'
#'   # Calculate the path with maximum radius 100
#'   ret <- dijkstras_surface_distance(
#'     start_node = 1,
#'     positions = mesh$vertices,
#'     faces = mesh$faces,
#'     max_search_distance = 100,
#'     verbose = TRUE
#'   )
#'
#'   # get shortest path from the first node to node 43144
#'   path <- surface_path(ret, target_node = 43144)
#'
#'   # plot
#'   from_nodes <- path$path[-nrow(path)]
#'   to_nodes <- path$path[-1]
#'   # calculate colors
#'   pal <- colorRampPalette(
#'     colors = c("red", "orange", "orange3", "purple3", "purple4")
#'   )(1001)
#'   col <- pal[ceiling(
#'     path$distance / max(path$distance, na.rm = TRUE) * 1000
#'   ) + 1]
#'   oldpar <- par(mfrow = c(2, 2), mar = c(0, 0, 0, 0))
#'   for(xdim in c(1, 2, 3)) {
#'     if ( xdim < 3 ) {
#'       ydim <- xdim + 1
#'     } else {
#'       ydim <- 3
#'       xdim <- 1
#'     }
#'     plot(
#'       mesh$vertices[, xdim], mesh$vertices[, ydim],
#'       pch = ".", col = "#BEBEBE33", axes = FALSE,
#'       xlab = "P - A", ylab = "S - I", asp = 1
#'     )
#'     segments(
#'       x0 = mesh$vertices[from_nodes, xdim],
#'       y0 = mesh$vertices[from_nodes, ydim],
#'       x1 = mesh$vertices[to_nodes, xdim],
#'       y1 = mesh$vertices[to_nodes, ydim],
#'       col = col
#'     )
#'   }
#'
#'   # plot distance map
#'   distances <- ret$paths$distance
#'   col <- pal[ceiling(distances / max(distances, na.rm = TRUE) * 1000) + 1]
#'   selection <- !is.na(distances)
#'
#'   plot(
#'     mesh$vertices[, 2], mesh$vertices[, 3],
#'     pch = ".", col = "#BEBEBE33", axes = FALSE,
#'     xlab = "P - A", ylab = "S - I", asp = 1
#'   )
#'   points(
#'     mesh$vertices[selection, c(2, 3)],
#'     col = col[selection],
#'     pch = "."
#'   )
#'
#'   # reset graphic state
#'   par(oldpar)
#'
#' }
#'
#'
#'
#'
#'
#'
#' @export
dijkstras_surface_distance <- function(positions, faces, start_node, face_index_start = NA, max_search_distance = NA, ...) {

  # DIPSAUS DEBUG START
  # read.fs.surface <- internal_rave_function(
  #   "read.fs.surface", "threeBrain")
  # default_template_directory <- internal_rave_function(
  #   "default_template_directory", "threeBrain")
  # surface_path <- file.path(default_template_directory(),
  #                           "N27", "surf", "lh.pial")
  # if(!file.exists(surface_path)) {
  #   internal_rave_function(
  #     "download_N27", "threeBrain")()
  # }
  #
  # mesh <- read.fs.surface(surface_path)
  # positions <- mesh$vertices
  # faces <- mesh$faces
  # start_node <- 1L
  # face_index_start <- NA
  # max_search_distance <- NA
#
#
  # # Position is 2D, total 6 points
  # positions <- matrix(runif(6 * 2), ncol = 2)
  #
  # # edges defines connected nodes
  # faces <- matrix(ncol = 2, byrow = TRUE, data = c(
  #   1,2,
  #   2,3,
  #   1,3,
  #   2,4,
  #   3,4,
  #   2,5,
  #   4,5,
  #   2,5,
  #   4,6,
  #   5,6
  # ))
  #


  stopifnot(is.matrix(positions))
  stopifnot(is.matrix(faces))

  if (!is.integer(faces)) {
    faces <- array(as.integer(faces), dim = dim(faces))
  }

  nc_face <- ncol(faces)
  nc_positions <- ncol(positions)
  if (!nc_face %in% c(2, 3)) {
    stop("`dijkstras_distance`: Face indices `faces` should have two or three columns")
  }
  if (!nc_positions %in% c(1, 2, 3)) {
    stop("`dijkstras_distance`: Vertex `positions` dimension should be 1, 2, or 3")
  }

  n_points <- nrow(positions)
  n_indices <- nrow(faces)
  max_index <- max(faces)
  min_index <- min(faces)

  if ( !n_points || !n_indices ) {
    stop("`dijkstras_distance`: cannot work with zero number vertices or empty face indices.")
  }

  if (!isTRUE(max_search_distance > 0 && is.finite(max_search_distance))) {
    max_search_distance <- -1.0
  }

  if (anyNA(positions)) {
    stop("`dijkstras_distance`: Cannot handle NA vertex positions")
  }
  if (is.na(min_index)) {
    stop("`dijkstras_distance`: Cannot handle NA face index")
  }

  # Check the face index start (0 or 1)
  face_index_start <- as.integer(face_index_start)
  if (length(face_index_start) != 1 || is.na(face_index_start)) {
    face_index_start <- min_index
  }

  faces <- faces - face_index_start

  if ( max_index - face_index_start + 1 > n_points ) {
    warning("`surface_distance_dijkstras`: face index starts from ", face_index_start,
            ". Found face index [", max_index, "] that is greater than maximum allowed [",
            n_points - 1 + face_index_start, "]. Some mesh triangles will be ignored.")
    max_fidx <- n_points - 1
    sel <- rowSums(faces < 0 | faces > max_fidx) == 0
    faces <- faces[sel, , drop = FALSE]
    n_indices <- nrow(faces)
  }
  if ( n_indices == 0 ) {
    stop("No valid face indices found.")
  }

  if (length(start_node) < 1) {
    stop("`dijkstras_distance`: `start_node` must not be empty")
  }
  start_node <- as.integer(start_node - face_index_start)
  if (!isTRUE( all( start_node >= 0 & start_node < n_points ) )) {
    stop("`dijkstras_distance`: `start_node` is invalid. `start_node - face_index_start` must be integer(s) (no NA allowed) between 0 ~ ", n_points - 1)
  }

  if ( nc_positions == 1 ) {
    positions <- cbind(positions, 0, 0)
  } else if ( nc_positions == 2 ) {
    positions <- cbind(positions, 0)
  }

  if ( nc_face == 2 ) {
    positions <- rbind(positions, Inf)
    faces <- rbind(
      cbind(faces, n_points),
      cbind(n_points, faces[, c(2, 1)])
    )
  }

  ret <- vcgDijkstra(
    t(positions),
    t(faces),
    start_node,
    max_search_distance
  )

  if ( max_search_distance <= 0 ) {
    max_search_distance <- Inf
  }

  structure(
    list(
      paths = data.frame(
        node_id = seq_len(n_points),
        prev_id = ret$parent[seq_len(n_points)] + 1L,
        distance = ret$geodist[seq_len(n_points)]
      ),
      start_node = start_node,
      face_index_start = face_index_start,
      max_search_distance = max_search_distance,
      n_nodes = n_points,
      n_faces = n_indices
    ),
    class = c("ravetools-surf-dist-dijkstras", "ravetools-surf-dist")
  )

}



# x <- dijkstras_surface_distance(1, mesh$vertices, mesh$faces, max_search_distance = NA, verbose = TRUE)

#' @rdname dijkstras-path
#' @export
surface_path <- function(x, target_node) {

  if (!inherits(x, "ravetools-surf-dist")) {
    stop("`surface_path`: input `x` must be a surface distance result. Please check `?dijkstras_surface_distance` to calculate surface distance first.")
  }

  target_node <- as.integer(target_node)
  if ( !isTRUE( target_node >= 1 & target_node <= x$n_nodes )) {
    stop("`dijkstras_findpath`: `target_node` must be 1~", x$n_nodes)
  }

  new_idx <- target_node
  idx <- NULL
  while (!is.na(new_idx)) {
    idx <- c(new_idx, idx)
    new_idx <- x$paths$prev_id[[new_idx]]
  }

  data.frame(
    path = idx,
    distance = x$paths$distance[idx]
  )
}

butter_cutoff <- function(type, n, w, r) {
  # DIPSAUS DEBUG START
  # w <- 60 / 250
  # type <- "high"
  # w <- c(1, 60) / 250
  # type <- "pass"
  # r <- 6
  # n <- 6

  Ws <- sort(w)
  Rs <- r
  Rc <- 10 * log10(2)

  # Wpw <- (2/T) * tan(pi * Wp/T)
  Wsw <- tan(pi * Ws / 2)

  qs <- log(10 ^ (Rs / 10) - 1)
  qc <- log(10 ^ (Rc / 10) - 1)

  # n <- 0.5 * (qs - qp)/log(ws/wp)
  # log(ws/wc) =
  log_ws_wc <- (qs - qc) / ( 2 * n )
  ws_wc <- exp( log_ws_wc )

  switch(
    type,
    "low" = {
      Wcw <- Wsw / ws_wc
    },
    "high" = {
      # wc <- Wsw
      # ws <- Wcw
      Wcw <- Wsw * ws_wc
    },
    "pass" = {
      w02 <- Wsw[[1]] * Wsw[[2]]
      # wp <- Wpw[2] - Wpw[1]
      ws <- Wsw[2] - Wsw[1]

      # ws <- ws/wp
      # wp <- 1
      wc <- ws / ws_wc

      # solve:
      # Wcw[2] - Wcw[1] = wc
      # Wcw[2] * Wcw[1] = w02
      #
      # i.e.
      # Wcw[1]^2 + wc * Wcw[1] - w02 = 0
      Wcw1 <- (-wc + sqrt(wc^2 + 4 * w02)) / 2
      Wcw2 <- Wcw1 + wc
      # 0.005631939 0.441716719
      Wcw <- c(Wcw1, Wcw2)
    },
    "stop" = {
      w02 <- Wsw[1] * Wsw[2]

      # wp <- w02/(Wpw[2] - Wpw[1])
      # ws <- w02/(Wsw[2] - Wsw[1])
      # ws <- ws/wp
      # wp <- 1

      ws <- w02 / (Wsw[2] - Wsw[1])
      wc <- ws / ws_wc
      wc <- w02 / wc

      # solve:
      # Wcw[2] - Wcw[1] = wc
      # Wcw[2] * Wcw[1] = w02
      #
      # i.e.
      # Wcw[1]^2 + wc * Wcw[1] - w02 = 0
      Wcw1 <- (-wc + sqrt(wc^2 + 4 * w02)) / 2
      Wcw2 <- Wcw1 + wc
      Wcw <- c(Wcw1, Wcw2)
    }
  )

  Wc <- atan( Wcw ) * 2 / pi

  # filter <- butter(n, Wc, type)
  # hw <- freqz2(filter$b, filter$a, fs = 500)
  # print(hw, cutoffs = -Rs)
  # filtmag_db(filter$b, filter$a, Wc) + Rc
  # filtmag_db(filter$b, filter$a, Ws) + Rs
  Wc

}

butter_check <- function(type, n, w, r) {
  r <- abs(r)
  Wc <- butter_cutoff(type, n, w, r)
  filter <- gsignal::butter(n = n, w = Wc, type = type, output = "Arma")

  half_pow <- 10 * log10(0.5)
  check_filter(b = filter$b, a = filter$a, w = c(w, Wc), r_expected = rep(c(-r, half_pow), each = length(w)))
}

#' @title 'Butterworth' filter with maximum order
#' @description
#' Large filter order might not be optimal, but at lease this function
#' provides a feasible upper bound for the order such that the
#' filter has a stable \code{AR} component.
#'
#' @param type filter type
#' @param w scaled frequency ranging from 0 to 1, where 1 is 'Nyquist' frequency
#' @param r decibel attenuation at frequency \code{w}, default is around
#' \code{3 dB} (half power)
#' @param tol tolerance of reciprocal condition number, default is
#' \code{.Machine$double.eps}.
#'
#' @returns 'Butterworth' filter in 'Arma' form.
#'
#' @examples
#'
#' # Find highest order (sharpest transition) of a band-pass filter
#' sample_rate <- 500
#' nyquist <- sample_rate / 2
#'
#' type <- "pass"
#' w <- c(1, 50) / nyquist
#' Rs <- 6     # power attenuation at w
#'
#' # max order filter
#' filter <- butter_max_order(w, "pass", Rs)
#'
#' # -6 dB cutoff should be around 1 ~ 50 Hz
#' diagnose_filter(filter$b, filter$a, fs = sample_rate)
#'
#' @export
butter_max_order <- function(w, type = c("low", "high", "pass", "stop"),
                             r = 10 * log10(2), tol =  .Machine$double.eps) {

  w <- sort(w)
  r <- abs(r)

  stopifnot2(
    all(is.finite(w) & w >= 0 & w <= 1),
    msg = "`w` must be scaled to [0, 1], where 1 is Nyquist frequency"
  )

  if (type %in% c("pass", "stop")) {
    stopifnot(length(w) == 2)
  } else {
    stopifnot(length(w) == 1)
  }

  validator <- function(n) {
    check <- butter_check(type, n, w, r)
    check$reciprocal_condition > tol
  }

  n <- guess_max_integer(validator, min_v = 1)

  if (is.na(n)) { n <- 1 }

  Wc <- butter_cutoff(type, n, w, r)

  filter <- butter(n, Wc, type)

  filter$n <- n
  filter$Wc <- Wc
  filter$type <- type
  filter$check <- butter_check(type, n, w, r)
  filter

}

# Float-precision Morlet wavelet implementation.
#
# Public entry point is `morlet_wavelet()` (defined in R/wavelet.R), which
# dispatches here when `precision = "float"`.

wavelet_kernels2_float <- function(freqs, srate, wave_num,
                             data_length, signature = NULL) {
  freqs <- as.double(freqs)
  srate <- as.double(srate)
  wave_num <- as.double(wave_num)
  data_length <- as.integer(data_length)
  kernel_info <- wavelet_kernels(freqs = freqs, srate = srate, wave_num = wave_num)
  digest <- digest::digest(list(freqs, srate, wave_num, data_length, signature = signature))
  root_dir <- file.path(tempdir2(check = TRUE), "ravetools")
  if (!dir.exists(root_dir)) {
    dir.create(root_dir, showWarnings = FALSE, recursive = TRUE)
  }
  path <- file.path(root_dir, sprintf("wavelet-%s", digest))


  arr_dim <- c(data_length, length(kernel_info$kernels))
  tryCatch({
    return(filearray::filearray_checkload(
      filebase = path, mode = "readonly", symlink_ok = FALSE,
      freqs = freqs, srate = srate, wave_num = wave_num, data_length = data_length,
      arr_dim = arr_dim, rave_data_type = "rave-wavelet-kernels-float"
    ))
  }, error = function(e) {

    if (getOption("ravetools.debug", FALSE)) {
      print(path)
      warning(e)
    }

    if (dir.exists(path)) {
      unlink(path, recursive = TRUE)
    }

  })

  arr <- filearray::filearray_create(
    filebase = path,
    dimension = arr_dim,
    type = "complex",
    partition_size = 1
  )
  arr$.mode <- "readwrite"

  tmp <- complex(data_length)
  lapply(seq_along(kernel_info$kernels), function(ii) {
    tmp_wavelet <- kernel_info$kernels[[ii]]
    w_l <- length(tmp_wavelet)
    n_pre  <- ceiling(data_length / 2) - floor(w_l / 2)
    n_post <- data_length - n_pre - w_l
    x <- c(rep(0, n_pre), tmp_wavelet, rep(0, n_post))
    fftw_c2c(data = x, inverse = 0L, ret = tmp)
    conjugate(tmp)
    arr[, ii] <- tmp
    NULL
  })


  arr$.header$freqs <- freqs
  arr$.header$srate <- srate
  arr$.header$wave_num <- wave_num
  arr$.header$data_length <- data_length
  arr$.header$arr_dim <- arr_dim
  arr$.header$freqs <- freqs
  arr$.header$rave_data_type <- "rave-wavelet-kernels-float"
  arr$.save_header()

  arr$.mode <- "readonly"

  return(arr)
}

morlet_wavelet_float <- function(data, freqs, srate, wave_num,
                                 trend = c("constant", "linear", "none"),
                                 signature = NULL, segment_length = NULL, ...) {

  trend <- match.arg(trend)
  freqs <- as.double(freqs)
  srate <- as.double(srate)
  wave_num <- as.double(wave_num)
  more_args <- list(...)
  data_digest <- digest::digest(data)

  f_l <- length(freqs)
  d_l <- length(data)

  segment_length_norm <- wavelet_normalize_segment_length(segment_length, d_l)
  segmented <- !is.na(segment_length_norm)

  # detrend on the full signal preserves legacy semantics regardless of path
  if (trend != "none") {
    data <- as.vector(detrend(data, trend = trend, ...))
  }

  # In segmented mode we need the segmentation plan up-front (it both
  # validates `segment_length` against the longest wavelet kernel and supplies
  # row indices used by the per-frequency fmap closure below).
  if (segmented) {
    kern_info <- wavelet_kernels(freqs = freqs, srate = srate, wave_num = wave_num)
    max_kernel_len <- max(vapply(kern_info$kernels, length, integer(1L)))
    plan <- wavelet_segment_plan(d_l, segment_length_norm, max_kernel_len)
    kern_data_length <- segment_length_norm
  } else {
    plan <- NULL
    kern_data_length <- d_l
  }

  # Kernel cache differentiates by `data_length`, so single-shot and segmented
  # builds get distinct kernel files. Output cache deliberately omits
  # `segment_length` since both paths yield (up to fp) identical coefficients.
  fft_waves <- wavelet_kernels2_float(freqs, srate, wave_num, kern_data_length,
                                      signature = signature)

  out_path <- tempfile2()
  output <- tryCatch({
    filearray::filearray_checkload(
      filebase = out_path, symlink_ok = FALSE,
      freqs = freqs, srate = srate, wave_num = wave_num,
      data_digest = data_digest, trend = trend,
      more_args = more_args,
      rave_data_type = "rave-wavelet-coefficients",
      precision = "float"
    )
  }, error = function(e) {
    if (getOption("ravetools.debug", FALSE)) {
      print(out_path)
      warning(e)
    }
    if (dir.exists(out_path)) { unlink(out_path, recursive = TRUE) }
    NULL
  })
  if (inherits(output, "FileArray")) { return(output) }

  output <- filearray::filearray_create(
    filebase = out_path, dimension = c(d_l, f_l),
    type = "complex", partition_size = 1
  )
  output$.mode <- "readwrite"

  wave_len <- nrow(fft_waves)
  ind <- seq_len(ceiling(wave_len / 2))
  norm_factor <- wave_len * sqrt(srate / 2)

  if (segmented) {
    seg_mat <- wavelet_build_segment_matrix(data, plan)
    fft_seg <- mvfftw_r2c(seg_mat, HermConj = 1L)
    n_seg <- plan$n_seg
    local_starts <- plan$local_starts
    local_ends <- plan$local_ends

    new_ind <- seq_len(plan$segment_length)
    new_ind <- c(new_ind[-ind], new_ind[ind])

    # Per frequency: batch-convolve the kernel against all segment columns,
    # then gather the fft-shifted interior of each segment into a length-d_l
    # vector that fmap writes to the corresponding output column. We index
    # `out_cplx` through the precomputed shift-permutation `new_ind` instead
    # of materializing a `shifted` matrix (avoids a large per-frequency alloc).
    
    # Creating buffer to avoid repeated allocations in the fmap loop
    buffer <- array(0 + 0i, dim = c(plan$segment_length, n_seg))
    fmap_fun <- function(input) {
      # out_cplx <- mvfftw_c2c(kern * fft_seg, inverse = 1L) / norm_factor
      # Hence out_cplx = buffer * norm_factor
      mvfftw_c2c(input[[1]] * fft_seg, inverse = 1L, ret = buffer)

      # The math is: shifted = rbind(out_cplx[-ind, , drop = FALSE], out_cplx[ind, , drop = FALSE])
      # or shifted = rbind(buffer[-ind, , drop = FALSE], buffer[ind, , drop = FALSE]) / norm_factor
      unlist(lapply(seq_len(n_seg), function(k) {
        # equivalent to: shifted[local_starts[k]:local_ends[k], k]
        # Division by norm_factor deferred until after subsetting to avoid repeated large allocs.
        buffer[new_ind[local_starts[k]:local_ends[k]], k] / norm_factor
      }))
    }
  } else {
    fft_data <- fftw_r2c(data)
    tmp <- complex(length(fft_data))
    fmap_fun <- function(input) {
      fftw_c2c(data = input[[1]] * fft_data, inverse = 1L, ret = tmp)
      c(tmp[-ind], tmp[ind]) / norm_factor
    }
  }

  filearray::fmap(x = fft_waves, fun = fmap_fun,
                  .y = output, .buffer_count = ncol(output))

  output$.header$freqs <- freqs
  output$.header$srate <- srate
  output$.header$wave_num <- wave_num
  output$.header$data_digest <- data_digest
  output$.header$trend <- trend
  output$.header$more_args <- more_args
  output$.header$rave_data_type <- "rave-wavelet-coefficients"
  output$.header$precision <- "float"
  output$.save_header()

  output$.mode <- "readonly"
  output
}

meshintegrity <- function(mesh, facecheck = FALSE, normcheck = FALSE) {

  mesh <- ensure_mesh3d(mesh)

  ##check vertices
  if (!is.null(mesh$vb) && is.matrix(mesh$vb)) {
    vdim <- dim(mesh$vb)
    if (vdim[2] < 1) {
      stop("mesh has no vertices")
    }
    if (!vdim[1] %in% c(3, 4)) {
      stop("vertices have invalid dimensionality")
    }
    if (!identical(storage.mode(mesh$vb), "double")) {
      storage.mode(mesh$vb) <- "double"
    }
    if ( anyNA(mesh$vb) ) {
      stop("vertex coords need to be numeric values")
    }
  } else {
    stop("mesh has no/invalid vertices")
  }
  if (is.matrix(mesh$it)) {
    if (ncol(mesh$it) == 0)
      mesh$it <- NULL
  }
  if (!is.null(mesh$it) && is.matrix(mesh$it)) {
    itdim <- dim(mesh$it)
    if (itdim[1] != 3) {
      stop("only triangular faces are valid")
    }

    if (!identical(storage.mode(mesh$it), "integer")) {
      storage.mode(mesh$it) <- "integer"
    }
    if ( anyNA(mesh$it) ) {
      stop("face indices need to be integer values")
    }

    itrange <- range(mesh$it)
    if (itrange[1] < 1 || itrange[2] > vdim[2]) {
      stop("faces reference non-existent vertices")
    }
  } else if (facecheck) {
    stop("mesh has no triangular face indices")
  }

  if (normcheck) {
    if (!is.null(mesh$normals) && is.matrix(mesh$normals)) {
      ndim <- dim(mesh$normals)
      if (!prod(ndim == vdim)) {
        stop("normals must be of same dimensionality as vertices")
      }

      if ( anyNA(mesh$normals) ) {
        stop("normal coords need to be numeric values")
      }
    } else {
      stop("mesh has no vertex normals")
    }
  }
  return(mesh)
}

cSizeMesh <- function(mesh) {
  x <- t(mesh$vb[1:3, ])
  X <- scale(x, scale = FALSE)
  y <- sqrt(sum(as.vector(X) ^ 2))
  return(y)
}

pcax <- function(mesh) {
  x <- t(mesh$vb[1:3, ])
  pc <- 3 * stats::prcomp(x, retx = FALSE)$sdev[1]
  return(pc)
}

meshOff <- function(x, offset) {
  x <- vcg_update_normals(x)
  x$vb[1:3, ] <- x$vb[1:3, ] + offset * x$normals[1:3, ]
  return(x)
}

bbox <- function(x) {
  # Bounding box

  bbox <- apply(x$vb[1:3, , drop = FALSE], 1L, range)
  bbox <- expand.grid(bbox[, 1], bbox[, 2], bbox[, 3])
  dia <- max(stats::dist(bbox))
  return(list(bbox = bbox, diag = dia))
}

checkFaceOrientation <- function(x, offset = NULL) {
  if (is.null(offset)) {
    offset <- pcax(x) / 15
  }
  out <- TRUE
  xoff <- meshOff(x, offset)
  cx <- cSizeMesh(x)
  cxoff <- cSizeMesh(xoff)
  if (cx > cxoff) {
    out <- FALSE
  }
  return(out)
}

invertFaces <- function(mesh) {
  mesh$it <- mesh$it[c(3, 2, 1), ]
  mesh <- vcg_update_normals(mesh)
  return(mesh)
}

#' @title Update vertex normal
#' @param mesh triangular mesh or a point-cloud (matrix of 3 columns)
#' @param weight method to compute per-vertex normal vectors: \code{"area"}
#' weighted average of surrounding face normal, or \code{"angle"} weighted
#' vertex normal vectors.
#' @param pointcloud integer vector of length 2: containing optional
#' parameters for normal calculation of point clouds; the first entry
#' specifies the number of neighboring points to consider; the second
#' entry specifies the amount of smoothing iterations to be performed.
#' @param verbose whether to verbose the progress
#' @returns A \code{'mesh3d'} object with normal vectors.
#' @examples
#'
#' if(is_not_cran()) {
#'
#' # Prepare mesh with no normal
#' data("left_hippocampus_mask")
#' mesh <- vcg_isosurface(left_hippocampus_mask)
#' mesh$normals <- NULL
#'
#' # Start: examples
#' new_mesh <- vcg_update_normals(mesh, weight = "angle",
#'                                pointcloud = c(10, 10))
#'
#' rgl_view({
#'   rgl_call("mfrow3d", 1, 2)
#'   rgl_call("shade3d", mesh, col = 2)
#'
#'   rgl_call("next3d")
#'   rgl_call("shade3d", new_mesh, col = 2)
#' })
#' }
#'
#'
#' @export
vcg_update_normals <- function(
    mesh,
    weight = c("area", "angle"),
    pointcloud = c(10, 0),
    verbose = FALSE
) {
  weight <- match.arg(weight)
  if (length(pointcloud) != 2) {
    stop("pointcloud must be an integer vector of length 2")
  }

  if ( weight == "area" ) {
    type <- 0L
  } else {
    type <- 1L
  }
  if ( is.matrix(mesh)) {
    tmp <- list()
    tmp$vb <- rbind(t(mesh), 1)
    mesh <- tmp
    class(mesh) <- "mesh3d"
  }
  mesh <- meshintegrity(mesh)
  vb <- mesh$vb[1:3, , drop = FALSE]
  if (!is.matrix(vb)) {
    stop("mesh has no vertices")
  }
  it <- mesh$it - 1L
  normals <- vcgUpdateNormals(vb, it, type, pointcloud, !verbose)
  mesh$normals <- rbind(normals, 1)
  return(mesh)
}


vcg_barycentric_subdivision <- function(mesh) {
  mesh <- meshintegrity(mesh = mesh, facecheck = TRUE)
  v0 <- mesh$vb[1:3, mesh$it[1, ], drop = FALSE]
  v1 <- mesh$vb[1:3, mesh$it[2, ], drop = FALSE]
  v2 <- mesh$vb[1:3, mesh$it[3, ], drop = FALSE]
  vb <- (v0 + v1 + v2) / 3.0

  # Adding n=ncol(vb) indices, hence 2n more faces
  vb_faces <- seq_len(ncol(vb)) + ncol(mesh$vb)

  f1 <- rbind(
    mesh$it[2, ],
    mesh$it[3, ],
    vb_faces,
    deparse.level = 0
  )

  f2 <- rbind(
    mesh$it[3, ],
    mesh$it[1, ],
    vb_faces,
    deparse.level = 0
  )

  mesh$it[3, ] <- vb_faces

  structure(
    class = "mesh3d",
    list(
      vb = cbind(mesh$vb[1:3, , drop = FALSE], vb, deparse.level = 0),
      it = cbind(mesh$it[1:3, , drop = FALSE], f1, f2, deparse.level = 0)
    )
  )

}

vcg_edge_subdivision <- function(mesh) {
  mesh <- meshintegrity(mesh = mesh, facecheck = TRUE)
  vb <- mesh$vb[1:3, , drop = FALSE]
  it <- mesh$it - 1L
  storage.mode(it) <- "integer"
  m <- vcgEdgeSubdivision(vb, it)
  return(m)
}

#' @name vcg_subdivision
#' @title Sub-divide (up-sample) a triangular mesh
#' @description
#' Up-sample a triangular mesh by adding a vertex at each edge or face center.
#' @param mesh triangular mesh stored as object of class 'mesh3d'.
#' @param method either \code{'edge'} (default) to add new mid-point vertices to
#' edge, or \code{'barycenter'} to add new vertices at face \code{'Bary'}
#' centers.
#' @returns An object of class "mesh3d"
#'
#' @examples
#'
#' mesh <- plane_geometry()
#'
#' # default
#' mesh_edge <- vcg_subdivision(mesh, "edge")
#'
#' # barycenter
#' mesh_face <- vcg_subdivision(mesh, "barycenter")
#'
#' if(is_not_cran()) {
#'
#'   rgl_view({
#'     rgl_call("wire3d", mesh, col = 1)
#'     rgl_call("wire3d", mesh_edge, col = 2)
#'     rgl_call("wire3d", mesh_face, col = 3)
#'   })
#'
#'
#' }
#'
#'
#'
#' @export
vcg_subdivision <- function(mesh, method = c("edge", "barycenter")) {
  method <- match.arg(method)

  mesh <- switch(
    method,
    "edge" = {
      vcg_edge_subdivision(mesh)
    },
    {
      vcg_barycentric_subdivision(mesh)
    }
  )
  mesh
}

#' @name vcg_smooth
#' @title Implicitly smooth a triangular mesh
#' @description
#' Applies smoothing algorithms on a triangular mesh.
#'
#' @param mesh triangular mesh stored as object of class 'mesh3d'.
#' @param use_mass_matrix logical: whether to use mass matrix to keep the mesh
#' close to its original position (weighted per area distributed on vertices);
#' default is \code{TRUE}
#' @param fix_border logical: whether to fix the border vertices of the mesh;
#' default is \code{FALSE}
#' @param use_cot_weight logical: whether to use cotangent weight; default is
#' \code{FALSE} (using uniform 'Laplacian')
#' @param laplacian_weight numeric: weight when \code{use_cot_weight} is \code{FALSE};
#' default is \code{1.0}
#' @param degree integer: degrees of 'Laplacian'; default is \code{1}
#' @param type method name of explicit smooth, choices are \code{'taubin'},
#' \code{'laplace'}, \code{'HClaplace'}, \code{'fujiLaplace'},
#' \code{'angWeight'}, \code{'surfPreserveLaplace'}.
#' @param iteration number of iterations
#' @param lambda In \code{vcg_smooth_implicit}, the amount of smoothness,
#' useful only if \code{use_mass_matrix} is \code{TRUE}; default is \code{0.2}.
#' In \code{vcg_smooth_explicit}, parameter for \code{'taubin'} smoothing.
#' @param mu parameter for \code{'taubin'} explicit smoothing.
#' @param delta parameter for scale-dependent 'Laplacian' smoothing or
#' maximum allowed angle (in 'Radian') for deviation between surface preserving
#' 'Laplacian'.
#' @returns An object of class "mesh3d" with:
#' \item{\code{vb}}{vertex coordinates}
#' \item{\code{normals}}{vertex normal vectors}
#' \item{\code{it}}{triangular face index}
#' @examples
#'
#' if(is_not_cran()) {
#'
#' # Prepare mesh with no normals
#' data("left_hippocampus_mask")
#'
#' # Grow 2mm on each direction to fill holes
#' volume <- grow_volume(left_hippocampus_mask, 2)
#'
#' # Initial mesh
#' mesh <- vcg_isosurface(volume)
#'
#' # Start: examples
#' rgl_view({
#'   rgl_call("mfrow3d", 2, 4)
#'   rgl_call("title3d", "Naive ISOSurface")
#'   rgl_call("shade3d", mesh, col = 2)
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Implicit Smooth")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_implicit(mesh, degree = 2))
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Explicit Smooth - taubin")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_explicit(mesh, "taubin"))
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Explicit Smooth - laplace")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_explicit(mesh, "laplace"))
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Explicit Smooth - angWeight")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_explicit(mesh, "angWeight"))
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Explicit Smooth - HClaplace")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_explicit(mesh, "HClaplace"))
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Explicit Smooth - fujiLaplace")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_explicit(mesh, "fujiLaplace"))
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "Explicit Smooth - surfPreserveLaplace")
#'   rgl_call("shade3d", col = 2,
#'            x = vcg_smooth_explicit(mesh, "surfPreserveLaplace"))
#' })
#'
#' }
#'
#' @export
vcg_smooth_implicit <- function(
    mesh, lambda = 0.2, use_mass_matrix = TRUE, fix_border = FALSE,
    use_cot_weight = FALSE, degree = 1L, laplacian_weight = 1.0
) {
  mesh <- meshintegrity(mesh)
  smooth_quality <- FALSE
  vb <- mesh$vb[1:3, , drop = FALSE]
  it <- (mesh$it - 1L)
  storage.mode(it) <- "integer"
  lambda <- as.double(lambda)[[1]]
  laplacian_weight <- as.double(laplacian_weight)[[1]]
  use_mass_matrix <- as.logical(use_mass_matrix)[[1]]
  fix_border <- as.logical(fix_border)[[1]]
  use_cot_weight <- as.logical(use_cot_weight)[[1]]
  smooth_quality <- as.logical(smooth_quality)[[1]]
  degree <- as.integer(degree)[[1]]
  tmp <- vcgSmoothImplicit(vb, it, lambda, use_mass_matrix, fix_border, use_cot_weight, degree, laplacian_weight, smooth_quality)
  mesh$vb[1:3, ] <- tmp$vb
  mesh$normals <- rbind(tmp$normals, 1)
  mesh$it <- tmp$it
  mesh
}

#' @rdname vcg_smooth
#' @export
vcg_smooth_explicit <- function(
  mesh, type = c("taubin", "laplace", "HClaplace", "fujiLaplace", "angWeight", "surfPreserveLaplace"),
  iteration = 10, lambda = 0.5, mu = -0.53, delta = 0.1
) {
  mesh <- meshintegrity(mesh)
  type <- match.arg(type)
  type <- substring(type[1], 1L, 1L)
  vb <- mesh$vb[1:3, , drop = FALSE]
  it <- (mesh$it - 1L)
  storage.mode(it) <- "integer"
  method <- 0
  if (type == "l" || type == "L") {
    method <- 1
  } else if (type == "H" || type == "h") {
    method <- 2
  } else if (type == "f" || type == "F") {
    method <- 3
  } else if (type == "a" || type == "A") {
    method <- 4
  } else if (type == "s" || type == "S") {
    method <- 5
  }

  stopifnot(is.integer(it))

  tmp <- vcgSmooth(vb, it, iteration, method, lambda, mu, delta)
  mesh$vb[1:3, ] <- tmp$vb
  mesh$normals <- rbind(tmp$normals, 1)
  mesh$it <- tmp$it
  invisible(meshintegrity(mesh))
}

#' @title Compute volume for manifold meshes
#' @param mesh triangular mesh of class \code{'mesh3d'}
#' @returns The numeric volume of the mesh
#'
#' @examples
#'
#' # Initial mesh
#' mesh <- vcg_sphere()
#'
#' vcg_mesh_volume(mesh)
#'
#' @export
vcg_mesh_volume <- function(mesh) {
  vcgVolume( meshintegrity(mesh) )
}

#' Simple 3-dimensional sphere mesh
#' @param sub_division density of vertex in the resulting mesh
#' @param normals whether the normal vectors should be calculated
#' @returns A \code{'mesh3d'} object
#' @examples
#'
#' vcg_sphere()
#'
#' @export
vcg_sphere <- function(sub_division = 3L, normals = TRUE) {
  vcgSphere(sub_division, normals)
}

#' @title Create surface mesh from 3D-array
#' @description
#' Create surface from 3D-array using marching cubes algorithm
#'
#' @param volume a volume or a mask volume
#' @param threshold_lb lower-bound threshold for creating the surface; default
#' is \code{0}
#' @param threshold_ub upper-bound threshold for creating the surface; default
#' is \code{NA} (no upper-bound)
#' @param vox_to_ras a \code{4x4} \code{'affine'} transform matrix indicating the
#' 'voxel'-to-world transform.
#'
#' @returns A triangular mesh of class \code{'mesh3d'}
#'
#' @examples
#'
#'
#' if(is_not_cran()) {
#'
#' library(ravetools)
#' data("left_hippocampus_mask")
#'
#' mesh <- vcg_isosurface(left_hippocampus_mask)
#'
#'
#' rgl_view({
#'
#'   rgl_call("mfrow3d", 1, 2)
#'
#'   rgl_call("title3d", "Direct ISOSurface")
#'   rgl_call("shade3d", mesh, col = 2)
#'
#'   rgl_call("next3d")
#'   rgl_call("title3d", "ISOSurface + Implicit Smooth")
#'
#'   rgl_call("shade3d",
#'            vcg_smooth_implicit(mesh, degree = 2),
#'            col = 3)
#' })
#'
#' }
#' @export
vcg_isosurface <- function(
    volume,
    threshold_lb = 0, threshold_ub = NA,
    vox_to_ras = diag(c(-1, -1, 1, 1))
) {
  if (!length(volume) || length(dim(volume)) != 3) {
    stop("vcg_isosurface: 3D non-empty `volume` array needed")
  }

  if (is.na(threshold_lb)) { threshold_lb <- 0 }
  sel <- volume > threshold_lb
  dimnames(sel) <- NULL

  if (!is.na(threshold_ub)) {
    sel <- sel & volume < threshold_ub
  }

  mesh <- structure(vcgIsoSurface( sel, 0.5 ), class = "mesh3d")

  # voxel (0-indexed) to RAS
  mesh$vb <- (vox_to_ras %*% rbind(mesh$vb, 1))[seq_len(3), ]

  if (!checkFaceOrientation(mesh)) {
    mesh <- invertFaces(mesh)
  }
  return(mesh)
}



#' Sample a surface mesh uniformly
#' @param x surface
#' @param voxel_size 'voxel' size for space 'discretization'
#' @param offset offset position shift of the new surface from the input
#' @param discretize whether to use step function(\code{TRUE}) instead of
#' linear interpolation (\code{FALSE}) to calculate the position of the
#' intersected edge of the marching cube; default is \code{FALSE}
#' @param multi_sample whether to calculate multiple samples for more accurate
#' results (at the expense of more computing time) to remove artifacts; default
#' is \code{FALSE}
#' @param absolute_distance whether an unsigned distance field should be
#' computed. When set to \code{TRUE}, non-zero offsets is to be set, and
#' double-surfaces will be built around the original surface, like a sandwich.
#' @param merge_clost whether to merge close vertices; default is \code{TRUE}
#' @param verbose whether to verbose the progress; default is \code{TRUE}
#' @returns A triangular mesh of class \code{'mesh3d'}
#' @examples
#'
#' sphere <- vcg_sphere()
#' mesh <- vcg_uniform_remesh(sphere, voxel_size = 0.45)
#'
#' if(is_not_cran()) {
#'
#' rgl_view({
#'
#'   rgl_call("mfrow3d", 1, 2)
#'
#'   rgl_call("title3d", "Input")
#'   rgl_call("wire3d", sphere, col = 2)
#'   rgl_call("next3d")
#'
#'   rgl_call("title3d", "Re-meshed to 0.1mm edge distance")
#'   rgl_call("wire3d", mesh, col = 3)
#' })
#'
#' }
#'
#' @export
vcg_uniform_remesh <- function(
    x, voxel_size = NULL, offset = 0, discretize = FALSE,
    multi_sample = FALSE, absolute_distance = FALSE, merge_clost = FALSE, verbose = TRUE) {

  x <- meshintegrity(mesh = x, facecheck = TRUE)
  if (is.null( voxel_size )) {
    voxel_size <- bbox(x)$dia / 50
  }
  vb <- x$vb
  it <- x$it - 1L
  out <- structure(
    vcgUniformResample( vb, it, voxel_size, offset, discretize,
                        multi_sample, absolute_distance, merge_clost, !verbose ),
    class = "mesh3d"
  )
  return(meshintegrity(out))
}



#' @title Cast rays to intersect with mesh
#' @param x surface mesh
#' @param ray_origin a matrix with 3 rows or a vector of length 3, the positions
#' of ray origin
#' @param ray_direction a matrix with 3 rows or a vector of length 3, the
#' direction of the ray, will be normalized to length 1
#' @param max_distance positive maximum distance to cast the normalized ray;
#' default is infinity. Any invalid distances (negative, zero, or \code{NA})
#' will be interpreted as unset.
#' @param both_sides whether to inverse the ray (search both positive and
#' negative ray directions); default is false
#' @returns A list of ray casting results: whether any intersection is found,
#' position and face normal of the intersection, distance of the ray, and the
#' index of the intersecting face (counted from 1)
#' @examples
#'
#' library(ravetools)
#' sphere <- vcg_sphere(normals = FALSE)
#' sphere$vb[1:3, ] <- sphere$vb[1:3, ] + c(10, 10, 10)
#' vcg_raycaster(
#'   x = sphere,
#'   ray_origin = array(c(0, 0, 0, 1, 0, 0), c(3, 2)),
#'   ray_direction = c(1, 1, 1)
#' )
#'
#' @export
vcg_raycaster <- function(
    x, ray_origin, ray_direction, max_distance = Inf, both_sides = FALSE) {

  x <- meshintegrity(mesh = x, facecheck = TRUE)

  if (is.matrix(ray_origin)) {
    ray_origin <- ray_origin[seq_len(3), , drop = FALSE]
    # ray_direction <- ray_direction[seq_len(3), , drop = FALSE]
  } else {
    # assuming ray_origin is a vector of 3
    ray_origin <- matrix(ray_origin[c(1, 2, 3)], ncol = 1L)
    # ray_direction <- matrix(ray_direction[c(1,2,3)], ncol = 1L)
  }
  n_rays <- ncol(ray_origin)

  if (length(ray_direction) == 3) {
    ray_direction <- matrix(ray_direction[c(1, 2, 3)], nrow = 3L, ncol = n_rays)
  } else {
    ray_direction <- ray_direction[seq_len(3), , drop = FALSE]
  }

  if (ncol(ray_direction) != n_rays) {
    stop("`vcg_raycaster`: number of rays is ", n_rays, " according to `ray_origin`. However `ray_direction` is different number of points. Please make sure these two variables have the same number of elements.")
  }

  # normalize the ray_direction
  ray_length <- sqrt(colSums(ray_direction^2))
  zero_length <- is.na(ray_length) | ray_length == 0

  ray_direction <- sweep(ray_direction, 2L, ray_length, FUN = "/", check.margin = TRUE)
  ray_direction[, zero_length] <- 0


  stopifnot(length(max_distance) == 1)
  if (is.na(max_distance) || max_distance <= 0) {
    max_distance <- Inf
  }


  results <- vcgRaycaster(
    vb_ = x$vb,
    it_ = x$it - 1L,
    rayOrigin = ray_origin,
    rayDirection = ray_direction,
    maxDistance = max_distance,
    bothSides = both_sides
    # threads =
  )

  list(
    has_intersection = as.logical(results$hitFlag),
    intersection = array(results$intersectPoints, dim = c(3L, n_rays)),
    normals = array(results$intersectNormals, dim = c(3L, n_rays)),
    face_index = results$intersectIndex + 1L,
    distance = results$castDistance,
    ray_origin = ray_origin,
    ray_direction = ray_direction
  )

}

#' @title Find nearest \code{k} points
#' @description
#' For each point in the query, find the nearest \code{k} points in target using
#' \code{K-D} tree.
#' @param target a matrix with \code{n} rows (number of points) and 2 or 3
#' columns, or a \code{mesh3d} object. This is the target point cloud where
#' nearest distances will be sought
#' @param query a matrix with \code{n} rows (number of points) and 2 or 3
#' columns, or a \code{mesh3d} object. This is the query point cloud where
#' for each point, the nearest \code{k} points in \code{target} will be sought.
#' @param k positive number of nearest neighbors to look for
#' @param leaf_size the suggested leaf size for the \code{K-D} tree; default is
#' \code{16}; larger leaf size will result in smaller depth
#' @param max_depth maximum depth of the \code{K-D} tree; default is \code{64}
#' @returns A list of two matrices: \code{index} is a matrix of indices of
#' \code{target} points, whose distances are close to the corresponding
#' \code{query} point. If no point in \code{target} is found, then \code{NA}
#' will be presented. Each \code{distance} is the corresponding distance
#' from the query point to the target point.
#'
#' @examples
#'
#' # Find nearest point in B with the smallest distance for each point in A
#'
#' library(ravetools)
#'
#' n <- 10
#' A <- matrix(rnorm(n * 2), nrow = n)
#' B <- matrix(rnorm(n * 4), nrow = n * 2)
#' result <- vcg_kdtree_nearest(
#'   target = B, query = A,
#'    k = 1
#' )
#'
#' plot(
#'   rbind(A, B),
#'   pch = 20,
#'   col = c(rep("red", n), rep("black", n * 2)),
#'   xlab = "x",
#'   ylab = "y",
#'   main = "Black: target; Red: query"
#' )
#'
#' nearest_points <- B[result$index, ]
#' arrows(A[, 1],
#'        A[, 2],
#'        nearest_points[, 1],
#'        nearest_points[, 2],
#'        col = "red",
#'        length = 0.1)
#'
#' # ---- Sanity check ------------------------------------------------
#' nearest_index <- apply(A, 1, function(pt) {
#'   which.min(colSums((t(B) - pt) ^ 2))
#' })
#'
#' result$index == nearest_index
#'
#'
#'
#' @export
vcg_kdtree_nearest <- function(
    target, query, k = 1, leaf_size = 16, max_depth = 64) {

  get_point_cloud <- function(x) {
    if (is.matrix(x) || is.array(x)) {
      x <- x[drop = FALSE]
      stopifnot2(
        is.matrix(x) && ncol(x) %in% c(2, 3),
        msg = "`vcg_kdtree_nearest`: input `x` must be a column-matrix with 2 or 3 columns and `n` rows as number of points."
      )
      if ( ncol(x)  == 2 ) {
        x <- cbind(x, 0)
      }
      x <- t(x)
    } else {
      x <- meshintegrity(mesh = x, facecheck = FALSE)
      x <- x$vb[1:3, , drop = FALSE]
    }
    x
  }

  target <- get_point_cloud(target)
  query <- get_point_cloud(query)

  k <- as.integer(k)
  if (!is.finite(k) || k <= 0) {
    stop("`vcg_kdtree_nearest`: `k` must be finite positive.")
  }

  leaf_size <- as.integer(leaf_size)
  if (!is.finite(leaf_size) || leaf_size <= 0) {
    stop("`vcg_kdtree_nearest`: `leaf_size` must be finite positive.")
  }

  max_depth <- as.integer(max_depth)
  if (!is.finite(max_depth) || max_depth <= 0) {
    stop("`vcg_kdtree_nearest`: `max_depth` must be finite positive.")
  }

  result <- vcgKDTreeSearch(
    target_ = target,
    query_ = query,
    k = k,
    nPointsPerCell = leaf_size,
    maxDepth = max_depth
  )
  result$index <- result$index + 1L
  result$index[result$index == 0] <- NA_integer_
  result

}

#' @title Subset mesh by vertex
#' @param x surface mesh
#' @param selector logical vector (must not contain NA), and length must be
#' consistent with the number of vertices in \code{x}: which nodes are
#' to be kept
#' @returns A triangular mesh of class \code{'mesh3d'}, a subset of \code{x}
#'
#' @examples
#'
#' sphere <- vcg_sphere()
#'
#' nv <- ncol(sphere$vb)
#'
#' selector <- seq_len(nv) > (nv / 2)
#'
#' sub <- vcg_subset_vertex(sphere, selector)
#'
#' if(is_not_cran()) {
#'   rgl_view({
#'
#'     # subset sphere will be displayed in red
#'     rgl_call("shade3d", sub, col = 'red')
#'
#'     # Original sphere will be displayed as wireframe
#'     rgl_call("wire3d", sphere, col = (2 - selector))
#'
#'   })
#' }
#'
#'
#' @export
vcg_subset_vertex <- function(x, selector) {
  x <- ensure_mesh3d(x)
  selector <- as.logical(selector)
  facecheck <- !is.null(x$it)
  x <- meshintegrity(x, facecheck = facecheck)

  ns <- length(selector)
  if (ncol(x$vb) != ns) {
    stop("`vcg_subset_vertex`: Number of vertices does not match the length of `selector`")
  }
  if (ns == 0) {
    return(x)
  }
  if (!facecheck) {
    x$vb <- x$vb[, selector, drop = FALSE]
    if (!is.null(x$normals)) {
      x$normals <- x$normals[, selector, drop = FALSE]
    }
    return(x)
  }

  selector[is.na(selector)] <- FALSE
  x <- vcgSubset(x$vb[1:3, , drop = FALSE], x$it - 1L, !selector)
  return(x)
}

test_that("VCG functions", {

  n <- 20
  A <- matrix(rnorm(n * 3), nrow = n)
  B <- matrix(rnorm(n * 6), nrow = n * 2)
  result <- vcg_kdtree_nearest(
    target = B, query = A,
     k = 1
  )
  nearest_index <- apply(A, 1, function(pt) {
    which.min(colSums((t(B) - pt) ^ 2))
  })

  expect_true(all(result$index == nearest_index))

})

#' @title Canonical Response Parameterization (\verb{CRP})
#' @description
#' Parameterizes single-trial evoked responses (e.g. cortico-cortical evoked
#' potentials, \verb{CCEPs}) using the Canonical Response Parameterization
#' method (see 'Citation'). The function estimates the response
#' duration \eqn{\tau_R}, the time after stimulus at which the evoked
#' response has its most consistent, shared structure across trials.
#' The estimator is obtained from the time course of cross-trial projection
#' magnitudes, extracts the canonical response shape \eqn{C(t)} via a linear
#' kernel-trick PCA on the trial matrix truncated at \eqn{\tau_R}, and
#' reports per-trial weights, residuals, signal-to-noise, explained variance
#' and extraction-significance statistics.
#'
#' This is an R translation of \code{CRP_method.m} (and the surrounding
#' artifact-rejection / duration-uncertainty logic in
#' \code{CRP_illustration.m}) from the upstream MATLAB reference
#' implementation; see \sQuote{References}.
#'
#' @param x numeric matrix of single-trial evoked voltages with shape
#' \code{time x trials} (i.e. rows are timepoints, columns are trials);
#' the matrix orientation matches the variable \code{V} / \code{data} in
#' the MATLAB reference. At least two trials are required.
#' @param time numeric vector of length \code{nrow(x)} giving the
#' stimulus-aligned time (in seconds) of each row of \code{x}; must be
#' monotonically increasing and span \code{[t_start, t_end]}.
#' @param t_start,t_end numeric scalars, post-stimulation start and end times
#' (in seconds) defining the analysis window. Defaults match the MATLAB
#' illustration (\code{0.015 s} to \code{1 s}).
#' @param remove_artifacts logical; if \code{TRUE} (the default), an initial
#' \verb{CRP} pass is run to identify outlier/artifact trials, which are then
#' dropped before the final pass. See \sQuote{Details}.
#' @param artifact_interval character, one of \code{"full"} (the default,
#' matching the active option in the MATLAB illustration) or \code{"tR"};
#' selects whether per-trial outlier statistics are computed on the
#' projection magnitudes for the full window or only at the response
#' duration \eqn{\tau_R}.
#' @param artifact_p_threshold numeric, p-value threshold below which a
#' trial is flagged as artifact (provided its mean projection is also
#' below the cohort mean); defaults to \code{1e-5}.
#' @param threshold_quantile numeric in \code{(0, 1)}; the fraction of the
#' peak mean projection magnitude used to derive the duration-uncertainty
#' bounds \code{tau_R_lower} and \code{tau_R_upper}. Defaults to
#' \code{0.98} as in the manuscript.
#' @param time_step integer, sampling step (in samples) used when sweeping
#' candidate response duration; defaults to \code{5L}, matching the
#' MATLAB \code{t_step}. Larger values are faster but smooth the
#' projection profile.
#' @param detect_onset logical; if \code{TRUE}, estimate the response
#' \emph{onset} \eqn{\tau_{onset}} via a reverse cumulative-projection scan
#' (see \sQuote{Details}). The forward canonical shape and per-trial weights
#' are left unchanged; only the \emph{reported} canonical shape \code{C} is
#' restricted to the active response support \eqn{[\tau_{onset}, \tau_R]}.
#' The onset may fall before \code{t_start} (down to \code{onset_search_start}).
#' Defaults to \code{FALSE}.
#' @param onset_search_start numeric scalar or \code{NULL}; the earliest time
#' (in seconds) the onset scan may reach when \code{detect_onset = TRUE}.
#' \code{NULL} (the default) uses the first loaded time point, so the scan can
#' look across the entire signal before \eqn{\tau_R}. Raise it to keep the scan
#' away from an early stimulation artifact. Clamped into
#' \code{[min(time), t_end]}.
#'
#' @returns A named list with the following elements:
#' \describe{
#' \item{\code{parameters}}{A list of single-trial parameterizations
#' (\code{crp_parms} in MATLAB):
#' \describe{
#'   \item{\code{C}}{Numeric vector, the reported canonical response shape
#'     \eqn{C(t)}, taken as the slice of \code{C_full} over its active support
#'     (so it shares identical values with \code{C_full}). By default this is
#'     \eqn{[t_{start}, \tau_R]}, where it equals the first eigenvector of the
#'     linear kernel PCA on \code{V_tR} (oriented to the mean trace, unit-norm,
#'     length \eqn{T_R}). When \code{detect_onset = TRUE} it is restricted to
#'     \eqn{[\tau_{onset}, \tau_R]} (so its norm is then \eqn{\le 1}). The
#'     matching time axis is \code{params_times}.}
#'   \item{\code{C_full}}{Numeric vector spanning the \emph{entire} loaded
#'     time range (every row of \code{time}, including any baseline at
#'     \eqn{t < 0}). Obtained by applying the trial-space \code{loading} to the
#'     full retained data, so \code{C_full} coincides exactly with the
#'     untrimmed forward \code{C} on \eqn{[t_{start}, \tau_R]} and extrapolates
#'     the shape everywhere else. Time axis is \code{params_times_full}.
#'     Outside \eqn{[t_{start}, \tau_R]} cross-trial consistency is not
#'     optimized, so those portions are more variable.}
#'   \item{\code{loading}}{Numeric vector of length \eqn{K} (retained
#'     trials), the trial-space loading \eqn{g = V_{tR}^\top C = s_1 v_1}
#'     recovered from the forward decomposition. Applying it to a
#'     time \eqn{\times} trials matrix and dividing by \eqn{\|g\|^2}
#'     reconstructs the canonical shape over that time range; this is how
#'     \code{C_full} is formed.}
#'   \item{\code{al}}{Numeric vector of length \eqn{K} (number of trials),
#'     the per-trial alpha coefficient \eqn{\alpha_k = C^\top V_k}: scalar
#'     projection of trial \eqn{k} onto \eqn{C(t)}. Larger magnitude means
#'     the trial resembles the canonical shape more strongly; sign reflects
#'     polarity relative to \eqn{C}.}
#'   \item{\code{al_p}}{Numeric vector of length \eqn{K}, alpha-prime
#'     \eqn{\alpha_k / \sqrt{T_R}}: \code{al} rescaled to remove the
#'     duration dependence from the unit-norm convention on \eqn{C}.
#'     Expressed in \eqn{\mu V} and comparable across electrodes or
#'     conditions with different \eqn{\tau_R}.}
#'   \item{\code{ep}}{Numeric matrix of shape \eqn{T_R \times K}, the
#'     per-trial residual \eqn{\epsilon_k(t) = V_k(t) - \alpha_k C(t)}
#'     after the shared component is removed, computed over the forward window
#'     \eqn{[t_{start}, \tau_R]} against the untrimmed forward canonical shape.
#'     Access trial \eqn{k} via \code{ep[, k]}. (When
#'     \code{detect_onset = TRUE} the reported \code{C} may be shorter than
#'     \eqn{T_R}; \code{ep} always stays on the full forward window.)}
#'   \item{\code{epep_root}}{Numeric vector of length \eqn{K},
#'     \eqn{\|\epsilon_k\| = \sqrt{\epsilon_k^\top \epsilon_k}}: L2 norm
#'     of the residual per trial. Smaller values indicate the canonical
#'     shape describes that trial more faithfully.}
#'   \item{\code{Vsnr}}{Numeric vector of length \eqn{K}, per-trial
#'     signal-to-noise \eqn{\alpha_k / \|\epsilon_k\|}. Values \eqn{> 1}
#'     indicate the canonical component is larger than the residual.}
#'   \item{\code{expl_var}}{Numeric vector of length \eqn{K}, per-trial
#'     explained variance \eqn{1 - \|\epsilon_k\|^2 / \|V_k\|^2}: fraction
#'     of each trial's energy accounted for by \eqn{\alpha_k C(t)}.
#'     Ranges in \eqn{[0, 1]}.}
#'   \item{\code{tR}}{Numeric scalar, response duration \eqn{\tau_R} in
#'     seconds: the time at which mean cross-trial projection magnitude is
#'     maximized.}
#'   \item{\code{params_times}}{Numeric vector, time axis for the reported
#'     \code{C}; length matches \code{C} (\eqn{T_R}, or the onset-trimmed
#'     support when \code{detect_onset = TRUE}).}
#'   \item{\code{params_times_full}}{Numeric vector, time axis for
#'     \code{C_full}; equals the full loaded \code{time} (every row,
#'     including the baseline at \eqn{t < 0}).}
#'   \item{\code{V_tR}}{Numeric matrix \eqn{T_R \times K}, trial matrix
#'     truncated to \eqn{\tau_R} - the data actually decomposed. Together
#'     with \code{ep} and \code{avg_trace_tR} it always spans the full
#'     forward window \eqn{[t_{start}, \tau_R]}, independently of
#'     \code{detect_onset}.}
#'   \item{\code{avg_trace_tR}}{Numeric vector of length \eqn{T_R}, simple
#'     trial average truncated to \eqn{\tau_R}.}
#' }}
#' \item{\code{projections}}{A list of projection-stage outputs
#' (\code{crp_projs} in MATLAB):
#' \describe{
#'   \item{\code{proj_tpts}}{Numeric vector, candidate duration time points
#'     (seconds) at which projection magnitudes were evaluated.}
#'   \item{\code{S_all}}{Numeric matrix; rows are non-redundant off-diagonal
#'     trial-pair projections, columns correspond to \code{proj_tpts}.
#'     Units: \eqn{\mu V \cdot s^{1/2}}.}
#'   \item{\code{mean_proj_profile}}{Numeric vector, mean of \code{S_all}
#'     across trial pairs at each candidate duration, the profile whose
#'     maximum defines \eqn{\tau_R}.}
#'   \item{\code{var_proj_profile}}{Numeric vector, variance of
#'     \code{S_all} across trial pairs at each candidate duration.}
#'   \item{\code{tR_index}}{Integer, column index into \code{S_all} and
#'     \code{proj_tpts} corresponding to \eqn{\tau_R}.}
#'   \item{\code{tR_sample}}{Integer, row (sample) index of \eqn{\tau_R}
#'     within the windowed data; \code{V[seq_len(tR_sample), ]} is the
#'     truncated trial matrix \code{V_tR}.}
#'   \item{\code{avg_trace_input}}{Numeric vector, simple trial average
#'     over the full analysis window (not truncated to \eqn{\tau_R}).}
#'   \item{\code{stat_indices}}{Integer vector, row indices of \code{S_all}
#'     used for the significance t-tests, constructed so each trial-pair
#'     comparison appears at most once.}
#'   \item{\code{t_value_tR}, \code{p_value_tR}}{t-statistic and one-sided
#'     p-value (H1: mean projection \eqn{> 0}) at \eqn{\tau_R}. Primary
#'     extraction-significance test reported in the manuscript.}
#'   \item{\code{t_value_full}, \code{p_value_full}}{Same test at the full
#'     analysis-window duration.}
#' }}
#' \item{\code{bad_trials}}{Integer vector of column indices into the
#' original \code{x} flagged and removed as artifacts; \code{integer(0)}
#' when none removed or when \code{remove_artifacts = FALSE}.}
#' \item{\code{tau_R}}{Numeric scalar, estimated response duration
#' \eqn{\tau_R} in seconds (convenience copy of \code{parameters$tR}).}
#' \item{\code{tau_R_lower}, \code{tau_R_upper}}{Numeric scalars, lower
#' and upper threshold-crossing times (seconds) bracketing \eqn{\tau_R}
#' at the \code{threshold_quantile} fraction of the peak mean projection
#' magnitude.}
#' \item{\code{tau_onset}, \code{tau_onset_lower},
#' \code{tau_onset_upper}}{Numeric scalars, the estimated response onset
#' time (seconds) and its lower/upper threshold-crossing bounds, from the
#' reverse projection scan; all \code{NA} unless \code{detect_onset = TRUE}.}
#' \item{\code{onset}}{List with the reverse-scan profile (\code{onset_tpts}
#' and \code{mean_proj_profile}, mapped onto original time); \code{NULL}
#' unless \code{detect_onset = TRUE} (also \code{NULL} when the search window
#' had too few samples).}
#' \item{\code{t_start}, \code{t_end}}{The analysis window used.}
#' \item{\code{sample_rate}}{Numeric, sampling rate inferred from
#' \code{time}.}
#' }
#'
#' @details
#' Briefly, the algorithm proceeds in three stages:
#' \enumerate{
#' \item For a sweep of candidate durations \eqn{k}, compute pairwise
#' L2-normalized cross-projection magnitudes between trials truncated to
#' \eqn{[0, k]}. The duration that maximizes the mean projection magnitude
#' is taken as the response duration \eqn{\tau_R}.
#' \item Apply linear kernel-trick PCA to the trial matrix truncated to
#' \eqn{\tau_R}; the first principal component is the canonical response
#' shape \eqn{C(t)}.
#' \item Project \eqn{C(t)} into each trial to obtain per-trial weights
#' \eqn{\alpha_k}; the residual \eqn{\epsilon_k = V_k - \alpha_k C}
#' summarizes trial-by-trial deviation from the canonical shape.
#' }
#' Significance is assessed by a one-sided t-test on the off-diagonal
#' projection magnitudes against zero, restricted to a non-overlapping
#' subset of comparison pairs to avoid double-counting.
#'
#' When \code{remove_artifacts = TRUE}, the function performs an initial
#' \verb{CRP} pass and runs an unpaired t-test for each trial comparing the
#' projections it participates in against all other off-diagonal
#' projections. Trials with \code{p < artifact_p_threshold} \emph{and}
#' mean projection below the cohort mean are dropped, and \verb{CRP} is re-run.
#'
#' When \code{detect_onset = TRUE}, a complementary reverse pass estimates the
#' response \emph{onset}. The retained trials between \code{onset_search_start}
#' and \eqn{\tau_R} are time-reversed, and the same cumulative cross-projection
#' profile is computed growing backward from \eqn{\tau_R}. The backward
#' duration that maximizes cross-trial consistency marks \eqn{\tau_{onset}},
#' the time at which trials begin to share structure - useful when the response
#' is delayed by an unknown latency. Because the scan can extend before
#' \code{t_start}, the onset may be earlier than the analysis window. This pass
#' only locates a time: it re-uses the forward loading, so the canonical shape
#' and per-trial weights are unchanged (the reported \code{C} is merely sliced
#' to \eqn{[\tau_{onset}, \tau_R]}).
#'
#' Time points whose data are not finite (any \code{NA}, \code{NaN} or
#' \code{Inf} across trials) are dropped before analysis. \code{t_start},
#' \code{t_end} and \code{onset_search_start} are clamped into the available
#' time range rather than triggering an error.
#'
#' @references
#' The \verb{CRP} algorithm is described in \doi{10.1371/journal.pcbi.1011105},
#' with a reference MATLAB implementation at
#' \url{https://github.com/kaijmiller/crp_scripts}. See
#' \code{citation("ravetools")} for the full bibliographic entry.
#'
#' @examples
#' set.seed(42)
#'
#' # Synthetic CCEP-like data: shared canonical shape with per-trial scaling
#' n_time   <- 500L
#' n_trials <- 20L
#' tt <- seq(-0.5, 1, length.out = n_time)
#' canonical <- exp(-((tt - 0.10) / 0.05)^2) -
#'              0.5 * exp(-((tt - 0.30) / 0.10)^2)
#' V <- (outer(canonical, runif(n_trials, 0.5, 1.5)) +
#'        matrix(rnorm(n_time * n_trials, sd = 0.3), n_time, n_trials)) * 2
#'
#' res <- crp(V, tt)
#'
#' op <- par(mfrow = c(1, 3), mar = c(4.5, 4, 3, 1))
#' on.exit({ par(op) })
#'
#' # ---- Panel 1: all trials (full window) + mean + C(t) overlay ----------
#' parms <- res$parameters
#' matplot(tt, V, type = "l", lty = 1,
#'         col = "#80808060", xlab = "Time (s)",
#'         ylab = expression(mu * V),
#'         main = expression("Canonical shape " * C(t)))
#' # scale C(t) to the amplitude of the mean trace for overlay;
#' # C(t) ends at tau_R so the line is cut off there naturally
#' C_scaled <- parms$C * max(abs(rowMeans(V))) / max(abs(parms$C))
#' lines(parms$params_times, C_scaled, col = "#FFFF0080", lwd = 3)
#'
#' # Mean
#' lines(tt, rowMeans(V), col = "black", lwd = 1)
#' legend("topright", c("mean", "C(t) scaled"),
#'        col = c("black", "#FFFF00"), lty = c(1, 2), lwd = 2,
#'        bty = "n", cex = 0.8)
#'
#' # ---- Panel 2: per-trial alpha-prime weights -----------------------------
#' barplot(sort(parms$al_p), col = "steelblue", border = NA, las = 1,
#'         xlab = "Trial (sorted)",
#'         ylab = expression(alpha * "'" ~ (mu * V)),
#'         main = expression("Per-trial " * alpha * "' (alpha-prime)"))
#' abline(h = c(0, mean(parms$al_p)), lty = 2)
#'
#' # ---- Panel 3: mean projection profile with tau_R bounds ----------------
#' proj <- res$projections
#' plot(proj$proj_tpts, proj$mean_proj_profile, type = "l", lwd = 2,
#'      xlab = "Candidate duration (s)",
#'      ylab = expression(bar(S) ~ (mu * V %.% s^{0.5})),
#'      main = expression("Projection profile & " * tau[R]),
#'      las = 1)
#' abline(v = c(res$tau_R_lower, res$tau_R, res$tau_R_upper),
#'        col = c("cyan3", "orange2", "red"),
#'        lty = c(2, 1, 2), lwd = 2)
#' legend("topright",
#'        legend = expression(tau[lb], tau[R], tau[ub]),
#'        col = c("cyan3", "orange2", "red"),
#'        lty = c(2, 1, 2), lwd = 2, bty = "n")
#'
#' par(op)
#'
#' # ---- Onset detection (reverse scan) on the same data -------------------
#' res_onset <- crp(V, tt, detect_onset = TRUE)
#' c(tau_onset = res_onset$tau_onset, tau_R = res_onset$tau_R)
#'
#' @export
crp <- function(
  x, time,
  t_start = 0.015, t_end = 1,
  remove_artifacts = TRUE,
  artifact_interval = c("full", "tR"),
  artifact_p_threshold = 1e-5,
  threshold_quantile = 0.98,
  time_step = 5L,
  detect_onset = FALSE,
  onset_search_start = NULL
) {

  # ---- input validation ----------------------------------------------------
  if (!is.matrix(x) || !is.numeric(x)) {
    stop("`x` must be a numeric matrix with shape time x trials.")
  }
  if (ncol(x) < 2L) {
    stop("`x` must have at least 2 trials (columns).")
  }
  time <- as.numeric(time)
  if (anyNA(time) || length(time) != nrow(x)) {
    stop("`time` must be a non-NA numeric vector with length equal to nrow(x).")
  }

  # ---- drop time points with non-finite data (NA / NaN / Inf) -------------
  finite_rows <- is.finite(rowSums(x))
  if (!all(finite_rows)) {
    x <- x[finite_rows, , drop = FALSE]
    time <- time[finite_rows]
  }
  if (length(time) < 11L) {
    stop("`x` has too few finite time points (need >= 11 after removing ",
         "non-finite rows).")
  }
  if (any(diff(time) <= 0)) {
    stop("`time` must be strictly monotonically increasing.")
  }
  artifact_interval <- match.arg(artifact_interval)
  if (!is.numeric(threshold_quantile) || length(threshold_quantile) != 1L ||
      threshold_quantile <= 0 || threshold_quantile >= 1) {
    stop("`threshold_quantile` must be a single number in (0, 1).")
  }
  time_step <- as.integer(time_step)
  if (length(time_step) != 1L || is.na(time_step) || time_step < 1L) {
    stop("`time_step` must be a positive integer scalar.")
  }
  detect_onset <- isTRUE(as.logical(detect_onset))

  # ---- clamp the analysis window into the available time range ------------
  # t_start / t_end only bound where the response may *start* / *end*; samples
  # outside the loaded `time` cannot constrain them, so clamp instead of
  # erroring (e.g. when non-finite rows trimmed the edges or t_end exceeds the
  # last sample).
  t_start <- max(t_start, min(time))
  t_end <- min(t_end, max(time))
  if (t_start >= t_end) {
    stop("`t_start` must be less than `t_end`.")
  }

  # onset search floor: how far before t_start the reverse scan may look
  if (is.null(onset_search_start)) {
    onset_search_start <- min(time)
  } else {
    onset_search_start <- as.numeric(onset_search_start)
    if (length(onset_search_start) != 1L || !is.finite(onset_search_start)) {
      stop("`onset_search_start` must be a single finite number or NULL.")
    }
    onset_search_start <- min(max(onset_search_start, min(time)), t_end)
  }

  # ---- subset to analysis window ------------------------------------------
  tpts <- which(time >= t_start & time <= t_end)
  if (length(tpts) < 11L) {
    stop("Analysis window contains too few samples (need >= 11).")
  }
  V <- x[tpts, , drop = FALSE]
  t_win <- time[tpts]
  srate <- 1 / mean(diff(time))

  # ---- optional artifact removal ------------------------------------------
  bad_trials <- integer(0)
  if (isTRUE(remove_artifacts)) {
    init <- crp_core(V, t_win, srate, time_step)
    mp <- crp_trial_tests(init$crp_projs, artifact_interval, ncol(V))
    flag <- mp$p < artifact_p_threshold & mp$m < mean(mp$m)
    bad_trials <- which(flag)
    if (length(bad_trials) > 0L) {
      if (ncol(V) - length(bad_trials) < 2L) {
        stop(
          "Artifact removal would leave fewer than 2 trials; ",
          "consider raising `artifact_p_threshold` or setting ",
          "`remove_artifacts = FALSE`."
        )
      }
      V <- V[, -bad_trials, drop = FALSE]
    }
  }
  retained_cols <- setdiff(seq_len(ncol(x)), bad_trials)

  # ---- final CRP pass ------------------------------------------------------
  out <- crp_core(V, t_win, srate, time_step)
  crp_projs <- out$crp_projs
  crp_params <- out$crp_params

  # ---- full-range canonical shape -----------------------------------------
  # Apply the trial-space loading to the entire loaded time range (every row of
  # the retained trials, including any baseline at t < 0). By construction
  # C_full equals the forward C exactly on [t_start, tau_R] and extrapolates the
  # shape everywhere else, so C_full, the reported C and the onset-trimmed C all
  # share identical values on their common support.
  x_ret <- x[, retained_cols, drop = FALSE]
  g <- crp_params$loading
  gg <- sum(g * g)
  C_full <- if (gg > 0) as.numeric(x_ret %*% g) / gg else numeric(nrow(x_ret))
  crp_params$C_full <- C_full
  crp_params$params_times_full <- time

  # ---- duration-uncertainty bounds (threshold_quantile of peak) ----------
  bnd <- crp_profile_bounds(crp_projs$mean_proj_profile, threshold_quantile)
  tau_R_lower <- crp_projs$proj_tpts[bnd$lb]
  tau_R_upper <- crp_projs$proj_tpts[bnd$hb]

  # ---- optional response-onset detection (reverse projection scan) -------
  # Re-uses the forward loading: the scan only *locates* the onset time and
  # never recomputes C or alpha. It may reach before t_start (down to
  # onset_search_start) so a delayed response with unknown latency can be found.
  tau_onset <- NA_real_
  tau_onset_lower <- NA_real_
  tau_onset_upper <- NA_real_
  onset <- NULL
  if (isTRUE(detect_onset)) {
    idx_tR_full <- tpts[crp_projs$tR_sample]
    idx_lo_full <- which(time >= onset_search_start)[1L]
    if (idx_tR_full - idx_lo_full + 1L >= 11L) {
      onset <- crp_onset(
        x_ret[idx_lo_full:idx_tR_full, , drop = FALSE],
        time[idx_lo_full:idx_tR_full],
        srate, time_step, threshold_quantile
      )
      tau_onset <- onset$tau_onset
      tau_onset_lower <- onset$tau_onset_lower
      tau_onset_upper <- onset$tau_onset_upper
    }
  }

  # ---- reported canonical shape (slice of C_full) -------------------------
  # Always derived from C_full so C_full, the reported C and the onset-trimmed C
  # share identical values on their common support. Defaults to
  # [t_start, tau_R]; restricted to [tau_onset, tau_R] when an onset was found.
  report_lo <- if (is.finite(tau_onset)) tau_onset else t_start
  sel <- crp_params$params_times_full >= report_lo &
    crp_params$params_times_full <= crp_params$tR
  crp_params$C <- crp_params$C_full[sel]
  crp_params$params_times <- crp_params$params_times_full[sel]

  structure(
    class = "ravetools_crp",
    list(
      parameters = crp_params,
      projections = crp_projs,
      bad_trials = bad_trials,
      tau_R_lower = tau_R_lower,
      tau_R = crp_params$tR,
      tau_R_upper = tau_R_upper,
      tau_onset = tau_onset,
      tau_onset_lower = tau_onset_lower,
      tau_onset_upper = tau_onset_upper,
      onset = onset,
      t_start = t_start,
      t_end = t_end,
      sample_rate = srate,
      .data = list(
        V = x_ret,
        time = time,
        time_range = range(time),
        window = c(t_start, t_end)
      )
    )
  )
}

#' @title Plot \verb{CRP} results
#' @description
#' S3 plot method for objects of class \code{ravetools_crp} returned by
#' \code{\link{crp}}.  Produces a three-panel figure:
#' \enumerate{
#' \item Single-trial traces over the entire loaded time range with the mean,
#'   the canonical shape \eqn{C(t)} on its active support (solid) and the
#'   full-range extension \eqn{C_{full}(t)} (dashed) overlaid; the analysis
#'   window edges and, when present, \eqn{\tau_{onset}} are marked.
#' \item Per-trial \eqn{\alpha'} weights sorted in ascending order.
#' \item Mean cross-trial projection profile with vertical lines marking
#'   \eqn{\tau_{lb}}, \eqn{\tau_R} and \eqn{\tau_{ub}}, plus the reverse
#'   onset profile and \eqn{\tau_{onset}} when \code{detect_onset} was used.
#' }
#' @param x an object of class \code{ravetools_crp} as returned by
#' \code{\link{crp}}.
#' @param ... additional graphical parameters passed to \code{\link{par}}
#' (e.g. \code{mar}, \code{cex.axis}); currently unused beyond restoring
#' the previous \code{par} state on exit.
#' @return Invisibly returns \code{x}.
#' @export
plot.ravetools_crp <- function(x, ...) {
  parms <- x$parameters
  proj  <- x$projections
  V  <- x$.data$V
  tt <- x$.data$time
  window <- x$.data$window
  avg_trace <- if (is.matrix(V) && length(V) > 0) {
    rowMeans(V)
  } else {
    parms$C_full
  }

  op <- graphics::par(mfrow = c(1, 3))
  on.exit(graphics::par(op), add = TRUE)

  # ---- Panel 1: all trials (entire time range) + mean + C(t) overlay -----
  # Scale C_full (and the reported C, a slice of it) by the SAME factor so they
  # coincide on shared support; the factor is derived from the in-window region
  # for a stable overlay even though C_full is drawn across the whole range.
  inwin <- tt >= window[1L] & tt <= window[2L]
  cden <- max(abs(parms$C_full[inwin]), na.rm = TRUE)
  sf <- if (cden > 0) max(abs(avg_trace[inwin]), na.rm = TRUE) / cden else 1
  C_full_scaled <- parms$C_full * sf
  C_scaled <- parms$C * sf
  has_onset <- is.finite(x$tau_onset)

  ylim <- range(c(V, C_full_scaled, avg_trace), na.rm = TRUE)

  plot(range(tt), ylim, type = "n", las = 1,
       xlab = "Time (s)", ylab = expression(mu * V),
       main = expression("Canonical shape " * C(t)))

  if (is.matrix(V) && length(V) > 0) {
    # We might trim V to save space, hence V might not be available
    graphics::matlines(tt, V, lty = 1, col = "#80808060")
  }

  # analysis-window edges (where the forward tau_R search operated)
  graphics::abline(v = window, col = "#cccccc", lty = 3, lwd = 1)

  graphics::abline(v = x$tau_R, col = "red", lwd = 1)
  graphics::text(
    x = x$tau_R, y = min(C_full_scaled),
    labels = bquote(tau[R] * "=" * .(sprintf("%.3fs", x$tau_R))),
    col = "red", pos = 4, offset = 0.5
  )

  # full-range canonical shape (dashed = extrapolation outside active support)
  graphics::lines(parms$params_times_full, C_full_scaled,
                  col = "#FFB300", lty = 3, lwd = 3)
  # canonical shape on the active response support (solid)
  graphics::lines(parms$params_times, C_scaled, col = "#FFD500", lwd = 3)
  graphics::lines(tt, avg_trace, col = "black", lwd = 1)

  # onset marker (only when detect_onset was requested)
  if (has_onset) {
    graphics::abline(v = x$tau_onset, col = "forestgreen", lwd = 1, lty = 2)
    graphics::text(
      x = x$tau_onset, y = max(C_full_scaled),
      labels = bquote(tau[onset] * "=" * .(sprintf("%.3fs", x$tau_onset))),
      col = "forestgreen", pos = 4, offset = 0.5
    )
  }

  legend_lab <- c("mean", "C(t)", "C_full(t)")
  legend_col <- c("black", "#FFD500", "#FFB300")
  legend_lty <- c(1, 1, 3)
  graphics::legend("topright", legend = legend_lab,
                   col = legend_col, lty = legend_lty, lwd = 2,
                   bty = "n", cex = 0.8)

  # ---- Panel 2: per-trial alpha-prime weights -----------------------------
  graphics::barplot(
    sort(parms$al_p),
    col = "steelblue",
    border = NA,
    las = 1,
    xlab = "Trial (sorted)",
    ylab = expression(alpha * "'" ~ (mu * V)),
    main = expression("Per-trial " * alpha * "' (alpha-prime)")
  )
  graphics::abline(h = c(0, mean(parms$al_p)), lty = c(1, 2))

  # ---- Panel 3: mean projection profile with tau_R (and onset) ----------
  onset <- x$onset
  if (has_onset && is.list(onset) &&
      length(onset$onset_tpts) > 0L) {
    has_onset_prof <- TRUE

    yr <- range(c(proj$mean_proj_profile, onset$mean_proj_profile), na.rm = TRUE)
    xr <- range(c(proj$proj_tpts, onset$onset_tpts), na.rm = TRUE)
  } else {
    has_onset_prof <- FALSE

    yr <- range(proj$mean_proj_profile, na.rm = TRUE)
    xr <- range(proj$proj_tpts, na.rm = TRUE)
  }

  graphics::plot(xr, yr, type = "n", xlab = "Candidate duration (s)", ylim = yr,
                 ylab = expression(bar(S) ~ (mu * V %.% s^{0.5})),
                 main = expression("Projection profile & " * tau[R]), las = 1)

  graphics::lines(proj$proj_tpts, proj$mean_proj_profile, lwd = 2)
  if (has_onset_prof) {
    # reverse (onset) projection profile, mapped back onto original time
    graphics::lines(onset$onset_tpts, onset$mean_proj_profile,
                    col = "gray", lwd = 2)
  }
  graphics::abline(v = x$tau_R, col = "#FF0000", lty = 1, lwd = 2)
  if (has_onset) {
    graphics::abline(v = x$tau_onset, col = "forestgreen", lwd = 2)
    leg <- c(
      bquote(tau[onset] * "=" * .(sprintf("%.3fs", x$tau_onset))),
      bquote(tau[R] * "=" * .(sprintf("%.3fs", x$tau_R)))
    )
    graphics::legend(
      "topright",
      legend = as.expression(leg),
      col = c("forestgreen", "red"),
      lwd = 2,
      bty = "n"
    )
  } else {
    leg <- c(bquote(tau[R] * "=" * .(sprintf("%.3fs", x$tau_R))))
    graphics::legend(
      "topright", legend = as.expression(leg),
      col = "red", lwd = 2, bty = "n")
  }

  invisible(x)
}


# --------------------------------------------------------------------------
# Internal helpers (not exported)
# --------------------------------------------------------------------------

# Single-trial cross-projection magnitudes (MATLAB `ccep_proj`).
# V is time x trials; returns a numeric vector of off-diagonal projections.
crp_ccep_proj <- function(V) {
  norms <- sqrt(colSums(V * V))
  # remove zero norms since 0/1=0
  norms[norms == 0] <- 1
  V0 <- t(t(V) / norms)
  V0[!is.finite(V0)] <- 0
  P <- crossprod(V0, V) # trials x trials
  diag(P) <- NA_real_
  S0 <- as.numeric(P)
  S0[!is.na(S0)]
}

# Linear kernel-trick PCA (MATLAB `kt_pca`).
# Returns columns of E as unit-normalized eigenvectors of X (descending),
# and S as corresponding singular values (sqrt of eigenvalues).
crp_kt_pca <- function(X) {
  XtX <- crossprod(X)
  eg <- eigen(XtX, symmetric = TRUE)
  # eigen() already returns values in descending order
  evals <- pmax(eg$values, 0)
  S <- sqrt(evals)
  ES <- X %*% eg$vectors
  # divide each column by S (avoid divide-by-zero)
  S_safe <- ifelse(S > 0, S, 1)
  E <- ES / rep(S_safe, each = nrow(X))
  E[, S == 0] <- 0
  list(E = E, S = S)
}

# Non-overlapping pair indices for significance testing
# (port of MATLAB `get_stat_indices`).
crp_stat_indices <- function(N) {
  # base indices: every other element of 1:(N^2 - N)
  stat_indices <- seq.int(1L, N * N - N, by = 2L)
  if (N %% 2L == 1L) {
    # odd N: offset every even-indexed column block by 1
    b <- integer(length(stat_indices))
    half <- (N - 1L) %/% 2L
    for (k in seq_len(N)) {
      if (k %% 2L == 0L) {
        idx <- ((k - 1L) * half + 1L):(k * half)
        b[idx] <- 1L
      }
    }
    stat_indices <- stat_indices + b
  }
  stat_indices
}

# Per-trial outlier statistics (port of `trial_tests` from CRP_illustration.m).
crp_trial_tests <- function(crp_projs, interval, num_trials) {
  S_test <- if (identical(interval, "tR")) {
    crp_projs$S_all[, crp_projs$tR_index]
  } else {
    crp_projs$S_all[, ncol(crp_projs$S_all)]
  }

  N <- num_trials
  L <- length(S_test) # = N^2 - N

  stat_indices <- crp_stat_indices(N)
  excl_indices <- setdiff(seq_len(L), stat_indices)

  m <- numeric(N)
  p <- numeric(N)

  for (q in seq_len(N)) {
    # projections of normalized other trials into this trial, before this
    # cluster's own normalized projections (only present for q > 1)
    if (q > 1L) {
      t_indices1 <- seq.int(q - 1L, q * (N - 1L), by = N - 1L)
    } else {
      t_indices1 <- integer(0)
    }
    # projections of normalized other trials into this trial, after this
    # cluster's own normalized projections
    t_indices2 <- if (q * N <= L) {
      seq.int(q * N, L, by = N - 1L)
    } else {
      integer(0)
    }
    # projections of this trial into other trials
    t_indices3 <- (q - 1L) * (N - 1L) + seq_len(N - 1L)

    g_in <- unique(c(t_indices1, t_indices2, t_indices3))
    g_out <- setdiff(seq_len(L), c(g_in, excl_indices))

    # unpaired t-test (Welch by default in R; MATLAB ttest2 is pooled, but
    # the threshold is heuristic so this is acceptable).
    a <- S_test[g_in]
    b <- S_test[g_out]
    if (length(a) < 2L || length(b) < 2L ||
        stats::var(a) == 0 && stats::var(b) == 0) {
      p[q] <- NA_real_
    } else {
      tt <- tryCatch(
        stats::t.test(a, b),
        error = function(e) { NULL }
      )
      p[q] <- if (is.null(tt)) NA_real_ else tt$p.value
    }
    m[q] <- mean(S_test[t_indices3])
  }

  list(m = m, p = p)
}

# Cumulative cross-trial projection profile over candidate truncation lengths.
# Sweeps a truncation length `k` over the rows (time) of `V`; for each `k` it
# computes the off-diagonal cross-trial projection magnitudes and their mean and
# variance. Shared by the forward CRP pass and the reverse onset scan (which
# passes a time-reversed `V`). Indices are returned in the sample domain so the
# caller can map them back onto its own time axis.
crp_proj_profile <- function(V, srate, time_step) {
  Tn <- nrow(V)
  K <- ncol(V)
  proj_idx <- seq.int(10L, Tn, by = time_step)
  if (length(proj_idx) < 1L) {
    stop("Window too short for the requested `time_step`.")
  }

  npairs <- K * K - K
  S_all <- matrix(0, nrow = npairs, ncol = length(proj_idx))
  m <- numeric(length(proj_idx))
  v2 <- numeric(length(proj_idx))

  # For each cumulative time point, calculate the projection
  for (i in seq_along(proj_idx)) {
    k <- proj_idx[i]
    S <- crp_ccep_proj(V[seq_len(k), , drop = FALSE]) / sqrt(srate)
    S_all[, i] <- S
    m[i] <- mean(S)
    v2[i] <- stats::var(S)
  }

  list(
    proj_idx = proj_idx,
    S_all = S_all,
    mean_proj_profile = m,
    var_proj_profile = v2
  )
}

# Core CRP computation (MATLAB `CRP_method`).
crp_core <- function(V, t_win, srate, time_step) {
  K <- ncol(V)

  # Cumulative cross-trial projection profile (shared with the onset scan)
  prof <- crp_proj_profile(V, srate, time_step)
  proj_tpts <- prof$proj_idx
  S_all <- prof$S_all
  m <- prof$mean_proj_profile
  v2 <- prof$var_proj_profile

  # tt is the time point where the trial consistency reaches the maximum
  tt <- which.max(m)
  tR_sample <- proj_tpts[tt]

  V_tR <- V[seq_len(tR_sample), , drop = FALSE]
  pca <- crp_kt_pca(V_tR)
  C <- pca$E[, 1L]
  # Eigenvector sign is arbitrary; orient C to align with the mean trace
  if (sum(C * rowMeans(V_tR)) < 0) { C <- -C }

  al <- as.numeric(crossprod(C, V_tR))
  ep <- V_tR - C %*% rbind(al)

  # Trial-space loading recovered from the unit-norm, sign-oriented C and V_tR
  # (g = s1 * v1). Applying it to any time x trials matrix and dividing by
  # ||g||^2 reconstructs the canonical shape there; crp() uses it to build the
  # full-range C_full. Kept in trial space (length K) so it generalizes beyond
  # the analysis window.
  loading <- as.numeric(crossprod(V_tR, C))   # length K (retained trials)

  stat_indices <- crp_stat_indices(K)

  s_tR <- S_all[stat_indices, tt]
  s_full <- S_all[stat_indices, ncol(S_all)]

  t_value_tR <- mean(s_tR) / (stats::sd(s_tR) / sqrt(length(s_tR)))
  p_value_tR <- stats::t.test(s_tR, mu = 0, alternative = "greater")$p.value

  t_value_full <- mean(s_full) / (stats::sd(s_full) / sqrt(length(s_full)))
  p_value_full <- stats::t.test(s_full, mu = 0, alternative = "greater")$p.value

  ep2 <- colSums(ep * ep)
  v2_tR <- colSums(V_tR * V_tR)

  crp_projs <- list(
    proj_tpts = t_win[proj_tpts],
    S_all = S_all,
    mean_proj_profile = m,
    var_proj_profile = v2,
    tR_index = tt,
    tR_sample = tR_sample,
    avg_trace_input = rowMeans(V),
    stat_indices = stat_indices,
    t_value_tR = t_value_tR,
    p_value_tR = p_value_tR,
    t_value_full = t_value_full,
    p_value_full = p_value_full
  )

  crp_params <- list(
    V_tR = V_tR,
    al = al,
    C = C,
    loading = loading,
    ep = ep,
    tR = t_win[tR_sample],
    params_times = t_win[seq_len(tR_sample)],
    avg_trace_tR = rowMeans(V_tR),
    al_p = al / sqrt(length(C)),
    epep_root = sqrt(ep2),
    Vsnr = al / sqrt(ep2),
    expl_var = 1 - ep2 / v2_tR
  )

  list(crp_projs = crp_projs, crp_params = crp_params)
}

# Threshold-crossing bounds around the peak of a projection profile.
# Returns the peak index and the lower/upper indices where `m` last (going up)
# and first (coming down) crosses `threshold_quantile * max(m)` on either side
# of the peak. Shared by the forward tau_R bounds and the reverse onset bounds.
crp_profile_bounds <- function(m, threshold_quantile) {
  n <- length(m)
  cidx <- which.max(m)
  thr <- threshold_quantile * m[cidx]

  up_cross <- 1L + which(m[-1L] > thr & m[-n] <= thr)
  up_cross <- up_cross[up_cross <= cidx]
  lb <- if (length(up_cross) > 0L) up_cross[length(up_cross)] else 1L

  down_cross <- which(m[-n] > thr & m[-1L] <= thr)
  down_cross <- down_cross[down_cross >= cidx]
  hb <- if (length(down_cross) > 0L) down_cross[1L] else n

  list(peak = cidx, lb = lb, hb = hb)
}

# Response-onset estimation by reverse cumulative projection.
# `Xo` is the retained trial matrix over [onset_search_start, tau_R] in forward
# time order, with matching times `to` (so the last row is tau_R). The trials
# are time-reversed, so the cumulative truncation grows backward from tau_R; the
# same cross-trial projection profile is computed and the backward duration that
# maximizes cross-trial consistency marks the response onset tau_onset, with the
# threshold crossings giving its uncertainty bounds. The scan can reach before
# t_start. It only locates a time - it never recomputes the canonical shape C or
# the weights alpha, which are taken from the forward pass.
crp_onset <- function(Xo, to, srate, time_step, threshold_quantile) {
  n <- nrow(Xo)

  # too few samples to run a backward sweep
  if (n < 11L) {
    return(list(
      tau_onset = NA_real_,
      tau_onset_lower = NA_real_,
      tau_onset_upper = NA_real_,
      onset_tpts = numeric(0),
      mean_proj_profile = numeric(0)
    ))
  }

  V_rev <- Xo[n:1L, , drop = FALSE]
  prof <- crp_proj_profile(V_rev, srate, time_step)
  m <- prof$mean_proj_profile
  bnd <- crp_profile_bounds(m, threshold_quantile)

  # map backward sample extents to forward-time rows of `Xo`/`to`: a backward
  # extent of `k` samples from tau_R (row n) reaches forward row (n - k + 1). A
  # larger extent reaches an earlier time, so the earliest (lower) onset bound
  # uses the upper extent hb and the latest (upper) bound uses the lower extent.
  onset_tpts <- to[n - prof$proj_idx + 1L]
  ord <- order(onset_tpts)

  list(
    tau_onset = to[n - prof$proj_idx[bnd$peak] + 1L],
    tau_onset_lower = to[n - prof$proj_idx[bnd$hb] + 1L],
    tau_onset_upper = to[n - prof$proj_idx[bnd$lb] + 1L],
    onset_tpts = onset_tpts[ord],
    mean_proj_profile = m[ord]
  )
}

# Double-precision Morlet wavelet implementation.
#
# Public entry point is `morlet_wavelet()` (defined in R/wavelet.R), which
# dispatches here when `precision = "double"`.

wavelet_kernels2_double <- function(freqs, srate, wave_num, data_length, signature = NULL) {
  freqs <- as.double(freqs)
  srate <- as.double(srate)
  wave_num <- as.double(wave_num)
  data_length <- as.integer(data_length)
  kernel_info <- wavelet_kernels(freqs = freqs, srate = srate, wave_num = wave_num)
  digest <- digest::digest(list(freqs, srate, wave_num, data_length, signature = signature))
  root_dir <- file.path(tempdir2(check = TRUE), "ravetools")
  if (!dir.exists(root_dir)) {
    dir.create(root_dir, showWarnings = FALSE, recursive = TRUE)
  }
  real_path <- file.path(root_dir, sprintf("wavelet-real-%s", digest))
  imag_path <- file.path(root_dir, sprintf("wavelet-imag-%s", digest))


  arr_dim <- c(data_length, length(kernel_info$kernels))
  tryCatch({
    return(list(
      real = filearray::filearray_checkload(
        filebase = real_path, mode = "readonly", symlink_ok = FALSE,
        freqs = freqs, srate = srate, wave_num = wave_num, data_length = data_length,
        arr_dim = arr_dim, rave_data_type = "rave-wavelet-kernels-double-real"
      ),
      imag = filearray::filearray_checkload(
        filebase = imag_path, mode = "readonly", symlink_ok = FALSE,
        freqs = freqs, srate = srate, wave_num = wave_num, data_length = data_length,
        arr_dim = arr_dim, rave_data_type = "rave-wavelet-kernels-double-imag"
      )
    ))
  }, error = function(e) {
    if (getOption("ravetools.debug", FALSE)) {
      print(real_path)
      print(imag_path)
      warning(e)
    }

    if (dir.exists(real_path)) { unlink(real_path, recursive = TRUE) }
    if (dir.exists(imag_path)) { unlink(imag_path, recursive = TRUE) }

  })

  arr_real <- filearray::filearray_create(
    filebase = real_path,
    dimension = arr_dim,
    type = "double",
    partition_size = 1
  )
  arr_imag <- filearray::filearray_create(
    filebase = imag_path,
    dimension = arr_dim,
    type = "double",
    partition_size = 1
  )
  arr_real$.mode <- "readwrite"
  arr_imag$.mode <- "readwrite"

  tmp <- complex(data_length)
  lapply(seq_along(kernel_info$kernels), function(ii) {
    tmp_wavelet <- kernel_info$kernels[[ii]]
    w_l <- length(tmp_wavelet)
    n_pre  <- ceiling(data_length / 2) - floor(w_l / 2)
    n_post <- data_length - n_pre - w_l
    x <- c(rep(0, n_pre), tmp_wavelet, rep(0, n_post))
    fftw_c2c(data = x, inverse = 0L, ret = tmp)
    conjugate(tmp)
    arr_real[, ii] <- Re(tmp)
    arr_imag[, ii] <- Im(tmp)
    NULL
  })

  arr_real$.header$freqs <- freqs
  arr_real$.header$srate <- srate
  arr_real$.header$wave_num <- wave_num
  arr_real$.header$data_length <- data_length
  arr_real$.header$arr_dim <- arr_dim
  arr_real$.header$freqs <- freqs
  arr_real$.header$rave_data_type <- "rave-wavelet-kernels-double-real"
  arr_real$.save_header()

  arr_imag$.header$freqs <- freqs
  arr_imag$.header$srate <- srate
  arr_imag$.header$wave_num <- wave_num
  arr_imag$.header$data_length <- data_length
  arr_imag$.header$arr_dim <- arr_dim
  arr_imag$.header$freqs <- freqs
  arr_imag$.header$rave_data_type <- "rave-wavelet-kernels-double-imag"
  arr_imag$.save_header()


  arr_real$.mode <- "readonly"
  arr_imag$.mode <- "readonly"

  return(list(
    real = arr_real,
    imag = arr_imag
  ))
}


morlet_wavelet_double <- function(data, freqs, srate, wave_num,
                                  trend = c("constant", "linear", "none"),
                                  signature = NULL, segment_length = NULL, ...) {

  trend <- match.arg(trend)
  freqs <- as.double(freqs)
  srate <- as.double(srate)
  wave_num <- as.double(wave_num)
  more_args <- list(...)
  data_digest <- digest::digest(data)

  f_l <- length(freqs)
  d_l <- length(data)

  segment_length_norm <- wavelet_normalize_segment_length(segment_length, d_l)
  segmented <- !is.na(segment_length_norm)

  # detrend on the full signal first; preserves legacy semantics
  if (trend != "none") {
    data <- as.vector(detrend(data, trend = trend, ...))
  }

  if (segmented) {
    kern_info <- wavelet_kernels(freqs = freqs, srate = srate, wave_num = wave_num)
    max_kernel_len <- max(vapply(kern_info$kernels, length, integer(1L)))
    plan <- wavelet_segment_plan(d_l, segment_length_norm, max_kernel_len)
    kern_data_length <- segment_length_norm
  } else {
    plan <- NULL
    kern_data_length <- d_l
  }

  # Kernel cache differentiates by `data_length`; output cache deliberately
  # omits `segment_length` since both paths produce the same coefficients.
  fft_waves <- wavelet_kernels2_double(freqs, srate, wave_num, kern_data_length,
                                       signature = signature)

  real_path <- tempfile2(pattern = "wavelet-double-real-")
  imag_path <- tempfile2(pattern = "wavelet-double-imag-")
  output <- tryCatch({
    list(
      real = filearray::filearray_checkload(
        filebase = real_path, symlink_ok = FALSE,
        freqs = freqs, srate = srate, wave_num = wave_num,
        data_digest = data_digest, trend = trend,
        more_args = more_args,
        rave_data_type = "rave-wavelet-coefficients-real",
        precision = "double"
      ),
      imag = filearray::filearray_checkload(
        filebase = imag_path, symlink_ok = FALSE,
        freqs = freqs, srate = srate, wave_num = wave_num,
        data_digest = data_digest, trend = trend,
        more_args = more_args,
        rave_data_type = "rave-wavelet-coefficients-imag",
        precision = "double"
      )
    )
  }, error = function(e) {
    if (getOption("ravetools.debug", FALSE)) {
      print(real_path)
      print(imag_path)
      warning(e)
    }
    if (dir.exists(real_path)) { unlink(real_path, recursive = TRUE) }
    if (dir.exists(imag_path)) { unlink(imag_path, recursive = TRUE) }
    NULL
  })
  if (length(output) == 2) { return(output) }

  output_real <- filearray::filearray_create(
    filebase = real_path, dimension = c(d_l, f_l),
    type = "double", partition_size = 1
  )
  output_imag <- filearray::filearray_create(
    filebase = imag_path, dimension = c(d_l, f_l),
    type = "double", partition_size = 1
  )
  output_real$.mode <- "readwrite"
  output_imag$.mode <- "readwrite"

  wave_len <- nrow(fft_waves$real)
  ind <- seq_len(ceiling(wave_len / 2))
  norm_factor <- wave_len * sqrt(srate / 2)

  if (segmented) {
    seg_mat <- wavelet_build_segment_matrix(data, plan)
    fft_seg <- mvfftw_r2c(seg_mat, HermConj = 1L)
    n_seg <- plan$n_seg
    local_starts <- plan$local_starts
    local_ends <- plan$local_ends

    new_ind <- seq_len(plan$segment_length)
    new_ind <- c(new_ind[-ind], new_ind[ind])

    # Pre-allocated buffer avoids a full segment_length × n_seg complex alloc
    # per frequency. mvfftw_c2c writes in-place via `ret`; norm_factor division
    # is deferred to the small per-segment subsets.
    buffer <- array(0 + 0i, dim = c(plan$segment_length, n_seg))
    ii <- 1L
    fmap_fun <- function(input) {
      kern <- input[[1]] + 1i * input[[2]]
      mvfftw_c2c(kern * fft_seg, inverse = 1L, ret = buffer)
      col_vec <- unlist(lapply(seq_len(n_seg), function(k) {
        buffer[new_ind[local_starts[k]:local_ends[k]], k] / norm_factor
      }))
      output_real[, ii] <- Re(col_vec)
      output_imag[, ii] <- Im(col_vec)
      ii <<- ii + 1L
      NULL
    }
  } else {
    fft_data <- fftw_r2c(data)
    tmp <- complex(length(fft_data))
    ii <- 1L
    fmap_fun <- function(input) {
      kern <- input[[1]] + 1i * input[[2]]
      fftw_c2c(data = kern * fft_data, inverse = 1L, ret = tmp)
      shifted <- c(tmp[-ind], tmp[ind]) / norm_factor
      output_real[, ii] <- Re(shifted)
      output_imag[, ii] <- Im(shifted)
      ii <<- ii + 1L
      NULL
    }
  }

  filearray::fmap2(
    x = fft_waves,
    fun = fmap_fun,
    .buffer_count = f_l,
    .simplify = FALSE
  )

  output_real$.header$freqs <- freqs
  output_real$.header$srate <- srate
  output_real$.header$wave_num <- wave_num
  output_real$.header$data_digest <- data_digest
  output_real$.header$trend <- trend
  output_real$.header$more_args <- more_args
  output_real$.header$rave_data_type <- "rave-wavelet-coefficients-real"
  output_real$.header$precision <- "double"
  output_real$.save_header()

  output_imag$.header$freqs <- freqs
  output_imag$.header$srate <- srate
  output_imag$.header$wave_num <- wave_num
  output_imag$.header$data_digest <- data_digest
  output_imag$.header$trend <- trend
  output_imag$.header$more_args <- more_args
  output_imag$.header$rave_data_type <- "rave-wavelet-coefficients-imag"
  output_imag$.header$precision <- "double"
  output_imag$.save_header()

  output_real$.mode <- "readonly"
  output_imag$.mode <- "readonly"
  list(
    real = output_real,
    imag = output_imag
  )
}

#' @importFrom gsignal butter
#' @export
gsignal::butter

#' @importFrom gsignal buttap
#' @export
gsignal::buttap

#' @importFrom gsignal buttord
#' @export
gsignal::buttord

#' @importFrom gsignal cheb1ap
#' @export
gsignal::cheb1ap

#' @importFrom gsignal cheb1ord
#' @export
gsignal::cheb1ord

#' @importFrom gsignal cheb2ap
#' @export
gsignal::cheb2ap

#' @importFrom gsignal cheb2ord
#' @export
gsignal::cheb2ord

#' @importFrom gsignal cheby1
#' @export
gsignal::cheby1

#' @importFrom gsignal cheby2
#' @export
gsignal::cheby2

#' @importFrom gsignal ellip
#' @export
gsignal::ellip

#' @importFrom gsignal ellipap
#' @export
gsignal::ellipap

#' @importFrom gsignal ellipord
#' @export
gsignal::ellipord


#' @importFrom gsignal resample
#' @export
gsignal::resample

#' @importFrom gsignal hilbert
#' @export
gsignal::hilbert

#' @title Plot one or more signal traces in the same figure
#' @param signals numerical matrix with each row to be a signal trace and each
#' column contains the signal values at a time point
#' @param sample_rate sampling frequency
#' @param col signal color, can be vector of one or more
#' @param space vertical spacing among the traces; for values greater than 1,
#' the spacing is absolute; default is \code{0.995}; for values less equal to 1,
#' this is the percentile of the whole data. However, the quantile mode can be
#' manually turned off is \code{"absolute"} is required; see \code{space_mode}
#' @param space_mode mode of spacing, only used when \code{space} is less equal
#' to one; default is quantile
#' @param start_time the time to start drawing relative to the first column
#' @param duration duration of the signal to draw
#' @param compress whether to compress signals if the data is too large
#' @param channel_names \code{NULL} or a character vector of channel names
#' @param time_shift the actual start time of the signal. Unlike
#' \code{start_time}, this should be the actual physical time represented
#' by the first column
#' @param new_plot whether to draw a new plot; default is true
#' @param xline,yline the gap between axis and label
#' @param xlab,ylab,lwd,xlim,cex,cex.lab,mar,mgp,xaxs,yaxs,tck,... plot
#' parameters; see \code{\link[base]{plot}} and \code{\link[graphics]{par}}
#' @examples
#'
#'
#' n <- 1000
#' base_signal <- c(rep(0, n/2), sin(seq(0,10,length.out = n/2))) * 10
#' signals <- rbind(rnorm(n) + base_signal,
#'                  rbinom(n, 10, 0.3) + base_signal,
#'                  rt(n, 5) + base_signal)
#' plot_signals(signals, sample_rate = 100)
#' plot_signals(signals, sample_rate = 100, start_time = 5)
#' plot_signals(signals, sample_rate = 100,
#'              start_time = 5, time_shift = 100)
#'
#'
#' @export
plot_signals <- function(
    signals, sample_rate = 1, col = graphics::par("fg"),
    space = 0.995, space_mode = c("quantile", "absolute"),
    start_time = 0, duration = NULL,
    compress = TRUE, channel_names = NULL, time_shift = 0,
    xlab = "Time (s)", ylab = "Electrode",
    lwd = 0.5, new_plot = TRUE, xlim = NULL,
    cex = 1, cex.lab = 1,
    mar = c(3.1, 2.1, 2.1, 0.8) * (0.25 + cex * 0.75) + 0.1,
    mgp = cex * c(2, 0.5, 0),
    xaxs = "r", yaxs = "i", xline = 1.5 * cex, yline = 1 * cex,
    tck = -0.005 * (3 + cex), ...)
{
  space_mode <- match.arg(space_mode)
  if (space_mode == "quantile" && space <= 1) {
    # quantile is slow: sample
    if (length(signals) > 100000) {
      space <- stats::quantile(
        signals[sample(length(signals), 100000)],
        space, na.rm = TRUE
      ) * 2
    } else {
      space <- stats::quantile(signals, space, na.rm = TRUE) * 2
    }
  }
  # make sure spacing is positive
  space <- abs(space)
  compress <- round(compress)
  if (!is.matrix(signals)) {
    if (is.vector(signals)) {
      signals <- matrix(signals, nrow = 1)
    } else {
      signals <- as.matrix(signals)
    }
  }
  ns <- nrow(signals)
  nt <- ncol(signals)

  if (compress == 1) {
    if (length(duration)) {
      n_tp <- round(duration * sample_rate)
    } else {
      n_tp <- round(nt - start_time * sample_rate)
    }
    if (n_tp * ns > 100000) {
      compress <- (n_tp * ns / 100000)
      if (n_tp / compress < 1000) {
        compress <- n_tp / 1000
      }
    }
  }

  if (compress > 1) {
    tidx <- round(seq(1, ncol(signals), by = compress))
    nt <- length(tidx)

    Time <- (tidx - 1) / sample_rate

    sample_rate <- sample_rate / compress
  } else {
    nt <- ncol(signals)
    tidx <- seq_len(nt)

    Time <- (tidx - 1) / sample_rate
  }

  start_time <- min(start_time, range(Time)[2] - 10 / sample_rate)
  if (is.null(duration)) {
    time_range <- c(start_time, range(Time)[2])
  } else {
    time_range <- c(start_time, start_time + duration)
  }
  tsl <- Time >= time_range[[1]] & Time <= time_range[[2]]

  tidx <- tidx[tsl]
  Time <- Time[tsl]

  signals <- signals[, tidx, drop = FALSE]
  nt <- ncol(signals)

  if (length(col) == 1) {
    col <- rep(col, ns)
  }
  y0 <- space * seq_len(ns)
  r <- y0 + signals

  if (is.null(channel_names)) {
    channel_names <- as.character(seq_len(ns))
    if (length(y0) > 30) {
      ind <- seq(0, length(y0), by = 5)
      ind <- unique(c(1, ind[-1], length(y0)))
      diff_ind <- diff(ind)
      if (diff_ind[[length(diff_ind)]] == 1) {
        ind <- ind[-(length(ind) - 1)]
      }
      y0 <- y0[ind]
      channel_names <- channel_names[ind]
    }
  }
  par_opt <- graphics::par(c("mai", "mar", "mgp", "cex.main", "cex.lab", "cex.axis", "cex.sub"))
  par_opt$cex.lab <- cex.lab

  if (new_plot) {
    graphics::par(mar = mar, mgp = mgp, cex.lab = par_opt$cex.lab)
    on.exit({graphics::par(par_opt)}, add = TRUE)

    graphics::matplot(time_shift + Time, t(r), type = "l",
                      col = col, lty = 1, lwd = lwd, frame.plot = FALSE,
                      yaxt = "n", xlab = "", ylab = "", ylim = space * c(0.5, ns + 0.5),
                      xlim = xlim, cex = cex, cex.main = par_opt$cex.main * cex,
                      cex.lab = par_opt$cex.lab * cex,
                      cex.axis = par_opt$cex.axis * cex,
                      xaxs = xaxs, yaxs = yaxs,
                      tck = tck, ...)
    graphics::axis(
      side = 2, at = y0, labels = channel_names, pos = start_time +
        time_shift, las = 1, tck = tck,
      cex = cex, cex.main = par_opt$cex.main * cex,
      cex.lab = par_opt$cex.lab * cex,
      cex.axis = par_opt$cex.axis * cex
    )
    graphics::mtext(side = 2, text = ylab, line = yline,
                    cex = par_opt$cex.lab * cex)
    graphics::mtext(side = 1, text = xlab, line = xline,
                    cex = par_opt$cex.lab * cex)
    # graphics::title(ylab = ylab, line = 1, cex.lab = cex * 0.8)
  } else {
    graphics::matpoints(time_shift + Time, t(r), type = "l",
                        col = col, lty = 1, lwd = lwd,
                        xlim = xlim, cex = cex,
                        cex.main = par_opt$cex.main * cex,
                        cex.lab = par_opt$cex.lab * cex,
                        cex.axis = par_opt$cex.axis * cex)
  }
  return(invisible(list(space = space, space_mode = space_mode, compress = compress,
                        time_range = time_range)))
}

#' @title Inflate a cortical surface mesh
#' @description
#' Iteratively relaxes a closed cortical-surface mesh into a smoother, more
#' compact "inflated" shape, while tracking how far each vertex moves inward
#' along the surface normal as it goes. Returns both the inflated mesh and
#' this per-vertex depth map (\code{sulc}).
#'
#' @details
#' The implementation follows the inflation procedure described in the
#' literature (see \strong{References}):
#' \enumerate{
#'   \item Optionally rescale the mesh to a canonical surface area
#'     (110,000 \eqn{mm^2}, the normalization target used by the reference
#'     procedure).
#'   \item Outer loop: a neighborhood-averaging size \code{n_averages} is
#'     halved at each level (16, 8, ..., 0).
#'   \item Inner loop (\code{niterations} repetitions per level):
#'     \describe{
#'       \item{Distance term}{Restoring force pulling each vertex back towards
#'         its original distances to its neighbors, with coefficient
#'         \code{l_dist * sqrt(n_averages)}.}
#'       \item{Gradient averaging}{Smooth the distance-term gradient over
#'         \code{n_averages} neighborhood passes before adding the spring term.}
#'       \item{Normalized spring term}{Laplacian (neighbor-averaging) smoothing
#'         scaled by \code{sqrt(orig_area / current_area)}, with coefficient
#'         \code{l_spring_norm}; added \emph{after} gradient averaging.}
#'       \item{Momentum step}{Update vertex positions using momentum
#'         integration with a 1 mm per-step displacement cap.}
#'       \item{Depth accumulation}{Accumulate the normal-projected component
#'         of each step into the per-vertex depth map (\code{sulc}).}
#'     }
#'   \item Center the mesh, rescale it, and zero-mean the depth map (\code{sulc}).
#' }
#'
#' @param mesh triangular mesh of class \code{'mesh3d'}, typically a white-matter
#'   surface in millimeters. The mesh \strong{must} be closed, manifold, and
#'   genus-0 (watertight, no boundary or non-manifold edges, single connected
#'   component) - this is a hard precondition of the underlying algorithm. An
#'   open or defective mesh (e.g. a raw surface extracted via
#'   \code{\link{vcg_isosurface}}, which often contains small extraction
#'   cracks) will silently inflate into a distorted, "sausage"-shaped result
#'   rather than a smooth sphere-like surface, and \code{mris_inflate} will
#'   raise an error if such defects are detected.
#' @param n_averages starting number of gradient-averaging passes (outer loop);
#'   halved each level down to 0. Default \code{16}.
#' @param niterations number of inner iterations per averaging level.
#'   Default \code{10}.
#' @param l_spring_norm normalized spring term coefficient. Default \code{1.0}.
#' @param l_dist distance-preservation coefficient (before per-level scaling).
#'   Default \code{0.1}.
#' @param momentum momentum coefficient. Default \code{0.9}.
#' @param dt time step. Default \code{0.9}.
#' @param desired_rms target \verb{RMS} tangent-plane height for early stopping
#'   of each inner loop. Default \code{0.015} (mm).
#' @param scale_brain logical; whether to scale the mesh to the canonical
#'   surface area (110,000 \eqn{mm^2}) before inflation. Set \code{FALSE} only
#'   for non-standard inputs. Default \code{TRUE}.
#' @param verbose logical; print per-iteration progress. Default \code{FALSE}.
#'
#' @returns A named list:
#' \describe{
#'   \item{\code{mesh}}{Inflated surface as a \code{'mesh3d'} object with
#'     \code{vb}, \code{it}, and \code{normals}.}
#'   \item{\code{sulc}}{Numeric vector of the per-vertex depth values described
#'     above (zero-mean, in mesh units).}
#' }
#'
#' @references
#' Cortical surface-based analysis II: Inflation, flattening, and a
#' surface-based coordinate system. \emph{NeuroImage}, 9(2), 195-207 (1999).
#'
#' @examples
#' if (is_not_cran()) {
#'
#'
#' data("left_hippocampus_mask")
#' mesh <- vcg_isosurface(left_hippocampus_mask)
#'
#' # Fix defects
#' mesh <- vcg_fix_defects(mesh, verbose = TRUE)
#'
#' # Center the mesh
#' mesh$vb[1:3, ] <- mesh$vb[1:3, ] - rowMeans(mesh$vb[1:3, ])
#'
#' # Inflate the surface while keeping the node distances
#' result <- mris_inflate(mesh, n_averages = 4L, niterations = 5L,
#'                        scale_brain = FALSE, verbose = TRUE)
#'
#' # Visualize with the sulcal values
#' pal <- colorRampPalette(c("black", "gray", "red"))(128)
#' col <- pal[pmax(pmin(round(result$sulc * 10 + 64), 128), 1)]
#'
#' oldpar <- par(mfrow = c(1, 2))
#' on.exit({ par(oldpar) })
#'
#' plot(
#'   mesh, col = col,
#'   eye = c(0, 100, 0),
#'   up = c(1, 0, 0))
#'
#' plot(
#'   result$mesh, col = col,
#'   eye = c(0, 100, 0),
#'   up = c(1, 0, 0))
#'
#'
#' }
#'
#' @export
mris_inflate <- function(mesh,
                         n_averages    = 16L,
                         niterations   = 10L,
                         l_spring_norm = 1.0,
                         l_dist        = 0.1,
                         momentum      = 0.9,
                         dt            = 0.9,
                         desired_rms   = 0.015,
                         scale_brain   = TRUE,
                         verbose       = FALSE) {
  mesh <- meshintegrity(mesh, facecheck = TRUE)

  vb <- mesh$vb[1:3, , drop = FALSE]
  storage.mode(vb) <- "double"

  it <- mesh$it
  storage.mode(it) <- "integer"

  tmp <- mrisInflate(
    vb_           = vb,
    it_           = it,
    n_averages    = as.integer(n_averages)[[1L]],
    niterations   = as.integer(niterations)[[1L]],
    l_spring_norm = as.double(l_spring_norm)[[1L]],
    l_dist        = as.double(l_dist)[[1L]],
    momentum      = as.double(momentum)[[1L]],
    dt            = as.double(dt)[[1L]],
    desired_rms   = as.double(desired_rms)[[1L]],
    scale_brain   = as.logical(scale_brain)[[1L]],
    verbose       = as.logical(verbose)[[1L]]
  )

  inflated <- structure(list(
    vb      = rbind(tmp$vb, 1),
    it      = it,
    normals = rbind(tmp$normals, 1)
  ),
  class = c("ravetools_mesh3d", "mesh3d"))

  list(mesh = inflated, sulc = as.numeric(tmp$sulc))
}

#' @title Collapse array
#' @param x A numeric multi-mode tensor (array), without \code{NA}
#' @param keep Which dimension to keep
#' @param average collapse to sum or mean
#' @param transform transform on the data before applying collapsing;
#' choices are \code{'asis'} (no change), \code{'10log10'} (used to calculate
#' decibel), \code{'square'} (sum-squared), \code{'sqrt'} (square-root and
#' collapse)
#' @param ... passed to other methods
#'
#' @returns a collapsed array with values to be mean or summation along
#' collapsing dimensions
#'
#'
#' @examples
#'
#' # Set ncores = 2 to comply to CRAN policy. Please don't run this line
#' ravetools_threads(n_threads = 2L)
#'
#' # Example 1
#' x = matrix(1:16, 4)
#'
#' # Keep the first dimension and calculate sums along the rest
#' collapse(x, keep = 1)
#' rowMeans(x)  # Should yield the same result
#'
#' # Example 2
#' x = array(1:120, dim = c(2,3,4,5))
#' result = collapse(x, keep = c(3,2))
#' compare = apply(x, c(3,2), mean)
#' sum(abs(result - compare)) # The same, yield 0 or very small number (1e-10)
#'
#'
#' \donttest{
#'
#' ravetools_threads(n_threads = -1)
#'
#' # Example 3 (performance)
#'
#' # Small data, no big difference
#' x = array(rnorm(240), dim = c(4,5,6,2))
#' microbenchmark::microbenchmark(
#'   result = collapse(x, keep = c(3,2)),
#'   compare = apply(x, c(3,2), mean),
#'   times = 1L, check = function(v) {
#'     max(abs(range(do.call('-', v)))) < 1e-10
#'   }
#' )
#'
#' # large data big difference
#' x = array(rnorm(prod(300,200,105)), c(300,200,105,1))
#' microbenchmark::microbenchmark(
#'   result = collapse(x, keep = c(3,2)),
#'   compare = apply(x, c(3,2), mean),
#'   times = 1L , check = function(v) {
#'     max(abs(range(do.call('-', v)))) < 1e-10
#'   })
#'
#' }
#'
#' @export
collapse <- function(x, keep, ...) {
  UseMethod("collapse")
}

#' @rdname collapse
#' @export
collapse.array <- function(
  x, keep, average = TRUE,
  transform = c("asis", "10log10", "square", "sqrt"), ...) {

  transform <- match.arg(transform)

  dim_x <- dim(x)
  keep <- as.integer(keep)
  if (!length(keep)) {
    stop("collapse.array: `keep` length must be positive")
  } else if (any(!is.finite(keep) | duplicated(keep))) {
    stop("collapse.array: `keep` cannot have duplicated margin indices or NAs")
  } else if (any(keep < 1 | keep > length(dim_x))) {
    stop("collapse.array: `keep` must be margin indices (from 1 to max margin)")
  }

  is_complex <- is.complex(x)
  mode_re <- "double"
  if (is_complex && transform == "asis") {
    mode_re <- "complex"
  }

  dim_keep <- dim_x[keep]
  if (!length(x)) {
    if (average) {
      v <- NaN
    } else {
      v <- 0
    }
    storage.mode(v) <- mode_re
    if (length(dim_keep) >= 2) {
      return(array(v, dim_keep))
    } else {
      return(rep(v, dim_keep[[1]]))
    }

  }

  method <- which(c("asis", "10log10", "square", "sqrt") == transform)
  average <- as.integer(as.logical(average))

  if (is_complex) {
    re <- collapser_cplx(x = x, keep = keep, method = method, average = average)
    if (transform != "asis") {
      re <- Re(re)
    }
  } else {
    re <- collapser_real(x = x, keep = keep, method = method, average = average)
  }

  if (inherits(re, "ravetools_error")) {
    stop(re)
  }

  return(re)

}

test_that("testing collapse with 100x100x30x10; keep=c(4,2); sum", {
  dim <- c(100, 10, 30, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 2)

  aggregate <- sum
  avg <- 0L
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x0x10; keep=c(4,2); sum", {
  dim <- c(100, 10, 0, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 2)

  aggregate <- sum
  avg <- 0L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x30x10; keep=4; sum", {
  dim <- c(100, 10, 0, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- 4

  aggregate <- sum
  avg <- 0L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x30x10; keep=c(4,3,2,1); sum", {
  dim <- c(100, 10, 30, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 3, 2, 1)

  aggregate <- sum
  avg <- 0L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x0x10; keep=c(4,3,2,1); sum", {
  dim <- c(100, 10, 0, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 3, 2, 1)

  aggregate <- sum
  avg <- 0L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})



test_that("testing collapse with 100x100x30x10; keep=c(4,2); mean", {
  dim <- c(100, 10, 30, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 2)

  aggregate <- mean
  avg <- 1L
  method <- "asis"
  r1 <- r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x0x10; keep=c(4,2); mean", {
  dim <- c(100, 10, 0, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 2)

  aggregate <- mean
  avg <- 1L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x30x10; keep=4; mean", {
  dim <- c(100, 10, 0, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- 4

  aggregate <- mean
  avg <- 1L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x30x10; keep=c(4,3,2,1); mean", {
  dim <- c(100, 10, 30, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 3, 2, 1)

  aggregate <- mean
  avg <- 1L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

test_that("testing collapse with 100x100x0x10; keep=c(4,3,2,1); mean", {
  dim <- c(100, 10, 0, 10)
  x <- array(rnorm(prod(dim)), dim = dim)

  keep <- c(4, 3, 2, 1)

  aggregate <- mean
  avg <- 1L

  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  x <- abs(x)
  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(10.0 * log10(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(x * x) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(sqrt(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)


  x <- array(rnorm(prod(dim)), dim = dim) * 1i + array(rnorm(prod(dim)), dim = dim)
  method <- "asis"
  r1 <- apply(x, keep, aggregate)
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "10log10"
  r1 <- apply(x, keep, function(x) { aggregate(20.0 * log10(Mod(x))) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "square"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)^2) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

  method <- "sqrt"
  r1 <- apply(x, keep, function(x) { aggregate(Mod(x)) })
  r2 <- ravetools::collapse(x, keep, avg, method)
  testthat::expect_equal(r1, r2)

})

#' @title Imaging registration using \code{'NiftyReg'}
#' @description Registers 'CT' to 'MRI', or 'MRI' to another 'MRI'
#' @param source source imaging data, or a \code{'nifti'} file path;
#' for example, 'CT'
#' @param target target imaging data to align to; for example, 'MRI'
#' @param method method of transformation, choices are \code{'rigid'},
#' \code{'affine'}, or \code{'nonlinear'}
#' @param interpolation how volumes should be interpolated, choices are
#' \code{'cubic'}, \code{'trilinear'}, or \code{'nearest'}
#' @param threads,symmetric,verbose,... see \code{\link[RNiftyReg]{niftyreg}}
#' @returns See \code{\link[RNiftyReg]{niftyreg}}
#' @examples
#'
#' \donttest{
#'
#' source <- system.file("extdata", "epi_t2.nii.gz", package="RNiftyReg")
#' target <- system.file("extdata", "flash_t1.nii.gz", package="RNiftyReg")
#' aligned <- register_volume(source, target, verbose = FALSE)
#'
#' source_img <- aligned$source[[1]]
#' target_img <- aligned$target
#' aligned_img <- aligned$image
#'
#' oldpar <- par(mfrow = c(2, 2), mar = c(0.1, 0.1, 3.1, 0.1))
#'
#' pal <- grDevices::grey.colors(256, alpha = 1)
#' image(source_img[,,30], asp = 1, axes = FALSE,
#'       col = pal, main = "Source image")
#' image(target_img[,,64], asp = 1, axes = FALSE,
#'       col = pal, main = "Target image")
#' image(aligned_img[,,64], asp = 1, axes = FALSE,
#'       col = pal, main = "Aligned image")
#'
#' # bucket fill and calculate differences
#' aligned_img[is.nan(aligned_img) | aligned_img <= 1] <- 1
#' target_img[is.nan(target_img) | aligned_img <= 1] <- 1
#' diff <- abs(aligned_img / target_img - 1)
#' image(diff[,,64], asp = 1, axes = FALSE,
#'       col = pal, main = "Percentage Difference")
#'
#' par(oldpar)
#'
#' }
#'
#' @export
register_volume <- function(
    source, target, method = c("rigid", "affine", "nonlinear"),
    interpolation = c("cubic", "trilinear", "nearest"),
    threads = detect_threads(), symmetric = TRUE, verbose = TRUE, ...) {
  method <- match.arg(method)
  interpolation <- match.arg(interpolation)
  interpolation <- list(
    cubic = 3L,
    trilinear = 1L,
    nearest = 0L
  )[[interpolation]]

  if (isTRUE(is.character(source))) {
    source <- RNiftyReg::readNifti(source)
  }
  if (isTRUE(is.character(target))) {
    target <- RNiftyReg::readNifti(target)
  }

  RNiftyReg::niftyreg(
    source = source,
    target = target,
    scope = method,
    interpolation = interpolation,
    threads = threads,
    internal = FALSE,
    symmetric = symmetric,
    estimateOnly = FALSE,
    verbose = verbose,
    ...
  )
}

#' @name convolve
#' @title Convolution of \code{1D}, \code{2D}, \code{3D} data via \code{FFT}
#' @description
#' Use the 'Fast-Fourier' transform to compute the convolutions of two data
#' with zero padding. This function is mainly designed for image convolution.
#' For forward and backward convolution/filter, see \code{\link{filtfilt}}.
#'
#' @param x one-dimensional signal vector, two-dimensional image, or
#' three-dimensional volume; numeric or complex
#' @param filter kernel with the same number of dimensions as \code{x}
#' @returns Convolution results with the same length and dimensions as \code{x}.
#' If \code{x} is complex, results will be complex, otherwise results will
#' be real numbers.
#' @details This implementation uses 'Fast-Fourier' transform to perform
#' \code{1D}, \code{2D}, or \code{3D} convolution. Compared to implementations
#' using original mathematical definition of convolution, this approach is
#' much faster, especially for image and volume convolutions.
#'
#' The input \code{x} is zero-padded beyond edges. This is most common in image
#' or volume convolution, but less optimal for periodic one-dimensional signals.
#' Please use other implementations if non-zero padding is needed.
#'
#' The convolution results might be different to the ground truth by a precision
#' error, usually at \code{1e-13} level, depending on the \code{'FFTW3'}
#' library precision and implementation.
#'
#' @examples
#'
#'
#' # ---- 1D convolution ------------------------------------
#' x <- cumsum(rnorm(100))
#' filter <- dnorm(-2:2)
#' # normalize
#' filter <- filter / sum(filter)
#' smoothed <- convolve_signal(x, filter)
#'
#' plot(x, pch = 20)
#' lines(smoothed, col = 'red')
#'
#' # ---- 2D convolution ------------------------------------
#' x <- array(0, c(100, 100))
#' x[
#'   floor(runif(10, min = 1, max = 100)),
#'   floor(runif(10, min = 1, max = 100))
#' ] <- 1
#'
#' # smooth
#' kernel <- outer(dnorm(-2:2), dnorm(-2:2), FUN = "*")
#' kernel <- kernel / sum(kernel)
#'
#' y <- convolve_image(x, kernel)
#'
#' oldpar <- par(mfrow = c(1,2))
#' image(x, asp = 1, axes = FALSE, main = "Origin")
#' image(y, asp = 1, axes = FALSE, main = "Smoothed")
#' par(oldpar)
#'
#'
NULL

#' @rdname convolve
#' @export
convolve_signal <- function(x, filter) {

  len_x <- length(x)
  len_y <- length(filter)

  padded_len <- len_x + len_y + 1L

  junk_len <- ceiling(len_y / 2) - 1

  x_ <- c(as.double(x), rep(0.0, len_y + 1L))
  y_ <- c(as.double(filter), rep(0.0, len_x + 1L))

  a <- fftw_c2c(x_)
  b <- fftw_c2c(y_)

  a <- fftw_c2c(a * b, inverse = TRUE, ret = a) / padded_len
  b <- a[junk_len + seq_len(len_x)]

  if (!is.complex(x)) {
    b <- Re(b)
  }
  b
}

#' @rdname convolve
#' @export
convolve_image <- function(x, filter) {
  # make sure x and filter are matrix
  if (!is.matrix(x)) {
    x <- as.matrix(x)
  }

  if (!is.matrix(filter)) {
    filter <- as.matrix(filter)
  }

  dim_x <- dim(x)
  dim_y <- dim(filter)

  padded_dim <- dim_x + dim_y + 1L

  junk_dim <- ceiling(dim_y / 2) - 1

  x_ <- array(0.0, padded_dim)
  x_[ seq_len(dim_x[[1]]), seq_len(dim_x[[2]]) ] <- x
  y_ <- array(0.0, padded_dim)
  y_[ seq_len(dim_y[[1]]), seq_len(dim_y[[2]]) ] <- filter

  a <- fftw_c2c_2d(x_)
  b <- fftw_c2c_2d(y_)

  a <- fftw_c2c_2d(a * b, inverse = TRUE, ret = a) / prod(padded_dim)
  b <- a[junk_dim[[1]] + seq_len(dim_x[[1]]),
         junk_dim[[2]] + seq_len(dim_x[[2]]),
         drop = FALSE]

  if (!is.complex(x)) {
    b <- Re(b)
  }
  b
}

#' @rdname convolve
#' @export
convolve_volume <- function(x, filter) {
  # make sure x and filter are matrix
  if (!is.array(x)) {
    x <- is.array(x)
  }

  if (!is.array(filter)) {
    filter <- is.array(filter)
  }

  dim_x <- dim(x)
  dim_y <- dim(filter)
  if (length(dim_x) < 3L) {
    dim_x <- c(dim_x, rep(1, 3 - length(dim_x)))
  } else if (length(dim_x) > 3L) {
    if (prod(dim_x) != prod(dim_x[1:3])) {
      stop("`convolve_volume`: x must be an array with 3 dimensions")
    }
    dim_x <- dim_x[1:3]
    dim(x) <- dim_x
  }
  if (length(dim_y) < 3L) {
    dim_y <- c(dim_y, rep(1, 3 - length(dim_y)))
  } else if (length(dim_y) > 3L) {
    if (prod(dim_y) != prod(dim_y[1:3])) {
      stop("`convolve_volume`: filter must be an array with 3 dimensions")
    }
    dim_y <- dim_y[1:3]
    dim(filter) <- dim_y
  }

  padded_dim <- dim_x + dim_y + 1L
  junk_dim <- ceiling(dim_y / 2) - 1L

  x_ <- array(0.0, padded_dim)
  x_[
    seq_len(dim_x[[1]]),
    seq_len(dim_x[[2]]),
    seq_len(dim_x[[3]])
  ] <- x
  y_ <- array(0.0, padded_dim)
  y_[
    seq_len(dim_y[[1]]),
    seq_len(dim_y[[2]]),
    seq_len(dim_y[[3]])
  ] <- filter

  a <- fftw_c2c_3d(x_)
  b <- fftw_c2c_3d(y_)

  a <- fftw_c2c_3d(a * b, inverse = TRUE, ret = a) / prod(padded_dim)
  b <- a[
    junk_dim[[1]] + seq_len(dim_x[[1]]),
    junk_dim[[2]] + seq_len(dim_x[[2]]),
    junk_dim[[3]] + seq_len(dim_x[[3]]),
    drop = FALSE
  ]

  if (!is.complex(x)) {
    b <- Re(b)
  }
  b
}


#' Grow volume mask
#' @param volume volume mask array, must be 3-dimensional array
#' @param x,y,z size of grow along each direction
#' @param threshold threshold after convolution
#' @returns A binary volume mask
#' @examples
#'
#'
#' oldpar <- par(mfrow = c(2,3), mar = c(0.1,0.1,3.1,0.1))
#'
#' mask <- array(0, c(21,21,21))
#' mask[11,11,11] <- 1
#' image(mask[11,,], asp = 1,
#'       main = "Original mask", axes = FALSE)
#' image(grow_volume(mask, 2)[11,,], asp = 1,
#'       main = "Dilated (size=2) mask", axes = FALSE)
#' image(grow_volume(mask, 5)[11,,], asp = 1,
#'       main = "Dilated (size=5) mask", axes = FALSE)
#'
#' mask[11, sample(11,2), sample(11,2)] <- 1
#' image(mask[11,,], asp = 1,
#'       main = "Original mask", axes = FALSE)
#' image(grow_volume(mask, 2)[11,,], asp = 1,
#'       main = "Dilated (size=2) mask", axes = FALSE)
#' image(grow_volume(mask, 5)[11,,], asp = 1,
#'       main = "Dilated (size=5) mask", axes = FALSE)
#'
#' par(oldpar)
#'
#'
#' @export
grow_volume <- function(volume, x, y = x, z = x, threshold = 0.5) {
  filter <- array(1.0, c(x, y, z))
  re <- convolve_volume(volume, filter)
  sel <- re > threshold
  mode(sel) <- "integer"
  sel
}

load_imager <- function() {
  tryCatch({
    imager <- asNamespace("imager")
    imager
  }, error = function(e) {})
}

test_that("convolve_image", {
  testthat::skip_on_cran()
  testthat::skip_if(system.file(package = "imager") == "")

  # Do NOT test on CRAN, `imager` does not install well, I don't want to
  # add it to `Suggests`
  imager <- load_imager()
  testthat::skip_if(is.null(imager), message = "Fail to load package imager")
  boats <- imager$as.cimg(imager$.__NAMESPACE__.$lazydata$boats[, , 1, 1])
  filter <- imager$as.cimg(function(x, y) sign(x - 5), 11, 11)

  x <- imager$convolve(boats, filter = filter)
  y <- imager$as.cimg(ravetools:::convolve_image(boats, filter = filter))
  testthat::expect_equal(y, x, tolerance = 1e-10)

  boats <- imager$as.cimg(t(imager$.__NAMESPACE__.$lazydata$boats[, , 1, 1]))
  filter <- imager$as.cimg(function(x, y) sign(x - 5), 12, 12)

  x <- imager$convolve(boats, filter = filter)
  y <- imager$as.cimg(ravetools:::convolve_image(boats, filter = filter))
  testthat::expect_equal(y, x, tolerance = 1e-10)

})



test_that("convolve_volume", {
  testthat::skip_on_cran()
  testthat::skip_if(system.file(package = "imager") == "")

  # Do NOT test on CRAN, `imager` does not install well, I don't want to
  # add it to `Suggests`
  imager <- load_imager()
  testthat::skip_if(is.null(imager), message = "Fail to load package imager")

  boats <- as.array(imager$.__NAMESPACE__.$lazydata$boats)
  dim(boats) <- dim(boats)[c(1, 2, 4, 3)]
  boats <- imager$as.cimg(boats)
  filter <- imager$as.cimg(function(x, y, z) {
    sign(x - 5)
  }, 11, 11, 11)

  x <- imager$convolve(boats, filter = filter)
  y <- imager$as.cimg(structure(ravetools:::convolve_volume(boats, filter = filter), dim = dim(boats)))
  testthat::expect_equal(y, x, tolerance = 1e-10)

  boats <- as.array(imager$.__NAMESPACE__.$lazydata$boats)
  dim(boats) <- dim(boats)[c(4, 2, 1, 3)]
  boats <- imager$as.cimg(boats)
  filter <- imager$as.cimg(function(x, y, z) {
    sign(x - 5)
  }, 12, 12, 12)

  x <- imager$convolve(boats, filter = filter)
  y <- imager$as.cimg(structure(ravetools:::convolve_volume(boats, filter = filter), dim = dim(boats)))
  testthat::expect_equal(y, x, tolerance = 1e-10)

  # imager <- asNamespace("imager")
  # boats <- imager$as.cimg(array(rnorm(256*256*256), c(rep(256,3), 1)))
  # filter <- imager$as.cimg(function(x,y,z) sign(x-5),11,11, 11)
  # microbenchmark::microbenchmark(
  #   {imager$convolve(boats, filter = filter)},
  #   {imager$as.cimg(structure(ravetools:::convolve_volume(boats, filter = filter), dim = dim(boats)))}, times = 1
  # )

})

find_outliers <- function(x, nsdev = 3, duration = 1, sym = FALSE) {

  if (sym) {
    tmp <- abs(x)
    nsd_ <- stats::sd(tmp, na.rm = TRUE) * nsdev
    idx <- which(is.na(tmp) | tmp > nsd_)
  } else {
    m <- mean(x, na.rm = TRUE)
    nsd_ <- stats::sd(x, na.rm = TRUE) * nsdev
    idx <- which(is.na(x) | abs(x - m) > nsd_)
  }
  outliers <- deparse_svec(idx, max_lag = duration, concatenate = FALSE)

  idx0 <- lapply(outliers, function(out) {
    re <- parse_svec(out)
    if (length(re) < 0.5 * duration) {
      return(NULL)
    } else {
      return(re)
    }
  })
  return(unlist(idx0))
}

smooth_signals <- function(x, time, ord = 4L, nknots = 60, avoid_timepoints = integer(0), regularization = 0.1) {
  is_matrix <- TRUE
  if (!is.matrix(x)) {
    is_matrix <- FALSE
    x <- matrix(x, nrow = 1)
  }
  ntp <- ncol(x)
  if (missing(time)) {
    time <- seq_len(ntp)
  }
  time_idx <- seq_along(time)
  avoid_timepoints <- unique(c(as.integer(avoid_timepoints), which(colSums(is.na(x)) > 0)))
  if (anyNA(avoid_timepoints)) {
    avoid_timepoints <- avoid_timepoints[!is.na(avoid_timepoints)]
  }
  time_idx <- time_idx[!time_idx %in% avoid_timepoints]

  time_slices <- deparse_svec(time_idx, concatenate = FALSE)
  time_range <- max(time) - min(time)
  plan <- sapply(time_slices, function(idx_char) {
    idx <- parse_svec(idx_char)
    time_start <- time[idx[[1]]]
    time_end <- time[idx[[length(idx)]]]
    c(time_start, time_end, (time_end - time_start) / time_range)
  })
  if (ncol(plan) == 1) {
    plan[3, ] <- nknots - ord
  } else {
    nknots2 <- ceiling(plan[3, ] / sum(plan[3, ]) * (nknots - ord))
    while (sum(nknots2) > nknots - ord) {
      tmp <- which.max(nknots2)
      if (nknots2[tmp] >= 2) {
        nknots2[tmp] <- nknots2[tmp] - 1
      } else {
        break
      }
    }
    plan[3, ] <- nknots2
    plan <- plan[, nknots2 > 0, drop = FALSE]
  }
  knots <- apply(plan, 2, function(item) {
    time_start <- item[[1]]
    time_end <- item[[2]]
    nknots_ <- item[[3]]
    # Chebychev points
    cos((2 * seq_len(nknots_) - 1) / (nknots_ * 2) * pi) * (time_end - time_start) / 2 + (time_start + time_end) / 2
  })
  knots <- sort(as.double(unlist(knots)))
  if (length(knots) > nknots - ord) {
    knots <- knots[seq_len(nknots - ord)]
  }

  knots <- c(rep(time[1], ord - 1), knots, rep(time[length(time)], nknots - length(knots) - 1))

  B <- t(splines::splineDesign(knots, time, ord = ord, outer.ok = TRUE, sparse = FALSE))
  if (!inherits(B, "matrix")) {
    class(B) <- "matrix"
  }
  bbt <- tcrossprod(B)
  diag(bbt) <- diag(bbt) + regularization
  bbtsolve <- qr.solve(bbt)

  # Initial fit
  if (length(avoid_timepoints)) {
    gamma <- tcrossprod(x[-avoid_timepoints], B[, -avoid_timepoints]) %*% bbtsolve
  } else {
    gamma <- tcrossprod(x, B) %*% bbtsolve
  }
  fitted <- gamma %*% B
  if (!is_matrix) {
    fitted <- as.vector(fitted)
  }
  fitted
}

interp_spline <- function(obj1, obj2, ord = 4L,
                           na.action = stats::na.omit, sparse = FALSE) {
  stopifnot(exprs = {
    (degree <- ord - 1L) >= 0
    length(degree) == 1L
    degree == as.integer(degree)
  })
  frm <- na.action(data.frame(x = as.numeric(obj1), y = as.numeric(obj2)))
  frm <- frm[order(frm$x), ]
  ndat <- nrow(frm)
  x <- frm$x
  if (anyDuplicated(x))
    stop("values of 'x' must be distinct")
  if (length(x) < ord)
    stop(gettextf("must have at least 'ord'=%d points", ord), domain = NA)
  iDeg <- seq_len(degree)

  nu <- ord %/% 2L
  knots <- c(x[iDeg] + x[1L] - x[ord], x, x[ndat + iDeg - degree] + x[ndat] - x[ndat - degree])

  derivs <- c(nu, integer(ndat + 2 * (nu - 2)), nu)
  x <- c(rep(x[1L], nu - 1), x, rep(x[ndat], nu - 1))
  y <- c(rep(0, nu - 1), frm$y, rep(0, nu - 1))
  des <- splines::splineDesign(knots, x, ord, derivs, sparse = sparse)
  # des <- crossprod(des)
  diag(des) <- diag(des) + 0.001
  coeff <- solve(des, y)
  value <- structure(list(knots = knots, coefficients = coeff, order = ord),
                     formula = do.call("~", list(substitute(obj2), substitute(obj1))),
                     class = c("nbSpline", "bSpline", "spline"))

  value <- splines::polySpline(value)
  coeff <- stats::coef(value)
  coeff[, 1L] <- frm$y
  coeff[1L, degree] <- coeff[nrow(coeff), degree] <- 0
  value$coefficients <- coeff
  return(value)
}


interpolate_missing_signal <- function(x, time, nknots = 100, ord = 4L, ...) {
  # Time
  if (missing(time)) {
    ntp <- length(x)
    time <- seq_len(ntp)
  }
  valid_points <- !is.na(x)

  if (sum(valid_points) > nknots) {
    valid_points <- which(valid_points)
    valid_points <- valid_points[round(seq(1, length(valid_points), length.out = nknots))]
  }

  # ispl <- tryCatch({
  #   splines::interpSpline( obj1 = time[valid_points], obj2 = x[valid_points], ord = 4L, na.action = na.omit, sparse = TRUE, ...)
  # }, error = function(e) {
  #   splines::interpSpline( obj1 = time[valid_points], obj2 = x[valid_points], ord = 4L, na.action = na.omit, sparse = FALSE, ...)
  # })

  ispl <- tryCatch({
    interp_spline( obj1 = time[valid_points], obj2 = x[valid_points], ord = ord, na.action = stats::na.omit, sparse = TRUE, ...)
  }, error = function(e) {
    interp_spline( obj1 = time[valid_points], obj2 = x[valid_points], ord = ord, na.action = stats::na.omit, sparse = FALSE, ...)
  })
  re <- stats::predict( ispl, time )$y
  re[!is.na(x)] <- x[!is.na(x)]
  return(re)
}

#' @title Find and interpolate stimulation signals
#' @param x numerical vector representing a analog signal
#' @param sample_rate sampling frequency
#' @param duration time in second: duration of interpolation
#' @param ord spline order, default is 4
#' @param nknots a rough number of knots to use, default is 100
#' @param nsd number of standard deviation to detect stimulation signals, default is 1
#' @param nstim number of stimulation pulses, default is to auto-detect
#' @param regularization regularization parameter in case of inverting singular matrices, default is 0.5
#' @returns Interpolated signal with an attribute of which sample points are interpolated
#' @examples
#'
#' x0 <- rnorm(1000) / 5 + sin(1:1000 / 300)
#'
#' # Simulates pulase signals
#' x <- x0
#' x[400:410] <- -100
#' x[420:430] <- 100
#'
#' fitted <- interpolate_stimulation(x, 100, duration = 0.3, nknots = 10, nsd = 2)
#'
#' oldpar <- par(mfrow = c(2, 1))
#'
#' plot(fitted, type = 'l', col = 'blue', lwd = 2)
#' lines(x, col = 'red')
#' lines(x0, col = 'black')
#' legend("topleft", c("Interpolated", "Observed", "Underlying"),
#'        lty = 1, col = c("blue", "red", "black"))
#'
#' pwelch(x0, 100, 200, 100, plot = 1, col = 'black', ylim = c(-50, 50))
#' pwelch(x, 100, 200, 100, plot = 2, col = 'red')
#' pwelch(fitted, 100, 200, 100, plot = 2, col = 'blue')
#'
#' par(oldpar)
#'
#'
#' @noRd
interpolate_stimulation <- function(x, sample_rate, duration = 40 / sample_rate,
                               ord = 4L, nknots = 100, nsd = 1, nstim = NULL, regularization = 0.5) {

  # DIPSAUS DEBUG START
  # nknots <- 60
  # ord <- 4
  # nsd <- 3
  # duration <- 4/30000
  # sample_rate <- 30000
  # nstim <- NULL

  ntp <- length(x)

  # Time
  time <- seq_len(ntp)

  # de-trend
  a <- x[[1]]
  b <- x[[ntp]]

  trend <- ((time - 1) / (ntp - 1) * (b - a) + a)
  x <- x - trend

  duration_npts <- duration * sample_rate

  # find outliers
  fitted <- smooth_signals(x = x, time = time, ord = ord, nknots = nknots, regularization = regularization)
  outliers <- find_outliers(x = x - fitted, nsdev = nsd, duration = duration_npts / 4, sym = TRUE)
  if (!length(nstim)) {
    nstim <- length(deparse_svec(outliers, concatenate = FALSE, max_lag = duration_npts / 4))
  }

  # DIPSAUS DEBUG START
  # plot(fitted, type = "l", lty = 1, col = 'red')
  # matlines(x, type = "l", lty = 1, col = 'gray')

  fitted <- smooth_signals(x = x, time = time, ord = ord, nknots = nknots, avoid_timepoints = outliers, regularization = regularization)
  tmp <- (x - fitted)
  tmp[outliers] <- NA_real_
  outliers_new <- find_outliers(x = x - fitted, nsdev = nsd, duration = duration_npts / 4, sym = TRUE)
  outliers <- unique(c(outliers, outliers_new))



  outliers_txt <- deparse_svec(outliers, concatenate = FALSE)
  if (length(outliers_txt) > nstim) {
    outliers_txt <- outliers_txt[order(sapply(outliers_txt, function(idx) {
      mean(abs(tmp[parse_svec(idx)]))
      # length(parse_svec(idx))
    }), decreasing = TRUE)]
    outliers_txt <- outliers_txt[seq_len(nstim)]
  }
  outliers <- unlist(lapply(outliers_txt, function(idx) {
    idx <- parse_svec(idx)
    if (length(idx) < duration_npts) {
      pad1 <- floor((duration_npts - length(idx)) / 4)
      if (pad1 > idx[1]) {
        pad1 <- idx[1]
      }
      idx <- (idx[1] - pad1) + seq_len(duration_npts)
    }
    idx
  }))
  # outliers <- parse_svec(idx)

  tmp <- x
  tmp[outliers] <- NA_real_
  # neg <- !is.na(tmp) & tmp < 0
  fitted <- interpolate_missing_signal(tmp, time = time, nknots = nknots, ord = max(2L, ord))
  # fitted[outliers] <- NA_real_
  # outliers_new <- find_outliers(x = x - fitted, nsdev = nsd, duration = duration_npts, sym = TRUE)
  # outliers <- unique(c(outliers, outliers_new))
  #
  # outliers_txt <- deparse_svec(outliers, concatenate = FALSE)
  # if(length(outliers_txt) > nstim) {
  #   outliers_txt <- outliers_txt[order(sapply(outliers_txt, function(idx) {
  #     mean(abs(tmp[parse_svec(idx)]))
  #     # length(parse_svec(idx))
  #   }), decreasing = TRUE)]
  #   outliers_txt <- outliers_txt[seq_len(nstim)]
  # }
  # outliers <- unlist(lapply(outliers_txt, function(idx) {
  #   idx <- parse_svec(idx)
  #   if (length(idx) < duration_npts) {
  #     idx <- idx[1] -1 + seq_len(duration_npts)
  #   }
  #   idx
  # }))
  # tmp <- x
  # tmp[outliers] <- NA_real_
  # # neg <- !is.na(tmp) & tmp < 0
  # fitted <- interpolate_missing_signal(tmp, time = time, ord = max(4L, ord), period = NULL)

  # spot outliers again

  # fitted[neg] <- -fitted[neg]

  # DIPSAUSE DEBUG START
  # plot(fitted, type = "l", lty = 1, col = 'red')
  # abline(v = outliers, col = "gray")
  # lines(fitted, type = "l", lty = 1, col = 'red')
  # lines(tmp, type = "l", lty = 1, col = 'black')

  fitted <- fitted + trend

  # fitted <- matrix(fitted, nrow = 1)
  attr(fitted, "interpolated") <- outliers
  fitted
}

# ---- fftfilt -----------------------------------------------------------------

fftfilt <- function(b, x, n = NULL, legacy = FALSE) {
  if (length(n) > 1) {
    stop("fftfilt: n has to be a scalar")
  }
  # DIPSAUS DEBUG START
  # b <- rnorm(5)
  # x <- rnorm(500)
  # n <- NULL

  N <- n
  l_b <- length(b)

  if (!is.matrix(x)) {
    l_x <- length(x)
    x <- matrix(x, ncol = 1L)
    is_vector <- TRUE
  } else {
    l_x <- nrow(x)
    is_vector <- FALSE
  }
  if (!length(n)) {
    N <- 2 ^ (ceiling(log(l_x + l_b - 1) / log(2)))
    B <- fft(postpad(b, N))
    if ( legacy ) {
      y <- apply(x, 2L, function(xi) {
        # y <- ifft(fft(postpad(x, N)) * B)
        ifft(fft(postpad(xi, N)) * B)
      })
    } else {
      y <- stats::mvfft(stats::mvfft(postpad(x, N)) * B, inverse = TRUE) / N
    }
  } else {
    N <- 2 ^ (ceiling(log(max(n, l_b)) / log(2)))
    L <- N - l_b + 1
    B <- fft(postpad(b, N))
    R <- ceiling(l_x / L)

    y <- array(0.0i, c(l_x, ncol(x)))
    for (r in seq_len(R)) {
      lo <- (r - 1) * L + 1
      hi <- min(r * L, l_x)
      tmp <- x[lo:hi, , drop = FALSE]
      if ( legacy ) {
        tmp <- apply(tmp, 2L, function(tmpi) {
          ifft(fft(postpad(tmpi, N)) * B)
        })
      } else {
        tmp <- stats::mvfft(stats::mvfft(postpad(tmp, N)) * B, inverse = TRUE) / N
      }
      hi <- min(lo + N - 1, l_x)
      y[lo:hi, ] <- y[lo:hi, ] + tmp[1:(hi - lo + 1), ]
    }
  }
  y <- y[1:l_x, , drop = is_vector]
  if (is.numeric(b) && is.numeric(x)) {
    y <- Re(y)
  }
  if (!any(as.logical(b - round(b)))) {
    idx <- !any(as.logical(x - round(x)))
    y[idx] <- round(y[idx])
  }
  return(y)

  # b <- rnorm(5)
  # x <- rnorm(500)
  # n <- NULL
  # range(fftfilt(b, x) - fftfilt(b, x, legacy = TRUE))
  # range(fftfilt(b, array(x, c(250,2))) - fftfilt(b, array(x, c(250,2)), legacy = TRUE))
  # fftfilt(b, array(x, c(250,2))) - fftfilt(b, array(x, c(250,2)), legacy = TRUE)

}

#' @title Estimate per-node curvature of a surface mesh
#' @description
#' Estimates, at every vertex of a closed triangular mesh, the local mean
#' curvature, Gaussian curvature, and the two principal curvatures, by fitting
#' an osculating quadratic surface to each vertex's neighborhood.
#'
#' @details
#' For each vertex, the function:
#' \enumerate{
#'   \item builds an orthonormal tangent frame \eqn{(e_1, e_2, n)}, where
#'     \eqn{n} is the vertex's (already-computed) unit normal;
#'   \item expresses each neighbor's offset from the vertex in this frame as
#'     tangential coordinates \eqn{(u, w)} and a height \eqn{h} above the
#'     tangent plane along \eqn{n};
#'   \item fits, by least squares over the vertex's 2-ring neighborhood, the
#'     osculating \verb{paraboloid} \eqn{h = a u^2 + b u w + c w^2};
#'   \item derives the mean curvature \eqn{H = a + c}, the Gaussian curvature
#'     \eqn{K = 4ac - b^2}, and the principal curvatures
#'     \eqn{k_{1,2} = H \pm \sqrt{\max(H^2 - K,\ 0)}}.
#' }
#' Because the tangent frame is orthonormal and the \verb{paraboloid} is fitted with
#' zero gradient at the vertex (by construction, since \eqn{n} is the fitted
#' normal direction), these are exactly the eigenvalues of the second
#' fundamental form, i.e. the principal curvatures.
#'
#' The sign of the result follows the orientation of the per-vertex outward
#' normal: a locally convex ('\verb{gyrus}-like') patch, where neighbors lie
#' toward the surface's interior relative to the outward normal, has negative
#' mean curvature, while a locally concave ('\verb{sulcus}-like') patch has
#' positive mean curvature. For example, a sphere of radius \eqn{r} with outward-pointing
#' normals has uniform curvature \eqn{H = -1/r} and \eqn{K = 1/r^2} everywhere.
#'
#' Vertices whose 2-ring neighborhood is degenerate (fewer than three
#' neighbors, or neighbor offsets that do not span the tangent plane, e.g.
#' nearly \verb{collinear}) are reported with all four curvature values set to zero.
#'
#' @param mesh triangular mesh of class \code{'mesh3d'} (or coercible via
#' \code{\link{ensure_mesh3d}}). The mesh \strong{must} be closed, manifold,
#' and genus-0 (watertight, no boundary or non-manifold edges, single
#' connected component); \code{mris_curvature} raises an error if such defects
#' are detected
#' @param verbose logical; print progress messages. Default \code{FALSE}
#'
#' @returns A named list of four numeric vectors, each of length
#' \code{ncol(mesh$vb)} (one entry per vertex, in vertex order):
#' \describe{
#'   \item{\code{mean}}{Mean curvature \eqn{H = (k_1 + k_2)/2}.}
#'   \item{\code{gaussian}}{Gaussian curvature \eqn{K = k_1 k_2}.}
#'   \item{\code{k1}}{First principal curvature (\eqn{k_1 \geq k_2}).}
#'   \item{\code{k2}}{Second principal curvature.}
#' }
#'
#' @references
#' Cortical surface-based analysis II: Inflation, flattening, and a
#' surface-based coordinate system. \emph{NeuroImage}, 9(2), 195-207 (1999).
#'
#' @examples
#'
#'
#' if (is_not_cran()) {
#'
#' data("left_hippocampus_mask")
#' mesh <- vcg_isosurface(left_hippocampus_mask)
#'
#' plot(mesh)
#'
#' # Fix defects
#' mesh <- vcg_fix_defects(mesh, verbose = TRUE)
#'
#' # Smooth
#' smoothed <- mris_smooth(mesh, verbose = TRUE)
#'
#' res <- mris_curvature(smoothed)
#'
#' range(res$k1)
#'
#' col <- color_ramp_continuous(
#'   res$k1, clim = c(-1, 1), alpha = TRUE,
#'   cmap = c("black", "gray", "red"))
#'
#' plot(smoothed, col = list(col),
#'      eye = c(-100, 100, 0), up = c(0, 0, 1))
#'
#' }
#'
#'
#' @export
mris_curvature <- function(mesh, verbose = FALSE) {
    mesh <- meshintegrity(mesh, facecheck = TRUE)

    vb <- mesh$vb[1:3, , drop = FALSE]
    storage.mode(vb) <- "double"

    it <- mesh$it
    storage.mode(it) <- "integer"

    tmp <- mrisCurvature(
        vb_     = vb,
        it_     = it,
        verbose = as.logical(verbose)[[1L]]
    )

    list(
        mean     = as.numeric(tmp$mean),
        gaussian = as.numeric(tmp$gaussian),
        k1       = as.numeric(tmp$k1),
        k2       = as.numeric(tmp$k2)
    )
}

Vector3 <- R6::R6Class(
  classname = "Vector3",
  portable = TRUE,
  cloneable = FALSE,
  lock_objects = FALSE,
  private = list(
    .extern_ptr = NULL
  ),
  active = list(
    is_vector3 = function() { TRUE },
    pointer = function() { private$.extern_ptr }
  ),
  public = list(
    initialize = function() {
      private$.extern_ptr <- Vector3__new()
    },
    format = function(...) {
      l <- self$get_size()
      s <- sprintf("<Vector3: len=%d>", l)
      if (l == 0) { return(s) }
      arr <- self$to_array(max_n_elems = 5)
      s <- c(
        s,
        utils::capture.output({
          print(data.frame(
            x = arr[1, ],
            y = arr[2, ],
            z = arr[3, ]
          ), ...)
        })
      )

      if ( l > 5 ) {
        s <- c(s, "...")
      }
      s
    },
    clone2 = function(...) {
      Vector3$new()$copy(self)
    },
    set = function(x, y, z) {
      data <- as.numeric(rbind(x, y, z))
      Vector3__from_array(private$.extern_ptr, data)
      self
    },
    from_array = function(data, offset = 0L, n_elems = -1L) {
      Vector3__from_array(private$.extern_ptr,
                          as.numeric(data),
                          offset = as.integer(offset),
                          n_elems = as.integer(n_elems))
      self
    },
    resize = function(n_elems) {
      Vector3__resize(private$.extern_ptr, n_elems = as.integer(n_elems))
      self
    },
    get_size = function() {
      Vector3__get_size(private$.extern_ptr)
    },
    to_array = function(n_skip = 0L, max_n_elems = -1L) {
      Vector3__to_array(private$.extern_ptr, n_skip = as.integer(n_skip), max_n_elems = as.integer(max_n_elems))
    },
    set_scalar = function(value) {
      Vector3__set_scalar(private$.extern_ptr, as.numeric(value))
      self
    },
    set_x = function(value) {
      Vector3__set_x(private$.extern_ptr, as.numeric(value))
      self
    },
    set_y = function(value) {
      Vector3__set_y(private$.extern_ptr, as.numeric(value))
      self
    },
    set_z = function(value) {
      Vector3__set_z(private$.extern_ptr, as.numeric(value))
      self
    },
    get_x = function(i) {
      Vector3__get_x(private$.extern_ptr, as.integer(i))
    },
    get_y = function(i) {
      Vector3__get_y(private$.extern_ptr, as.integer(i))
    },
    get_z = function(i) {
      Vector3__get_z(private$.extern_ptr, as.integer(i))
    },
    get_item = function(i) {
      Vector3__get_item(private$.extern_ptr, as.integer(i))
    },
    copy = function(v) {
      stopifnot(R6::is.R6(v) && isTRUE(v$is_vector3))
      Vector3__copy(private$.extern_ptr, v$pointer)
      self
    },
    add = function(v) {
      v <- as_vector3(v)
      Vector3__add(private$.extern_ptr, v$pointer)
      self
    },
    add_scalar = function(s) {
      Vector3__add_scalar(private$.extern_ptr, as.numeric(s))
      self
    },
    add_vectors = function(a, b) {
      a <- as_vector3(a)
      b <- as_vector3(b)
      Vector3__add_vectors(private$.extern_ptr, a$pointer, b$pointer)
      self
    },
    add_scaled = function(v, s) {
      v <- as_vector3(v)
      Vector3__add_scaled(private$.extern_ptr, v$pointer, as.numeric(s))
      self
    },
    sub = function(v) {
      v <- as_vector3(v)
      Vector3__sub(private$.extern_ptr, v$pointer)
      self
    },
    sub_scalar = function(s) {
      Vector3__sub_scalar(private$.extern_ptr, as.numeric(s))
      self
    },
    sub_vectors = function(a, b) {
      a <- as_vector3(a)
      b <- as_vector3(b)
      Vector3__sub_vectors(private$.extern_ptr, a$pointer, b$pointer)
      self
    },
    multiply = function(v) {
      v <- as_vector3(v)
      Vector3__multiply(private$.extern_ptr, v$pointer)
      self
    },
    multiply_scalar = function(s) {
      Vector3__multiply_scalar(private$.extern_ptr, as.numeric(s))
      self
    },
    multiply_vectors = function(a, b) {
      a <- as_vector3(a)
      b <- as_vector3(b)
      Vector3__multiply_vectors(private$.extern_ptr, a$pointer, b$pointer)
      self
    },
    divide = function(v) {
      v <- as_vector3(v)
      Vector3__divide(private$.extern_ptr, v$pointer)
      self
    },
    divide_scalar = function(s) {
      Vector3__divide_scalar(private$.extern_ptr, as.numeric(s))
      self
    },
    min = function(v) {
      v <- as_vector3(v)
      Vector3__min(private$.extern_ptr, v$pointer)
      self
    },
    max = function(v) {
      v <- as_vector3(v)
      Vector3__max(private$.extern_ptr, v$pointer)
      self
    },
    clamp = function(min_v, max_v) {
      min_v <- as_vector3(min_v)
      max_v <- as_vector3(max_v)
      Vector3__clamp(private$.extern_ptr, min_v$pointer, max_v$pointer)
      self
    },
    floor = function() {
      Vector3__floor(private$.extern_ptr)
      self
    },
    ceil = function() {
      Vector3__ceil(private$.extern_ptr)
      self
    },
    round = function() {
      Vector3__round(private$.extern_ptr)
      self
    },
    round_to_zero = function() {
      Vector3__round_to_zero(private$.extern_ptr)
      self
    },
    negate = function() {
      Vector3__negate(private$.extern_ptr)
      self
    },
    dot = function(v) {
      v <- as_vector3(v)
      Vector3__dot(private$.extern_ptr, v$pointer)
    },
    length_squared = function() {
      Vector3__length_squared(private$.extern_ptr)
    },
    length = function() {
      Vector3__length(private$.extern_ptr)
    },
    length_manhattan = function() {
      Vector3__length_manhattan(private$.extern_ptr)
    },
    normalize = function() {
      Vector3__normalize(private$.extern_ptr)
      self
    },
    set_length = function(len) {
      Vector3__set_length(private$.extern_ptr, as.numeric(len))
      self
    },
    lerp = function(v, alpha) {
      v <- as_vector3(v)
      Vector3__lerp(private$.extern_ptr, v$pointer, as.numeric(alpha))
      self
    },
    lerp_vectors = function(v1, v2, alpha) {
      v1 <- as_vector3(v1)
      v2 <- as_vector3(v2)
      Vector3__lerp_vectors(private$.extern_ptr, v1$pointer, v2$pointer, as.numeric(alpha))
      self
    },
    cross = function(v) {
      v <- as_vector3(v)
      Vector3__cross(private$.extern_ptr, v$pointer)
      self
    },
    cross_vectors = function(a, b) {
      a <- as_vector3(a)
      b <- as_vector3(b)
      Vector3__cross_vectors(private$.extern_ptr, a$pointer, b$pointer)
      self
    },

    apply_matrix3 = function(m) {
      stopifnot(length(m) == 9)
      if (!is.numeric(m)) { m <- as.numeric(m) }
      Vector3__apply_matrix3(private$.extern_ptr, m)
      self
    },
    apply_matrix4 = function(m) {
      m <- as_matrix4(m)
      Vector3__apply_matrix4(private$.extern_ptr, m$pointer)
      self
    },
    apply_quaternion = function(q) {
      q <- as_quaternion(q)
      Vector3__apply_quaternion(private$.extern_ptr, q$pointer)
      self
    },
    apply_axis_angle = function(axis, angle) {
      axis <- as_vector3(axis)
      angle <- as.double(angle)[[1]]
      Vector3__apply_axis_angle(axis, angle)
      self
    },
    transform_direction = function(m) {
      m <- as_matrix4(m)
      Vector3__transform_direction(private$.extern_ptr, m$pointer)
      self
    },
    project_on_vector = function(v) {
      v <- as_vector3(v)
      Vector3__project_on_vector(private$.extern_ptr, v$pointer)
      self
    },
    project_on_plane = function(normal) {
      normal <- as_vector3(normal)
      Vector3__project_on_plane(private$.extern_ptr, normal$pointer)
      self
    },
    reflect = function(normal) {
      normal <- as_vector3(normal)
      Vector3__reflect(private$.extern_ptr, normal$pointer)
      self
    },
    angle_to = function(v) {
      v <- as_vector3(v)
      return( Vector3__angle_to(private$.extern_ptr, v$pointer) )
    },
    distance_to = function(v) {
      v <- as_vector3(v)
      return( Vector3__distance_to(private$.extern_ptr, v$pointer) )
    },
    distance_to_squared = function(v) {
      v <- as_vector3(v)
      return( Vector3__distance_to_squared(private$.extern_ptr, v$pointer) )
    },
    distance_to_manhattan = function(v) {
      v <- as_vector3(v)
      return( Vector3__distance_to_manhattan(private$.extern_ptr, v$pointer) )
    },
    set_from_matrix_position = function(m) {
      m <- as_matrix4(m)
      Vector3__set_from_matrix_position(private$.extern_ptr, m$pointer)
      self
    },
    set_from_matrix_scale = function(m) {
      m <- as_matrix4(m)
      Vector3__set_from_matrix_scale(private$.extern_ptr, m$pointer)
      self
    },
    set_from_spherical_coords = function(radius, phi, theta) {
      Vector3__set_from_spherical_coords(private$.extern_ptr,
                                         as.numeric(radius)[[1]],
                                         as.numeric(phi)[[1]],
                                         as.numeric(theta)[[1]])
      self
    }
  )
)

#' Create a \code{Vector3} instance to store '3D' points
#' @description
#' Create instances that mimic the \code{'three.js'} syntax.
#'
#' @param x,y,z numeric, must have the same length, \code{'xyz'} positions
#' @param v R object to be converted to \code{Vector3} instance
#' @returns A \code{Vector3} instance
#'
#' @examples
#'
#' vec3 <- new_vector3(
#'   x = 1:9,
#'   y = 9:1,
#'   z = rep(c(1,2,3), 3)
#' )
#'
#' vec3[]
#'
#' # transform
#' m <- new_matrix4()
#'
#' # rotation xy plane by 30 degrees
#' m$make_rotation_z(pi / 6)
#'
#' vec3$apply_matrix4(m)
#'
#' vec3[]
#'
#' as_vector3(c(1,2,3))
#'
#' @seealso \code{\link{new_matrix4}}, \code{\link{new_quaternion}}
#' @export
new_vector3 <- function(x = 0.0, y = 0.0, z = 0.0) {
  Vector3$new()$set(x = x, y = y, z = z)
}

#' @export
as.matrix.Vector3 <- function(x, ...) {
  if (R6::is.R6(x) && isTRUE(x$is_vector3)) {
    return(x$to_array(...))
  }
  NextMethod("as.matrix")
}

#' @export
`[.Vector3` <- function(x, i, ..., drop = TRUE) {
  if (missing(i)) {
    x$to_array()[, ..., drop = drop]
  } else {
    x$to_array()[i, ..., drop = drop]
  }
}

#' @export
dim.Vector3 <- function(x) {
  if (R6::is.R6(x) && isTRUE(x$is_vector3)) {
    return(c(3L, x$get_size()))
  }
  NextMethod("dim")
}

#' @export
`==.Vector3` <- function(e1, e2) {
  e2 == e1$to_array()
}

#' @rdname new_vector3
#' @export
as_vector3 <- function(v) {
  if ( R6::is.R6(v) && isTRUE(v$is_vector3) ) { return(v) }
  x_ <- as.double(v)
  if (length(x_) == 1) {
    return(new_vector3(x_, x_, x_))
  }
  if (length(x_) < 3) {
    stop("Input cannot be converted to a Vector3 instance. Please make sure input is a numeric vector of which the length is greater equal to 3.")
  }
  new_vector3()$from_array(x_)
}

test_that("resample3D", {
  arrayDim <- rev(c(256, 128, 16))
  fromArray <- array(rnorm(prod(arrayDim)), rev(arrayDim) + 1)
  oldVoxToWorld <- matrix(nrow = 4, byrow = TRUE, c(
    1.5, 0, 0.8, -200,
    0.01, 0, -1, 10,
    0, -0.8, 0.1, 60,
    0, 0, 0, 1
  ))
  newVoxToWorld <- matrix(nrow = 4, byrow = TRUE, c(
    15, 0, 0.08, -140,
    0.1, -0.5, 0, 30,
    0, 0.1, 0.1, -13.6,
    0, 0, 0, 1
  ))

  # Make sure the resulting data has valid data so this test can be valid
  expect_true(all(solve(oldVoxToWorld) %*% newVoxToWorld %*% c(arrayDim / 2, 1) > 0))
  expect_true(all((
    solve(oldVoxToWorld) %*% newVoxToWorld %*% c(arrayDim / 2, 1)
  )[1:3] < dim(fromArray)))

  reList <- resample3D(arrayDim, fromArray, t(newVoxToWorld), t(oldVoxToWorld), na = NA_real_)

  # reList[[2]] is the point transform of indexing space from new array to old array
  dim(reList[[2]]) <- c(4, 4)
  mat_actual <- reList[[2]]
  mat_expected <- solve(oldVoxToWorld) %*% newVoxToWorld

  mat <- mat_actual %*% solve(mat_expected)
  diag(mat) <- diag(mat) - 1

  # should be about 1e-14, but 1e-5 should be fine
  expect_lt(max(abs(mat)), 1e-5)

  get_expected_array <- function(na_fill = NA_real_) {
    # compare with manual
    expected_array <- array(na_fill, arrayDim)

    # construct index set in ijk1 format
    idx <- t(cbind(arrayInd(seq_len(prod(arrayDim)), arrayDim) - 1, 1))

    # calculate index in old array
    idx <- round(solve(oldVoxToWorld) %*% newVoxToWorld %*% idx)[1:3, , drop = FALSE]

    # remove outofbound indexings
    idx[idx < 0 | idx >= dim(fromArray)] <- NA

    # get 1-index
    idx <- colSums(idx * c(1, cumprod(dim(fromArray)))[1:3]) + 1

    expected_array[!is.na(idx)] <- fromArray[idx[!is.na(idx)]]
    expected_array
  }

  expected_array <- get_expected_array(NA_real_)
  expect_equal(storage.mode(reList[[1]]), storage.mode(fromArray))
  expect_equal(reList[[1]], expected_array)



  # check if other formats work too

  # integer
  fromArray <- fromArray * 10
  storage.mode(fromArray) <- "integer"
  reList <- resample3D(arrayDim, fromArray, t(newVoxToWorld), t(oldVoxToWorld), na = NA_integer_)
  dim(reList[[2]]) <- c(4, 4)
  expect_equal(reList[[2]], expected = mat_actual)
  expect_equal(storage.mode(reList[[1]]), "integer")
  expect_equal(reList[[1]], get_expected_array(NA_integer_))

  # string
  storage.mode(fromArray) <- "character"
  reList <- resample3D(arrayDim, fromArray, t(newVoxToWorld), t(oldVoxToWorld), na = NA_character_)
  dim(reList[[2]]) <- c(4, 4)
  expect_equal(reList[[2]], expected = mat_actual)
  expect_equal(storage.mode(reList[[1]]), "character")
  expect_equal(reList[[1]], get_expected_array(NA_character_))

  # complex
  storage.mode(fromArray) <- "complex"
  reList <- resample3D(arrayDim, fromArray, t(newVoxToWorld), t(oldVoxToWorld), na = NA_complex_)
  dim(reList[[2]]) <- c(4, 4)
  expect_equal(reList[[2]], expected = mat_actual)
  expect_equal(storage.mode(reList[[1]]), "complex")
  expect_equal(reList[[1]], get_expected_array(NA_complex_))

  # raw
  fromArray <- array(sample(255, size = length(fromArray), replace = TRUE), dim(fromArray))
  storage.mode(fromArray) <- "raw"
  reList <- resample3D(arrayDim, fromArray, t(newVoxToWorld), t(oldVoxToWorld), na = as.raw(0))
  dim(reList[[2]]) <- c(4, 4)
  expect_equal(reList[[2]], expected = mat_actual)
  expect_equal(storage.mode(reList[[1]]), "raw")
  expect_equal(reList[[1]], get_expected_array(as.raw(0)))

  # logical
  fromArray <- array(sample(c(TRUE, FALSE, NA_integer_), size = length(fromArray), replace = TRUE), dim(fromArray))
  storage.mode(fromArray) <- "logical"
  reList <- resample3D(arrayDim, fromArray, t(newVoxToWorld), t(oldVoxToWorld), na = as.logical(NA))
  dim(reList[[2]]) <- c(4, 4)
  expect_equal(reList[[2]], expected = mat_actual)
  expect_equal(storage.mode(reList[[1]]), "logical")
  expect_equal(reList[[1]], get_expected_array(as.logical(NA)))

})

# Synthetic CCEP-like data with a shared canonical shape plus per-trial scaling.
# `delay` shifts the response onset so the pre-response window is pure noise.
crp_sim <- function(n_time = 600L, n_trials = 30L, delay = 0,
                    t0 = -0.05, t1 = 1.2, noise = 0.25, seed = 1L) {
  set.seed(seed)
  tt <- seq(t0, t1, length.out = n_time)
  peak1 <- delay + 0.10
  peak2 <- delay + 0.30
  shape <- exp(-((tt - peak1) / 0.05)^2) - 0.5 * exp(-((tt - peak2) / 0.10)^2)
  shape[tt < delay] <- 0
  V <- outer(shape, runif(n_trials, 0.6, 1.4)) +
    matrix(rnorm(n_time * n_trials, sd = noise), n_time, n_trials)
  list(V = V * 3, time = tt)
}

test_that("crp returns C_full spanning the entire loaded time, matching C", {
  sim <- crp_sim(seed = 11L)
  res <- crp(sim$V, sim$time)

  parms <- res$parameters
  eps <- sqrt(.Machine$double.eps)

  # C_full spans the entire loaded time range (incl. pre-stimulus t < 0)
  expect_equal(length(parms$C_full), length(sim$time))
  expect_equal(length(parms$params_times_full), length(parms$C_full))
  expect_equal(parms$params_times_full, sim$time)
  expect_equal(res$.data$time, sim$time)

  # reported C (detect_onset = FALSE) is the C_full slice over [t_start, tau_R]
  sel <- parms$params_times_full >= res$t_start &
    parms$params_times_full <= res$tau_R
  expect_equal(parms$C, parms$C_full[sel], tolerance = eps)
  expect_equal(parms$params_times, parms$params_times_full[sel])

  # default C equals the forward unit-norm eigenvector
  expect_equal(sqrt(sum(parms$C^2)), 1, tolerance = eps)

  # tR_sample (window index) matches the reported-C length and tau_R
  tr <- length(parms$C)
  expect_identical(res$projections$tR_sample, tr)
  expect_equal(res$tau_R, parms$params_times[tr])
})

test_that("detect_onset leaves the forward loading (alpha, C_full) untouched", {
  sim <- crp_sim(delay = 0.3, seed = 7L)
  res_off <- crp(sim$V, sim$time, detect_onset = FALSE)
  res_on  <- crp(sim$V, sim$time, detect_onset = TRUE)

  # forward quantities are byte-for-byte identical regardless of onset
  expect_equal(res_on$parameters$al, res_off$parameters$al)
  expect_equal(res_on$parameters$al_p, res_off$parameters$al_p)
  expect_equal(res_on$parameters$Vsnr, res_off$parameters$Vsnr)
  expect_equal(res_on$parameters$expl_var, res_off$parameters$expl_var)
  expect_equal(res_on$parameters$C_full, res_off$parameters$C_full)
  expect_equal(res_on$parameters$ep, res_off$parameters$ep)
  expect_equal(res_on$tau_R, res_off$tau_R)
  expect_identical(res_on$bad_trials, res_off$bad_trials)
})

test_that("detect_onset trims the reported C to [tau_onset, tau_R]", {
  sim <- crp_sim(delay = 0.3, seed = 7L)
  res <- crp(sim$V, sim$time, detect_onset = TRUE)

  expect_true(is.finite(res$tau_onset))
  expect_gte(res$tau_onset, res$t_start)
  expect_lte(res$tau_onset, res$tau_R)
  expect_true(res$tau_onset_lower <= res$tau_onset)
  expect_true(res$tau_onset_upper >= res$tau_onset)

  parms <- res$parameters
  pt <- parms$params_times
  expect_equal(pt[1L], res$tau_onset)
  expect_equal(pt[length(pt)], res$tau_R)
  expect_equal(length(parms$C), length(pt))

  # the trimmed C is exactly the matching slice of C_full
  full_times <- parms$params_times_full
  sel <- full_times >= res$tau_onset & full_times <= res$tau_R
  expect_equal(parms$C, parms$C_full[sel], tolerance = sqrt(.Machine$double.eps))
})

test_that("onset scan recovers a delayed response latency", {
  delay <- 0.30
  sim <- crp_sim(delay = delay, noise = 0.2, seed = 3L)
  res <- crp(sim$V, sim$time, t_start = 0.02, t_end = 1.1,
             detect_onset = TRUE)

  # estimated onset should land near the true delay
  expect_lt(abs(res$tau_onset - delay), 0.1)
  # and well before the response duration tau_R
  expect_lt(res$tau_onset, res$tau_R)
})

test_that("onset can be detected before t_start", {
  delay <- 0.05
  sim <- crp_sim(delay = delay, noise = 0.2, seed = 3L)
  # t_start deliberately set above the true onset
  res <- crp(sim$V, sim$time, t_start = 0.1, t_end = 1.1,
             detect_onset = TRUE)

  expect_true(is.finite(res$tau_onset))
  expect_lt(res$tau_onset, res$t_start)        # earlier than the window start
  expect_lt(abs(res$tau_onset - delay), 0.1)   # near the true latency

  # reported C is sliced from C_full and starts at tau_onset (< t_start)
  parms <- res$parameters
  sel <- parms$params_times_full >= res$tau_onset &
    parms$params_times_full <= res$tau_R
  expect_equal(parms$C, parms$C_full[sel], tolerance = sqrt(.Machine$double.eps))
  expect_lt(parms$params_times[1L], res$t_start)
})

test_that("onset_search_start floors the reverse scan and validates", {
  sim <- crp_sim(delay = 0.05, noise = 0.2, seed = 3L)
  res <- crp(sim$V, sim$time, t_start = 0.1, t_end = 1.1,
             detect_onset = TRUE, onset_search_start = 0.08)
  expect_gte(res$tau_onset, 0.08)

  expect_error(
    crp(sim$V, sim$time, detect_onset = TRUE, onset_search_start = c(0, 1)),
    "onset_search_start"
  )
  expect_error(
    crp(sim$V, sim$time, detect_onset = TRUE, onset_search_start = NA),
    "onset_search_start"
  )
})

test_that("crp drops non-finite time points (NA / NaN / Inf)", {
  sim <- crp_sim(seed = 8L)
  V <- sim$V
  bad <- c(5L, 17L, 100L)
  V[bad[1L], 1L] <- NA
  V[bad[2L], 3L] <- NaN
  V[bad[3L], 2L] <- Inf

  res <- crp(V, sim$time)
  expect_equal(nrow(res$.data$V), length(sim$time) - length(bad))
  expect_equal(length(res$.data$time), length(sim$time) - length(bad))
  expect_false(any(sim$time[bad] %in% res$.data$time))
  expect_equal(length(res$parameters$C_full), length(sim$time) - length(bad))
})

test_that("crp clamps t_start / t_end instead of erroring", {
  sim <- crp_sim(seed = 8L)

  # t_end beyond the last sample -> clamped to max(time), no error
  res <- crp(sim$V, sim$time, t_end = 99)
  expect_equal(res$t_end, max(sim$time))

  # t_start below the first sample -> clamped to min(time)
  res2 <- crp(sim$V, sim$time, t_start = -10, t_end = 1)
  expect_equal(res2$t_start, min(sim$time))

  # impossible window still errors
  expect_error(crp(sim$V, sim$time, t_start = 0.9, t_end = 0.5),
               "t_start.*less than.*t_end")
})

test_that("crp coerces detect_onset silently (isTRUE(as.logical(...)))", {
  sim <- crp_sim(seed = 5L)
  # NA, multi-value, and non-logical all coerce to FALSE — no error
  expect_error(crp(sim$V, sim$time, detect_onset = NA), NA)
  expect_error(crp(sim$V, sim$time, detect_onset = c(TRUE, FALSE)), NA)
  expect_error(crp(sim$V, sim$time, detect_onset = "yes"), NA)
  # "yes" coerces to NA via as.logical, then isTRUE -> FALSE
  r <- crp(sim$V, sim$time, detect_onset = "yes")
  expect_true(is.na(r$tau_onset))
})

test_that("plot.ravetools_crp runs with and without onset", {
  sim <- crp_sim(delay = 0.3, seed = 9L)
  res_off <- crp(sim$V, sim$time, detect_onset = FALSE)
  res_on  <- crp(sim$V, sim$time, detect_onset = TRUE)

  tmp <- tempfile(fileext = ".pdf")
  grDevices::pdf(tmp)
  on.exit({
    grDevices::dev.off()
    unlink(tmp)
  }, add = TRUE)

  expect_error(plot(res_off), NA)
  expect_error(plot(res_on), NA)
})

test_that("vcg_detect_collision matches an independent oracle", {

  # ---- Independent oracle -------------------------------------------------
  # Densely sample each geometry and take the minimum pairwise distance. This
  # shares no algorithm with the implementation, so it catches the failure mode
  # that motivates the hand-written primitives: vcglib's SegmentSegmentDistance
  # clamps the two line-line closest points independently and overestimates by
  # up to ~4x on skew pairs.
  #
  # Sampling can only overestimate, and by at most half a sample spacing per
  # side, which gives a two-sided bound on the implementation.

  sample_point <- function(p) list(pts = matrix(p, nrow = 1L), h = 0)

  sample_segment <- function(a, b, n = 400L) {
    t <- seq(0, 1, length.out = n)
    list(
      pts = cbind(a[1] + t * (b[1] - a[1]),
                  a[2] + t * (b[2] - a[2]),
                  a[3] + t * (b[3] - a[3])),
      h = sqrt(sum((b - a)^2)) / (n - 1L)
    )
  }

  sample_triangle <- function(v0, v1, v2, k = 40L) {
    grid <- expand.grid(i = 0:k, j = 0:k)
    grid <- grid[grid$i + grid$j <= k, , drop = FALSE]
    u <- grid$i / k
    v <- grid$j / k
    w <- 1 - u - v
    edge <- max(sqrt(sum((v1 - v0)^2)), sqrt(sum((v2 - v1)^2)), sqrt(sum((v0 - v2)^2)))
    list(
      pts = cbind(w * v0[1] + u * v1[1] + v * v2[1],
                  w * v0[2] + u * v1[2] + v * v2[2],
                  w * v0[3] + u * v1[3] + v * v2[3]),
      h = edge / k
    )
  }

  min_pair_dist <- function(A, B) {
    best <- Inf
    for (i in seq_len(nrow(A))) {
      best <- min(best, min(sqrt(colSums((t(B) - A[i, ])^2))))
    }
    best
  }

  # `impl` is the implementation's answer, `sa`/`sb` the sampled geometries.
  expect_matches_oracle <- function(impl, sa, sb, label) {
    sampled <- min_pair_dist(sa$pts, sb$pts)
    tol <- (sa$h + sb$h) / 2 + 1e-5
    # Must never overestimate (the vcglib bug class) ...
    expect_lt(impl, sampled + 1e-5)
    # ... and must not underestimate beyond the sampling resolution.
    expect_lt(sampled - impl, tol)
    invisible(label)
  }

  measure <- function(x, y, mode_x, mode_y) {
    # `radius = Inf` is not allowed, so use a bound larger than any distance here
    res <- vcg_detect_collision(x, y, mode_x = mode_x, mode_y = mode_y,
                                radius = 1000)
    res$distance[[1L]]
  }

  set.seed(20260819)

  # -- point vs point -------------------------------------------------------
  for (i in 1:20) {
    p <- matrix(runif(6, -3, 3), 2L, 3L)
    d <- measure(p[1, , drop = FALSE], p[2, , drop = FALSE], "points", "points")
    expect_equal(d, sqrt(sum((p[1, ] - p[2, ])^2)), tolerance = 1e-5)
  }

  # -- point vs segment -----------------------------------------------------
  for (i in 1:20) {
    p <- matrix(runif(9, -3, 3), 3L, 3L)
    d <- measure(p[1, , drop = FALSE], p[2:3, , drop = FALSE], "points", "segments")
    expect_matches_oracle(d, sample_point(p[1, ]), sample_segment(p[2, ], p[3, ]),
                          "point-segment")
  }

  # -- segment vs segment ---------------------------------------------------
  for (i in 1:40) {
    p <- matrix(runif(12, -3, 3), 4L, 3L)
    d <- measure(p[1:2, , drop = FALSE], p[3:4, , drop = FALSE], "segments", "segments")
    expect_matches_oracle(d, sample_segment(p[1, ], p[2, ]),
                          sample_segment(p[3, ], p[4, ]), "segment-segment")
  }

  # -- point vs triangle ----------------------------------------------------
  one_triangle <- function(v) {
    list(vb = rbind(t(v), 1), it = matrix(1:3, nrow = 3L),
         class = "mesh3d")
  }
  as_mesh <- function(v) {
    structure(list(vb = rbind(t(v), 1), it = matrix(1:3, ncol = 1L)),
              class = c("ravetools_mesh3d", "mesh3d"))
  }

  for (i in 1:20) {
    v <- matrix(runif(9, -3, 3), 3L, 3L)
    p <- matrix(runif(3, -3, 3), 1L, 3L)
    d <- measure(p, as_mesh(v), "points", "mesh")
    expect_matches_oracle(d, sample_point(p[1, ]),
                          sample_triangle(v[1, ], v[2, ], v[3, ]), "point-triangle")
  }

  # -- segment vs triangle --------------------------------------------------
  for (i in 1:20) {
    v <- matrix(runif(9, -3, 3), 3L, 3L)
    s <- matrix(runif(6, -3, 3), 2L, 3L)
    d <- measure(s, as_mesh(v), "segments", "mesh")
    expect_matches_oracle(d, sample_segment(s[1, ], s[2, ]),
                          sample_triangle(v[1, ], v[2, ], v[3, ]), "segment-triangle")
  }

  # -- triangle vs triangle -------------------------------------------------
  for (i in 1:20) {
    a <- matrix(runif(9, -3, 3), 3L, 3L)
    b <- matrix(runif(9, -3, 3), 3L, 3L)
    d <- measure(as_mesh(a), as_mesh(b), "mesh", "mesh")
    expect_matches_oracle(d, sample_triangle(a[1, ], a[2, ], a[3, ]),
                          sample_triangle(b[1, ], b[2, ], b[3, ]), "triangle-triangle")
  }
})


test_that("vcg_detect_collision is exact where vcglib's own routine is not", {
  # vcg::SegmentSegmentDistance returns 4.123 for this pair; the truth is 3.0.
  x <- rbind(c(0, 0, 0), c(1, 0, 0))
  y <- rbind(c(5, 0, 1), c(1, 4, 1))
  res <- vcg_detect_collision(x, y, mode_x = "segments", mode_y = "segments",
                              radius = 10)
  expect_equal(res$distance[[1L]], 3, tolerance = 1e-5)

  # The case the tractography viewer gets wrong: a segment passing straight
  # through a single-point ROI, with both of its vertices offset. Vertex
  # sampling reports 0.781; the true segment distance is 0.
  roi <- rbind(c(0, 0, 0))
  seg <- rbind(c(-0.781, 0, 0), c(0.781, 0, 0))
  res <- vcg_detect_collision(roi, seg, mode_y = "segments", radius = 1)
  expect_true(res$hit[[1L]])
  expect_equal(res$distance[[1L]], 0, tolerance = 1e-6)
})


test_that("vcg_detect_collision agrees with vcg_kdtree_nearest on point clouds", {
  set.seed(7)
  target <- matrix(rnorm(300), ncol = 3L)
  query <- matrix(rnorm(600), ncol = 3L)

  for (radius in c(0.25, 0.5, 1, 2)) {
    res <- vcg_detect_collision(target, query, mode_x = "points",
                                mode_y = "points", radius = radius)
    knn <- vcg_kdtree_nearest(target = target, query = query, k = 1L)

    expect_identical(as.logical(res$hit), as.vector(knn$distance <= radius))

    # Where there is a hit, both the distance and the reported target row must
    # match the nearest neighbour.
    hit <- which(res$hit)
    expect_equal(res$distance[hit], as.vector(knn$distance)[hit], tolerance = 1e-5)
    expect_identical(res$index[hit], as.integer(knn$index)[hit])
  }
})


test_that("vcg_detect_collision is symmetric in its two sides", {
  set.seed(11)
  a <- matrix(rnorm(60), ncol = 3L)
  b <- matrix(rnorm(90), ncol = 3L)

  for (modes in list(c("points", "points"), c("points", "segments"),
                     c("segments", "segments"))) {
    forward <- vcg_detect_collision(a, b, mode_x = modes[[1L]],
                                    mode_y = modes[[2L]], radius = 1.5)
    reverse <- vcg_detect_collision(b, a, mode_x = modes[[2L]],
                                    mode_y = modes[[1L]], radius = 1.5)

    expect_identical(any(forward$hit, na.rm = TRUE),
                     any(reverse$hit, na.rm = TRUE))
    expect_equal(min(forward$distance, na.rm = TRUE),
                 min(reverse$distance, na.rm = TRUE), tolerance = 1e-5)
  }
})


test_that("vcg_detect_collision accepts NA rows as polyline separators", {
  set.seed(3)
  roi <- matrix(rnorm(30), ncol = 3L)

  tract1 <- cbind(seq(-2, 2, by = 0.4), 0, 0)
  tract2 <- cbind(0, seq(-2, 2, by = 0.4), 1)
  combined <- rbind(tract1, c(NA, NA, NA), tract2)

  joint <- vcg_detect_collision(roi, combined, mode_y = "segments", radius = 1)

  sep <- nrow(tract1) + 1L
  expect_true(is.na(joint$hit[[sep]]))            # the separator row
  expect_true(is.na(joint$hit[[nrow(tract1)]]))   # last vertex of tract 1
  expect_true(is.na(joint$hit[[nrow(combined)]])) # last vertex of tract 2

  # Splitting the matrix into one call per polyline must give the same answers.
  solo1 <- vcg_detect_collision(roi, tract1, mode_y = "segments", radius = 1)
  solo2 <- vcg_detect_collision(roi, tract2, mode_y = "segments", radius = 1)

  expect_identical(joint$hit[seq_len(nrow(tract1))], solo1$hit)
  expect_equal(joint$distance[seq_len(nrow(tract1))], solo1$distance)
  expect_identical(joint$hit[(sep + 1L):nrow(combined)], solo2$hit)
  expect_equal(joint$distance[(sep + 1L):nrow(combined)], solo2$distance)

  # A separator on the indexed side must not create a phantom segment that
  # bridges the end of one polyline to the start of the next. This probe sits
  # near the midpoint of that would-be bridge and far from either polyline.
  a <- rbind(c(0, 0, 0), c(1, 0, 0))     # polyline 1
  b <- rbind(c(1, 20, 0), c(0, 20, 0))   # polyline 2, well away
  probe <- rbind(c(1, 10, 0))            # midway along the phantom bridge

  separated <- vcg_detect_collision(rbind(a, c(NA, NA, NA), b), probe,
                                    mode_x = "segments", radius = 1)
  bridged <- vcg_detect_collision(rbind(a, b), probe,
                                  mode_x = "segments", radius = 1)
  expect_false(separated$hit)   # nearest real segment is 10 away
  expect_true(bridged$hit)      # without the separator, the bridge passes through

  # Rows that are only partly finite are treated as separators too.
  partial <- combined
  partial[3L, 2L] <- NA_real_
  expect_true(is.na(vcg_detect_collision(roi, partial, mode_y = "segments",
                                         radius = 1)$hit[[3L]]))
})


test_that("vcg_detect_collision early_stop agrees with a full scan", {
  set.seed(5)
  roi <- matrix(rnorm(30), ncol = 3L)
  tracts <- rbind(
    cbind(seq(-2, 2, by = 0.25), 0, 0),
    c(NA, NA, NA),
    cbind(0.2, seq(-2, 2, by = 0.25), 0)
  )

  full <- vcg_detect_collision(roi, tracts, mode_y = "segments", radius = 0.8)
  early <- vcg_detect_collision(roi, tracts, mode_y = "segments", radius = 0.8,
                                early_stop = TRUE)

  # Every element the early-stopping scan did evaluate must agree.
  evaluated <- !is.na(early$hit)
  expect_identical(early$hit[evaluated], full$hit[evaluated])
  expect_equal(early$distance[evaluated], full$distance[evaluated])

  # Once it stops, it stops: nothing after the first hit in a group is
  # evaluated, and at most one hit is reported per group.
  expect_lte(sum(early$hit, na.rm = TRUE), 2L)
  expect_true(any(full$hit, na.rm = TRUE))

  # The unit is the NA-delimited group, not the row. A point cloud carrying no
  # separators is therefore ONE group, and stops after a single hit overall
  # rather than reporting every near point.
  origin <- rbind(c(0, 0, 0))
  cloud <- rbind(c(0, 0, 0), c(0.1, 0, 0), c(9, 9, 9), c(0.2, 0, 0))

  flat <- vcg_detect_collision(origin, cloud, mode_y = "points", radius = 0.5,
                               early_stop = TRUE)
  expect_identical(flat$hit, c(TRUE, NA, NA, NA))

  # Splitting the same cloud with a separator yields one hit per group. The
  # mode is stated explicitly throughout because these assertions are about
  # point-cloud scoping specifically, not about what `auto` happens to guess.
  grouped <- vcg_detect_collision(origin,
                                  rbind(cloud[1:2, ], c(NA, NA, NA), cloud[3:4, ]),
                                  mode_y = "points", radius = 0.5,
                                  early_stop = TRUE)
  expect_identical(grouped$hit, c(TRUE, NA, NA, FALSE, TRUE))

  # A target mesh is likewise a single group: it stops at the first colliding
  # face, where a full scan finds many.
  a <- vcg_sphere()
  b <- vcg_sphere()
  b$vb[1L, ] <- b$vb[1L, ] + 1
  expect_gt(sum(vcg_detect_collision(a, b, radius = 0)$hit), 1L)
  expect_identical(sum(vcg_detect_collision(a, b, radius = 0,
                                            early_stop = TRUE)$hit,
                       na.rm = TRUE), 1L)

  # The hit reported is the first in row order, not the closest approach. This
  # chain grazes the origin at 0.5 before returning at 0.1.
  chain <- rbind(c(-2, 0.5, 0), c(2, 0.5, 0), c(2, 0.1, 0), c(-2, 0.1, 0))
  scan_all <- vcg_detect_collision(origin, chain, mode_y = "segments",
                                   radius = 0.6)
  scan_early <- vcg_detect_collision(origin, chain, mode_y = "segments",
                                     radius = 0.6, early_stop = TRUE)

  expect_equal(min(scan_all$distance, na.rm = TRUE), 0.1, tolerance = 1e-5)
  expect_equal(scan_early$distance[which(scan_early$hit)], 0.5, tolerance = 1e-5)
  expect_gt(scan_early$distance[which(scan_early$hit)],
            min(scan_all$distance, na.rm = TRUE))
})


test_that("vcg_detect_collision accepted set grows monotonically with radius", {
  set.seed(13)
  roi <- matrix(rnorm(60), ncol = 3L)
  query <- matrix(rnorm(300), ncol = 3L)

  previous <- integer(0)
  for (radius in c(0.1, 0.3, 0.6, 1, 1.5, 3)) {
    hit <- which(vcg_detect_collision(roi, query, radius = radius)$hit)
    expect_true(all(previous %in% hit))
    previous <- hit
  }
  expect_gt(length(previous), 0L)

  # A radius that spans the whole bounding box must accept everything.
  span <- vcg_detect_collision(roi, query, radius = 1e6)
  expect_true(all(span$hit))
})


test_that("vcg_detect_collision handles mesh-to-mesh proximity", {
  a <- vcg_sphere()
  b <- vcg_sphere()
  b$vb[1L, ] <- b$vb[1L, ] + 3

  # Two unit spheres 3 apart leave a surface gap of about 1.
  expect_false(any(vcg_detect_collision(a, b, radius = 0.9)$hit))
  expect_true(any(vcg_detect_collision(a, b, radius = 1.1)$hit))

  # Overlapping spheres genuinely intersect, so some faces are at distance 0.
  c <- vcg_sphere()
  c$vb[1L, ] <- c$vb[1L, ] + 1
  overlap <- vcg_detect_collision(a, c, radius = 0)
  expect_true(any(overlap$hit))
  expect_equal(min(overlap$distance, na.rm = TRUE), 0, tolerance = 1e-6)

  # The reported index must point at a real face of `a`.
  hit <- which(overlap$hit)
  expect_true(all(overlap$index[hit] >= 1L & overlap$index[hit] <= ncol(a$it)))
})


test_that("vcg_detect_collision can count the interior of a closed surface", {
  roi <- vcg_sphere()
  centre <- rbind(c(0, 0, 0))

  expect_false(vcg_detect_collision(roi, centre, radius = 0.1)$hit)

  inside <- vcg_detect_collision(roi, centre, radius = 0.1,
                                 include_interior = TRUE)
  expect_true(inside$hit)
  expect_equal(inside$distance, 0)

  # A point well outside stays a miss either way.
  outside <- rbind(c(10, 10, 10))
  expect_false(vcg_detect_collision(roi, outside, radius = 0.1,
                                    include_interior = TRUE)$hit)

  # A segment that lies wholly inside is only caught with the interior test.
  seg <- rbind(c(-0.2, 0, 0), c(0.2, 0, 0))
  expect_false(any(vcg_detect_collision(roi, seg, mode_y = "segments",
                                        radius = 0.1)$hit, na.rm = TRUE))
  expect_true(any(vcg_detect_collision(roi, seg, mode_y = "segments",
                                       radius = 0.1,
                                       include_interior = TRUE)$hit,
                  na.rm = TRUE))

  # It needs a mesh on the indexed side.
  expect_error(
    vcg_detect_collision(matrix(rnorm(30), ncol = 3L), centre, radius = 1,
                         include_interior = TRUE),
    "include_interior"
  )
})


test_that("vcg_detect_collision validates and resolves its arguments", {
  roi <- vcg_sphere()
  pts <- matrix(rnorm(30), ncol = 3L)

  # `auto` picks mesh for a surface with faces, points for any matrix. Separator
  # rows do not change that guess, so a matrix of connected segments must say so.
  expect_identical(
    vcg_detect_collision(roi, pts, radius = 1),
    vcg_detect_collision(roi, pts, mode_x = "mesh", mode_y = "points",
                         radius = 1)
  )
  chained <- rbind(pts[1:4, ], c(NA, NA, NA), pts[5:8, ])
  expect_identical(
    vcg_detect_collision(roi, chained, radius = 1),
    vcg_detect_collision(roi, chained, mode_x = "mesh", mode_y = "points",
                         radius = 1)
  )

  # A matrix cannot be a mesh.
  expect_error(vcg_detect_collision(pts, pts, mode_x = "mesh", radius = 1),
               "no faces")
  expect_error(vcg_detect_collision(roi, pts, mode_y = "mesh", radius = 1),
               "no faces")

  expect_error(vcg_detect_collision(roi, pts, radius = -1), "radius")
  expect_error(vcg_detect_collision(roi, pts, radius = NA), "radius")
  expect_error(vcg_detect_collision(roi, pts, radius = Inf), "radius")

  # 2-column input is padded to 3D, as elsewhere in the package.
  flat <- vcg_detect_collision(matrix(c(0, 0), ncol = 2L),
                               matrix(c(3, 4), ncol = 2L), radius = 10)
  expect_equal(flat$distance, 5)
})


test_that("vcg_detect_collision handles degenerate input", {
  roi <- matrix(rnorm(30), ncol = 3L)

  # Empty target.
  empty <- vcg_detect_collision(roi, matrix(numeric(0), ncol = 3L), radius = 1)
  expect_length(empty$hit, 0L)
  expect_length(empty$distance, 0L)

  # Empty ROI: nothing can collide.
  none <- vcg_detect_collision(matrix(numeric(0), ncol = 3L), roi, radius = 1e6)
  expect_false(any(none$hit))

  # A single row in `segments` mode carries no segment at all.
  lone <- vcg_detect_collision(roi, rbind(c(0, 0, 0)), mode_y = "segments",
                               radius = 10)
  expect_true(is.na(lone$hit))

  # A zero-length segment behaves like the point it degenerates to.
  degenerate <- vcg_detect_collision(rbind(c(0, 0, 0)),
                                     rbind(c(1, 0, 0), c(1, 0, 0)),
                                     mode_y = "segments", radius = 10)
  expect_equal(degenerate$distance[[1L]], 1, tolerance = 1e-6)

  # Coincident points are at distance 0, and radius 0 still counts as contact.
  touching <- vcg_detect_collision(rbind(c(1, 2, 3)), rbind(c(1, 2, 3)),
                                   radius = 0)
  expect_true(touching$hit)
  expect_equal(touching$distance, 0)
})

# ---- vcg_max_edge_length -----------------------------------------------

test_that("vcg_max_edge_length returns a positive scalar", {
  sphere <- vcg_sphere()
  mel <- vcg_max_edge_length(sphere)
  expect_length(mel, 1L)
  expect_true(is.finite(mel))
  expect_gt(mel, 0)
})

test_that("vcg_max_edge_length >= vcg_average_edge_length", {
  sphere <- vcg_sphere()
  expect_gte(vcg_max_edge_length(sphere), vcg_average_edge_length(sphere))
})

test_that("vcg_max_edge_length agrees with brute-force calculation", {
  sphere <- vcg_sphere()
  vb <- sphere$vb[1:3, ]
  it <- sphere$it
  e1 <- vb[, it[1, ]] - vb[, it[2, ]]
  e2 <- vb[, it[2, ]] - vb[, it[3, ]]
  e3 <- vb[, it[3, ]] - vb[, it[1, ]]
  expected <- max(sqrt(colSums(e1^2)), sqrt(colSums(e2^2)), sqrt(colSums(e3^2)))
  expect_equal(vcg_max_edge_length(sphere), expected, tolerance = 1e-5)
})

# ---- vcg_subdivide_max_edge_length -------------------------------------

test_that("vcg_subdivide_max_edge_length satisfies the threshold", {
  sphere    <- vcg_sphere()
  cur_max   <- vcg_max_edge_length(sphere)
  threshold <- cur_max * 0.4
  result    <- vcg_subdivide_max_edge_length(sphere, max_edge_len = threshold)
  expect_lte(vcg_max_edge_length(result), threshold * 1.001)
})

test_that("vcg_subdivide_max_edge_length is a no-op when threshold exceeds current max", {
  sphere  <- vcg_sphere()
  cur_max <- vcg_max_edge_length(sphere)
  result  <- vcg_subdivide_max_edge_length(sphere, max_edge_len = cur_max * 2)
  expect_equal(sphere$it, result$it)
  expect_equal(sphere$vb, result$vb)
})

test_that("vcg_subdivide_max_edge_length returns a valid mesh3d", {
  sphere <- vcg_sphere()
  result <- vcg_subdivide_max_edge_length(sphere, max_edge_len = 0.1)
  expect_true(inherits(result, "mesh3d"))
  expect_equal(nrow(result$vb), 4L)
  expect_equal(nrow(result$it), 3L)
  expect_gt(ncol(result$it), ncol(sphere$it))
})

test_that("vcg_subdivide_max_edge_length explicit max_iter=1 may leave some long edges", {
  sphere    <- vcg_sphere()
  cur_max   <- vcg_max_edge_length(sphere)
  threshold <- cur_max * 0.3
  # one pass may not fully satisfy the threshold
  result1   <- vcg_subdivide_max_edge_length(sphere, max_edge_len = threshold, max_iter = 1L)
  result_full <- vcg_subdivide_max_edge_length(sphere, max_edge_len = threshold)
  # full should have at least as many faces as single-pass
  expect_gte(ncol(result_full$it), ncol(result1$it))
})

# ---- vcg_mesh_patch ----------------------------------------------------
# waypoints are now an n x 3 coordinate matrix (x, y, z per row).
# Construct by transposing columns of sphere$vb.

test_that("vcg_mesh_patch returns a length-2 list of mesh3d objects", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.25), as.integer(n * 0.5), as.integer(n * 0.75))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])
  result <- vcg_mesh_patch(sphere, waypoints = wp)
  expect_length(result, 2L)
  expect_true(inherits(result[[1]], "mesh3d"))
  expect_true(inherits(result[[2]], "mesh3d"))
})

test_that("vcg_mesh_patch face counts partition the original mesh", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.25), as.integer(n * 0.5), as.integer(n * 0.75))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])
  result <- vcg_mesh_patch(sphere, waypoints = wp)
  expect_equal(ncol(result[[1]]$it) + ncol(result[[2]]$it), ncol(sphere$it))
})

test_that("vcg_mesh_patch orig_vertex maps correctly to original vertices", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.25), as.integer(n * 0.5), as.integer(n * 0.75))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])
  result <- vcg_mesh_patch(sphere, waypoints = wp)
  p1 <- result[[1]]
  p2 <- result[[2]]
  expect_equal(p1$vb[1:3, ], sphere$vb[1:3, p1$orig_vertex], tolerance = 1e-6)
  expect_equal(p2$vb[1:3, ], sphere$vb[1:3, p2$orig_vertex], tolerance = 1e-6)
})

test_that("vcg_mesh_patch seed_vertex selects the correct side", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.25), as.integer(n * 0.5), as.integer(n * 0.75))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])

  result_auto <- vcg_mesh_patch(sphere, waypoints = wp)
  p1_auto <- result_auto[[1]]

  # A vertex from p1_auto should stay in the first patch
  inside_v <- p1_auto$orig_vertex[ceiling(length(p1_auto$orig_vertex) / 2L)]
  result_seed <- vcg_mesh_patch(sphere, waypoints = wp, seed_vertex = inside_v)
  p1_seed <- result_seed[[1]]

  expect_setequal(p1_auto$orig_vertex, p1_seed$orig_vertex)
})

test_that("vcg_mesh_patch p1 centroid is closer to mean waypoint than p2", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.25), as.integer(n * 0.5), as.integer(n * 0.75))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])
  result  <- vcg_mesh_patch(sphere, waypoints = wp)
  p1 <- result[[1]]
  p2 <- result[[2]]
  mean_wp <- colMeans(wp)   # mean of the 4 waypoint coordinates
  d1 <- sqrt(sum((rowMeans(p1$vb[1:3, , drop = FALSE]) - mean_wp)^2))
  d2 <- sqrt(sum((rowMeans(p2$vb[1:3, , drop = FALSE]) - mean_wp)^2))
  expect_lte(d1, d2)
})

test_that("vcg_mesh_patch errors with fewer than 3 waypoints", {
  sphere <- vcg_sphere()
  wp2 <- t(sphere$vb[1:3, 1:2, drop = FALSE])   # 2-row matrix
  expect_error(
    vcg_mesh_patch(sphere, waypoints = wp2),
    "at least 3"
  )
})

test_that("vcg_mesh_patch errors with wrong waypoints column count", {
  sphere <- vcg_sphere()
  expect_error(
    vcg_mesh_patch(sphere, waypoints = matrix(1, nrow = 4, ncol = 2)),
    "3 columns"
  )
})

test_that("vcg_mesh_patch works with 3 waypoints", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.33), as.integer(n * 0.66))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])
  result <- vcg_mesh_patch(sphere, waypoints = wp)
  expect_equal(ncol(result[[1]]$it) + ncol(result[[2]]$it), ncol(sphere$it))
})

test_that("vcg_mesh_patch max_edge_length refines the output patches", {
  sphere <- vcg_sphere()
  n      <- ncol(sphere$vb)
  wp_idx <- c(1L, as.integer(n * 0.25), as.integer(n * 0.5), as.integer(n * 0.75))
  wp     <- t(sphere$vb[1:3, wp_idx, drop = FALSE])
  mel    <- vcg_max_edge_length(sphere) * 0.5

  result <- vcg_mesh_patch(sphere, waypoints = wp, max_edge_length = mel)
  p1 <- result[[1]]
  p2 <- result[[2]]

  # Both patches should respect the max_edge_length
  expect_lte(vcg_max_edge_length(p1), mel * 1.001)
  expect_lte(vcg_max_edge_length(p2), mel * 1.001)

  # More faces than without refinement
  result_plain <- vcg_mesh_patch(sphere, waypoints = wp)
  expect_gt(ncol(p1$it) + ncol(p2$it),
            ncol(result_plain[[1]]$it) + ncol(result_plain[[2]]$it))

  # orig_vertex values are valid indices
  expect_true(all(p1$orig_vertex >= 1L))
  expect_true(all(p2$orig_vertex >= 1L))
})

test_that("trilinear resampling: values and nearest-neighbor default unchanged", {
  set.seed(1)
  nd <- c(12L, 11L, 10L)
  x <- array(rnorm(prod(nd)), nd)
  v2r <- diag(4)
  v2r[1:3, 4] <- c(-3, -2, -1)

  # nearest-neighbor (default) must equal the explicit nearest call
  a <- resample3D(nd, x, t(v2r), t(v2r), na = 0)
  b <- resample3D(nd, x, t(v2r), t(v2r), na = 0, interpolation = 0L)
  expect_identical(a[[1]], b[[1]])
  # identity transform reproduces the input
  expect_equal(as.vector(a[[1]]), as.vector(x), tolerance = 1e-9)

  # trilinear at integer coordinates equals the sampled voxel values
  tl <- resample3D(nd, x, t(v2r), t(v2r), na = 0, interpolation = 1L)
  expect_equal(as.vector(tl[[1]]), as.vector(x), tolerance = 1e-9)

  # trilinear midpoint between two voxels is their average
  half <- v2r
  half[1, 4] <- half[1, 4] + 0.5            # shift sampling by half a voxel in x
  mid <- resample3D(nd, x, t(half), t(v2r), na = NA_real_, interpolation = 1L)[[1]]
  expect_equal(mid[3, 4, 5], (x[3, 4, 5] + x[4, 4, 5]) / 2, tolerance = 1e-9)
})

# helper: smooth multi-blob phantom
phantom <- function(nd, seed = 7, nblob = 6) {
  set.seed(seed)
  g <- expand.grid(x = 0:(nd[1] - 1),
                   y = 0:(nd[2] - 1),
                   z = 0:(nd[3] - 1))
  val <- numeric(nrow(g))
  for (k in seq_len(nblob)) {
    c0 <- runif(3, nd / 3, 2 * nd / 3)
    val <- val + runif(1, .5, 1) *
      exp(-((g$x - c0[1])^2 + (g$y - c0[2])^2 + (g$z - c0[3])^2) / (2 * runif(1, 3, 6)^2))
  }
  array(val, nd)
}

recenter <- function(M, nd, v2r) {
  ctr <- v2r %*% c((nd - 1) / 2, 1)
  Tc <- diag(4)
  Tc[1:3, 4] <- ctr[1:3]
  Tc %*% M %*% solve(Tc)
}

rotz <- function(deg) {
  th <- deg * pi / 180
  R <- diag(4)
  R[1:2, 1:2] <- matrix(c(cos(th), sin(th), -sin(th), cos(th)), 2)
  R
}

test_that("rigid registration recovers a known translation (all metrics)", {
  nd <- c(40L, 40L, 40L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -20
  target <- phantom(nd)
  shift <- c(3, -2, 1)
  Ttrue <- diag(4)
  Ttrue[1:3, 4] <- shift
  source <- apply_transform3d(
    target,
    v2r,
    solve(Ttrue),
    reference_dim = nd,
    reference_vox2ras = v2r,
    interpolation = "trilinear"
  )

  for (m in c("meansquares", "cc", "mattes")) {
    res <- register_volume3d(
      source, target, v2r, v2r, type = "rigid", metric = m,
      shrink_factors = c(4, 2, 1), smoothing_sigmas = c(2, 1, 0),
      iterations = c(200, 150, 80), sampling_rate = 0.4, verbose = FALSE)
    expect_equal(res$transform[1:3, 4], shift, tolerance = 0.25,
                 info = paste("metric:", m))
    expect_lt(max(abs(res$transform[1:3, 1:3] - diag(3))), 0.03)
  }
})

test_that("rigid registration recovers a known rotation", {
  skip_on_cran()
  nd <- c(48L, 48L, 48L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -24
  target <- phantom(nd)
  Ttrue <- recenter(rotz(12), nd, v2r)
  Ttrue[1:3, 4] <- Ttrue[1:3, 4] + c(2, -3, 1)
  source <- apply_transform3d(target,
                              v2r,
                              solve(Ttrue),
                              reference_dim = nd,
                              reference_vox2ras = v2r)

  res <- register_volume3d(
    source, target, v2r, v2r, type = "rigid", metric = "meansquares",
    shrink_factors = c(4, 2, 1), smoothing_sigmas = c(2, 1, 0),
    iterations = c(300, 200, 100), sampling_rate = 0.4, verbose = FALSE)
  expect_lt(max(abs(res$transform - Ttrue)), 0.1)
})

test_that("affine registration recovers scale + shear + rotation", {
  skip_on_cran()
  nd <- c(48L, 48L, 48L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -24
  target <- phantom(nd)
  A <- diag(4)
  A[1, 1] <- 1.08
  A[2, 2] <- 0.93
  A[1, 2] <- 0.06
  A[2, 3] <- -0.04
  Ttrue <- recenter(A %*% rotz(5), nd, v2r)
  Ttrue[1:3, 4] <- Ttrue[1:3, 4] + c(1, 1, -1)
  source <- apply_transform3d(target,
                              v2r,
                              solve(Ttrue),
                              reference_dim = nd,
                              reference_vox2ras = v2r)

  res <- register_volume3d(
    source, target, v2r, v2r, type = "affine", metric = "cc",
    shrink_factors = c(4, 2, 1), smoothing_sigmas = c(2, 1, 0),
    iterations = c(300, 200, 120), sampling_rate = 0.4, verbose = FALSE)
  expect_lt(max(abs(res$transform - Ttrue)), 0.06)
})

test_that("registration handles different vox2ras (spacing) on source and target", {
  skip_on_cran()
  nd_t <- c(40L, 40L, 40L)
  v2r_t <- diag(4)
  v2r_t[1:3, 4] <- -20
  target <- phantom(nd_t)

  # source on a finer grid (0.5 mm) with an origin shift, holding the same anatomy
  nd_s <- c(80L, 80L, 80L)
  v2r_s <- diag(c(0.5, 0.5, 0.5, 1))
  v2r_s[1:3, 4] <- -20
  source <- apply_transform3d(target,
                              v2r_t,
                              diag(4),
                              reference_dim = nd_s,
                              reference_vox2ras = v2r_s)

  shift <- c(2.5, -1.5, 1)
  Ttrue <- diag(4)
  Ttrue[1:3, 4] <- shift
  source <- apply_transform3d(source, v2r_s, solve(Ttrue),
                              reference_dim = nd_s, reference_vox2ras = v2r_s)

  res <- register_volume3d(
    source, target, v2r_s, v2r_t, type = "rigid", metric = "cc",
    shrink_factors = c(4, 2, 1), smoothing_sigmas = c(2, 1, 0),
    iterations = c(200, 150, 100), sampling_rate = 0.5, verbose = FALSE)
  expect_equal(res$transform[1:3, 4], shift, tolerance = 0.4)

  # warped source lives on the target grid and correlates with the target
  expect_identical(dim(res$image), dim(target))
  ok <- target > 0.05
  expect_gt(cor(res$image[ok], target[ok]), 0.9)
})

test_that("SyN recovers a known smooth deformation", {
  skip_on_cran()
  nd <- c(32L, 32L, 32L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -16
  target <- phantom(nd, seed = 3, nblob = 8)

  # vectorized trilinear sampler (0-based continuous coords)
  tril <- function(vol, cx, cy, cz) {
    d <- dim(vol)
    inb <- cx >= 0 & cy >= 0 & cz >= 0 & cx <= d[1] - 1 & cy <= d[2] - 1 & cz <= d[3] - 1
    cx <- pmin(pmax(cx, 0), d[1] - 1)
    cy <- pmin(pmax(cy, 0), d[2] - 1)
    cz <- pmin(pmax(cz, 0), d[3] - 1)
    x0 <- floor(cx)
    y0 <- floor(cy)
    z0 <- floor(cz)
    x1 <- pmin(x0 + 1, d[1] - 1)
    y1 <- pmin(y0 + 1, d[2] - 1)
    z1 <- pmin(z0 + 1, d[3] - 1)
    fx <- cx - x0
    fy <- cy - y0
    fz <- cz - z0
    idx <- function(x, y, z) {
      x + d[1] * (y + d[2] * z) + 1
    }
    v <- function(x, y, z) {
      vol[idx(x, y, z)]
    }
    c00 <- v(x0, y0, z0) * (1 - fx) + v(x1, y0, z0) * fx
    c10 <- v(x0, y1, z0) * (1 - fx) + v(x1, y1, z0) * fx
    c01 <- v(x0, y0, z1) * (1 - fx) + v(x1, y0, z1) * fx
    c11 <- v(x0, y1, z1) * (1 - fx) + v(x1, y1, z1) * fx
    c0 <- c00 * (1 - fy) + c10 * fy
    c1 <- c01 * (1 - fy) + c11 * fy
    out <- c0 * (1 - fz) + c1 * fz
    out[!inb] <- 0
    out
  }

  g <- expand.grid(x = 0:(nd[1] - 1),
                   y = 0:(nd[2] - 1),
                   z = 0:(nd[3] - 1))
  amp <- 1.4
  L <- 16
  moving <- array(tril(
    target,
    g$x + amp * sin(2 * pi * g$y / L),
    g$y + amp * sin(2 * pi * g$z / L),
    g$z + amp * sin(2 * pi * g$x / L)
  ), nd)

  base_cor <- cor(as.vector(moving), as.vector(target))
  res <- register_volume3d(
    moving, target, v2r, v2r, type = "syn", metric = "cc",
    shrink_factors = c(4, 2, 1), smoothing_sigmas = c(2, 1, 0),
    iterations = c(200, 150, 80), sampling_rate = 1,
    syn_iterations = c(50, 40, 20), syn_sigma = 3, verbose = FALSE)

  expect_true(all(is.finite(res$forward_field)))
  expect_identical(dim(res$forward_field), c(nd, 3L))
  expect_gt(cor(as.vector(res$image), as.vector(target)), base_cor + 0.01)
  expect_lt(mean((res$image - target)^2), mean((moving - target)^2) / 2)
})

test_that("metric improves over the identity transform", {
  nd <- c(40L, 40L, 40L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -20
  target <- phantom(nd)
  Ttrue <- recenter(rotz(8), nd, v2r)
  Ttrue[1:3, 4] <- Ttrue[1:3, 4] + c(2, -2, 1)
  source <- apply_transform3d(target, v2r, solve(Ttrue),
                              reference_dim = nd, reference_vox2ras = v2r)

  res <- register_volume3d(
    source, target, v2r, v2r, type = "rigid", metric = "mattes",
    shrink_factors = c(4, 2, 1), smoothing_sigmas = c(2, 1, 0),
    iterations = c(200, 150, 80), sampling_rate = 0.4, verbose = FALSE)
  # metric is a cost (negative MI); the final value should be below the first
  expect_lt(tail(res$metric_trace, 1), res$metric_trace[1])
})

test_that("masks: all-ones equals unmasked, partial mask runs, bad dims error", {
  nd <- c(32L, 32L, 32L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -16
  target <- phantom(nd, seed = 5)
  shift <- c(2, -1, 1.5)
  Ttrue <- diag(4)
  Ttrue[1:3, 4] <- shift
  source <- apply_transform3d(target, v2r, solve(Ttrue),
                              reference_dim = nd, reference_vox2ras = v2r)

  lin <- function(tm = NULL) {
    register_volume3d(
      source,
      target,
      v2r,
      v2r,
      target_mask = tm,
      type = "rigid",
      metric = "cc",
      shrink_factors = c(4, 2),
      smoothing_sigmas = c(2, 1),
      iterations = c(100, 60),
      sampling_rate = 1,
      verbose = FALSE
    )$transform
  }

  # an all-ones mask includes every voxel, so it must reproduce the unmasked
  # result bit-for-bit (the sampler skips nothing and the RNG stream is the same)
  expect_identical(lin(array(1, nd)), lin())

  # a partial mask still runs and recovers roughly the right translation
  m <- array(0, nd)
  m[6:26, 6:26, 6:26] <- 1
  expect_equal(lin(m)[1:3, 4], shift, tolerance = 0.5)

  # SyN with a mask (target + source) yields a finite field of the right shape
  syn <- register_volume3d(
    source, target, v2r, v2r, target_mask = m, source_mask = m, type = "syn",
    shrink_factors = c(2, 1), smoothing_sigmas = c(1, 0), iterations = c(50, 30),
    syn_iterations = c(10, 5), metric = "cc", verbose = FALSE)
  expect_true(all(is.finite(syn$forward_field)))
  expect_identical(dim(syn$forward_field), c(nd, 3L))

  # a mask must match its image dimensions
  expect_error(
    register_volume3d(source, target, v2r, v2r, target_mask = array(1, c(8, 8, 8)),
                      type = "rigid", verbose = FALSE),
    "must match its image dimensions")
})

test_that("cortical landmark term pulls the SyN warp onto correspondences", {
  nd <- c(32L, 32L, 32L)
  v2r <- diag(4)
  v2r[1:3, 4] <- -16
  set.seed(11)
  img <- array(rnorm(prod(nd)), nd)   # identical fixed = moving

  # corresponding points at integer voxels; target = fixed + a known +2.5 mm in x
  set.seed(12)
  ijk <- unique(cbind(sample(8:24, 90, TRUE), sample(8:24, 90, TRUE), sample(8:24, 90, TRUE)))
  nP <- nrow(ijk)
  ras_fix <- t(v2r %*% rbind(t(ijk), 1))[, 1:3]
  off <- 2.5
  ras_mov <- ras_fix
  ras_mov[, 1] <- ras_mov[, 1] + off

  # points_weight = 0 disables the term -> bit-identical to supplying no points
  pars <- list(type = "syn_only", init_transform = diag(4), metric = "cc",
               shrink_factors = c(2, 1), smoothing_sigmas = c(1, 0),
               syn_iterations = c(8, 8), verbose = FALSE)
  a <- do.call(register_volume3d, c(list(img, img, v2r, v2r), pars))$forward_field
  b <- do.call(register_volume3d, c(list(img, img, v2r, v2r,
        target_points = ras_fix, source_points = ras_mov, points_weight = 0), pars))$forward_field
  expect_identical(a, b)

  # identical images => the image force is ~0, so the landmarks drive the field
  res <- register_volume3d(
    img, img, v2r, v2r, type = "syn_only", init_transform = diag(4),
    target_points = ras_fix, source_points = ras_mov, points_weight = 2,
    shrink_factors = c(2, 1), smoothing_sigmas = c(1, 0),
    syn_iterations = c(40, 40), syn_sigma = 2, metric = "cc", verbose = FALSE)
  ff <- res$forward_field
  disp <- t(vapply(seq_len(nP), function(p) {
    ff[ijk[p, 1] + 1, ijk[p, 2] + 1, ijk[p, 3] + 1, ]
  }, numeric(3)))
  # the recovered displacement at the points should approach the known offset
  expect_lt(mean(sqrt(rowSums((
    matrix(c(off, 0, 0), nP, 3, byrow = TRUE) - disp
  )^2))), 0.8)
  expect_gt(mean(disp[, 1]), 0.7 * off)

  # points must be supplied together and with matching row counts
  expect_error(
    register_volume3d(img, img, v2r, v2r, type = "syn",
                      target_points = ras_fix, verbose = FALSE),
    "supplied together")
  expect_error(
    register_volume3d(img, img, v2r, v2r, type = "syn", target_points = ras_fix,
                      source_points = ras_mov[1:4, ], verbose = FALSE),
    "same number of rows")
})

test_that("C++ signal filter", {
  initialize_filter <- function(b, a, x) {
    # make sure a, b share the same order
    na <- length(a)
    nb <- length(b)

    if (na > nb) {
      b <- c(b, rep(0, na - nb))
      n <- na
    } else {
      a <- c(a, rep(0, nb - na))
      n <- nb
    }

    # length of edge transients
    nf <- max(1, 3 * (n - 1))

    # compute the initial condition if n > 1
    if (n > 1) {
      z1 <- diag(1, n - 1) - cbind(-a[-1], rbind(diag(1, n - 2), 0))
      z2 <- b[-1] - b[1] * a[-1]
      z <- solve(z1, z2)
    } else {
      z <- numeric(0)
    }
    list(
      a = a,
      b = b,
      z = z,
      nfilt = n,
      nfact = nf
    )
  }
  myFilter <- function(b, a, x, z) {
    # make sure a, b share the same order
    na <- length(a)
    nb <- length(b)
    if (na > nb) {
      b <- c(b, rep(0, na - nb))
      n <- na
    } else {
      a <- c(a, rep(0, nb - na))
      n <- nb
    }
    b <- b / a[1]
    a <- a / a[1]
    y <- rep(0, length(x))
    if (missing(z)) {
      z <- rep(0, n - 1)
    }

    for (m in seq_along(y)) {
      xm <- x[m]
      y[m] <- b[1] * xm + z[1]
      ym <- y[m]
      for (i in 2:(n - 1)) {
        z[i - 1] <- b[i] * xm + z[i] - a[i] * ym
      }
      z[n - 1] <- b[n] * xm - a[n] * ym
    }
    list(y, z)
  }

  bf <- gsignal::butter(10, c(0.15, 0.3))
  t <- seq(0, 1, by = 0.005)
  x <- as.double(sin(2 * pi * t * 2.3)) + rnorm(length(t), mean = t)
  b <- as.double(bf$b)
  a <- as.double(bf$a)
  nx <- length(x)
  init <- initialize_filter(b, a, x)
  nfact <- init$nfact
  z <- as.double(init$z)

  expected <- myFilter(b, a, x, z)
  my_result <- ravetools:::cpp_filter(b, a, x, z)

  testthat::expect_lt(max(abs((expected[[1]] - my_result[[1]]) / (expected[[1]] + my_result[[1]] + 1)
  )), 1e-3)
  testthat::expect_lt(max(abs((expected[[2]] - my_result[[2]]) / (expected[[2]] + my_result[[2]] + 1)
  )), 1e-3)

})


test_that("multi-signal fftfilt", {
  # sanity-check
  x <- array(sample_signal(5000) + 1:5000, c(1000, 5))
  filter <- design_filter_fir(
    sample_rate = 2000,
    high_pass_freq = 0.5,
    high_pass_trans_freq = 0.4,
    data_size = 1000
  )
  b <- filter$b
  expected <- apply(x, 2, function(xi) {
    re <- fftfilt(b, xi, legacy = TRUE)
    expect_vector(re)
    expect_length(re, length(xi))
    re
  })
  expect_equal(fftfilt(b, x, legacy = TRUE), expected)
  expect_equal(fftfilt(b, x, legacy = FALSE), expected)


  # scaled FIR
  filter$a <- 2
  expected <- apply(x, 2, function(xi) {
    filtfilt(filter$b, filter$a, xi)
  })


  actual <- filtfilt(filter$b, filter$a, x)

  testthat::expect_equal(actual, expected)


  # # speed check
  # microbenchmark::microbenchmark(
  #   expected = apply(x, 2, function(xi) {
  #     filtfilt(filter$b, filter$a, xi)
  #   }),
  #   actual = filtfilt(filter$b, filter$a, x),
  #   times = 10
  # )
})

test_that("carla_zmin_all matches the R reference", {
  # Pure-R reference for the C++ shared-bootstrap evaluation. `x_ord` is
  # time x trials x channels (channels in ranking order). For each column of
  # `ind`: trial-average every channel once (-> Useg, time x channel); then for
  # every subset size ii = 2..n_good re-reference the first ii channels by their
  # channel mean, correlate (over time) each unreferenced channel against every
  # re-referenced one, Fisher-z, and record the mean z of the most globally
  # anti-correlated channel. Row 1 (subset size 1) is structurally NA.
  ref_zmin_all <- function(x_ord, ind) {
    d <- dim(x_ord)
    n_t <- d[1]
    n_tr <- d[2]
    n_good <- d[3]
    nboot <- ncol(ind)
    out <- matrix(NA_real_, n_good, nboot)
    for (b in seq_len(nboot)) {
      w <- tabulate(ind[, b], n_tr) / nrow(ind)
      Useg <- vapply(seq_len(n_good), function(c) {
        matrix(x_ord[, , c], n_t, n_tr) %*% w
      }, numeric(n_t))                          # n_t x n_good (time x channel)
      for (ii in seq.int(2L, n_good)) {
        U    <- Useg[, seq_len(ii), drop = FALSE]  # n_t x ii
        car  <- rowMeans(U)                        # length n_t (candidate CAR)
        Uref <- U - car                            # re-reference
        r <- stats::cor(U, Uref)                   # ii x ii
        diag(r) <- NA_real_
        z <- atanh(r)
        kk <- which.min(rowMeans(z, na.rm = TRUE))
        out[ii, b] <- mean(z[kk, ], na.rm = TRUE)
      }
    }
    out
  }

  set.seed(101)
  # A few shapes, read as (n_t, n_tr, n_good); smallest subset size ii = 2.
  for (cfg in list(c(30, 8, 4), c(60, 20, 6), c(120, 40, 12), c(200, 50, 20))) {
    n_t <- cfg[1]
    n_tr <- cfg[2]
    n_good <- cfg[3]
    nboot <- 20L
    x_ord <- array(stats::rnorm(n_t * n_tr * n_good), dim = c(n_t, n_tr, n_good))
    ind <- matrix(sample.int(n_tr, n_tr * nboot, replace = TRUE),
                  nrow = n_tr, ncol = nboot)
    got <- carla_zmin_all(x_ord, ind)
    expect_equal(dim(got), c(n_good, nboot))
    expect_true(all(is.na(got[1, ])))
    expect_equal(got, ref_zmin_all(x_ord, ind), tolerance = 1e-8)
  }
})

test_that("carla_zmin_all is deterministic and thread-count invariant", {
  set.seed(202)
  # time x trials x channels
  x_ord <- array(stats::rnorm(150 * 30 * 20), dim = c(150, 30, 20))
  ind <- matrix(sample.int(30, 30 * 40, replace = TRUE), nrow = 30, ncol = 40)

  ravetools_threads(n_threads = 1L)
  one <- carla_zmin_all(x_ord, ind)
  ravetools_threads(n_threads = 4L)
  four <- carla_zmin_all(x_ord, ind)
  expect_equal(one, four, tolerance = 1e-12)
})

test_that("carla() runs end-to-end and selects the non-responsive channels", {
  set.seed(303)
  nchan <- 16L
  n_t <- 200L
  ntrial <- 30L
  # time x trials x channels
  x <- array(stats::rnorm(n_t * ntrial * nchan), dim = c(n_t, ntrial, nchan))
  # Channels 1:3 carry a phase-locked evoked response; the rest are noise.
  ep <- 6 * exp(-seq(0, 2, length.out = n_t)) * sin(2 * pi * 5 * seq(0, 0.2, length.out = n_t))
  for (ch in 1:3) for (k in seq_len(ntrial)) x[, k, ch] <- x[, k, ch] + ep

  fit <- carla(x, nboot = 50L)
  expect_type(fit, "list")
  expect_true(all(c("channels", "order", "vars", "n_optimum", "zmin_mean") %in% names(fit)))
  # zmin_mean: n_good x nboot, first row structurally NA, rest finite.
  expect_equal(dim(fit$zmin_mean), c(length(fit$order), 50L))
  expect_true(all(is.na(fit$zmin_mean[1, ])))
  expect_true(all(is.finite(fit$zmin_mean[-1, ])))
  # Responsive channels should NOT be picked into the common-average reference.
  expect_false(any(1:3 %in% fit$channels))
})

test_that("carla() determinism under a fixed seed", {
  set.seed(404)
  # time x trials x channels
  x <- array(stats::rnorm(100 * 20 * 12), dim = c(100, 20, 12))
  set.seed(1)
  a <- carla(x, nboot = 30L, sensitive = TRUE)
  set.seed(1)
  b <- carla(x, nboot = 30L, sensitive = TRUE)
  expect_equal(a$zmin_mean, b$zmin_mean)
  expect_equal(a$channels, b$channels)
  expect_equal(a$n_optimum, b$n_optimum)
})

test_that("carla() handles single-trial and nboot = 1 inputs", {
  set.seed(505)
  # time x trials x channels
  x <- array(stats::rnorm(80 * 15 * 10), dim = c(80, 15, 10))

  # Single trial (matrix input, time x channels) -> variance ranking, single-pass.
  fit1 <- carla(x[, 1, ])
  expect_equal(length(fit1$zmin_mean), length(fit1$order))
  expect_true(fit1$n_optimum >= 1L)

  # Multiple trials but nboot = 1 -> single-pass, no bootstrap.
  fitb1 <- carla(x, nboot = 1L)
  expect_equal(ncol(matrix(fitb1$zmin_mean, ncol = 1L)), 1L)
  expect_true(fitb1$n_optimum >= 1L)
})

test_that("carla() accepts a FileArray and matches the in-memory result", {
  skip_if_not_installed("filearray")

  set.seed(606)
  nchan <- 14L
  n_t <- 120L
  ntrial <- 25L
  x <- array(stats::rnorm(n_t * ntrial * nchan), dim = c(n_t, ntrial, nchan))
  ep <- 5 * exp(-seq(0, 2, length.out = n_t)) * sin(2 * pi * 6 * seq(0, 0.2, length.out = n_t))
  for (ch in 1:3) for (k in seq_len(ntrial)) x[, k, ch] <- x[, k, ch] + ep

  xf <- filearray::as_filearray(x)
  on.exit(xf$delete(force = TRUE), add = TRUE)

  for (vref in c(FALSE, TRUE)) {
    set.seed(7)
    mem <- carla(x, nboot = 40L, virtual_reference = vref)
    set.seed(7)
    disk <- carla(xf, nboot = 40L, virtual_reference = vref)
    expect_equal(disk$channels,  mem$channels)
    expect_equal(disk$n_optimum, mem$n_optimum)
    expect_equal(disk$zmin_mean, mem$zmin_mean, tolerance = 1e-9)
    expect_equal(disk$vars,      mem$vars,      tolerance = 1e-9)
  }
})

test_that("pwelch", {
  rave_pwelch <- function(x,
                          fs,
                          window = 64,
                          noverlap = 8,
                          nfft = 256,
                          col = "black",
                          xlim = NULL,
                          ylim = NULL,
                          main = "Welch periodogram",
                          plot = TRUE,
                          log = "xy",
                          spec_func = stats::spectrum,
                          cex = 1,
                          ...)
  {
    x <- as.vector(x)
    x_len <- length(x)
    nfft <- max(min(nfft, length(x)), window)
    window <- hanning(window)
    window_len <- length(window)
    step <- max(floor(window_len - noverlap + 0.99), 1)
    # print(c(1, x_len - window_len + 1, by = step))
    offset <- seq(1, x_len - window_len + 1, by = step)
    N <- length(offset)
    re <- sapply(seq_len(N), function(i) {
      a <- detrend_naive(x[offset[i] - 1 + seq_len(window_len)])
      a <- fftwtools::fftw_r2c(postpad(a$Y * window, nfft))
      Mod(a)^2
    })
    NN <- ceiling((nfft + 1) / 2)
    freq <- seq(0, fs / 2, length.out = NN)
    spec <- rowMeans(re[seq_len(NN), , drop = FALSE]) / window_len #(window_len/2)^2
    res <- list(freq = freq,
                spec = spec,
                method = "Welch")
    return(invisible(res))
  }


  x <- rnorm(10000)
  a <- pwelch(
    x,
    fs = 100,
    window = 64,
    noverlap = 6,
    nfft = 256,
    plot = FALSE,
    window_family = hanning
  )
  b <- rave_pwelch(
    x,
    fs = 100,
    window = 64,
    noverlap = 6,
    nfft = 256,
    log = ""
  )

  expect_equal(a$freq, b$freq)
  expect_equal(a$spec, b$spec)

  # if(FALSE) {
  #   x <- rnorm(600000)
  #   microbenchmark::microbenchmark(
  #     rave = {
  #       rave_pwelch(x, 2000, log = '')
  #     },
  #     ravetools = {
  #       pwelch(x, 2000, plot = FALSE)
  #     }, times = 20
  #   )
  # }

})

test_that("Vector3", {
  library(testthat)
  x <- array(1:30, c(3, 10)) + 0.3
  v0 <- matrix(x[3:29], nrow = 3)
  v <- new_vector3()
  v$from_array(x, offset = 2)

  expect_equal(v$to_array(), v0)
  expect_equal(v$to_array(3, 3), v0[, 4:6])

  expect_equal(v$get_size(), ncol(v0))

  expect_error(v$resize(0))

  v$resize(6)
  v0 <- v0[, 1:6]
  expect_equal(v$to_array(0, -1), v0)
  expect_equal(v$to_array(3, 3), v0[, 4:6])
  expect_equal(v$get_size(), ncol(v0))

  expect_equal(range(v$set_scalar(0.1)$to_array()), c(0.1, 0.1))

  v1 <- v$set_x(0.2)
  expect_equal(as.vector(v1$to_array(
    n_skip = 0, max_n_elems = 1
  )), c(0.2, 0.1, 0.1))
  v$set_y(0.3)
  v$set_z(0.4)
  expect_equal(as.vector(v1$to_array(
    n_skip = 3, max_n_elems = 1
  )), c(0.2, 0.3, 0.4))

  v$from_array(v0)
  expect_equal(as.vector(v1$to_array(
    n_skip = 3, max_n_elems = 1
  )), v$get_item(i = 4))

  v1 <- new_vector3()
  v1$copy(v)

  expect_equal(v1$to_array(), v0)

  v1$copy(v1)
  expect_equal(v1$to_array(), v0)

  expect_equal(v$get_x(4), v0[1, 4])
  expect_equal(v$get_y(5), v0[2, 5])
  expect_equal(v$get_z(6), v0[3, 6])
  expect_equal(v$get_item(3), v0[, 3])

  expect_equal(v1$add(v)[], v0 * 2)
  expect_equal(v1$add(v1)[], v0 * 4)

  v1$copy(v)
  expect_equal(v1$add_scalar(0.1)[], v0 + 0.1)
  expect_error(v1$add_scalar(c(1, 2)))

  s <- seq_len(ncol(v)) / 10
  ss <- t(replicate(3, seq_len(ncol(v)) / 10))
  expect_equal(v1$add_scalar(s)[], v0 + 0.1 + ss)

  expect_equal(v1$add_vectors(v, v)[], v0 * 2)
  v2 <- new_vector3()$set(10.2, 3.8, 4.6)
  expect_error(v1$add_vectors(v, v2))
  v2$copy(v)$add_scalar(0.1)
  expect_equal(v1$copy(v)$add_vectors(v2, v1)[], v0 * 2 + 0.1)

  expect_equal(v1$copy(v)$add_scaled(v, 0.4)[], v0 * 1.4)

  expect_equal(v1$sub(v)[], v0 * 0.4)
  expect_equal(v1$copy(v)$sub_scalar(0.1)[], v0 - 0.1)
  expect_error(v1$sub_scalar(c(1, 2)))
  expect_equal(v1$sub_scalar(s)[], v0 - 0.1 - ss)

  expect_equal(v1$copy(v)$multiply_scalar(1.5)$sub_vectors(v1, v)[],
               v0 * 0.5)
  expect_equal(v1$copy(v)$multiply_scalar(1.5)$sub_vectors(v, v1)[],
               v0 * -0.5)
  expect_equal(v1$copy(v)$multiply(v)[], v0^2)
  expect_equal(v1$copy(v)$multiply_scalar(s)[], v0 * ss)
  expect_error(v1$multiply_scalar(c(1, 2)))
  expect_equal(v1$multiply_vectors(v, v)[], v0^2)
  expect_equal(v1$copy(v)$add_scalar(0.1)$multiply_vectors(v, v1)[],
               v0 * (v0 + 0.1))
  expect_equal(v1$copy(v)$add_scalar(0.1)$multiply_vectors(v1, v)[],
               v0 * (v0 + 0.1))

  expect_equal(v1$copy(v)$multiply_scalar(s)$divide(v)[], ss)
  expect_equal(v1$copy(v)$divide_scalar(s)[], v0 / ss)

  v2 <- new_vector3()$set(5.2, 3.8, 4.6)
  v00 <- v0
  v00[1, v00[1, ] > v2$get_x(1)] <- v2$get_x(1)
  v00[2, v00[2, ] > v2$get_y(1)] <- v2$get_y(1)
  v00[3, v00[3, ] > v2$get_z(1)] <- v2$get_z(1)
  expect_equal(v1$copy(v)$min(v2)[], v00)
  expect_equal(v1$max(v)[], v0)

  v00 <- v0
  v00[1, v00[1, ] < v2$get_x(1)] <- v2$get_x(1)
  v00[2, v00[2, ] < v2$get_y(1)] <- v2$get_y(1)
  v00[3, v00[3, ] < v2$get_z(1)] <- v2$get_z(1)
  expect_equal(v1$max(v2)[], v00)
  expect_equal(v1$min(v)[], v0)

  v2$set(10.2, 3.8, 4.6)
  v3 <- new_vector3()$set(5.2, 13.8, 14.6)
  v00 <- v0
  v00[1, v00[1, ] > v2$get_x(1)] <- v2$get_x(1)
  v00[2, v00[2, ] < v2$get_y(1)] <- v2$get_y(1)
  v00[3, v00[3, ] < v2$get_z(1)] <- v2$get_z(1)
  v00[1, v00[1, ] < v3$get_x(1)] <- v3$get_x(1)
  v00[2, v00[2, ] > v3$get_y(1)] <- v3$get_y(1)
  v00[3, v00[3, ] > v3$get_z(1)] <- v3$get_z(1)
  expect_equal(v1$copy(v)$clamp(v2, v3)[], v00)

  v2$from_array(c(
    0.1,
    -0.1,
    0.4999,
    -0.4999,
    0.5,
    -0.5,
    0.999,
    -0.999,
    1.0001,
    -1.0001,
    1.5,
    -1.5
  ))
  expected_ <- floor(v2[])
  expect_equal(v2$floor()[], expected_)
  v2$from_array(c(
    0.1,
    -0.1,
    0.4999,
    -0.4999,
    0.5,
    -0.5,
    0.999,
    -0.999,
    1.0001,
    -1.0001,
    1.5,
    -1.5
  ))
  expected_ <- ceiling(v2[])
  expect_equal(v2$ceil()[], expected_)
  v2$from_array(c(
    0.1,
    -0.1,
    0.4999,
    -0.4999,
    0.5,
    -0.5,
    0.999,
    -0.999,
    1.0001,
    -1.0001,
    1.5,
    -1.5
  ))
  expected_ <- round(v2[])
  expect_equal(v2$round()[], expected_)
  v2$from_array(c(
    0.1,
    -0.1,
    0.4999,
    -0.4999,
    0.5,
    -0.5,
    0.999,
    -0.999,
    1.0001,
    -1.0001,
    1.5,
    -1.5
  ))
  expected_ <- floor(abs(v2[]))
  expect_equal(abs(v2$round_to_zero()[]), expected_)
  expect_equal(v1$copy(v)$negate()[], -v0)

  expect_equal(v1$copy(v)$dot(v), colSums(v0^2))
  expect_error(v1$dot(v2))
  expect_equal(v1$length_squared(), colSums(v0^2))
  expect_equal(v1$length(), sqrt(colSums(v0^2)))
  expect_equal(v1$length_manhattan(), (colSums(abs(v0))))
  expect_equal(v1$normalize()$length(), rep(1, ncol(v)))
  expect_equal(v1$set_length(s)$length(), s)

  expect_equal(v1$copy(v)$add_scalar(0.5)$negate()$lerp(v, alpha = 0.6)[],
               v0 * 0.2 - 0.2)
  v1$copy(v)$add_scalar(0.5)$negate()
  v00 <- v0 * 0.4 + v1[] * 0.6
  v1$lerp_vectors(v, v1, 0.6)
  expect_equal(v1[], v00)

  cross_prod <- function(x, y) {
    x <- x[seq_len(3)]
    y <- y[seq_len(3)]
    return(c(x[2] * y[3] - x[3] * y[2], x[3] * y[1] - x[1] * y[3], x[1] * y[2] - x[2] * y[1]))
  }

  v2$from_array(c(10.2, 3.8, 4.6, -10.5, 3.6, -4.2))
  v3$from_array(c(5.2, 13.8, 14.6, -5.3, 5.67, 10.55))
  v00 <- cbind(cross_prod(v2[, 1], v3[, 1]), cross_prod(v2[, 2], v3[, 2]))
  expect_equal(v1$cross_vectors(v2, v3)[], v00)
  expect_equal(v1$copy(v2)$cross(v3)[], v00)
  expect_equal(v1$copy(v3)$cross(v2)[], -v00)
  v2$resize(1)
  v3$resize(1)
  expect_equal(v1$copy(v3)$cross(v2)[], -v00[, 1, drop = FALSE])
  expect_equal(v1$copy(v2)$cross_vectors(v3, v1)[], -v00[, 1, drop = FALSE])

  v2 <- v1$copy(v)$clone2()$add(v)
  expect_equal(v2[], v0 * 2)
  expect_equal(v1[], v0)

  # // Vector3& applyMatrix3(const Matrix3& m);
  m <- matrix(c(1:9), nrow = 3, byrow = TRUE)
  expect_equal(v1$copy(v)$apply_matrix3(m)[], m %*% v0)

  # // Vector3& applyMatrix4(const Matrix4& m);
  m <- matrix(c(1:12, 0, 0, 0, 1), nrow = 4, byrow = TRUE)
  expect_equal(v1$copy(v)$apply_matrix4(m)[], (m %*% rbind(v0, 1))[1:3, ])

  # // Vector3& applyQuaternion(const Quaternion& q);
  q <- c(0.1, 0.2, 0.3, 0.4)
  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4))$apply_quaternion(q)[], cbind(c(0.3, 0.6, 0.9), c(0.12, 2.34, 1.2)))

  # // Vector3& transformDirection(const Matrix4& m);
  # JS Code:
  # new THREE.Vector3().set(1,2,3).transformDirection(new THREE.Matrix4().makeRotationX(Math.PI/2)).toArray()
  expect_equal(
    v1$from_array(c(1, 2, 3, 6, 5, 4))$transform_direction(new_matrix4()$make_rotation_x(pi /
                                                                                           2))[],
    cbind(
      c(
        0.2672612419124244,
        -0.8017837257372732,
        0.5345224838248488
      ),
      c(
        0.6837634587578276,
        -0.45584230583855173,
        0.5698028822981898
      )
    )
  )

  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$project_on_vector(new_vector3(1, 1, 1))[],
               cbind(rep(1, 3) * (sum(c(
                 1, 2, 3
               )) / sum(c(
                 1, 1, 1
               ))) , rep(1, 3) * (sum(c(
                 6, 5, 4
               )) / sum(c(
                 1, 1, 1
               ))), c(0, 0, 0)))
  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$project_on_vector(new_vector3(0, 0, 0))[],
               cbind(c(0, 0, 0), c(0, 0, 0), c(0, 0, 0)))

  expect_equal(
    v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$project_on_plane(new_vector3(1, 1, 1))[],
    cbind(
      c(1, 2, 3) - rep(1, 3) * (sum(c(1, 2, 3)) / sum(c(1, 1, 1))),
      c(6, 5, 4) - rep(1, 3) * (sum(c(6, 5, 4)) / sum(c(1, 1, 1))),
      c(0, 0, 0)
    )
  )
  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$project_on_plane(new_vector3(0, 0, 0))[],
               cbind(c(1, 2, 3), c(6, 5, 4), c(0, 0, 0)))

  # // Vector3& reflect(const Vector3& normal);
  ## JS Code:
  # new THREE.Vector3().set(1,2,3).reflect(new THREE.Vector3().set(1,1,1).normalize()).toArray()
  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$reflect(new_vector3(1, 1, 1))[], cbind(c(-3, -2, -1), c(-4, -5, -6), c(0, 0, 0)))

  expect_equal(
    v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$angle_to(new_vector3(1, 1, 1))[],
    acos(
      v2$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$normalize()$dot(new_vector3(1, 1, 1)$normalize())
    )
  )
  expect_error(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$angle_to(new_vector3(1:2, 1:2, 1:2)))
  expect_equal(
    v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$angle_to(new_vector3(1:3, 1:3, 1:3))[],
    acos(
      v2$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$normalize()$dot(new_vector3(1, 1, 1)$normalize())
    )
  )

  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$distance_to(new_vector3(1, 1, 1)),
               c(sqrt(sum((
                 c(1, 2, 3) - c(1, 1, 1)
               )^2)), sqrt(sum((
                 c(6, 5, 4) - c(1, 1, 1)
               )^2)), sqrt(sum((
                 c(0, 0, 0) - c(1, 1, 1)
               )^2))))
  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$distance_to_squared(new_vector3(1, 1, 1)),
               c((sum((
                 c(1, 2, 3) - c(1, 1, 1)
               )^2)), (sum((
                 c(6, 5, 4) - c(1, 1, 1)
               )^2)), (sum((
                 c(0, 0, 0) - c(1, 1, 1)
               )^2))))

  expect_equal(v1$from_array(c(1, 2, 3, 6, 5, 4, 0, 0, 0))$distance_to_manhattan(new_vector3(1, 1, 1)),
               c((sum(abs(
                 c(1, 2, 3) - c(1, 1, 1)
               ))), (sum(abs(
                 c(6, 5, 4) - c(1, 1, 1)
               ))), (sum(abs(
                 c(0, 0, 0) - c(1, 1, 1)
               )))))

  m1 <- new_matrix4()$set(1:16)

  expect_equal(v2$set_from_matrix_position(m1)[], m1[1:3, 4, drop = FALSE])
  expect_equal(v2$set_from_matrix_scale(m1)[, 1], apply(m1[1:3, 1:3], 2, norm, "2"))

  # JS Code
  # new THREE.Vector3().setFromSphericalCoords(1,2,3).toArray()
  expect_equal(v2$set_from_spherical_coords(1, 2, 3)[], cbind(
    c(
      0.12832006020245673,
      -0.4161468365471424,
      -0.9001976297355174
    )
  ))

  # // Vector3& applyAxisAngle(const Vector3& axis, double angle);
  # // Vector3& applyEuler(const Euler& euler);
  # // Vector3& applyNormalMatrix(const Matrix3& m);
  # // Vector3& project(const Camera& camera);
  # // Vector3& unproject(const Camera& camera);
  # // Vector3& setFromSpherical(const Spherical& s);
  # // Vector3& setFromMatrix3Column(const Matrix3& m, unsigned int index);

})

# Helper: gain magnitude at normalized frequency f0 in [0,1]
fir_gain <- function(b, f0) {
  L <- length(b)
  Mod(sum(exp(-1i * 2 * pi * seq(0, L - 1) * (f0 / 2)) * b))
}

# -- Unity-gain tests (self-contained, no MATLAB needed) ----------------------

test_that("fir1: scale=TRUE produces unity gain at the correct reference frequency", {

  # Low-pass  -> DC gain = 1
  b <- fir1(20, 0.3, "low", hamming, scale = TRUE)$b
  expect_equal(fir_gain(b, 0), 1.0, tolerance = 1e-10)

  # High-pass -> Nyquist gain = 1
  b <- fir1(20, 0.3, "high", hamming, scale = TRUE)$b
  expect_equal(fir_gain(b, 1), 1.0, tolerance = 1e-10)

  # Band-pass -> gain at center of passband = 1
  # fir1 internally converts "pass" (2 cutoffs) to "DC-0" and uses mean(w)
  b <- fir1(20, c(0.2, 0.5), "pass", hamming, scale = TRUE)$b
  expect_equal(fir_gain(b, mean(c(0.2, 0.5))), 1.0, tolerance = 1e-10)

  # Band-stop -> DC gain = 1 (MATLAB fir1 convention: fband=TRUE for stop)
  b <- fir1(20, c(0.2, 0.5), "stop", hamming, scale = TRUE)$b
  expect_equal(fir_gain(b, 0), 1.0, tolerance = 1e-10)

  # scale=FALSE must not change the scaling -> gain differs from 1 noticeably
  b_unscaled <- fir1(20, 0.3, "low", hamming, scale = FALSE)$b
  b_scaled   <- fir1(20, 0.3, "low", hamming, scale = TRUE)$b
  expect_false(isTRUE(all.equal(b_unscaled, b_scaled)))
})


test_that("design_filter_fir: scale=TRUE produces unity gain at reference frequency (all methods x all types)", {

  fs   <- 1000
  nyq  <- fs / 2

  for (method in c("kaiser", "firls", "remez")) {

    # -- Low-pass: reference = DC (f0 = 0) --------------------------------
    f <- design_filter_fir(
      sample_rate = fs, low_pass_freq = 100,
      filter_order = 60, scale = TRUE, method = method
    )
    expect_equal(
      fir_gain(f$b, 0), 1.0, tolerance = 1e-9,
      label = sprintf("low-pass %s DC gain", method)
    )

    # -- High-pass: reference = Nyquist (f0 = 1) --------------------------
    f <- design_filter_fir(
      sample_rate = fs, high_pass_freq = 200,
      filter_order = 60, scale = TRUE, method = method
    )
    expect_equal(
      fir_gain(f$b, 1), 1.0, tolerance = 1e-9,
      label = sprintf("high-pass %s Nyquist gain", method)
    )

    # -- Band-pass: reference = mean(kaisprm$Wc) = mean of transition-band midpoints ----
    # high_pass_freq=100 <= low_pass_freq=300 -> ftype="pass"
    f <- design_filter_fir(
      sample_rate = fs, high_pass_freq = 100, low_pass_freq = 300,
      filter_order = 60, scale = TRUE, method = method
    )
    # mean(Wc) = (w_stop[1]+w_pass[1]+w_pass[2]+w_stop[2]) / 4
    f0_pass <- (f$parameters$stopband[[1]] + f$parameters$passband[[1]] +
                f$parameters$passband[[2]] + f$parameters$stopband[[2]]) / 4
    expect_equal(
      fir_gain(f$b, f0_pass), 1.0, tolerance = 1e-9,
      label = sprintf("band-pass %s transition-midpoint gain", method)
    )

    # -- Band-stop: reference = DC (f0 = 0) for both low- and high-freq notch --
    # MATLAB fir1 always scales band-stop at DC (fband=TRUE -> b/sum(b))
    f_lo <- design_filter_fir(
      sample_rate = fs,
      low_pass_freq = 100, low_pass_trans_freq = 10,
      high_pass_freq = 200, high_pass_trans_freq = 10,
      filter_order = 60, scale = TRUE, method = method
    )
    expect_equal(
      fir_gain(f_lo$b, 0), 1.0, tolerance = 1e-9,
      label = sprintf("band-stop (low-freq notch) %s DC gain", method)
    )

    f_hi <- design_filter_fir(
      sample_rate = fs,
      low_pass_freq = 400, low_pass_trans_freq = 10,
      high_pass_freq = 450, high_pass_trans_freq = 10,
      filter_order = 60, scale = TRUE, method = method
    )
    expect_equal(
      fir_gain(f_hi$b, 0), 1.0, tolerance = 1e-9,
      label = sprintf("band-stop (high-freq notch) %s DC gain", method)
    )
  }
})


test_that("design_filter_fir: scale=FALSE does not normalise", {
  f_on  <- design_filter_fir(sample_rate = 1000, low_pass_freq = 100,
                              filter_order = 40, scale = TRUE)
  f_off <- design_filter_fir(sample_rate = 1000, low_pass_freq = 100,
                              filter_order = 40, scale = FALSE)
  # Scaled must have DC = 1; unscaled coefficients must differ
  expect_equal(fir_gain(f_on$b, 0), 1.0, tolerance = 1e-9)
  expect_false(isTRUE(all.equal(f_on$b, f_off$b)))
})


# -- MATLAB coefficient comparison --------------------------------------------
#
# Run the block below in MATLAB, paste the printed vectors back here, then
# remove the skip() calls.
#
# MATLAB code (paste into MATLAB command window or script):
# ------------------------------------------------------------------------------
#   % fir1 defaults: Hamming window + scale=true, identical to ravetools defaults
#   b_low  = fir1(20, 0.3);
#   b_high = fir1(20, 0.3, 'high');
#   b_pass = fir1(20, [0.2 0.5], 'bandpass');
#   b_stop = fir1(20, [0.2 0.5], 'stop');
#
#   fprintf('b_low  <- c('); fprintf('%.15g, ', b_low);  fprintf(')\n');
#   fprintf('b_high <- c('); fprintf('%.15g, ', b_high); fprintf(')\n');
#   fprintf('b_pass <- c('); fprintf('%.15g, ', b_pass); fprintf(')\n');
#   fprintf('b_stop <- c('); fprintf('%.15g, ', b_stop); fprintf(')\n');
# ------------------------------------------------------------------------------

test_that("fir1 low-pass matches MATLAB (hamming window, scale=TRUE)", {
  b_matlab <- c(
    0, 0.00292890317900091, 0.00634234732232293, 0.00378304197175637,
    -0.0123878480564484, -0.0343265728270622, -0.0318598610890541,
    0.0265312166066068, 0.137863164556997, 0.251347679146418,
    0.299555858378927, 0.251347679146418, 0.137863164556997,
    0.0265312166066068, -0.0318598610890541, -0.0343265728270622,
    -0.0123878480564484, 0.00378304197175637, 0.00634234732232293,
    0.00292890317900091, 0
  )

  b_r <- fir1(20, 0.3, "low", hamming, scale = TRUE)$b
  expect_equal(b_r, b_matlab, tolerance = 1e-7)
})


test_that("fir1 high-pass matches MATLAB (hamming window, scale=TRUE)", {
  b_matlab <- c(
    0, -0.00293014380119296, -0.00634503380813584, -0.00378464438929522,
    0.0123930952900616, 0.0343411128461149, 0.0318733562605525,
    -0.0265424546756356, -0.137921560513081, -0.251454144771811,
    0.699259735957564, -0.251454144771811, -0.137921560513081,
    -0.0265424546756356, 0.0318733562605525, 0.0343411128461149,
    0.0123930952900616, -0.00378464438929522, -0.00634503380813584,
    -0.00293014380119296, 0
  )

  b_r <- fir1(20, 0.3, "high", hamming, scale = TRUE)$b
  expect_equal(b_r, b_matlab, tolerance = 1e-7)
})


test_that("fir1 band-pass matches MATLAB (hamming window, scale=TRUE)", {
  b_matlab <- c(
    0, 0.0058661414341214, 0.00647237119566064, -0.00061145861614182,
    0.0126418101787922, 0.0350302987081212, -0.0325130171418959,
    -0.17094564692336, -0.140689484487979, 0.130693550172259,
    0.305697025206915, 0.130693550172259, -0.140689484487979,
    -0.17094564692336, -0.0325130171418959, 0.0350302987081212,
    0.0126418101787922, -0.00061145861614182, 0.00647237119566064,
    0.0058661414341214, 0
  )

  b_r <- fir1(20, c(0.2, 0.5), "pass", hamming, scale = TRUE)$b
  expect_equal(b_r, b_matlab, tolerance = 1e-7)
})


test_that("fir1 band-stop matches MATLAB (hamming window, scale=TRUE)", {
  b_matlab <- c(
    0, -0.00574321408319063, -0.00633674005647684, 0.000598645255139493,
    -0.0123768959666647, -0.0342962247225443, 0.0318316938030531,
    0.167363412201366, 0.137741279807586, -0.127954814312097,
    0.698345716147657, -0.127954814312097, 0.137741279807586,
    0.167363412201366, 0.0318316938030531, -0.0342962247225443,
    -0.0123768959666647, 0.000598645255139493, -0.00633674005647684,
    -0.00574321408319063, 0
  )

  b_r <- fir1(20, c(0.2, 0.5), "stop", hamming, scale = TRUE)$b
  expect_equal(b_r, b_matlab, tolerance = 1e-7)
})


# -- MATLAB comparison for design_filter_fir (kaiser method) ------------------
#
# The kaiser method calls fir1 with cutoffs = kaisprm$Wc and
# window = kaiser(n+1, beta).  Extract these from R, then verify in MATLAB.
#
# Run in R first:
#   f <- design_filter_fir(sample_rate=1000, low_pass_freq=100,
#                           filter_order=40, data_size=5000)
#   cat("n   =", f$parameters$order, "\n")
#   cat("Wc  =", f$parameters$cutoff, "\n")
#   cat("beta=", f$parameters$beta,  "\n")
#
# Then in MATLAB:
#   n = <n from R>; Wc = <Wc from R>; beta = <beta from R>;
#   b_lp_k = fir1(n, Wc, kaiser(n+1, beta), 'scale');
#   fprintf('b_lp_kaiser <- c('); fprintf('%.15g, ', b_lp_k); fprintf(')\n');

test_that("design_filter_fir kaiser low-pass matches MATLAB fir1", {
  # MATLAB run with: n=40; Wc=0.2013379; beta=3.395321;
  # b_lp_k = fir1(n, Wc, kaiser(n+1, beta), 'scale');
  # Note: Wc/beta were printed by R's cat() with ~7 sig figs; tolerance is loose accordingly.
  b_matlab <- c(
    0.000198963867774825, -0.00169241569059748, -0.00396307591421208,
    -0.00534352793206251, -0.0044412511715439, -0.000541195062692525,
    0.00567067137122843, 0.0118991132812478, 0.0148359902757939,
    0.0115083955960501, 0.000926301843132593, -0.0146875568050966,
    -0.0297746373755591, -0.0367515288539371, -0.0285015519433629,
    -0.00123060885531412, 0.0434159169471656, 0.0979264203726522,
    0.150565797367715, 0.188682526152012, 0.202594505059212,
    0.188682526152012, 0.150565797367715, 0.0979264203726522,
    0.0434159169471656, -0.00123060885531412, -0.0285015519433629,
    -0.0367515288539371, -0.0297746373755591, -0.0146875568050966,
    0.000926301843132593, 0.0115083955960501, 0.0148359902757939,
    0.0118991132812478, 0.00567067137122843, -0.000541195062692525,
    -0.0044412511715439, -0.00534352793206251, -0.00396307591421208,
    -0.00169241569059748, 0.000198963867774825
  )

  f <- design_filter_fir(
    sample_rate = 1000, low_pass_freq = 100,
    filter_order = 40, data_size = 5000
  )
  # Loose tolerance: Wc/beta were entered in MATLAB with only ~7 significant figures
  expect_equal(f$b, b_matlab, tolerance = 1e-4)
})

# -- Narrow band-stop (notch narrower than the auto transition heuristic) -----

test_that("design_filter_fir: narrow band-stop with auto transitions succeeds", {
  # Notch is only 1 Hz wide (24-25 Hz) at sr=2000.  With data_size=NA the
  # auto suggested_trans_bandwidth is 0.01 (normalized) = 10 Hz, which is
  # 10x the notch.  The function must silently cap it rather than error.
  # Note: "remez" requires non-degenerate (non-zero-width) stop bands in its
  # internal grid and errors on near-zero transitions; that is a separate
  # gsignal constraint unrelated to this fix.
  for (method in c("kaiser", "firls")) {
    f <- design_filter_fir(
      sample_rate = 2000,
      high_pass_freq = 25, low_pass_freq = 24,
      filter_order = 200, method = method
    )
    expect_true(inherits(f, "ravetools-design_filter_fir"),
                label = sprintf("narrow notch returns filter (%s)", method))
    # Band-stop filters normalise at DC (f0 = 0)
    expect_equal(
      fir_gain(f$b, 0), 1.0, tolerance = 1e-6,
      label = sprintf("narrow notch DC gain = 1 (%s)", method)
    )
  }
})

test_that("design_filter_fir: narrow notch with one explicit transition succeeds", {
  # low_pass_trans_freq explicitly set to 0.3 Hz (< notch width 1 Hz);
  # high_pass_trans_freq is NA and will be capped to the remaining room.
  f <- design_filter_fir(
    sample_rate = 2000,
    high_pass_freq = 25, low_pass_freq = 24,
    low_pass_trans_freq = 0.3,
    filter_order = 200
  )
  expect_true(inherits(f, "ravetools-design_filter_fir"))
  expect_equal(fir_gain(f$b, 0), 1.0, tolerance = 1e-6)
})

test_that("design_filter_fir: user-supplied transitions that overlap throw an error", {
  # Combined explicit transition (1 + 1 = 2 Hz) exceeds 1 Hz notch -> error.
  expect_error(
    design_filter_fir(
      sample_rate = 2000,
      high_pass_freq = 25, low_pass_freq = 24,
      low_pass_trans_freq = 1, high_pass_trans_freq = 1,
      filter_order = 200
    ),
    regexp = "transition bandwidths overlap"
  )
  # A single explicit transition (1.5 Hz) that alone exceeds the notch -> error.
  expect_error(
    design_filter_fir(
      sample_rate = 2000,
      high_pass_freq = 25, low_pass_freq = 24,
      low_pass_trans_freq = 1.5,
      filter_order = 200
    ),
    regexp = "transition bandwidths overlap"
  )
})

scale_filter <- function(b, fband, freq, len) {

  if ( fband ) {
    b <- b / sum(b)
  } else {
    if ( freq[[4]] == 1 ) {
      # length(w) == 1
      f0 <- 1
    } else {
      f0 <- mean(freq[c(3, 4)])
    }
    # b = b / abs( exp(-1i*2*pi*(0:L-1)*(f0/2))*(b.') );
    b <- b / Mod(sum(
      exp(-1i * 2 * pi * seq(0, len - 1) * (f0 / 2)) * b
    ))
  }

  b

}

fir_validate <- function(n, max_freq, magnitude, odd_allowed = FALSE) {

  # order
  if (missing(n) || length(n) != 1 || !is.numeric(n) || n - round(n) != 0 || n <= 0 ) {
    stop("Filter order `n` must be a positive integer")
  }

  n_is_odd <- n %% 2 == 1

  if ( magnitude[[length(magnitude)]] != 0 && max_freq == 1 && n_is_odd && !odd_allowed ) {
    warning("IR filters must have a gain of zero at the Nyquist frequency. Increasing the order by 1")
    n <- n + 1
  }

  n
}


#' Window-based \code{FIR} filter design
#' @description
#' Generate a \code{fir1} filter that is checked against \code{Matlab}
#' \code{fir1} function.
#'
#' @param n filter order
#' @param w band edges, non-decreasing vector in the range 0 to 1, where 1 is
#' the \code{Nyquist} frequency. A scalar for high-pass or low-pass filters,
#' a vector pair for band-pass or band-stop, or a vector for an
#' alternating pass/stop filter.
#' @param type type of the filter, one of \code{"low"} for a low-pass filter,
#' \code{"high"} for a high-pass filter, \code{"stop"} for a stop-band
#' (band-reject) filter, \code{"pass"} for a pass-band filter, \code{"DC-0"}
#' for a band-pass as the first band of a multi-band filter, or \code{"DC-1"}
#' for a band-stop as the first band of a multi-band filter; default \code{"low"}
#' @param window smoothing window function or a numerical vector. The filter is
#' the same shape as the smoothing window. When \code{window} is a function,
#' \code{window(n+1)} will be called, otherwise the length of the window
#' vector needs to have length of \code{n+1}; default: \code{hamming}
#' @param scale whether to scale the filter; default is true
#' @param hilbert whether to use 'Hilbert' transformer; default is false
#' @returns The \code{FIR} filter coefficients with class \code{'Arma'}.
#' The moving average coefficient is a vector of length \code{n+1}.
#' @export
fir1 <- function(
    n, w, type = c("low", "high", "stop", "pass", "DC-0", "DC-1"),
    window = hamming, scale = TRUE, hilbert = FALSE
) {
  type_missing <- missing(type)

  nw <- length(w)
  if (!nw || any(w < 0 | w > 1)) {
    stop("`fir1`: w must be a real vector from range 0.0 to 1.0")
  }
  if (nw > 1 && any(diff(w) < 0)) {
    stop("`fir1`: w must be non-decreasing")
  }

  if (w[[length(w)]] >= 1) {
    w <- w[-length(w)]
    if (length(w) == 1) {
      type <- "high"
    } else {
      type <- "DC-0"
    }
  } else if (type_missing) {
    if (nw == 1) {
      type <- "low"
    } else if (nw == 2) {
      type <- "pass"
    } else {
      type <- "DC-0"
    }
  } else {
    type <- match.arg(type)
  }

  nw <- length(w)

  nbands <- nw + 1

  # make sure default 3 band filter is bandpass
  if ( type == "pass" ) {
    if ( nbands > 2 ) {
      type <- "DC-0"
    }
  }

  # frequency vector
  freq <- c(0, rep(w, each = 2), 1)

  # magnitude vector
  use_first_band <- !type %in% c("DC-0", "high")
  magnitude <- rep( (use_first_band + seq(0, nbands - 1)) %% 2, each = 2)

  n <- fir_validate(n, freq[[length(freq)]], magnitude, odd_allowed = hilbert)

  # filter length is order+1
  filter_len <- n + 1

  if (is.function(window)) {
    wind <- window(filter_len)
    if ( length(wind) != filter_len ) {
      stop("`fir1`: window length is invalid given requested length: (n+1)")
    }
  } else {
    if (length(window) != filter_len) {
      stop("`fir1`: window must be either a function or a vector of length (n+1)")
    }
    wind <- window
  }

  if (hilbert) {
    hh <- firls(n, freq, magnitude, ftype = "hilbert")$b
  } else {
    hh <- firls(n, freq, magnitude)$b
  }

  b <- wind * hh

  if (scale) {
    b <- scale_filter(b, use_first_band, freq, filter_len)
  }

  return(structure(list(b = b, a = 1), class = "Arma"))
}

sinc <- function(x) {
  v <- pi * x
  re <- sin(v) / v
  if (length(v)) {
    re[ v == 0 ] <- 1
  }
  re
}

#' Least-squares linear-phase \code{FIR} filter design
#' @description
#' Produce a linear phase filter from the weighted mean squared such that error
#'    in the specified bands is minimized.
#' @param N filter order, must be even (if odd, then will be increased by one)
#' @param freq vector of frequency points in the range from 0 to 1, where 1
#'    corresponds to the \code{Nyquist} frequency.
#' @param A vector of the same length as \code{freq} containing the desired
#'    amplitude at each of the points specified in \code{freq}.
#' @param W weighting function that contains one value for each band that
#'    weights the mean squared error in that band. \code{W} must be half the
#'    length of \code{freq}.
#' @param ftype transformer type; default is \code{""}; alternatively,
#'    \code{'h'} or \code{'hilbert'} for 'Hilbert' transformer.
#' @param legacy whether to use the legacy implementations, which uses
#'    \code{\link{qr.solve}} instead of faster \code{\link{chol}}
#' @returns The \code{FIR} filter coefficients with class \code{'Arma'}.
#'    The moving average coefficient is a vector of length \code{n+1}.
#' @export
firls <- function(N, freq, A, W = NULL, ftype = "", legacy = FALSE) {

  if (!is.numeric(N) || length(N) != 1 || !is.finite(N) || N <= 0) {
    stop("Invalid input for N.")
  }
  N <- as.integer(N)

  if (!is.numeric(freq) || !is.numeric(A) || length(freq) != length(A) || any(freq < 0) || any(freq > 1)) {
    stop("Invalid input for freq or A.")
  }
  freq <- as.double(freq)
  A <- as.double(A)

  if (!is.null(W)) {
    if (length(W) == 0) {
      W <- rep(1, floor(length(freq) / 2))
    } else {
      W <- as.double(W)
    }
  } else {
    W <- rep(1, floor(length(freq) / 2))
  }

  if (identical(ftype, "h") || identical(ftype, "hilbert")) {
    filtype <- 1
    differ <- 0
  } else if (identical(ftype, "d") || identical(ftype, "differentiator")) {
    stop("Filter type 'differentiator' has not been implemented yet")
    filtype <- 1
    differ <- 1
  } else {
    filtype <- 0
    differ <- 0
  }

  if (length(W) == 0) {
    weight <- rep(1, floor(length(freq) / 2))
  } else {
    weight <- W
  }

  freq_length <- length(freq)
  wt <- abs(sqrt(weight))

  N <- N + 1

  Fn <- freq / 2

  dF <- diff(Fn)

  if (any(dF < 0)) {
    stop("Invalid frequency vector.")
  }

  lendF <- freq_length - 1
  fullband <- FALSE

  if (lendF > 1) {
    if (legacy) {
      # Original check: effectively always FALSE for standard fir1 inputs
      fullband <- all(dF == 0)
    } else {
      # Correct check: the "gaps" between adjacent bands are the even-indexed
      # diffs (1-based). fir1 builds freq = c(0, w1, w1, w2, w2, ..., 1) so
      # those gap diffs are exactly 0 -> fullband = TRUE -> fast path, no matrix.
      gap_idx <- seq(2, length(dF), by = 2)
      fullband <- all(dF[gap_idx] == 0)
    }
  }

  tempW <- wt - wt[1]
  constant_weights <- sum(tempW) == 0

  L <- (N - 1) / 2
  Nodd <- (N %% 2 == 1)

  b0 <- 0

  if (filtype == 0) {
    if (!Nodd) {
      m <- (0:L) + 0.5
    } else {
      m <- 0:L
    }

    k <- m

    need_matrix <- !fullband || !constant_weights

    if (need_matrix) {
      result <- initMatrices(m)
      I1 <- result$I1
      I2 <- result$I2
      G <- result$G
    }

    if (Nodd) {
      k <- k[-1]
      b0 <- 0
    }

    b <- numeric(length(k))

    for (s in seq(1, length(Fn), by = 2)) {
      if (Fn[s + 1] == Fn[s]) {
        next
      }
      m_s <- (A[s + 1] - A[s]) / (Fn[s + 1] - Fn[s])
      b1 <- A[s] - m_s * Fn[s]

      if (Nodd) {
        b0 <- b0 + (b1 * (Fn[s + 1] - Fn[s]) + m_s / 2 * (Fn[s + 1]^2 - Fn[s]^2)) * wt[(s + 1) / 2]^2
      }

      b <- b + (m_s / (4 * pi^2) * (cos(2 * pi * k * Fn[s + 1]) - cos(2 * pi * k * Fn[s])) / (k^2)) * wt[(s + 1) / 2]^2
      b <- b + (Fn[s + 1] * (m_s * Fn[s + 1] + b1) * sinc(2 * k * Fn[s + 1]) -
                  Fn[s] * (m_s * Fn[s] + b1) * sinc(2 * k * Fn[s])) * wt[(s + 1) / 2]^2

      if (need_matrix) {
        G <- G + (0.5 * Fn[s + 1] * (sinc(2 * I1 * Fn[s + 1]) + sinc(2 * I2 * Fn[s + 1])) -
                    0.5 * Fn[s] * (sinc(2 * I1 * Fn[s]) + sinc(2 * I2 * Fn[s]))) * wt[(s + 1) / 2]^2
      }
    }

    if (Nodd) {
      b <- c(b0, b)
    }

    if (need_matrix) {
      if (legacy) {
        a <- qr.solve(G, b, tol = 1e-30)
      } else {
        # G is symmetric positive (semi-)definite; Cholesky is ~3x faster than QR.
        # Fall back to QR if G is numerically singular.
        R <- tryCatch(chol(G), error = function(e) { NULL })
        if (is.null(R)) {
          a <- qr.solve(G, b, tol = 1e-30)
        } else {
          a <- backsolve(R, backsolve(R, b, transpose = TRUE))
        }
      }
    } else {
      a <- (wt[1]^2) * 4 * b
      if (Nodd) {
        a[1] <- a[1] / 2
      }
    }

    if (Nodd) {
      h <- c(a[seq(L + 1, 2, by = -1)] / 2, a[1], a[2:(L + 1)] / 2)
    } else {
      h <- 0.5 * c(rev(a), a)
    }

  } else if (filtype == 1) {
    if (differ) {
      do_weight <- as.double(abs(A[seq(2, length(A), by = 2)]) + abs(A[seq(1, length(A), by = 2)]) > 0)
    } else {
      do_weight <- numeric(length(Fn))
    }

    need_matrix <- !fullband || any(do_weight) || !constant_weights

    if (Nodd) {
      m <- 1:L
    } else {
      m <- 0:L + 0.5
    }

    k <- m
    b <- numeric(length(k))

    if (need_matrix) {
      result <- initMatrices(m)
      I1 <- result$I1
      I2 <- result$I2
      G <- result$G
    } else {
      G <- matrix(0, nrow = 0, ncol = 0)
    }

    for (s in seq(1, length(Fn), by = 2)) {
      if ( Fn[s + 1] == Fn[s] ) { next }
      if (do_weight[(s + 1) / 2]) {
        if (Fn[s] == 0) {
          Fn[s] <- 1e-5
        }

        m_s <- (A[s + 1] - A[s]) / (Fn[s + 1] - Fn[s])
        b1 <- A[s] - m_s * Fn[s]

        snint1 <- sineint(2 * pi * k * Fn[s + 1]) - sineint(2 * pi * k * Fn[s])
        csint1 <- Re((-1 / 2) * (pracma::expint(1i * 2 * pi * k * Fn[s + 1]) +
                                 pracma::expint(-1i * 2 * pi * k * Fn[s + 1]) -
                                 pracma::expint(1i * 2 * pi * k * Fn[s]) -
                                 pracma::expint(-1i * 2 * pi * k * Fn[s])))

        b <- b + (m_s * snint1 +
                    b1 * 2 * pi * k * (-sinc(2 * k * Fn[s + 1]) + sinc(2 * k * Fn[s]) + csint1)) *
          wt[(s + 1) / 2]^2

        snint1 <- sineint(2 * pi * Fn[s + 1] * (-I2))
        snint2 <- sineint(2 * pi * Fn[s + 1] * I1)
        snint3 <- sineint(2 * pi * Fn[s] * (-I2))
        snint4 <- sineint(2 * pi * Fn[s] * I1)

        G <- G - ((-1 / 2) * (cos(2 * pi * Fn[s + 1] * (-I2)) /  Fn[s + 1] -
                              2 * snint1 * pi * I2 -
                              cos(2 * pi * Fn[s + 1] * I1) / Fn[s + 1] +
                              2 * snint2 * pi * I1) -
                    (-1 / 2) * (cos(2 * pi * Fn[s] * (-I2)) / Fn[s] -
                                2 * snint3 * pi * I2 -
                                cos(2 * pi * Fn[s] * I1) / Fn[s] +
                                2 * snint4 * pi * I1)) * wt[(s + 1) / 2]^2
      } else {
        m_s <- (A[s + 1] - A[s]) / (Fn[s + 1] - Fn[s])
        b1 <- A[s] - m_s * Fn[s]

        b <- b + (m_s / (4 * pi^2) * (sin(2 * pi * k * Fn[s + 1]) - sin(2 * pi * k * Fn[s])) / (k^2)) * wt[(s + 1) / 2]^2
        b <- b + ((m_s * Fn[s] + b1) * cos(2 * pi * k * Fn[s]) -
                    (m_s * Fn[s + 1] + b1) * cos(2 * pi * k * Fn[s + 1])) / (2 * pi * k) * wt[(s + 1) / 2]^2

        if (need_matrix) {
          G <- G + (0.5 * Fn[s + 1] * (sinc(2 * I1 * Fn[s + 1]) - sinc(2 * I2 * Fn[s + 1])) -
                      0.5 * Fn[s] * (sinc(2 * I1 * Fn[s]) - sinc(2 * I2 * Fn[s]))) * wt[(s + 1) / 2]^2
        }
      }
    }

    if (need_matrix) {
      a <- solve(G, b)
    } else {
      a <- -4 * b * wt[1]^2
    }

    if (Nodd) {
      h <- 0.5 * c(rev(a), 0, -a)
    } else {
      h <- 0.5 * c(rev(a), -a)
    }

    if (differ) {
      h <- -h
    }

  } else {
    h <- 0
  }

  return(structure(list(b = h, a = 1), class = "Arma"))
}


initMatrices <- function(m) {
  k <- m
  x <- matrix(k, ncol = length(m), nrow = length(m), byrow = TRUE)
  y <- matrix(m, ncol = length(k), nrow = length(k), byrow = FALSE)

  I1 <- x + y
  I2 <- x - y
  G <- matrix(0, ncol = length(m), nrow = length(m))

  return(list(I1 = I1, I2 = I2, G = G))
}


# Might not be implemented correctly, not used
fir2 <- function(n, f, m, grid_n = 512,
                 ramp_n = grid_n / 20,
                  window = hamming(n + 1))
{
  t <- length(f)
  if (t < 2 || f[1] != 0 || f[t] != 1 || any(diff(f) < 0))
    stop("frequency must be nondecreasing starting from 0 and ending at 1")
  if (t != length(m))
    stop("frequency and magnitude vectors must be the same length")
  if (length(grid_n) > 1 || length(ramp_n) > 1)
    stop("grid_n and ramp_n must be integers")
  if (length(window) != n + 1)
    stop("window must be of length n+1")
  if (2 * grid_n < n + 1)
    grid_n <- 2^ceiling(log2(abs(n + 1)))
  if (ramp_n > 0) {
    basef <- f
    basem <- m
    idx <- which(diff(f) == 0)
    f[idx] <- f[idx] - ramp_n / grid_n / 2
    f[idx + 1] <- f[idx + 1] + ramp_n / grid_n / 2
    f <- c(f, basef[idx])
    f[f < 0] <- 0
    f[f > 1] <- 1
    f <- sort(unique(c(f, basef[idx])))
    m <- approx(basef, basem, f, ties = "ordered")$y
  }
  grid <- approx(f, m, seq(0, 1, length = grid_n + 1), ties = "ordered")$y
  if ((n %% 2) == 0) {
    b <- ifft(c(grid, grid[seq(grid_n, 2, by = -1)]))
    mid <- (n + 1) / 2
    b <- Re(c(b[(2 * grid_n - floor(mid) + 1):(2 * grid_n)],
              b[1:ceiling(mid)]))
  } else {
    b <- ifft(c(grid, rep(0, grid_n * 2), grid[seq(grid_n, 2, by = -1)]))
    b <- 2 * Re(c(b[seq(length(b) - n + 1, length(b), by = 2)],
                  b[seq(2, n + 2, by = 2)]))
  }
  b <- b * window
  b
}

# Tests for the firls() fast-path fix (fullband check + Cholesky solver).
#
# Two invariants to verify:
#   1. Correctness: new path produces coefficients numerically identical to
#      legacy (within ~1e-12 for well-conditioned cases, noting the two paths
#      are algebraically equivalent when fullband=TRUE).
#   2. fullband detection fires correctly:
#      - fir1-style specs (no gaps)  -> fullband = TRUE  -> fast path
#      - specs with inter-band gaps  -> fullband = FALSE -> matrix path
#
# Helper: normalised frequency gain at f0 in [0, 1]
fir_gain <- function(b, f0) {
  L <- length(b)
  Mod(sum(exp(-1i * 2 * pi * seq(0, L - 1) * (f0 / 2)) * b))
}

# -- 1. fir1-compatible specs: no inter-band gaps ------------------------------

test_that("firls new vs legacy agree for no-gap (fir1-compatible) specs", {

  cases <- list(
    # name,  N,   freq,                    A
    list("low-pass even",     100, c(0, 0.3, 0.3, 1),       c(1, 1, 0, 0)),
    list("low-pass odd",      101, c(0, 0.3, 0.3, 1),       c(1, 1, 0, 0)),
    list("high-pass even",    80,  c(0, 0.6, 0.6, 1),       c(0, 0, 1, 1)),
    list("band-pass even",    120, c(0, 0.2, 0.2, 0.5, 0.5, 1), c(0, 0, 1, 1, 0, 0)),
    list("band-stop even",    120, c(0, 0.2, 0.2, 0.5, 0.5, 1), c(1, 1, 0, 0, 1, 1)),
    list("multi-band even",   60,  c(0, 0.1, 0.1, 0.3, 0.3, 0.6, 0.6, 1),
                                   c(1, 1, 0, 0, 1, 1, 0, 0))
  )

  for (case in cases) {
    label <- case[[1]]
    N <- case[[2]]
    freq <- case[[3]]
    A <- case[[4]]

    b_new    <- firls(N, freq, A)$b
    b_legacy <- firls(N, freq, A, legacy = TRUE)$b

    expect_equal(
      b_new, b_legacy, tolerance = 1e-12,
      label = sprintf("firls no-gap case '%s'", label)
    )
  }
})

# -- 2. fir1() output unchanged ------------------------------------------------
# fir1 calls firls internally; verify that the window-based filter coefficients
# are unchanged across several filter types.

test_that("fir1 coefficients are unchanged by the fast-path fix", {

  for (n in c(20, 50, 101)) {
    b_new    <- fir1(n, 0.3, "low")$b
    b_legacy <- firls(n, c(0, 0.3, 0.3, 1), c(1, 1, 0, 0), legacy = TRUE)$b
    # fir1 applies a hamming window on top, so compare the raw firls output
    # rather than the final windowed b.  Instead just check the gain is the same.
    # (fir1 always applies the hamming window, so we re-derive for legacy too)
    wind <- hamming(n + 1)
    b_legacy_win <- b_legacy * wind
    # Normalise both at DC the same way (scale=TRUE in fir1 = b/sum(b) for LP)
    b_legacy_win <- b_legacy_win / sum(b_legacy_win)

    expect_equal(
      b_new, b_legacy_win, tolerance = 1e-12,
      label = sprintf("fir1 low-pass n=%d", n)
    )
  }
})

# -- 3. Gain correctness after fix --------------------------------------------
# Independent check: filters produced by the new path must satisfy their
# passband / stopband specifications to within reasonable tolerance.

test_that("firls (new path) satisfies passband gain requirements", {

  # Low-pass: gain near 1 in passband, near 0 in stopband
  b <- firls(100, c(0, 0.3, 0.3, 1), c(1, 1, 0, 0))$b
  expect_gt(fir_gain(b, 0),   0.99)   # DC gain close to 1
  expect_gt(fir_gain(b, 0.15), 0.9)   # mid-passband
  expect_lt(fir_gain(b, 0.5),  0.05)  # well into stopband

  # High-pass
  b <- firls(100, c(0, 0.6, 0.6, 1), c(0, 0, 1, 1))$b
  expect_lt(fir_gain(b, 0.1),  0.05)
  expect_gt(fir_gain(b, 0.8),  0.9)
  expect_gt(fir_gain(b, 1.0),  0.99)

  # Band-pass
  b <- firls(120, c(0, 0.2, 0.2, 0.6, 0.6, 1), c(0, 0, 1, 1, 0, 0))$b
  expect_lt(fir_gain(b, 0.05), 0.1)
  expect_gt(fir_gain(b, 0.4),  0.9)
  expect_lt(fir_gain(b, 0.85), 0.1)
})

# -- 4. Specs with inter-band gaps must still give correct results -------------
# When freq has genuine gaps (transition bands), fullband=FALSE -> matrix path.
# Verify the Cholesky solver gives the same answer as QR for these cases.

test_that("firls with inter-band gaps: Cholesky agrees with legacy QR", {

  gap_cases <- list(
    list("gap low-pass",  80,  c(0, 0.25, 0.35, 1),       c(1, 1, 0, 0)),
    list("gap band-pass", 80,  c(0, 0.15, 0.25, 0.5, 0.6, 1), c(0, 0, 1, 1, 0, 0)),
    list("gap notch",     80,  c(0, 0.2, 0.3, 0.5, 0.6, 1),   c(1, 1, 0, 0, 1, 1))
  )

  for (case in gap_cases) {
    label <- case[[1]]
    N <- case[[2]]
    freq <- case[[3]]
    A <- case[[4]]

    b_new    <- firls(N, freq, A)$b
    b_legacy <- firls(N, freq, A, legacy = TRUE)$b

    # Both answers are within the backward-error bound of the ill-conditioned
    # system; 1e-10 is tighter than the actual error floor for small N.
    expect_equal(
      b_new, b_legacy, tolerance = 1e-10,
      label = sprintf("firls gap case '%s'", label)
    )
  }
})

# -- 5. design_filter_fir ("firls" method) coefficients unchanged by fix -------
# design_filter_fir builds the firls freq spec as:
#   w   = c(0, rep(bands, each = 2), 1)       # no gaps between adjacent bands
#   a   = rep(mag2, each = 2)[-c(1, end)]
# then applies a hamming window and scales.  The test reproduces that exact call
# with legacy=TRUE and confirms bit-for-bit agreement (within 1e-12).

test_that("design_filter_fir method='firls' coefficients unchanged by fix", {

  fs <- 500

  reconstruct_legacy <- function(f_new) {
    # Recover bands and mag2 from stored parameters
    freq_stored <- f_new$parameters$frequency   # c(0, bands..., 1)
    amp_stored  <- f_new$parameters$amplitude   # mag2
    bands <- freq_stored[-c(1, length(freq_stored))]   # drop 0 and 1
    mag2  <- amp_stored

    # Reproduce design_filter_fir's internal firls call exactly
    w <- c(0, rep(bands, each = 2), 1)
    a <- rep(mag2, each = 2)
    a <- a[-c(1, length(a))]

    b <- firls(N = f_new$parameters$order, freq = w, A = a, legacy = TRUE)$b

    # Same hamming window as design_filter_fir
    wind <- hamming(length(b))
    b <- b * wind

    # Reproduce design_filter_fir's reference-frequency scaling (matches the
    # outer `if (scale)` block in design_filter_fir, not fir1's scale_filter):
    #   low-pass  -> f0 = 0  (DC,      b / sum(b))
    #   high-pass -> f0 = 1  (Nyquist)
    #   band-pass -> f0 = mean(kaisprm$Wc)
    #                      = mean(c(passband, stopband))  [for the firls method]
    #   band-stop -> f0 = 0
    ftype <- f_new$parameters$type[[2]]
    f0 <- switch(
      ftype,
      "low"  = 0,
      "high" = 1,
      "pass" = mean(c(f_new$parameters$passband, f_new$parameters$stopband)),
      0   # stop / default -> DC
    )
    L <- length(b)
    b / Mod(sum(exp(-1i * 2 * pi * seq(0, L - 1) * (f0 / 2)) * b))
  }

  # Low-pass
  f_lp <- design_filter_fir(
    sample_rate = fs, low_pass_freq = 50,
    filter_order = 100, method = "firls"
  )
  expect_equal(f_lp$b, reconstruct_legacy(f_lp), tolerance = 1e-12,
               label = "design_filter_fir firls low-pass coeff equality")

  # Band-pass
  f_bp <- design_filter_fir(
    sample_rate = fs, high_pass_freq = 80, low_pass_freq = 150,
    filter_order = 100, method = "firls"
  )
  expect_equal(f_bp$b, reconstruct_legacy(f_bp), tolerance = 1e-12,
               label = "design_filter_fir firls band-pass coeff equality")
})

#' @title Compute 'multitaper' spectral densities of time-series data
#' @name multitaper
#'
#' @param data numerical vector, signal traces
#' @param data_length length of data
#' @param fs sampling frequency in 'Hz'
#' @param frequency_range frequency range to look at; length of two
#' @param time_bandwidth a number indicating time-half bandwidth product; i.e.
#' the window duration times the half bandwidth of main lobe; default is
#' \code{5}
#' @param num_tapers number of 'DPSS' tapers to use; default is \code{NULL} and
#' will be automatically computed from \code{floor(2*time_bandwidth - 1)}
#' @param window_params vector of two numbers; the first number is the
#' window size in seconds; the second number if the step size; default is
#' \code{c(5, 1)}
#' @param detrend_opt how you want to remove the trend from data window; options
#' are \code{'linear'} (default), \code{'constant'}, and \code{'off'}
#' @param nfft 'NFFT' size, positive; see 'Details'
#' @returns \code{multitaper_config} returns a list of configuration parameters
#' for the filters; \code{multitaper} also returns the time, frequency and
#' corresponding spectral power.
#'
#' @details The original source code comes from 'Prerau' Lab (see 'Github'
#' repository \code{'multitaper_toolbox'} under user \code{'preraulab'}).
#' The results tend to agree with their 'Python' implementation with precision
#' on the order of at \code{1E-7} with standard deviation at most \code{1E-5}.
#' The original copy was licensed under a Creative Commons Attribution
#' 'NC'-'SA' 4.0 International License
#' (\url{https://creativecommons.org/licenses/by-nc-sa/4.0/}).
#'
#' This package (\code{'ravetools'}) redistributes the \code{multitaper}
#' function under minor modifications on \code{nfft}. In the original copy
#' there is no parameter to control the exact numbers of \code{nfft}, and
#' the \code{nfft} is always the power of 2. While choosing
#' \code{nfft} to be the power of 2 is always recommended, the modified code
#' allows other choices.
#'
#' @examples
#'
#' \donttest{
#'
#' # Takes long to run
#'
#' time <- seq(0, 3, by = 0.001)
#' x <- sin(time * 20*pi) + exp(-time^2) * cos(time * 10*pi)
#'
#' res <- multitaper(
#'   x, 1000, frequency_range = c(0,15),
#'   time_bandwidth=1.5,
#'   window_params=c(2,0.01)
#' )
#'
#'
#' image(
#'   x = res$time,
#'   y = res$frequency,
#'   z = 10 * log10(res$spec),
#'   xlab = "Time (s)",
#'   ylab = 'Frequency (Hz)',
#'   col = matlab_palette()
#' )
#'
#' }
#'
NULL

multitaper_process_input <- function(
    len_data, fs, frequency_range = NULL, time_bandwidth = 5,
    num_tapers = NULL, window_params = c(5, 1), nfft = NA,
    detrend_opt = "linear") {


  # Set frequency range if not provided
  if (is.null(frequency_range)) {
    frequency_range <- c(0, fs / 2)
  }

  # Set detrend method
  detrend_opt <- tolower(detrend_opt)
  if (detrend_opt != "linear") {
    if (detrend_opt == "const") {
      detrend_opt <- "constant"
    } else if (detrend_opt == "none" || detrend_opt == "false") {
      detrend_opt <- "off"
    } else {
      stop(paste("'", toString(detrend_opt),
                 "' is not a valid detrend_opt argument. The",
                 " choices are: 'constant', 'linear', or 'off'.",
                 sep = ""))
    }
  }

  # Check if frequency range is valid
  if (frequency_range[2] > fs / 2) {
    frequency_range[2] <- fs / 2
    warning(paste(
      "Upper frequency range greater than Nyquist, setting range to [",
      toString(frequency_range[1]), ",",
      toString(frequency_range[2]), "].",
      sep = ""))
  }

  # Set number of tapers if none provided
  optimal_num_tapers <- floor(2 * time_bandwidth) - 1
  if (is.null(num_tapers)) {
    num_tapers <- optimal_num_tapers
  }

  # Warn if number of tapers is suboptimal
  if (num_tapers != optimal_num_tapers) {
    warning(paste("Suboptimal number of tapers being used. Number of tapers is optimal at floor(2*TW) - 1 which is ",
                  toString(optimal_num_tapers), " in this case.", sep = ""))
  }


  # Check if window size is valid, fix if not
  if ((window_params[1] * fs) %% 1 != 0) {
    winsize_samples <- round(window_params[1] * fs)
    warning(paste("Window size is not divisible by sampling frequency. Adjusting window",
                  " size to ", toString(winsize_samples / fs), " seconds.", sep = ""))
  } else {
    winsize_samples <- window_params[1] * fs
  }

  # Check if window step size is valid, fix if not
  if ((window_params[2] * fs) %% 1 != 0) {
    winstep_samples <- round(window_params[2] * fs)
    warning(paste("Window step size is not divisible by sampling frequency. Adjusting window",
                  " step size to ", toString(winstep_samples / fs), " seconds.", sep = ""))
  } else {
    winstep_samples <- window_params[2] * fs
  }


  # Check if length of data is smaller than window (bad)
  if (len_data < winsize_samples) {
    stop(paste("Data length (", toString(len_data), ") is shorter than the window size (",
               toString(winsize_samples), "). Either increase data length or decrease",
               " window size.", sep = ""))
  }

  # Find window start indices and num of windows
  window_start <- seq(1, len_data - winsize_samples + 1, by = winstep_samples)
  num_windows <- length(window_start)

  # Get num points in FFT
  nfft <- as.integer(nfft)
  if (!is.na(nfft) && nfft < 1) {
    stop("Invalid nfft")
  }
  if (is.na(nfft)) {
    nfft <- max(2^ceiling(log2(abs(winsize_samples))), winsize_samples)
  }


  return(
    list(
      fs = fs,
      frequency_range = frequency_range,
      time_bandwidth = time_bandwidth,
      num_tapers = num_tapers,
      winsize_samples = winsize_samples,
      winstep_samples = winstep_samples,
      window_start = window_start,
      num_windows = num_windows,
      nfft = nfft,
      detrend_opt = detrend_opt
    )
  )
}

multitaper_process_spectrogram_params <- function(
  fs, nfft, frequency_range, window_start, datawin_size
) {

  # Create frequency vector
  df <- fs / nfft
  sfreqs <- seq(df / 2, fs - (df / 2), by = df)

  # Get frequencies for given frequency range
  freq_inds <- (sfreqs >= frequency_range[1]) & (sfreqs <= frequency_range[2])
  sfreqs <- sfreqs[freq_inds]

  # Compute times in middle of each spectrum
  window_middle_times <- window_start + round(datawin_size / 2)
  stimes <- window_middle_times / fs

  # Get indices for each window
  window_idxs <- lapply(window_start, function(start) {
    seq(start, start + datawin_size - 1, by = 1)
  }) # list of indices for n windows


  return(list(
    window_idxs = window_idxs,
    stimes = stimes,
    sfreqs = sfreqs,
    freq_inds = freq_inds
  ))

}

#' @rdname multitaper
#' @export
multitaper_config <- function(
    data_length, fs, frequency_range = NULL, time_bandwidth = 5,
    num_tapers = NULL, window_params = c(5, 1),
    nfft = NA, detrend_opt = "linear"
) {
  res <- multitaper_process_input(
    data_length, fs, frequency_range, time_bandwidth, num_tapers,
    window_params, nfft, detrend_opt)

  # Set up spectrogram parameters
  res2 <- with(res, {
    multitaper_process_spectrogram_params(fs, nfft, frequency_range, window_start, winsize_samples)
  })

  structure(c(res, res2), class = "ravetools-multitaper-config")

}

#' @export
`print.ravetools-multitaper-config` <- function(x, ...) {
  # display_spectrogram_properties(fs, time_bandwidth, num_tapers, c(winsize_samples, winstep_samples), frequency_range,
  #                                detrend_opt)
  data_window_params <- c(x$winsize_samples, x$winstep_samples)
  data_window_params <- data_window_params / x$fs

  # Print spectrogram properties
  cat(
    sep = "",
    "Multitaper Spectrogram Configuration: \n",

    "     Spectral Resolution: ",
    toString(2 * x$time_bandwidth / data_window_params[1]), "Hz\n",

    "     Window Length: ",
    toString(data_window_params[1]), "s\n",


    "     Window Step: ",
    toString(data_window_params[2]), "s\n",

    "     Time Half-Bandwidth Product: ",
    toString(x$time_bandwidth), "\n",

    "     Number of Tapers: ",
    toString(x$num_tapers), "\n",

    "     Frequency Range: ",
    toString(x$frequency_range[1]), "-",
    toString(x$frequency_range[2]), "Hz\n",

    "     Detrend: ", x$detrend_opt, "\n"
  )

}



multitaper_calc_mts_segment <- function(data_segment, dpss_tapers, nfft, freq_inds, detrend_opt) {

  # If segment has all zeros, return vector of zeros
  if (all(data_segment == 0)) {
    if (is.logical(freq_inds)) {
      ret <- rep(0, sum(freq_inds))
    } else {
      ret <- rep(0, length(freq_inds))
    }

    return(ret)
  }

  # Optionally detrend data to remove low freq DC component
  if (detrend_opt != "off") {
    data_segment <- detrend(data_segment, trend = detrend_opt)
  }

  # Multiply data by dpss tapers (STEP 2)
  if (is.matrix(dpss_tapers)) {
    tapered_data <- sweep(dpss_tapers, 1, data_segment, "*",
                          check.margin = FALSE)
  } else {
    tapered_data <- matrix(data_segment * dpss_tapers, ncol = 1L)
  }


  # Manually add nfft zero-padding (R's fft function does not support)
  npad <- nfft - nrow(tapered_data)
  # npad_left <- ceiling(npad / 2)
  # npad_right <- npad - npad_left
  tapered_padded_data <- rbind(
    # array(0, c(
    #   npad_left,
    #   ncol(tapered_data)
    # )),
    tapered_data,
    array(0, c(
      npad,
      ncol(tapered_data)
    ))
  )

  # Compute the FFT (STEP 3)
  # fft_data <- apply(tapered_padded_data, 2, stats::fft)
  # fft_range = fft_data[freq_inds,,drop = FALSE]

  # nr1 <- nrow(tapered_padded_data)
  fft_data <- mvfftw_r2c(tapered_padded_data)
  # nr2 <- nrow(fft_data)
  #
  # if( nr1 %% 2 == 0 ) {
  #   fft_data <- rbind(fft_data, Conj(
  #     fft_data[nr2 - seq_len(nr1 - nr2), ,drop=FALSE]
  #   ))
  # } else {
  #   fft_data <- rbind(fft_data, Conj(
  #     fft_data[(nr2 + 1) - seq_len(nr1 - nr2), ,drop=FALSE]
  #   ))
  # }
  if (is.logical(freq_inds)) {
    freq_inds <- which(freq_inds)
  }
  fft_range <- fft_data[freq_inds, , drop = FALSE]

  # Take the FFT magnitude (STEP 4.1)
  magnitude <- Im(fft_range)^2 + Re(fft_range)^2
  mt_spectrum <- rowSums(magnitude)

  return(mt_spectrum)
}

#' @rdname multitaper
#' @export
multitaper <- function(
    data, fs, frequency_range = NULL, time_bandwidth = 5,
    num_tapers = NULL, window_params = c(5, 1),
    nfft = NA, detrend_opt = "linear"
) {

  # Make sure data is 1D atomic vector
  if ((is.atomic(data) == FALSE) || is.list(data)) {
    stop("data must be a 1D atomic vector")
  }

  dlen <- length(data)

  conf <- multitaper_config(
    dlen, fs, frequency_range, time_bandwidth, num_tapers,
    window_params, nfft, detrend_opt
  )

  window_idxs <- conf$window_idxs
  winsize_samples <- conf$winsize_samples
  freq_inds <- conf$freq_inds
  stimes <- conf$stimes
  sfreqs <- conf$sfreqs
  list2env(conf, envir = environment())


  # Split data into window segments
  split_data_helper <- function(indices, data) { # for sapply when splitting data into windows
    data_seg <- data[indices]
    return(data_seg)
  }
  data_segments <- t(sapply(window_idxs, split_data_helper, data = data))


  # COMPUTE THE MULTITAPER SPECTROGRAM
  #     STEP 1: Compute DPSS tapers based on desired spectral properties
  #     STEP 2: Multiply the data segment by the DPSS Tapers
  #     STEP 3: Compute the spectrum for each tapered segment
  #     STEP 4: Take the mean of the tapered spectra

  # Compute DPSS tapers (STEP 1)
  dpss_tapers <- waveslim::dpss.taper(winsize_samples, num_tapers, time_bandwidth) * sqrt(fs)

  # Compute multitaper
  mt_spectrogram <- apply(
    data_segments, 1, multitaper_calc_mts_segment,
    dpss_tapers = dpss_tapers, nfft = nfft,
    freq_inds = freq_inds, detrend_opt = detrend_opt)

  # Compute mean fft magnitude (STEP 4.2)
  mt_spectrogram <- Conj(t(mt_spectrogram)) / fs^2 / num_tapers


  # if(all(as.vector(mt_spectrogram) == 0)) {
  #   print("Spectrogram calculated as all zeros, no plot shown")
  # }else if(plot_on) {
    # fields::image.plot(x=conf$time, y=conf$frequency, 10*log10(conf$spec), xlab="Time (s)",
    #            ylab='Frequency (Hz)')
  # }

  return(list(spec = mt_spectrogram, time = stimes, frequency = sfreqs, configuration = conf))
}

hertz_to_erb_scale <- function(f) {
  21.4 * log10(1 + 4.37e-3 * f)
}

erb_scale_to_hertz <- function(erb_scale) {
  (10 ^ (erb_scale / 21.4) - 1) / 4.37e-3
}

hertz_to_erb <- function(f) {
  24.7 * (4.37 * f / 1000 + 1)
}

nextpow2 <- function(x) {
  ceiling(log2(x))
}

#' @title Apply gamma-tone filters to obtain auditory envelopes
#' @param x a numeric vector or matrix; if \code{x} is a matrix, it should
#' be column-major (each column is a sound track)
#' @param sample_rate sampling frequency
#' @param center_frequencies center frequencies at which the envelopes will
#' be derived; can be either a length of two defining the lower and
#' upper bound, and using \code{n_bands} to interpolate automatically, or
#' a length of multiple, with the frequencies specified explicitly
#' @param n_bands number of the center frequencies, can be missing if
#' \code{center_frequencies} is explicit and no interpolation is needed;
#' if specified, then the frequencies will be interpolated using
#' equivalent rectangular bandwidth rate (\code{'ERB'})
#' @param use_hilbert whether to apply 'Hilbert' transform; default is true,
#' which calculates the magnitude; set to false when only the filter is needed
#' @param downsample whether to down-sample the envelopes after the filters;
#' default is \code{NA} (no down-sample).
#' @param downsample_before_hilbert whether the down-sample happens before
#' or after the 'Hilbert' transform so speed up the computation if the signal
#' is too long; only used when \code{downsample} is greater than 1; default
#' is \code{FALSE}. Use with caution, especially when the voice center
#' frequency is close to the 'Nyquist' frequency. However, if used properly,
#' there will be significant performance boost on large signals with
#' high sampling rates
#' @returns A file-array object of filtered and potentially down-sampled
#' data; see 'Examples' on how to use this function.
#' @examples
#'
#'
#' fs <- 4000
#' time <- seq_len(8000) / fs
#' x <- sin(160 * pi * time) +
#'   sin(1000 * pi * time) * dnorm(time, mean = 1, sd = 0.1) +
#'   0.5 * rnorm(length(time))
#'
#' # envelope
#' result <- gammatone_fast(
#'   x,
#'   sample_rate = fs,
#'   center_frequencies = c(20, 1000),
#'   n_bands = 128,
#'
#'   # default downsample happens after hilbert
#'   downsample = 40
#' )
#'
#'
#' oldpar <- par(mfrow = c(2, 1))
#'
#' plot(
#'   time,
#'   x,
#'   type = "l",
#'   xlab = "Time",
#'   ylab = "",
#'   main = "Original mixed 80Hz and 500Hz"
#' )
#'
#' # only one channel
#' envelope <- subset(result, Channel ~ Channel == 1, drop = TRUE)
#' dnames <-  dimnames(envelope)
#' image(
#'   x = as.numeric(dnames$Time),
#'   y = as.numeric(dnames$Frequency),
#'   z = envelope,
#'   xlab = "Time",
#'   ylab = "Frequency",
#'   main = "Envelope from 20Hz to 1000Hz"
#' )
#'
#'
#'
#' par(oldpar) # reset graphics state
#'
#'
#' @export
gammatone_fast <- function(x, sample_rate, center_frequencies, n_bands,
                           use_hilbert = TRUE, downsample = NA,
                           downsample_before_hilbert = FALSE) {

  # x <- as.vector(ieegio::io_read_mat("~/rave_data/raw_dir/DemoSubject/008/DemoSubjectDatafile008_ch13.mat")$analogTraces)[]
  # sample_rate <- 2000
  # center_frequencies <- c(20, 1000)
  # n_bands <- 128
  # downsample <- 20
  # use_hilbert <- TRUE

  filter_order <- 4
  center_frequencies <- unname(as.double(center_frequencies))
  if (missing(n_bands)) {
    n_bands <- length(center_frequencies)
  } else {
    if (length(center_frequencies) != 2) {
      stop("`gammatone_fast`: `n_bands` is provided but length of `center_frequencies` is not two. Please either leave `n_bands` empty or set `center_frequencies` to be the frequency ranges.")
    }
    center_frequencies <- sort(center_frequencies)
    centr_erbs <- hertz_to_erb_scale(center_frequencies)
    centr_erbs <- seq(centr_erbs[[1]], centr_erbs[[2]], length.out = n_bands)
    center_frequencies <- erb_scale_to_hertz(centr_erbs)
  }

  gammatone_length <- 2 ^ nextpow2(0.128 * sample_rate)

  # Bandwidth factor for gammatone is 1.019
  erb <- hertz_to_erb(center_frequencies)
  b <- 1.019 * erb

  phase <- 0

  two_pi_per_timepoint <- 2 * pi / sample_rate

  # closed form (?) integral of impulse
  # gain <- ((1.019 * b * two_pi_per_timepoint) ^ filter_order) / 6

  time_seq <- (seq_len(gammatone_length) - 1L) / sample_rate

  # for each frequency center calculate IR
  tcs <- NULL
  gammatone_filters <- sapply(center_frequencies, function(f) {
    # f <- center_frequencies[[1]]
    erb <- hertz_to_erb(f)
    b <- 1.019 * erb
    gain <- ((1.019 * b * two_pi_per_timepoint) ^ filter_order) / 6
    # make sure no phase shift
    tc <- (filter_order - 1) / (2 * pi * b)
    tcs <<- c(tcs, tc)
    phase <- -2 * pi * f * tc

    gain * sample_rate^3 * time_seq^(filter_order - 1) * exp(-2 * pi * b * time_seq) * cos(2 * pi * f * time_seq + phase)
  })

  # Use fftfilt() but more efficient for large vectors
  l_filter <- gammatone_length
  x_is_mat <- is.matrix(x)
  if ( !x_is_mat ) {
    x <- matrix(x, ncol = 1)
  }
  l_x <- nrow(x)
  c_x <- ncol(x)
  N <- 2 ^ nextpow2(l_x + l_filter - 1)


  downsample <- as.double(downsample)
  final_sample_rate <- sample_rate

  n_timepoints <- l_x

  if (!is.na(downsample)) {
    if (downsample <= 1) {
      downsample <- NA_real_
    } else {
      n_timepoints <- ceiling(n_timepoints / downsample)
      final_sample_rate <- sample_rate / downsample
    }
  } else {
    downsample <- NA_real_
  }


  root_dir <- file.path(tempdir2(check = TRUE), "ravetools")
  digest <- digest::digest(list(
    center_frequencies = center_frequencies,
    filter_order = filter_order,
    N = N,
    sample_rate = sample_rate
  ))
  x_digest <- digest::digest(list(
    x_digest = digest::digest(x),
    filter_digest = digest,
    use_hilbert = use_hilbert,
    downsample = downsample
  ))

  # if the filters < 256MB
  if (!dir.exists(root_dir)) {
    dir.create(root_dir, showWarnings = FALSE, recursive = TRUE)
  }
  # use filearray to store the signals on disk
  filters_path <- file.path(root_dir, sprintf("gammatone-%s", digest))
  gammatone_filters_fft <- filearray::filearray_load_or_create(
    filebase = filters_path,
    dimension = c(N, n_bands),
    type = "complex",
    symlink_ok = FALSE,
    partition_size = 1L,
    initialize = TRUE,
    ready = TRUE,
    on_missing = function(arr) {
      lapply(seq_len(n_bands), function(ii) {
        arr[, ii] <- fft(postpad(gammatone_filters[, ii], N))
        NULL
      })
      arr$set_header("ready", TRUE)
      arr
    }
  )

  # delay
  delay <- round(tcs * sample_rate)

  # result <- apply(x, 2L, function(x_slice) {
  #   # print(ii <<- ii + 1)
  #   # x_slice <- x[,1]
  #   result <- stats::mvfft(gammatone_filters_fft[drop = FALSE] * fft(postpad(x_slice, N)), inverse = TRUE) / N
  #   result <- Re(result[seq_len(l_x), , drop = FALSE])
  #
  #   # offset delay
  #   result <- sapply(seq_len(n_bands), function(jj) {
  #     res <- result[, jj]
  #     res <- c(res[-seq_len(delay[[jj]])], res[seq_len(delay[[jj]])])
  #
  #     if (isTRUE(downsample > 1)) {
  #       # downsample the filtered results before hilbert or returning to speed up
  #       res <- decimate(res, q = downsample)
  #     }
  #     res
  #   })
  #
  #   if (use_hilbert) {
  #     # apply hilbert transform
  #     result <- abs(hilbert(result))
  #   }
  #
  #   result
  #
  # })
  # dim(result) <- c(l_x, n_bands, c_x)
  # result <- aperm(result, c(1, 3, 2))


  # if( N * n_bands * c_x <= 33554432 ) {
  #   # using around 2-3GB RAM so it's fine
  #   result <- apply(x, 2L, function(x_slice) {
  #     # print(ii <<- ii + 1)
  #     # x_slice <- x[,1]
  #     result <- stats::mvfft(gammatone_filters_fft[drop = FALSE] * fft(postpad(x_slice, N)), inverse = TRUE) / N
  #     result <- Re(result[seq_len(l_x), , drop = FALSE])
  #     if (use_hilbert) {
  #       # apply hilbert transform
  #       result <- abs(hilbert(result))
  #     }
  #     # offset delay
  #     result <- sapply(seq_len(n_bands), function(jj) {
  #       res <- result[, jj]
  #       c(res[-seq_len(delay[[jj]])], res[seq_len(delay[[jj]])])
  #     })
  #   })
  #   dim(result) <- c(l_x, n_bands, c_x)
  #   result <- aperm(result, c(1, 3, 2))
  #
  #   # samp <- result[, 1, ]
  #   # n_plot <- 10000
  #   # matplot(x = seq_len(n_plot) / sample_rate, y = samp[seq_len(n_plot), ], lty = 1, type = "l")
  #   # diagnose_channel(s1 = as.vector(x), s2 = rowSums(samp), srate = sample_rate, which = 1, col = c("black", "red"))
  #   # plot_signals(
  #   #   t(cbind(x, rowSums(samp))), sample_rate = sample_rate, space = 0, duration = 1
  #   # )
  #
  # } else {

  result_path <- file.path(root_dir, sprintf("gammatone-results-%s", x_digest))

  result <- filearray::filearray_load_or_create(
    filebase = result_path,
    dimension = c(n_timepoints, c_x, n_bands),
    type = "double",
    symlink_ok = FALSE,
    partition_size = 1L,
    initialize = FALSE,
    ready = FALSE,
    verbose = FALSE,
    on_missing = function(arr) {

      design_table <- expand.grid(
        x = seq_len(c_x),
        f = seq_len(n_bands)
      )

      fft_x <- filearray::as_filearray(stats::mvfft(postpad(x, N)))
      on.exit({
        fft_x$.mode <- "readwrite"
        fft_x$delete()
      })

      # Avoid using gsignal functions
      ravetools <- asNamespace("ravetools")

      lapply(seq_len(nrow(design_table)), function(row_ii) {
        # row_ii <- 1
        design_row <- design_table[row_ii, ]

        ii <- design_row$x
        jj <- design_row$f

        result <- stats::fft(fft_x[, ii] * gammatone_filters_fft[, jj], inverse = TRUE) / N
        result <- Re(result[seq_len(l_x)])

        # offset delay
        result <- c( result[-seq_len(delay[[jj]])], result[seq_len(delay[[jj]])] )

        # downsample
        # TODO: make sure n_timepoints is accurate and not with possible 1-off?
        if ( downsample_before_hilbert && isTRUE(downsample > 1) ) {
          result <- ravetools$decimate(x = result, q = downsample)
          if (length(result) > n_timepoints) {
            result <- result[seq_len(n_timepoints)]
          } else if (length(result) < n_timepoints) {
            result <- postpad(result, n_timepoints)
          }
        }

        if (use_hilbert) {
          # apply hilbert transform
          result <- abs(hilbert(result))
        }

        # downsample
        # TODO: make sure n_timepoints is accurate and not with possible 1-off?
        if ( !downsample_before_hilbert && isTRUE(downsample > 1) ) {
          result <- ravetools$decimate(x = result, q = downsample)
          if (length(result) > n_timepoints) {
            result <- result[seq_len(n_timepoints)]
          } else if (length(result) < n_timepoints) {
            result <- postpad(result, n_timepoints)
          }
        }

        arr[, ii, jj] <- result

        return()
      })

      arr$set_header("center_frequencies", center_frequencies, save = FALSE)
      arr$set_header("filter_order", as.integer(filter_order), save = FALSE)
      arr$set_header("orig_sample_rate", sample_rate, save = FALSE)
      arr$set_header("use_hilbert", use_hilbert, save = FALSE)
      arr$set_header("downsample", downsample, save = FALSE)

      time <- seq_len(n_timepoints) / sample_rate
      if (!is.na(downsample)) {
        time <- time * downsample
      }

      dimnames(arr) <- list(
        Time = time,
        Channel = seq_len(c_x),
        Frequency = center_frequencies
      )
      arr$set_header("ready", TRUE, save = TRUE)
      arr
    }
  )
  # }

  result


  # n_plot <- 1:min(10000, length(x))
  # samp <- result[n_plot, 1, ]
  # matplot(x = n_plot / sample_rate, y = samp, lty = 1, type = "l")
  # sp <- ravetools::plot_signals(
  #   as.vector(scale(as.vector(x)[n_plot], scale = FALSE)), sample_rate = sample_rate, space = 1, duration = 1
  # )
  # ravetools::plot_signals(
  #   rowMeans(samp) * 10, sample_rate = sample_rate, space = sp$space, space_mode = "absolute", duration = 1, new_plot = F,
  #   col = "red"
  # )
  # abline(h = sp$space , col = "gray", lty=2)
  # for(ii in 1:128) {
  #   ravetools::plot_signals(
  #     samp[, ii], sample_rate = sample_rate, space = sp$space, space_mode = "absolute", duration = 1, new_plot = F,
  #     col = ii+1
  #   )
  # }

}

# --------------------------------------------------------------------------- #
#  Catmull-Rom 3D spline – mirrors THREE.CatmullRomCurve3
#  Supports: centripetal (default), chordal, uniform (catmullrom + tension)
# --------------------------------------------------------------------------- #

# ---- cubic Hermite polynomial helpers (internal) --------------------------

# Coefficients of  f(t) = c0 + c1*t + c2*t^2 + c3*t^3
# with f(0)=x0, f(1)=x1, f'(0)=v0, f'(1)=v1.
crm_cubic_init <- function(x0, x1, v0, v1) {
  c(x0,
    v0,
    -3 * x0 + 3 * x1 - 2 * v0 - v1,
     2 * x0 - 2 * x1 +     v0 + v1)
}

crm_cubic_eval <- function(coef, t) {
  coef[[1L]] + t * (coef[[2L]] + t * (coef[[3L]] + t * coef[[4L]]))
}

# Non-uniform Barry-Goldman reparametrization
crm_nonuniform <- function(x0, x1, x2, x3, dt0, dt1, dt2) {
  v1 <- (x1 - x0) / dt0 - (x2 - x0) / (dt0 + dt1) + (x2 - x1) / dt1
  v2 <- (x2 - x1) / dt1 - (x3 - x1) / (dt1 + dt2) + (x3 - x2) / dt2
  crm_cubic_init(x1, x2, v1 * dt1, v2 * dt1)
}

# Classic uniform Catmull-Rom (tension-scaled)
crm_uniform <- function(x0, x1, x2, x3, tension) {
  crm_cubic_init(x1, x2, tension * (x2 - x0), tension * (x3 - x1))
}

# --------------------------------------------------------------------------- #

#' @title \verb{Catmull-Rom} 3D Spline Curve
#' @description
#' Creates a smooth \verb{Catmull-Rom} spline curve through a set of 3D key points.
#'
#' @param points numeric matrix with at least 2 rows and exactly 3 columns
#'   (\code{x}, \code{y}, \code{z}), giving the key (control) points through
#'   which the curve passes.
#' @param curve_type character; One of
#'   \code{"centripetal"} (default), \code{"chordal"}, or \code{"uniform"}.
#'   \code{"centripetal"} uses \eqn{\alpha = 0.5} (square-root of chord
#'   length), \code{"chordal"} uses \eqn{\alpha = 1} (chord length), and
#'   \code{"uniform"} is the classic formulation controlled by
#'   \code{tension}.
#' @param tension numeric scalar in \eqn{[0, 1]}; tangent scaling factor used
#'   only when \code{curve_type = "uniform"}. At \code{0.5} (default) the
#'   curve matches the standard \verb{Catmull-Rom} formulation.
#' @param closed logical; if \code{TRUE} the curve closes on itself by
#'   connecting the last point back to the first. Default is \code{FALSE}.
#'
#' @returns An object of class \code{"ravetools_curve"} (a list) with the
#'   following elements:
#' \describe{
#'   \item{\code{points}}{The input key-point matrix (\eqn{n \times 3}).}
#'   \item{\code{curve_type}}{Character, the parameterization type.}
#'   \item{\code{tension}}{Numeric, the tension value (relevant for
#'     \code{"uniform"} only).}
#'   \item{\code{closed}}{Logical, whether the curve is closed.}
#'   \item{\code{get_point}}{A \code{function(t)} that accepts a scalar
#'     \code{t} in \eqn{[0, 1]} and returns a named numeric vector
#'     on the curve.}
#'   \item{\code{get_points}}{A \code{function(n)} that returns an
#'     \eqn{n \times 3} matrix of \code{n} evenly spaced points along the
#'     curve, with column names \code{"x"}, \code{"y"}, \code{"z"}.}
#'   \item{\code{get_closest_t}}{A \code{function(query, coarse_n = 200L)}
#'     that, given a 3-element numeric vector \code{query} (\code{x}, \code{y},
#'     \code{z}), returns a list with elements \code{t} (the parameter value in
#'     \eqn{[0, 1]} of the nearest point), \code{point} (the closest point on
#'     the curve as a named numeric vector), and \code{distance} (Euclidean
#'     distance from \code{query} to the curve).  The search uses
#'     \code{coarse_n} uniform samples for an initial bracket followed by
#'     scalar optimization.}
#'   \item{\code{t_keypoints}}{Numeric vector of length \eqn{n} with the
#'     \code{t} parameter value where each key point lies on the curve.
#'     First element is always \code{0}, last is always \code{1}.}
#'   \item{\code{segment_lengths}}{Numeric vector of length \eqn{n-1} (open
#'     curve) or \eqn{n} (closed curve) containing the arc length of each
#'     spline segment, estimated by numerical integration.}
#' }
#'
#' @seealso \code{\link{print.ravetools_curve}},
#'   \code{\link{plot.ravetools_curve}}
#'
#' @examples
#'
#' pts <- matrix(c(
#'   -33.0534, -10.6213, -21.8328,
#'   -34.7526, -25.5089, -14.5390,
#'   -41.2002, -10.4606, -22.0032,
#'   -46.4717, -10.3567, -22.1134,
#'   -51.7431, -10.2528, -22.2237,
#'   -57.0146, -10.1488, -22.3339,
#'   -62.2860, -10.0449, -22.4442,
#'   -67.5575,  -9.9410, -22.5544
#' ), ncol = 3, byrow = TRUE)
#'
#' curve <- catmull_rom_3d(pts)
#' print(curve)
#'
#' # Sample 100 evenly spaced points along the curve
#' smooth <- curve$get_points(100)
#' head(smooth)
#'
#' # Evaluate the curve at t = 0.5 (midpoint)
#' curve$get_point(0.5)
#'
#' # get closest point on curve
#' curve$get_closest_t(c(-49, -10, -22))
#'
#' plot(curve, use_rgl = FALSE)
#'
#' @export
catmull_rom_3d <- function(
    points,
    curve_type = c("centripetal", "chordal", "uniform"),
    tension    = 0.5,
    closed     = FALSE
) {
  curve_type <- match.arg(curve_type)

  if (!is.matrix(points)) {
    points <- as.matrix(points)
  }
  stopifnot2(is.numeric(points),
             msg = "`points` must be a numeric matrix.")
  stopifnot2(ncol(points) == 3L,
             msg = "`points` must have exactly 3 columns (x, y, z).")
  n_pts <- nrow(points)
  stopifnot2(n_pts >= 2L,
             msg = "`points` must have at least 2 rows.")

  tension <- as.numeric(tension)[[1L]]
  stopifnot2(is.finite(tension) && tension >= 0 && tension <= 1,
             msg = "`tension` must be a single finite number in [0, 1].")

  closed <- isTRUE(closed)

  # Number of key points captured at construction time
  l <- n_pts

  # t parameter values at each key point
  # Open:   t_i = i / (l-1)  => t[1]=0, t[l]=1
  # Closed: t_i = i / l      => t[1]=0, t[l]=(l-1)/l
  t_keypoints <- seq.int(0L, l - 1L) / (if (closed) l else (l - 1L))

  # Helper: 0-based index i → 1-based row, wrapping with modulo
  get_ctrl <- function(i) {
    points[(i %% l) + 1L, ]
  }

  # ---- get_point -----------------------------------------------------------
  get_point <- function(t) {
    t <- as.numeric(t[[1L]])

    # Map t ∈ [0,1] to floating segment index + intra-segment weight
    p_raw <- (l - (if (closed) 0L else 1L)) * t
    seg   <- floor(p_raw)
    w     <- p_raw - seg

    # Open curve: clamp t == 1 exactly onto the last segment
    if (!closed && w == 0 && seg == l - 1L) {
      seg <- l - 2L
      w   <- 1
    }

    # Four bounding control points (ghost extrapolation at open endpoints)
    # Note: R's %% is always non-negative for positive modulus, so negative
    # indices wrap correctly without extra adjustment (unlike JS).
    if (closed || seg > 0L) {
      p0 <- get_ctrl(seg - 1L)
    } else {
      # Reflect: p0 = 2*p1 - p2
      p0 <- 2 * points[1L, ] - points[2L, ]
    }

    p1 <- get_ctrl(seg)
    p2 <- get_ctrl(seg + 1L)

    if (closed || (seg + 2L) < l) {
      p3 <- get_ctrl(seg + 2L)
    } else {
      # Reflect: p3 = 2*p_{n} - p_{n-1}
      p3 <- 2 * points[l, ] - points[l - 1L, ]
    }

    # Build cubic Hermite polynomials per axis
    if (curve_type == "centripetal" || curve_type == "chordal") {
      # pow applied to squared-distance:
      #   centripetal: (d^2)^0.25 = d^0.5  (alpha = 0.5)
      #   chordal    : (d^2)^0.5  = d      (alpha = 1)
      pow_exp <- if (curve_type == "chordal") 0.5 else 0.25

      dt0 <- sum((p1 - p0)^2)^pow_exp
      dt1 <- sum((p2 - p1)^2)^pow_exp
      dt2 <- sum((p3 - p2)^2)^pow_exp

      # Safety check for degenerate (duplicate / very close) points
      if (dt1 < 1e-4) dt1 <- 1.0
      if (dt0 < 1e-4) dt0 <- dt1
      if (dt2 < 1e-4) dt2 <- dt1

      cx <- crm_nonuniform(p0[[1L]], p1[[1L]], p2[[1L]], p3[[1L]], dt0, dt1, dt2)
      cy <- crm_nonuniform(p0[[2L]], p1[[2L]], p2[[2L]], p3[[2L]], dt0, dt1, dt2)
      cz <- crm_nonuniform(p0[[3L]], p1[[3L]], p2[[3L]], p3[[3L]], dt0, dt1, dt2)
    } else {
      # uniform (classic Catmull-Rom with tension)
      cx <- crm_uniform(p0[[1L]], p1[[1L]], p2[[1L]], p3[[1L]], tension)
      cy <- crm_uniform(p0[[2L]], p1[[2L]], p2[[2L]], p3[[2L]], tension)
      cz <- crm_uniform(p0[[3L]], p1[[3L]], p2[[3L]], p3[[3L]], tension)
    }

    c(x = crm_cubic_eval(cx, w),
      y = crm_cubic_eval(cy, w),
      z = crm_cubic_eval(cz, w))
  }

  # ---- get_points ----------------------------------------------------------
  get_points <- function(n = 50L) {
    n <- as.integer(n[[1L]])
    if (is.na(n) || n < 2L) {
      stop("`n` must be an integer >= 2.")
    }
    ts  <- seq.int(0L, n - 1L) / (n - 1L)
    mat <- matrix(0, nrow = n, ncol = 3L)
    for (i in seq_len(n)) {
      mat[i, ] <- get_point(ts[[i]])
    }
    colnames(mat) <- c("x", "y", "z")
    mat
  }

  # ---- closest point on curve to a query 3D point -------------------------
  get_closest_t <- function(query, coarse_n = 200L) {
    query    <- as.numeric(query)[seq_len(3L)]
    coarse_n <- max(as.integer(coarse_n[[1L]]), 2L)

    # Coarse pass: evaluate the curve uniformly and find the nearest sample
    pts_c  <- get_points(coarse_n)
    d2_c   <- rowSums(sweep(pts_c, 2L, query)^2)
    best_i <- which.min(d2_c)

    # Narrow the search bracket to ±1 coarse step around the best sample
    dt    <- 1.0 / (coarse_n - 1L)
    ts_c  <- seq.int(0L, coarse_n - 1L) / (coarse_n - 1L)
    lo    <- max(0.0, ts_c[[best_i]] - dt)
    hi    <- min(1.0, ts_c[[best_i]] + dt)

    # Fine pass: scalar optimization within the bracket
    res   <- stats::optimize(function(t) sum((get_point(t) - query)^2),
                             interval = c(lo, hi))
    t_opt <- res$minimum
    pt    <- get_point(t_opt)
    list(t = t_opt, point = pt, distance = sqrt(res$objective))
  }

  # ---- segment arc lengths (numerical integration, 50 sub-samples each) ---
  n_segs    <- if (closed) l else l - 1L
  t_seg_end <- c(t_keypoints[-1L], if (closed) 1.0)
  sub_n     <- 50L
  segment_lengths <- numeric(n_segs)
  for (s in seq_len(n_segs)) {
    ts_sub  <- seq(t_keypoints[[s]], t_seg_end[[s]], length.out = sub_n + 1L)
    pts_sub <- matrix(0, nrow = sub_n + 1L, ncol = 3L)
    for (j in seq_len(sub_n + 1L)) {
      pts_sub[j, ] <- get_point(ts_sub[[j]])
    }
    diffs <- diff(pts_sub)
    segment_lengths[[s]] <- sum(sqrt(rowSums(diffs * diffs)))
  }

  structure(
    list(
      points          = points,
      curve_type      = curve_type,
      tension         = tension,
      closed          = closed,
      t_keypoints     = t_keypoints,
      segment_lengths = segment_lengths,
      get_point       = get_point,
      get_points      = get_points,
      get_closest_t   = get_closest_t
    ),
    class = "ravetools_curve"
  )
}

# --------------------------------------------------------------------------- #
#  S3 methods
# --------------------------------------------------------------------------- #

#' @title Print method for \code{ravetools_curve}
#' @description Prints a concise summary of a \code{ravetools_curve} object
#' returned by \code{\link{catmull_rom_3d}}.
#' @param x an object of class \code{ravetools_curve}.
#' @param ... additional arguments passed to \code{\link{print.data.frame}}.
#' @return Invisibly returns \code{x}.
#' @export
print.ravetools_curve <- function(x, ...) {
  type_str <- x$curve_type
  if (x$curve_type == "uniform") {
    type_str <- sprintf("uniform (tension=%.2f)", x$tension)
  }
  cat(sprintf("<ravetools_curve: %d key points, type=%s%s>\n",
              nrow(x$points),
              type_str,
              if (x$closed) ", closed" else ""))
  cat("Key points (t / x / y / z / seg_length):\n")
  seg_len <- c(x$segment_lengths, NA_real_)
  print(data.frame(
    t          = x$t_keypoints,
    x          = x$points[, 1L],
    y          = x$points[, 2L],
    z          = x$points[, 3L],
    seg_length = seg_len
  ), ...)
  cat(sprintf("Total arc length: %.4g\n", sum(x$segment_lengths)))
  invisible(x)
}

#' @title Plot method for \code{ravetools_curve}
#' @description
#' Plots a \code{ravetools_curve} object created by
#' \code{\link{catmull_rom_3d}}.  When the \code{rgl} package is available
#' and \code{use_rgl = TRUE} (default), an interactive 3D scene is opened.
#' Otherwise three 2-D projection panels (\emph{x-y}, \emph{x-z},
#' \emph{y-z}) are drawn using base \R{} graphics.
#'
#' @param x an object of class \code{ravetools_curve}.
#' @param n integer; number of sample points used to draw the smooth curve.
#'   Default is \code{200L}.
#' @param col color for the spline curve line. Default \code{"steelblue"}.
#' @param pch,cex plotting character and scaling for the key control points
#'   (base-R fallback only). Default \code{pch = 19}, \code{cex = 1}.
#' @param use_rgl logical; if \code{TRUE} (default) and the \code{rgl}
#'   package is installed, an interactive 3D window is used. Set to
#'   \code{FALSE} to force the base-R 2D projection panels.
#' @param ... additional graphical parameters forwarded to the underlying
#'   plot calls.
#' @return Invisibly returns \code{x}.
#' @export
plot.ravetools_curve <- function(x, n = 200L, col = "steelblue",
                                 pch = 19L, cex = 1,
                                 use_rgl = TRUE, ...) {
  smooth <- x$get_points(as.integer(n))
  kp     <- x$points

  # ---- rgl 3D path --------------------------------------------------------
  if (isTRUE(use_rgl) && check_rgl(strict = FALSE)) {
    return(rgl_view({
      rgl_call("lines3d",
               smooth[, 1L], smooth[, 2L], smooth[, 3L],
               col = col, lwd = 2, ...)
      rgl_call("points3d",
               kp[, 1L], kp[, 2L], kp[, 3L],
               col = "orange", size = 8)
      rgl_call("axes3d")
      rgl_call("title3d",
               xlab = "x", ylab = "y", zlab = "z",
               main = sprintf("Catmull-Rom (%s)", x$curve_type))
    }))
  }

  # ---- base-R 2D projection fallback --------------------------------------
  op <- graphics::par(mfrow = c(1L, 3L), mar = c(4, 4, 3, 1))
  on.exit(graphics::par(op), add = TRUE)

  proj <- function(xi, yi, xlab, ylab) {
    graphics::plot(smooth[, xi], smooth[, yi],
                   type = "l", col = col, lwd = 2,
                   xlab = xlab, ylab = ylab,
                   main = sprintf("%s-%s projection", xlab, ylab), ...)
    graphics::points(kp[, xi], kp[, yi],
                     col = "orange", pch = pch, cex = cex)
  }

  proj(1L, 2L, "x", "y")
  proj(1L, 3L, "x", "z")
  proj(2L, 3L, "y", "z")

  invisible(x)
}

# Internal helper: build a sub-mesh from a set of face indices of orig_mesh.
# Vertex indices are remapped; orig_vertex stores the original column indices.
build_patch_mesh <- function(orig_mesh, face_indices) {
  vb <- orig_mesh$vb           # 4 x nv (homogeneous)
  it <- orig_mesh$it           # 3 x nf, 1-based
  patch_faces <- it[, face_indices, drop = FALSE]
  orig_v_idx  <- sort(unique(as.integer(patch_faces)))

  remap <- integer(ncol(vb))
  remap[orig_v_idx] <- seq_along(orig_v_idx)
  new_it <- matrix(remap[as.integer(patch_faces)], nrow = 3L)
  new_vb <- vb[, orig_v_idx, drop = FALSE]

  m <- list(vb = new_vb, it = new_it)
  if (!is.null(orig_mesh$normals))
    m$normals <- orig_mesh$normals[, orig_v_idx, drop = FALSE]
  m$orig_vertex <- orig_v_idx
  class(m) <- c("ravetools_mesh3d", "mesh3d")
  m
}

#' @title Split a mesh into two patches along a geodesic boundary
#' @description
#' Connects a set of surface \verb{waypoints} with geodesic paths to form a
#' closed boundary loop, then splits the mesh into the two regions created by
#' that loop.
#'
#' Consecutive \verb{waypoints} are joined by the geodesic (Dijkstra) shortest
#' path along the mesh surface; the last \verb{waypoint} connects back to the
#' first.
#'
#' @param mesh triangular mesh of class \code{'mesh3d'}.
#' @param waypoints numeric matrix with exactly 3 columns (\code{x}, \code{y},
#'   \code{z}) and at least 3 rows; each row is a 3-D coordinate defining a
#'   boundary corner. After any mesh subdivision, each coordinate is snapped to
#'   the nearest vertex in the (possibly refined) mesh. Consecutive snapped
#'   vertices are joined by geodesic paths to form the closed boundary loop.
#' @param seed_vertex integer (optional, 1-based). A vertex known to be inside
#'   the desired first patch. When \code{NULL} (default) the smaller of the
#'   two regions is returned first.
#' @param max_edge_length numeric (optional). When positive and finite, the mesh
#'   is refined before patching so that no edge exceeds this length. Global
#'   edge subdivision (\code{\link{vcg_subdivision}}) is applied repeatedly
#'   until the average edge length falls below the threshold (fast), then
#'   \code{\link{vcg_subdivide_max_edge_length}} handles any remaining outlier
#'   edges. \verb{Waypoint} coordinates are snapped to vertices after
#'   refinement, so finer vertices improve boundary accuracy.
#'   Default \code{NA} (no refinement).
#'
#' @returns A length-2 list of \code{mesh3d} objects. Each contains:
#' \describe{
#'   \item{\code{$orig_vertex}}{1-based integer vector: new vertex index \code{i}
#'     corresponds to column \code{orig_vertex[i]} of the original \code{mesh$vb}.}
#' }
#' The first element is the patch whose centroid is closest to the mean
#' \verb{waypoint} position; the second is the complementary connected patch.
#' On a multi-manifold mesh, disconnected components not adjacent to the
#' boundary loop appear in neither patch. When the boundary loop does not
#' divide the mesh (degenerate \verb{waypoints}), the second element is
#' \code{NULL}.
#'
#' @note All \verb{waypoints} must lie on the same connected component of the mesh.
#'   The mesh should be manifold; run \code{\link{vcg_fix_defects}} first
#'   if needed.
#'
#' @seealso \code{\link{dijkstras_surface_distance}}, \code{\link{surface_path}},
#'   \code{\link{vcg_fix_defects}}
#'
#' @examples
#'
#' mesh <- vcg_sphere()
#' mesh <- vcg_uniform_remesh(mesh)
#'
#' waypoints <- diag(1, 3)
#' patches <- vcg_mesh_patch(mesh, waypoints)
#'
#' plot_mesh_polygon(
#'   patches,
#'   col = list("red", 'gray'),
#'   alpha = list(1, 0.5),
#'   eye = c(10, 10, 10)
#' )
#'
#'
#' @export
vcg_mesh_patch <- function(mesh, waypoints, seed_vertex = NULL,
                            max_edge_length = NA) {
  mesh <- meshintegrity(mesh, facecheck = TRUE)

  waypoints <- as.matrix(waypoints)
  if (!is.numeric(waypoints) || ncol(waypoints) != 3L || nrow(waypoints) < 3L)
    stop("vcg_mesh_patch: `waypoints` must be a numeric matrix with 3 columns (x,y,z) and at least 3 rows")

  # ---- optional mesh refinement ----
  if (isTRUE(max_edge_length > 0) && is.finite(max_edge_length)) {
    avg_len <- vcg_average_edge_length(mesh)
    while (avg_len >= max_edge_length) {
      mesh    <- vcg_subdivision(mesh, "edge")
      avg_len <- vcg_average_edge_length(mesh)
    }
    mesh <- vcg_subdivide_max_edge_length(mesh, max_edge_len = max_edge_length)
  }

  vb <- mesh$vb[1:3, , drop = FALSE]
  it <- mesh$it          # 3 x nf, 1-based

  # ---- snap waypoint coordinates to nearest vertices (after subdivision) ----
  vb3 <- vb  # 3 x nv
  waypoint_verts <- apply(waypoints, 1L, function(coord) {
    which.min(colSums((vb3 - coord)^2))
  })
  if (anyDuplicated(waypoint_verts))
    warning("vcg_mesh_patch: two or more waypoints snapped to the same vertex; boundary may be degenerate")

  # dijkstras_surface_distance expects positions as N x 3 (rows = vertices)
  # and faces as M x 3 (rows = faces), 1-indexed
  positions <- t(vb)
  faces     <- t(it)

  # ---- build boundary loop via consecutive geodesic segments ----
  n_wp <- length(waypoint_verts)
  boundary_vertices <- integer(0)
  for (i in seq_len(n_wp)) {
    src <- waypoint_verts[i]
    dst <- waypoint_verts[if (i == n_wp) 1L else i + 1L]
    dist_res <- dijkstras_surface_distance(
      positions        = positions,
      faces            = faces,
      start_node       = src,
      face_index_start = 1L
    )
    seg <- surface_path(dist_res, target_node = dst)
    if (seg$path[1L] != src)
      stop(sprintf(
        "vcg_mesh_patch: cannot find geodesic path from waypoint %d (vertex %d) to waypoint %d (vertex %d). Are they on the same connected component?",
        i, src,
        if (i == n_wp) 1L else i + 1L, dst
      ))
    # Drop the last vertex (= first vertex of the next segment) to avoid duplicates
    boundary_vertices <- c(boundary_vertices, seg$path[-length(seg$path)])
  }

  # ---- resolve seed face (0-based for C++) ----
  seed_face <- -1L
  if (!is.null(seed_vertex)) {
    sv <- as.integer(seed_vertex)
    fc <- which(colSums(it == sv) > 0L)
    if (!length(fc))
      stop("vcg_mesh_patch: seed_vertex ", sv, " does not belong to any face")
    seed_face <- fc[1L] - 1L   # convert to 0-based
  }

  # ---- C++ BFS flood fill for fi1 ----
  fi1 <- vcgMeshPatchFaces(
    vb_          = vb,
    it_          = it - 1L,                  # 0-based faces
    boundary_seq = boundary_vertices - 1L,   # 0-based vertices
    seed_face    = seed_face
  )

  # ---- fi2: BFS from the other side of the first boundary edge ----
  # Using setdiff would incorrectly include disconnected components on a
  # multi-manifold mesh. Instead, seed the second BFS from a face adjacent
  # to the first boundary edge that is not in fi1.
  v0_1 <- boundary_vertices[1L]   # already 1-based
  v1_1 <- boundary_vertices[2L]
  seed2_candidates <- setdiff(which(colSums(it == v0_1) > 0L & colSums(it == v1_1) > 0L), fi1)
  fi2 <- if (length(seed2_candidates) == 0L) {
    integer(0)   # boundary lies on mesh boundary, or loop fully encloses the mesh
  } else {
    vcgMeshPatchFaces(
      vb_          = vb,
      it_          = it - 1L,
      boundary_seq = boundary_vertices - 1L,
      seed_face    = seed2_candidates[1L] - 1L   # 0-based
    )
  }

  # ---- construct sub-meshes ----
  p1 <- build_patch_mesh(mesh, fi1)
  p2 <- if (length(fi2) > 0L) build_patch_mesh(mesh, fi2) else NULL

  # ---- order: patch closest to mean waypoint coordinate comes first ----
  mean_wp  <- colMeans(waypoints)   # 3-vector from the n x 3 input matrix
  centroid <- function(m3d) {
    rowMeans(m3d$vb[1:3, , drop = FALSE])
  }
  d1 <- sqrt(sum((centroid(p1) - mean_wp)^2))
  if (!is.null(p2)) {
    d2 <- sqrt(sum((centroid(p2) - mean_wp)^2))
    if (d2 < d1) {
      tmp <- p1
      p1 <- p2
      p2 <- tmp
    }
  }

  list(p1, p2)
}

# Internal: apply user-supplied world-space clipping planes to a 3 x N matrix
# of points. Returns a logical keep mask of length N. `eye_world` is used
# only for planes whose `side` element is 0 (auto camera-facing). A NULL or
# zero-length `clipping_plane` returns an all-TRUE mask. A bare length-5
# numeric vector is accepted as shorthand for a one-plane list.
apply_clipping_planes <- function(points_world, eye_world, clipping_plane) {
  n_pts <- ncol(points_world)
  keep <- rep(TRUE, n_pts)
  if (!length(clipping_plane)) return(keep)
  if (is.numeric(clipping_plane) && !is.list(clipping_plane)) {
    clipping_plane <- list(clipping_plane)
  } else if (!is.list(clipping_plane)) {
    stop("`clipping_plane` must be NULL, a length-5 numeric vector, or a list of such vectors")
  }
  for (cp in clipping_plane) {
    cp <- as.numeric(cp)
    if (length(cp) != 5L || anyNA(cp)) {
      stop("`clipping_plane`: each plane must be a numeric vector of length 5 with no NA")
    }
    nrm <- cp[1:3]
    nlen <- sqrt(sum(nrm * nrm))
    if (!is.finite(nlen) || nlen == 0) {
      stop("`clipping_plane`: the first three elements (normal vector) must not be zero")
    }
    nrm <- nrm / nlen
    d <- cp[[4L]] / nlen
    sd <- cp[[5L]]
    if (!isTRUE(sd %in% c(-1, 0, 1))) {
      stop("`clipping_plane`: the fifth element must be -1, 0, or 1")
    }
    if (sd == 0) {
      signed_eye <- sum(nrm * eye_world) - d
      sd <- if (signed_eye >= 0) 1 else -1
    }
    signed_pts <- as.numeric(nrm %*% points_world) - d
    if (sd > 0) {
      keep <- keep & (signed_pts >= 0)
    } else {
      keep <- keep & (signed_pts <= 0)
    }
  }
  keep
}

# Internal: resolve auto-computed xlim/ylim with a `zoom` magnification that
# is aware of the current device's plot region (`par("pin")`) and the
# requested aspect ratio `asp`. When both `asp` and `pin` are usable, the
# visible data window after `asp`-induced expansion is shrunk by `zoom`, so
# the limiting axis is zoomed by exactly `zoom` and the other axis is
# zoomed less (filling extra margin) while preserving `asp`. When `asp` or
# `pin` is unavailable, falls back to a symmetric shrink by `zoom` around
# the data midpoint. User-supplied (non-NULL) `xlim`/`ylim` are passed
# through unchanged.
zoom_limits <- function(xlim, ylim, x_all, y_all, zoom, asp) {
  x_user <- length(xlim) > 0
  y_user <- length(ylim) > 0
  if (!x_user) xlim <- range(x_all, na.rm = TRUE)
  if (!y_user) ylim <- range(y_all, na.rm = TRUE)
  if (x_user && y_user) return(list(xlim = xlim, ylim = ylim))
  if (!(is.finite(zoom) && zoom > 0) || zoom == 1) {
    return(list(xlim = xlim, ylim = ylim))
  }

  dx <- xlim[2L] - xlim[1L]
  dy <- ylim[2L] - ylim[1L]
  mx <- mean(xlim)
  my <- mean(ylim)

  pin <- tryCatch(graphics::par("pin"), error = function(e) NULL)
  asp_ok <- length(asp) == 1L && is.finite(asp) && asp > 0
  pin_ok <- !is.null(pin) && length(pin) == 2L &&
            all(is.finite(pin)) && all(pin > 0)
  data_ok <- is.finite(dx) && is.finite(dy) && dx > 0 && dy > 0

  if (asp_ok && pin_ok && data_ok) {
    pw <- pin[[1L]]
    ph <- pin[[2L]]
    # Physical inches per x-data-unit when both axes fit (asp y/x ratio):
    #   scale_x = min(pw/dx, ph/(asp*dy))
    # Visible ranges: vis_dx = pw/scale_x, vis_dy = ph/(asp*scale_x).
    # Zooming in by `zoom` scales `scale_x` by `zoom`.
    scale_x <- min(pw / dx, ph / (asp * dy))
    new_dx <- pw / (scale_x * zoom)
    new_dy <- ph / (asp * scale_x * zoom)
  } else {
    new_dx <- dx / zoom
    new_dy <- dy / zoom
  }

  if (!x_user) xlim <- c(mx - new_dx / 2, mx + new_dx / 2)
  if (!y_user) ylim <- c(my - new_dy / 2, my + new_dy / 2)
  list(xlim = xlim, ylim = ylim)
}

#' @title Render one or more meshes as an orthographic dot cloud in base R
#' @description
#' Projects mesh vertices onto a 2D plane using an orthographic camera defined
#' by an eye position, a look-at point, and an up direction, then draws the
#' projected dots with base-R \code{\link[graphics]{plot}}.  Each dot is
#' rendered opaque, but its \code{cex} is modulated by a rim-light weight
#' \eqn{1 - |n \cdot z_{cam}|}: front- and back-facing vertices shrink toward
#' zero size, while grazing-edge vertices keep their full size.  This
#' size-modulated trick gives a rim-light look without paying R's per-point
#' transparency-blending cost.
#'
#' Multiple meshes share a single depth space: all vertices are projected
#' together and sorted globally by depth before a single \code{plot()} call,
#' so the painter's algorithm works correctly across meshes.  Meshes without
#' faces (point clouds) are rendered at full size and are not affected
#' by the \code{side} filter.
#'
#' @param mesh a \code{'mesh3d'} object, or a list of \code{'mesh3d'} objects.
#'   Meshes without a face matrix (\code{$it}) are treated as point clouds.
#'   If \code{mesh$normals} is absent and the mesh has faces, normals are
#'   recomputed with \code{\link{vcg_update_normals}}.
#' @param eye numeric vector of length 3 - camera position in world space.
#' @param lookat numeric vector of length 3 - the world-space point the camera
#'   is looking at.
#' @param up numeric vector of length 3 - a world-space vector indicating
#'   which direction is "up" for the camera; defaults to \code{c(0, 1, 0)}.
#' @param col base color(s) for the dots.  Accepted forms:
#'   \itemize{
#'     \item A single color string - applied to all meshes.
#'     \item A character vector of length 2 or more (not matching any vertex
#'       count) - used as a depth-ordered color gradient applied to all meshes.
#'     \item A character vector of length equal to the number of vertices in a
#'       single mesh - per-vertex colors for that mesh.
#'     \item A list of length equal to the number of meshes - each element is
#'       one of the above, applied to the corresponding mesh.
#'   }
#'   Default \code{"gray30"}.
#' @param pch point character; a scalar or vector/list with one value per mesh,
#'   recycled as necessary.  Default \code{16L}.
#' @param cex point expansion factor; a scalar or vector/list with one value
#'   per mesh, recycled as necessary.  Default \code{0.1}.
#' @param add logical; if \code{TRUE} the dots are added to an existing plot
#'   instead of opening a new one.  Default \code{FALSE}.
#' @param axes logical; whether to draw axes on a new plot.  Ignored when
#'   \code{add = TRUE}.  Default \code{FALSE}.
#' @param asp aspect ratio of the new plot; default \code{1} (equal scaling).
#'   Ignored when \code{add = TRUE}.
#' @param xlim,ylim axis limits for the new plot; \code{NULL} (default) lets R
#'   choose automatically from all meshes' projected vertices.  Ignored when
#'   \code{add = TRUE}.
#' @param zoom positive numeric magnification applied to the auto-computed
#'   axis limits when \code{xlim} or \code{ylim} is \code{NULL}.  Values
#'   greater than \code{1} zoom in, values in \code{(0, 1)} zoom out.
#'   When \code{asp} is set, the zoom is plot-region aware (queried via
#'   \code{par("pin")}): the axis that already fills the device after
#'   \code{asp}-induced expansion is zoomed by exactly \code{zoom}, and
#'   the other axis is zoomed less so the data fills more of the plot
#'   region while preserving the requested \code{asp}.  Ignored for any
#'   axis whose limit was supplied explicitly, and when \code{add = TRUE}.
#'   Default \code{1}.
#' @param xlab,ylab axis labels for the new plot.  Ignored when
#'   \code{add = TRUE}.
#' @param normal_weight passed to \code{\link{vcg_update_normals}} when normals
#'   must be recomputed; one of \code{"auto"} (area if no normals present,
#'   otherwise skip), \code{"area"}, or \code{"angle"}.  Default \code{"auto"}.
#' @param side which side of meshed surfaces to render.  One of \code{"front"}
#'   (default), \code{"back"}, or \code{"both"} (renders all vertices).  Point clouds are always
#'   rendered regardless of this setting.
#' @param mesh_clipping numeric in \code{[0, 1]} controlling how much of
#'   the surface is kept relative to the camera-facing direction.
#'   Vertices whose rim-light weight (\eqn{1 - |n \cdot z_{cam}|}) is at
#'   or below \code{1 - mesh_clipping} are dropped, leaving only the
#'   silhouette/grazing band.  \code{mesh_clipping = 1} (no clipping)
#'   keeps all vertices; smaller values peel away more of the
#'   front/back-facing dots.  Surviving vertices are drawn opaque, and
#'   their \code{cex} is multiplied by the rim-light weight so
#'   grazing-edge dots appear larger than near-front-facing dots.
#'   Drawing opaque points with size-modulated weight is dramatically
#'   faster than R's true per-point transparency.  Default \code{0.7}.
#' @param alpha numeric in \code{[0, 1]}; one value per mesh (recycled),
#'   sets the transparency of each mesh (\code{1} = fully opaque,
#'   \code{0} = fully transparent).  Vertices belonging to a mesh with
#'   \code{alpha = 0} are dropped entirely.  Default \code{1}.
#' @param clipping_plane optional list of world-space clipping planes used to
#'   hide parts of the scene.  Each plane is a numeric vector of length 5:
#'   the first three entries are the plane normal \eqn{(n_x, n_y, n_z)}
#'   (must be non-zero; normalized internally), the fourth is the signed
#'   distance from the world origin to the plane along that normal (so the
#'   plane equation is \eqn{n \cdot x = d}), and the fifth indicates which
#'   half-space is kept: \code{1} keeps the front side (the side the normal
#'   points to), \code{-1} keeps the back side, and \code{0} keeps whichever
#'   side currently faces the camera (auto-flipped per call based on
#'   \code{eye}).  Multiple planes are intersected.  Clipping is applied
#'   per vertex.  A single length-5 numeric vector is also accepted as
#'   shorthand for a one-plane list.  Default \code{NULL} (no clipping).
#' @param clipping_plane_enabled logical vector, one entry per mesh
#'   (recycled), controlling whether \code{clipping_plane} is applied to
#'   that mesh.  \code{TRUE} (default) means the mesh participates in
#'   clipping; \code{FALSE} exempts the mesh entirely (all of its vertices
#'   are kept regardless of the clipping planes).  Has no effect when
#'   \code{clipping_plane} is \code{NULL}.
#' @param ... additional graphical parameters forwarded to
#'   \code{\link[graphics]{plot.default}} (new plot) or
#'   \code{\link[graphics]{points}} (when \code{add = TRUE}).
#'
#' @return Invisibly returns a list with components \code{xlim} and
#'   \code{ylim} (the plot limits used).
#'
#' @examples
#'
#' mesh <- vcg_isosurface(left_hippocampus_mask)
#'
#' # Side view - rim-light shows the outline of the hippocampus
#' plot_mesh_dotcloud(
#'   mesh,
#'   eye = c(150, 0, 0),
#'   lookat = c(0, 0, 0),
#'   up = c(0, 0, 1),
#'   col = "steelblue",
#'   cex = 2
#' )
#'
#' # Two meshes: surface + electrode point cloud
#' n_elec <- 20
#' electrodes <- structure(
#'   list(vb = matrix(rnorm(3 * n_elec, sd = 5), 3, n_elec) +
#'          rowMeans(mesh$vb)[1:3]),
#'   class = "mesh3d"
#' )
#' plot_mesh_dotcloud(
#'   mesh  = list(mesh, electrodes),
#'   eye    = c(150, 0, 0),
#'   lookat = c(0, 0, 0),
#'   up     = c(0, 0, 1),
#'   col    = list("steelblue", "red"),
#'   pch    = c(16L, 17L),
#'   cex    = c(2, 1.2)
#' )
#'
#' @seealso \code{\link{vcg_update_normals}}, \code{\link{vcg_isosurface}}
#'
#' @inheritSection ensure_mesh3d Coercing Surface Inputs
#'
#' @export
plot_mesh_dotcloud <- function(
    mesh,
    eye = c(0, 0, 1000),
    lookat = c(0, 0, 0),
    up = c(0, 1, 0),
    col = c("white", "gray30"),
    pch = 16L,
    cex = 0.1,
    add = FALSE,
    axes = FALSE,
    asp = 1,
    xlim = NULL,
    ylim = NULL,
    zoom = 1,
    xlab = "",
    ylab = "",
    normal_weight = c("auto", "area", "angle"),
    side = c("front", "back", "both"),
    mesh_clipping = 0.7,
    alpha = 1,
    clipping_plane = NULL,
    clipping_plane_enabled = TRUE,
    ...
) {
  normal_weight <- match.arg(normal_weight)
  side <- match.arg(side)

  # DIPSAUS DEBUG START
  # mesh <- ieegio::read_surface("~/rave_data/raw_dir/yael_demo_001/rave-imaging/fs/surf/lh.pial")
  # normal_weight = "auto"
  # side <- "front"
  # list2env(envir = .GlobalEnv, list(eye = c(0, 0, 1),
  #                                   lookat = c(0, 0, 0),
  #                                   up = c(0, 1, 0),
  #                                   col = "gray30",
  #                                   pch = 16L,
  #                                   cex = 0.1,
  #                                   add = FALSE,
  #                                   axes = FALSE,
  #                                   asp = 1,
  #                                   xlim = NULL,
  #                                   ylim = NULL,
  #                                   xlab = "",
  #                                   ylab = ""))
  # DIPSAUS DEBUG END

  # ---- 0. Normalize mesh to a list --------------------------------------
  if (inherits(mesh, "mesh3d") || (!is.null(mesh$vb))) {
    mesh_list <- list(mesh)
  } else {
    mesh_list <- mesh
  }
  n_meshes <- length(mesh_list)

  # col: list -> per mesh; scalar/vector -> same col spec for every mesh
  if (is.list(col)) {
    col_list <- rep_len(col, n_meshes)
  } else {
    col_list <- rep(list(col), n_meshes)
  }
  # Normalize every color spec to #RRGGBB up front so downstream code never
  # needs col2rgb / adjustcolor again.
  col_list <- lapply(col_list, function(ci) {
    substr(grDevices::adjustcolor(ci), 1L, 7L)
  })

  # pch/cex/alpha: one value per mesh (recycled)
  pch_vec   <- rep_len(as.integer(unlist(pch)),   n_meshes)
  cex_vec   <- rep_len(as.numeric(unlist(cex)),   n_meshes)
  alpha_vec <- pmax(0, pmin(1, rep_len(as.numeric(unlist(alpha)), n_meshes)))
  clip_enabled_vec <- rep_len(as.logical(clipping_plane_enabled), n_meshes)
  clip_enabled_vec[is.na(clip_enabled_vec)] <- TRUE

  # ---- 1. View matrix: world -> camera (R6 Vector3 / Matrix4) -----------
  # look_at builds R_c2w (camera-to-world, rotation only, three.js convention).
  # invert() == transpose() for a pure rotation, giving R_w2c.
  # set_position() adds the translation column: t = -R_w2c * eye.
  eye_v3    <- as_vector3(eye[1:3])
  lookat_v3 <- as_vector3(lookat[1:3])
  up_v3     <- as_vector3(up[1:3])

  rot_w2c <- Matrix4$new()$look_at(eye_v3, lookat_v3, up_v3)$invert()
  t_vec   <- Vector3$new()$from_array(eye[1:3])$apply_matrix4(rot_w2c)$multiply_scalar(-1)
  view    <- rot_w2c$clone2()$set_position(t_vec)

  # ---- 2. Process each mesh: extract geometry and normals ---------------
  # Only the per-mesh normal recomputation needs a loop; everything else
  # below is vectorized over the concatenated arrays.
  mesh_data <- vector("list", n_meshes)

  for (i in seq_len(n_meshes)) {
    m   <- ensure_mesh3d(mesh_list[[i]])
    vb  <- m$vb
    if (nrow(vb) > 3L) vb <- vb[1:3, , drop = FALSE]
    n_v <- ncol(vb)

    has_faces <- !is.null(m$it) && length(m$it) > 0L

    if (has_faces) {
      nw <- if (normal_weight == "auto") "area" else normal_weight
      if (is.null(m$normals) || normal_weight != "auto") {
        m <- vcg_update_normals(m, weight = nw)
      }
      nm <- m$normals
      if (nrow(nm) > 3L) nm <- nm[1:3, , drop = FALSE]
    } else {
      # Placeholder normals for point clouds (zero columns - facing and rim
      # weight are overridden via the has_faces_per_vert mask below).
      nm <- matrix(0, nrow = 3L, ncol = n_v)
    }

    mesh_data[[i]] <- list(
      vb        = vb,
      normals   = nm,
      has_faces = has_faces,
      n_verts   = n_v,
      pch       = pch_vec[[i]],
      cex       = cex_vec[[i]],
      col       = col_list[[i]]
    )
  }

  # ---- 3. Concatenate everything, then run vectorized ops ---------------
  all_vb      <- do.call(cbind, lapply(mesh_data, `[[`, "vb"))      # 3 x N
  all_normals <- do.call(cbind, lapply(mesh_data, `[[`, "normals")) # 3 x N
  n_verts_vec <- vapply(mesh_data, `[[`, 0L, "n_verts")
  total_n     <- sum(n_verts_vec)

  # per-vertex has_faces mask (TRUE for vertices belonging to a meshed surface)
  has_faces_per_vert <- rep(
    vapply(mesh_data, `[[`, logical(1), "has_faces"),
    n_verts_vec
  )

  # Single view transform for ALL vertices
  cam_pts <- Vector3$new()$from_array(all_vb)$apply_matrix4(view)
  x_all     <- cam_pts[1L, ]
  y_all     <- cam_pts[2L, ]
  depth_all <- -cam_pts[3L, ]    # positive = in front of camera

  # Single direction transform for ALL normals (zero columns for point
  # clouds are fine - they get overridden by the has_faces mask).
  cam_normals <- Vector3$new()$from_array(all_normals)$transform_direction(view)
  facing_all  <- cam_normals[3L, ]   # cam Z = -fwd component

  # Rim-light weight: 1 at grazing edges, 0 at front/back.
  # Point clouds: rim_weight = 1 (override) so they render at full size.
  rim_weight <- pmax(0, pmin(1, 1 - abs(facing_all)))
  rim_weight[!has_faces_per_vert] <- 1.0
  facing_all[!has_faces_per_vert] <- NA_real_

  # pch / cex / alpha: replicate the per-mesh scalar n_v times
  pch_all   <- rep(pch_vec,   n_verts_vec)
  cex_all   <- rep(cex_vec,   n_verts_vec)
  alpha_all <- rep(alpha_vec, n_verts_vec)

  # ---- 4. Per-mesh base colors (only the color step needs per-mesh logic
  # because length(col_i) is compared to that mesh's n_v) -----------------
  # Normalized depth used for gradient colors (global across all meshes).
  dr      <- range(depth_all, na.rm = TRUE)
  dr_span <- dr[2L] - dr[1L]
  if (dr_span < .Machine$double.eps) {
    depth_norm_all <- rep(0.5, total_n)
  } else {
    depth_norm_all <- (depth_all - dr[1L]) / dr_span
  }

  col_all <- character(total_n)
  offsets <- cumsum(c(0L, n_verts_vec))

  for (i in seq_len(n_meshes)) {
    n_v <- n_verts_vec[[i]]
    if (n_v == 0L) next
    idx   <- seq.int(offsets[[i]] + 1L, offsets[[i + 1L]])
    col_i <- col_list[[i]]

    if (length(col_i) == n_v) {
      # Per-vertex colors (already #RRGGBB from upfront normalization)
      base_col <- col_i
    } else if (length(col_i) == 1L) {
      base_col <- rep(col_i[[1L]], n_v)
    } else {
      # Multi-color gradient mapped to global depth of this mesh's vertices.
      # colorRampPalette returns #RRGGBB strings already.
      ramp     <- grDevices::colorRampPalette(rev(col_i))(256L)
      base_col <- ramp[ceiling(depth_norm_all[idx] * 255) + 1L]
    }
    col_all[idx] <- base_col
  }

  # Append per-vertex alpha to the #RRGGBB strings -> #RRGGBBAA.
  # Skip the paste entirely when every mesh is fully opaque so the
  # graphics device can take the no-alpha fast path.
  if (any(alpha_vec < 1)) {
    alpha_hex <- sprintf("%02X", as.integer(round(alpha_all * 255)))
    col_all   <- paste0(col_all, alpha_hex)
  }

  # ---- 5. mesh_clipping cull + side filter (point clouds always pass) ---
  keep <- !is.na(rim_weight) & (rim_weight >= (1 - mesh_clipping)) & (alpha_all > 0)
  if (side == "front") {
    keep <- keep & (is.na(facing_all) | facing_all > 0)
  } else if (side == "back") {
    keep <- keep & (is.na(facing_all) | facing_all < 0)
  }

  # User-supplied world-space clipping planes (per-vertex cull). Meshes
  # with `clipping_plane_enabled = FALSE` are exempted from this cull.
  if (length(clipping_plane)) {
    cp_keep <- apply_clipping_planes(all_vb, eye[1:3], clipping_plane)
    clip_enabled_all <- rep(clip_enabled_vec, n_verts_vec)
    cp_keep[!clip_enabled_all] <- TRUE
    keep <- keep & cp_keep
  }

  # ---- 6. Open canvas + screen limits -----------------------------------
  # Call plot.new() before zoom_limits so par("pin") reflects the true
  # plot-region dimensions of this device (margins applied). Then
  # plot.window() locks in the zoomed limits + asp before drawing.
  if (!add) {
    graphics::plot.new()
  }
  lim <- zoom_limits(xlim, ylim, x_all, y_all, zoom, asp)
  xlim <- lim$xlim
  ylim <- lim$ylim
  if (!add) {
    graphics::plot.window(xlim = xlim, ylim = ylim, asp = asp, ...)
    if (axes) {
      graphics::axis(1)
      graphics::axis(2)
      graphics::box()
    }
    if (nzchar(xlab) || nzchar(ylab)) {
      graphics::title(xlab = xlab, ylab = ylab)
    }
  }

  # ---- 7. Filter, painter's sort (far first), fold rim weight into cex --
  # Dots are drawn at full color; cex is scaled by the per-vertex rim
  # weight so grazing-edge dots remain visually heavier than near
  # front-facing dots without paying R's slow transparency-blending cost.
  x_k   <- x_all[keep]
  y_k   <- y_all[keep]
  col_k <- col_all[keep]
  pch_k <- pch_all[keep]
  cex_k <- cex_all[keep] * rim_weight[keep]

  ord   <- order(depth_all[keep], decreasing = TRUE)
  x_k   <- x_k[ord]
  y_k   <- y_k[ord]
  col_k <- col_k[ord]
  pch_k <- pch_k[ord]
  cex_k <- cex_k[ord]

  # ---- 8. Draw points ---------------------------------------------------
  graphics::points(x_k, y_k, col = col_k, pch = pch_k, cex = cex_k, ...)

  invisible(list(xlim = xlim, ylim = ylim))
}

#' @title Calculate Contrasts of Arrays in Different Methods
#' @description Provides seven methods to baseline an array and calculate
#' contrast.
#' @param x array (tensor) to calculate contrast
#' @param along_dim integer range from 1 to the maximum dimension of \code{x}.
#' baseline along this dimension, this is usually the time dimension.
#' @param baseline_indexpoints integer vector, which index points are counted
#' into baseline window? Each index ranges from 1 to \code{dim(x)[[along_dim]]}.
#' See Details.
#' @param unit_dims integer vector, baseline unit: see Details.
#' @param method character, baseline method; one of \code{"percentage"},
#' \code{"sqrt_percentage"}, \code{"decibel"}, \code{"zscore"},
#' \code{"sqrt_zscore"}, \code{"db_zscore"}, or \code{"subtract_mean"}.
#' See Details.
#' @param baseline_subarray sub-arrays that should be used to calculate
#' baseline; default is \code{NULL} (automatically determined by
#' \code{baseline_indexpoints}).
#' @param ... passed to other methods
#'
#' @returns Contrast array with the same dimension as \code{x}.
#' @details
#' Consider a scenario where we want to baseline a bunch of signals recorded
#' from different locations. For each location, we record \code{n} sessions.
#' For each session, the signal is further decomposed into frequency-time
#' domain. In this case, we have the input \code{x} in the following form:
#' \deqn{session \times frequency \times time \times location}
#' Now we want to calibrate signals for each session, frequency and location
#' using the first 100 time points as baseline points, then the code will be
#' \code{baseline_array(x, along_dim=3, baseline_window=1:100, unit_dims=c(1,2,4))}
#' \code{along_dim=3} is dimension of time, in this case, it's the
#' third dimension of \code{x}. \code{baseline_indexpoints=1:100}, meaning
#' the first 100 time points are used to calculate baseline.
#' \code{unit_dims} defines the unit signal. Its value \code{c(1,2,4)}
#' means the unit signal is per session (first dimension), per frequency
#' (second) and per location (fourth).
#'
#' In some other cases, we might want to calculate baseline across frequencies
#' then the unit signal is \eqn{frequency x time}, i.e. signals that share the
#' same session and location also share the same baseline. In this case,
#' we assign \code{unit_dims=c(1,4)}.
#'
#' There are seven baseline methods. They fit for different types of data.
#' Denote \eqn{z} is a unit signal and \eqn{z_0} is its baseline slice. Then
#' these baseline methods are:
#'
#' \describe{
#' \item{\code{"percentage"}}{
#' \deqn{
#'   \frac{z - \bar{z_{0}}}{\bar{z_{0}}} \times 100\%
#' }{
#'   (z / mean(z_0) - 1) x 100\%
#' }
#' }
#' \item{\code{"sqrt_percentage"}}{
#' \deqn{
#'   \frac{\sqrt{z} - \bar{\sqrt{z_{0}}}}{\bar{\sqrt{z_{0}}}} \times 100\%
#' }{
#'   (sqrt(z) / mean(sqrt(z_0)) - 1) x 100\%
#' }
#' }
#' \item{\code{"decibel"}}{
#' \deqn{
#'   10 \times ( \log_{10}(z) - \bar{\log_{10}(z_{0})} )
#' }{
#'   10 * ( log10(z) - mean( log10(z_0) ) )
#' }
#' }
#' \item{\code{"zscore"}}{
#' \deqn{
#'   \frac{z-\bar{z_{0}}}{sd(z_{0})}
#' }{
#'   (z - mean( z_0 )) / sd( z_0 )
#' }
#' }
#' \item{\code{"sqrt_zscore"}}{
#' \deqn{
#'   \frac{\sqrt{z}-\bar{\sqrt{z_{0}}}}{sd(\sqrt{z_{0}})}
#' }{
#'   (sqrt(z) - mean( sqrt(z_0) )) / sd( sqrt(z_0) )
#' }
#' }
#' \item{\code{"db_zscore"}}{
#' Z-score applied in the decibel (10log10) domain:
#' \deqn{
#'   \frac{10\log_{10}(z) - \bar{10\log_{10}(z_{0})}}{sd(10\log_{10}(z_{0}))}
#' }{
#'   ( log10(z) - mean( log10(z_0) ) ) / sd( log10(z_0) )
#' }
#' }
#' \item{\code{"subtract_mean"}}{
#' Simple mean subtraction with no scaling:
#' \deqn{
#'   z - \bar{z_{0}}
#' }{
#'   z - mean(z_0)
#' }
#' }
#'
#'
#' }
#'
#' @examples
#'
#' # Set ncores = 2 to comply to CRAN policy. Please don't run this line
#' ravetools_threads(n_threads = 2L)
#'
#'
#' library(ravetools)
#' set.seed(1)
#'
#' # Generate sample data
#' dims = c(10,20,30,2)
#' x = array(rnorm(prod(dims))^2, dims)
#'
#' # Set baseline window to be arbitrary 10 timepoints
#' baseline_window = sample(30, 10)
#'
#' # ----- baseline percentage change ------
#'
#' # Using base functions
#' re1 <- aperm(apply(x, c(1,2,4), function(y) {
#'   m <- mean(y[baseline_window])
#'   (y/m - 1) * 100
#' }), c(2,3,1,4))
#'
#' # Using ravetools
#' re2 <- baseline_array(x, 3, c(1,2,4),
#'                       baseline_indexpoints = baseline_window,
#'                       method = 'percentage')
#'
#' # Check different, should be very tiny (double precisions)
#' range(re2 - re1)
#'
#' \donttest{
#' # Check speed for large dataset, might take a while to profile
#'
#' ravetools_threads(n_threads = -1)
#'
#' dims <- c(200,20,300,2)
#' x <- array(rnorm(prod(dims))^2, dims)
#' # Set baseline window to be arbitrary 10 timepoints
#' baseline_window <- seq_len(100)
#' f1 <- function() {
#'   aperm(apply(x, c(1,2,4), function(y) {
#'     m <- mean(y[baseline_window])
#'     (y/m - 1) * 100
#'   }), c(2,3,1,4))
#' }
#' f2 <- function() {
#'   # equivalent as bl = x[,,baseline_window, ]
#'   #
#'   baseline_array(x, along_dim = 3,
#'                  baseline_indexpoints = baseline_window,
#'                  unit_dims = c(1,2,4), method = 'percentage')
#' }
#' range(f1() - f2())
#' microbenchmark::microbenchmark(f1(), f2(), times = 10L)
#'
#' }
#'
#'
#'
#' @export
baseline_array <- function(x, along_dim, unit_dims = seq_along(dim(x))[-along_dim], ...) {

  dims <- dim(x)
  along_dim <- as.integer(along_dim)
  unit_dims <- as.integer(unit_dims)

  stopifnot2(along_dim >= 1 && along_dim <= length(dims), msg = paste0(
    sQuote("along_dim"), " is invalid, it must be an integer from 1 to ",
    length(dims)
  ))

  # unit_dims is baseline unit,
  stopifnot2(!any(is.na(unit_dims)), msg = paste0(sQuote("unit_dims"), " contains NAs"))
  stopifnot2(all(unit_dims %in% seq_along(dims)), msg = paste0(sQuote("unit_dims"), " has invalid dimensions"))
  stopifnot2(!along_dim %in% unit_dims, msg = paste0(
    sQuote("along_dim"), " cannot be inside of ", sQuote("unit_dims")))
  unit_dims <- sort(unit_dims)

  UseMethod("baseline_array")

}

#' @rdname baseline_array
#' @export
baseline_array.array <- function(
    x, along_dim, unit_dims = seq_along(dim(x))[-along_dim],
    method = c("percentage", "sqrt_percentage", "decibel", "zscore",
               "sqrt_zscore", "db_zscore", "subtract_mean"),
    baseline_indexpoints = NULL, baseline_subarray = NULL, ...) {

  along_dim <- as.integer(along_dim)
  unit_dims <- as.integer(unit_dims)
  method <- match.arg(method)
  method_int <- which(c(
    "percentage", "sqrt_percentage", "decibel", "zscore", "sqrt_zscore",
    "db_zscore", "subtract_mean") == method)

  dims <- dim(x)
  ntimepoints <- dims[along_dim]

  stopifnot2(!(is.null(baseline_indexpoints) && is.null(baseline_subarray)),
             msg = "Either `baseline_indexpoints` or `baseline_subarray` must be specified")
  if (!is.null(baseline_indexpoints)) {
    baseline_indexpoints <- as.integer(baseline_indexpoints)
    # calculate baseline window
    baseline_indexpoints <- baseline_indexpoints[!is.na(baseline_indexpoints) & baseline_indexpoints > 0 & baseline_indexpoints <= ntimepoints]
    baseline_indexpoints <- sort(baseline_indexpoints)
    stopifnot2(length(baseline_indexpoints) > 0, msg = paste0("Baseline window is invalid: cannot find any valid time points. \nPlease makesure ", sQuote("baseline_indexpoints"), " contains at least one integer from 1 to ", ntimepoints))
    call_args <- lapply(seq_along(dims), function(ii) {
      if (ii == along_dim) {
        return(quote(baseline_indexpoints))
      }
      .missing_arg[[1]]
    })
    call_args <- c(list(quote(x)), call_args, list(drop = FALSE))
    bl <- do.call("[", call_args)
    bldims <- dim(bl)
  } else {
    stopifnot2(
      is.numeric(baseline_subarray) && is.array(baseline_subarray) &&
        length(dim(baseline_subarray)) == length(dims),
      msg = "`baseline_subarray` must be an array with same margin size as `x`"
    )
    bldims2 <- dim(baseline_subarray)
    bldims <- dims
    bldims[[along_dim]] <- bldims2[[along_dim]]
    bldims <- as.integer(bldims)
    stopifnot2(all(bldims2 == bldims),
               msg = "`baseline_subarray` dimension is invalid")
    bl <- baseline_subarray
  }

  rest <- seq_along(dims)[-unit_dims]
  #
  # if(!all(c(1, 3, 4) %in% unit_dims)) {
  #   bl <- ravetools::collapse(bl, keep = unit_dims)
  # }
  #
  #
  # dim(bl2) <- c(20,1,1,1)
  # bldims2 <- dim(bl2)
  # re <- .Call(`_ravetools_baselineArray`, x, bl2, dims, bldims2, along_dim - 1L, unit_dims - 1L, rest - 1L, method_int - 1L)

  re <- .Call(`_ravetools_baselineArray`, x, bl, dims, bldims, along_dim - 1L, unit_dims - 1L, rest - 1L, method_int - 1L)
  if (inherits(re, "ravetools_error")) {
    stop(re)
  }
  re
}

test_that("Baseline", {
  set.seed(0)
  dm <- c(5, 40, 10, 6)
  x <- array(rnorm(prod(dm))^2, dm)

  # trial-baseline
  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "percentage"
  )
  exp <- aperm(apply(x, c(1, 3, 4), function(slice) {
    (slice / mean(slice[c(20:30)]) - 1) * 100
  }), c(2, 1, 3, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_percentage"
  )
  exp <- aperm(apply(x, c(1, 3, 4), function(slice) {
    slice <- sqrt(slice)
    (slice / mean(slice[c(20:30)]) - 1) * 100
  }), c(2, 1, 3, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "decibel"
  )
  exp <- aperm(apply(x, c(1, 3, 4), function(slice) {
    slice <- 10 * log10(slice)
    slice - mean(slice[c(20:30)])
  }), c(2, 1, 3, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "zscore"
  )
  exp <- aperm(apply(x, c(1, 3, 4), function(slice) {
    bl <- slice[c(20:30)]
    (slice - mean(bl)) / sd(bl)
  }), c(2, 1, 3, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_zscore"
  )
  exp <- aperm(apply(x, c(1, 3, 4), function(slice) {
    slice <- sqrt(slice)
    bl <- slice[c(20:30)]
    (slice - mean(bl)) / sd(bl)
  }), c(2, 1, 3, 4))
  expect_equal(act, exp)

  # trial-baseline, also with zero values

  x <- array(0, dm)
  exp <- x
  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "percentage"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_percentage"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "decibel"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x + 1,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "zscore"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x + 1,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_zscore"
  )
  expect_equal(act, exp)


  # global baseline
  set.seed(0)
  dm <- c(5, 40, 10, 6)
  x <- array(rnorm(prod(dm))^2, dm)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "percentage"
  )
  exp <- aperm(array(apply(x, c(1, 4), function(slice) {
    (slice / mean(slice[c(20:30), ]) - 1) * 100
  }), dm[c(2, 3, 1, 4)]), c(3, 1, 2, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_percentage"
  )
  exp <- aperm(array(apply(x, c(1, 4), function(slice) {
    slice <- sqrt(slice)
    (slice / mean(slice[c(20:30), ]) - 1) * 100
  }), dm[c(2, 3, 1, 4)]), c(3, 1, 2, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "decibel"
  )
  exp <- aperm(array(apply(x, c(1, 4), function(slice) {
    slice <- 10 * log10(slice)
    (slice - mean(slice[c(20:30), ]))
  }), dm[c(2, 3, 1, 4)]), c(3, 1, 2, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "zscore"
  )
  exp <- aperm(array(apply(x, c(1, 4), function(slice) {
    bl <- slice[c(20:30), ]
    (slice - mean(bl)) / sd(bl)
  }), dm[c(2, 3, 1, 4)]), c(3, 1, 2, 4))
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_zscore"
  )
  exp <- aperm(array(apply(x, c(1, 4), function(slice) {
    slice <- sqrt(slice)
    bl <- slice[c(20:30), ]
    (slice - mean(bl)) / sd(bl)
  }), dm[c(2, 3, 1, 4)]), c(3, 1, 2, 4))
  expect_equal(act, exp)

  # global-baseline, also with zero values

  x <- array(0, dm)
  exp <- x
  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "percentage"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_percentage"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "decibel"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x + 1,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "zscore"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x + 1,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = c(20:30),
    method = "sqrt_zscore"
  )
  expect_equal(act, exp)
})

test_that("Baseline - db_zscore and guard conditions", {

  # ── Normal cases ──────────────────────────────────────────────────────────

  set.seed(0)
  dm <- c(5, 40, 10, 6)
  x <- array(rnorm(prod(dm))^2, dm)

  # trial baseline
  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = 20:30,
    method = "db_zscore"
  )
  exp <- aperm(apply(x, c(1, 3, 4), function(slice) {
    bl <- log10(slice[20:30])
    (log10(slice) - mean(bl)) / sd(bl)
  }), c(2, 1, 3, 4))
  expect_equal(act, exp)

  # global baseline
  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = 20:30,
    method = "db_zscore"
  )
  exp <- aperm(array(apply(x, c(1, 4), function(slice) {
    bl <- log10(slice[20:30, ])
    (log10(slice) - mean(bl)) / sd(bl)
  }), dm[c(2, 3, 1, 4)]), c(3, 1, 2, 4))
  expect_equal(act, exp)

  # ── db_zscore guard: all-zero baseline (bl_sd = 0) ───────────────────────

  x <- array(0, dm)
  exp <- x

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = 20:30,
    method = "db_zscore"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = 20:30,
    method = "db_zscore"
  )
  expect_equal(act, exp)

  # ── db_zscore guard: constant baseline (bl_sd = 0) ───────────────────────
  # All baseline values identical (> 0) -> sd(log10(bl)) = 0 -> output 0

  x <- array(1, dm)
  exp <- array(0, dm)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = 20:30,
    method = "db_zscore"
  )
  expect_equal(act, exp)

  act <- ravetools::baseline_array(
    x = x,
    along_dim = 2,
    unit_dims = c(1, 4),
    baseline_indexpoints = 20:30,
    method = "db_zscore"
  )
  expect_equal(act, exp)

  # ── decibel guard fix: geometric mean of baseline = 1 ────────────────────
  # mean(log10(bl)) == 0 but there are valid baseline values.
  # Old guard (bl_mean == 0.0) would incorrectly zero out all output.
  # New guard (bl_valid_len == 0) correctly computes 10*(log10(x) - 0).

  dm_small <- c(1, 10, 1, 1)
  # Baseline points 1-4: c(10, 0.1, 10, 0.1) -> log10 values c(1,-1,1,-1) -> mean = 0
  x_small <- array(c(10, 0.1, 10, 0.1, 10, 100, 1000, 0.1, 0.01, 0.001), dm_small)
  exp_small <- array(10 * log10(as.vector(x_small)), dm_small)

  act_small <- ravetools::baseline_array(
    x = x_small,
    along_dim = 2,
    unit_dims = c(1, 3, 4),
    baseline_indexpoints = 1:4,
    method = "decibel"
  )
  expect_equal(act_small, exp_small)
})

# Documentation-only file. The functions below are defined in
# R/RcppExports.R (auto-generated). This file adds roxygen documentation
# and `@export` tags so the wrappers appear in the package NAMESPACE.

#' Low-level \pkg{FFTW3} wrappers
#'
#' @name fftw-internal
#' @aliases fftw_r2c fftw_c2r fftw_c2c mvfftw_r2c mvfftw_c2c mvfftw_c2r
#' @aliases fftw_r2c_2d fftw_c2c_2d fftw_r2c_3d fftw_c2c_3d
#'
#' @description
#' Thin R bindings around the \pkg{FFTW3} library. These are
#' \strong{low-level} routines exposed primarily for advanced users and other
#' packages that need maximum throughput. They perform minimal input checking
#' and follow \pkg{FFTW} conventions (e.g. unnormalized inverse transforms,
#' one-sided real-to-complex spectra). For most user code prefer
#' \code{\link[stats]{fft}}, \code{\link[stats]{mvfft}}, or higher-level
#' helpers in this package such as \code{\link{convolve}}, \code{\link{pwelch}},
#' \code{\link{multitaper}}, and the filtering utilities.
#'
#' \strong{Warning:} the API is intentionally close to \pkg{FFTW}'s C interface
#' and may change between releases. Outputs match the corresponding base R
#' transforms up to floating-point round-off.
#'
#' @param data Numeric (real) or complex input. For 2D/3D variants, a matrix
#'   or 3-dimensional array. For \code{mvfftw_*} functions, a matrix whose
#'   columns are transformed independently.
#' @param HermConj Integer \code{0} or \code{1}. When \code{1}, return the
#'   full Hermitian-symmetric spectrum of length \code{N} (matching
#'   \code{stats::fft}); when \code{0}, return only the non-redundant
#'   one-sided half of length \code{floor(N/2) + 1}.
#' @param inverse Integer \code{0} (forward) or \code{1} (inverse). The
#'   inverse transform is \strong{unnormalized}; divide by the number of
#'   elements to invert a forward transform, mirroring
#'   \code{stats::fft(., inverse = TRUE)}.
#' @param fftwplanopt Integer planner effort: \code{0} = \code{FFTW_ESTIMATE}
#'   (default, fast planning), \code{1} = \code{FFTW_MEASURE},
#'   \code{2} = \code{FFTW_PATIENT}, \code{3} = \code{FFTW_EXHAUSTIVE}.
#' @param retrows Integer; expected number of rows of the time-domain signal
#'   for \code{mvfftw_c2r} when reconstructing from a one-sided spectrum.
#' @param ret Optional reusable output buffer of the correct type and
#'   length; pass \code{NULL} (default) to let the function allocate one.
#'   This input is for advanced users; leave \code{NULL} if you do not know
#'   what you are d
#'
#' @return A complex (or real, for \code{*_c2r}) vector / matrix / array
#'   matching the corresponding base R transform up to floating-point error.
#'
#' @details
#' All functions preserve their \code{data} argument: the input buffer is
#' copied internally before planning when needed, so callers may safely reuse
#' \code{data} after the call. For multi-dimensional variants, axis ordering
#' follows R (column-major) conventions.
#'
#' @examples
#' set.seed(1)
#'
#' ## --- 1D real-to-complex --------------------------------------------------
#' x <- rnorm(16)
#' a <- ravetools::fftw_r2c(x, HermConj = 1)
#' b <- stats::fft(x)
#' all.equal(a, b)                                  # TRUE (within tol)
#'
#' # one-sided spectrum (length floor(N/2)+1)
#' a_half <- ravetools::fftw_r2c(x, HermConj = 0)
#' all.equal(a_half, b[seq_len(length(x) %/% 2 + 1)])
#'
#' ## --- 1D complex-to-complex ----------------------------------------------
#' z <- complex(real = rnorm(16), imaginary = rnorm(16))
#' all.equal(ravetools::fftw_c2c(z, inverse = 0), stats::fft(z))
#' all.equal(ravetools::fftw_c2c(z, inverse = 1), stats::fft(z, inverse = TRUE))
#'
#' ## --- 1D complex-to-real (inverse of fftw_r2c) ---------------------------
#' # Using the full Hermitian spectrum:
#' xr <- ravetools::fftw_c2r(a, HermConj = 1) / length(x)
#' all.equal(xr, x)
#'
#' ## --- Multivariate (column-wise) ----------------------------------------
#' M <- matrix(rnorm(32), nrow = 8, ncol = 4)
#' all.equal(ravetools::mvfftw_r2c(M, HermConj = 1), stats::mvfft(M + 0i))
#'
#' Mz <- matrix(complex(real = rnorm(32), imaginary = rnorm(32)),
#'              nrow = 8, ncol = 4)
#' all.equal(ravetools::mvfftw_c2c(Mz, inverse = 0), stats::mvfft(Mz))
#' all.equal(ravetools::mvfftw_c2c(Mz, inverse = 1),
#'           stats::mvfft(Mz, inverse = TRUE))
#'
#' # one-sided -> back to real signal
#' Mh <- ravetools::mvfftw_r2c(M, HermConj = 0)
#' Mr <- ravetools::mvfftw_c2r(Mh, retrows = nrow(M)) / nrow(M)
#' all.equal(Mr, M)
#'
#' ## --- 2D ----------------------------------------------------------------
#' X2 <- matrix(rnorm(20), nrow = 5, ncol = 4)
#' all.equal(ravetools::fftw_r2c_2d(X2, HermConj = 1), stats::fft(X2 + 0i))
#'
#' Z2 <- matrix(complex(real = rnorm(20), imaginary = rnorm(20)),
#'              nrow = 5, ncol = 4)
#' all.equal(ravetools::fftw_c2c_2d(Z2, inverse = 0), stats::fft(Z2))
#' all.equal(ravetools::fftw_c2c_2d(Z2, inverse = 1),
#'           stats::fft(Z2, inverse = TRUE))
#'
#' ## --- 3D ----------------------------------------------------------------
#' X3 <- array(rnorm(60), dim = c(5, 4, 3))
#' all.equal(ravetools::fftw_r2c_3d(X3, HermConj = 1), stats::fft(X3 + 0i))
#'
#' Z3 <- array(complex(real = rnorm(60), imaginary = rnorm(60)),
#'             dim = c(5, 4, 3))
#' all.equal(ravetools::fftw_c2c_3d(Z3, inverse = 0), stats::fft(Z3))
#' all.equal(ravetools::fftw_c2c_3d(Z3, inverse = 1),
#'           stats::fft(Z3, inverse = TRUE))
#'
#' @export fftw_r2c
#' @export fftw_c2r
#' @export fftw_c2c
#' @export mvfftw_r2c
#' @export mvfftw_c2c
#' @export mvfftw_c2r
#' @export fftw_r2c_2d
#' @export fftw_c2c_2d
#' @export fftw_r2c_3d
#' @export fftw_c2c_3d
NULL

#' @title Cluster electrodes by their canonical \verb{CRP} response shape
#' @description
#' Groups recording electrodes by the \emph{shape} of their canonical evoked
#' response. Each electrode is summarized by \code{\link{crp}} into one
#' amplitude-normalized canonical shape; an electrode-by-electrode similarity
#' matrix is factorized with \code{\link{naive_nmf}} to find clusters that share
#' a response shape, and a representative basis profile curve is extracted per
#' cluster. This applies the basis-profile-curve (\verb{BPC}) approach
#' \emph{across electrodes} rather than across stimulation sites; see
#' \sQuote{References}.
#'
#' @details
#' \enumerate{
#' \item \strong{Common window.} Electrodes may have different time axes
#'   (\code{\link{crp}} drops \code{NA} samples); the overlapping time domain is
#'   used, \code{time_window} is clipped into it, and each electrode is subset on
#'   the fly with no interpolation.
#' \item \strong{Similarity.} With \code{paired = TRUE} each entry is the
#'   one-sample t-statistic, across an electrode pair's common trials, of the
#'   per-trial \verb{CCEP} cross-projection of the \eqn{1/\alpha}-rescaled
#'   responses; negatives are zeroed, the matrix made symmetric and scaled to a
#'   maximum of one. With \code{paired = FALSE} it is the cosine cross-projection
#'   of the \code{C_full} curves.
#' \item \strong{Rank selection.} \code{\link{naive_nmf}} factorizes the
#'   similarity at rank \eqn{Q} (from \code{initial_rank}); each rank is re-run a
#'   few times and the lowest-error fit kept. \eqn{Q} is reduced while
#'   \eqn{\zeta} - the sum of the upper off-diagonal of the row-normalized
#'   \eqn{HH^\top} - exceeds \code{zeta_threshold}.
#' \item \strong{Assignment.} Each electrode takes its winner-take-all cluster
#'   over the normalized \verb{NMF} loadings; with \code{null_class}, loadings
#'   below \eqn{1/(2\sqrt{N})} are left unassigned.
#' \item \strong{Basis curves.} Per cluster, the first linear kernel \verb{PCA}
#'   component of its members' \code{C_full} curves (as in \code{\link{crp}}).
#' }
#'
#' @param crp_list a named list of \code{ravetools_crp} objects (one per
#' electrode) from \code{\link{crp}}; list names are used as electrode labels.
#' Time axes need not match - the common overlapping domain is used (an error is
#' raised if the domains do not overlap). See \sQuote{Details}.
#' @param paired logical; \code{TRUE} (the default) builds the trial-level
#' \verb{BPC} t-statistic similarity and requires the single trials in
#' \code{.data$V}; \code{FALSE} uses the cosine similarity of the \code{C_full}
#' curves. See \sQuote{Details}.
#' @param time_window numeric \code{c(lo, hi)} analysis window, clipped into the
#' common domain; an \code{NA} extends to that domain edge. Defaults to
#' \code{c(0, NA)} (stimulus onset to the end of the domain).
#' @param n_clusters integer or \code{NULL}; if given, fixes the number of
#' clusters and skips the automatic selection (which otherwise uses
#' \code{zeta_threshold}).
#' @param initial_rank integer or \code{NULL}; the starting \verb{NMF} rank for
#' automatic selection. \code{NULL} uses \code{min(length(crp_list) - 1L, 10L)}.
#' @param zeta_threshold numeric \eqn{> 0}; rank-selection cutoff on \eqn{\zeta}
#' (see \sQuote{Details}); smaller values give fewer clusters. Defaults to
#' \code{1}, matching the \verb{BPC} reference.
#' @param null_class logical; if \code{TRUE} (the default) electrodes whose
#' winner-take-all loading falls below \eqn{1/(2\sqrt{N})} are left unassigned
#' (\code{NA}); if \code{FALSE} every electrode is forced into a cluster. See
#' \sQuote{Details}.
#' @param nmf_max_iters,nmf_tol passed to \code{\link{naive_nmf}}.
#' @param verbose logical; whether to report progress.
#'
#' @returns A named list of class \code{ravetools_crp_cluster}:
#' \describe{
#' \item{\code{clusters}}{Integer vector, the cluster index assigned to each
#'   electrode (winner-take-all over the row-normalized \verb{NMF} loadings).
#'   When \code{null_class = TRUE}, electrodes whose top loading falls below
#'   \eqn{1/(2\sqrt{N})} are left unassigned (\code{NA}).}
#' \item{\code{basis_curves}}{Numeric matrix, time \eqn{\times} number of
#'   clusters; column \eqn{q} is the basis profile curve \eqn{B_q(t)} for
#'   cluster \eqn{q}, the first linear-kernel \verb{PCA} component of its member
#'   \code{C_full} curves, sign-oriented to the cluster mean.}
#' \item{\code{basis_times}}{Numeric vector, the time axis for
#'   \code{basis_curves} (the common overlapping time axis, restricted to
#'   \code{time_window}).}
#' \item{\code{similarity}}{Numeric matrix, the electrode-by-electrode
#'   similarity \eqn{\Xi} that was factorized (non-negative, scaled to a maximum
#'   of one).}
#' \item{\code{nmf}}{The \code{\link{naive_nmf}} result for the selected rank.}
#' \item{\code{n_clusters}}{Integer, the number of clusters found.}
#' \item{\code{domain}}{Numeric \code{c(lo, hi)}, the overlapping time domain
#'   shared by all electrodes.}
#' \item{\code{paired},\code{time_window}}{The settings used; \code{time_window}
#'   is the effective window after clipping into \code{domain}.}
#' }
#'
#' @references
#' The \verb{BPC} method is described in \doi{10.1371/journal.pcbi.1008710}; the
#' underlying \verb{CRP} method in \doi{10.1371/journal.pcbi.1011105}.
#'
#' @seealso \code{\link{crp}}, \code{\link{naive_nmf}}
#'
#' @examples
#' # Four response shapes; shapes 3 and 4 differ only in amplitude, so they
#' # cluster together once shapes are amplitude-normalized.
#' \donttest{
#'
#' n_time <- 300L
#' tt <- seq(-0.2, 1, length.out = n_time)
#' shapes <- list(
#'   exp(-((tt - 0.08) / 0.03)^2) - 0.5 * exp(-((tt - 0.18) / 0.04)^2),
#'   exp(-((tt - 0.38) / 0.03)^2) - 0.5 * exp(-((tt - 0.50) / 0.04)^2),
#'   exp(-((tt - 0.70) / 0.04)^2),
#'   exp(-((tt - 0.70) / 0.04)^2) * 2
#' )
#'
#' # 4 electrodes per shape, each parameterized with crp()
#' crp_list <- list()
#' for (g in seq_along(shapes)) {
#'   for (e in seq_len(4L)) {
#'     V <- outer(shapes[[g]], runif(15L, 0.5, 1.5)) +
#'       matrix(rnorm(n_time * 15L, sd = 0.15), n_time, 15L)
#'     crp_list[[sprintf("elec_%d_%d", g, e)]] <-
#'       crp(V, tt, remove_artifacts = FALSE)
#'   }
#' }
#'
#' res <- crp_cluster(crp_list, verbose = TRUE)
#' res$n_clusters
#' table(res$clusters)
#' plot(res)
#'
#' }
#'
#' @export
crp_cluster <- function(
  crp_list,
  paired = TRUE,
  time_window = c(0, NA),
  n_clusters = NULL,
  initial_rank = NULL,
  zeta_threshold = 1,
  null_class = TRUE,
  nmf_max_iters = 1e4,
  nmf_tol = c(1e-4, 1e-8),
  verbose = TRUE
) {

  # DIPSAUS DEBUG START
  # pipeline <- ravepipeline::pipeline(
  #   pipeline_name = "voltage_explorer",
  #   paths = file.path(
  #     rstudioapi::getActiveProject(),
  #     "../rave-pipelines/modules/"
  #   ),
  #   temporary = TRUE
  # )
  # crp_results <- pipeline$run("crp_results")
  # crp_list <- unname(lapply(crp_results, function(x) { x[[1]] }))
  # paired = TRUE
  # time_window = c(0, 0.5)
  # normalize = TRUE
  # n_clusters = NULL
  # initial_rank = NULL
  # zeta_threshold = 0.5
  # nmf_max_iters = 1e4
  # nmf_tol = c(1e-4, 1e-8)
  # verbose = TRUE


  paired <- isTRUE(as.logical(paired))

  # ---- validate inputs ----------------------------------------------------
  if (!is.list(crp_list) || length(crp_list) < 2L) {
    stop("`crp_list` must be a list of at least 2 `ravetools_crp` objects.")
  }
  is_crp <- vapply(crp_list, inherits, logical(1L), what = "ravetools_crp")
  if (!all(is_crp)) {
    stop("Every element of `crp_list` must be a `ravetools_crp` object ",
         "returned by `crp()`.")
  }
  labels <- names(crp_list)
  if (is.null(labels) || any(!nzchar(labels)) || anyDuplicated(labels)) {
    labels <- sprintf("electrode_%d", seq_along(crp_list))
  }
  names(crp_list) <- labels
  N <- length(crp_list)

  # ---- common (overlapping) time domain across electrodes -----------------
  # crp() drops non-finite (NA) rows, so electrodes may carry different time
  # axes. We only require their domains to overlap; subsetting is done on the
  # fly (by time value, per electrode) rather than assuming shared row indices.
  domain <- crp_cluster_domain(crp_list)

  # ---- comparison window: validate, then clip into the common domain ------
  time_window <- as.numeric(time_window)
  if (!length(time_window)) {
    window <- domain
  } else {
    if (length(time_window) == 1) {
      time_window <- c(0, time_window)
    } else {
      time_window <- time_window[c(1, 2)]
    }
    time_window[is.na(time_window)] <- domain[is.na(time_window)]
    time_window[time_window < domain[[1]]] <- domain[[1]]
    time_window[time_window > domain[[2]]] <- domain[[2]]
    if (is.unsorted(time_window)) {
      stop(
        sprintf(
          "`time_window` must be a numeric c(lo, hi) with lo < hi, and must overlap with the common time domain [%.4g, %.4g].",
          domain[1L], domain[2L])
      )
    }
    window <- time_window
  }

  # ---- electrode-by-electrode similarity ----------------------------------
  # paired = TRUE: trial-level cross-projection t-statistic over pairwise common
  # trials; paired = FALSE: cosine of the C_full curves. Returned directly.
  similarity <- crp_cluster_similarity_matrix(crp_list, window, paired)

  # ---- NMF clustering with optional automatic rank selection ---------------
  if (is.null(initial_rank)) {
    initial_rank <- min(N - 1L, 10L)
  }
  initial_rank <- max(1L, min(as.integer(initial_rank), N - 1L))
  if (!is.null(n_clusters)) {
    n_clusters <- max(1L, min(as.integer(n_clusters), N - 1L))
  }

  sel <- crp_cluster_nmf_select(
    similarity,
    n_clusters = n_clusters,
    initial_rank = initial_rank,
    zeta_threshold = zeta_threshold,
    nmf_max_iters = nmf_max_iters,
    nmf_tol = nmf_tol,
    verbose = verbose
  )
  nmf <- sel$nmf

  # ---- winner-take-all assignment, renumbered to present clusters ----------
  # Each electrode joins its top cluster on the row-normalized NMF loadings. When
  # `null_class = TRUE` (the BPC reference behavior) a membership is kept only if
  # that normalized loading exceeds 1/(2*sqrt(N)) - half the value a uniform
  # loading 1/sqrt(N) would give - else the electrode is left unassigned (NA, the
  # "null" class). When FALSE every electrode is forced into its winner cluster.
  Hn <- nmf$H
  rn <- sqrt(rowSums(Hn * Hn))
  rn[rn == 0] <- 1
  Hn <- Hn / rn
  raw_cluster <- apply(Hn, 2L, which.max)
  if (isTRUE(null_class)) {
    winner <- Hn[cbind(raw_cluster, seq_len(N))]
    raw_cluster[winner <= 1 / (2 * sqrt(N))] <- NA_integer_
  }
  present <- sort(unique(raw_cluster))
  clusters <- match(raw_cluster, present)
  n_found <- length(present)

  # ---- per-cluster basis curves + per-electrode metrics --------------------
  # Subsets each electrode's C_full on the fly to `window` using its own time.
  fm <- crp_cluster_feature_matrix(crp_list, window, TRUE)
  feature_mat <- fm$features
  basis_times <- fm$times
  basis <- crp_cluster_basis(feature_mat, clusters, n_found)
  basis_curves <- basis$basis_curves
  # metrics <- crp_cluster_metrics(feature_mat, clusters, basis_curves)

  rownames(basis_curves) <- NULL

  structure(
    class = "ravetools_crp_cluster",
    list(
      clusters = clusters,
      basis_curves = basis_curves,
      basis_times = basis_times,
      similarity = similarity,
      # metrics = metrics,
      nmf = nmf,
      n_clusters = n_found,
      paired = paired,
      domain = domain,
      time_window = window
    )
  )
}

#' @title Plot electrode clustering results
#' @description S3 plot method for \code{ravetools_crp_cluster} objects returned
#' by \code{\link{crp_cluster}}. Draws a two-panel figure: the per-cluster basis
#' profile curves, and the electrode-by-electrode similarity matrix ordered by
#' cluster.
#' @param x a \code{ravetools_crp_cluster} object.
#' @param ... ignored.
#' @return Invisibly returns \code{x}.
#' @export
plot.ravetools_crp_cluster <- function(x, ...) {
  Q <- x$n_clusters
  cols <- grDevices::hcl.colors(max(Q, 2L), palette = "Dark 3")

  op <- graphics::par(mfrow = c(1L, 2L))
  on.exit(graphics::par(op), add = TRUE)

  # ---- Panel 1: basis profile curves --------------------------------------
  bt <- x$basis_times
  bc <- x$basis_curves
  graphics::matplot(
    bt, bc, type = "l", lty = 1L, lwd = 2L, col = cols[seq_len(Q)],
    las = 1L, xlab = "Time (s)", ylab = "Basis amplitude (a.u.)",
    main = "Basis profile curves"
  )
  graphics::abline(h = 0, col = "#cccccc", lty = 3L)
  graphics::legend(
    "topright", bty = "n", cex = 0.8, lwd = 2L, col = cols[seq_len(Q)],
    legend = sprintf("cluster %d (n=%d)", seq_len(Q),
                     tabulate(x$clusters, nbins = Q))
  )

  # ---- Panel 2: similarity matrix ordered by cluster ----------------------
  ord <- order(x$clusters)
  S <- x$similarity[ord, ord, drop = FALSE]
  graphics::image(
    seq_len(nrow(S)), seq_len(ncol(S)), S,
    col = grDevices::hcl.colors(64L, palette = "YlOrRd", rev = TRUE),
    xlab = "Electrode (ordered by cluster)", ylab = "Electrode",
    main = "Similarity matrix", las = 1L
  )
  # cluster block boundaries
  brk <- cumsum(tabulate(x$clusters, nbins = Q))
  brk <- brk[-length(brk)] + 0.5
  graphics::abline(v = brk, h = brk, col = "#202020", lty = 1L, lwd = 1L)

  invisible(x)
}


# --------------------------------------------------------------------------
# Internal helpers (not exported)
# --------------------------------------------------------------------------

# Overlapping time domain (range) shared by all electrodes. Different crp objects
# may carry different time axes (crp() drops non-finite rows), so the common
# domain is [max of the per-electrode minima, min of the per-electrode maxima].
crp_cluster_domain <- function(crp_list) {
  rng <- vapply(crp_list, function(obj) {
    range(obj$parameters$params_times_full)
  }, numeric(2L))
  lo <- max(rng[1L, ])
  hi <- min(rng[2L, ])
  if (!is.finite(lo) || !is.finite(hi) || lo >= hi) {
    stop("The `crp` objects have no overlapping time domain.")
  }
  c(lo, hi)
}

# Assemble the time x electrode feature matrix from each electrode's C_full,
# subsetting on the fly to the (domain-clipped) `window` using each electrode's
# own time axis. Because the axes may be heterogeneous, the selected timepoints
# are checked for agreement across electrodes (no interpolation). Returns the
# feature matrix and its shared time axis. Optionally unit-normalized so
# electrodes cluster by waveform shape rather than amplitude.
crp_cluster_feature_matrix <- function(crp_list, window, normalize) {
  picked <- lapply(crp_list, function(obj) {
    tt <- obj$parameters$params_times_full
    sel <- tt >= window[1L] & tt <= window[2L]
    list(times = tt[sel], values = obj$parameters$C_full[sel])
  })

  ref_times <- picked[[1L]]$times
  if (length(ref_times) < 2L) {
    stop("Fewer than 2 timepoints fall within the common time domain.")
  }
  for (i in seq_along(picked)) {
    if (length(picked[[i]]$times) != length(ref_times) ||
        !isTRUE(all.equal(picked[[i]]$times, ref_times, tolerance = 1e-8))) {
      stop("Electrodes have misaligned timepoints within the common domain; ",
           "`C_full` cannot be stacked without interpolation (electrode '",
           names(crp_list)[i], "' differs).")
    }
  }

  feature_mat <- vapply(picked, function(p) {
    p$values
  }, numeric(length(ref_times)))
  feature_mat <- as.matrix(feature_mat)
  if (isTRUE(normalize)) {
    norms <- sqrt(colSums(feature_mat * feature_mat))
    norms[norms == 0] <- 1
    feature_mat <- t(t(feature_mat) / norms)
  }
  feature_mat[!is.finite(feature_mat)] <- 0
  list(features = feature_mat, times = ref_times)
}

# Electrode-by-electrode similarity from the cross-projection of normalized
# C_full curves (BPC's Xi). Negatives are zeroed and the matrix is scaled to a
# maximum of one, matching the BPC significance-matrix convention.
crp_cluster_cross_matrix <- function(feature_mat) {
  norms <- sqrt(colSums(feature_mat * feature_mat))
  norms[norms == 0] <- 1
  F0 <- t(t(feature_mat) / norms)
  P <- crossprod(F0)            # cosine similarity, electrodes x electrodes
  P <- (P + t(P)) / 2           # enforce symmetry
  P[P < 0] <- 0
  mx <- max(P)
  if (mx > 0) P <- P / mx
  P
}

# Electrode-by-electrode similarity matrix (BPC significance matrix), returned
# directly. When `paired` is TRUE it is the one-sample t-statistic, across each
# electrode pair's *common* trials, of the per-trial CCEP cross-projection
# between the alpha-rescaled single-trial responses. When FALSE it falls back to
# the cosine cross-projection of the C_full curves (no trials needed). Both are
# made non-negative, symmetric and scaled to a maximum of one for `naive_nmf`.
crp_cluster_similarity_matrix <- function(crp_list, window, paired) {
  if (!isTRUE(paired)) {
    fm <- crp_cluster_feature_matrix(crp_list, window, normalize = TRUE)
    return(crp_cluster_cross_matrix(fm$features))
  }

  N <- length(crp_list)

  # alpha-rescaled, window-subset single trials per electrode (~ canonical shape
  # per trial, so amplitude variance is removed and the t-statistic sharpens)
  similarity <- lapply(seq_len(N), function(i) {
    obj_i <- crp_list[[i]]
    # al <- obj$parameters$al
    # al[!is.finite(al) | al == 0] <- 1
    tt_i <- obj_i$parameters$params_times
    # tt_i <- obj_i$.data$time
    sel_i <- tt_i >= window[1L] & tt_i <= window[2L]
    bad_trials_i <- obj_i$bad_trials

    tp_i <- sum(sel_i)
    window_i <- range(tt_i[sel_i])
    trial_i <- seq_len(ncol(obj_i$parameters$V_tR) + length(bad_trials_i))
    # trial_i <- seq_len(ncol(obj_i$.data$V) + length(bad_trials_i))
    if (length(bad_trials_i)) {
      trial_i <- trial_i[-bad_trials_i]
    }


    X_i <- vapply(seq_len(N), function(j) {

      # if (i == j) { return(0.0) }

      obj_j <- crp_list[[j]]

      bad_trials_j <- obj_j$bad_trials
      trial_j <- seq_len(ncol(obj_j$parameters$V_tR) + length(bad_trials_j))
      # trial_j <- seq_len(ncol(obj_j$.data$V) + length(bad_trials_j))
      if (length(bad_trials_j)) {
        trial_j <- trial_j[-bad_trials_j]
      }

      common_trials <- intersect(trial_i, trial_j)

      if (length(common_trials) <= 1) { return(0.0) }

      # crp() requires increasing time, paired=TRUE already assumed time consistency
      tt_j <- obj_j$parameters$params_times
      # tt_j <- obj_j$.data$time
      sel_j <- tt_j >= window_i[1L] & tt_j <= window_i[2L]

      tp_j <- sum(sel_j)
      if (!tp_j) { return(0.0) }

      window_j <- range(tt_j[sel_j])
      sel_i <- tt_i >= window_j[1L] & tt_i <= window_j[2L]
      tp_i <- sum(sel_i)
      if (!tp_i) { return(0.0) }

      # Assuming sum(sel_i) == sum(sel_j)
      if (tp_i != tp_j) {
        stop(sprintf("Number of time-point mismatch between object %d (%d points) vs %d (%d points)",
                     i, tp_i, j, tp_j))
      }

      trial_sel_j <- trial_j %in% common_trials
      vi <- obj_i$parameters$V_tR[sel_i, trial_i %in% common_trials, drop = FALSE]
      vj <- obj_j$parameters$V_tR[sel_j, trial_sel_j, drop = FALSE]
      # vi <- obj_i$.data$V[sel_i, trial_i %in% common_trials, drop = FALSE]
      # vj <- obj_j$.data$V[sel_j, trial_sel_j, drop = FALSE]

      # Calculate rescales
      # for vi, normalize to norm_2 = 1
      norm_i <- sqrt(colSums(vi ^ 2))
      norm_i[norm_i == 0] <- 1
      ui <- t(vi) / norm_i

      # For vj, rescale by `al`
      al_j <- obj_j$parameters$al
      al_j[al_j == 0] <- 1
      uj <- t(vj) / al_j[trial_sel_j]

      S <- rowSums(ui * uj)
      S <- S[is.finite(S)]
      if (length(S) <= 1) { return(0.0) }

      sd_s <- stats::sd(S)
      if (!is.finite(sd_s) || sd_s == 0) { return(0.0) }

      # standardized S
      mean(S) / (sd_s / sqrt(length(S)))
    }, 0.0)
    X_i
  })

  similarity <- t(simplify2array(unname(similarity), higher = TRUE))
  similarity[similarity < 0 | !is.finite(similarity)] <- 0
  # diag(similarity) <- 0

  # off-diagonal structure drives clustering: scale those to [0, 1] and set a
  # unit self-similarity (the diagonal t-statistic of trial norms is huge and
  # would otherwise dominate the max-scaling).
  m <- max(similarity)
  if (m > 0) {
    similarity <- similarity / m
  }

  # diag(similarity) <- 1

  similarity
}

# Run NMF on the similarity matrix, optionally reducing the rank automatically.
# When `n_clusters` is NULL the rank starts at `initial_rank` and drops by one
# while zeta - the sum of the upper-half off-diagonal elements of (row-normalized)
# HH^T - exceeds `zeta_threshold` (the BPC degeneracy criterion; see
# `crp_cluster_zeta`).
crp_cluster_nmf_select <- function(
  similarity, n_clusters, initial_rank, zeta_threshold,
  nmf_max_iters, nmf_tol, verbose
) {
  # NNMF is sensitive to its random initialization, so - following the BPC
  # reference (which uses 20) - re-run it a few times per rank and keep the
  # lowest-reconstruction-error factorization. 5 is sufficient here.
  n_restarts <- 5L
  fit_rank <- function(k) {
    fits <- lapply(seq_len(n_restarts), function(.) {
      naive_nmf(similarity, k = k, tol = nmf_tol,
                max_iters = nmf_max_iters, verbose = FALSE)
    })
    err <- vapply(fits, function(f) f$error[[1L]], numeric(1L))
    fits[[which.min(err)]]
  }

  if (!is.null(n_clusters)) {
    return(list(nmf = fit_rank(n_clusters), zeta = NA_real_, rank = n_clusters))
  }

  k <- initial_rank
  nmf <- fit_rank(k)
  zeta <- crp_cluster_zeta(nmf$H)
  while (k > 1L && is.finite(zeta) && zeta > zeta_threshold) {
    if (isTRUE(verbose)) {
      message(sprintf("rank %d: HH^T off-diagonal sum = %.3f > %.2f, reducing",
                      k, zeta, zeta_threshold))
    }
    k <- k - 1L
    nmf <- fit_rank(k)
    zeta <- crp_cluster_zeta(nmf$H)
  }
  if (isTRUE(verbose)) {
    message(sprintf("selected rank %d (HH^T off-diagonal sum = %.3f)", k, zeta))
  }
  list(nmf = nmf, zeta = zeta, rank = k)
}

# NMF degeneracy, per the BPC reference (Miller et al.): the rows of H (the NNMF
# components' loadings over electrodes) are unit-normalized, then zeta is the SUM
# of the upper-half off-diagonal elements of HH^T - the total shared structure
# between components. (The reference notebook normalizes H rows and sums
# triu(HH^T, 1); the paper also mentions a max variant, but the released code
# uses the sum.) A degenerate factorization has high overlap; once the rank
# matches the true clusters the components become near-disjoint and the sum drops.
crp_cluster_zeta <- function(H) {
  if (nrow(H) < 2L) return(0)
  rn <- sqrt(rowSums(H * H))
  rn[rn == 0] <- 1
  Hn <- H / rn                            # unit-normalize each component (row)
  HHt <- tcrossprod(Hn)                   # normalized HH^T
  sum(HHt[upper.tri(HHt)])                # BPC: sum of upper off-diagonals
}

# Per-cluster basis profile curve: the first linear-kernel PCA component of the
# member C_full curves, sign-oriented to the cluster mean. Reuses `crp_kt_pca`.
crp_cluster_basis <- function(feature_mat, clusters, n_found) {
  Tn <- nrow(feature_mat)
  basis_curves <- matrix(0, nrow = Tn, ncol = n_found)
  for (q in seq_len(n_found)) {
    members <- which(clusters == q)
    Xq <- feature_mat[, members, drop = FALSE]
    pca <- crp_kt_pca(Xq)
    Bq <- pca$E[, 1L]
    if (sum(Bq * rowMeans(Xq)) < 0) Bq <- -Bq
    basis_curves[, q] <- Bq
  }
  list(basis_curves = basis_curves)
}

# Per-electrode parameterization against its assigned basis curve: alpha
# (projection), signal-to-noise and explained variance, defined exactly as in
# `crp_core`.
crp_cluster_metrics <- function(feature_mat, clusters, basis_curves) {
  N <- ncol(feature_mat)
  alpha <- numeric(N)
  Vsnr <- numeric(N)
  expl_var <- numeric(N)
  for (i in seq_len(N)) {
    f <- feature_mat[, i]
    B <- basis_curves[, clusters[i]]
    a <- sum(B * f)
    ep <- f - a * B
    epep <- sum(ep * ep)
    ff <- sum(f * f)
    alpha[i] <- a
    Vsnr[i] <- if (epep > 0) a / sqrt(epep) else NA_real_
    expl_var[i] <- if (ff > 0) 1 - epep / ff else NA_real_
  }
  data.frame(
    cluster = unname(clusters),
    alpha = alpha,
    Vsnr = Vsnr,
    expl_var = expl_var,
    stringsAsFactors = FALSE
  )
}

## usethis namespace: start
#' @importFrom stats approx
#' @importFrom stats median
#' @importFrom stats quantile
#' @importFrom stats sd
#' @importFrom Rcpp sourceCpp
#' @importFrom graphics image
#' @importFrom R6 R6Class
#' @importFrom R6 is.R6
#' @useDynLib ravetools, .registration = TRUE
## usethis namespace: end
NULL

`%?<-%` <- function(lhs, value) {
  env <- parent.frame()
  lhs <- substitute(lhs)

  isnull <- tryCatch({
    is.null(eval(lhs, envir = env))
  }, error = function(e) {
    return(TRUE)
  })

  if (isnull) {
    eval(as.call(list( quote(`<-`), lhs, value )), envir = env)
  }
}

`%||%` <- function(a, b) {
  if (is.null(a)) {
    b
  } else {
    a
  }
}

deparse1 <- function(expr, collapse = " ") {
  paste(deparse(expr), collapse = collapse)
}


stopifnot2 <- function(..., msg = "Condition not satisfied") {
  if (!all(c(...))) {
    stop(msg)
  }
}

rand_string <- function(length = 10) {
  paste(sample(c(letters, LETTERS, 0:9), length, replace = TRUE), collapse = "")
}


tempdir2 <- function(check = TRUE) {
  path <- getOption("ravetools.tempdir",
            default = Sys.getenv("RAVETOOLS_TEMPDIR",
                                 unset = tempdir(check = FALSE)))
  if (check && !dir.exists(path)) {
    dir.create(path, showWarnings = FALSE, recursive = TRUE)
  }
  path
}

tempfile2 <- function(
  pattern = "ravetmp-", tmpdir = file.path(tempdir2(check = TRUE), "ravetools"),
  fileext = "") {
  if (!dir.exists(tmpdir)) {
    dir.create(tmpdir, showWarnings = FALSE, recursive = TRUE)
  }
  if (getOption("ravetools.debug", FALSE)) {
    file.path(tmpdir, sprintf("%s%s%s", pattern, rand_string(), fileext))
  } else {
    tempfile(pattern = pattern, tmpdir = tmpdir, fileext = fileext)
  }

}


.missing_arg <- alist(x = )

.matlab_palette <- c("#00008F", "#00009F", "#0000AF", "#0000BF", "#0000CF",
          "#0000DF", "#0000EF", "#0000FF", "#0010FF", "#0020FF",
          "#0030FF", "#0040FF", "#0050FF", "#0060FF", "#0070FF",
          "#0080FF", "#008FFF", "#009FFF", "#00AFFF", "#00BFFF",
          "#00CFFF", "#00DFFF", "#00EFFF", "#00FFFF", "#10FFEF",
          "#20FFDF", "#30FFCF", "#40FFBF", "#50FFAF", "#60FF9F",
          "#70FF8F", "#80FF80", "#8FFF70", "#9FFF60", "#AFFF50",
          "#BFFF40", "#CFFF30", "#DFFF20", "#EFFF10", "#FFFF00",
          "#FFEF00", "#FFDF00", "#FFCF00", "#FFBF00", "#FFAF00",
          "#FF9F00", "#FF8F00", "#FF8000", "#FF7000", "#FF6000",
          "#FF5000", "#FF4000", "#FF3000", "#FF2000", "#FF1000",
          "#FF0000", "#EF0000", "#DF0000", "#CF0000", "#BF0000",
          "#AF0000", "#9F0000", "#8F0000", "#800000")

#' 'Matlab' heat-map plot palette
#' @returns vector of 64 colors
#' @export
matlab_palette <- function() {
  .matlab_palette
}

#' @name parallel-options
#' @title Set or get thread options
#' @param n_threads number of threads to set
#' @param stack_size Stack size (in bytes) to use for worker threads. The
#' default used for \code{"auto"} is 2MB on 32-bit systems and 4MB on 64-bit
#' systems.
#' @returns \code{detect_threads} returns an integer of default threads that
#' is determined by the number of \code{CPU} cores; \code{ravetools_threads}
#' returns nothing.
#'
#' @examples
#'
#' detect_threads()
#'
#' ravetools_threads(n_threads = 2)
#'
#' @export
detect_threads <- function() {
  getDefaultNumThreads()
}

#' @rdname parallel-options
#' @export
ravetools_threads <- function(n_threads = "auto", stack_size = "auto") {
  if (identical(n_threads, "auto"))
    n_threads <- -1L
  else if (!is.numeric(n_threads))
    stop("n_threads must be an integer")
  else n_threads <- as.integer(n_threads)
  if (identical(stack_size, "auto"))
    stack_size <- 0L
  else if (!is.numeric(stack_size))
    stop("stack_size must be an integer")
  else stack_size <- as.integer(stack_size)
  if (n_threads == -1L)
    Sys.unsetenv("RAVETOOLS_NUM_THREADS")
  else Sys.setenv(RAVETOOLS_NUM_THREADS = n_threads)
  if (stack_size == 0L)
    Sys.unsetenv("RAVETOOLS_STACK_SIZE")
  else Sys.setenv(RAVETOOLS_STACK_SIZE = stack_size)
  invisible()
}

#' @title Internal function
#' @description
#' Do not call this function directly
#' @param if_interactive,verbose default is \code{TRUE}
#' @returns logical
#' @keywords internal
#' @export
is_not_cran <- function(if_interactive = TRUE, verbose = FALSE) {
  not_cran_flag <- identical(toupper(as.character(Sys.getenv("NOT_CRAN", ""))), "TRUE")
  limit_core_flag <- identical(toupper(Sys.getenv("_R_CHECK_LIMIT_CORES_")), "TRUE")
  interactive_flag <- interactive()
  if (limit_core_flag) {
    if (verbose) {
      message("_R_CHECK_LIMIT_CORES_ is TRUE/true (on CRAN)")
    }
    return(FALSE)
  }
  if (not_cran_flag) {
    if (verbose) {
      message("NOT_CRAN is TRUE/true (not on CRAN)")
    }
    return(TRUE)
  }
  if (interactive_flag) {
    if_interactive <- isTRUE(if_interactive)
    if (verbose) {
      message(sprintf("Session is interactive (%son CRAN)",
                      ifelse(if_interactive, "", "not ")))
    }
    return(if_interactive)
  }
  if (verbose) {
    message("No flag detected, default is on CRAN")
  }
  return(FALSE)
}

#' Get external function from 'RAVE'
#' @description
#' Internal function used for examples relative to 'RAVE' project and should
#' not be used directly.
#' @param name function or variable name
#' @param pkg 'RAVE' package name
#' @param inherit passed to \code{\link{get0}}
#' @param on_missing default value to return of no function is found
#' @returns Function object if found, otherwise \code{on_missing}.
#' @export
internal_rave_function <- function(name, pkg, inherit = TRUE, on_missing = NULL) {
  if (!pkg %in% c("ravecore", "raveio", "ravedash", "ravebuiltins", "rave", "threeBrain",
                 "dipsaus", "filearray", "readNSx", "rpymat", "rpyANTs")) {
    stop("`extern_function`: Package [", pkg, "] is not a RAVE package.")
  }
  if (system.file(package = pkg) == "") { return(on_missing) }
  ns <- asNamespace(pkg)
  get0(x = name, envir = ns, inherits = inherit, ifnotfound = on_missing)
}

#' @export
`print.ravetools-printable` <- function(x, ...) {
  cat(paste(c(format(x, ...), ""), collapse = "\n"))
}

#' @export
`format.ravetools-printable` <- function(x, ...) {
  printable_message <- attr(x, "printable_message")
  if (length(printable_message)) {
    return(printable_message)
  }
  NextMethod("format")
}

#' Left 'Hippocampus' of 'N27-Collin' brain
#'
#' @format
#' A three-mode integer mask array with values of \code{1} ('Hippocampus')
#' and \code{0} (other brain tissues)
#'
"left_hippocampus_mask"


#' Sample stimulation recording
#'
#' @format
#' A list of one-second signal trace and sample rate (30000)
#'
"stimulation_signal"


lapply_async <- function(x, FUN, FUN.args = list(), callback = NULL, ncores = NULL,
                         on_failure = "multisession", ...) {
  if (system.file(package = "raveio") == "") {
    do.call("lapply", c(list(X = x, FUN = FUN), FUN.args))
    ret <- lapply(x, FUN)
  } else {
    raveio <- asNamespace("raveio")
    ret <- raveio$lapply_async(x = x, FUN = FUN, FUN.args = FUN.args, callback = callback, ncores = ncores, on_failure = on_failure, ...)
  }
  ret
}


#' @title Map continuous values to colors
#' @description
#' Linearly maps a numeric vector onto a color ramp, clamping values outside
#' the given range to the range's endpoints.
#'
#' @param values numeric vector of values to map to colors
#' @param clim length-two numeric vector giving the value range to map from;
#' values outside \code{[clim[1], clim[2]]} are clamped to the nearer endpoint
#' before mapping. Default is \code{range(values, na.rm = TRUE)}
#' @param cmap the color ramp to map onto: either a vector of colors (passed
#' to \code{\link[grDevices]{colorRamp}} to build the ramp function) or a
#' function such as the one returned by \code{\link[grDevices]{colorRamp}}
#' that takes a numeric vector with elements in \code{[0, 1]} and returns an
#' \code{n x 3} (or \code{n x 4}, with alpha) matrix of color-channel values
#' in \code{[0, 255]}. Default is \code{grDevices::hcl.colors(11)}
#' @param ... passed to \code{\link[grDevices]{colorRamp}} when \code{cmap} is
#' a vector of colors (for example, \code{alpha = TRUE} to build an
#' alpha-aware ramp)
#'
#' @returns A character vector of \code{'#RRGGBB'} (or \code{'#RRGGBBAA'})
#' color strings, the same length as \code{values}.
#'
#' @examples
#'
#' x <- rnorm(100)
#' col <- color_ramp_continuous(x)
#'
#' plot(x, col = col, pch = 16)
#'
#'
#' # Change color palettes with vector of colors
#' col <- color_ramp_continuous(
#'   x, cmap = c("lightgreen", "white", "pink"))
#' plot(x, col = col, pch = 16)
#'
#' # Using colorRamp
#' col <- color_ramp_continuous(
#'   x, cmap = colorRamp(c("black", "orangered", "orange")))
#' plot(x, col = col, pch = 16)
#'
#' # Using color ramp palette `function(n) { ... }`
#' col <- color_ramp_continuous(
#'   x, cmap = hcl.colors, palette = "Blue-Red 3")
#' plot(x, col = col, pch = 16)
#'
#' # Set range
#' col <- color_ramp_continuous(
#'   x, clim = c(0, 1),
#'   cmap = c("black", "orangered", "orange"))
#' plot(x, col = col, pch = 16)
#'
#'
#'
#' @export
color_ramp_continuous <- function(values, clim = range(values, na.rm = TRUE),
                                  cmap = grDevices::hcl.colors(11), ...) {
  clim[is.finite(clim)]
  if (!length(clim)) {
    stop("Value range is invalid: the value range must be finite values")
  }
  clim <- range(clim, na.rm = TRUE)
  span <- clim[[2]] - clim[[1]]
  if (span <= 0) { span <- 1}
  values_01 <- (values - clim[[1]]) / span

  values_01[values_01 < 0] <- 0
  values_01[values_01 > 1] <- 1

  # cmap is either a vector or a function returned by colorRamp
  if (!is.function(cmap)) {
    cmap <- grDevices::colorRamp(colors = cmap, ...)(values_01)
  } else {
    if (names(formals(cmap))[[1]] == "n") {
      # colorRampPalette
      cmap <- cmap(256, ...)
      return(cmap[round(values_01 * 255) + 1])
    } else {
      cmap <- cmap(values_01, ...)
    }
  }
  # cmap <- cmap(values_01)
  if (ncol(cmap) >= 4) {
    col <- grDevices::rgb(
      red = cmap[, 1],
      green = cmap[, 2],
      blue = cmap[, 3],
      alpha =  cmap[, 4],
      maxColorValue = 255
    )
  } else {
    col <- grDevices::rgb(
      red = cmap[, 1],
      green = cmap[, 2],
      blue = cmap[, 3],
      maxColorValue = 255
    )
  }
  col
}


# color_ramp_

#' @name stimpulse_interpolate
#' @title Find and interpolate stimulation pulses
#' @param signal a channel signal trace
#' @param sample_rate sample rate
#' @param pulse_duration stimulation pulse duration in seconds
#' @param n_pulses suggested number of pulses
#' @param threshold suggested suggested threshold of responses to find
#' stimulation pulses
#' @param pulse_info a list containing number of pulses \code{n_pulse},
#' onset index (\code{onset_index}, first time-point is 1), offset index
#' (\code{offset_index}); this can be generated by \code{stimpulse_find};
#' see 'Examples' below
#' @param expand_timepoints point offsets allowed to align the pulses
#' @param center whether to center the pulses by median; default is true
#' @param max_offset maximum (edge) offsets in seconds to interpolate the
#' pulses; default is \code{-0.0002} seconds before stimulation onset
#' and \code{0.0005} seconds after the stimulation offset
#' @returns \code{stimpulse_find} and \code{stimpulse_align} returns the pulse
#' information (\code{pulse_info}) with the time-points of detected or corrected
#' stimulation on-set and off-set. The time-points are 1-indexed.
#' \code{stimpulse_extract} extract the signals around pulses;
#' \code{stimpulse_interpolate} returns interpolated signals.
#' @examples
#'
#'
#'
#' data("stimulation_signal")
#'
#' signal <- stimulation_signal$signal
#' sample_rate <- stimulation_signal$sample_rate
#'
#' # each pulse is roughly <0.001 seconds
#' pulse_durations <- 0.001
#'
#' # Initial pulses
#' pulse_info <- stimpulse_find(signal, sample_rate, pulse_durations)
#'
#' # number of pulses detected
#' pulse_info$n_pulses
#'
#' # extract responses -10 points before onset ~ 20 points after offset
#' expand_timepoints <- c(-20, 80)
#' pulses_snippets <- stimpulse_extract(
#'   signal = signal,
#'   pulse_info = pulse_info,
#'   expand_timepoints = expand_timepoints
#' )
#'
#' # Visualize the pulses
#' snippet_time <- seq(
#'   expand_timepoints[[1]], by = 1,
#'   length.out = nrow(pulses_snippets)) / sample_rate * 1000
#' matplot(snippet_time, pulses_snippets, type = 'l', lty = 1, col = 'gray80',
#'         xlab = "Time (ms)", ylab = "uV", main = "Initial find")
#'
#' # Align the pulses
#' pulse_info <- stimpulse_align(signal, pulse_info)
#'
#' # Estimated pulse duration
#' estimated_duration <-
#'   (pulse_info$offset_index - pulse_info$onset_index + 1) / sample_rate
#'
#' # reload aligned pulses
#' pulses_snippets <- stimpulse_extract(
#'   signal = signal,
#'   pulse_info = pulse_info,
#'   expand_timepoints = expand_timepoints
#' )
#' matplot(snippet_time, pulses_snippets, type = 'l', lty = 1, col = 'gray80',
#'         xlab = "Time (ms)", ylab = "uV", main = "Aligned pulses")
#' lines(snippet_time, rowMeans(pulses_snippets), col = 'red')
#'
#'
#' # Interpolate the pulses
#' interpolated <- stimpulse_interpolate(
#'   signal = signal,
#'   sample_rate = sample_rate,
#'   pulse_info = pulse_info,
#'   max_offset = c(-0.0003, 0.0005)
#' )
#'
#'
#' interp_snippets <- stimpulse_extract(
#'   signal = interpolated,
#'   pulse_info = pulse_info,
#'   expand_timepoints = expand_timepoints
#' )
#'
#' oldpar <- par(mfrow = c(1, 2))
#' on.exit(par(oldpar))
#' matplot(snippet_time, pulses_snippets, type = 'l', lty = 1,
#'         col = 'gray80', xlab = "Time (ms)", ylab = "uV",
#'         main = "Stim pulses", ylim = c(-600, 400))
#' lines(snippet_time, rowMeans(pulses_snippets), col = 'red')
#' abline(v = max(estimated_duration) * 1000, lty = 2)
#'
#' matplot(snippet_time, interp_snippets, type = 'l', lty = 1,
#'         col = 'gray80', xlab = "Time (ms)", ylab = "uV",
#'         main = "Interpolated 0.5 ms bandwidth")
#' lines(snippet_time, rowMeans(interp_snippets), col = 'red')
#' abline(v = max(estimated_duration) * 1000, lty = 2, col = "gray40")
#' abline(v = max(estimated_duration) * 1000 + 0.5, lty = 2)
#'
#'
#'
#'
#'
#'
NULL

#' @rdname stimpulse_interpolate
#' @export
stimpulse_find <- function(signal, sample_rate, pulse_duration, n_pulses = NA, threshold = NA) {

  # pulse_duration <- 0.003
  # n_pulses <- NA
  # threshold <- NA

  # diff by one makes brain signal more flat
  signal_diff <- diff(signal)
  signal_diff_ceiling <- max(abs(signal_diff)) + 1

  min_idx_gap <- ceiling(pulse_duration * sample_rate)

  # throttle
  find_onsets_throttle <- function(threshold, sdff) {
    pulse_onset_idx <- which(abs(sdff) > threshold) - 1L
    pulse_onset_idx[pulse_onset_idx <= 0] <- 1L

    # find all the onsets
    last_idx <- -Inf
    pulse_onset_idx <- pulse_onset_idx[vapply(pulse_onset_idx, function(idx) {

      last_idx2 <- last_idx
      if (idx - last_idx2 >= min_idx_gap) {
        last_idx <<- idx
        return(TRUE)
      }
      return(FALSE)

    }, FALSE)]

    pulse_onset_idx
  }

  # debounce
  find_onsets <- function(threshold, sdff) {
    pulse_onset_idx <- which(abs(sdff) > threshold) - 1L
    pulse_onset_idx[pulse_onset_idx <= 0] <- 1L

    # find all the onsets
    last_idx <- -Inf
    pulse_onset_idx <- pulse_onset_idx[vapply(pulse_onset_idx, function(idx) {

      last_idx2 <- last_idx
      last_idx <<- idx
      if (idx - last_idx2 >= min_idx_gap) {
        return(TRUE)
      }
      return(FALSE)

    }, FALSE)]

    pulse_onset_idx
  }

  match_n_pulses <- function(min_threshold, max_threshold, n_pulses, sdff) {
    # print(c(min_threshold, max_threshold))
    pulse_onset_idx1 <- find_onsets(min_threshold, sdff = sdff)
    if (length(pulse_onset_idx1) <= n_pulses || (max_threshold - min_threshold) <= 0) {
      return(list(
        pulse_onset_idx = pulse_onset_idx1,
        threshold = min_threshold
      ))
    }

    pulse_onset_idx2 <- find_onsets(max_threshold, sdff = sdff)
    if (length(pulse_onset_idx2) >= n_pulses ||
       length(pulse_onset_idx2) == length(pulse_onset_idx1)) {
      return(list(
        pulse_onset_idx = pulse_onset_idx2,
        threshold = max_threshold
      ))
    }

    re <- match_n_pulses(
      min_threshold = min_threshold,
      max_threshold = (min_threshold + max_threshold) / 2,
      n_pulses = n_pulses,
      sdff = sdff
    )
    if (length(re$pulse_onset_idx) == n_pulses) { return(re) }

    re <- match_n_pulses(
      min_threshold = (min_threshold + max_threshold) / 2,
      max_threshold = max_threshold,
      n_pulses = n_pulses,
      sdff = sdff
    )
    if (length(re$pulse_onset_idx) == n_pulses) { return(re) }
  }

  if (is.na(threshold)) {
    if (is.na(n_pulses)) {
      threshold <- abs(median(signal_diff)) + 4 * stats::sd(signal_diff)
      pulse_onset_idx <- find_onsets(threshold, signal_diff)
      n_pulses <- length(pulse_onset_idx)

      # again, find a better threshold
      matched <- match_n_pulses(
        abs(median(signal_diff)) + 3 * stats::sd(signal_diff),
        signal_diff_ceiling,
        n_pulses = n_pulses,
        sdff = signal_diff
      )
      if (
        length(matched$pulse_onset_idx) == n_pulses &&
        all(abs(matched$pulse_onset_idx - pulse_onset_idx) < min_idx_gap)
      ) {
        pulse_onset_idx <- matched$pulse_onset_idx
        threshold <- matched$threshold
      }
    } else {

      matched <- match_n_pulses(
        abs(median(signal_diff)) + 3 * stats::sd(signal_diff),
        signal_diff_ceiling,
        n_pulses = n_pulses,
        sdff = signal_diff
      )
      pulse_onset_idx <- matched$pulse_onset_idx
      threshold <- matched$threshold
    }
  }

  # pulse onset are found pulse_onset_idx
  pulse_onset_idx <- as.integer(pulse_onset_idx)
  n_pulses <- length(pulse_onset_idx)

  # TODO: what if no pulse detected?
  if (!n_pulses) {
    stop("No stimulation pulse detected. Please set the signal threshold manually")
  }

  # revese the signal and find the offset
  match <- match_n_pulses(
    abs(median(signal_diff)) + 3 * stats::sd(signal_diff),
    signal_diff_ceiling,
    n_pulses = n_pulses,
    sdff = rev(signal_diff)
  )
  pulse_offset_idx <- as.integer(sort(length(signal) - match$pulse_onset_idx))

  # pulse_offset_idx might mismatch with pulse_onset_idx
  duration_idx <- vapply(pulse_onset_idx, function(idx) {
    tmp <- pulse_offset_idx[pulse_offset_idx > idx]
    if (length(tmp)) { return( tmp[[1]] - idx ) }
    return(NA_integer_)
  }, FUN.VALUE = 0L)

  if (all(is.na(duration_idx))) {
    duration_idx[] <- min_idx_gap
  } else {
    duration_idx[is.na(duration_idx)] <- max(duration_idx, na.rm = TRUE)
  }

  list(
    n_pulses = n_pulses,
    threshold = threshold,
    onset_index = pulse_onset_idx,
    offset_index = pulse_onset_idx + duration_idx
  )
}

#' @rdname stimpulse_interpolate
#' @export
stimpulse_extract <- function(signal, pulse_info, expand_timepoints = c(-10, 20), center = TRUE) {
  expand_timepoints <- round(expand_timepoints)
  expand_timepoints[[1]] <- min(expand_timepoints[[1]], 0)
  expand_timepoints[[2]] <- max(expand_timepoints[[2]], 0)
  pre_points <- -expand_timepoints[[1]]
  post_points <- expand_timepoints[[2]]
  onset_index <- pulse_info$onset_index + expand_timepoints[[1]]

  max_pulse_npoints <- max(pulse_info$offset_index - pulse_info$onset_index) + 1

  offset_index <- pulse_info$onset_index + max_pulse_npoints + expand_timepoints[[2]] - 1

  time_points_delta <- seq_len(max_pulse_npoints + expand_timepoints[[2]] - expand_timepoints[[1]]) - 1L

  pulses <- sapply(onset_index, function(idx) {
    signal[idx + time_points_delta]
  }, simplify = TRUE)

  if ( center ) {
    m <- apply(pulses, 2, median, na.rm = TRUE)
    pulses <- t(t(pulses) - m)
  }
}

#' @rdname stimpulse_interpolate
#' @export
stimpulse_align <- function(signal, pulse_info, expand_timepoints = c(-10, 20)) {

  # DIPSAUS DEBUG START
  # data("stimulation_signal")
  #
  # signal <- stimulation_signal$signal
  # sample_rate <- stimulation_signal$sample_rate
  # pulse_info <- stimpulse_find(signal, sample_rate, 0.001)
  #
  # pulses_snippets <- stimpulse_extract(signal, pulse_info, center = TRUE)
  # matplot(pulses_snippets, type = 'l', lty = 1, col = 'gray80')
  # lines(rowMeans(pulses_snippets), col = 'red')
  # expand_timepoints <- c(-10, 20)

  if (length(pulse_info$onset_index) == 0) {
    stop("No stimulation pulse is specified")
  }

  expand_timepoints <- round(expand_timepoints)
  expand_timepoints[[1]] <- min(expand_timepoints[[1]], 0)
  expand_timepoints[[2]] <- max(expand_timepoints[[2]], 0)
  pre_points <- -expand_timepoints[[1]]
  post_points <- expand_timepoints[[2]]

  max_pulse_npoints <- max(pulse_info$offset_index - pulse_info$onset_index) + 1

  onset_index <- pulse_info$onset_index + expand_timepoints[[1]]
  offset_index <- pulse_info$onset_index + max_pulse_npoints + expand_timepoints[[2]] - 1

  time_points_delta <- seq_len(max_pulse_npoints + expand_timepoints[[2]] - expand_timepoints[[1]]) - 1L

  pulses_centered <- stimpulse_extract(signal,
                                    pulse_info,
                                    expand_timepoints = expand_timepoints,
                                    center = TRUE)
  # stim_averaged <- apply(pulses_centered, 1L, median)
  # stim_averaged <- rowMeans(pulses_centered)
  stim_averaged <- apply(pulses_centered, 1L, function(x) {
    if (mean(x) > 0) {
      quantile(x, 0.75)
    } else {
      quantile(x, 0.25)
    }
  })
  # matplot(pulses_centered, type = 'l', lty = 1, col = 'gray80')
  # lines(stim_averaged, col = 'red')

  # for each pulses_centered, convolute with `stim_averaged` to find time offset
  # within +-10 points
  offsets_template <- seq(-pre_points, post_points)

  peaks <- find_peaks(abs(stim_averaged), min_distance = 2)$index

  if (length(peaks) > 0) {
    neg_mask <- rep(TRUE, length(offsets_template))
    neg_mask[unlist(lapply(peaks, function(idx) {
      idx + offsets_template
    }))] <- FALSE
  } else {
    neg_mask <- rep(FALSE, length(offsets_template))
  }
  offsets <- apply(pulses_centered, 2L, function(slice) {

    # slice <- pulses_centered[,10]


    conv <- sapply(offsets_template, function(offset) {

      if (offset > 0) {
        slice <- c(rep(NA_real_, offset), slice)[seq_along(stim_averaged)]
      } else if (offset < 0) {
        slice <- rev(c(rep(NA_real_, -offset), rev(slice))[seq_along(stim_averaged)])
      }
      # slice[neg_mask] <- NA_real_
      mean(stim_averaged * slice, na.rm = TRUE)

    })
    conv[is.na(conv)] <- -Inf
    offset <- offsets_template[which.max(conv)][[1]]
    offset
  })

  # local({
  #   onset_index2 <- onset_index - offsets
  #
  #   pulses <- sapply(onset_index2, function(idx) {
  #     signal[idx + time_points_delta]
  #   }, simplify = TRUE)
  #
  #   m <- apply(pulses, 2, median, na.rm = TRUE)
  #   pulses_centered <- t(t(pulses) - m)
  #   stim_averaged <- apply(pulses_centered, 1L, median)
  #   matplot(pulses_centered, type = 'l', lty = 1, col = 'gray80')
  #   lines(stim_averaged, col = 'red')
  # })

  pulse_info$onset_index <- pulse_info$onset_index - offsets
  pulse_info$offset_index <- pulse_info$offset_index - offsets

  pulse_info$aligned <- TRUE
  pulse_info
}

#' @rdname stimpulse_interpolate
#' @export
stimpulse_interpolate <- function(signal, sample_rate, pulse_info, max_offset = c(-0.0002, 0.0005)) {

  if (length(pulse_info$onset_index) == 0) {
    stop("No stimulation pulse is specified")
  }

  onset_index <- pulse_info$onset_index
  offset_index <- pulse_info$offset_index

  start <- c(0, offset_index)
  finish <- c(onset_index, length(signal))

  # load -pre_points to post_points to interpolate
  max_duration_npts <- max(offset_index - onset_index) + 1
  pre_points <- min(finish - start) - 1
  post_points <- max_duration_npts + min(finish - start, 100) - 1


  time_points_template <- seq(-pre_points, post_points)
  max_offset_pts <- ceiling(abs(max_offset) * sample_rate) * c(-1, 1)

  # # obtain signal slices and masks
  # mask <- time_points_template >= 0 & time_points_template <= max_duration_npts
  # signal_slices <- sapply(onset_index, function(idx) {
  #   slice <- signal[idx + time_points_template]
  #   slice
  # })
  #
  # # intercept <- apply(signal_slices, 2L, median, na.rm = TRUE)
  #
  # # centered_signals <- t(t(signal_slices) - intercept)
  #
  # mask_idx <- which(mask)
  # mask_idx_range <- range(mask_idx)
  # bandwidth_pts <- ceiling(min(mask_idx_range[[1]], length(time_points_template) - mask_idx_range[[2]], max_offset_pts)) - 1
  # bandwidth <- length(mask_idx) / length(time_points_template) * 2
  #
  # mask_expanded <- mask
  # mask_expanded_start <- mask_idx_range[[1]] - bandwidth_pts
  # mask_expanded_finish <- mask_idx_range[[2]] + bandwidth_pts
  # mask_expanded[seq(mask_expanded_start, mask_expanded_finish)] <- TRUE
  #
  #
  # fitted <- apply(signal_slices, 2, function(slice) {
  #   # slice <- signal_slices[,2]; slice0 <- slice
  #
  #   smoothed <- stats::lowess(time_points_template, slice, f = bandwidth, iter = 10)
  #   slice[mask_expanded] <- smoothed$y[mask_expanded]
  #
  #   # plot(slice0, type = 'l')
  #   # lines(smoothed$y, col = 'blue')
  #   # lines(slice, col = 'red')
  #
  #   slice
  # })
  #
  # env <- new.env(parent = emptyenv())
  # env$signal <- signal
  #
  # sapply(seq_along(onset_index), function(ii) {
  #   idx <- onset_index[[ii]]
  #   env$signal[idx + time_points_template] <- fitted[, ii]
  #   return()
  # })
  #
  # env$signal

  # prepare interpolation splines
  ntp <- length(time_points_template)

  plan <- cbind(
    c(-pre_points, -1 + max_offset_pts[[1]], (pre_points - max_offset_pts[[2]]) / ntp),
    c(max_duration_npts + max_offset_pts[[2]], post_points, (post_points - max_duration_npts - max_offset_pts[[2]] + 1) / ntp)
  )

  nknots <- 50
  ord <- 4
  regularization <- 0.01
  nknots2 <- ceiling(plan[3, ] / sum(plan[3, ]) * (nknots - ord))
  while (sum(nknots2) > nknots - ord) {
    tmp <- which.max(nknots2)
    if (nknots2[tmp] >= 2) {
      nknots2[tmp] <- nknots2[tmp] - 1
    } else {
      break
    }
  }
  plan[3, ] <- nknots2
  plan <- plan[, nknots2 > 0, drop = FALSE]


  knots <- apply(plan, 2, function(item) {
    time_start <- item[[1]]
    time_end <- item[[2]]
    nknots_ <- item[[3]]
    # Chebychev points
    # cos((2 * seq_len(nknots_) - 1) / (nknots_ * 2) * pi) * (time_end - time_start) / 2 + (time_start + time_end) / 2

    seq(time_start, time_end, length.out = nknots_)
  })

  knots <- sort(as.double(unlist(knots)))
  if (length(knots) > nknots - ord) {
    knots <- knots[seq_len(nknots - ord)]
  }

  knots <- c(
    rep(time_points_template[1], ord - 1),
    knots,
    rep(time_points_template[length(time_points_template)], nknots - length(knots) - 1)
  )

  B <- t(
    splines::splineDesign(
      knots,
      time_points_template,
      ord = ord,
      outer.ok = TRUE,
      sparse = FALSE
    )
  )
  if (!inherits(B, "matrix")) {
    class(B) <- "matrix"
  }
  bbt <- tcrossprod(B)
  diag(bbt) <- diag(bbt) + regularization
  bbtsolve <- qr.solve(bbt)

  # obtain masked signals
  mask <- time_points_template >= -1 + max_offset_pts[[1]] & time_points_template <= (max_duration_npts + max_offset_pts[[2]])
  signal_slices <- sapply(onset_index, function(idx) {
    slice <- signal[idx + time_points_template]
    slice
  })

  intercept <- apply(signal_slices, 2L, median, na.rm = TRUE)

  centered_signals <- t(t(signal_slices) - intercept)

  # slice <- centered_signals[,1]
  # plot(time_points_template, slice, type = "l")
  # lines(lowess(time_points_template, slice, f = (sum(mask) + 5) / length(mask)), col = 'red')

  # Initial fit
  gamma <- tcrossprod(t(centered_signals[!mask, , drop = FALSE]), B[, !mask]) %*% bbtsolve
  fitted <- t(gamma %*% B + intercept)

  masked_timepoints <- time_points_template[mask]
  # env <- new.env(parent = emptyenv())
  # env$signal <- signal

  replacements <- lapply(seq_along(onset_index), function(ii) {
    idx <- onset_index[[ii]]

    cbind(idx + masked_timepoints, fitted[mask, ii])
  })
  replacements <- do.call("rbind", replacements)

  signal[replacements[, 1]] <- replacements[, 2]

  structure(
    signal,
    onset_timepoints = onset_index,
    masked_offsets = time_points_template[mask]
  )
}


# data("stimulation_signal")
#
# signal <- stimulation_signal$signal
# sample_rate <- stimulation_signal$sample_rate
# pulse_info <- stimpulse_find(signal, sample_rate, 0.001)
#
# pulses_snippets <- stimpulse_extract(signal, pulse_info, center = TRUE)
# matplot(pulses_snippets, type = 'l', lty = 1, col = 'gray80')
# lines(rowMeans(pulses_snippets), col = 'red')
#
# pulse_info <- stimpulse_align(signal, pulse_info)
# pulses_snippets <- stimpulse_extract(signal, pulse_info, center = TRUE)
# matplot(pulses_snippets, type = 'l', lty = 1, col = 'gray80')
# lines(rowMeans(pulses_snippets), col = 'red')
#
# # pulse_info$offset_index <- pulse_info$offset_index + 30
# interpolated <- stimpulse_interpolate(signal, sample_rate, pulse_info, max_offset = c(5, 40) / sample_rate)
#
# # interpolated[is.na(interpolated)] <- 0
# ravetools::diagnose_channel((interpolated[seq_len(20000)]), signal, srate = sample_rate, try_compress = FALSE, name = c("interpotad", "orig"))

# # visualize
# time = seq_along(signal) / 30000
# plotly::plot_ly(type = "scattergl", mode = "lines") |>
#   plotly::add_trace(
#     # data = data.frame(time = tpoints / 30000,
#     #                   snippet = snippet,
#     #                   interpolated = tmp),
#     x = time,
#     y = signal,
#     line = list(
#       color = "gray"
#     )
#   ) |>
#   plotly::add_trace(
#     x = time,
#     y = interpolated,
#     line = list(
#       color = "red"
#     )
#   )
# plotly::add_trace(
#   x = ravetools::decimate(time, 3, ftype = 'fir'),
#   y = ravetools::decimate(signal, 3, ftype = 'iir'),
#   line = list(
#     color = "black"
#   )
# ) |>
# plotly::add_trace(
#   x = ravetools::decimate(time, 3, ftype = 'iir'),
#   y = ravetools::decimate(interpolated, 3, ftype = 'iir'),
#   line = list(
#     color = "blue"
#   )
# )

#' @title Sample '3D' volume in the world (anatomical \code{'RAS'}) space
#' @description
#' Low-level implementation to sample a '3D' volume into a given orientation and
#' shape using nearest-neighbor, \verb{trilinear}, or cubic B-spline
#' interpolation.
#' @param x image (volume) to be sampled: \code{dim(x)} must have length of 3
#' @param new_dim target dimension, integers of length 3
#' @param vox2ras_old from volume index (column-row-slice) to \code{'RAS'}
#' (right-anterior-superior) transform: the volume index starts from 0 (C-style)
#' instead of 1 (R-style) to comply with \code{'NIfTI'} transform.
#' @param vox2ras_new the targeting transform from volume index to \code{'RAS'}
#' @param na_fill default numbers to fill if a pixel is out of bound; default is
#' \code{NA} or \code{as.raw(0)} if input \code{x} is raw type
#' @param interpolation interpolation method: \code{'nearest'} (default),
#' \code{'trilinear'}, or \code{'bspline'} (cubic \verb{Catmull-Rom}).
#' \verb{Trilinear} and B-spline are only applied when \code{x} is a
#' \code{double} volume; other storage types silently fall back to
#' nearest-neighbor.
#' @returns A newly sampled volume that aligns with \code{x} in the anatomical
#' \code{'RAS'} coordinate system. The underlying storage mode is the same as
#' \code{x}
#'
#' @examples
#'
#'
#' # up-sample and rotate image
#' x <- array(0, c(9, 9, 9))
#' x[4:6, 4:6, 4:6] <- 1
#' vox2ras <- matrix(nrow = 4, byrow = TRUE, c(
#'   0.7071, -0.7071, 0, 0,
#'   0.7071, 0.7071, 0, -5.5,
#'   0, 0, 1, -4,
#'   0, 0, 0, 1
#' ))
#'
#' new_vox2ras <- matrix(nrow = 4, byrow = TRUE, c(
#'   0, 0.5, 0, -4,
#'   0, 0, -0.5, 4,
#'   0.5, 0, 0, -4,
#'   0, 0, 0, 1
#' ))
#'
#' y <- resample_3d_volume(
#'   x,
#'   c(17, 17, 17),
#'   vox2ras_old = vox2ras,
#'   vox2ras_new = new_vox2ras,
#'   na_fill = 0
#' )
#'
#'
#' image(y[9,,])
#'
#'
#'
#' @export
resample_3d_volume <- function(x, new_dim, vox2ras_old, vox2ras_new = vox2ras_old, na_fill = NA,
                               interpolation = c("nearest", "trilinear", "bspline")) {
  verify_dim <- function(dm, name) {
    if (length(dm) < 3) {
      stop(sprintf("`resample_3d_volume`: %s must be a 3-dimensional volume.", name))
    }
    if (length(dm) > 3 &&
        !all(dm[-c(1, 2, 3)] == 1)) {
      stop(
        sprintf("`resample_3d_volume`: %s has more than 3-dimensions and the extra ", name),
        "dimension has more than 1 component. Please make sure the volume dimension ",
        "has length of 3.")
    }
    dm[1:3]
  }
  verify_dim(dim(x), "Input `x`")
  new_dim <- verify_dim(new_dim, "Output")

  vox2ras_old <- as.matrix(vox2ras_old)
  if (!length(vox2ras_old) %in% c(12, 16) || ncol(vox2ras_old) != 4) {
    stop(
      "`resample_3d_volume`: `vox2ras_old` must be a 3x4 or 4x4 matrix ",
      "from the 0-indexing space to Right-Anterior-Superior coordinate system."
    )
  }

  vox2ras_new <- as.matrix(vox2ras_new)
  if (!length(vox2ras_new) %in% c(12, 16) || ncol(vox2ras_new) != 4) {
    stop(
      "`resample_3d_volume`: `vox2ras_new` must be a 3x4 or 4x4 matrix ",
      "from the 0-indexing space to Right-Anterior-Superior coordinate system."
    )
  }

  interpolation <- match.arg(interpolation)
  interp_code <- switch(interpolation, trilinear = 1L, bspline = 2L, 0L)

  if (storage.mode(x) == "raw" ) {
    if (is.na(na_fill)) {
      na_fill <- as.raw(0)
    } else {
      storage.mode(na_fill) <- storage.mode(x)
    }
  } else {
    storage.mode(na_fill) <- storage.mode(x)
  }

  re <- resample3D(
    arrayDim = new_dim,
    fromArray = x,
    newVoxToWorldTransposed = t(vox2ras_new),
    oldVoxToWorldTransposed = t(vox2ras_old),
    na = na_fill,
    interpolation = interp_code
  )
  resampled <- re[[1]]

  vox2vox_orig <- re[[2]]
  dim(vox2vox_orig) <- c(4, 4)
  attr(resampled, "vox2ras") <- vox2ras_new
  attr(resampled, "vox2vox_orig") <- vox2vox_orig
  resampled
}

#' @name raw-to-sexp
#' @title Convert raw vectors to R vectors
#' @param x raw vector of bytes
#' @returns Numeric vectors, except for \code{raw_to_string}, which returns
#' a string.
#' @details For numeric conversions, the function names are straightforward.
#' For example,
#' \code{raw_to_uintN} converts raw vectors to unsigned integers, and
#' \code{raw_to_intN} converts raw vectors to signed integers. The number
#' \code{'N'} stands for the number of bits used to store the integer.
#' For example \code{raw_to_uint8} uses 8 bits (1 byte) to store an integer,
#' hence the value range is \code{0-255}.
#'
#' The input data length must be multiple of the element size represented by
#' the underlying data. For example \code{uint16} integer uses 16 bites, and
#' one raw number uses 8 bits, hence two raw vectors can form one unsigned
#' integer-16. That is, \code{raw_to_uint16} requires the length of input
#' to be multiple of two. An easy calculation is: the length of \code{x} times
#' 8, must be divided by \code{'N'} (see last paragraph for definition).
#'
#' The returned data uses the closest available R native data type that can
#' fully represent the data. For example, R does not have single \code{float}
#' type, hence \code{raw_to_float} returns \code{double} type, which can
#' represent all possible values in \code{float}. For \code{raw_to_uint32},
#' the potential value range is \code{0 - (2^32-1)}. This exceeds the limit of
#' R integer type \code{(-2^31) - (2^31-1)}. Therefore, the returned values
#' will be real (double float) data type.
#'
#' There is no native data type that can store integer-64 data in R, package
#' \code{bit64} provides \code{integer64} type, which will be used by
#' \code{raw_to_int64}. Currently there is no solution to convert raw to
#' unsigned integer-64 type.
#'
#' \code{raw_to_string} converts raw to character string. This function respects
#' \code{null} character, hence is slightly different than the native
#' \code{\link{rawToChar}}, which translates raw byte-by-byte. If each
#' raw byte represents a valid character, then the above two functions returns
#' the same result. However, when the characters represented by raw bytes are
#' invalid, \code{raw_to_string} will stop parsing and returns only the valid
#' characters, while \code{\link{rawToChar}} will still try to parse, and
#' most likely to result in errors.
#' Please see Examples for comparisons.
#'
#' @examples
#'
#' # 0x00, 0x7f, 0x80, 0xFF
#' x <- as.raw(c(0, 127, 128, 255))
#'
#' raw_to_uint8(x)
#'
#' # The first bit becomes the integer sign
#' # 128 -> -128, 255 -> -1
#' raw_to_int8(x)
#'
#' ## Comments based on little endian system
#'
#' # 0x7f00 (32512), 0xFF80 (65408 unsigned, or -128 signed)
#' raw_to_uint16(x)
#' raw_to_int16(x)
#'
#' # 0xFF807F00 (4286611200 unsigned, -8356096 signed)
#' raw_to_uint32(x)
#' raw_to_int32(x)
#'
#' # ---------------------------- String ---------------------------
#'
#' # ASCII case: all valid
#' x <- charToRaw("This is an ASCII string")
#'
#' raw_to_string(x)
#' rawToChar(x)
#'
#' x <- c(charToRaw("This is the end."),
#'        as.raw(0),
#'        charToRaw("*** is invalid"))
#'
#' # rawToChar will raise error
#' raw_to_string(x)
#'
#' # ---------------------------- Integer64 ------------------------
#' # Runs on little endian system
#' x <- as.raw(c(0x80, 0x00, 0x7f, 0x80, 0xFF, 0x50, 0x7f, 0x00))
#'
#' # Calculate bitstring, which concaternates the followings
#' # 10000000 (0x80), 00000000 (0x00), 01111111 (0x7f), 10000000 (0x80),
#' # 11111111 (0xFF), 01010000 (0x50), 01111111 (0x7f), 00000000 (0x00)
#'
#' if(.Platform$endian == "little") {
#'   bitstring <- paste0(
#'     "00000000011111110101000011111111",
#'     "10000000011111110000000010000000"
#'   )
#' } else {
#'   bitstring <- paste0(
#'     "00000001000000001111111000000001",
#'     "11111111000010101111111000000000"
#'   )
#' }
#'
#' # This is expected value
#' bit64::as.integer64(structure(
#'   bitstring,
#'   class = "bitstring"
#' ))
#'
#' # This is actual value
#' raw_to_int64(x)
#'
#'
NULL


#' @rdname raw-to-sexp
#' @export
raw_to_uint8 <- function(x) { rawToUInt8(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_uint16 <- function(x) { rawToUInt16(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_uint32 <- function(x) { rawToUInt32(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_int8 <- function(x) { rawToInt8(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_int16 <- function(x) { rawToInt16(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_int32 <- function(x) { rawToInt32(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_int64 <- function(x) { rawToInt64(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_float <- function(x) { rawToFloat(x) }

#' @rdname raw-to-sexp
#' @export
raw_to_string <- function(x) { rawToString(x) }

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

baselineArray <- function(x, bl, dims, bldims, tidx, per, rest, method) {
    .Call(`_ravetools_baselineArray`, x, bl, dims, bldims, tidx, per, rest, method)
}

bucketFillVolume <- function(volume, x, y, z, fill) {
    .Call(`_ravetools_bucketFillVolume`, volume, x, y, z, fill)
}

collapser_cplx <- function(x, keep, method = 1L, average = 0L) {
    .Call(`_ravetools_collapser_cplx`, x, keep, method, average)
}

collapser_real <- function(x, keep, method = 1L, average = 0L) {
    .Call(`_ravetools_collapser_real`, x, keep, method, average)
}

columnQuantile <- function(x, prob, naRm) {
    .Call(`_ravetools_columnQuantile`, x, prob, naRm)
}

columnMedian <- function(x, naRm) {
    .Call(`_ravetools_columnMedian`, x, naRm)
}

fastColMeans <- function(x, col, xDim) {
    .Call(`_ravetools_fastColMeans`, x, col, xDim)
}

quickQuantile <- function(x, prob, naRm) {
    .Call(`_ravetools_quickQuantile`, x, prob, naRm)
}

quickMedian <- function(x, naRm) {
    .Call(`_ravetools_quickMedian`, x, naRm)
}

fastcov <- function(x1, x2, col1 = NULL, col2 = NULL, df = -1.0) {
    .Call(`_ravetools_fastcov`, x1, x2, col1, col2, df)
}

fftw_r2c <- function(data, HermConj = 1L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_r2c`, data, HermConj, fftwplanopt, ret)
}

fftw_c2c <- function(data, inverse = 0L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_c2c`, data, inverse, fftwplanopt, ret)
}

fftw_c2r <- function(data, HermConj = 1L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_c2r`, data, HermConj, fftwplanopt, ret)
}

mvfftw_r2c <- function(data, fftwplanopt = 0L, HermConj = 0L, ret = NULL) {
    .Call(`_ravetools_mvfftw_r2c`, data, fftwplanopt, HermConj, ret)
}

mvfftw_c2c <- function(data, inverse = 0L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_mvfftw_c2c`, data, inverse, fftwplanopt, ret)
}

mvfftw_c2r <- function(data, fftwplanopt = 0L, retrows = 0L, ret = NULL) {
    .Call(`_ravetools_mvfftw_c2r`, data, fftwplanopt, retrows, ret)
}

fftw_r2c_2d <- function(data, HermConj = 1L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_r2c_2d`, data, HermConj, fftwplanopt, ret)
}

fftw_c2c_2d <- function(data, inverse = 0L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_c2c_2d`, data, inverse, fftwplanopt, ret)
}

fftw_r2c_3d <- function(data, HermConj = 1L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_r2c_3d`, data, HermConj, fftwplanopt, ret)
}

fftw_c2c_3d <- function(data, inverse = 0L, fftwplanopt = 0L, ret = NULL) {
    .Call(`_ravetools_fftw_c2c_3d`, data, inverse, fftwplanopt, ret)
}

conjugate <- function(data) {
    .Call(`_ravetools_conjugate`, data)
}

cpp_filter <- function(b, a, x, z) {
    .Call(`_ravetools_cpp_filter`, b, a, x, z)
}

Matrix4__new <- function() {
    .Call(`_ravetools_Matrix4__new`)
}

Matrix4__to_array <- function(self) {
    .Call(`_ravetools_Matrix4__to_array`, self)
}

Matrix4__from_array <- function(self, array, offset = 0L) {
    invisible(.Call(`_ravetools_Matrix4__from_array`, self, array, offset))
}

Matrix4__identity <- function(self) {
    invisible(.Call(`_ravetools_Matrix4__identity`, self))
}

Matrix4__copy <- function(self, m) {
    invisible(.Call(`_ravetools_Matrix4__copy`, self, m))
}

Matrix4__copy_position <- function(self, m) {
    invisible(.Call(`_ravetools_Matrix4__copy_position`, self, m))
}

Matrix4__extract_basis <- function(self, x, y, z) {
    invisible(.Call(`_ravetools_Matrix4__extract_basis`, self, x, y, z))
}

Matrix4__make_basis <- function(self, x, y, z) {
    invisible(.Call(`_ravetools_Matrix4__make_basis`, self, x, y, z))
}

Matrix4__extract_rotation <- function(self, m) {
    invisible(.Call(`_ravetools_Matrix4__extract_rotation`, self, m))
}

Matrix4__look_at <- function(self, eye, target, up) {
    invisible(.Call(`_ravetools_Matrix4__look_at`, self, eye, target, up))
}

Matrix4__multiply_matrices <- function(self, a, b) {
    invisible(.Call(`_ravetools_Matrix4__multiply_matrices`, self, a, b))
}

Matrix4__multiply_scalar <- function(self, s) {
    invisible(.Call(`_ravetools_Matrix4__multiply_scalar`, self, s))
}

Matrix4__determinant <- function(self) {
    .Call(`_ravetools_Matrix4__determinant`, self)
}

Matrix4__transpose <- function(self) {
    invisible(.Call(`_ravetools_Matrix4__transpose`, self))
}

Matrix4__set_position <- function(self, x, y, z) {
    invisible(.Call(`_ravetools_Matrix4__set_position`, self, x, y, z))
}

Matrix4__invert <- function(self) {
    invisible(.Call(`_ravetools_Matrix4__invert`, self))
}

Matrix4__scale <- function(self, v) {
    invisible(.Call(`_ravetools_Matrix4__scale`, self, v))
}

Matrix4__get_max_scale_on_axis <- function(self) {
    .Call(`_ravetools_Matrix4__get_max_scale_on_axis`, self)
}

Matrix4__make_translation <- function(self, x, y, z) {
    invisible(.Call(`_ravetools_Matrix4__make_translation`, self, x, y, z))
}

Matrix4__make_rotation_x <- function(self, theta) {
    invisible(.Call(`_ravetools_Matrix4__make_rotation_x`, self, theta))
}

Matrix4__make_rotation_y <- function(self, theta) {
    invisible(.Call(`_ravetools_Matrix4__make_rotation_y`, self, theta))
}

Matrix4__make_rotation_z <- function(self, theta) {
    invisible(.Call(`_ravetools_Matrix4__make_rotation_z`, self, theta))
}

Matrix4__make_rotation_axis <- function(self, axis, angle) {
    invisible(.Call(`_ravetools_Matrix4__make_rotation_axis`, self, axis, angle))
}

Matrix4__make_scale <- function(self, x, y, z) {
    invisible(.Call(`_ravetools_Matrix4__make_scale`, self, x, y, z))
}

Matrix4__make_shear <- function(self, xy, xz, yx, yz, zx, zy) {
    invisible(.Call(`_ravetools_Matrix4__make_shear`, self, xy, xz, yx, yz, zx, zy))
}

Matrix4__make_perspective <- function(self, left, right, top, bottom, near, far) {
    invisible(.Call(`_ravetools_Matrix4__make_perspective`, self, left, right, top, bottom, near, far))
}

Matrix4__make_orthographic <- function(self, left, right, top, bottom, near, far) {
    invisible(.Call(`_ravetools_Matrix4__make_orthographic`, self, left, right, top, bottom, near, far))
}

Quaternion__new <- function() {
    .Call(`_ravetools_Quaternion__new`)
}

Quaternion__set <- function(self, x, y, z, w) {
    invisible(.Call(`_ravetools_Quaternion__set`, self, x, y, z, w))
}

Quaternion__copy <- function(self, quaternion) {
    invisible(.Call(`_ravetools_Quaternion__copy`, self, quaternion))
}

Quaternion__to_array <- function(self) {
    .Call(`_ravetools_Quaternion__to_array`, self)
}

Quaternion__getX <- function(self) {
    .Call(`_ravetools_Quaternion__getX`, self)
}

Quaternion__setX <- function(self, v) {
    invisible(.Call(`_ravetools_Quaternion__setX`, self, v))
}

Quaternion__getY <- function(self) {
    .Call(`_ravetools_Quaternion__getY`, self)
}

Quaternion__setY <- function(self, v) {
    invisible(.Call(`_ravetools_Quaternion__setY`, self, v))
}

Quaternion__getZ <- function(self) {
    .Call(`_ravetools_Quaternion__getZ`, self)
}

Quaternion__setZ <- function(self, v) {
    invisible(.Call(`_ravetools_Quaternion__setZ`, self, v))
}

Quaternion__getW <- function(self) {
    .Call(`_ravetools_Quaternion__getW`, self)
}

Quaternion__setW <- function(self, v) {
    invisible(.Call(`_ravetools_Quaternion__setW`, self, v))
}

Quaternion__set_from_axis_angle <- function(self, axis, angle) {
    invisible(.Call(`_ravetools_Quaternion__set_from_axis_angle`, self, axis, angle))
}

Quaternion__set_from_rotation_matrix <- function(self, m, fix_qfac) {
    invisible(.Call(`_ravetools_Quaternion__set_from_rotation_matrix`, self, m, fix_qfac))
}

Quaternion__set_from_unit_vectors <- function(self, v_from, v_to) {
    invisible(.Call(`_ravetools_Quaternion__set_from_unit_vectors`, self, v_from, v_to))
}

Quaternion__angle_to <- function(self, q) {
    .Call(`_ravetools_Quaternion__angle_to`, self, q)
}

Quaternion__rotate_towards <- function(self, q, step) {
    invisible(.Call(`_ravetools_Quaternion__rotate_towards`, self, q, step))
}

Quaternion__slerp <- function(self, qb, t) {
    invisible(.Call(`_ravetools_Quaternion__slerp`, self, qb, t))
}

Quaternion__identity <- function(self) {
    invisible(.Call(`_ravetools_Quaternion__identity`, self))
}

Quaternion__invert <- function(self) {
    invisible(.Call(`_ravetools_Quaternion__invert`, self))
}

Quaternion__conjugate <- function(self) {
    invisible(.Call(`_ravetools_Quaternion__conjugate`, self))
}

Quaternion__dot <- function(self, v) {
    .Call(`_ravetools_Quaternion__dot`, self, v)
}

Quaternion__length_squared <- function(self) {
    .Call(`_ravetools_Quaternion__length_squared`, self)
}

Quaternion__length <- function(self) {
    .Call(`_ravetools_Quaternion__length`, self)
}

Quaternion__normalize <- function(self) {
    invisible(.Call(`_ravetools_Quaternion__normalize`, self))
}

Quaternion__multiply <- function(self, q) {
    invisible(.Call(`_ravetools_Quaternion__multiply`, self, q))
}

Quaternion__premultiply <- function(self, q) {
    invisible(.Call(`_ravetools_Quaternion__premultiply`, self, q))
}

Quaternion__multiply_quaternions <- function(self, a, b) {
    invisible(.Call(`_ravetools_Quaternion__multiply_quaternions`, self, a, b))
}

Vector3__new <- function() {
    .Call(`_ravetools_Vector3__new`)
}

Vector3__from_array <- function(self, array, offset = 0L, n_elems = -1L) {
    invisible(.Call(`_ravetools_Vector3__from_array`, self, array, offset, n_elems))
}

Vector3__resize <- function(self, n_elems) {
    invisible(.Call(`_ravetools_Vector3__resize`, self, n_elems))
}

Vector3__get_size <- function(self) {
    .Call(`_ravetools_Vector3__get_size`, self)
}

Vector3__to_array <- function(self, n_skip = 0L, max_n_elems = -1L) {
    .Call(`_ravetools_Vector3__to_array`, self, n_skip, max_n_elems)
}

Vector3__set_scalar <- function(self, value) {
    invisible(.Call(`_ravetools_Vector3__set_scalar`, self, value))
}

Vector3__set_x <- function(self, value) {
    invisible(.Call(`_ravetools_Vector3__set_x`, self, value))
}

Vector3__set_y <- function(self, value) {
    invisible(.Call(`_ravetools_Vector3__set_y`, self, value))
}

Vector3__set_z <- function(self, value) {
    invisible(.Call(`_ravetools_Vector3__set_z`, self, value))
}

Vector3__get_x <- function(self, i) {
    .Call(`_ravetools_Vector3__get_x`, self, i)
}

Vector3__get_y <- function(self, i) {
    .Call(`_ravetools_Vector3__get_y`, self, i)
}

Vector3__get_z <- function(self, i) {
    .Call(`_ravetools_Vector3__get_z`, self, i)
}

Vector3__get_item <- function(self, i) {
    .Call(`_ravetools_Vector3__get_item`, self, i)
}

Vector3__copy <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__copy`, self, v))
}

Vector3__add <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__add`, self, v))
}

Vector3__add_scalar <- function(self, s) {
    invisible(.Call(`_ravetools_Vector3__add_scalar`, self, s))
}

Vector3__add_vectors <- function(self, a, b) {
    invisible(.Call(`_ravetools_Vector3__add_vectors`, self, a, b))
}

Vector3__add_scaled <- function(self, v, s) {
    invisible(.Call(`_ravetools_Vector3__add_scaled`, self, v, s))
}

Vector3__sub <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__sub`, self, v))
}

Vector3__sub_scalar <- function(self, s) {
    invisible(.Call(`_ravetools_Vector3__sub_scalar`, self, s))
}

Vector3__sub_vectors <- function(self, a, b) {
    invisible(.Call(`_ravetools_Vector3__sub_vectors`, self, a, b))
}

Vector3__multiply <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__multiply`, self, v))
}

Vector3__multiply_scalar <- function(self, s) {
    invisible(.Call(`_ravetools_Vector3__multiply_scalar`, self, s))
}

Vector3__multiply_vectors <- function(self, a, b) {
    invisible(.Call(`_ravetools_Vector3__multiply_vectors`, self, a, b))
}

Vector3__apply_matrix3 <- function(self, m) {
    invisible(.Call(`_ravetools_Vector3__apply_matrix3`, self, m))
}

Vector3__apply_matrix4 <- function(self, m) {
    invisible(.Call(`_ravetools_Vector3__apply_matrix4`, self, m))
}

Vector3__apply_quaternion <- function(self, q) {
    invisible(.Call(`_ravetools_Vector3__apply_quaternion`, self, q))
}

Vector3__transform_direction <- function(self, m) {
    invisible(.Call(`_ravetools_Vector3__transform_direction`, self, m))
}

Vector3__divide <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__divide`, self, v))
}

Vector3__divide_scalar <- function(self, s) {
    invisible(.Call(`_ravetools_Vector3__divide_scalar`, self, s))
}

Vector3__min <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__min`, self, v))
}

Vector3__max <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__max`, self, v))
}

Vector3__clamp <- function(self, min, max) {
    invisible(.Call(`_ravetools_Vector3__clamp`, self, min, max))
}

Vector3__floor <- function(self) {
    invisible(.Call(`_ravetools_Vector3__floor`, self))
}

Vector3__ceil <- function(self) {
    invisible(.Call(`_ravetools_Vector3__ceil`, self))
}

Vector3__round <- function(self) {
    invisible(.Call(`_ravetools_Vector3__round`, self))
}

Vector3__round_to_zero <- function(self) {
    invisible(.Call(`_ravetools_Vector3__round_to_zero`, self))
}

Vector3__negate <- function(self) {
    invisible(.Call(`_ravetools_Vector3__negate`, self))
}

Vector3__dot <- function(self, v) {
    .Call(`_ravetools_Vector3__dot`, self, v)
}

Vector3__length_squared <- function(self) {
    .Call(`_ravetools_Vector3__length_squared`, self)
}

Vector3__length <- function(self) {
    .Call(`_ravetools_Vector3__length`, self)
}

Vector3__length_manhattan <- function(self) {
    .Call(`_ravetools_Vector3__length_manhattan`, self)
}

Vector3__normalize <- function(self) {
    invisible(.Call(`_ravetools_Vector3__normalize`, self))
}

Vector3__set_length <- function(self, length) {
    invisible(.Call(`_ravetools_Vector3__set_length`, self, length))
}

Vector3__lerp <- function(self, v, alpha) {
    invisible(.Call(`_ravetools_Vector3__lerp`, self, v, alpha))
}

Vector3__lerp_vectors <- function(self, v1, v2, alpha) {
    invisible(.Call(`_ravetools_Vector3__lerp_vectors`, self, v1, v2, alpha))
}

Vector3__cross <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__cross`, self, v))
}

Vector3__cross_vectors <- function(self, a, b) {
    invisible(.Call(`_ravetools_Vector3__cross_vectors`, self, a, b))
}

Vector3__project_on_vector <- function(self, v) {
    invisible(.Call(`_ravetools_Vector3__project_on_vector`, self, v))
}

Vector3__project_on_plane <- function(self, planeNormal) {
    invisible(.Call(`_ravetools_Vector3__project_on_plane`, self, planeNormal))
}

Vector3__reflect <- function(self, normal) {
    invisible(.Call(`_ravetools_Vector3__reflect`, self, normal))
}

Vector3__angle_to <- function(self, v) {
    .Call(`_ravetools_Vector3__angle_to`, self, v)
}

Vector3__distance_to <- function(self, v) {
    .Call(`_ravetools_Vector3__distance_to`, self, v)
}

Vector3__distance_to_squared <- function(self, v) {
    .Call(`_ravetools_Vector3__distance_to_squared`, self, v)
}

Vector3__distance_to_manhattan <- function(self, v) {
    .Call(`_ravetools_Vector3__distance_to_manhattan`, self, v)
}

Vector3__set_from_spherical_coords <- function(self, radius, phi, theta) {
    invisible(.Call(`_ravetools_Vector3__set_from_spherical_coords`, self, radius, phi, theta))
}

Vector3__set_from_matrix_position <- function(self, m) {
    invisible(.Call(`_ravetools_Vector3__set_from_matrix_position`, self, m))
}

Vector3__set_from_matrix_scale <- function(self, m) {
    invisible(.Call(`_ravetools_Vector3__set_from_matrix_scale`, self, m))
}

rawToUInt8 <- function(x) {
    .Call(`_ravetools_rawToUInt8`, x)
}

rawToInt8 <- function(x) {
    .Call(`_ravetools_rawToInt8`, x)
}

rawToUInt16 <- function(x) {
    .Call(`_ravetools_rawToUInt16`, x)
}

rawToInt16 <- function(x) {
    .Call(`_ravetools_rawToInt16`, x)
}

rawToUInt32 <- function(x) {
    .Call(`_ravetools_rawToUInt32`, x)
}

rawToInt32 <- function(x) {
    .Call(`_ravetools_rawToInt32`, x)
}

rawToInt64 <- function(x) {
    .Call(`_ravetools_rawToInt64`, x)
}

rawToFloat <- function(x) {
    .Call(`_ravetools_rawToFloat`, x)
}

rawToString <- function(x) {
    .Call(`_ravetools_rawToString`, x)
}

resample3D <- function(arrayDim, fromArray, newVoxToWorldTransposed, oldVoxToWorldTransposed, na) {
    .Call(`_ravetools_resample3D`, arrayDim, fromArray, newVoxToWorldTransposed, oldVoxToWorldTransposed, na)
}

shiftArray <- function(x, alongIdx, unitIdx, shiftAmount) {
    .Call(`_ravetools_shiftArray`, x, alongIdx, unitIdx, shiftAmount)
}

getDefaultNumThreads <- function() {
    .Call(`_ravetools_getDefaultNumThreads`)
}

vcgIsoSurface <- function(array_, thresh) {
    .Call(`_ravetools_vcgIsoSurface`, array_, thresh)
}

vcgSmoothImplicit <- function(vb_, it_, lambda_, useMassMatrix, fixBorder, useCotWeight, degree, lapWeight_, SmoothQ) {
    .Call(`_ravetools_vcgSmoothImplicit`, vb_, it_, lambda_, useMassMatrix, fixBorder, useCotWeight, degree, lapWeight_, SmoothQ)
}

vcgSmooth <- function(vb_, it_, iter, method, lambda, mu, delta_) {
    .Call(`_ravetools_vcgSmooth`, vb_, it_, iter, method, lambda, mu, delta_)
}

vcgUniformResample <- function(vb_, it_, voxelSize, offsetThr, discretizeFlag, multiSampleFlag, absDistFlag, mergeCloseVert, silent) {
    .Call(`_ravetools_vcgUniformResample`, vb_, it_, voxelSize, offsetThr, discretizeFlag, multiSampleFlag, absDistFlag, mergeCloseVert, silent)
}

vcgUpdateNormals <- function(vb_, it_, select, pointcloud, silent) {
    .Call(`_ravetools_vcgUpdateNormals`, vb_, it_, select, pointcloud, silent)
}

vcgEdgeSubdivision <- function(vb_, it_) {
    .Call(`_ravetools_vcgEdgeSubdivision`, vb_, it_)
}

vcgVolume <- function(mesh_) {
    .Call(`_ravetools_vcgVolume`, mesh_)
}

vcgSphere <- function(subdiv, normals) {
    .Call(`_ravetools_vcgSphere`, subdiv, normals)
}

vcgDijkstra <- function(vb_, it_, source, maxdist_) {
    .Call(`_ravetools_vcgDijkstra`, vb_, it_, source, maxdist_)
}

vcgRaycaster <- function(vb_, it_, rayOrigin, rayDirection, maxDistance, bothSides, threads = 1L) {
    .Call(`_ravetools_vcgRaycaster`, vb_, it_, rayOrigin, rayDirection, maxDistance, bothSides, threads)
}

vcgKDTreeSearch <- function(target_, query_, k, nPointsPerCell = 16L, maxDepth = 64L) {
    .Call(`_ravetools_vcgKDTreeSearch`, target_, query_, k, nPointsPerCell, maxDepth)
}

vcgSubset <- function(vb_, it_, selector_) {
    .Call(`_ravetools_vcgSubset`, vb_, it_, selector_)
}

# Register entry points for exported C++ functions
methods::setLoadAction(function(ns) {
    .Call(`_ravetools_RcppExport_registerCCallable`)
})

test_that("fftw_r2c_2d", {
  # even margins
  x <- array(rnorm(1000), c(20, 50))

  a <- fftwtools::fftw_r2c_2d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_2d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1000)
  a <- fftwtools::fftw_r2c_2d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_2d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1000, 1)
  a <- t(fftwtools::fftw_r2c_2d(t(x), HermConj = 1))
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- t(fftwtools::fftw_r2c_2d(t(x), HermConj = 0))[1:501, 1, drop = FALSE]
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)


  # odd
  x <- array(rnorm(1071), c(21, 51))

  a <- fftwtools::fftw_r2c_2d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_2d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1071)
  a <- fftwtools::fftw_r2c_2d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_2d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1071, 1)
  a <- t(fftwtools::fftw_r2c_2d(t(x), HermConj = 1))
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- t(fftwtools::fftw_r2c_2d(t(x), HermConj = 0))[1:536, 1, drop = FALSE]
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  # zero length/margin
  x <- numeric(0)
  dim(x) <- c(0, 100)

  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(dim(b), dim(x))

  dim(x) <- c(100, 0)
  b <- ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(dim(b), c(51, 0))

})



test_that("fftw_c2c_2d", {
  # even margins
  x <- array(rnorm(1000), c(20, 50))

  a <- fftwtools::fftw_c2c_2d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_2d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1000)
  a <- fftwtools::fftw_c2c_2d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_2d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1000, 1)
  a <- fftwtools::fftw_c2c_2d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_2d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)


  # odd
  x <- array(rnorm(1071), c(21, 51))

  a <- fftwtools::fftw_c2c_2d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_2d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1071)
  a <- fftwtools::fftw_c2c_2d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_2d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1071, 1)
  a <- fftwtools::fftw_c2c_2d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_2d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  # zero length/margin
  x <- numeric(0)
  dim(x) <- c(0, 100)

  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(dim(b), dim(x))

  dim(x) <- c(100, 0)
  b <- ravetools:::fftw_c2c_2d(x, inverse = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(dim(b), dim(x))

})

test_that("fftw_c2c_2d input immutability and roundtrip", {
  set.seed(7)
  x <- matrix(complex(real = rnorm(200), imaginary = rnorm(200)),
              nrow = 20, ncol = 10)
  x_orig <- x + 0  # deep copy

  fwd <- ravetools:::fftw_c2c_2d(x, inverse = 0)
  testthat::expect_equal(x, x_orig)

  fwd_meas <- ravetools:::fftw_c2c_2d(x, inverse = 0, fftwplanopt = 1L)
  testthat::expect_equal(x, x_orig)
  testthat::expect_equal(fwd_meas, fwd, tolerance = 1e-10)

  inv <- ravetools:::fftw_c2c_2d(fwd, inverse = 1)
  testthat::expect_equal(inv / length(x), x, tolerance = 1e-10)

  testthat::expect_equal(x, x_orig)
})

test_that("fftw_r2c_2d input immutability", {
  set.seed(8)
  x <- matrix(rnorm(200), nrow = 20, ncol = 10)
  x_orig <- x + 0

  ravetools:::fftw_r2c_2d(x, HermConj = 1)
  testthat::expect_equal(x, x_orig)

  ravetools:::fftw_r2c_2d(x, HermConj = 0)
  testthat::expect_equal(x, x_orig)

  ravetools:::fftw_r2c_2d(x, HermConj = 1, fftwplanopt = 1L)
  testthat::expect_equal(x, x_orig)
})

test_that("fftw_*_2d match stats::fft", {
  set.seed(91)
  # real input
  X <- matrix(rnorm(120), nrow = 10, ncol = 12)
  ref <- stats::fft(X + 0i)
  testthat::expect_equal(
    ravetools:::fftw_r2c_2d(X, HermConj = 1), ref, tolerance = 1e-10
  )
  testthat::expect_equal(
    ravetools:::fftw_c2c_2d(X + 0i, inverse = 0), ref, tolerance = 1e-10
  )

  # complex input
  Z <- matrix(complex(real = rnorm(120), imaginary = rnorm(120)),
              nrow = 10, ncol = 12)
  testthat::expect_equal(
    ravetools:::fftw_c2c_2d(Z, inverse = 0),
    stats::fft(Z),
    tolerance = 1e-10
  )
  testthat::expect_equal(
    ravetools:::fftw_c2c_2d(Z, inverse = 1),
    stats::fft(Z, inverse = TRUE),
    tolerance = 1e-10
  )
})

require(testthat)
test_that("fftw_r2c", {
  set.seed(1)
  x <- rnorm(1000) + 1
  c <- complex(1000)

  expect_error(ravetools:::fftw_r2c(
    data = x,
    HermConj = 1,
    ret = complex(999)
  ))
  expect_error(ravetools:::fftw_r2c(
    data = x,
    HermConj = 0,
    ret = complex(500)
  ))

  expect_equal(ravetools:::fftw_r2c(x, HermConj = 1), stats::fft(x))
  expect_equal(ravetools:::fftw_r2c(x, HermConj = 0, fftwplanopt = 1),
               stats::fft(x)[1:501])

  # make sure not edit in-place
  set.seed(1)
  xx <- rnorm(1000)
  expect_equal(x, xx + 1)

  c <- complex(1000)
  ravetools:::fftw_r2c(data = x,
                       HermConj = 1,
                       ret = c)
  expect_equal(c, stats::fft(x))

  expect_equal(c[1:501], fftwtools::fftw_r2c(x, 0))


  expect_equal(x, xx + 1)


})


test_that("mvfftw_r2c", {
  set.seed(1)
  x <- rnorm(1000)
  dim(x) <- c(100, 10)
  c <- apply(x, 2, stats::fft)[1:51, ]
  a <- ravetools:::mvfftw_r2c(x, 0)
  cc <- fftwtools::mvfftw_r2c(x, 1)[1:51, ]
  expect_equal(a, c)
  expect_equal(a, cc)

  set.seed(1)
  xx <- rnorm(1000)
  dim(xx) <- c(100, 10)
  expect_equal(x, xx)

  e <- complex(length(a))
  ravetools:::mvfftw_r2c(x, ret = e)
  expect_equal(e, c)

  set.seed(1)
  x <- rnorm(1000)
  dim(x) <- c(100, 10)
  b <- ravetools:::mvfftw_r2c(x, 1L)
  # d <- fftwtools::mvfftw_r2c(x, 1, 1)[1:51,]
  expect_equal(b, c)
  # expect_equal(b, d)

  set.seed(1)
  xx <- rnorm(1000)
  dim(xx) <- c(100, 10)
  expect_equal(x, xx)
})

test_that("mvfftw_r2c HermConj", {
  set.seed(1)
  x <- matrix(rnorm(1000), nrow = 100, ncol = 10)
  xx <- x + 1

  ref <- stats::mvfft(x)  # 100 x 10 complex

  # --- HermConj=1 returns full n x m matrix matching stats::mvfft ---
  full <- ravetools:::mvfftw_r2c(x, HermConj = 1L)
  expect_true(is.matrix(full))
  expect_equal(dim(full), c(100L, 10L))
  expect_equal(full, ref)

  # --- input immutability ---
  expect_equal(xx - 1, x)

  # --- half-spectrum (default HermConj=0) is consistent prefix ---
  half <- ravetools:::mvfftw_r2c(x)
  expect_equal(half, full[1:51, ])

  # --- pre-allocated ret for HermConj=1 ---
  ret_full <- matrix(0 + 0i, nrow = 100, ncol = 10)
  ravetools:::mvfftw_r2c(x, HermConj = 1L, ret = ret_full)
  expect_equal(ret_full, ref)

  expect_equal(xx - 1, x)

  # --- fftwplanopt=1 with HermConj=1 (exercises non-ESTIMATE path) ---
  full_meas <- ravetools:::mvfftw_r2c(x, fftwplanopt = 1L, HermConj = 1L)
  expect_equal(full_meas, ref)

  expect_equal(xx - 1, x)

  # --- odd nrows (101): conjugate mirror for odd-length DFT ---
  x_odd <- matrix(rnorm(1010), nrow = 101, ncol = 10)
  ref_odd <- stats::mvfft(x_odd)
  full_odd <- ravetools:::mvfftw_r2c(x_odd, HermConj = 1L)
  expect_equal(dim(full_odd), c(101L, 10L))
  expect_equal(full_odd, ref_odd)
})


test_that("fftw_c2r", {
  set.seed(1)
  x <- rnorm(1000) + 1i * rnorm(1000)
  xx <- x + 1
  c <- double(1000)

  expect_error(ravetools:::fftw_c2r(
    data = x,
    HermConj = 1,
    ret = double(999)
  ))
  expect_error(ravetools:::fftw_c2r(
    data = x,
    HermConj = 0,
    ret = double(1000)
  ))

  expect_equal(x + 1, xx)

  expect_equal(ravetools:::fftw_c2r(x, HermConj = 1),
               fftwtools::fftw_c2r((xx - 1), HermConj = 1, n = 1000))
  expect_equal(x + 1, xx)

  expect_equal(
    ravetools:::fftw_c2r(x, HermConj = 0, fftwplanopt = 1),
    fftwtools::fftw_c2r((xx - 1), HermConj = 0, n = 1998)
  )
  expect_equal(x + 1, xx)

  ravetools:::fftw_c2r(data = x,
                       HermConj = 1,
                       ret = c)
  expect_equal(c, fftwtools::fftw_c2r((xx - 1), HermConj = 1))
  expect_equal(x + 1, xx)

  expect_equal(c, fftwtools::fftw_c2r((xx - 1), HermConj = 1))

  expect_equal(x + 1, xx)


})

test_that("fftw_c2c", {
  set.seed(1)
  x <- rnorm(1000) + 1i * rnorm(1000)
  xx <- x + 1
  c <- complex(1000)

  expect_error(ravetools:::fftw_c2c(data = x, ret = double(1000)))
  expect_error(ravetools:::fftw_c2c(data = x, ret = complex(999)))

  expect_equal(x + 1, xx)

  expect_equal(ravetools:::fftw_c2c(x), fftwtools::fftw_c2c((xx - 1)))
  expect_equal(x + 1, xx)

  expect_equal(ravetools:::fftw_c2c(x, fftwplanopt = 1L, ret = c),
               fftwtools::fftw_c2c((xx - 1)))
  expect_equal(x + 1, xx)

  expect_equal(
    ravetools:::fftw_c2c(
      x,
      inverse = TRUE,
      fftwplanopt = 1L,
      ret = c
    ),
    fftwtools::fftw_c2c((xx - 1), inverse = TRUE)
  )
  expect_equal(x + 1, xx)

  expect_equal(c, fftwtools::fftw_c2c((xx - 1), inverse = TRUE))

  # inplace with ret == data
  expect_val <- fftwtools::fftw_c2c((xx - 1), inverse = TRUE)
  expect_equal(ravetools:::fftw_c2c(x, inverse = TRUE, ret = x), expect_val)

  expect_equal(x, expect_val)


})

test_that("mvfftw_c2c", {
  set.seed(1)
  x <- matrix(rnorm(1000) + 1i * rnorm(1000), nrow = 100, ncol = 10)
  xx <- x + 1

  # --- input validation ---
  expect_error(ravetools:::mvfftw_c2c(data = x, ret = double(1000)))
  expect_error(ravetools:::mvfftw_c2c(data = x, ret = complex(999)))

  # input not modified
  expect_equal(x + 1, xx)

  # --- forward transform matches stats::mvfft ---
  fwd <- ravetools:::mvfftw_c2c(x)
  expect_true(is.matrix(fwd))
  expect_equal(dim(fwd), dim(x))
  expect_equal(fwd, stats::mvfft(xx - 1))

  # input still unmodified
  expect_equal(x + 1, xx)

  # --- inverse transform matches stats::mvfft(inverse = TRUE) ---
  inv <- ravetools:::mvfftw_c2c(x, inverse = 1L)
  expect_true(is.matrix(inv))
  expect_equal(dim(inv), dim(x))
  expect_equal(inv, stats::mvfft(xx - 1, inverse = TRUE))

  expect_equal(x + 1, xx)

  # --- round-trip: inverse(forward(x)) / nrow == x ---
  rt <- ravetools:::mvfftw_c2c(
    ravetools:::mvfftw_c2c(xx - 1),
    inverse = 1L
  ) / nrow(x)
  expect_equal(rt, xx - 1)

  # --- pre-allocated ret (forward) ---
  ret_fwd <- matrix(0 + 0i, nrow = 100, ncol = 10)
  result <- ravetools:::mvfftw_c2c(x, ret = ret_fwd)
  expect_equal(result, stats::mvfft(xx - 1))
  expect_equal(ret_fwd, stats::mvfft(xx - 1))   # ret filled in-place

  expect_equal(x + 1, xx)

  # --- pre-allocated ret (inverse) ---
  ret_inv <- matrix(0 + 0i, nrow = 100, ncol = 10)
  result_inv <- ravetools:::mvfftw_c2c(x, inverse = 1L, ret = ret_inv)
  expect_equal(result_inv, stats::mvfft(xx - 1, inverse = TRUE))
  expect_equal(ret_inv, stats::mvfft(xx - 1, inverse = TRUE))

  expect_equal(x + 1, xx)

  # --- non-ESTIMATE plan (fftwplanopt = 1) ---
  fwd_meas <- ravetools:::mvfftw_c2c(x, fftwplanopt = 1L)
  expect_equal(fwd_meas, stats::mvfft(xx - 1))

  inv_meas <- ravetools:::mvfftw_c2c(x, inverse = 1L, fftwplanopt = 1L)
  expect_equal(inv_meas, stats::mvfft(xx - 1, inverse = TRUE))

  expect_equal(x + 1, xx)
})

test_that("mvfftw_c2r", {
  set.seed(1)
  x <- matrix(rnorm(1000), nrow = 100, ncol = 10)  # real input
  xx <- x + 1

  # --- half-spectrum from r2c (nrows=51 -> retrows=100) ---
  half <- ravetools:::mvfftw_r2c(x)                 # 51 × 10 complex
  half2 <- ravetools:::mvfftw_r2c(xx - 1)

  # input validation
  expect_error(ravetools:::mvfftw_c2r(data = half, ret = complex(510)))  # wrong type
  expect_error(ravetools:::mvfftw_c2r(data = half, ret = double(509)))   # wrong length
  expect_error(ravetools:::mvfftw_c2r(data = half, retrows = 99L))       # invalid retrows (nc=51 -> only 100 or 101 allowed)

  expect_equal(xx - 1, x)
  expect_equal(half, half2)
  gc()
  expect_equal(half, half2)

  # --- default retrows=0 auto-resolves to even (100) ---
  result <- ravetools:::mvfftw_c2r(half)
  expect_equal(half, half2)

  expect_true(is.matrix(result))
  expect_equal(dim(result), c(100L, 10L))
  # unnormalised inverse: matches stats::fft round-trip (Re only, no /n)
  expect_equal(result, Re(stats::mvfft(stats::mvfft(x), inverse = TRUE)))

  expect_equal(xx - 1, x)

  # --- round-trip: c2r(r2c(x)) / nrow == x ---
  rt <- ravetools:::mvfftw_c2r(ravetools:::mvfftw_r2c(x)) / nrow(x)
  expect_equal(rt, x)

  # --- explicit even retrows ---
  expect_equal(half, half2)
  result_even <- ravetools:::mvfftw_c2r(half, retrows = 100L)
  expect_equal(half, half2)
  expect_equal(result_even, result)

  # --- odd retrows (101 rows of output) ---
  # build half-spectrum from a 101-row real matrix
  x_odd <- matrix(rnorm(1010), nrow = 101, ncol = 10)
  half_odd <- ravetools:::mvfftw_r2c(x_odd)         # 51 × 10 (nc = 51)
  result_odd <- ravetools:::mvfftw_c2r(half_odd, retrows = 101L)
  expect_true(is.matrix(result_odd))
  expect_equal(dim(result_odd), c(101L, 10L))
  rt_odd <- ravetools:::mvfftw_c2r(half_odd, retrows = 101L) / 101
  expect_equal(rt_odd, x_odd)

  # --- pre-allocated ret ---
  ret_mat <- matrix(0.0, nrow = 100, ncol = 10)
  result2 <- ravetools:::mvfftw_c2r(half, ret = ret_mat)
  expect_equal(result2, result)
  expect_equal(ret_mat, result)      # filled in-place

  expect_equal(xx - 1, x)

  # --- non-ESTIMATE plan (fftwplanopt = 1) exercises the fixed memcpy ---
  result_meas <- ravetools:::mvfftw_c2r(half, fftwplanopt = 1L)
  expect_equal(result_meas, result)

  rt_meas <- ravetools:::mvfftw_c2r(ravetools:::mvfftw_r2c(x, fftwplanopt = 1L),
                                    fftwplanopt = 1L) / nrow(x)
  expect_equal(rt_meas, x)

  expect_equal(xx - 1, x)
})

library(testthat)

test_that("check conversion values are correct", {

  if (.Platform$endian == "little") {
    rev2 <- rev
  } else {
    rev2 <- function(x) { x }
  }


  bytes <- c("00", "02", "02", "ff", "80", "00", "7f", "80", "ff", "ff", "ff", "ff")
  bytes <- c(bytes, rev(bytes))

  num_ <- eval(parse(text = paste("c(", paste(sprintf("0x%s", bytes), collapse = ", "), ")")))
  x_ <- as.raw(num_)
  bits <- matrix(as.numeric(rawToBits(x_)), nrow = 8L)

  num_16_ <- eval(parse(text = paste("c(", paste(sprintf("0x%s", local({
    dim(bytes) <- c(2, length(bytes) / 2)
    apply(bytes, 2, function(x) { paste(rev2(x), sep = "", collapse = "") })
  })), collapse = ", "), ")")))

  num_32_ <- eval(parse(text = paste("c(", paste(sprintf("0x%s", local({
    dim(bytes) <- c(4, length(bytes) / 4)
    apply(bytes, 2, function(x) { paste(rev2(x), sep = "", collapse = "") })
  })), collapse = ", "), ")")))

  num_64_ <- local({
    dim(x_) <-  c(8, length(x_) / 8)
    bitstrings <- apply(x_, 2, function(x) {
      bits <- rev2(rawToBits(x))
      paste(as.integer(bits), collapse = "")
    })
    bit64::as.integer64(structure(bitstrings, class = "bitstring"))
  })

  float_ <- local({
    dim(bits) <- c(32, length(bits) / 32)
    b <- bits[, 2]
    apply(bits, 2, function(b) {
      b <- rev2(b)
      sign <- ifelse(b[1] == 0, 1, -1)
      e <- sum(b[2:9] * 2^(7:0))
      mantissa <- b[-(1:9)]
      if (!length(mantissa)) {
        mantissa <- 0
      } else {
        mantissa <- sum(mantissa * 2^(-seq_along(mantissa)))
      }
      if (e == 0) {
        # denormalization
        e <- -126
      } else {
        e <- e - 127
        mantissa <- mantissa + 1L
      }
      if (e == 128) {
        if (mantissa == 0) {
          return(sign * Inf)
        } else {
          return(NaN)
        }
      }
      sign * mantissa * 2^e
    })
  })

  for (len in seq_along(num_)) {

    num <- num_[seq_len(len)]

    x <- x_[seq_len(len)]

    expect_equal(rawToUInt8(x), num)
    expect_equal(rawToInt8(x), local({
      num[num > 127] <- num[num > 127] - 256L
      num
    }))

    if (len %% 2 != 0) {
      expect_error(rawToUInt16(x))
      expect_error(rawToInt16(x))
    } else {
      num_16 <- num_16_[seq_len(floor(len / 2))]
      expect_equal(rawToUInt16(x), num_16)
      expect_equal(rawToInt16(x), local({
        num_16[num_16 >= 2L^15L] <- num_16[num_16 >= 2L^15L] - 2L^16L
        num_16
      }))
    }

    if (len %% 4 != 0) {
      expect_error(rawToUInt32(x))
      expect_error(rawToInt32(x))
      expect_error(rawToFloat(x))
    } else {

      num_32 <- num_32_[seq_len(floor(len / 4))]
      float <- float_[seq_len(floor(len / 4))]

      expect_equal(rawToUInt32(x), num_32)
      expect_equal(rawToInt32(x), local({
        num_32[num_32 >= 2L^31L] <- num_32[num_32 >= 2L^31L] - 2L^32L
        num_32
      }))
      expect_equal(rawToFloat(x), float)
    }

    if (len %% 8 != 0) {
      expect_error(rawToInt64(x))
    } else {
      num_64 <- num_64_[seq_len(floor(len / 8))]
      expect_equal(rawToInt64(x), num_64)
    }

  }




})

test_that("fftw_r2c_3d", {
  # even margins
  x <- array(rnorm(10000), c(20, 50, 10))
  y <- as.vector(x)

  a <- fftwtools::fftw_r2c_3d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_3d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1000, 10)
  x1 <- x
  a <- fftwtools::fftw_r2c_3d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_3d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1000, 1, 10)
  a <- aperm(fftwtools::fftw_r2c_3d(x1, HermConj = 1), c(2, 1, 3))
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- a[1:501, 1, 1:10, drop = FALSE]
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1000, 10, 1)
  a <- aperm(fftwtools::fftw_r2c_3d(x1, HermConj = 1), c(2, 3, 1))
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- a[1:501, 1:10, 1, drop = FALSE]
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  # make sure x is not altered
  testthat::expect_equal(as.vector(x), y)

  # odd
  x <- array(rnorm(11781), c(21, 51, 11))
  y <- as.vector(x)
  a <- fftwtools::fftw_r2c_3d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_3d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1071, 11)
  x1 <- x
  a <- fftwtools::fftw_r2c_3d(x, HermConj = 1)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_r2c_3d(x, HermConj = 0)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1071, 1, 11)
  a <- aperm(fftwtools::fftw_r2c_3d(x1, HermConj = 1), c(2, 1, 3))
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- a[1:536, 1, 1:11, drop = FALSE]
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1071, 11, 1)
  a <- aperm(fftwtools::fftw_r2c_3d(x1, HermConj = 1), c(2, 3, 1))
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- a[1:536, 1:11, 1, drop = FALSE]
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  testthat::expect_equal(as.vector(x), y)

  # zero length/margin
  x <- numeric(0)
  dim(x) <- c(0, 100, 10)

  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(dim(b), dim(x))

  dim(x) <- c(100, 0, 10)
  b <- ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(dim(b), c(51, 0, 10))

})


test_that("fftw_c2c_3d", {
  # even margins
  x <- array(rnorm(10000), c(20, 50, 10))

  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1000, 10)
  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1000, 1, 10)
  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1000, 10, 1)
  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)


  # odd
  x <- array(rnorm(11781), c(21, 51, 11))

  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1, 1071, 11)
  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1071, 1, 11)
  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  dim(x) <- c(1071, 11, 1)
  a <- fftwtools::fftw_c2c_3d(x, inverse = 1)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  a <- fftwtools::fftw_c2c_3d(x, inverse = 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(b, a, tolerance = 1e-10)

  # zero length/margin
  x <- numeric(0)
  dim(x) <- c(0, 100, 10)

  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(dim(b), dim(x))

  dim(x) <- c(100, 0, 10)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(dim(b), dim(x))

  dim(x) <- c(100, 10, 0)
  b <- ravetools:::fftw_c2c_3d(x, inverse = 1)
  testthat::expect_equal(dim(b), dim(x))

  b <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(dim(b), dim(x))

})

test_that("fftw_c2c_3d input immutability and roundtrip", {
  set.seed(11)
  x <- array(complex(real = rnorm(60), imaginary = rnorm(60)), c(5, 4, 3))
  x_orig <- x + 0

  fwd <- ravetools:::fftw_c2c_3d(x, inverse = 0)
  testthat::expect_equal(x, x_orig)

  fwd_meas <- ravetools:::fftw_c2c_3d(x, inverse = 0, fftwplanopt = 1L)
  testthat::expect_equal(x, x_orig)
  testthat::expect_equal(fwd_meas, fwd, tolerance = 1e-10)

  inv <- ravetools:::fftw_c2c_3d(fwd, inverse = 1)
  testthat::expect_equal(inv / length(x), x, tolerance = 1e-10)

  testthat::expect_equal(x, x_orig)
})

test_that("fftw_r2c_3d input immutability", {
  set.seed(12)
  x <- array(rnorm(60), c(5, 4, 3))
  x_orig <- x + 0

  ravetools:::fftw_r2c_3d(x, HermConj = 1)
  testthat::expect_equal(x, x_orig)

  ravetools:::fftw_r2c_3d(x, HermConj = 0)
  testthat::expect_equal(x, x_orig)

  ravetools:::fftw_r2c_3d(x, HermConj = 1, fftwplanopt = 1L)
  testthat::expect_equal(x, x_orig)
})

test_that("fftw_*_3d match stats::fft", {
  set.seed(95)
  # real input
  X <- array(rnorm(120), dim = c(5, 4, 6))
  ref <- stats::fft(X + 0i)
  testthat::expect_equal(
    ravetools:::fftw_r2c_3d(X, HermConj = 1), ref, tolerance = 1e-10
  )
  testthat::expect_equal(
    ravetools:::fftw_c2c_3d(X + 0i, inverse = 0), ref, tolerance = 1e-10
  )

  # complex input
  Z <- array(complex(real = rnorm(120), imaginary = rnorm(120)),
             dim = c(5, 4, 6))
  testthat::expect_equal(
    ravetools:::fftw_c2c_3d(Z, inverse = 0),
    stats::fft(Z),
    tolerance = 1e-10
  )
  testthat::expect_equal(
    ravetools:::fftw_c2c_3d(Z, inverse = 1),
    stats::fft(Z, inverse = TRUE),
    tolerance = 1e-10
  )
})

mesh_volume <- function(mesh) {
  suppressWarnings({
    tryCatch({
      return(vcg_mesh_volume(mesh))
    }, error = function(e) {
      NA_real_
    })
  })
}

#' @title Generate 3D mesh surface from volume data
#' @description This function is soft-deprecated. Please use
#' \code{\link{vcg_mesh_volume}}, \code{\link{vcg_uniform_remesh}}, and
#' \code{\link{vcg_smooth_explicit}} or \code{\link{vcg_smooth_implicit}}.
#' @param volume 3-dimensional volume array
#' @param output_format resulting data format, choices are \code{'rgl'} and
#' \code{'freesurfer'}
#' @param IJK2RAS volume 'IJK' (zero-indexed coordinate index) to
#' \code{'tkrRAS'} transform, default is automatically determined
#' @param threshold threshold used to create volume mask; the surface will be
#' created to fit the mask boundaries
#' @param verbose whether to verbose the progress
#' @param remesh whether to re-sample the mesh using \code{\link{vcg_uniform_remesh}}
#' @param remesh_voxel_size,remesh_multisample,remesh_automerge see
#' arguments in \code{\link{vcg_uniform_remesh}}
#' @param smooth whether to smooth the mesh via \code{\link{vcg_smooth_explicit}}
#' @param smooth_lambda,smooth_delta,smooth_method see \code{\link{vcg_smooth_explicit}}
#' @returns A \code{'mesh3d'} surface if \code{output_format} is 'rgl', or
#' \code{'fs.surface'} surface otherwise.
#' @examples
#'
#'
#' volume <- array(0, dim = c(8,8,8))
#' volume[4:5, 4:5, 4:5] <- 1
#'
#' graphics::image(x = volume[4,,])
#'
#' # you can use rgl::wire3d(mesh) to visualize the mesh
#' mesh <- mesh_from_volume(volume, verbose = FALSE)
#'
#'
#' @export
mesh_from_volume <- function(
    volume, output_format = c("rgl", "freesurfer"),
    IJK2RAS = NULL, threshold = 0, verbose = TRUE,
    remesh = TRUE, remesh_voxel_size = 1,
    remesh_multisample = TRUE, remesh_automerge = TRUE,
    smooth = FALSE, smooth_lambda = 10, smooth_delta = 20,
    smooth_method = "surfPreserveLaplace"
) {

  output_format <- match.arg(output_format)

  debug <- function(..., appendLF = TRUE) {
    if (verbose) {
      appendLF <- ifelse(appendLF, "\n", "")
      cat(..., appendLF)
    }
  }

  volume <- as.array(volume)
  dm <- dim(volume)[c(1, 2, 3)]
  dim(volume) <- dm

  if (length(IJK2RAS) != 16L) {
    IJK2RAS <- matrix(
      nrow = 4, byrow = TRUE,
      c(-1, 0, 0, dm[[1]] / 2,
        0, 0, 1, -dm[[3]] / 2,
        0, -1, 0, dm[[2]] / 2,
        0, 0, 0, 1))
  }

  mesh <- vcg_isosurface(volume, threshold_lb = threshold, vox_to_ras = IJK2RAS)

  debug(sprintf("The initial reconstructed surface volume is %.1f mm^3", mesh_volume(mesh)))

  if ( remesh ) {
    mesh <- vcg_uniform_remesh(
      mesh, voxel_size = remesh_voxel_size, multi_sample = remesh_multisample,
      merge_clost = remesh_automerge, verbose = verbose)
    debug(sprintf("The re-meshed surface volume is %.1f mm^3", mesh_volume(mesh)))
  }
  if ( smooth ) {
    mesh <- vcg_smooth_explicit(
      mesh = mesh, type = smooth_method,
      lambda = smooth_lambda, delta = smooth_delta
    )
    debug(sprintf("The smoothed surface volume is %.1f mm^3", mesh_volume(mesh)))
  }

  switch(
    output_format,
    "freesurfer" = {
      face_index <- t(mesh$it)
      face_index_start <- min(face_index)
      face_index <- face_index - (face_index_start - 1L)

      vertices <- t(mesh$vb[c(1, 2, 3), ])
      mesh <- structure(
        list(
          mesh_face_type = "tris",
          vertices = vertices,
          faces = face_index
        ),
        class = c("fs.surface", "list")
      )
    }, {
    }
  )
  mesh

}

#' @title Remove the trend for one or more signals
#' @description 'Detrending' is often used before the signal power calculation.
#' @param x numerical or complex, a vector or a matrix
#' @param trend the trend of the signal; choices are \code{'constant'} and
#' \code{'linear'}
#' @param break_points integer vector, or \code{NULL}; only used when
#' \code{trend} is \code{'linear'} to remove piecewise linear trend; will
#' throw warnings if \code{trend} is \code{'constant'}
#' @returns The signals with trend removed in matrix form; the number of columns
#' is the number of signals, and number of rows is length of the signals
#' @examples
#'
#' x <- rnorm(100, mean = 1) + c(
#'   seq(0, 5, length.out = 50),
#'   seq(5, 3, length.out = 50))
#' plot(x)
#'
#' plot(detrend(x, 'constant'))
#' plot(detrend(x, 'linear'))
#' plot(detrend(x, 'linear', 50))
#'
#' @export
detrend <- function(x, trend = c("constant", "linear"), break_points = NULL) {

  trend <- match.arg(trend)

  if (!is.numeric(x) && !is.complex(x)) {
    stop("'x' must be a numeric or complex vector or matrix.")
  }

  if (!is.matrix(x)) {
    x <- as.matrix(x)
  }
  n <- nrow(x)

  if (length(break_points)) {
    if (trend == "constant") {
      warning("Breakpoints not used for 'constant' trend type.")
    } else {
      break_points <- as.integer(break_points)
      if (any(is.na(break_points) | break_points < 1 | break_points > n)) {
        stop("'break_points' must be a subset of 1:nrow(x).")
      }
    }
  }

  if (trend == "constant") {
    x_mean <- colMeans(x, na.rm = TRUE)
    re <- sweep(x, 2, x_mean, "-", check.margin = FALSE)
    return(re)
  } else {

    break_points <- sort(unique(c(0, break_points, n - 1)))
    # print(break_points)
    nbps <- length(break_points)

    a <- sapply(seq_len(nbps), function(ii) {
      if (ii == nbps) { return(rep(1, n)) }
      pad <- break_points[ii]
      m <- n - pad
      c(rep(0, pad), seq(1 / m, 1, length.out = m))
    })

    # print(a)
    # print(dim(x))
    re <- x - tcrossprod(a, t(qr.solve(a, x)))

    return(re)
  }

}

#' @title Localize white-matter and \verb{pial} surfaces from an intensity volume
#' @description
#' Starting from a single closed surface mesh (typically a smoothed estimate
#' of the white-matter boundary, in the same physical coordinate space as
#' \code{volume}) and a co-registered intensity volume (such as a normalized
#' \verb{T1} scan), iteratively deforms the surface in two passes to localize
#' the white/gray-matter and gray-matter/\verb{CSF} tissue-intensity
#' boundaries, producing a \code{white} and a \code{pial} surface.
#'
#' @details
#' The implementation keeps the two dominant ideas of the surface-placement
#' procedure described in the literature (see \strong{References}):
#' \enumerate{
#'   \item \strong{Intensity-target localization}: for each vertex, sample
#'     \code{volume} along the vertex's current normal at offsets spanning
#'     \eqn{\pm}\code{max_thickness} in steps of \code{step_size}, and pull
#'     the vertex toward the offset whose sampled intensity is closest to a
#'     single target value (\code{white_intensity} for the white-surface pass,
#'     \code{pial_intensity} for the \verb{pial}-surface pass).
#'   \item \strong{Smoothness}: a 1-ring Laplacian spring keeps the mesh
#'     regular while the per-vertex intensity term, which reacts independently
#'     to noisy image data, pulls vertices toward the tissue boundary.
#' }
#' Both terms are integrated with the same gradient-averaging and
#' momentum-integration machinery \code{\link{mris_inflate}} and
#' \code{\link{mris_sphere}} use: each inner iteration clears the gradient,
#' adds the intensity-target term, smooths it over \code{n_averages} passes of
#' 1-ring averaging, adds the locally-acting smoothness term (which is not
#' itself smoothed),
#' then takes a momentum-integration step with a 1 \verb{mm} per-step
#' displacement cap, and refreshes the vertex normals.
#'
#' The white-surface pass runs first, for \code{niterations} iterations; the
#' \verb{pial}-surface pass then continues from its result, with the momentum
#' velocity reset to rest, toward \code{pial_intensity} for another
#' \code{niterations} iterations.
#'
#' This is a \emph{reduced} port: the literature's procedure is a
#' multi-resolution optimization over roughly seven weighted energy terms
#' (intensity, intensity gradient, smoothness, self-intersection repulsion,
#' curvature, and more), using per-vertex gray/white/\verb{CSF} intensity
#' statistics derived from a prior segmentation. Reproducing that faithfully
#' is out of scope for this package; \code{white_intensity} and
#' \code{pial_intensity} are supplied directly here instead, for example the
#' midpoints between the typical white-matter/gray-matter and
#' gray-matter/\verb{CSF} intensities of \code{volume}.
#'
#' @param mesh triangular mesh of class \code{'mesh3d'} (or coercible via
#' \code{\link{ensure_mesh3d}}), in the same physical (\verb{RAS}) coordinate
#' space as \code{volume}. As with \code{\link{mris_inflate}}, the mesh
#' \strong{must} be closed, manifold, and genus-0 (watertight, no boundary or
#' non-manifold edges, single connected component); \code{mris_make_surfaces}
#' raises an error if such defects are detected
#' @param volume a 3-dimensional numeric array of image intensities (such as
#' a normalized \verb{T1} volume), co-registered with \code{mesh}
#' @param white_intensity target intensity for the white/gray-matter boundary;
#' the value the white-surface pass searches for along each vertex normal. For
#' a normalized \verb{T1} volume this is typically close to the midpoint
#' between the white-matter and gray-matter intensities
#' @param pial_intensity target intensity for the gray-matter/\verb{CSF}
#' boundary, analogous to \code{white_intensity} for the \verb{pial}-surface pass
#' @param IJK2RAS volume \verb{IJK} (zero-indexed voxel index) to surface
#' \verb{tkrRAS} transform (a \code{4x4} matrix); default is \code{NULL},
#' which assumes \code{volume} is a conformed volume (\verb{LIA} orientation,
#' \verb{1mm} isotropic 'voxels', centered at the volume midpoint, the same
#' convention \code{\link{fill_surface}} defaults to) and derives the
#' transform from \code{dim(volume)}; set this explicitly when \code{volume}
#' uses a different orientation or voxel size
#' @param max_thickness half-width, in \verb{mm}, of the per-vertex normal
#' search window the intensity-target term samples. Default \code{5}
#' @param step_size sampling step, in \verb{mm}, along the search window.
#' Default \code{0.4}
#' @param n_averages number of 1-ring gradient-averaging passes applied to the
#' intensity-target gradient before the smoothness term is added (mirrors
#' \code{\link{mris_inflate}}'s gradient-averaging step). Default \code{4}
#' @param niterations number of deformation iterations \emph{per pass}
#' (white, then \verb{pial}). Default \code{10}
#' @param l_intensity intensity-target term coefficient. Default \code{1.0}
#' @param l_spring smoothness (1-ring Laplacian spring) term coefficient.
#' Default \code{0.5}
#' @param momentum momentum coefficient. Default \code{0.9}
#' @param dt time step. Default \code{0.5}
#' @param verbose logical; print per-pass progress. Default \code{FALSE}
#'
#' @returns A named list of two \code{'mesh3d'} surfaces (each with
#' \code{vb}, \code{it}, and \code{normals}):
#' \describe{
#'   \item{\code{white}}{Surface localized to \code{white_intensity}.}
#'   \item{\code{pial}}{Surface localized to \code{pial_intensity}, continuing
#'     from \code{white}.}
#' }
#'
#' @references
#' Cortical surface-based analysis I: Segmentation and surface
#' reconstruction. \emph{NeuroImage}, 9(2), 179-194 (1999).
#'
#' @examples
#'
#' if (is_not_cran()) {
#'
#'
#' data("left_hippocampus_mask")
#' n_vox <- length(left_hippocampus_mask)
#'
#' volume <- left_hippocampus_mask + runif(n = n_vox, 0, 1)
#' vox2ras <- diag(1, 4)
#' mesh <- vcg_isosurface(volume, threshold_lb = 0.99)
#'
#' plot(mesh)
#'
#' # Fix defects
#' mesh <- vcg_fix_defects(mesh, verbose = TRUE, merge_tolerance = 1.75)
#'
#'
#' res <- mris_make_surfaces(
#'   mesh,
#'   volume,
#'   pial_intensity = 1.1,
#'   white_intensity = 1,
#'   IJK2RAS = vox2ras
#' )
#'
#' plot(res$pial)
#'
#'
#' }
#'
#'
#' @export
mris_make_surfaces <- function(
    mesh,
    volume,
    white_intensity,
    pial_intensity,
    IJK2RAS       = NULL,
    max_thickness = 5,
    step_size     = 0.4,
    n_averages    = 4L,
    niterations   = 10L,
    l_intensity   = 1.0,
    l_spring      = 0.5,
    momentum      = 0.9,
    dt            = 0.5,
    verbose       = FALSE
) {
  mesh <- meshintegrity(mesh, facecheck = TRUE)

  vb <- mesh$vb[1:3, , drop = FALSE]
  storage.mode(vb) <- "double"

  it <- mesh$it
  storage.mode(it) <- "integer"

  volume <- as.array(volume)
  if (length(dim(volume)) != 3L) {
    stop("`mris_make_surfaces`: `volume` must be a 3-dimensional array")
  }
  storage.mode(volume) <- "double"

  if (length(IJK2RAS) != 16L) {
    # Conformed-volume ('LIA', 1mm isotropic, centered) default; the same
    # convention `fill_surface` assumes when `IJK2RAS` is not supplied,
    # generalized here from a fixed cube edge to `dim(volume)`.
    vol_dim <- dim(volume)
    IJK2RAS <- matrix(
      nrow = 4, byrow = TRUE,
      c(-1, 0, 0, vol_dim[[1]] / 2,
        0, 0, 1, -vol_dim[[3]] / 2,
        0, -1, 0, vol_dim[[2]] / 2,
        0, 0, 0, 1))
  } else {
    IJK2RAS <- matrix(as.double(IJK2RAS), nrow = 4)
  }

  ras2ijk <- solve(IJK2RAS)
  storage.mode(ras2ijk) <- "double"

  tmp <- mrisMakeSurfaces(
    vb_             = vb,
    it_             = it,
    volume_         = volume,
    ras2ijk_        = ras2ijk,
    white_intensity = as.double(white_intensity)[[1L]],
    pial_intensity  = as.double(pial_intensity)[[1L]],
    max_thickness   = as.double(max_thickness)[[1L]],
    step_size       = as.double(step_size)[[1L]],
    n_averages      = as.integer(n_averages)[[1L]],
    niterations     = as.integer(niterations)[[1L]],
    l_intensity     = as.double(l_intensity)[[1L]],
    l_spring        = as.double(l_spring)[[1L]],
    momentum        = as.double(momentum)[[1L]],
    dt              = as.double(dt)[[1L]],
    verbose         = as.logical(verbose)[[1L]]
  )

  pack <- function(part) {
    structure(
      list(
        vb      = rbind(part$vb, 1),
        it      = it,
        normals = rbind(part$normals, 1)
      ),
      class = c("ravetools_mesh3d", "mesh3d")
    )
  }

  list(
    white = pack(tmp$white),
    pial  = pack(tmp$pial)
  )
}

#' @name fast_quantile
#' @title Compute quantiles
#' @param x numerical-value vector for \code{fast_quantile} and
#' \code{fast_median}, and column-major matrix for \code{fast_mvquantile} and
#' \code{fast_mvmedian}
#' @param prob a probability with value from 0 to 1
#' @param na.rm logical; if true, any \code{NA} are removed from \code{x}
#' before the quantiles are computed
#' @param ... reserved for future use
#' @returns \code{fast_quantile} and \code{fast_median} calculate univariate
#' quantiles (single-value return); \code{fast_mvquantile} and \code{fast_mvmedian}
#' calculate multivariate quantiles (for each column, result lengths equal to
#' the number of columns).
#'
#' @examples
#'
#' fast_quantile(runif(1000), 0.1)
#' fast_median(1:100)
#'
#' x <- matrix(rnorm(100), ncol = 2)
#' fast_mvquantile(x, 0.2)
#' fast_mvmedian(x)
#'
#' # Compare speed for vectors (usually 30% faster)
#' x <- rnorm(10000)
#' microbenchmark::microbenchmark(
#'   fast_median = fast_median(x),
#'   base_median = median(x),
#'   # bioc_median = Biobase::rowMedians(matrix(x, nrow = 1)),
#'   times = 100, unit = "milliseconds"
#' )
#'
#' # Multivariate cases
#' # (5~7x faster than base R)
#' # (3~5x faster than Biobase rowMedians)
#' x <- matrix(rnorm(100000), ncol = 20)
#' microbenchmark::microbenchmark(
#'   fast_median = fast_mvmedian(x),
#'   base_median = apply(x, 2, median),
#'   # bioc_median = Biobase::rowMedians(t(x)),
#'   times = 10, unit = "milliseconds"
#' )
#'
#' @export
fast_quantile <- function(x, prob = 0.5, na.rm = FALSE, ...) {
  if (length(prob) != 1L) {
    stop("`fast_quantile`: `prob` length can only be one")
  }
  .Call(`_ravetools_quickQuantile`, x, prob, isTRUE(as.logical(na.rm)))
}

#' @rdname fast_quantile
#' @export
fast_median <- function(x, na.rm = FALSE, ...) {
  .Call(`_ravetools_quickMedian`, x, isTRUE(as.logical(na.rm)))
}

#' @rdname fast_quantile
#' @export
fast_mvquantile <- function(x, prob = 0.5, na.rm = FALSE, ...) {
  if (length(prob) != 1L) {
    stop("`fast_quantile`: `prob` length can only be one")
  }
  .Call(`_ravetools_columnQuantile`, x, prob, isTRUE(as.logical(na.rm)))
}

#' @rdname fast_quantile
#' @export
fast_mvmedian <- function(x, na.rm = FALSE, ...) {
  .Call(`_ravetools_columnMedian`, x, isTRUE(as.logical(na.rm)))
}

# guess a upper-bound non-negative integer

#' Guess the max of non-negative numbers that pass the test
#' @param validator function that returns true if the number passes the test
#' or false if fails
#' @param initial initial value to guess, must be within \code{(min_v, max_v)}
#' @param min_v minimum number that passes the test
#' @param max_v maximum possible number
#' @returns The upper bound of the number that passes the test
#'
#' @examples
#'
#'
#' # 42
#' guess_max_integer(
#'   function(x) {
#'     x <= 42
#'   }
#' )
#'
#' # 42
#' guess_max_integer(
#'   function(x) {
#'     x <= 100
#'   }, max_v = 42
#' )
#'
#' # 42
#' guess_max_integer(
#'   function(x) {
#'     x <= 42
#'   }, min_v = 42
#' )
#'
#' # NA
#' guess_max_integer(
#'   function(x) {
#'     x <= 4
#'   }, min_v = 42
#' )
#'
#'
#'
#' @noRd
guess_max_integer <- function(validator, initial = NA, min_v = 0, max_v = Inf) {

  # DIPSAUS DEBUG START
  # validator <- function(x) { x <= 42 }
  # initial = NA
  # min_v = 0
  # max_v = Inf

  if (!validator(min_v)) { return(NA) }

  if (is.na(initial)) {
    if (is.infinite(max_v)) {
      initial <- max(min_v, 1)
    } else {
      initial <- ceiling((min_v + max_v) / 2.)
    }
  }

  min_v <- min_v
  max_v <- max_v

  if (is.finite( max_v )) {
    if ( validator( max_v ) ) {
      return( max_v )
    }
  } else {
    while ( validator(initial) ) {
      min_v <- initial
      initial <- initial * 2
    }
    max_v <- initial
    initial <- ceiling((min_v + max_v) / 2.)
  }



  test_number <- function(v) {

    if ( min_v + 1 >= max_v ) { return( min_v ) }

    if ( validator(v) ) {
      min_v <<- v
    } else {
      max_v <<- v
    }

    v <- ceiling((min_v + max_v) / 2.)

    return(Recall(v))
  }

  test_number(initial)

}

#' @title Render one or more meshes as flat-shaded triangles in base R
#' @description
#' Projects each triangular face onto a 2D plane using an orthographic camera,
#' shades it with a single color proportional to how directly it faces the
#' camera (Lambert), depth-sorts all faces across all meshes, and draws them
#' in a single \code{\link[graphics]{polygon}} call.
#'
#' Meshes without faces (point clouds) are substituted by a small sphere
#' (\code{\link{vcg_sphere}}) centered at each point and scaled by
#' \code{cex}; they then participate in the same rendering pipeline as ordinary
#' faced meshes.
#'
#' A camera-facing clipping pass discards triangles whose outward normal
#' points along the camera ray (signed \eqn{n \cdot z_{cam}} \eqn{> 1 - \mathrm{mesh\_clipping}}),
#' peeling the front cap off the surface so the back wall (and any
#' interior meshes) become visible. Set \code{mesh_clipping = 1} to disable
#' clipping. Point-cloud meshes (those rendered as substitute
#' \code{\link{vcg_sphere}} instances) are exempt from this clip so they
#' remain solid even when the enclosing surface is peeled.
#'
#' Multiple meshes share a single depth space: all faces are projected,
#' sorted, and drawn together so the painter's algorithm works correctly
#' across meshes.
#'
#' @param mesh a \code{'mesh3d'} object, or a list of \code{'mesh3d'} objects.
#'   Meshes without a face matrix (\code{$it}) are rendered as sphere
#'   instances (one \code{\link{vcg_sphere}} per vertex, radius \code{cex}).
#' @param eye numeric vector of length 3 - camera position in world space.
#' @param lookat numeric vector of length 3 - the world-space point the camera
#'   is looking at.
#' @param up numeric vector of length 3 - world-space "up" direction; defaults
#'   to \code{c(0, 1, 0)}.
#' @param col base color(s) per mesh.  Same forms as
#'   \code{\link{plot_mesh_dotcloud}}: a single color, a depth-gradient vector,
#'   a per-vertex character vector, or a list of any of these (one element per
#'   mesh).  Default \code{c("white", "gray30")}.
#' @param cex radius of the substitute sphere used for point-cloud meshes
#'   (world units).  Has no effect on meshes that already have faces.  Default
#'   \code{1}.
#' @param sphere_subdivision integer subdivision level forwarded to
#'   \code{\link{vcg_sphere}} when substituting point clouds.  Higher values
#'   give smoother spheres at the cost of more triangles.  Default \code{1}
#'   (80 triangles per sphere).
#' @param add logical; if \code{TRUE} the faces are added to an existing plot
#'   instead of opening a new one.  Default \code{FALSE}.
#' @param axes,asp,xlim,ylim,xlab,ylab,main passed to
#'   \code{\link[graphics]{plot.default}} when \code{add = FALSE}.
#' @param zoom positive numeric magnification applied to the auto-computed
#'   axis limits when \code{xlim} or \code{ylim} is \code{NULL}.  Values
#'   greater than \code{1} zoom in, values in \code{(0, 1)} zoom out.
#'   When \code{asp} is set, the zoom is plot-region aware (queried via
#'   \code{par("pin")}): the axis that already fills the device after
#'   \code{asp}-induced expansion is zoomed by exactly \code{zoom}, and
#'   the other axis is zoomed less so the data fills more of the plot
#'   region while preserving the requested \code{asp}.  Ignored for any
#'   axis whose limit was supplied explicitly, and when \code{add = TRUE}.
#'   Default \code{1}.
#' @param side which side of each triangle to render.  One of \code{"front"}
#'   (default), \code{"back"}, or \code{"both"} (shows all triangles).
#' @param mesh_clipping numeric in \code{[0, 1]} controlling camera-facing
#'   clipping: any triangle whose signed Lambert dot product
#'   (\eqn{n \cdot z_{cam}}, i.e. positive for front-facing) is at or
#'   above \code{mesh_clipping} is discarded.  A value of \code{1}
#'   (default) keeps the full surface; \code{0.5} peels a 60-degree cap
#'   off the front and reveals the back wall (whose absolute Lambert
#'   shade is still large, so it renders brightly); \code{0} clips
#'   every front-facing triangle and shows the interior only.
#'   Back-facing triangles are never clipped by this rule.  Default
#'   \code{1}.
#' @param alpha numeric in \code{[0, 1]}; one value per mesh (recycled),
#'   sets the transparency of each mesh (\code{1} = fully opaque,
#'   \code{0} = fully transparent).  Faces belonging to a mesh with
#'   \code{alpha = 0} are dropped entirely.  Because faces are drawn in
#'   back-to-front (painter's) order, alpha blending across multiple
#'   meshes follows that order.  Default \code{1}.
#' @param clipping_plane optional list of world-space clipping planes used to
#'   hide parts of the scene.  Each plane is a numeric vector of length 5:
#'   the first three entries are the plane normal \eqn{(n_x, n_y, n_z)}
#'   (must be non-zero; normalized internally), the fourth is the signed
#'   distance from the world origin to the plane along that normal (so the
#'   plane equation is \eqn{n \cdot x = d}), and the fifth indicates which
#'   half-space is kept: \code{1} keeps the front side (the side the normal
#'   points to), \code{-1} keeps the back side, and \code{0} keeps whichever
#'   side currently faces the camera (auto-flipped per call based on
#'   \code{eye}).  Multiple planes are intersected.  Clipping is applied
#'   per face using the face centroid (a face is kept only when its
#'   centroid lies on the kept side of every plane).  A single length-5
#'   numeric vector is also accepted as shorthand for a one-plane list.
#'   Default \code{NULL} (no clipping).
#' @param clipping_plane_enabled logical vector, one entry per mesh
#'   (recycled), controlling whether \code{clipping_plane} is applied to
#'   that mesh.  \code{TRUE} (default) means the mesh participates in
#'   clipping; \code{FALSE} exempts the mesh entirely (all of its faces
#'   are kept regardless of the clipping planes).  Has no effect when
#'   \code{clipping_plane} is \code{NULL}.
#' @param shadow_color color used for fully unlit (grazing/back) faces.
#'   The Lambert shade linearly interpolates from \code{shadow_color}
#'   (at \code{shade = 0}) toward the face color lit by a white light
#'   (at \code{shade = 1}), so unlit faces fade to this color instead
#'   of an implicit black background.  Default \code{par("fg")}, so the
#'   shadow contrasts with the current device's canvas.
#' @param light_intensity non-negative scalar controlling the brightness
#'   of the (white) light source: at \code{shade = 1} the face color is
#'   multiplied by \code{light_intensity} (clamped to \code{[0, 1]}).
#'   \code{1} (default) reproduces the face color exactly when fully
#'   lit; smaller values darken the whole mesh, larger values saturate.
#' @param ambient_intensity scalar in \code{[0, 1]} acting as a lower
#'   bound on the Lambert shade so grazing/back faces still receive some
#'   light and never collapse fully to \code{shadow_color}.  An effective
#'   shade of \code{max(shade, ambient_intensity)} is used in the
#'   background-to-light blend.  Default \code{0.2}.
#' @param ... additional graphical parameters forwarded to
#'   \code{\link[graphics]{plot.default}} (new plot only).
#'
#' @return Invisibly returns a list with components \code{xlim} and
#'   \code{ylim} (the plot limits used).
#'
#' @details
#' Limitations of the base-R polygon path (no \code{rgl}):
#' \itemize{
#'   \item Flat shading only - one color per triangle.  No per-vertex
#'     color interpolation.
#'   \item No depth buffer - faces are depth-sorted by centroid (painter's
#'     algorithm).  Interpenetrating triangles can render in the wrong
#'     order.
#'   \item Anti-aliasing seams can appear between adjacent triangles on
#'     raster devices; Cairo-based devices (\code{png(type = "cairo")},
#'     \code{svg()}, \code{pdf()}) produce cleaner output than the default
#'     quartz/X11 path.
#' }
#'
#' @examples
#'
#' mesh <- vcg_isosurface(left_hippocampus_mask)
#'
#' # Surface alone
#' plot(
#'   mesh,
#'   eye    = c(150, 30, 0),
#'   lookat = c(0, 0, 0),
#'   up     = c(0, 0, 1),
#'   col    = "steelblue"
#' )
#'
#' # Surface + electrode point cloud (rendered as small icospheres)
#' n_elec <- 20
#' electrodes <- structure(
#'   list(vb = matrix(rnorm(3 * n_elec, sd = 5), 3, n_elec) +
#'          rowMeans(mesh$vb)[1:3]),
#'   class = "mesh3d"
#' )
#' plot_mesh_polygon(
#'   mesh = list(mesh, electrodes),
#'   eye    = c(150, -30, 0),
#'   lookat = c(0, 0, 0),
#'   up     = c(0, 0, 1),
#'   alpha  = c(0.5, 1),
#'   col    = list("steelblue", "red"),
#'   cex    = 1.5
#' )
#'
#' @seealso \code{\link{plot_mesh_dotcloud}}, \code{\link{vcg_isosurface}}
#'
#' @inheritSection ensure_mesh3d Coercing Surface Inputs
#'
#' @export
plot_mesh_polygon <- function(
    mesh,
    eye    = c(0, 0, 1000),
    lookat = c(0, 0, 0),
    up     = c(0, 1, 0),
    col    = c("white", "gray30"),
    cex    = 1,
    add    = FALSE,
    axes   = FALSE,
    asp    = 1,
    xlim   = NULL,
    ylim   = NULL,
    zoom   = 1,
    xlab   = "",
    ylab   = "",
    main   = "",
    side   = c("front", "back", "both"),
    mesh_clipping = 1,
    sphere_subdivision = 1L,
    alpha  = 1,
    shadow_color = NULL,
    light_intensity = 1,
    ambient_intensity = 0.2,
    clipping_plane = NULL,
    clipping_plane_enabled = TRUE,
    ...
) {
  side <- match.arg(side)

  # DIPSAUS DEBUG START
  # mesh <- vcg_isosurface(ravetools::left_hippocampus_mask)
  # list2env(envir = .GlobalEnv, list(
  #   eye = c(150, 0, 0), lookat = c(0, 0, 0), up = c(0, 0, 1),
  #   col = "steelblue", cex = 1, add = FALSE, axes = FALSE, asp = 1,
  #   xlim = NULL, ylim = NULL, xlab = "", ylab = "", mesh_clipping = 1
  # ))
  # DIPSAUS DEBUG END

  # ---- 0. Normalize mesh + col list -------------------------------------
  if (inherits(mesh, c("mesh3d", "fs.surface", "ieegio_surface"))) {
    mesh_list <- list(ensure_mesh3d(mesh))
  } else {
    mesh_list <- lapply(mesh, ensure_mesh3d)
  }
  n_meshes <- length(mesh_list)

  if (is.list(col)) {
    col_list <- rep_len(col, n_meshes)
  } else {
    col_list <- rep(list(col), n_meshes)
  }
  col_list <- lapply(col_list, function(ci) {
    substr(grDevices::adjustcolor(ci), 1L, 7L)
  })

  alpha_vec <- pmax(0, pmin(1, rep_len(as.numeric(unlist(alpha)), n_meshes)))
  clip_enabled_vec <- rep_len(as.logical(clipping_plane_enabled), n_meshes)
  clip_enabled_vec[is.na(clip_enabled_vec)] <- TRUE

  # Shading parameters: shadow color (defaults to par("fg")) and white-light
  # intensity. Resolved once here so par() is only queried per call.
  if (is.null(shadow_color)) {
    shadow_color <- graphics::par("fg")
  }
  shadow_rgb      <- grDevices::col2rgb(shadow_color)[, 1L] / 255
  light_intensity <- max(0, as.numeric(light_intensity)[[1L]])
  ambient_intensity <- max(0, min(1, as.numeric(ambient_intensity)[[1L]]))

  cex <- as.numeric(cex)[[1L]]
  sphere_subdivision <- as.integer(sphere_subdivision)[[1L]]

  # ---- 1. View matrix (R6, same as plot_mesh_dotcloud) ------------------
  rot_w2c <- Matrix4$new()$look_at(
    as_vector3(eye[1:3]),
    as_vector3(lookat[1:3]),
    as_vector3(up[1:3])
  )$invert()
  t_vec <- Vector3$new()$from_array(eye[1:3])$apply_matrix4(rot_w2c)$multiply_scalar(-1)
  view  <- rot_w2c$clone2()$set_position(t_vec)

  # ---- 2. Extract vb/faces (substitute small spheres for point clouds) ----
  sphere_tmpl <- vcg_sphere(sphere_subdivision, normals = FALSE)
  sphere_v <- sphere_tmpl$vb[1:3, , drop = FALSE] * cex   # 3 x n_sv (scaled)
  sphere_f <- sphere_tmpl$it                              # 3 x n_sf
  storage.mode(sphere_f) <- "integer"
  n_sv <- ncol(sphere_v)
  n_sf <- ncol(sphere_f)

  mesh_data <- vector("list", n_meshes)
  for (i in seq_len(n_meshes)) {
    m  <- mesh_list[[i]]
    vb <- m$vb
    if (nrow(vb) > 3L) vb <- vb[1:3, , drop = FALSE]

    has_faces <- !is.null(m$it) && length(m$it) > 0L

    if (has_faces) {
      it <- m$it
      storage.mode(it) <- "integer"
    } else {
      # Instance a vcg_sphere at each point
      n_pts <- ncol(vb)
      if (n_pts == 0L) {
        vb <- matrix(numeric(0), nrow = 3L, ncol = 0L)
        it <- matrix(integer(0), nrow = 3L, ncol = 0L)
      } else {
        # Vertices: repeat sphere verts and add point position
        rep_sv  <- sphere_v[, rep.int(seq_len(n_sv), n_pts), drop = FALSE]
        rep_pts <- vb[, rep(seq_len(n_pts), each = n_sv), drop = FALSE]
        vb <- rep_sv + rep_pts
        # Faces: repeat sphere_f, offset by n_sv per point
        it_flat <- rep.int(as.integer(sphere_f), n_pts) +
          rep(seq.int(0L, by = n_sv, length.out = n_pts), each = 3L * n_sf)
        it <- matrix(it_flat, nrow = 3L)
      }
    }

    mesh_data[[i]] <- list(
      vb         = vb,
      it         = it,
      col        = col_list[[i]],
      n_verts    = ncol(vb),
      n_faces    = ncol(it),
      pointcloud = !has_faces
    )
  }

  # ---- 3. Concatenate vb and faces (offset face indices) ----------------
  n_verts_vec <- vapply(mesh_data, `[[`, 0L, "n_verts")
  n_faces_vec <- vapply(mesh_data, `[[`, 0L, "n_faces")
  pc_vec      <- vapply(mesh_data, `[[`, NA, "pointcloud")
  # Per-face flag: was this face part of an originally-point-cloud mesh?
  is_pointcloud <- rep.int(pc_vec, n_faces_vec)
  # Per-face alpha (one value per mesh, replicated across that mesh's faces)
  alpha_face <- rep(alpha_vec, n_faces_vec)
  total_verts <- sum(n_verts_vec)
  total_faces <- sum(n_faces_vec)

  all_vb <- do.call(cbind, lapply(mesh_data, `[[`, "vb"))     # 3 x V
  vert_offsets <- c(0L, cumsum(n_verts_vec)[-n_meshes])
  if (length(vert_offsets) < n_meshes) vert_offsets <- 0L
  all_it <- do.call(cbind, lapply(seq_len(n_meshes), function(i) {
    if (n_faces_vec[[i]] == 0L) return(matrix(integer(0), 3L, 0L))
    mesh_data[[i]]$it + vert_offsets[[i]]
  }))

  if (total_faces == 0L) {
    if (!add) {
      if (!length(xlim)) xlim <- c(-1, 1)
      if (!length(ylim)) ylim <- c(-1, 1)
      graphics::plot.default(NA, NA, xlim = xlim, ylim = ylim,
                             asp = asp, axes = axes, type = "n",
                             xlab = xlab, ylab = ylab, ...)
    }
    return(invisible(list(xlim = xlim, ylim = ylim)))
  }

  # ---- 4. Single view transform on all vertices -------------------------
  cam_pts <- Vector3$new()$from_array(all_vb)$apply_matrix4(view)
  x_all <- cam_pts[1L, ]
  y_all <- cam_pts[2L, ]
  z_all <- cam_pts[3L, ]

  # ---- 5. Per-face: normal (cam space), facing, centroid depth ----------
  i1 <- all_it[1L, ]
  i2 <- all_it[2L, ]
  i3 <- all_it[3L, ]

  e1x <- x_all[i2] - x_all[i1]
  e1y <- y_all[i2] - y_all[i1]
  e1z <- z_all[i2] - z_all[i1]
  e2x <- x_all[i3] - x_all[i1]
  e2y <- y_all[i3] - y_all[i1]
  e2z <- z_all[i3] - z_all[i1]

  # Cross product, normalize z-component for cos(angle to camera axis)
  nx <- e1y * e2z - e1z * e2y
  ny <- e1z * e2x - e1x * e2z
  nz <- e1x * e2y - e1y * e2x
  n_len <- sqrt(nx * nx + ny * ny + nz * nz)
  facing <- nz / n_len    # in [-1, 1]; NaN for degenerate triangles

  depth_face <- -(z_all[i1] + z_all[i2] + z_all[i3]) / 3   # +ve = in front

  # Lambert shade: |facing| works for both sides (back-faces are covered by
  # front-faces after painter's sort anyway, so this matches a closed mesh).
  shade <- abs(facing)
  shade[is.na(shade)] <- 0

  # ---- 6. Camera-facing clip (peel away the cap facing the camera) -----
  #         + side filter
  # Discard triangles whose outward normal points along the camera ray:
  # signed facing = n . z_cam >= mesh_clipping. mesh_clipping = 1 keeps
  # the full surface; smaller values peel a wider cap and reveal the back
  # wall (whose Lambert shade |facing| is still large, so it renders
  # brightly).
  keep <- facing <= mesh_clipping
  if (side == "front") {
    keep <- keep & (facing > 0)
  } else if (side == "back") {
    keep <- keep & (facing < 0)
  }
  keep <- is_pointcloud | keep   # point clouds are exempt from camera-facing clip
  keep <- keep & (alpha_face > 0)

  # User-supplied world-space clipping planes (per-face cull via centroid).
  # Applies to point-cloud sphere instances as well. Meshes with
  # `clipping_plane_enabled = FALSE` are exempted from this cull.
  if (length(clipping_plane)) {
    centroids <- (all_vb[, i1, drop = FALSE] +
                  all_vb[, i2, drop = FALSE] +
                  all_vb[, i3, drop = FALSE]) / 3
    cp_keep <- apply_clipping_planes(centroids, eye[1:3], clipping_plane)
    clip_enabled_face <- rep(clip_enabled_vec, n_faces_vec)
    cp_keep[!clip_enabled_face] <- TRUE
    keep <- keep & cp_keep
  }

  # ---- 7. Per-face base color (one entry per mesh) ----------------------
  dr <- range(depth_face, na.rm = TRUE)
  dr_span <- dr[2L] - dr[1L]
  depth_face_norm <- if (dr_span < .Machine$double.eps) {
    rep(0.5, total_faces)
  } else {
    (depth_face - dr[1L]) / dr_span
  }

  col_face       <- character(total_faces)
  face_offsets   <- cumsum(c(0L, n_faces_vec))

  for (i in seq_len(n_meshes)) {
    n_f <- n_faces_vec[[i]]
    if (n_f == 0L) next
    fidx  <- seq.int(face_offsets[[i]] + 1L, face_offsets[[i + 1L]])
    col_i <- col_list[[i]]
    n_v_i <- n_verts_vec[[i]]

    if (length(col_i) == n_v_i) {
      # Per-vertex: take face's first vertex color (flat shading)
      face_v1_local <- all_it[1L, fidx] - vert_offsets[[i]]
      col_face[fidx] <- col_i[face_v1_local]
    } else if (length(col_i) == 1L) {
      col_face[fidx] <- col_i[[1L]]
    } else {
      ramp <- grDevices::colorRampPalette(rev(col_i))(256L)
      col_face[fidx] <- ramp[ceiling(depth_face_norm[fidx] * 255) + 1L]
    }
  }

  # ---- 8. Filter + painter's sort (far first) ---------------------------
  keep_idx <- which(keep)
  if (!length(keep_idx)) {
    if (!add) {
      if (!length(xlim)) xlim <- range(x_all, na.rm = TRUE)
      if (!length(ylim)) ylim <- range(y_all, na.rm = TRUE)
      graphics::plot.default(NA, NA, xlim = xlim, ylim = ylim,
                             asp = asp, axes = axes, type = "n",
                             xlab = xlab, ylab = ylab, ...)
    }
    return(invisible(list(xlim = xlim, ylim = ylim)))
  }

  ord <- keep_idx[order(depth_face[keep_idx], decreasing = TRUE)]

  i1k <- all_it[1L, ord]
  i2k <- all_it[2L, ord]
  i3k <- all_it[3L, ord]
  shade_k <- shade[ord]
  col_k   <- col_face[ord]
  alpha_k <- alpha_face[ord]

  # ---- 9. Apply Lambert shading -----------------------------------------
  # Light is white at brightness `light_intensity`; the fully-lit color
  # is therefore col * light_intensity. The fully-unlit color is
  # `shadow_color` (default par("fg")). The Lambert term `shade_k` is
  # lower-bounded by `ambient_intensity` so rim/back faces still receive
  # some light:
  #   eff_shade = max(shade, ambient_intensity)
  #   shaded    = shadow + eff_shade * (col * light_intensity - shadow)
  # The final pmin clamp is applied to the output channels (not lit_*),
  # so values stay in [0, 1] only after the lerp.
  rgb_mat <- grDevices::col2rgb(col_k) / 255
  lit_r   <- pmin(1, rgb_mat[1L, ] * light_intensity)
  lit_g   <- pmin(1, rgb_mat[2L, ] * light_intensity)
  lit_b   <- pmin(1, rgb_mat[3L, ] * light_intensity)

  eff_shade <- pmax(shade_k, ambient_intensity)

  r_out <- shadow_rgb[1L] + eff_shade * (lit_r - shadow_rgb[1L])
  g_out <- shadow_rgb[2L] + eff_shade * (lit_g - shadow_rgb[2L])
  b_out <- shadow_rgb[3L] + eff_shade * (lit_b - shadow_rgb[3L])

  # Skip the alpha channel when every mesh is fully opaque (faster device path).
  if (any(alpha_vec < 1)) {
    col_shaded <- grDevices::rgb(r_out, g_out, b_out, alpha = alpha_k)
  } else {
    col_shaded <- grDevices::rgb(r_out, g_out, b_out)
  }

  # ---- 10. Open canvas + plot limits ------------------------------------
  # Call plot.new() before zoom_limits so par("pin") reflects the true
  # plot-region dimensions of this device (margins applied). Then
  # plot.window() locks in the zoomed limits + asp before drawing.
  if (!add) {
    graphics::plot.new()
  }
  lim <- zoom_limits(xlim, ylim, x_all, y_all, zoom, asp)
  xlim <- lim$xlim
  ylim <- lim$ylim
  if (!add) {
    graphics::plot.window(xlim = xlim, ylim = ylim, asp = asp, ...)
    if (axes) {
      graphics::axis(1)
      graphics::axis(2)
      graphics::box()
    }
    if (nzchar(xlab) || nzchar(ylab) || nzchar(main)) {
      graphics::title(main = main, xlab = xlab, ylab = ylab)
    }
  }

  # 4 rows per triangle: v1, v2, v3, NA. Flatten column-major for polygon().
  tri_x <- rbind(x_all[i1k], x_all[i2k], x_all[i3k], NA_real_)
  tri_y <- rbind(y_all[i1k], y_all[i2k], y_all[i3k], NA_real_)

  # ---- 11. Render -------------------------------------------------------
  grDevices::dev.hold()
  on.exit(grDevices::dev.flush(), add = TRUE)
  # Hide the anti-aliasing seam between adjacent triangles by drawing the
  # border in the fill color - but only when fully opaque. With alpha < 1,
  # the border stroke would blend on top of the fill and produce visibly
  # darker triangle edges, so fall back to no border in that case.
  border_col <- if (any(alpha_vec < 1)) NA else col_shaded
  graphics::polygon(c(tri_x), c(tri_y),
                    col = col_shaded, border = border_col)

  invisible(list(xlim = xlim, ylim = ylim))
}

#' @export
plot.ravetools_mesh3d <- function(x, method = c("polygon", "dotcloud", "rgl"), ...) {
  method <- match.arg(method)
  switch(
    method,
    "polygon" = {
      plot_mesh_polygon(mesh = x, ...)
    },
    "dotcloud" = {
      plot_mesh_dotcloud(mesh = x, ...)
    },
    {
      # rgl
      rgl_view({
        rgl_call("shade3d", x, ...)
      })
    }
  )
}

#' @export
print.ravetools_mesh3d <- function(x, prefix = "", ...) {
  cat(prefix, " mesh3d object with ", ncol(x$vb), " vertices, ",
      paste(c(if (length(x$ip)) paste(length(x$ip), "points"),
              if (length(x$is)) paste(ncol(x$is), "segments"),
              if (length(x$it)) paste(ncol(x$it), "triangles"),
              if (length(x$ib)) paste(ncol(x$ib), "quads")), collapse = ", "),
      ".\n", sep = "")
}

#' @title Native 3D volume registration (\code{'rigid'}, \code{'affine'}, or \code{'SyN'})
#' @description
#' Self-contained image registration for 3D volumes, implemented purely in
#' \pkg{RcppEigen} (no other external registration library). It mirrors the
#' core behavior of \pkg{'ANTs'} \code{antsRegistration}: a multi-resolution,
#' physical-shift scaled gradient-descent optimizer driving a similarity metric,
#' working entirely in the anatomical \verb{RAS} (right-anterior-superior)
#' space. Each volume carries its own \code{vox2ras} (0-indexed voxel index to
#' anatomical \verb{RAS}) \eqn{4\times 4} transform, so volumes with
#' different sampling, orientation, or field of view are aligned correctly.
#'
#' @param source the moving volume to be aligned, a 3D array (for example a
#' \code{'CT'}); the result transform maps \code{target} into this image's space.
#' May also be a \code{list} of already co-registered 3D arrays (multiple
#' modalities such as \code{'T1'}, \code{'T2'}, an atlas constraint); all source
#' channels must share the same dimensions and \code{vox2ras}
#' @param target the fixed/reference volume to align to, a 3D array (for example
#' a \code{'MRI'}), or a \code{list} of co-registered arrays matching
#' \code{source} channel-for-channel
#' @param weights optional numeric weights, one per source/target pair,
#' controlling each channel's contribution to the deformable cost; default is
#' equal weighting. Weights are normalized internally to sum to 1. Only the
#' deformable (\code{"syn"}/\code{"syn_only"}) stage is multivariate; the linear
#' stage always uses the first (primary) pair
#' @param source_vox2ras,target_vox2ras 4x4 (or 3x4) matrices mapping the
#' 0-indexed voxel coordinate (column-row-slice, \code{'C'}-style starting from
#' 0, complying with \code{'NIfTI'}) to the \verb{RAS} coordinate system; if
#' \code{NULL}, the function looks for a \code{"vox2ras"} attribute on the array
#' @param source_mask,target_mask optional 3D mask arrays restricting where the
#' metric is evaluated; default \code{NULL} (no mask, evaluate everywhere). A
#' \code{target_mask} (on the target/fixed grid, same dimensions as
#' \code{target}) limits which target voxels drive the registration; a
#' \code{source_mask} (on the source/moving grid) drops samples that map outside
#' it. Non-zero voxels are included. Besides focusing the alignment on a region
#' of interest, masks speed things up by skipping background: the linear stage
#' samples only inside the mask, and the deformable stage skips warping voxels
#' outside the (dilated) mask. One mask per grid, shared across channels
#' @param source_points,target_points optional \code{N x 3} matrices of
#' corresponding landmark coordinates that add a surface/landmark term to the
#' \strong{deformable} (\code{"syn"}/\code{"syn_only"}) stage; default
#' \code{NULL} (no term). Row \code{i} of \code{target_points} (in the
#' target/fixed \verb{RAS}) and row \code{i} of \code{source_points} (in the
#' source/moving \verb{RAS}) must be the \emph{same} anatomical location, for
#' example corresponding cortical-surface vertices from a \code{'FreeSurfer'}
#' spherical registration. The term pulls the warp of each target point onto its
#' source correspondent, recovering cortical folding (\verb{gyrification})
#' that the intensity metric blurs over, while the image metric still drives
#' deep-brain \verb{subcortical} structures. Must be supplied together with
#' equal row counts. \strong{Points must be in the same \verb{RAS} frame as
#' \code{source_vox2ras}/\code{target_vox2ras}} (note \code{'FreeSurfer'}
#' surfaces use \verb{surface/tkr} \verb{RAS}, which differs from scanner
#' \verb{RAS} by \code{c_ras})
#' @param points_weight relative weight of the landmark term against the image
#' metric in the deformable stage; default \code{0.5}. Larger values follow the
#' landmarks more closely. The sparse landmark force is attenuated by
#' \code{syn_sigma} smoothing, so this typically needs tuning for a given point
#' count and spacing
#' @param type type of transform to estimate; one of \code{'rigid'} (6 degrees
#' of freedom), \code{'affine'} (12), \code{"syn"} (\verb{affine} followed
#' by a \verb{SDR} - symmetric \verb{diffeomorphic} deformation), or
#' \code{'syn_only'} (\verb{deformable} stage only — no \verb{affine} is
#' estimated; \code{init_transform} is used directly as the starting
#' \verb{affine}, useful when you already have a good linear alignment)
#' @param metric similarity metric: \code{"mattes"} (Mattes mutual information,
#' the default, best for cross-modal such as \code{'CT'}-\code{'MRI'}),
#' \code{"cc"} (normalized cross-correlation, for same-modality), or
#' \code{"meansquares"} (mean squared intensity difference). With multiple
#' channels, supply either a single metric (used for every pair) or a vector
#' with one metric per source/target pair
#' @param shrink_factors integer down-sampling factors, one per resolution level
#' (coarsest first); default \code{c(4, 2, 1)}
#' @param smoothing_sigmas Gaussian smoothing applied at each level, in voxels,
#' same length as \code{shrink_factors}; default \code{c(2, 1, 0)}
#' @param iterations maximum optimizer iterations per level; default
#' \code{c(1000, 500, 250)}
#' @param sampling_rate fraction of fixed voxels sampled to evaluate the metric
#' (speeds up large volumes); default \code{0.2}
#' @param interpolation output interpolation used when warping each modality
#' onto the target grid: \code{'trilinear'} (default), \code{'nearest'} (keeps
#' label or segmentation values intact), or \code{"bspline"} (cubic
#' \verb{Catmull-Rom}, higher quality, slower). Like \code{metric}, supply a
#' single value (applied to every channel) or one per source/target pair. This
#' affects only the returned warped images, never the optimization's internal
#' sampling (always \code{'trilinear'}, so every channel still produces
#' smooth gradients)
#' @param number_of_bins number of histogram bins for the \code{"mattes"}
#' metric; default \code{32}
#' @param seed random seed for the voxel sampler, for reproducibility
#' @param init_transform optional 4x4 initial \verb{RAS}-to-\verb{RAS}
#' transform (fixed to moving) to start from
#' @param syn_iterations,syn_sigma deformable stage controls (only used when
#' \code{type = "syn"} or \code{"syn_only"}): per-level iteration counts and the
#' Gaussian regularization sigma (in voxels) applied to the update field. Both
#' are recycled to the number of levels, so \code{syn_sigma} may be a vector to
#' vary the regularization per stage (e.g. \code{c(3, 3, 1)} to relax it at the
#' finest level for sharper detail)
#' @param verbose logical; if \code{TRUE} (default) print per-level and
#' per-iteration progress to the console, including the current stage, shrink
#' factor, smoothing sigma, cost metric, and step size (linear) or maximum
#' displacement (deformable); useful to monitor convergence on large volumes
#' @returns A list with:
#' \describe{
#' \item{\code{transform}}{the estimated 4x4 \verb{RAS}-to-\verb{RAS} linear
#' transform mapping \code{target} (fixed) coordinates to \code{source} (moving)
#' coordinates}
#' \item{\code{image}}{the (primary) \code{source} resampled onto the
#' \code{target} grid}
#' \item{\code{images}}{a list with every source channel resampled onto the
#' \code{target} grid using its own \code{interpolation}; \code{image} is the
#' first element. For single-image input this is a length-1 list}
#' \item{\code{forward_field},\code{inverse_field}}{(only for \code{"syn"}) the
#' deformation fields}
#' \item{\code{metric_trace}}{the metric value across optimizer iterations}
#' \item{\code{type},\code{metric}}{echoes of the inputs}
#' }
#' @seealso \code{\link{apply_transform3d}}, \code{\link{resample_3d_volume}}
#' @examples
#' \donttest{
#'
#' # --- synthetic same-modality example -------------------------------------
#' nd <- c(50, 50, 50)
#' vox2ras <- diag(4); vox2ras[1:3, 4] <- -25
#' blob <- function(cx, cy, cz, s = 6) {
#'   g <- expand.grid(x = 0:(nd[1]-1), y = 0:(nd[2]-1), z = 0:(nd[3]-1))
#'   array(exp(-((g$x-cx)^2 + (g$y-cy)^2 + (g$z-cz)^2) / (2*s^2)), nd)
#' }
#' target <- blob(25, 25, 25)
#' source <- blob(28, 23, 26)            # shifted by a known (3, -2, 1) mm
#'
#' res <- register_volume3d(
#'   source, target,
#'   source_vox2ras = vox2ras, target_vox2ras = vox2ras,
#'   type = "rigid", metric = "cc"
#' )
#' res$transform[1:3, 4]                 # ~ c(3, -2, 1)
#'
#'
#' # --- multimodal registration (several co-registered channels) ------------
#' # Two aligned modalities (e.g. a T1 and a T2) jointly drive the deformable
#' # stage. Each pair may use its own metric, and `weights` set their relative
#' # influence (normalized internally to sum to 1). All source channels must
#' # share a grid; likewise all target channels.
#' t1_target <- blob(25, 25, 25, s = 6)
#' t2_target <- blob(25, 25, 25, s = 9)  # same anatomy, different contrast
#' t1_source <- blob(27, 24, 26, s = 6)  # both channels share the same warp
#' t2_source <- blob(27, 24, 26, s = 9)
#'
#' res_mm <- register_volume3d(
#'   source = list(t1_source, t2_source),
#'   target = list(t1_target, t2_target),
#'   source_vox2ras = vox2ras, target_vox2ras = vox2ras,
#'   weights = c(2, 1),                  # T1 counts twice as much as T2
#'   metric = c("mattes", "cc"),         # one metric per channel
#'   type = "syn"
#' )
#' res_mm$transform[1:3, 4]
#'
#' }
#' @export
register_volume3d <- function(
    source, target,
    source_vox2ras = NULL, target_vox2ras = NULL,
    source_mask = NULL, target_mask = NULL,
    source_points = NULL, target_points = NULL, points_weight = 0.5,
    weights = NULL,
    type = c("rigid", "affine", "syn", "syn_only"),
    metric = "mattes",
    shrink_factors = c(4, 2, 1),
    smoothing_sigmas = c(2, 1, 0),
    iterations = c(1000, 500, 250),
    sampling_rate = 0.2,
    interpolation = "trilinear",
    number_of_bins = 32L,
    seed = 1L,
    init_transform = NULL,
    syn_iterations = c(40, 20, 0),
    syn_sigma = 3,
    verbose = TRUE) {

  type <- match.arg(type)

  # Accept either a single array or a list of co-registered modalities.
  if (!is.list(source)) source <- list(source)
  if (!is.list(target)) target <- list(target)
  n_pairs <- length(source)
  if (length(target) != n_pairs) {
    stop("`register_volume3d`: `source` and `target` must have the same number of channels.")
  }

  # Per-channel metric. A single metric (the default "mattes") is recycled to
  # every channel; otherwise supply one metric per pair. Each entry is validated
  # against the available options, so three channels left at the default all
  # use "mattes" rather than being silently read as c("mattes","cc","meansquares").
  metric_choices <- c("mattes", "cc", "meansquares")
  metric <- as.character(metric)
  if (length(metric) == 1L) metric <- rep(metric, n_pairs)
  if (length(metric) != n_pairs) {
    stop("`register_volume3d`: `metric` must have length 1 or length(source).")
  }
  if (!all(metric %in% metric_choices)) {
    stop("`register_volume3d`: `metric` must be one of ",
         paste(metric_choices, collapse = ", "), ".")
  }

  # Per-channel output interpolation, validated like `metric`. Controls how each
  # warped modality is resampled onto the target grid, e.g. "nearest" to keep a
  # segmentation's label values intact. The optimizer's internal metric sampling
  # is always trilinear, so this does not change what drives the registration.
  interp_choices <- c("trilinear", "nearest", "bspline")
  interpolation <- as.character(interpolation)
  if (length(interpolation) == 1L) interpolation <- rep(interpolation, n_pairs)
  if (length(interpolation) != n_pairs) {
    stop("`register_volume3d`: `interpolation` must have length 1 or length(source).")
  }
  if (!all(interpolation %in% interp_choices)) {
    stop("`register_volume3d`: `interpolation` must be one of ",
         paste(interp_choices, collapse = ", "), ".")
  }
  interp_codes <- vapply(interpolation,
    function(x) switch(x, nearest = 0L, bspline = 2L, 1L), integer(1))

  # Weights: default equal, normalized to sum to 1.
  if (is.null(weights)) weights <- rep(1.0, n_pairs)
  weights <- as.double(weights)
  if (length(weights) != n_pairs || anyNA(weights) ||
      any(weights < 0) || sum(weights) <= 0) {
    stop("`register_volume3d`: `weights` must be non-negative, length(source), and not all zero.")
  }
  weights <- weights / sum(weights)

  # A single vox2ras applies to every channel in each group (co-registered).
  source_vox2ras <- resolve_vox2ras(source[[1]], source_vox2ras, "source")
  target_vox2ras <- resolve_vox2ras(target[[1]], target_vox2ras, "target")

  # Convert each channel; all channels in a group must share the same grid.
  src_list <- lapply(source, as_double_volume, "source")
  tgt_list <- lapply(target, as_double_volume, "target")
  src_dim <- dim(src_list[[1]])
  tgt_dim <- dim(tgt_list[[1]])
  for (k in seq_len(n_pairs)) {
    if (!identical(dim(src_list[[k]]), src_dim)) {
      stop("`register_volume3d`: all `source` channels must share the same dimensions.")
    }
    if (!identical(dim(tgt_list[[k]]), tgt_dim)) {
      stop("`register_volume3d`: all `target` channels must share the same dimensions.")
    }
  }

  # Optional masks (NULL = none, unchanged behavior). A mask restricts where the
  # metric is evaluated and skips background voxels: `target_mask` (fixed/target
  # grid) limits which target voxels drive the registration; `source_mask`
  # (moving/source grid) drops samples that map outside it. Nonzero = include.
  # One mask per grid (shared across channels, which are co-registered).
  target_mask <- validate_mask(target_mask, tgt_dim, "target_mask")
  source_mask <- validate_mask(source_mask, src_dim, "source_mask")

  # Optional cortical landmark correspondences (only used by the deformable
  # stage). `target_points` (fixed/target RAS) and `source_points` (moving/source
  # RAS) are N x 3 matrices whose row i is the SAME cortical vertex in each space
  # (e.g. corresponding vertices from FreeSurfer spherical registration). They
  # add a weighted term that pulls the warp of each target point onto its source
  # correspondent, recovering gyrification the intensity metric blurs over.
  # IMPORTANT: points must be in the same RAS frame as the vox2ras matrices.
  source_points <- validate_points(source_points, "source_points")
  target_points <- validate_points(target_points, "target_points")
  if (is.null(source_points) != is.null(target_points)) {
    stop("`register_volume3d`: `source_points` and `target_points` must be supplied together.")
  }
  if (!is.null(source_points) && nrow(source_points) != nrow(target_points)) {
    stop("`register_volume3d`: `source_points` and `target_points` must have the same number of rows (corresponding vertices).")
  }
  points_weight <- as.double(points_weight)[[1L]]

  # The linear (rigid/affine) stage is driven by the primary pair only; extra
  # channels contribute only to the deformable stage (ANTs convention).
  if (n_pairs > 1 && type %in% c("rigid", "affine")) {
    warning("`register_volume3d`: multiple channels only affect the deformable ",
            "(syn / syn_only) stage; using the first pair for type = '", type, "'.")
  }
  src <- src_list[[1]]
  tgt <- tgt_list[[1]]
  metric1 <- metric[[1]]

  shrink_factors <- as.integer(shrink_factors)
  nlev <- length(shrink_factors)
  smoothing_sigmas <- recycle(as.double(smoothing_sigmas), nlev)
  iterations <- recycle(as.integer(iterations), nlev)

  # Default initialization: align the centers of mass of the primary pair in
  # RAS. Without this, volumes whose scanner origins differ (e.g. CT vs MRI from
  # different sessions) start with no overlap and the optimizer cannot move.
  if (is.null(init_transform)) {
    com_fixed <- center_of_mass(tgt, target_vox2ras)
    com_moving <- center_of_mass(src, source_vox2ras)
    init_transform <- diag(4)
    init_transform[1:3, 4] <- com_moving - com_fixed
  } else {
    init_transform <- as.matrix(init_transform)
  }

  run_linear <- function(linear_type, init) {
    register_linear_cpp(
      fixed = as.double(tgt), fixedDim = tgt_dim, fixedVox2Ras = target_vox2ras,
      moving = as.double(src), movingDim = src_dim, movingVox2Ras = source_vox2ras,
      type = linear_type, metric = metric1,
      shrinkFactors = shrink_factors, smoothingSigmas = smoothing_sigmas,
      iterations = iterations, samplingRate = as.double(sampling_rate),
      learningRate = 0, numberOfBins = as.integer(number_of_bins),
      seed = as.integer(seed), initTransform = init,
      fixedMask = target_mask, movingMask = source_mask,
      verbose = isTRUE(verbose))
  }

  syn_iters <- recycle(as.integer(syn_iterations), nlev)
  syn_sigma <- recycle(as.double(syn_sigma), nlev)   # per-level flow regularization

  # syn_only: skip all linear stages, use init_transform as the affine init
  if (type == "syn_only") {
    transform <- init_transform
    syn <- register_syn(
      tgt_list, tgt_dim, target_vox2ras, src_list, src_dim, source_vox2ras,
      metrics = metric, weights = weights, init_transform = transform,
      shrink_factors = shrink_factors, smoothing_sigmas = smoothing_sigmas,
      iterations = syn_iters, interp_codes = interp_codes,
      field_sigma = syn_sigma, seed = as.integer(seed),
      fixed_mask = target_mask, moving_mask = source_mask,
      fixed_points = target_points, moving_points = source_points,
      points_weight = points_weight,
      verbose = isTRUE(verbose))
    return(finalize_registration(list(
      transform = transform,
      forward_field = syn$forward_field,
      inverse_field = syn$inverse_field,
      image = syn$image,
      images = syn$images,
      metric_trace = syn$metric_trace,
      type = type,
      metric = metric
    ), source_vox2ras, target_vox2ras, tgt_dim))
  }

  # Staged initialization (ANTs-style): rigid -> affine -> deformable
  init <- init_transform
  trace <- numeric(0)

  if (type %in% c("affine", "syn")) {
    rigid <- run_linear("rigid", init)
    init <- rigid$transform
    trace <- c(trace, rigid$metric_trace)
    linear <- run_linear("affine", init)
  } else {
    linear <- run_linear("rigid", init)
  }
  trace <- c(trace, linear$metric_trace)
  transform <- linear$transform

  result <- list(
    transform = transform,
    type = type,
    metric = metric,
    metric_trace = trace
  )

  if (type == "syn") {
    syn <- register_syn(
      tgt_list, tgt_dim, target_vox2ras, src_list, src_dim, source_vox2ras,
      metrics = metric, weights = weights, init_transform = transform,
      shrink_factors = shrink_factors, smoothing_sigmas = smoothing_sigmas,
      iterations = syn_iters, interp_codes = interp_codes,
      field_sigma = syn_sigma, seed = as.integer(seed),
      fixed_mask = target_mask, moving_mask = source_mask,
      fixed_points = target_points, moving_points = source_points,
      points_weight = points_weight,
      verbose = isTRUE(verbose))
    result$forward_field <- syn$forward_field
    result$inverse_field <- syn$inverse_field
    result$metric_trace <- c(trace, syn$metric_trace)
    result$image <- syn$image
    result$images <- syn$images
    return(finalize_registration(result, source_vox2ras, target_vox2ras, tgt_dim))
  }

  # resample each moving channel onto the fixed grid using the linear transform,
  # each with its own output interpolation
  images <- lapply(seq_len(n_pairs), function(k) {
    apply_transform3d(
      src_list[[k]], source_vox2ras, transform,
      reference_dim = tgt_dim, reference_vox2ras = target_vox2ras,
      interpolation = interpolation[k])
  })
  result$image <- images[[1]]
  result$images <- images

  finalize_registration(result, source_vox2ras, target_vox2ras, tgt_dim)
}

# Tag a registration result with its class, attach the fixed-grid vox2ras to the
# deformation fields (so they are self-describing for save_registration), and
# record the grid geometry needed to re-apply or export the transform.
finalize_registration <- function(result, source_vox2ras, target_vox2ras, target_dim) {
  if (!is.null(result$forward_field)) {
    attr(result$forward_field, "vox2ras") <- target_vox2ras
  }
  if (!is.null(result$inverse_field)) {
    attr(result$inverse_field, "vox2ras") <- target_vox2ras
  }
  result$geometry <- list(
    source_vox2ras = source_vox2ras,
    target_vox2ras = target_vox2ras,
    target_dim = target_dim
  )
  class(result) <- "ravetools_register_volume3d"
  result
}


#' @title Apply a linear \verb{RAS} transform to resample a 3D volume
#' @description
#' Warps a moving volume onto a reference grid given a 4x4
#' \verb{RAS}-to-\verb{RAS} transform (such as the \code{transform} returned
#' by \code{\link{register_volume3d}}). The transform maps reference (fixed)
#' \verb{RAS} coordinates to moving \verb{RAS} coordinates.
#' @param volume moving 3D array to resample
#' @param vox2ras the moving volume's voxel-to-\verb{RAS} 4x4 transform
#' @param transform 4x4 \verb{RAS}-to-\verb{RAS} transform (fixed to moving)
#' @param reference_dim output dimension (the fixed grid); defaults to
#' \code{dim(volume)}
#' @param reference_vox2ras the fixed grid's voxel-to-\verb{RAS} transform;
#' defaults to \code{vox2ras}
#' @param interpolation \code{'trilinear'} (default), \code{'nearest'}, or
#' \code{'bspline'} (cubic \verb{Catmull-Rom})
#' @param na_fill value for out-of-bounds voxels; default \code{0}
#' @returns The resampled volume on the reference grid, with a \code{'vox2ras'}
#' attribute equal to \code{reference_vox2ras}.
#' @seealso \code{\link{register_volume3d}}
#' @export
apply_transform3d <- function(
    volume, vox2ras, transform,
    reference_dim = dim(volume), reference_vox2ras = vox2ras,
    interpolation = c("trilinear", "nearest", "bspline"), na_fill = 0) {

  interpolation <- match.arg(interpolation)
  interp_code <- switch(interpolation, trilinear = 1L, bspline = 2L, 0L)

  volume <- as_double_volume(volume, "volume")
  vox2ras <- as.matrix(vox2ras)
  reference_vox2ras <- as.matrix(reference_vox2ras)
  transform <- as.matrix(transform)

  # for a fixed voxel f: moving voxel = solve(vox2ras) %*% transform %*%
  #   reference_vox2ras %*% f  =>  newVoxToWorld = transform %*% reference_vox2ras
  new_vox_to_world <- transform %*% reference_vox2ras

  storage.mode(na_fill) <- "double"
  re <- resample3D(
    arrayDim = as.integer(reference_dim[1:3]),
    fromArray = volume,
    newVoxToWorldTransposed = t(new_vox_to_world),
    oldVoxToWorldTransposed = t(vox2ras),
    na = na_fill,
    interpolation = interp_code)

  out <- re[[1]]
  attr(out, "vox2ras") <- reference_vox2ras
  out
}


# ---- internal helpers -------------------------------------------------------

resolve_vox2ras <- function(x, v, name) {
  if (is.null(v)) v <- attr(x, "vox2ras")
  if (is.null(v)) {
    stop(sprintf(
      "`register_volume3d`: %s_vox2ras is missing. Provide a 4x4 voxel-to-RAS matrix (or set a 'vox2ras' attribute on `%s`).",
      name, name))
  }
  v <- as.matrix(v)
  if (nrow(v) == 3 && ncol(v) == 4) v <- rbind(v, c(0, 0, 0, 1))
  if (!all(dim(v) == c(4, 4))) {
    stop(sprintf("`register_volume3d`: %s_vox2ras must be a 4x4 (or 3x4) matrix.", name))
  }
  v
}

as_double_volume <- function(x, name) {
  d <- dim(x)
  if (length(d) < 3 || any(d[1:3] <= 1)) {
    stop(sprintf("`register_volume3d`: `%s` must be a 3D volume.", name))
  }
  if (length(d) > 3 && !all(d[-(1:3)] == 1)) {
    stop(sprintf("`register_volume3d`: `%s` has more than 3 non-singleton dimensions.", name))
  }
  storage.mode(x) <- "double"
  dim(x) <- d[1:3]
  x
}

# Validate an optional mask (NULL passes through) and return it as a flat double
# vector matching `ref_dim` (the corresponding image grid). Nonzero = include.
validate_mask <- function(m, ref_dim, name) {
  if (is.null(m)) return(NULL)
  m <- as_double_volume(m, name)
  if (!identical(dim(m), ref_dim)) {
    stop(sprintf("`register_volume3d`: `%s` must match its image dimensions (%s).",
                 name, paste(ref_dim, collapse = " x ")))
  }
  as.double(m)
}

# Validate an optional landmark point set (NULL passes through) and return it as
# an N x 3 double matrix of RAS coordinates.
validate_points <- function(p, name) {
  if (is.null(p)) return(NULL)
  p <- as.matrix(p)
  storage.mode(p) <- "double"
  if (ncol(p) != 3L) {
    stop(sprintf("`register_volume3d`: `%s` must be an N x 3 matrix of RAS coordinates.", name))
  }
  p
}

recycle <- function(v, n) {
  if (length(v) == n) return(v)
  if (length(v) == 1L) return(rep(v, n))
  if (length(v) > n) return(v[seq_len(n)])
  c(v, rep(v[length(v)], n - length(v)))
}

# Intensity-weighted center of the volume, in RAS. Negative intensities (e.g.
# CT air around -1000) are clamped to 0 so the center reflects tissue/bone.
# Falls back to the geometric (field-of-view) center if the volume is empty.
center_of_mass <- function(vol, vox2ras) {
  d <- dim(vol)
  m <- as.double(vol)
  m[!is.finite(m) | m < 0] <- 0
  dim(m) <- d
  xm <- .rowSums(m, d[1], d[2] * d[3])           # mass per x slice
  yz <- .colSums(m, d[1], d[2] * d[3])           # mass per (y, z), summed over x
  dim(yz) <- c(d[2], d[3])
  ym <- rowSums(yz)
  zm <- colSums(yz)
  tot <- sum(xm)
  if (!is.finite(tot) || tot <= 0) {
    ijk <- (d - 1) / 2                            # geometric center fallback
  } else {
    ijk <- c(
      sum((0:(d[1] - 1)) * xm),
      sum((0:(d[2] - 1)) * ym),
      sum((0:(d[3] - 1)) * zm)) / tot
  }
  as.numeric(vox2ras %*% c(ijk, 1))[1:3]
}

# Deformable (symmetric diffeomorphic) stage; see src/reg_syn.cpp.
# fixed_list / moving_list are lists of co-registered channels (one or more);
# metrics has one entry per channel. weights are normalized contributions.
register_syn <- function(fixed_list, fixed_dim, fixed_vox2ras,
                          moving_list, moving_dim, moving_vox2ras,
                          metrics, weights, init_transform,
                          shrink_factors, smoothing_sigmas, iterations,
                          interp_codes, field_sigma, seed,
                          fixed_mask = NULL, moving_mask = NULL,
                          fixed_points = NULL, moving_points = NULL, points_weight = 0.5,
                          grad_step = 0.2, total_sigma = 0, cc_radius = 2L,
                          verbose = FALSE) {
  # Mattes drives the linear stage; the deformable refinement uses local CC
  # (ANTs SyN's metric) unless mean squares was requested, applied per channel.
  deform_metric <- ifelse(metrics == "meansquares", "meansquares", "cc")
  fixed_dbl <- lapply(fixed_list, as.double)
  moving_dbl <- lapply(moving_list, as.double)
  # NULL point sets become 0 x 3 matrices (the C++ side treats N = 0 as "none").
  fixed_points_mat <- if (is.null(fixed_points)) matrix(numeric(0), 0L, 3L) else as.matrix(fixed_points)
  moving_points_mat <- if (is.null(moving_points)) matrix(numeric(0), 0L, 3L) else as.matrix(moving_points)
  register_syn_cpp(
    fixedList = fixed_dbl, fixedDim = fixed_dim, fixedVox2Ras = fixed_vox2ras,
    movingList = moving_dbl, movingDim = moving_dim, movingVox2Ras = moving_vox2ras,
    affineTransform = init_transform, metricNames = deform_metric,
    weights = as.double(weights),
    shrinkFactors = shrink_factors, smoothingSigmas = smoothing_sigmas,
    iterations = iterations, gradStep = as.double(grad_step),
    flowSigma = as.double(field_sigma),   # one entry per level (recycled upstream)
    totalSigma = as.double(total_sigma),
    ccRadius = as.integer(cc_radius),
    interpCodes = as.integer(interp_codes),
    fixedMask = fixed_mask, movingMask = moving_mask,
    fixedPoints = fixed_points_mat, movingPoints = moving_points_mat,
    pointsWeight = as.double(points_weight),
    verbose = isTRUE(verbose))
}

#' Decimate with 'FIR' or 'IIR' filter
#' @param x signal to be decimated
#' @param q integer factor to down-sample by
#' @param n filter order used in the down-sampling; default is \code{30}
#' if \code{ftype='fir'}, or \code{8} if \code{ftype='iir'}
#' @param ftype filter type, choices are \code{'fir'} (default) and
#' \code{'iir'}
#' @returns Decimated signal
#' @details This function is migrated from \code{gsignal} package,
#' but with padding and indexing fixed. The results agree with 'Matlab'.
#'
#' @examples
#'
#' x <- 1:100
#' y <- decimate(x, 2, ftype = "fir")
#' y
#'
#' # compare with signal package
#' z <- gsignal::decimate(x, 2, ftype = "fir")
#'
#' # Compare decimated results
#' plot(x, type = 'l')
#' points(seq(1,100, 2), y, col = "green")
#' points(seq(1,100, 2), z, col = "red")
#'
#'
#' @export
decimate <- function(
  x, q, n = if (ftype == "iir") 8 else 30, ftype = "fir") {
  if (q != round(q))
    stop("decimate only works with integer q.")
  l_x <- length(x)


  if (ftype == "fir") {
    npad <- ceiling(n / 2)
    lpad <- 2 * x[1] - x[(npad + 1):2]
    rpad <- 2 * x[l_x] - x[l_x - (1:npad)]
    inp <- c(lpad, x, rpad)

    b <- fir1(n, 1 / q)$b
    y <- fftfilt(b, inp)
    y <- y[ceiling(npad + n / 2) + (1:l_x)]
    y <- y[seq(1, length(x), by = q)]
  } else {
    # y <- gsignal::decimate(x, q, n, "iir")

    rip <- 0.05
    w <- 0.8 / q
    ba <- gsignal::cheby1(n, rip, w)

    while (
      n > 1 &&
      (
        all(ba$b == 0) ||
        abs(filtmag_db(ba$b, ba$a, w) + rip) > 1e-6
      )
    ) {
      n <- n - 1
      ba <- gsignal::cheby1(n, rip, w)
    }

    y <- filtfilt(ba$b, ba$a, x)

    ny <- ceiling(l_x / q)

    nbeg <- q - (q * ny - l_x)

    y <- y[seq(nbeg, l_x, by = q)]
  }

  y

}

# detect: The detection method ('neg', 'pos', 'both'). Default is 'neg'.
# segment_duration: Duration of each segment in seconds. Default is 5 minutes.
# stdmin: Threshold factor for detection. Default is 5.
# stdmax: Threshold factor for artifact removal. Default is 50.
# w_pre: Pre-event window size.
# w_post: Post-event window size.
# min_ref_per: Minimum refractory period in milliseconds. Default is 1.5.

lillie.test <- function(x) {
  DNAME <- deparse(substitute(x))
  x <- sort(x[stats::complete.cases(x)])
  n <- length(x)
  if (n < 5)
    stop("sample size must be greater than 4")
  p <- stats::pnorm((x - mean(x)) / sd(x))
  Dplus <- max(seq(1:n) / n - p)
  Dminus <- max(p - (seq(1:n) - 1) / n)
  K <- max(Dplus, Dminus)
  if (n <= 100) {
    Kd <- K
    nd <- n
  } else {
    Kd <- K * ((n / 100)^0.49)
    nd <- 100
  }
  pvalue <- exp(-7.01256 * Kd^2 * (nd + 2.78019) + 2.99587 *
                  Kd * sqrt(nd + 2.78019) - 0.122119 + 0.974598 / sqrt(nd) +
                  1.67997 / nd)
  if (pvalue > 0.1) {
    KK <- (sqrt(n) - 0.01 + 0.85 / sqrt(n)) * K
    if (KK <= 0.302) {
      pvalue <- 1
    } else if (KK <= 0.5) {
      pvalue <- 2.76773 - 19.828315 * KK + 80.709644 *
        KK^2 - 138.55152 * KK^3 + 81.218052 * KK^4
    } else if (KK <= 0.9) {
      pvalue <- -4.901232 + 40.662806 * KK - 97.490286 *
        KK^2 + 94.029866 * KK^3 - 32.355711 * KK^4
    } else if (KK <= 1.31) {
      pvalue <- 6.198765 - 19.558097 * KK + 23.186922 *
        KK^2 - 12.234627 * KK^3 + 2.423045 * KK^4
    } else {
      pvalue <- 0
    }
  }
  RVAL <- list(statistic = c(D = K), p.value = pvalue, method = "Lilliefors (Kolmogorov-Smirnov) normality test",
               data.name = DNAME)
  class(RVAL) <- "htest"
  return(RVAL)
}

spike_detection <- function(
    x,
    sample_rate,
    x_bp2 = NULL,
    x_bp4 = NULL,
    detect_method = c("negative", "positive", "both"),
    segment_duration = 5,
    std_min = 5,
    std_max = 50,
    w_pre = 20,
    w_post = 44,
    min_refractory = 1.5
) {

  detect_method <- match.arg(detect_method)

  spike_signal <- x
  nyquist <- sample_rate / 2

  if (is.null(x_bp2)) {
    ellip2 <- ravetools::ellip(
      n = 2,
      Rp = 0.1,
      Rs = 40,
      w = c(300, 3000) / nyquist,
      type = "pass"
    )
    signal_bp2 <- ravetools::filtfilt(b = ellip2$b, a = ellip2$a, x = spike_signal)
  } else {
    signal_bp2 <- x_bp2
  }

  if (is.null(x_bp4)) {
    ellip4 <- ravetools::ellip(
      n = 4,
      Rp = 0.1,
      Rs = 40,
      w = c(300, 3000) / nyquist,
      type = "pass"
    )
    signal_bp4 <- ravetools::filtfilt(b = ellip4$b, a = ellip4$a, x = spike_signal)
  } else {
    signal_bp4 <- x_bp4
  }

  ref <- ceiling(min_refractory * sample_rate / 1000)
  sample_ref <- floor(ref / 2)

  # We process data by segments
  total_n_timepoints <- length(spike_signal)
  segment_n_timepoints <- sample_rate * segment_duration * 60
  n_segments <- ceiling(total_n_timepoints / segment_n_timepoints)
  pre_to_post <- seq(-w_pre, w_post)


  spike_timepoints <- lapply(seq_len(n_segments), function(ii_seg) {
    timepoint_start <- max((ii_seg - 1) *  segment_n_timepoints + 1 - w_pre, 1)
    timepoint_end <- min(ii_seg * segment_n_timepoints + w_pre, total_n_timepoints)
    time_sel <- seq(timepoint_start, timepoint_end)

    slice_orig <- spike_signal[time_sel]
    slice_bp2 <- signal_bp2[time_sel]
    slice_bp4 <- signal_bp4[time_sel]

    threshold <- std_min * median(abs(slice_bp4)) / 0.6745
    threhsold_max <- threshold * std_min

    switch(
      detect_method,
      "negative" = {
        slice_threshold <- -slice_bp4
      },
      "positive" = {
        slice_threshold <- slice_bp4
        xaux <- which(slice_bp4 > threshold)
      },
      {
        # both
        slice_threshold <- abs(slice_bp4)
      }
    )

    xaux <- which(slice_threshold > threshold)
    xaux <- xaux[xaux > w_pre & xaux < (length(time_sel) - w_post)]

    tmp_env <- new.env(parent = emptyenv())
    tmp_env$xaux0 <- 0
    tmp_env$index <- NULL

    lapply(seq_along(xaux), function(ii) {

      xaux_ii <- xaux[[ii]]

      if (xaux_ii >= tmp_env$xaux0 + ref) {
        iaux <- which.min(slice_bp2[(xaux_ii - 1) + seq_len(sample_ref)])

        if ( max(abs(slice_bp2[pre_to_post + xaux_ii])) < threhsold_max ) {
          tmp_env$xaux0 <- xaux_ii + iaux - 1
          tmp_env$index <- c(tmp_env$index, tmp_env$xaux0)
        }
      }
      return()
    })


    spike_timepoints <- time_sel[tmp_env$index]

    spike_timepoints
  })

  spike_timepoints <- unlist(spike_timepoints)

  unique(spike_timepoints)
}



extract_waveforms_for_channel <- function(
    spike_timepoints, x_bp2, w_pre = 20, w_post = 44, int_factor = 5,
    detect_method = c("negative", "positive", "both")) {

  detect_method <- match.arg(detect_method)

  signal_bp2 <- x_bp2

  total_spikes <- length(spike_timepoints)

  len_waveform <- w_pre + w_post

  # Range to extract for each spike (with padding for interpolation)
  extra <- 2
  window_len <- len_waveform + 2 * extra
  s <- seq_len(window_len) - 1
  ints <- seq(0, window_len - 1, by = 1 / int_factor)

  # time-index offset
  pre_to_post <- seq(-w_pre - extra, w_post + extra - 1)

  # subset and super-sample
  interpolated_spikes <- sapply(seq_len(total_spikes), function(ii) {

    idx <- pre_to_post + spike_timepoints[[ii]]
    if (ii == 1) {
      idx[idx < 1] <- 1
    } else if (ii == total_spikes) {
      idx[idx > total_spikes] <- total_spikes
    }
    spike <- signal_bp2[idx]

    # Spline interpolation (like scipy's splrep/splev)
    spline_fit <- stats::splinefun(s, spike, method = "natural")
    spline_fit(ints)
  }, simplify = TRUE, USE.NAMES = FALSE)
  interpolated_spikes <- t(interpolated_spikes)

  # Find alignment within window
  align_win <- seq((w_pre + extra - 1) * int_factor + 1, (w_pre + extra + 1) * int_factor + 1)

  switch(
    detect_method,
    "positive" = {
      iaux <- apply(interpolated_spikes[, align_win, drop = FALSE], 1, which.max)
    },
    "negative" = {
      iaux <- apply(interpolated_spikes[, align_win, drop = FALSE], 1, which.min)
    },
    {
      iaux <- apply(abs(interpolated_spikes[, align_win, drop = FALSE]), 1, which.max)
    }
  )

  iaux <- iaux + (w_pre + extra - 1) * int_factor
  pre_to_post_interp <- seq(- w_pre * int_factor + int_factor, w_post * int_factor,
                            by = int_factor)

  spike_waveforms <- sapply(seq_along(iaux), function(ii) {
    interpolated_spikes[ii, iaux[[ii]] + pre_to_post_interp]
  }, simplify = TRUE, USE.NAMES = FALSE)
  spike_waveforms <- t(spike_waveforms)

  # validate
  # matplot(t(spike_waveforms[sample(total_spikes, 1000, ), ]),
  #         type = "l", lty = 1, col = "gray80")
  spike_waveforms
}


haar_feature_extraction <- function(
    spike_waveforms,
    level = 4,
    ls = 64,
    max_inputs = ceiling(0.75 * 64),
    min_inputs = 10,
    nd = 10
) {
  # spikes: matrix or list of spike waveforms, each row or element is a waveform

  nspk <- nrow(spike_waveforms)
  spike_mat <- spike_waveforms

  cc <- matrix(0, nrow = nspk, ncol = ls)

  cc <- t(apply(spike_mat, 1L, function(spike) {
    # Haar wavelet decomposition
    w <- waveslim::dwt(spike, wf = "haar", n.levels = level, boundary = "periodic")
    coeffs <- c(unlist(w[paste0("d", seq_len(level))]), w[[paste0("s", level)]])
    unname(coeffs[seq_len(ls)])
  }))

  cc[is.na(cc)] <- 0
  # ls <- ncol(cc)

  ks <- sapply(seq_len(ls), function(ii) {

    coef_ii <- cc[, ii]
    m <- mean(coef_ii)
    thr_dist <- stats::sd(coef_ii) * 3
    thr_dist_min <- m - thr_dist
    thr_dist_max <- m + thr_dist
    aux <- cc[coef_ii > thr_dist_min & coef_ii < thr_dist_max, ii]

    stats <- 0
    if (length(aux) > 10) {
      stats <- lillie.test(aux)$statistic
    }

    unname(stats)

  })

  sorted_indices <- order(ks)
  # ind <- sorted_indices
  A <- ks[sorted_indices][(length(ks) - max_inputs + 1):length(ks)]
  ncoeff <- length(A)
  maxA <- max(A)

  if (nd > 1 && ncoeff >= nd) {
    d <- ((A[nd:length(A)] - A[1:(length(A) - nd + 1)]) / maxA) * (ncoeff / nd)
    all_above1 <- which(d >= 1)
  } else {
    d <- numeric(0)
    all_above1 <- integer(0)
  }

  if (length(all_above1) >= 2) {
    aux2 <- diff(all_above1)
    # temp_bla <- stats::filter(c(0, aux2, 0), rep(1/3, 3), sides = 2, circular = TRUE)
    temp_bla <- ravetools::convolve_signal(c(0, aux2, 0), rep(1 / 3, 3))
    temp_bla[1] <- aux2[1]
    temp_bla[length(temp_bla)] <- aux2[length(aux2)]
    thr_knee_diff <- all_above1[which(temp_bla[-1] == 1)[1]] + (nd / 2)
    inputs <- max_inputs - thr_knee_diff
  } else {
    inputs <- min_inputs
  }

  if (inputs > max_inputs) {
    inputs <- max_inputs
  } else if (inputs < min_inputs) {
    inputs <- min_inputs
  }

  coeff <- sorted_indices[(length(sorted_indices) - inputs + 1):length(sorted_indices)]

  return(cc[, coeff])
}

filter_initialize <- function(b, a, x) {
  # DIPSAUS DEBUG START
  # filter <- ravetools::butter(1, c(50 / 250), "low")
  # b <- filter$b
  # a <- filter$a
  # x <- sample_signal(500)

  if (a[[1]] != 0) {
    a <- a / a[[1]]
    b <- b / a[[1]]
  }

  # make sure a, b share the same order
  na <- length(a)
  nb <- length(b)

  if ( na > nb ) {
    b <- c(b, rep(0, na - nb))
    n <- na
  } else {
    a <- c(a, rep(0, nb - na))
    n <- nb
  }

  # length of edge transients
  nf <- max(1, 3 * (n - 1))

  if (length(x) <= nf) {
    stop("Cannot apply the filter: the input signal is too short")
  }

  # n is nfilt in matlab filtfilt -> getCoeffsAndInitialConditions
  # nf is nfact in matlab filtfilt -> getCoeffsAndInitialConditions

  # compute the initial condition if n > 1
  if ( n > 2 ) {
    z1 <- diag(1, n - 1) - cbind( -a[-1], rbind(diag(1, n - 2), 0) )
    z2 <- b[-1] - b[1] * a[-1]
    z <- qr.solve(z1, z2, tol = 1e-30)

  } else if ( n == 2 ) {
    z1 <- 1 + a[-1]
    z2 <- b[-1] - b[1] * a[-1]
    z <- z2 / z1
  } else {
    z <- numeric(0)
  }
  list(
    a = a,
    b = b,
    z = z,
    nfilt = n,
    nfact = nf
  )
}

filtfilt_naive <- function(b, a, y, z, nfact) {

  # pad edges with reverse signals
  ny <- length(y)
  ytmp <- c(
    2.0 * y[[1]] - y[seq(nfact + 1, 2, by = -1)],
    y,
    2.0 * y[[ny]] - y[seq(ny - 1, ny - nfact, by = -1)]
  )

  ytmp <- filter_signal(b = b, a = a, x = ytmp, z = z * ytmp[[1]])[[1]]
  ytmp <- rev(ytmp)
  ytmp <- filter_signal(b = b, a = a, x = ytmp, z = z * ytmp[[1]])[[1]]

  re <- ytmp[seq(length(ytmp) - nfact, nfact + 1, by = -1)]
  return(re)
}

filtfilt_naive2 <- function(b, a, y, z, nfact) {

  xt <- 2.0 * y[[1]] - y[seq(nfact + 1, 2, by = -1)]
  z0 <- filter_signal(b = b, a = a, x = xt, z = z * xt[[1]])[[2]]
  yc2 <- filter_signal(b = b, a = a, x = y, z = z0)
  z0 <- yc2[[2]]
  yc2 <- yc2[[1]]

  ny <- length(y)
  xt <- 2.0 * y[[ny]] - y[seq(ny - 1, ny - nfact, by = -1)]
  yc3 <- filter_signal(b = b, a = a, x = xt, z = z0)[[1]]

  z0 <- filter_signal(b = b, a = a, x = rev(yc3), z = z * yc3[[length(yc3)]])[[2]]
  yc5 <- filter_signal(b = b, a = a, x = rev(yc2), z = z0)[[1]]

  return(rev( yc5 ))
}

#' @title Forward and reverse filter a one-dimensional signal
#' 
#' @description The result has been tested against 'Matlab' \code{filtfilt}
#'    function. Currently this function only supports one filter at a time.
#' 
#' @param b one-dimensional real numerical vector, the moving-average
#'    coefficients of an \code{ARMA} filter; alternatively, a \code{Sos}
#'    (second-order sections) object from \code{gsignal}, in which case
#'    \code{a} is ignored and \code{gsignal::filtfilt} is used internally
#' 
#' @param a the auto-regressive (recursive) coefficients of an \code{ARMA}
#'    filter; ignored when \code{b} is a \code{Sos} object
#' 
#' @param x numerical vector or matrix input (real value)
#' 
#' @returns The filtered signal, normally the same length as the input signal
#'    \code{x}.
#' @examples
#'
#' t <- seq(0, 1, by = 0.01)
#' x <- sin(2 * pi * t * 2.3)
#' bf <- gsignal::butter(2, c(0.15, 0.3))
#'
#' res <- filtfilt(bf$b, bf$a, x)
#'
#' ## Matlab (2022a) equivalent:
#' # t = [0:0.01:1];
#' # x = sin(2 * pi * t * 2.3);
#' # [b,a] = butter(2,[.15,.3]);
#' # res = filtfilt(b, a, x)
#'
#' @export
filtfilt <- function(b, a = 1, x) {

  # DIPSAUS DEBUG START
  # filter <- ravetools::butter(5, c(1 / 250, 50 / 250), "pass")
  # b <- filter$b
  # a <- filter$a
  # x <- sample_signal(500)

  if (!inherits(b, "Sos") && length(a) > 1 && rcond_filter_ar(a) <= .Machine$double.eps) {
    # This filter is ARMA filter with unstable AR part, which cannot be applied with the current implementation of filtfilt.
    # Conversion to SOS form may help if the filter is IIR
    b <- gsignal::as.Sos(gsignal::Arma(b = b, a = a))
  }

  # SOS dispatch: b is a Sos object, use gsignal::filtfilt
  if (inherits(b, "Sos")) {
    if (is.matrix(x)) {
      dn <- dimnames(x)
      out <- apply(x, 2L, function(col) gsignal::filtfilt(b, x = col))
      dimnames(out) <- dn
      return(out)
    }
    return(gsignal::filtfilt(b, x = x))
  }

  if (is.matrix(x)) {
    # DIPSAUS DEBUG START
    # b <- rnorm(6)
    # a <- 1
    # x <- array(rnorm(500), c(250, 2))

    init <- filter_initialize(b, a, x[, 1])
    nx <- nrow(x)

    if (length(a) == 1) {
      # FIR filter
      nfact <- init$nfact
      pad_pre <- sweep(
        - x[seq(nfact + 1, 2, by = -1), , drop = FALSE],
        MARGIN = 2L, FUN = "+",
        STATS = 2 * as.vector(x[1, ])
      )
      pad_post <- sweep(
        - x[seq(nx - 1, nx - nfact, by = -1), , drop = FALSE],
        MARGIN = 2L, FUN = "+",
        STATS = 2 * as.vector(x[nx, ])
      )
      padded_x <- rbind(pad_pre, x, pad_post)
      nr_px <- as.integer(nrow(padded_x))

      cond <- outer(init$z, padded_x[1, ], "*")
      ncond <- nrow(cond)
      re <- fftfilt(b = init$b, x = padded_x)

      # add initial condition response
      re[seq_len(ncond), ] <- re[seq_len(ncond), ] + cond

      # reverse
      re <- re[seq.int(nr_px, 1L, by = -1L), , drop = FALSE]
      cond <- outer(init$z, re[1, ], "*")
      re <- fftfilt(b = init$b, x = re)
      re[seq_len(ncond), ] <- re[seq_len(ncond), ] + cond

      # subset
      re <- re[seq(nr_px - nfact, nfact + 1, by = -1), , drop = FALSE]
      return(re)
    } else {
      re <- apply(x, 2L, function(xi) {
        if (nx < 10000) {
          filtfilt_naive(b = init$b, a = init$a, y = x, z = init$z, nfact = init$nfact)
        } else {
          filtfilt_naive2(b = init$b, a = init$a, y = x, z = init$z, nfact = init$nfact)
        }
      })
    }
  } else {
    nx <- length(x)
    init <- filter_initialize(b, a, x)

    if (nx < 10000) {
      re <- filtfilt_naive(b = init$b, a = init$a, y = x, z = init$z, nfact = init$nfact)
    } else {
      re <- filtfilt_naive2(b = init$b, a = init$a, y = x, z = init$z, nfact = init$nfact)
    }
  }

  return(re)

  # # sanity-check
  # x <- array(sample_signal(5000), c(1000, 5))
  # filter <- design_filter_fir(200, high_pass_freq = 0.5, filter_order = 50)
  # expected <- apply(x, 2, function(xi) {
  #   filtfilt(filter$b, filter$a, xi)
  # })
  # actual <- filtfilt(filter$b, filter$a, x)
  # range(actual - expected)
  #
  # # speed check
  # microbenchmark::microbenchmark(
  #   expected = apply(x, 2, function(xi) {
  #     filtfilt(filter$b, filter$a, xi)
  #   }),
  #   actual = filtfilt(filter$b, filter$a, x)
  # )


  re
}


#' @title Filter one-dimensional signal
#' @description The function is written from the scratch. The result has been
#' compared against the 'Matlab' \code{filter} function with one-dimensional
#' real inputs. Other situations such as matrix \code{b} or multi-dimensional
#' \code{x} are not implemented. For double filters (forward-backward),
#' see \code{\link{filtfilt}}.
#' @param b one-dimensional real numerical vector, the moving-average
#' coefficients of an \code{ARMA} filter
#' @param a the auto-regressive (recursive) coefficients of an \code{ARMA} filter
#' @param x numerical vector input (real value)
#' @param z initial condition, must have length of \code{n-1}, where \code{n}
#' is the maximum of lengths of \code{a} and \code{b}; default is all zeros
#' @returns A list of two vectors: the first vector is the filtered signal;
#' the second vector is the final state of \code{z}
#'
#' @examples
#'
#'
#' t <- seq(0, 1, by = 0.01)
#' x <- sin(2 * pi * t * 2.3)
#' bf <- gsignal::butter(2, c(0.15, 0.3))
#'
#' res <- filter_signal(bf$b, bf$a, x)
#' y <- res[[1]]
#' z <- res[[2]]
#'
#' ## Matlab (2022a) equivalent:
#' # t = [0:0.01:1];
#' # x = sin(2 * pi * t * 2.3);
#' # [b,a] = butter(2,[.15,.3]);
#' # [y,z] = filter(b, a, x)
#'
#'
#' @export
filter_signal <- function(b, a, x, z) {

  na <- length(a)
  nb <- length(b)

  if ( na > nb ) {
    b <- c(b, rep(0, na - nb))
    n <- na
  } else {
    a <- c(a, rep(0, nb - na))
    n <- nb
  }

  if (missing(z)) {
    z <- rep(0.0, n - 1)
  } else {
    if (length(z) < n - 1) {
      stop(sprintf("`filter`: initial condition `z` must have length >= %d", n - 1))
    }
  }

  if (!is.double(a)) {
    a <- as.double(a)
  }
  if (!is.double(b)) {
    b <- as.double(b)
  }
  if (!is.double(z)) {
    z <- as.double(z)
  }
  if (!is.double(x)) {
    x <- as.double(x)
  }
  return(cpp_filter(b, a, x, z))
}

# # Design filters

#' Design a digital filter
#' @description
#' Provides 'FIR' and 'IIR' filter options; default is 'FIR', see also
#' \code{\link{design_filter_fir}}; for 'IIR' filters, see
#' \code{\link{design_filter_iir}}.
#'
#' @param sample_rate data sample rate
#' @param data data to be filtered, can be optional (\code{NULL})
#' @param method filter method, options are \code{"fir"} (default),
#' \code{"butter"}, \code{"cheby1"}, \code{"cheby2"}, and \code{"ellip"}
#' @param high_pass_freq,low_pass_freq high-pass or low-pass frequency,
#' see \code{\link{design_filter_fir}} or \code{\link{design_filter_iir}}
#' @param high_pass_trans_freq,low_pass_trans_freq transition bandwidths,
#' see \code{\link{design_filter_fir}} or \code{\link{design_filter_iir}}
#' @param passband_ripple allowable pass-band ripple in decibel; default is
#' \code{0.1}
#' @param stopband_attenuation minimum stop-band attenuation (in decibel) at
#' transition frequency; default is \code{40} dB.
#' @param filter_order suggested filter order; for 'IIR' methods, see
#' \code{\link{design_filter_iir}} for details on how this interacts with
#' \code{use_sos}; for 'FIR' methods 'RAVE' may or may not adopt this
#' suggestion depending on the data and numerical feasibility
#' @param use_sos logical; passed to \code{\link{design_filter_iir}} for 'IIR'
#' methods (ignored for 'FIR'). When \code{TRUE} (default), filtering is done
#' via 'SOS' form using \code{gsignal::filtfilt}; when \code{FALSE}, the
#' ravetools 'ARMA' \code{filtfilt} is used.
#' @param data_size used by 'FIR' filter design to determine maximum order,
#' ignored in 'IIR' filters; automatically derived from \code{data}
#' @param ... passed to filter generator functions
#' @returns If \code{data} is specified and non-empty, this function returns
#' filtered data via forward and backward \code{filtfilt}; if \code{data} is
#' \code{NULL}, then returns the generator function.
#'
#' @examples
#'
#'
#' sample_rate <- 200
#' t <- seq(0, 10, by = 1 / sample_rate)
#' x <- sin(t * 4 * pi) + sin(t * 20 * pi) +
#'   2 * sin(t * 120 * pi) + rnorm(length(t), sd = 0.4)
#'
#' # ---- Using FIR ------------------------------------------------
#'
#' # Low-pass filter
#' y1 <- design_filter(
#'   data = x,
#'   sample_rate = sample_rate,
#'   low_pass_freq = 3, low_pass_trans_freq = 0.5
#' )
#'
#' # Band-pass cheby1 filter 8-12 Hz with custom transition
#' y2 <- design_filter(
#'   data = x,
#'   method = "cheby1",
#'   sample_rate = sample_rate,
#'   low_pass_freq = 12, low_pass_trans_freq = .25,
#'   high_pass_freq = 8, high_pass_trans_freq = .25
#' )
#'
#' y3 <- design_filter(
#'   data = x,
#'   sample_rate = sample_rate,
#'   low_pass_freq = 80,
#'   high_pass_freq = 30
#' )
#'
#' oldpar <- par(mfrow = c(2, 1),
#'               mar = c(3.1, 2.1, 3.1, 0.1))
#' plot(t, x, type = 'l', xlab = "Time", ylab = "",
#'      main = "Mixture of 2, 10, and 60Hz", xlim = c(0,1))
#' # lines(t, y, col = 'red')
#' lines(t, y3, col = 'green')
#' lines(t, y2, col = 'blue')
#' lines(t, y1, col = 'red')
#' legend(
#'   "topleft", c("Input", "Low: 3Hz", "Pass 8-12Hz", "Pass 30-80Hz"),
#'   col = c(par("fg"), "red", "blue", "green"), lty = 1,
#'   cex = 0.6
#' )
#'
#' # plot pwelch
#' pwelch(x, fs = sample_rate, window = sample_rate * 2,
#'        noverlap = sample_rate, plot = 1, ylim = c(-100, 10))
#' pwelch(y1, fs = sample_rate, window = sample_rate * 2,
#'        noverlap = sample_rate, plot = 2, col = "red")
#' pwelch(y2, fs = sample_rate, window = sample_rate * 2,
#'        noverlap = sample_rate, plot = 2, col = "blue")
#' pwelch(y3, fs = sample_rate, window = sample_rate * 2,
#'        noverlap = sample_rate, plot = 2, col = "green")
#'
#'
#' # ---- Clean this demo --------------------------------------------------
#' par(oldpar)
#'
#'
#' @export
design_filter <- function(
    sample_rate,
    data = NULL,
    method = c("fir_kaiser", "firls", "fir_remez", "butter", "cheby1", "cheby2", "ellip"),
    high_pass_freq = NA, high_pass_trans_freq = NA,
    low_pass_freq = NA, low_pass_trans_freq = NA,
    passband_ripple = 0.1, stopband_attenuation = 40,
    filter_order = NA, use_sos = TRUE,
    ..., data_size = length(data)
) {
  method <- match.arg(method)
  if (data_size <= 0) {
    data_size <- NA
  }
  if (length(data) && is.na(data_size)) {
    data_size <- floor(length(data) - 1) / 3
  }

  if ( startsWith(method, "fir") ) {
    method <- list(
      fir_kaiser = "kaiser",
      firls = "firls",
      fir_remez = "remez"
    )[[ method ]]
    filter <- design_filter_fir(
      method = method,
      sample_rate = sample_rate,
      filter_order = filter_order,
      data_size = data_size,
      high_pass_freq = high_pass_freq,
      high_pass_trans_freq = high_pass_trans_freq,
      low_pass_freq = low_pass_freq,
      low_pass_trans_freq = low_pass_trans_freq,
      stopband_attenuation = stopband_attenuation,
      ...
    )
  } else {
    filter <- design_filter_iir(
      method = method,
      sample_rate = sample_rate,
      filter_order = filter_order,
      use_sos = use_sos,
      high_pass_freq = high_pass_freq,
      high_pass_trans_freq = high_pass_trans_freq,
      low_pass_freq = low_pass_freq,
      low_pass_trans_freq = low_pass_trans_freq,
      passband_ripple = passband_ripple,
      stopband_attenuation = stopband_attenuation
    )
  }

  if (length(data)) {
    if (use_sos && !is.null(filter$sos)) {
      re <- filtfilt(b = filter$sos, x = data)
    } else {
      re <- filtfilt(b = filter$b, a = filter$a, x = data)
    }
  } else {
    re <- filter
  }

  re

}

#' @title Design an 'IIR' filter
#' @param sample_rate sampling frequency
#' @param method filter method name, choices are \code{"butter"},
#' \code{"cheby1"}, \code{"cheby2"}, and \code{"ellip"}
#' @param filter_order suggested filter order. When \code{use_sos=TRUE}
#' (default) this order is used as-is since 'SOS' form is numerically stable
#' at any order. When \code{use_sos=FALSE} and 'ARMA' form becomes numerically
#' unstable, the order will be automatically reduced with a warning;
#' leave \code{NA} (default) to let the function choose automatically.
#' @param use_sos logical; when \code{TRUE} (default), the filter is also
#' stored as a \code{Sos} (second-order sections) object in \code{filter$sos},
#' and no stability-based order correction is performed. When \code{FALSE},
#' the order is capped at the maximum stable 'ARMA' order (with a warning if
#' a specific \code{filter_order} was requested).
#' @param high_pass_freq high-pass frequency; default is \code{NA} (no high-pass
#' filter will be applied)
#' @param high_pass_trans_freq high-pass frequency band-width; default
#' is automatically inferred from filter type.
#' @param low_pass_freq low-pass frequency; default is \code{NA} (no
#' low-pass filter will be applied)
#' @param low_pass_trans_freq low-pass frequency band-width; default
#' is automatically inferred from filter type.
#' @param passband_ripple allowable pass-band ripple in decibel; default is
#' \code{0.1}
#' @param stopband_attenuation minimum stop-band attenuation (in decibel) at
#' transition frequency; default is \code{40} dB.
#' @returns A filter object with \code{$b}/\code{$a} 'ARMA' coefficients and
#' \code{$sos} second-order sections (a \code{gsignal} \code{Sos} object).
#'
#' @examples
#'
#' sample_rate <- 500
#'
#' my_diagnose <- function(
#'     filter, vlines = c(8, 12), cutoffs = c(-3, -6)) {
#'   diagnose_filter(
#'     b = filter$b,
#'     a = filter$a,
#'     fs = sample_rate,
#'     vlines = vlines,
#'     cutoffs = cutoffs
#'   )
#' }
#'
#' # ---- Default using butterworth to generate 8-12 bandpass filter ----
#'
#' # Butterworth filter with cut-off frequency
#' # 7 ~ 13 (default transition bandwidth is 1Hz) at -3 dB
#' filter <- design_filter_iir(
#'   method = "butter",
#'   low_pass_freq = 12,
#'   high_pass_freq = 8,
#'   sample_rate = 500
#' )
#'
#' filter
#'
#' my_diagnose(filter)
#'
#' ## explicit bandwidths and attenuation (sharper transition)
#'
#' # Butterworth filter with cut-off frequency
#' # passband ripple is 0.5 dB (8-12 Hz)
#' # stopband attenuation is 40 dB (5-18 Hz)
#' filter <- design_filter_iir(
#'   method = "butter",
#'   low_pass_freq = 12, low_pass_trans_freq = 6,
#'   high_pass_freq = 8, high_pass_trans_freq = 3,
#'   sample_rate = 500,
#'   passband_ripple = 0.5,
#'   stopband_attenuation = 40
#' )
#'
#' filter
#'
#' my_diagnose(filter)
#'
#' # ---- cheby1 --------------------------------
#'
#' filter <- design_filter_iir(
#'   method = "cheby1",
#'   low_pass_freq = 12,
#'   high_pass_freq = 8,
#'   sample_rate = 500
#' )
#'
#' my_diagnose(filter)
#'
#' # ---- cheby2 --------------------------------
#'
#' filter <- design_filter_iir(
#'   method = "cheby2",
#'   low_pass_freq = 12,
#'   high_pass_freq = 8,
#'   sample_rate = 500
#' )
#'
#' my_diagnose(filter)
#'
#' # ----- ellip ---------------------------------
#'
#' filter <- design_filter_iir(
#'   method = "ellip",
#'   low_pass_freq = 12,
#'   high_pass_freq = 8,
#'   sample_rate = 500
#' )
#'
#' my_diagnose(filter)
#'
#'
#'
#'
#' @export
design_filter_iir <- function(
    method = c("butter", "cheby1", "cheby2", "ellip"),
    sample_rate, filter_order = NA, use_sos = TRUE,
    high_pass_freq = NA, high_pass_trans_freq = NA,
    low_pass_freq = NA, low_pass_trans_freq = NA,
    passband_ripple = 0.1, stopband_attenuation = 40
) {
  method <- match.arg(method)

  # DIPSAUS DEBUG START
  # sample_rate <- 500
  # list2env(list(sample_rate = sample_rate,
  #                 high_pass_freq = 1, high_pass_trans_freq = 0.45,
  #                 low_pass_freq = 50, low_pass_trans_freq = 10), envir=.GlobalEnv)
  # list2env(list(filter_order = NA, low_pass_freq = NA, high_pass_freq = 4,
  #               low_pass_trans_freq = NA, high_pass_trans_freq = NA), envir=.GlobalEnv)
  # passband_ripple <- 0.1
  # stopband_attenuation <- 40
  # method <- "butter"


  nyquist <- sample_rate / 2
  passband_ripple <- abs(passband_ripple)
  stopband_attenuation <- abs(stopband_attenuation)

  r_pass <- passband_ripple
  r_stop <- stopband_attenuation

  stopifnot2(
    !is.na(low_pass_freq) || !is.na(high_pass_freq),
    msg = "Please specify low-pass and/or high-pass frequencies"
  )
  stopifnot2(is.finite(passband_ripple) && passband_ripple > 0,
             msg = "`passband_ripple` must be a finite positive number")
  stopifnot2(is.finite(stopband_attenuation) && stopband_attenuation > 0,
             msg = "`stopband_attenuation` must be a finite positive number")

  # Track which transition bandwidths were auto-inferred (NA) vs user-supplied;
  # used in the band-stop branch to decide whether to cap or error on overlap.
  auto_low_trans <- is.na(low_pass_trans_freq)
  auto_high_trans <- is.na(high_pass_trans_freq)

  # suggested bandwidth if bandwidths are NA

  if (is.na(low_pass_trans_freq)) {
    low_pass_trans_freq <- min(max(0.25 * low_pass_freq, 2), low_pass_freq)
  }
  if (is.na(high_pass_trans_freq)) {
    high_pass_trans_freq <- min(max(0.25 * high_pass_freq, 2), sample_rate / 2 - high_pass_freq)
  }

  # determine the filter type
  if ( is.na(high_pass_freq) ) {
    ftype <- "low"
    w_pass <- low_pass_freq / nyquist
    w_trans <- low_pass_trans_freq / nyquist

    w_stop <- clamp(w_pass + abs(w_trans), min = 0, max = 1)
  } else if ( is.na(low_pass_freq) ) {
    ftype <- "high"
    w_pass <- high_pass_freq / nyquist
    w_trans <- high_pass_trans_freq / nyquist

    w_stop <- clamp(w_pass - abs(w_trans), min = 0, max = 1)
  } else if ( isTRUE(high_pass_freq <= low_pass_freq) ) {
    ftype <- "pass"
    w_pass <- c(high_pass_freq, low_pass_freq) / nyquist
    w_trans <- c(high_pass_trans_freq, low_pass_trans_freq) / nyquist

    w_stop <- clamp(w_pass + c(-1, 1) * abs(w_trans), min = 0, max = 1)
  } else {
    ftype <- "stop"
    w_pass <- c(low_pass_freq, high_pass_freq) / nyquist
    auto_trans <- c(auto_low_trans, auto_high_trans)
    w_trans <- c(low_pass_trans_freq, high_pass_trans_freq) / nyquist

    w_stop <- clamp(w_pass + c(1, -1) * abs(w_trans), min = 0, max = 1)

    # Guard: stopband edges must not cross. Silently cap auto-inferred
    # transitions when the notch is narrower than the heuristic bandwidth.
    # Only error when user-supplied values are the cause.
    if (w_stop[[1]] >= w_stop[[2]]) {
      notch_width <- w_pass[[2]] - w_pass[[1]]
      fixed_total <- sum(abs(w_trans[!auto_trans]))
      if (!any(auto_trans) || fixed_total >= notch_width) {
        stopifnot2(FALSE, msg = sprintf(
          paste0(
            "Band-stop transition bandwidths overlap: notch is %.4g-%.4g Hz ",
            "(width %.4g Hz) but the combined specified transition bandwidth ",
            "exceeds it. Set each of `low_pass_trans_freq` and ",
            "`high_pass_trans_freq` to less than %.4g Hz."
          ),
          low_pass_freq, high_pass_freq,
          notch_width * nyquist,
          notch_width * nyquist / 2
        ))
      }
      n_auto <- sum(auto_trans)
      w_trans[auto_trans] <- (notch_width - fixed_total) * (1 - 1e-6) / n_auto
      w_stop <- clamp(w_pass + c(1, -1) * abs(w_trans), min = 0, max = 1)
    }
  }

  w_stop[w_stop < 0] <- 0
  w_stop[w_stop > 1] <- 1

  # get filter specs
  get_specs <- function(fun_ord, fun_make, min_order = 1) {
    spec <- fun_ord(Wp = w_pass, Ws = w_stop, Rp = r_pass, Rs = r_stop)
    # apply user-specified order if provided
    if (!is.na(filter_order) && is.finite(filter_order) && filter_order >= 1) {
      spec$n <- as.integer(filter_order)
    }
    if (!use_sos) {
      # ARMA stability correction
      requested_n <- spec$n
      validator <- function(n) {
        spec$n <- n
        filter <- gsignal::as.Arma(fun_make(spec))
        reciprocal_condition <- rcond_filter_ar(a = filter$a)
        reciprocal_condition > .Machine$double.eps
      }
      max_order <- guess_max_integer(validator, initial = spec$n, min_v = min_order)
      if (is.na(max_order)) {
        max_order <- 1
      }
      if (spec$n > max_order) {
        if (!is.na(filter_order) && is.finite(filter_order) && filter_order >= 1) {
          warning(sprintf(
            paste0("Requested filter_order=%d reduced to %d for numerical stability ",
                   "in ARMA form. Set use_sos=TRUE to avoid order correction."),
            requested_n, max_order
          ))
        }
        spec$n <- max_order
      }
    }
    spec
  }
  switch(
    method,
    "cheby1" = {
      fun_ord <- gsignal::cheb1ord
      fun_make <- gsignal::cheby1
      spec <- get_specs(fun_ord, fun_make)
      filter <- fun_make(spec)
    },
    "cheby2" = {
      fun_ord <- gsignal::cheb2ord
      fun_make <- gsignal::cheby2
      spec <- get_specs(fun_ord, fun_make, min_order = length(w_stop) * 2)
      filter <- fun_make(spec)
    },
    "ellip" = {
      fun_ord <- gsignal::ellipord
      fun_make <- gsignal::ellip
      spec <- get_specs(fun_ord, fun_make, min_order = 1)
      filter <- fun_make(spec)
    },
    {
      # "butter"
      spec <- gsignal::buttord(Wp = w_pass, Ws = w_stop, Rp = r_pass, Rs = r_stop)
      if (!is.na(filter_order) && is.finite(filter_order) && filter_order >= 1) {
        order <- as.integer(filter_order)
        if (!use_sos) {
          max_stable <- butter_max_order(w = w_pass, r = r_pass, type = ftype)$n
          if (order > max_stable) {
            warning(sprintf(
              paste0("Requested filter_order=%d reduced to %d for numerical stability ",
                     "in ARMA form. Set use_sos=TRUE to avoid order correction."),
              order, max_stable
            ))
            order <- max_stable
          }
        }
      } else if (use_sos) {
        order <- spec$n
      } else {
        max_stable <- butter_max_order(w = w_pass, r = r_pass, type = ftype)$n
        order <- min(spec$n, max_stable)
      }
      w_cutoff <- butter_cutoff(type = ftype, n = order, w = w_pass, r = r_pass)
      spec <- gsignal::FilterSpecs(n = order, Wc = w_cutoff, type = ftype)
      filter <- gsignal::butter(spec)
    }
  )

  checks <- check_filter(
    b = filter$b,
    a = filter$a,
    w = c(w_pass, w_stop),
    r_expected = -rep(c(r_pass, r_stop), each = length(w_pass)),
    fs = sample_rate
  )

  params <- list(
    type = c("iir", ftype),
    method = method,
    order = spec$n,
    passband = w_pass,
    stopband = w_stop,
    cutoff = spec$Wc,
    sample_rate = sample_rate
  )


  filter$parameters <- params
  filter$checks <- checks
  filter$sos <- gsignal::as.Sos(filter)


  class(filter) <- c(
    "ravetools-design_filter_iir",
    sprintf("ravetools-design_filter_%s", method),
    "ravetools-design_filter",
    "ravetools-printable",
    class(filter)
  )

  filter
}

#' @title A naive implementation of non-negative matrix factorization
#' @description
#' A pure-R vanilla implementation assuming inputs are non-negative matrices
#' without \code{NA}.
#' @param x a matrix, or can be converted into a matrix; all negative or missing
#' values will be treated as zero
#' @param k decomposition rank
#' @param tol stop criteria, a numeric of two; the first number is the
#' tolerance for root-mean-squared residuals, relative to the largest number in
#' \code{x}; the second number is the tolerance for weight differences; any
#' stopping criteria met will result in the stop of iteration
#' @param max_iters maximum iterations
#' @param verbose whether to report the progress; logical or a positive integer
#' (of step intervals)
#' @returns A list of weights (non-negative template matrix \code{W} and
#' non-negative \code{H}) and errors (root mean squared error of fitted,
#' matrix \code{W}, and \code{W} versus their previous iteration, respectively).
#'
#' @examples
#'
#'
#'
#' x <- stats::toeplitz(.9 ^ (0:31))
#'
#' nmf <- naive_nmf(x, k = 7, verbose = FALSE)
#'
#' fitted <- nmf$W %*% nmf$H
#'
#' oldpar <- par(mfrow = c(1, 2))
#' on.exit({ par(oldpar )})
#'
#' image(x, zlim = c(0, 1), main = "Input")
#' image(fitted, zlim = c(0, 1),
#'       main = sprintf("Fitted with rank=%d", nmf$k))
#'
#'
#'
#'
#' @export
naive_nmf <- function(x, k, tol = c(1e-4, 1e-8), max_iters = 1e4, verbose = TRUE) {
  max_iters <- as.integer(max_iters)
  stopifnot(isTRUE(max_iters >= 0))
  if (max_iters == 0) {
    max_iters <- 1e4
  }

  if (!is.matrix(x)) {
    x <- as.matrix(x)
  }
  dimnames(x) <- NULL
  x[is.na(x) | x < 0] <- 0
  tol1 <- max(x) * tol[[1]]
  tol2 <- tol[[2]]

  dm <- dim(x)
  nr <- dm[[1]]
  nc <- dm[[2]]

  results <- new.env(parent = emptyenv())
  results$k <- k
  results$iters <- 0L
  # nr x k
  results$W <- matrix(stats::runif(nr * k), nrow = nr, ncol = k)

  # k x nc
  results$H <- matrix(stats::runif(nc * k), nrow = k, ncol = nc)

  results$WH <- results$W %*% results$H

  results$error <- c(Inf, Inf, Inf)
  results$converged <- FALSE

  if (isTRUE(verbose)) {
    verbose_interval <- ceiling(max(max_iters / 100, 10))
    verbose <- TRUE
  } else if (!isFALSE(verbose)) {
    verbose_interval <- min(max(as.integer(verbose), 1), max_iters)
    verbose <- TRUE
  }

  eps <- .Machine$double.eps

  # errs <- dipsaus::fastqueue2()

  update <- function() {

    tmp <- crossprod(results$W, results$WH)
    diag(tmp) <- diag(tmp) + eps
    new_H <- results$H * ( crossprod(results$W, x) / tmp )

    new_H <- new_H * 0.1 + results$H * 0.9

    tmp <- tcrossprod( results$W %*% new_H, new_H )
    diag(tmp) <- diag(tmp) + eps
    new_W <- results$W * ( tcrossprod(x, new_H) / tmp )

    new_W <- new_W * 0.1 + results$W * 0.9

    new_WH <- new_W %*% new_H

    err_H <- sqrt(mean((new_H - results$H)^2, na.rm = FALSE))
    err_W <- sqrt(mean((new_W - results$W)^2, na.rm = FALSE))
    err_WH <- sqrt(mean((x - new_WH)^2, na.rm = FALSE))
    new_errors <- c(err_WH, err_W, err_H)

    results$H <- new_H
    results$W <- new_W
    results$WH <- new_WH

    # results$error_diff <- new_errors - results$error
    results$error <- new_errors
    # errs$add(new_errors)

    results$iters <- results$iters + 1L
    return()
  }

  lapply(seq_len(max_iters), function(iter) {
    if ( results$converged ) { return() }
    update()

    new_errors <- results$error

    if ((new_errors[[2]] < tol2 && new_errors[[3]] < tol2) || new_errors[[1]] < tol1) {
      results$converged <- TRUE
    }

    if ( verbose && (iter %% verbose_interval == 0 || iter == max_iters || results$converged) ) {
      cat(sprintf("Iteration %04d: [fitted err=%.2e] [W err=%.2e] [H err=%.2e]        \r",
                  iter, new_errors[[1]], new_errors[[2]], new_errors[[3]]))
    }
    return()

  })
  if (verbose) {
    cat("\n")
  }

  # results$error_trace <- do.call("rbind", errs$as_list())
  as.list(results)

}

Quaternion <- R6::R6Class(
  classname = "Quaternion",
  portable = TRUE,
  cloneable = FALSE,
  lock_objects = FALSE,
  private = list(
    .extern_ptr = NULL
  ),
  active = list(
    is_quaternion = function() { TRUE },
    pointer = function() { private$.extern_ptr },
    x = function( v ) {
      if (!missing(v)) {
        v <- as.double(v)[[1]]
        Quaternion__setX(private$.extern_ptr, v)
      }
      Quaternion__getX(private$.extern_ptr)
    },
    y = function( v ) {
      if (!missing(v)) {
        v <- as.double(v)[[1]]
        Quaternion__setY(private$.extern_ptr, v)
      }
      Quaternion__getY(private$.extern_ptr)
    },
    z = function( v ) {
      if (!missing(v)) {
        v <- as.double(v)[[1]]
        Quaternion__setZ(private$.extern_ptr, v)
      }
      Quaternion__getZ(private$.extern_ptr)
    },
    w = function( v ) {
      if (!missing(v)) {
        v <- as.double(v)[[1]]
        Quaternion__setW(private$.extern_ptr, v)
      }
      Quaternion__getW(private$.extern_ptr)
    }
  ),
  public = list(
    initialize = function() {
      private$.extern_ptr <- Quaternion__new()
    },
    format = function(...) {
      q <- format(self$to_array())
      sprintf(
        "<Quaternion>\nx = %s, y = %s, z = %s, w = %s",
        q[[1]], q[[2]], q[[3]], q[[4]]
      )
    },
    clone2 = function(...) {
      Quaternion$new()$copy(self)
    },
    set = function(x, y, z, w) {
      data <- as.numeric(c(x, y, z, w))[1:4]
      Quaternion__set(private$.extern_ptr, data[[1]], data[[2]], data[[3]], data[[4]])
      self
    },
    to_array = function() {
      Quaternion__to_array(private$.extern_ptr)
    },
    copy = function(m) {
      stopifnot(R6::is.R6(m) && isTRUE(m$is_quaternion))
      Quaternion__copy(private$.extern_ptr, m$pointer)
      self
    },
    normalize = function() {
      Quaternion__normalize(private$.extern_ptr)
      self
    },
    set_from_axis_angle = function(axis, angle) {
      axis <- as_vector3(axis)
      angle <- as.numeric(angle)[[1]]
      Quaternion__set_from_axis_angle(private$.extern_ptr, axis$pointer, angle)
      self
    },
    set_from_rotation_matrix = function(m, fix_qfac = FALSE) {
      m <- as_matrix4(m)
      Quaternion__set_from_rotation_matrix(private$.extern_ptr, m$pointer, as.integer(fix_qfac))
      self
    },
    set_from_unit_vectors = function(v_from, v_to) {
      v_from <- as_vector3(v_from)
      v_to <- as_vector3(v_to)
      Quaternion__set_from_unit_vectors(private$.extern_ptr, v_from$pointer, v_to$pointer)
      self
    },
    angle_to = function(q) {
      q <- as_quaternion(q)
      Quaternion__angle_to(private$.extern_ptr, q$pointer)
    },
    rotate_towards = function(q, step) {
      q <- as_quaternion(q)
      step <- as.double(step)[[1]]
      Quaternion__rotate_towards(private$.extern_ptr, q$pointer, step)
      self
    },
    slerp = function(qb, t) {
      qb <- as_quaternion(qb)
      t <- as.double(t)[[1]]
      Quaternion__slerp(private$.extern_ptr, qb$pointer, t)
      self
    },
    identity = function() {
      Quaternion__identity(private$.extern_ptr)
      self
    },
    invert = function() {
      Quaternion__invert(private$.extern_ptr)
      self
    },
    conjugate = function() {
      Quaternion__conjugate(private$.extern_ptr)
      self
    },
    dot = function(q) {
      q <- as_quaternion(q)
      Quaternion__dot(private$.extern_ptr, q$pointer)
    },
    length_squared = function() {
      Quaternion__length_squared(private$.extern_ptr)
    },
    length = function() {
      Quaternion__length(private$.extern_ptr)
    },
    multiply = function(q) {
      q <- as_quaternion(q)
      Quaternion__multiply_quaternions(private$.extern_ptr, private$.extern_ptr, q$pointer)
      self
    },
    premultiply = function(q) {
      q <- as_quaternion(q)
      Quaternion__multiply_quaternions(private$.extern_ptr, q$pointer, private$.extern_ptr)
      self
    },
    multiply_quaternions = function(qa, qb) {
      qa <- as_quaternion(qa)
      qb <- as_quaternion(qb)
      Quaternion__multiply_quaternions(private$.extern_ptr, qa$pointer, qb$pointer)
      self
    }
  )
)

#' Create a \code{Quaternion} instance to store '3D' rotation
#' @description
#' Create instances that mimic the \code{'three.js'} syntax.
#' @param x,y,z,w numeric of length one
#' @param q R object to be converted to \code{Quaternion}
#' @returns A \code{Quaternion} instance
#' @seealso \code{\link{new_vector3}}, \code{\link{new_matrix4}}
#' @export
new_quaternion <- function(x = 0, y = 0, z = 0, w = 1) {
  Quaternion$new()$set(x, y, z, w)
}

#' @export
`[.Quaternion` <- function(x, i, ..., drop = TRUE) {
  if (missing(i)) {
    x$to_array()
  } else {
    x$to_array()[i]
  }
}


#' @export
`==.Quaternion` <- function(e1, e2) {
  e2 == e1$to_array()
}

#' @rdname new_quaternion
#' @export
as_quaternion <- function(q) {
  if (R6::is.R6(q) && isTRUE(q$is_quaternion)) { return(q) }
  if (length(q) != 4) {
    stop("`as_quaternion`: length of `q` must be a vector of 4 numbers.")
  }
  new_quaternion(q[[1]], q[[2]], q[[3]], q[[4]])
}

#' Find peaks of a signal
#' @param x a numeric vector without missing values
#' @param min_val find peaks that are greater than this value
#' @param min_distance merge peaks that are less than \code{min_distance}
#' time-points away
#' @param min_width search radius (time-points) on whether the peak is
#' "local"; this is for seasonal oscillations.
#' @returns A list of peak index (1-based) and the corresponding value.
#' @examples
#'
#'
#' # Basic example
#' x <- sin(seq(0, 10, 0.01)) + rnorm(1001) * 0.1
#'
#' peaks <- find_peaks(x)
#'
#' plot(x, type = 'l')
#' abline(v = peaks$index, col = 'red')
#'
#' # merge peaks that are close
#' peaks <- find_peaks(x, min_distance = 400)
#'
#' plot(x, type = 'l')
#' abline(v = peaks$index, col = 'red')
#'
#' # with or without min_width
#' x <- c(0, 1, 0.5, 0.9, 0.2, 0.8, 0.2, 0.75, 0)
#'
#' # without min_width
#' peaks <- find_peaks(x, min_width = 0)
#' plot(x, type = 'l')
#' abline(v = peaks$index, col = 'red')
#'
#'
#' # with min_width=2: t=4 is greater than t=6
#' peaks <- find_peaks(x, min_width = 2)
#' plot(x, type = 'l')
#' abline(v = peaks$index, col = 'red')
#'
#'
#' @export
find_peaks <- function(x, min_val = NA, min_distance = 4, min_width = 2) {
  df1 <- c(1, diff(x))
  df2 <- c(1, 1, diff(x, differences = 2))

  # necessary condition to be peaks (1st & 2nd derivatives)
  idx <- (df1 * c(df1[-1], 0)) <= 0 & c(df2[-1], 0) < 0

  # Get peaks that are beyond min_val
  if ( is.na(min_val) ) {
    # automatically decide min_val from standard error
    min_val <- sd(x)
    if (all(x >= 0)) {
      min_val <- min(x) + min_val
    } else {
      min_val <- median(x) + min_val
    }
  }
  idx <- which(idx & (x > min_val))

  ord <- order(x[idx], decreasing = TRUE)
  idx_desc <- idx[ord]

  # merge peaks that are less than min_distance away
  for (ii in seq_along(idx_desc)) {
    elem <- idx_desc[[ii]]
    idx_desc[idx_desc > 0 & abs(idx_desc - elem) < min_distance] <- elem
  }
  idx_desc <- unique(idx_desc)

  # check +- min_width to see if the it is still a peak
  p_idx <- idx_desc + min_width
  p_idx[p_idx > length(x)] <- length(x)
  m_idx <- idx_desc - min_width
  m_idx[m_idx < 1] <- 1
  sign <- (x[p_idx] - x[idx_desc]) * (x[idx_desc] - x[m_idx])
  idx_desc <- idx_desc[sign <= 0]
  list(
    index = idx_desc,
    values = x[idx_desc]
  )
}

#' @title Common Average Re-referencing by Least Anti-Correlation (CARLA)
#' @description
#' Selects an optimal subset of channels to use as the common average reference
#' (CAR) for cortico-cortical evoked potential (\verb{CCEP}) data, following the
#' \verb{CARLA} (see 'Reference' and 'Citation'). Channels are ranked in
#' increasing order of their cross-trial covariance (or variance when only one
#' trial is available); subsets are then iteratively grown and the size that
#' yields the least anti-correlation between the candidate reference and the
#' remaining unreferenced channels is selected as optimal.
#'
#' @param x numeric array of shape \code{time x trials x channels}; if a
#' matrix is provided, it is treated as \code{time x channels} (single trial).
#' A \code{filearray} (class \code{FileArray}) is also accepted, allowing very
#' large inputs to stay on disk instead of in memory. The signal must already
#' be cropped to the responsive time window (typically \code{0.01 s} to
#' \code{0.3 s} post-stimulus).
#' @param nboot integer, number of bootstrapped trial resamplings used to
#' estimate the optimization statistic; defaults to \code{100}. Ignored when
#' a single trial is supplied.
#' @param sensitive logical; if \code{TRUE} (and more than one trial is
#' supplied), the more sensitive cutoff is used, corresponding to the channel
#' count just before the first statistically significant decrease in the mean
#' anti-correlation curve. The default \code{FALSE} returns the global maximum
#' of that curve.
#' @param min_size integer, minimum subset size considered when
#' \code{sensitive = TRUE}; defaults to
#' \code{max(2, ceiling(0.1 * nchan_good))}, where \code{nchan_good} is
#' the number of usable channels after the bad-channel mask has been
#' applied (see Details). Floored at 2 because subset size 1 is never
#' evaluated.
#' @param absolute_rank logical; if \code{FALSE} (the default), the
#' per-channel ranking statistic is the signed mean of the off-diagonal
#' trial-to-trial covariances, as in the original CARLA manuscript. This
#' assumes a phase-locked evoked response, so trial-pair covariances on
#' responsive channels are positive and a clean signed mean separates them
#' from non-responsive channels. Set to \code{TRUE} to use the mean of the
#' \strong{absolute} covariances instead, which is more robust to evoked
#' responses whose polarity flips across trials (e.g. alternating-polarity
#' stimulation, biphasic \verb{CCEPs} with jitter), at the cost of an upward
#' bias on the non-responsive floor. Ignored when only one trial is supplied
#' (the per-channel variance is used in that case).
#' @param virtual_reference logical; if \code{TRUE}, runs the modified
#' CARLA before the iterative subset
#' evaluation: rank channels once, designate the channel with the median
#' rank as a "virtual reference", subtract its signal from every channel,
#' then re-rank on the subtracted data. The virtual channel itself is
#' pinned to the very end of the new ordering so it is never picked into
#' the CAR until the final subset size. This is intended for data where
#' the recording reference is contaminated by stimulation artifact or
#' evoked activity: a contaminated reference makes every channel look
#' artificially similar and biases the covariance/variance ranking statistic,
#' but subtracting a mid-rank proxy of that contamination unbiases the
#' ranking so genuinely responsive channels rise to the top. Default
#' \code{FALSE} reproduces the original CARLA algorithm bit-for-bit.
#'
#' @returns A list with the following elements:
#' \describe{
#' \item{\code{channels}}{integer vector, sorted indices (1-based) of the
#' channels chosen to construct the common average reference. Compute the
#' reference yourself as the channel-wise mean (or median) over these channels
#' and subtract it from the original signal to obtain the re-referenced data.}
#' \item{\code{order}}{integer vector, indices of the \strong{good}
#' channels sorted in increasing order of the ranking statistic. Bad
#' channels (zero variance / all-\code{NA}; see Details) are excluded.}
#' \item{\code{vars}}{numeric vector of length \code{nchan} containing the
#' per-channel ranking statistic (mean cross-trial covariance, or variance
#' for a single trial). Bad channels keep their raw value (zero or
#' \code{NA}) so callers can audit the mask.}
#' \item{\code{n_optimum}}{integer, the optimal subset size selected
#' (indexes into \code{order}).}
#' \item{\code{zmin_mean}}{numeric matrix of shape
#' \code{length(order) x nboot} (or a length-\code{length(order)} vector
#' for a single trial / \code{nboot = 1}) holding, for each subset size
#' and bootstrap, the mean Fisher z-transformed correlation of the most
#' globally anti-correlated unreferenced channel against the candidate CAR.
#' Row 1 is always \code{NA} (subset size 1 is not evaluated).}
#' \item{\code{bad_channels}}{integer vector of channel indices that were
#' excluded from the analysis because their ranking statistic was zero or
#' \code{NA} (flat / dead channels, plus the virtual channel itself when
#' \code{virtual_reference = TRUE}).}
#' \item{\code{virtual_channel}}{integer, the index (1-based) of the
#' channel used as the virtual reference when \code{virtual_reference =
#' TRUE}; \code{NA_integer_} otherwise.}
#' \item{\code{vars1}}{numeric vector of the first-pass ranking statistic
#' when \code{virtual_reference = TRUE} (the post-subtraction statistic is
#' returned in \code{vars}); \code{NULL} otherwise.}
#' }
#'
#' @details
#' The function is a faithful port of the core \code{CARLA.m} routine; it does
#' not perform notch filtering, time-window cropping,
#' or grouping by stimulation site. Those steps belong to the surrounding
#' preprocess pipeline (see the example below).
#'
#' For each candidate subset size \eqn{n = 2, \ldots, N}{n = 2, ..., N}, the
#' candidate reference is computed as the channel-wise mean of the \eqn{n}
#' lowest-ranked channels. Each of those \eqn{n} channels is then correlated,
#' in its unreferenced form, against every channel of the candidate
#' re-referenced subset. The resulting Pearson correlations are
#' Fisher z-transformed; the row corresponding to the most globally
#' anti-correlated channel is recorded as \code{zmin}. The optimal \eqn{n}
#' is the one that maximizes (i.e. makes least negative) the mean of
#' \code{zmin}.
#'
#' \strong{Bad-channel mask.} Channels whose ranking statistic is exactly
#' zero or \code{NA} are flagged as "bad" and excluded both from the CAR
#' candidate pool and from the target-channel correlation rows. This
#' catches flat / dead channels (constant signal, no usable variance) and
#', when \code{virtual_reference = TRUE}, automatically removes the virtual
#' channel itself, since subtracting it from itself yields a zero trace.
#' All references to channel indices in the returned \code{order},
#' \code{n_optimum}, and \code{zmin_mean} are with respect to the
#' \strong{good} channels only; \code{vars} is full-length so callers can
#' inspect the raw scores.
#'
#' @references
#' The CARLA algorithm (\code{virtual_reference = FALSE}) is described in
#' \doi{10.1016/j.jneumeth.2024.110153}; the modified CARLA precursor
#' (\code{virtual_reference = TRUE}) is described in
#' \doi{10.1016/j.jneumeth.2025.110461}, with a reference implementation
#' at \url{https://github.com/hharveygit/SPES_reference_contam}. See
#' \code{citation("ravetools")} for the full bibliographic entries of
#' both manuscripts.
#'
#' @examples
#'
#' # ---- Simulate a small CCEP-like dataset --------------------------------
#' # 16 channels, 12 trials, sampled at 1 kHz, 0.5 s peri-stimulus epoch.
#' # Channels 1:4 are "responsive" (carry an evoked potential); the rest are
#' # noise-only and should make up the optimal CAR.
#' srate <- 1000
#' tt <- seq(-0.1, 0.4 - 1 / srate, by = 1 / srate)   # time, seconds
#' nchan <- 16
#' ntrial <- 30
#' resp_ch <- 1:4
#' noise_ch <- 4:7
#'
#' # Evoked potential template: damped sinusoid starting at t = 0
#' ep <- ifelse(tt >= 0,
#'              80 * exp(-tt / 0.05) * sin(2 * pi * 12 * tt),
#'              0)
#'
#' # time x trials x channels
#' x_full <- array(rnorm(length(tt) * ntrial * nchan, sd = 5),
#'                 dim = c(length(tt), ntrial, nchan))
#' for (ch in resp_ch) {
#'   for (k in seq_len(ntrial)) {
#'     x_full[, k, ch] <- x_full[, k, ch] + ep * runif(1, -0.8, 1.2)
#'   }
#' }
#' for (ch in noise_ch) {
#'   for (k in seq_len(ntrial)) {
#'     tmp <- x_full[, k, ch]
#'     x_full[, k, ch] <- tmp + sign(ch %% 2 - 0.5) * 5 *
#'       runif(length(tmp), 0.8, 1.2)
#'   }
#' }
#' # Add artifacts common to all channels and trials
#' artifacts <- 6 * sin(2 * pi * 60 * tt) + 7 * sin(2 * pi * 24 * tt)
#' x_full <- sweep(x_full, 1L, artifacts, "+")
#'
#' # ---- 1. Notch filter line noise (per channel, per trial) ---------------
#' # The CARLA paper notch-filters before ranking; the re-reference itself
#' # is applied to the original (unfiltered) signal.
#' x_clean <- x_full
#' for (ch in seq_len(nchan)) {
#'   for (k in seq_len(ntrial)) {
#'     x_clean[, k, ch] <- notch_filter(
#'       x_full[, k, ch], sample_rate = srate,
#'       lb = c(59, 119, 179), ub = c(61, 121, 181)
#'     )
#'   }
#' }
#'
#' # ---- 2. Crop to the responsive window (0.01 s to 0.3 s post-stim) ------
#' resp_idx <- which(tt > 0.0 & tt <= 0.3)
#' x_resp <- x_clean[resp_idx, , , drop = FALSE]
#'
#' # ---- 3. Run CARLA to pick reference channels ---------------------------
#' fit <- carla(x_resp, sensitive = TRUE, absolute_rank = TRUE,
#'              virtual_reference = TRUE)
#' fit$channels        # selected reference channels (should exclude 1:4)
#' fit$n_optimum       # number of channels in the optimal CAR
#'
#' # ---- 4. Re-reference the ORIGINAL (unfiltered) signal ------------------
#' # mean or median, your choice! (time x trials)
#' car_full <- apply(x_full[, , fit$channels, drop = FALSE], c(1, 2), mean)
#'
#' # old-style: using all channels for CAR
#' car_old <- apply(x_full, c(1, 2), mean)
#'
#' x_reref <- sweep(x_full, c(1, 2), car_full, "-")
#' x_compare <- sweep(x_full, c(1, 2), car_old, "-")
#'
#' # ---- 5. Inspect: evoked potential is preserved on responsive channels --
#' # `plot_signals` expects channels x time, so transpose each trial-1 slice.
#' op <- graphics::par(mfrow = c(2, 4), mar = c(4, 4, 2, 1))
#' ravetools::plot_signals(
#'   signals = t(x_full[, 1, ]),
#'   sample_rate = srate,
#'   main = "Trial 1 - (raw)")
#'
#' ravetools::plot_signals(
#'   signals = t(x_clean[, 1, ]),
#'   sample_rate = srate,
#'   main = "Notch-filtered")
#'
#' ravetools::plot_signals(
#'   signals = t(x_reref[, 1, ]),
#'   sample_rate = srate,
#'   main = sprintf("CARLA-ref (n=%d)", length(fit$channels)))
#'
#' ravetools::plot_signals(
#'   signals = t(x_compare[, 1, ]),
#'   sample_rate = srate,
#'   main = "Conventional CAR for comparison")
#'
#' col <- adjustcolor(seq_len(nchan))
#' col[resp_ch] <- adjustcolor(col[resp_ch], alpha.f = 0.2)
#'
#' # Trial-average -> time x channels, ready for `matplot(tt, .)`
#' graphics::matplot(tt, apply(x_full, c(1, 3), mean),
#'                   type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
#'                   main = "Trial-averaged (raw)", col = col)
#'
#' graphics::matplot(tt, apply(x_clean, c(1, 3), mean),
#'                   type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
#'                   main = "Notch-filtered", col = col)
#'
#' graphics::matplot(tt, apply(x_reref, c(1, 3), mean),
#'                   type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
#'                   main = "CARLA-referenced", col = col)
#'
#' graphics::matplot(tt, apply(x_compare, c(1, 3), mean),
#'                   type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
#'                   main = "conventional CAR-referenced", col = col)
#'
#' graphics::par(op)
#'
#' @export
carla <- function(x, nboot = 100L, sensitive = FALSE, min_size = NULL,
                  absolute_rank = FALSE, virtual_reference = FALSE) {

  # DIPSAUS DEBUG START
  # dm <- c(136, 2001,  105)
  # x <- array(rnorm(prod(dm)), dm)
  # list2env(list(nboot = 100L, sensitive = TRUE, min_size = NULL,
  #               absolute_rank = TRUE, virtual_reference = TRUE), .GlobalEnv)

  # DIPSAUS DEBUG START
  # repository <- ravecore::prepare_subject_raw_voltage_with_epochs(
  # subject = "demo@bids:ds005953/01",
  # electrodes = 1:118,
  # epoch_name = "default",
  #   time_windows = c(0.1, 1)
  # )
  # container <- repository$get_container()
  # epoch_voltage <- lapply(container$data_list, function(x) {
  #   x[dimnames = FALSE, drop = FALSE]
  # })
  # epoch_voltage <- simplify2array(epoch_voltage)
  # # Time x trial x channel
  # dim(epoch_voltage) <- dim(epoch_voltage)[c(1, 2, 4)]
  # x <- filearray::as_filearray(epoch_voltage)
  # nboot = 100L;
  # absolute_rank = TRUE
  # virtual_reference = TRUE


  stopifnot(is.numeric(x) || inherits(x, "FileArray"))

  # Time x trial x channel
  d <- dim(x)
  if (length(d) == 2) {
    d <- c(d[[1]], 1L, d[[2]])
    if (inherits(x, "FileArray")) {
      x <- x[drop = FALSE, dimnames = NULL]
    }
    dim(x) <- d
  }
  if (length(d) != 3L) {
    stop("`x` must be a numeric matrix (time x channels) or 3-D array ",
         "(time x trials x channels).")
  }
  n_t  <- d[[1L]] # Number of time points
  n_tr <- d[[2L]] # Number of trials
  n_ch <- d[[3L]] # Number of channels
  if (n_ch < 2L) {
    stop("`x` must have at least 2 channels.")
  }
  if (n_t < 2L) {
    stop("`x` must have at least 2 time points in the response window.")
  }

  nboot <- as.integer(nboot)
  if (is.na(nboot) || nboot < 1L) {
    stop("`nboot` must be a positive integer.")
  }
  if (n_tr == 1L) {
    nboot_eff <- 1L
  } else {
    nboot_eff <- nboot
  }

  # ---- 1. Per-channel ranking statistic --------------------------------------
  # In the original manuscript there is no `abs()`: the assumption is a
  # phase-locked evoked response, so trial-pair covariances are positive
  # and their signed mean is a clean score. That signed-mean form is the
  # default here (`absolute_rank = FALSE`). Setting `absolute_rank = TRUE`
  # takes the magnitude instead, which is robust to polarity-flipping
  # evoked responses (e.g. alternating-polarity stimulation, biphasic
  # CCEPs) where signed covariances can cancel.
  cov_summary <- if (isTRUE(absolute_rank)) {
    function(v) {
      mean(abs(v), na.rm = TRUE)
    }
  } else {
    function(v) {
      mean(v, na.rm = TRUE)
    }
  }
  if (n_tr == 1L) {
    rank_fn <- function(signal) {
      if (inherits(signal, "FileArray")) {
        filearray::fmap2(
          list(signal),
          .buffer_count = n_ch,
          .simplify = TRUE,
          fun = function(input) {
            stats::var(as.numeric(input[[1]]))
          }
        )
      } else {
        apply(signal, 3L, function(row) {
          stats::var(as.numeric(row))
        })
      }
    }
  } else {
    rank_fn <- function(signal) {

      if (inherits(signal, "FileArray")) {
        # 11.948s
        filearray::fmap2(
          list(signal),
          .buffer_count = n_ch,
          .simplify = TRUE,
          fun = function(input) {
            slice <- array(input[[1]], c(n_t, n_tr))
            cv <- fast_cov(slice)
            cov_summary(cv[upper.tri(cv)])
          }
        )
      } else {

        # system.time({
          vapply(seq_len(n_ch), function(ii) {
            slice <- matrix(signal[, , ii], nrow = n_t, ncol = n_tr) # time x trials
            cv <- fast_cov(slice)                                   # trials x trials
            cov_summary(cv[upper.tri(cv)])
          }, numeric(1L))
        # })
      }

    }
  }

  vars <- rank_fn(x)

  # Signal model (denote i: channel, j: trial, t: time):
  #   signal_ij(t) =
  #       r_i(t)        <- evoked response, shared across trials of channel i
  #                        (zero or near-zero on non-responsive channels)
  #     + a(t)          <- global artifact / common-mode signal we want to
  #                        remove via re-referencing (line noise, stim leak,
  #                        ground/ref drift); shared across channels
  #     + e_ij(t)       <- zero-mean noise, independent across trials and
  #                        (approximately) across channels
  #
  # The ranking statistic `vars[i]` estimates the across-trial covariance of
  # channel i on the response window. Taking the expectation:
  #   Cov(signal_i,j1, signal_i,j2)  ~=  Var(r_i) + Var(a)
  # because r_i and a are constant across trials while e_ij averages out.
  # Therefore:
  #   - Non-responsive channels  -> vars[i] ~= Var(a)            (small, common floor)
  #   - Responsive channels      -> vars[i] ~= Var(r_i) + Var(a) (larger)
  # Sorting `vars` in increasing order puts the least-responsive channels
  # first; growing the CAR subset from the top of `ord` keeps the evoked
  # response out of the reference for as long as possible.

  # ---- 1b. Optional virtual-reference pre-step (modified CARLA) -------------
  # Huang et al. (2025): when the recording reference is contaminated by
  # stim/evoked activity, every channel inherits a copy of the
  # contamination, which makes the ranking statistic above biased --
  # responsive and non-responsive channels look more similar than they are.
  # The fix is to pick the channel whose first-pass rank is in the middle
  # (so it is unlikely to be either pure noise or a strongly responsive
  # channel, but should still carry the contamination), subtract it from
  # every channel, and re-rank on the subtracted signal. The virtual
  # channel itself becomes identically zero after subtraction, so we
  # remove it from the new order and append it at the very end: it is
  # never picked into the CAR until the final subset size and never
  # influences any of the correlations downstream.
  if (virtual_reference) {
    vars1 <- vars
    # MATLAB `round(nChs/2)` rounds halves AWAY from zero; R's base
    # `round` is banker's (half-to-even). Use `floor(n/2 + 0.5)` to match
    # the reference implementation exactly.
    virtual_channel <- order(vars1)[as.integer(floor(n_ch / 2 + 0.5))]
    # Force n_t x n_tr matrix even when n_tr == 1 (drop=TRUE would
    # collapse to a length-n_t vector and break sweep over c(2, 3)).
    vref <- matrix(x[, , virtual_channel], nrow = n_t, ncol = n_tr)
    dimnames(vref) <- NULL

    if (inherits(x, "FileArray")) {
      x_work <- filearray::fmap(
        x = x,
        .buffer_count = n_ch,
        fun = function(slice) {
          slice[[1]] - vref
        }
      )
      on.exit({
        x_work$delete(force = TRUE)
      }, add = TRUE)
    } else {
      x_work <- sweep(x, c(1L, 2L), vref, "-")
    }
    # The virtual channel is now identically zero in `x_work`, so the
    # bad-channel mask below will (correctly) drop it from `ord`. There is
    # no need for the explicit `c(ord[ord != vc], vc)` reordering trick.
    vars <- rank_fn(x_work)
  } else {
    vars1 <- NULL
    virtual_channel <- NA_integer_
    x_work <- x
  }

  # ---- 1c. Bad-channel mask --------------------------------------------------
  # Flat / dead channels (zero variance) and channels that produced an NA
  # ranking statistic (e.g. all-NA input rows) cannot meaningfully appear in
  # the CAR - their correlation rows would be NaN - and including them
  # introduces noise into the bootstrap statistic. Drop them from `ord` and
  # shrink the iterative loop accordingly.
  bad_mask <- is.na(vars) | vars == 0
  bad_channels <- which(bad_mask)
  ord <- order(vars)
  ord <- ord[!bad_mask[ord]]
  n_good <- length(ord)
  if (n_good < 2L) {
    stop("`carla` needs at least 2 good channels (found ", n_good,
         " after dropping flat / NA channels).")
  }

  # ---- 2. Iterative subset evaluation ----------------------------------------
  # For each candidate subset size ii = 2..n_good the candidate CAR is the mean
  # of the ii lowest-ranked channels; every one of those channels is then
  # correlated (over time) in its unreferenced form against every re-referenced
  # channel, Fisher z-transformed, and the mean z of the most globally anti-
  # correlated channel is recorded. The anti-correlation score is the CARLA
  # diagnostic: as ii grows, a newly added responsive channel leaks its evoked
  # response into the CAR, which flips part of that response into the residual
  # of the other subset channels with the opposite sign, driving the score
  # negative. Tracking when that happens is what `n_optimum` picks up on.
  #
  # All subset sizes are evaluated in C++ (`carla_zmin_all`, see src/carla.cpp)
  # against ONE shared set of bootstrap trial-resamples: the per-channel trial-
  # average and the channel x channel Gram matrix are computed once per boot-
  # strap and each subset size is read out of the Gram in O(ii^2) work, instead
  # of re-trial-averaging every subset size. The resampling indices are drawn
  # here in R (so the RNG stays reproducible and thread-safe) and handed to
  # C++, which parallelizes over bootstraps via TinyParallel
  # (`ravetools_threads()`).
  #
  # When there is no bootstrap (a single trial, or `nboot = 1`) the "resample"
  # is the uniform mean over all trials -- `ind = seq_len(n_tr)` -- so the same
  # code path covers the single-pass case.
  if (nboot_eff == 1L) {
    ind <- matrix(seq_len(n_tr), nrow = n_tr, ncol = 1L)
  } else {
    # Column-major fill of a single `sample.int` draw = independent per-boot
    # sampling of `n_tr` trials with replacement.
    ind <- matrix(sample.int(n_tr, n_tr * nboot_eff, replace = TRUE),
                  nrow = n_tr, ncol = nboot_eff)
  }

  # Materialise the good channels in ranking order so subset size ii is exactly
  # the first ii columns (this also pulls a `filearray` into memory once).
  x_ord <- x_work[, , ord, drop = FALSE]           # n_t x n_tr x n_good

  # `carla_zmin_all` returns an n_good x nboot_eff matrix already in the (ii,
  # boot) layout used downstream; row 1 is all-NA (subset size 1 is never
  # evaluated and never selected by `which.max`).
  zmin_mean <- carla_zmin_all(x_ord, ind)

  # ---- 3. Pick the optimum subset size ---------------------------------------
  if (sensitive && nboot_eff > 1L) {
    if (is.null(min_size)) {
      # MATLAB: `nMin = ceil(0.1 * nChs)`. We follow that on the GOOD-channel
      # count, but floor at 2 because subset size 1 is never evaluated (row
      # 1 of `zmin_mean` is structurally NA), so a 1-floor would short-
      # circuit the sensitive search to `n_optimum = 1`.
      n_min <- max(2L, as.integer(ceiling(0.1 * n_good)))
    } else {
      n_min <- max(2L, as.integer(min_size))
    }
    zmm_x_trs <- rowMeans(zmin_mean, na.rm = TRUE)              # length n_good
    ii <- n_min
    n_optimum <- n_good
    while (ii <= n_good) {
      if (ii == n_good) {
        n_optimum <- ii
        break
      }
      if (!is.na(zmm_x_trs[ii + 1L]) &&
          zmm_x_trs[ii + 1L] > zmm_x_trs[ii]) {
        ii <- ii + 1L
        next
      }
      if (ii > 1L) {
        zmm_x_trs[seq_len(ii - 1L)] <- NA_real_
      }
      next_greater <- which(!is.na(zmm_x_trs) & zmm_x_trs > zmm_x_trs[ii])[1L]
      if (is.na(next_greater)) {
        n_optimum <- ii
        break
      }
      zmm_curr   <- zmin_mean[ii, ]
      idx        <- which.min(zmm_x_trs[seq_len(next_greater - 1L)])
      zmm_trough <- zmin_mean[idx, ]
      diffs <- as.numeric(outer(zmm_trough, zmm_curr, "-"))
      # left-tailed 95% CI on the bootstrap difference (trough - peak)
      conf_upper <- stats::quantile(diffs, 0.95, names = FALSE, type = 7L,
                                    na.rm = TRUE)
      if (!is.na(conf_upper) && conf_upper < 0) {
        n_optimum <- ii
        break
      }
      ii <- next_greater
    }
    if (n_optimum == n_min) {
      warning("CARLA optimum was detected at the minimum subset size floor.")
    }
  } else {
    if (n_tr == 1L) {
      zmm <- as.numeric(zmin_mean)
    } else {
      zmm <- rowMeans(zmin_mean, na.rm = TRUE)
    }
    n_optimum <- which.max(zmm)
  }

  channels <- sort(ord[seq_len(n_optimum)])
  # car <- apply(x[channels, , , drop = FALSE], c(2L, 3L), agg_fun)
  if (n_tr == 1L) {
    # car <- as.numeric(car)
    zmin_mean <- as.numeric(zmin_mean)
  }

  list(
    channels        = channels,
    # car             = car,
    order           = ord,
    vars            = vars,
    n_optimum       = as.integer(n_optimum),
    zmin_mean       = zmin_mean,
    bad_channels    = as.integer(bad_channels),
    virtual_channel = as.integer(virtual_channel),
    vars1           = vars1
  )
}

#' @name band_pass
#' @title Band-pass signals
#' @param x input signals, numeric vector or matrix. \code{x} must be
#' row-major if input is a matrix: each row is a channel, and each column is
#' a time-point.
#' @param sample_rate sampling frequency
#' @param lb lower frequency bound of the band-passing filter, must be positive
#' @param ub upper frequency bound of the band-passing filter, must be greater
#' than the lower bound and smaller than the half of sampling frequency
#' @param domain 1 if \code{x} is in time-domain, or 0 if \code{x} is in
#' frequency domain
#' @param method filter type, choices are \code{'fir'} and \code{'butter'}
#' @param direction filter direction, choices are \code{'forward'},
#' \code{'backward'}, and \code{'both'} directions
#' @param window window type, can be a character, a function, or a vector.
#' For character, \code{window} is a function name in the
#' \code{signal} package, for example, \code{'hanning'}; for a function,
#' \code{window} takes one integer argument and returns a numeric vector
#' with length of that input; for vectors, \code{window} is a numeric vector
#' o length \code{order+1}.
#' @param order the order of the filter, must be positive integer and be
#' less than one-third of the sample rate
#' @param ... ignored
#' @returns Filtered signals, vector if \code{x} is a vector, or matrix of
#' the same dimension as \code{x}
#' @examples
#'
#'
#' t <- seq(0, 1, by = 0.0005)
#' x <- sin(t * 0.4 * pi) + sin(t * 4 * pi) + 2 * sin(t * 120 * pi)
#'
#' oldpar <- par(mfrow = c(2, 2), mar = c(3.1, 2.1, 3.1, 0.1))
#' # ---- Using band_pass1 ------------------------------------------------
#'
#' y1 <- band_pass1(x, 2000, 0.1, 1)
#' y2 <- band_pass1(x, 2000, 1, 5)
#' y3 <- band_pass1(x, 2000, 10, 80)
#'
#' plot(t, x, type = 'l', xlab = "Time", ylab = "",
#'      main = "Mixture of 0.2, 2, and 60Hz")
#' lines(t, y1, col = 'red')
#' lines(t, y2, col = 'blue')
#' lines(t, y3, col = 'green')
#' legend(
#'   "topleft", c("Input", "Pass: 0.1-1Hz", "Pass 1-5Hz", "Pass 10-80Hz"),
#'   col = c(par("fg"), "red", "blue", "green"), lty = 1,
#'   cex = 0.6
#' )
#'
#' # plot pwelch
#' pwelch(x, fs = 2000, window = 4000, noverlap = 2000, plot = 1)
#' pwelch(y1, fs = 2000, window = 4000, noverlap = 2000,
#'        plot = 2, col = "red")
#' pwelch(y2, fs = 2000, window = 4000, noverlap = 2000,
#'        plot = 2, col = "blue")
#' pwelch(y3, fs = 2000, window = 4000, noverlap = 2000,
#'        plot = 2, col = "green")
#'
#'
#' # ---- Using band_pass2 with FIR filters --------------------------------
#'
#' order <- floor(2000 / 3)
#' z1 <- band_pass2(x, 2000, 0.1, 1, method = "fir", order = order)
#' z2 <- band_pass2(x, 2000, 1, 5, method = "fir", order = order)
#' z3 <- band_pass2(x, 2000, 10, 80, method = "fir", order = order)
#'
#' plot(t, x, type = 'l', xlab = "Time", ylab = "",
#'      main = "Mixture of 0.2, 2, and 60Hz")
#' lines(t, z1, col = 'red')
#' lines(t, z2, col = 'blue')
#' lines(t, z3, col = 'green')
#' legend(
#'   "topleft", c("Input", "Pass: 0.1-1Hz", "Pass 1-5Hz", "Pass 10-80Hz"),
#'   col = c(par("fg"), "red", "blue", "green"), lty = 1,
#'   cex = 0.6
#' )
#'
#' # plot pwelch
#' pwelch(x, fs = 2000, window = 4000, noverlap = 2000, plot = 1)
#' pwelch(z1, fs = 2000, window = 4000, noverlap = 2000,
#'        plot = 2, col = "red")
#' pwelch(z2, fs = 2000, window = 4000, noverlap = 2000,
#'        plot = 2, col = "blue")
#' pwelch(z3, fs = 2000, window = 4000, noverlap = 2000,
#'        plot = 2, col = "green")
#'
#' # ---- Clean this demo --------------------------------------------------
#' par(oldpar)
#'
#' @export
band_pass1 <- function(x, sample_rate, lb, ub, domain = 1, ...) {

  max_freq <- sample_rate / 2

  if (lb >= ub) {
    stop("`band_pass`: Band-passing filtering frequency must have lower bound < upper bound")
  }
  if ( lb < 0 ) {
    stop("`band_pass`: Band-passing frequency lower bound must be positive")
  }
  if ( max_freq < ub ) {
    stop("`band_pass`: Band-passing frequency higher bound must not exceed half of sampling frequency")
  }

  # check if x is a matrix
  if (!is.matrix(x)) {
    x_is_vec <- TRUE
    x <- matrix(x, nrow = 1)
  } else {
    x_is_vec <- FALSE
  }
  n_tp <- ncol(x)


  df <- sample_rate / n_tp

  center_freq <- (lb + ub) / 2
  filter_width <- ub - lb

  y <- seq(0, max_freq, by = df)
  gauss <- exp( - (y - center_freq)^2 )
  cnt_gauss <- round( center_freq / df )
  flat_padd <- round( filter_width / df )

  padd_left <- floor( flat_padd / 2 )
  padd_right <- ceiling( flat_padd / 2 )

  # if( padd_left >= cnt_gauss ) {
  #   padd_left <- cnt_gauss - 1
  #   padd_right <- flat_padd - cnt_gauss + 1
  # }
  # if( length(gauss) - padd_right < cnt_gauss + 1) {
  #   padd_right <- length(gauss) - cnt_gauss - 1
  #   padd_left <- cnt_gauss + flat_padd + 1 - length(gauss)
  # }

  our_wind <- c(
    gauss[ seq(padd_left + 1, cnt_gauss) ],
    rep( 1.0, flat_padd ),
    gauss[ seq(cnt_gauss + 1, length(gauss) - padd_right) ]
  )

  if ( n_tp %% 2 == 0 ) {
    our_wind <- our_wind[-length(our_wind)]
    our_wind <- c(our_wind, rev(our_wind))
  } else {
    our_wind <- c(our_wind[-length(our_wind)], rev(our_wind))
  }

  re <- apply(x, 1, function(slice) {
    if (domain > 0) {
      slice <- fft(slice)
    }
    slice <- slice * our_wind
    ifft(slice)
  })
  if (x_is_vec) {
    re <- drop(re)
  } else {
    re <- t(re)
  }
  re
}

#' @rdname band_pass
#' @export
band_pass2 <- function(x, sample_rate, lb, ub, order,
                       method = c("fir", "butter"),
                       direction = c("both", "forward", "backward"),
                       window = "hamming",
                       ...) {

  method <- match.arg(method)
  direction <- match.arg(direction)

  if ( missing(order) ) {
    if (method == "fir") {
      order <- 25
    } else {
      order <- 4
    }
  } else {
    if ( length(order) != 1 || !is.numeric(order) || is.na(order) || order != as.integer(order) || order < 0 ) {
      stop("`band_pass2`: filter order must be an positive integer")
    }
  }

  if (is.character(window)) {
    gsignal <- asNamespace("gsignal")
    window <- gsignal[[window]]
  }


  fn <- sample_rate / 2
  if (lb >= ub) {
    stop("`band_pass`: Band-passing filtering frequency must have lower bound < upper bound")
  }
  if ( lb < 0 ) {
    stop("`band_pass`: Band-passing frequency lower bound must be positive")
  }
  if ( fn < ub ) {
    stop("`band_pass`: Band-passing frequency higher bound must not exceed half of sampling frequency")
  }

  w <- c(lb, ub) / fn
  if (method == "fir") {
    b <- fir1(order, w = w, type = "pass", window = window)$b
    a <- 1
  } else {
    if (is.function(window)) {
      window <- window(order + 1)
    }
    bf <- gsignal::butter(order, w, type = "pass", window = window)
    b <- bf$b
    a <- bf$a
  }

  # check if x is a matrix
  if (!is.matrix(x)) {
    x_is_vec <- TRUE
    x <- matrix(x, nrow = 1)
  } else {
    x_is_vec <- FALSE
  }

  re <- apply(x, 1, function(slice) {
    switch(
      direction,
      "forward" = {
        re <- filter_signal(b = b, a = a, x = slice)
        re[[1]]
      },
      "backward" = {
        re <- filter_signal(b = b, a = a, x = rev(slice))
        rev(re[[1]])
      },
      {
        filtfilt(b = b, a = a, x = slice)
      }
    )
  })

  if ( x_is_vec ) {
    re <- drop(re)
  } else {
    re <- t(re)
  }

  re
}

#' @title Basis Profile Curve (\verb{BPC}) identification
#' @description
#' Identifies a small set of canonical temporal response shapes - \emph{basis
#' profile curves} (\verb{BPCs}) - from the single-trial stimulation-evoked
#' responses recorded at one measurement electrode, where the trials are grouped
#' by stimulation site (or any other condition). Each stimulation group is
#' assigned to the \verb{BPC} that best explains its trials, and the projection
#' strength of every group is quantified. This is the across-stimulation-site
#' \verb{BPC} method (see \sQuote{References}); the related
#' \code{\link{crp_cluster}} applies the same idea across electrodes.
#'
#' @details
#' Let \eqn{V} be the (windowed) \eqn{T \times K} matrix of all single trials and
#' let \code{groups} map each of the \eqn{K} columns to one of \eqn{n} stimulation
#' subgroups.
#' \enumerate{
#' \item \strong{Window.} When \code{time} and \code{time_window} are supplied the
#'   rows of \code{x} are cropped to that window; otherwise all rows are used.
#'   Rows with non-finite values are dropped.
#' \item \strong{Internal projections.} Each trial is \eqn{L_2}-normalized into
#'   \eqn{V_0}, and \eqn{P = V_0^\top V} collects the projection of every native
#'   trial onto every normalized trial.
#' \item \strong{Significance matrix.} For each ordered pair of subgroups
#'   \eqn{(k, l)} the set of relevant entries of \eqn{P} is gathered (the
#'   off-diagonal of the within-group block when \eqn{k = l}, the whole cross
#'   block otherwise) and reduced to a one-sample t-statistic versus zero. These
#'   form the \eqn{n \times n} significance matrix \eqn{\Xi}.
#' \item \strong{Rank selection.} \eqn{\Xi} is made non-negative and rescaled, then
#'   factorized with \code{\link{naive_nmf}} at decreasing inner rank while the
#'   degeneracy score \eqn{\zeta} - the sum of the upper off-diagonal of the
#'   row-normalized \eqn{HH^\top} - exceeds \code{zeta_threshold}.
#' \item \strong{Assignment.} Each subgroup takes its winner-take-all \verb{BPC}
#'   over the normalized \verb{NMF} loadings; with \code{null_class}, loadings
#'   below \eqn{1/(2\sqrt{n})} are left unassigned (the \dQuote{null} class).
#' \item \strong{Basis curves.} Per \verb{BPC}, the first linear-kernel \verb{PCA}
#'   component of all trials in its member subgroups, sign-oriented to a positive
#'   mean projection.
#' \item \strong{Weights.} For each member subgroup the per-trial coefficient
#'   \eqn{\alpha} (projection onto the basis curve) is normalized by the residual
#'   magnitude; the mean is the projection weight and a one-sample t-test gives a
#'   significance value.
#' }
#'
#' @param x numeric matrix of single-trial evoked voltages with shape
#' \code{time x trials} (rows are timepoints, columns are trials). All trials
#' across all stimulation groups are stacked column-wise.
#' @param groups vector of length \code{ncol(x)} (integer, character or factor)
#' giving the stimulation subgroup of each column of \code{x}. At least two
#' subgroups are required and each subgroup needs at least two trials.
#' @param time optional numeric vector of length \code{nrow(x)} giving the
#' stimulus-aligned time (in seconds) of each row of \code{x}. Only used to crop
#' to \code{time_window} and as the plotting axis; the method itself does not
#' require it.
#' @param time_window optional numeric \code{c(lo, hi)} analysis window (seconds);
#' requires \code{time}. An \code{NA} extends to the corresponding data edge.
#' \code{NULL} (the default) uses all rows.
#' @param n_bpc integer or \code{NULL}; if given, fixes the number of basis
#' curves and skips automatic rank selection.
#' @param initial_rank integer or \code{NULL}; the starting \verb{NMF} rank for
#' automatic selection. \code{NULL} uses \code{min(n - 1L, 10L)} where \eqn{n} is
#' the number of subgroups.
#' @param zeta_threshold numeric \eqn{> 0}; rank-selection cutoff on \eqn{\zeta}
#' (see \sQuote{Details}); smaller values give fewer curves. Defaults to \code{1},
#' matching the \verb{BPC} reference.
#' @param null_class logical; if \code{TRUE} (the default) subgroups whose
#' winner-take-all loading falls below \eqn{1/(2\sqrt{n})} are left unassigned
#' (\code{NA}); if \code{FALSE} every subgroup is forced into a curve.
#' @param nmf_max_iters,nmf_tol passed to \code{\link{naive_nmf}}.
#' @param verbose logical; whether to report progress.
#'
#' @returns A named list of class \code{ravetools_bpc}:
#' \describe{
#' \item{\code{curves}}{Numeric matrix, time \eqn{\times} number of \verb{BPCs};
#'   column \eqn{q} is the basis profile curve \eqn{B_q(t)}, the first
#'   linear-kernel \verb{PCA} component of its member trials (unit-norm,
#'   sign-oriented to a positive mean projection).}
#' \item{\code{time}}{Numeric vector, the (windowed) time axis for \code{curves};
#'   the sample index when \code{time} was not supplied.}
#' \item{\code{group_labels}}{The unique stimulation subgroup labels, in the order
#'   used by \code{clusters} and \code{xi}.}
#' \item{\code{clusters}}{Integer vector, the \verb{BPC} index assigned to each
#'   subgroup; \code{NA} for subgroups left in the null class.}
#' \item{\code{excluded_groups}}{The labels of subgroups not represented by any
#'   \verb{BPC}.}
#' \item{\code{weights}}{A \code{data.frame} with one row per (\verb{BPC},
#'   subgroup) membership: \code{bpc}, \code{group}, \code{n_trials},
#'   \code{weight} (mean residual-normalized \eqn{\alpha}) and \code{p_value}.}
#' \item{\code{alpha}}{Numeric matrix, trials \eqn{\times} number of \verb{BPCs};
#'   the per-trial projection \eqn{\alpha} of every trial onto each basis curve.}
#' \item{\code{xi}}{Numeric matrix, the \eqn{n \times n} significance matrix
#'   \eqn{\Xi}.}
#' \item{\code{nmf}}{The \code{\link{naive_nmf}} result for the selected rank, with
#'   \code{H} (raw loadings) and \code{H0} (winner-take-all, thresholded).}
#' \item{\code{n_bpc}}{Integer, the number of basis curves found.}
#' \item{\code{time_window}}{The effective analysis window used (or \code{NULL}).}
#' }
#'
#' @references
#' The \verb{BPC} method is described in \doi{10.1371/journal.pcbi.1008710}, with
#' a reference Python implementation at
#' \url{https://github.com/MultimodalNeuroimagingLab/bpc_jupyter}.
#'
#' @seealso \code{\link{crp}}, \code{\link{crp_cluster}}, \code{\link{naive_nmf}}
#'
#' @examples
#' # Three response shapes, several stimulation groups per shape.
#' \donttest{
#' set.seed(1)
#' n_time <- 300L
#' tt <- seq(-0.2, 1, length.out = n_time)
#' shapes <- list(
#'   exp(-((tt - 0.08) / 0.03)^2) - 0.5 * exp(-((tt - 0.18) / 0.04)^2),
#'   exp(-((tt - 0.38) / 0.03)^2) - 0.5 * exp(-((tt - 0.50) / 0.04)^2),
#'   exp(-((tt - 0.70) / 0.04)^2)
#' )
#'
#' # 3 stimulation groups per shape, 12 trials each
#' V <- NULL
#' groups <- NULL
#' g <- 0L
#' for (s in seq_along(shapes)) {
#'   for (rep in seq_len(3L)) {
#'     g <- g + 1L
#'     trials <- outer(shapes[[s]], runif(12L, 0.5, 1.5)) +
#'       matrix(rnorm(n_time * 12L, sd = 0.2), n_time, 12L)
#'     V <- cbind(V, trials)
#'     groups <- c(groups, rep(g, 12L))
#'   }
#' }
#'
#' res <- bpc(V, groups, time = tt, time_window = c(0, 1), verbose = TRUE)
#' res$n_bpc
#' res$clusters
#' plot(res)
#' }
#'
#' @export
bpc <- function(
  x,
  groups,
  time = NULL,
  time_window = NULL,
  n_bpc = NULL,
  initial_rank = NULL,
  zeta_threshold = 1,
  null_class = TRUE,
  nmf_max_iters = 1e4,
  nmf_tol = c(1e-4, 1e-8),
  verbose = TRUE
) {

  # ---- validate inputs ----------------------------------------------------
  if (!is.matrix(x) || !is.numeric(x)) {
    stop("`x` must be a numeric matrix with shape time x trials.")
  }
  if (ncol(x) < 2L) {
    stop("`x` must have at least 2 trials (columns).")
  }
  if (length(groups) != ncol(x)) {
    stop("`groups` must have length `ncol(x)` (one label per trial/column).")
  }

  # ---- crop rows to the analysis window (optional) ------------------------
  if (!is.null(time)) {
    time <- as.numeric(time)
    if (length(time) != nrow(x)) {
      stop("`time` must have length `nrow(x)`.")
    }
  }
  if (!is.null(time_window)) {
    if (is.null(time)) {
      stop("`time_window` requires `time` to be supplied.")
    }
    time_window <- as.numeric(time_window)
    if (length(time_window) == 1L) {
      time_window <- c(0, time_window)
    } else {
      time_window <- time_window[c(1L, 2L)]
    }
    rng <- range(time)
    time_window[is.na(time_window)] <- rng[is.na(time_window)]
    if (is.unsorted(time_window)) {
      stop("`time_window` must be a numeric c(lo, hi) with lo < hi.")
    }
    row_sel <- time >= time_window[1L] & time <= time_window[2L]
  } else {
    row_sel <- rep(TRUE, nrow(x))
  }

  V <- x[row_sel, , drop = FALSE]
  axis <- if (is.null(time)) seq_len(nrow(x))[row_sel] else time[row_sel]

  # drop rows with any non-finite trial value
  keep <- rowSums(!is.finite(V)) == 0L
  V <- V[keep, , drop = FALSE]
  axis <- axis[keep]
  if (nrow(V) < 2L) {
    stop("Fewer than 2 finite timepoints remain after windowing.")
  }

  # ---- subgroup index lists -----------------------------------------------
  group_labels <- sort(unique(groups))
  n <- length(group_labels)
  if (n < 2L) {
    stop("`groups` must contain at least 2 distinct subgroups.")
  }
  idx_list <- lapply(group_labels, function(g) which(groups == g))
  group_sizes <- lengths(idx_list)
  if (any(group_sizes < 2L)) {
    small <- group_labels[group_sizes < 2L]
    stop("Every subgroup needs at least 2 trials; offending subgroup(s): ",
         paste(small, collapse = ", "), ".")
  }

  # ---- normalized internal projections P = V0^T V -------------------------
  norms <- sqrt(colSums(V * V))
  norms[norms == 0] <- 1
  V0 <- t(t(V) / norms)
  V0[!is.finite(V0)] <- 0
  P <- crossprod(V0, V)               # trials x trials

  # ---- significance matrix Xi ---------------------------------------------
  xi <- bpc_significance_matrix(P, idx_list)

  # ---- NMF on the non-negative, rescaled Xi -------------------------------
  t0 <- xi
  t0[t0 < 0 | !is.finite(t0)] <- 0
  mx <- max(t0)
  if (mx > 0) t0 <- t0 / mx

  if (is.null(initial_rank)) {
    initial_rank <- min(n - 1L, 10L)
  }
  initial_rank <- max(1L, min(as.integer(initial_rank), n - 1L))
  if (!is.null(n_bpc)) {
    n_bpc <- max(1L, min(as.integer(n_bpc), n - 1L))
  }

  sel <- crp_cluster_nmf_select(
    t0,
    n_clusters = n_bpc,
    initial_rank = initial_rank,
    zeta_threshold = zeta_threshold,
    nmf_max_iters = nmf_max_iters,
    nmf_tol = nmf_tol,
    verbose = verbose
  )
  nmf <- sel$nmf

  # ---- winner-take-all assignment + null class ----------------------------
  # Subgroups join the BPC with the largest row-normalized NMF loading; with
  # `null_class` a loading below 1/(2*sqrt(n)) - half the uniform loading
  # 1/sqrt(n) - leaves the subgroup unassigned (NA, the null class).
  Hn <- nmf$H
  rn <- sqrt(rowSums(Hn * Hn))
  rn[rn == 0] <- 1
  Hn <- Hn / rn
  raw_cluster <- apply(Hn, 2L, which.max)
  if (isTRUE(null_class)) {
    winner <- Hn[cbind(raw_cluster, seq_len(n))]
    raw_cluster[winner <= 1 / (2 * sqrt(n))] <- NA_integer_
  }
  present <- sort(unique(raw_cluster))
  clusters <- match(raw_cluster, present)
  n_found <- length(present)

  # thresholded winner-take-all loadings (rows = BPC, cols = subgroup)
  H0 <- matrix(0, nrow = nrow(Hn), ncol = n)
  assigned <- !is.na(raw_cluster)
  if (any(assigned)) {
    H0[cbind(raw_cluster[assigned], which(assigned))] <-
      Hn[cbind(raw_cluster[assigned], which(assigned))]
  }

  # ---- basis profile curves -----------------------------------------------
  basis_curves <- bpc_basis(V, clusters, idx_list, n_found)

  # ---- per-trial projection weights and significance ----------------------
  stat <- bpc_curve_statistics(V, basis_curves, clusters, idx_list, group_labels)

  excluded_groups <- group_labels[is.na(clusters)]

  rownames(basis_curves) <- NULL
  dimnames(xi) <- list(as.character(group_labels), as.character(group_labels))

  nmf$H0 <- H0

  structure(
    class = "ravetools_bpc",
    list(
      curves = basis_curves,
      time = axis,
      group_labels = group_labels,
      clusters = clusters,
      excluded_groups = excluded_groups,
      weights = stat$weights,
      alpha = stat$alpha,
      xi = xi,
      nmf = nmf,
      n_bpc = n_found,
      time_window = if (is.null(time_window)) NULL else time_window
    )
  )
}

#' @title Plot Basis Profile Curve results
#' @description S3 plot method for \code{ravetools_bpc} objects returned by
#' \code{\link{bpc}}. Draws a three-panel figure: the basis profile curves, the
#' subgroup-by-subgroup significance matrix \eqn{\Xi} ordered by assignment, and
#' the per-subgroup projection weights.
#' @param x a \code{ravetools_bpc} object.
#' @param ... ignored.
#' @return Invisibly returns \code{x}.
#' @export
plot.ravetools_bpc <- function(x, ...) {
  Q <- x$n_bpc
  cols <- grDevices::hcl.colors(max(Q, 2L), palette = "Dark 3")

  op <- graphics::par(mfrow = c(1L, 3L))
  on.exit(graphics::par(op), add = TRUE)

  # ---- Panel 1: basis profile curves --------------------------------------
  graphics::matplot(
    x$time, x$curves, type = "l", lty = 1L, lwd = 2L, col = cols[seq_len(Q)],
    las = 1L, xlab = "Time (s)", ylab = "Basis amplitude (a.u.)",
    main = "Basis profile curves"
  )
  graphics::abline(h = 0, col = "#cccccc", lty = 3L)
  graphics::legend(
    "topright", bty = "n", cex = 0.8, lwd = 2L, col = cols[seq_len(Q)],
    legend = sprintf("BPC %d (n=%d)", seq_len(Q),
                     tabulate(x$clusters, nbins = Q))
  )

  # ---- Panel 2: significance matrix ordered by assignment -----------------
  ord <- order(x$clusters, na.last = TRUE)
  S <- x$xi[ord, ord, drop = FALSE]
  graphics::image(
    seq_len(nrow(S)), seq_len(ncol(S)), S,
    col = grDevices::hcl.colors(64L, palette = "YlOrRd", rev = TRUE),
    xlab = "Subgroup (ordered by BPC)", ylab = "Subgroup",
    main = expression("Significance matrix " * Xi), las = 1L
  )
  brk <- cumsum(tabulate(x$clusters, nbins = Q))
  brk <- brk[brk > 0 & brk < nrow(S)] + 0.5
  graphics::abline(v = brk, h = brk, col = "#202020", lty = 1L, lwd = 1L)

  # ---- Panel 3: per-subgroup projection weights ---------------------------
  w <- x$weights
  if (nrow(w)) {
    w <- w[order(w$bpc), , drop = FALSE]
    graphics::barplot(
      w$weight, col = cols[w$bpc], border = NA, las = 1L,
      names.arg = w$group, cex.names = 0.7,
      ylab = "Projection weight", main = "Projection weights"
    )
  } else {
    graphics::plot.new()
    graphics::title(main = "Projection weights")
  }

  invisible(x)
}


# --------------------------------------------------------------------------
# Internal helpers (not exported)
# --------------------------------------------------------------------------

# Significance matrix Xi (BPC `nativeNormalized`). `P` is the trials x trials
# projection matrix (normalized x native); `idx_list` gives the trial columns of
# each subgroup. Entry (k, l) is the one-sample t-statistic versus zero of the
# off-diagonal within-group block (k == l) or the full cross block (k != l). The
# resulting n x n matrix is generally asymmetric.
bpc_significance_matrix <- function(P, idx_list) {
  n <- length(idx_list)
  xi <- matrix(0, nrow = n, ncol = n)
  for (k in seq_len(n)) {
    ik <- idx_list[[k]]
    for (l in seq_len(n)) {
      il <- idx_list[[l]]
      if (k == l) {
        block <- P[ik, ik, drop = FALSE]
        b <- block[row(block) != col(block)]   # all off-diagonal elements
      } else {
        b <- as.numeric(P[ik, il, drop = FALSE])
      }
      b <- b[is.finite(b)]
      if (length(b) < 2L) next
      s <- stats::sd(b)
      if (!is.finite(s) || s == 0) next
      xi[k, l] <- mean(b) / (s / sqrt(length(b)))
    }
  }
  xi
}

# Per-BPC basis profile curve: the first linear-kernel PCA component of all
# trials in the BPC's member subgroups, sign-oriented to a positive mean
# projection. Reuses `crp_kt_pca`; mirrors `crp_cluster_basis`.
bpc_basis <- function(V, clusters, idx_list, n_found) {
  Tn <- nrow(V)
  basis_curves <- matrix(0, nrow = Tn, ncol = n_found)
  for (q in seq_len(n_found)) {
    members <- which(clusters == q)
    cols <- unlist(idx_list[members], use.names = FALSE)
    Vq <- V[, cols, drop = FALSE]
    pca <- crp_kt_pca(Vq)
    Bq <- pca$E[, 1L]
    if (mean(crossprod(Bq, Vq)) < 0) Bq <- -Bq
    basis_curves[, q] <- Bq
  }
  basis_curves
}

# Per-trial coefficients and projection weights (BPC `curvesStatistics` +
# `projectionWeights`). For each basis curve B: alpha = B^T V (one value per
# trial), residual ep = V - B alpha^T, and the per-trial residual power is the
# diagonal of ep^T ep. For each member subgroup the residual-normalized
# coefficients alpha / sqrt(ep2) give a mean projection weight and a one-sample
# t-test p-value.
bpc_curve_statistics <- function(V, basis_curves, clusters, idx_list, group_labels) {
  n_found <- ncol(basis_curves)
  K <- ncol(V)
  alpha <- matrix(0, nrow = K, ncol = n_found)
  rows <- list()
  for (q in seq_len(n_found)) {
    B <- basis_curves[, q]
    al <- as.numeric(crossprod(B, V))            # length K
    alpha[, q] <- al
    ep <- V - outer(B, al)
    ep2 <- colSums(ep * ep)                      # per-trial residual power
    members <- which(clusters == q)
    for (g in members) {
      idx <- idx_list[[g]]
      ratio <- al[idx] / sqrt(ep2[idx])
      ratio <- ratio[is.finite(ratio)]
      pval <- NA_real_
      if (length(ratio) >= 2L) {
        tt <- tryCatch(stats::t.test(ratio, mu = 0), error = function(e) NULL)
        if (!is.null(tt)) pval <- tt$p.value
      }
      rows[[length(rows) + 1L]] <- data.frame(
        bpc = q,
        group = group_labels[g],
        n_trials = length(idx),
        weight = if (length(ratio)) mean(ratio) else NA_real_,
        p_value = pval,
        stringsAsFactors = FALSE
      )
    }
  }
  weights <- if (length(rows)) {
    do.call(rbind, rows)
  } else {
    data.frame(
      bpc = integer(0), group = group_labels[0], n_trials = integer(0),
      weight = numeric(0), p_value = numeric(0), stringsAsFactors = FALSE
    )
  }
  list(weights = weights, alpha = alpha)
}

#' @title Project a surface onto a sphere and relax metric distortion
#' @description
#' Projects a closed triangular surface mesh radially onto a sphere and then
#' iteratively relaxes the metric distortion this introduces ("unfolding"), so
#' that geodesic distances and face orientations stay close to those of the
#' input surface - a step typically used to prepare an inflated cortical
#' surface for spherical registration.
#'
#' @details
#' \strong{This is a reduced procedure, not a complete reproduction of any
#' particular reference implementation.} The full unfolding procedure
#' described in the literature integrates seven weighted energy terms against
#' a separate reference surface through a multi-resolution, multi-thousand
#' line optimization pipeline, which is out of scope for this package. Instead,
#' this implementation keeps the two dominant terms and integrates them with
#' the same momentum-based machinery \code{\link{mris_inflate}} uses:
#' \describe{
#'   \item{Distance term (\code{l_dist})}{Restoring force pulling each vertex
#'     back towards the input mesh's own original distances to its
#'     neighbors.}
#'   \item{Area term (\code{l_area})}{Repulsive force that acts only on
#'     folded/negative-area faces, pushing their vertices apart - this is the
#'     actual "unfolding" mechanism.}
#' }
#' Both terms are integrated via momentum integration with gradient averaging
#' and a 1 mm per-step displacement cap, exactly as \code{\link{mris_inflate}}
#' does.
#'
#' A further simplification: the reference procedure relaxes a freshly
#' spherical-projected surface against a separate, previously-loaded
#' white-matter reference surface for the distance term; since this function
#' takes a single mesh as input, the input mesh's own metric (captured before
#' projection) is used as the reference instead - the same convention
#' \code{\link{mris_inflate}} already uses, and the practical analogue (the
#' white-matter surface is normally what gets inflated to produce a typical
#' input to this kind of unfolding step).
#'
#' @param mesh triangular mesh of class \code{'mesh3d'}, typically an inflated
#'   surface (see \code{\link{mris_inflate}}).  Must be closed, manifold, and
#'   genus-0 - the same hard precondition as \code{\link{mris_inflate}}.
#'   \code{mris_sphere} will raise an error if such defects are detected.
#' @param target_radius radius of the target sphere.  Default \code{100}.
#' @param n_averages number of gradient-averaging passes applied to the
#'   distance-term gradient each iteration (the same neighborhood-averaging
#'   mechanism \code{\link{mris_inflate}} uses).  Default \code{64}; reduced
#'   from the much larger averaging size used in production pipelines (tuned
#'   for cortical surfaces with hundreds of thousands of vertices) to a size
#'   more practical for typical \code{'mesh3d'} inputs.
#' @param niterations number of unfolding iterations.  Default \code{25}.
#' @param l_dist distance-preservation coefficient.  Default \code{1.0}.
#' @param l_area folded-face repulsion coefficient.  Default \code{1.0}.
#' @param momentum momentum coefficient.  Default \code{0.9}.
#' @param dt time step.  Default \code{0.05}.
#' @param verbose logical; print per-iteration progress.  Default \code{FALSE}.
#'
#' @returns The projected and relaxed surface as a \code{'mesh3d'} object with
#'   \code{vb}, \code{it}, and \code{normals}.
#'
#' @examples
#' if (is_not_cran()) {
#'
#' sphere <- vcg_sphere(sub_division = 3L)
#'
#' # deform so there is metric distortion to relax
#' sphere$vb[1, ] <- sphere$vb[1, ] * (1 + 0.2 * rnorm(ncol(sphere$vb)))
#'
#' # Fix defects
#' sphere <- vcg_fix_defects(sphere)
#'
#' result <- mris_sphere(
#'   sphere,
#'   n_averages = 8L,
#'   niterations = 10L,
#'   target_radius = 1,
#'   verbose = TRUE
#' )
#'
#' plot_mesh_polygon(list(sphere, result),
#'                   col = list("gray", "red"),
#'                   alpha = c(0.3, 0.9))
#'
#'
#'
#' }
#'
#' @export
mris_sphere <- function(
    mesh,
    target_radius = 100,
    n_averages    = 64L,
    niterations   = 25L,
    l_dist        = 1.0,
    l_area        = 1.0,
    momentum      = 0.9,
    dt            = 0.05,
    verbose       = FALSE
) {
  mesh <- meshintegrity(mesh, facecheck = TRUE)

  vb <- mesh$vb[1:3, , drop = FALSE]
  storage.mode(vb) <- "double"

  it <- mesh$it
  storage.mode(it) <- "integer"

  tmp <- mrisSphere(
    vb_           = vb,
    it_           = it,
    target_radius = as.double(target_radius)[[1L]],
    n_averages    = as.integer(n_averages)[[1L]],
    niterations   = as.integer(niterations)[[1L]],
    l_dist        = as.double(l_dist)[[1L]],
    l_area        = as.double(l_area)[[1L]],
    momentum      = as.double(momentum)[[1L]],
    dt            = as.double(dt)[[1L]],
    verbose       = as.logical(verbose)[[1L]]
  )

  structure(
    list(
      vb      = rbind(tmp$vb, 1),
      it      = it,
      normals = rbind(tmp$normals, 1)
    ),
    class = c("ravetools_mesh3d", "mesh3d")
  )
}

#' @name rgl-call
#' @title Safe ways to call package \code{'rgl'} without requiring \code{'x11'}
#' @description
#' Internally used for example show-cases. Please install package \code{'rgl'}
#' manually to use these functions.
#'
#' @param FUN \code{'rgl'} function name
#' @param ... passed to \code{'rgl'} function
#' @param expr expression within which \code{'rgl'} functions are called
#' @param quoted whether \code{expr} is quoted
#' @param env environment in which \code{expr} is evaluated
#' @param x triangular \code{'mesh3d'} object
#' @param length,lwd,col normal vector length, size, and color
#' @examples
#'
#'
#' # Make sure the example does not run when compiling
#' # or check the package
#' if(FALSE) {
#'
#'   volume <- array(0, dim = c(8,8,8))
#'   volume[4:5, 4:5, 4:5] <- 1
#'   mesh <- mesh_from_volume(volume, verbose = FALSE)
#'
#'   rgl_view({
#'
#'     rgl_call("shade3d", mesh, col = 3)
#'     rgl_plot_normals(mesh)
#'
#'   })
#'
#' }
#'
#'
NULL

check_rgl <- function(strict = NA) {
  rgl_unavailable <- function() {
    msg <- "Package `rgl` is not installed. Please install `rgl` to use this function."
    if (isTRUE(strict)) {
      stop(msg)
    } else if (is.na(strict)) {
      message(msg)
    }
    FALSE
  }
  if (identical(Sys.getenv("RAVETOOLS_RGL_DISABLED"), "1")) { return(rgl_unavailable()) }
  if ( getOption("ravetools.rgl.disabled", FALSE) ) { return(rgl_unavailable()) }
  if ( system.file(package = "rgl") == "" ) { return(rgl_unavailable()) }
  TRUE
}

#' @rdname rgl-call
#' @export
rgl_call <- function(FUN, ...) {
  check_rgl()

  # Cannot set this option back with on.exit, rgl may throw warning/errors
  # when X11 is unavailable. This option forces rgl to use webgl renderers.
  options(rgl.useNULL = TRUE)
  rgl <- asNamespace("rgl")
  f <- rgl[[FUN]]
  if (!is.function(f)) {
    stop("Function ", FUN, " is not a `rgl` function.")
  }
  rgl[[FUN]](...)
}

#' @rdname rgl-call
#' @export
rgl_view <- function(expr, quoted = FALSE, env = parent.frame()) {
  # Suppress RGL
  if (!quoted) {
    expr <- substitute(expr)
  }
  dev <- rgl_call("open3d")
  on.exit({
    rgl_call("close3d", dev = dev)
  }, add = TRUE, after = TRUE)
  eval(expr, envir = env)
  rgl_call("rglwidget")
}

#' @rdname rgl-call
#' @export
rgl_plot_normals <- function(x, length = 1, lwd = 1, col = 1, ...) {
  if (!"mesh3d" %in% class(x)) {
    stop("please provide object of class mesh3d")
  }
  args <- list(...)
  if ("long" %in% names(args)) {
    length <- args$long
    warning("argument 'long' is deprecated, please use 'length' instead")
  }
  if (is.null(x$normals)) {
    if (!is.null(x$it)) {
      x <- vcg_update_normals(x)
    } else {
      stop("mesh has neither normals nor faces")
    }
  }
  n.mesh <- list()
  lvb <- dim(x$vb)[2]
  vb <- x$vb
  vb.norm <- vb[1:3, , drop = FALSE] + t(length * t(x$normals[1:3, , drop = FALSE]))
  vb <- cbind(vb[1:3, , drop = FALSE], vb.norm)
  vb <- rbind(vb, 1)
  it <- rbind(1:lvb, 1:lvb, (1:lvb) + lvb)
  n.mesh$vb <- vb
  n.mesh$it <- it
  class(n.mesh) <- c("mesh3d", "shape3d")
  rgl_call("wire3d", n.mesh, color = col, lwd = lwd, lit = FALSE)
}

#' Design \code{'FIR'} filter using \code{\link{firls}}
#' @param method method to generate 'FIR' filter, default is using
#' \code{\link[gsignal]{kaiser}} estimate, other choices are
#' \code{\link{firls}} (with \code{hamming} window) and
#' \code{\link[gsignal]{remez}} design.
#' @param sample_rate sampling frequency
#' @param filter_order filter order, leave \code{NA} (default) if undecided
#' @param data_size minimum length of data to apply the filter, used to
#' decide the maximum filter order. For 'FIR' filter, data length must be
#' greater than \code{3xfilter_order}
#' @param high_pass_freq high-pass frequency; default is \code{NA} (no
#' high-pass filter will be applied)
#' @param high_pass_trans_freq high-pass frequency band-width; default
#' is automatically inferred from data size.
#' Frequency \code{high_pass_freq - high_pass_trans_freq} is the corner
#' of the stop-band
#' @param low_pass_freq low-pass frequency; default is \code{NA} (no
#' low-pass filter will be applied)
#' @param low_pass_trans_freq low-pass frequency band-width; default
#' is automatically inferred from data size.
#' Frequency \code{low_pass_freq + low_pass_trans_freq} is the corner
#' of the stop-band
#' @param stopband_attenuation allowable power attenuation (in decibel) at
#' transition frequency; default is \code{40} dB.
#' @param scale whether to scale the filter for unity gain
#'
#' @details
#' Filter type is determined from \code{high_pass_freq} and
#' \code{low_pass_freq}. High-pass frequency is ignored if \code{high_pass_freq}
#' is \code{NA}, hence the filter is low-pass filter. When
#' \code{low_pass_freq} is \code{NA}, then
#' the filter is high-pass filter. When both \code{high_pass_freq} and
#' \code{low_pass_freq} are valid (positive, less than 'Nyquist'), then
#' the filter is a band-pass filter if band-pass is less than low-pass
#' frequency, otherwise the filter is band-stop.
#'
#' Although the peak amplitudes are set at 1 by \code{low_pass_freq} and
#' \code{high_pass_freq}, the transition from peak amplitude to zero require
#' a transition, which is tricky but also important to set.
#' When 'FIR' filters have too steep transition boundaries, the filter tends to
#' have ripples in peak amplitude, introducing artifacts to the final signals.
#' When the filter is too flat, components from unwanted frequencies may also
#' get aliased into the filtered signals. Ideally, the transition bandwidth
#' cannot be too steep nor too flat. In this function, users may control
#' the transition frequency bandwidths via \code{low_pass_trans_freq} and
#' \code{high_pass_trans_freq}. The power at the end of transition is defined
#' by \code{stopband_attenuation}, with default value of \code{40} (i.e.
#' -40 dB, this number is automatically negated during the calculation).
#' By design, a low-pass 5 Hz filter with 1 Hz transition bandwidth results in
#' around -40 dB power at 6 Hz.
#'
#'
#' @returns 'FIR' filter in 'Arma' form.
#'
#' @examples
#'
#' # ---- Basic -----------------------------
#'
#' sample_rate <- 500
#' data_size <- 1000
#'
#' # low-pass at 5 Hz, with auto transition bandwidth
#' # from kaiser's method, with default stopband attenuation = 40 dB
#' filter <- design_filter_fir(
#'   low_pass_freq = 5,
#'   sample_rate = sample_rate,
#'   data_size = data_size
#' )
#'
#' # Passband ripple is around 0.08 dB
#' # stopband attenuation is around 40 dB
#' print(filter)
#'
#' diagnose_filter(
#'   filter$b, filter$a,
#'   fs = sample_rate,
#'   n = data_size,
#'   cutoffs = c(-3, -6, -40),
#'   vlines = 5
#' )
#'
#' # ---- Advanced ---------------------------------------------
#'
#' sample_rate <- 500
#' data_size <- 1000
#'
#' # Rejecting 3-8 Hz, with transition bandwidth 0.5 Hz at both ends
#' # Using least-square (firls) to generate FIR filter
#' # Suggesting the filter order n=160
#' filter <- design_filter_fir(
#'   low_pass_freq = 3, low_pass_trans_freq = 0.5,
#'   high_pass_freq = 8, high_pass_trans_freq = 0.5,
#'   filter_order = 160,
#'   sample_rate = sample_rate,
#'   data_size = data_size,
#'   method = "firls"
#' )
#'
#' #
#' print(filter)
#'
#' diagnose_filter(
#'   filter$b, filter$a,
#'   fs = sample_rate,
#'   n = data_size,
#'   cutoffs = c(-1, -40),
#'   vlines = c(3, 8)
#' )
#'
#'
#'
#' @export
design_filter_fir <- function(
    sample_rate, filter_order = NA, data_size = NA,
    high_pass_freq = NA, high_pass_trans_freq = NA,
    low_pass_freq = NA, low_pass_trans_freq = NA,
    stopband_attenuation = 40, scale = TRUE,
    method = c("kaiser", "firls", "remez")
) {

  method <- match.arg(method)

  # DIPSAUS DEBUG START
  # debug <- function(x, ...) { message(x, ...) }
  # sample_rate <- 200
  # data_size <- 2000
  # list2env(list(sample_rate = sample_rate,
  #                 high_pass_freq = NA, high_pass_trans_freq = NA,
  #                 low_pass_freq = 3, low_pass_trans_freq = NA), envir=.GlobalEnv)
  # list2env(list(filter_order = NA, low_pass_freq = 50, high_pass_freq = 4,
  #               low_pass_trans_freq = NA, high_pass_trans_freq = NA), envir=.GlobalEnv)
  # stopband_attenuation <- 40


  nyquist <- sample_rate / 2
  stopband_attenuation <- abs(stopband_attenuation)


  stopifnot2(
    !is.na(low_pass_freq) || !is.na(high_pass_freq),
    msg = "Please specify low-pass and/or high-pass frequencies"
  )
  stopifnot2(is.finite(stopband_attenuation) && stopband_attenuation > 0,
             msg = "`stopband_attenuation` must be a finite positive number")

  if (is.na(data_size)) {
    max_nfilters <- Inf
    # Huristic: 32 / pi / 2.285 / 333 (stopband atten 40 dB with data size 1000)
    suggested_trans_bandwidth <- 0.01
  } else {
    max_nfilters <- max(floor((data_size - 1) / 3), 2)
    if ( max_nfilters %% 2 == 1 ) {
      max_nfilters <- max_nfilters - 1
    }

    if ( isTRUE(filter_order > max_nfilters) ) {
      filter_order <- max_nfilters
    }
    # dw <- 2 * pi * diff(f)/fs
    # n <- max(1, ceiling((A - 8)/(2.285 * dw)))
    if ( stopband_attenuation > 8 ) {
      suggested_trans_bandwidth <- (stopband_attenuation - 8) / pi / 2.285 / max_nfilters
    } else {
      suggested_trans_bandwidth <- stopband_attenuation / 22 / max_nfilters
    }
  }

  # determine the filter type
  if ( is.na(high_pass_freq) ) {
    ftype <- "low"
    w_pass <- low_pass_freq / nyquist
    w_trans <- low_pass_trans_freq / nyquist
    w_trans[is.na(w_trans)] <- suggested_trans_bandwidth

    w_stop <- clamp(w_pass + abs(w_trans), min = 0, max = 1)

    # kaiserord
    bands <- c(w_pass, w_stop)
    mag <- c(1, 0)

    # firls
    mag2 <- c(1, 1, 0, 0)

  } else if ( is.na(low_pass_freq) ) {
    ftype <- "high"
    w_pass <- high_pass_freq / nyquist
    w_trans <- high_pass_trans_freq / nyquist
    w_trans[is.na(w_trans)] <- suggested_trans_bandwidth

    w_stop <- clamp(w_pass - abs(w_trans), min = 0, max = 1)

    # kaiserord
    bands <- c(w_stop, w_pass)
    mag <- c(0, 1)

    # firls
    mag2 <- c(0, 0, 1, 1)


  } else if ( isTRUE(high_pass_freq <= low_pass_freq) ) {
    ftype <- "pass"
    w_pass <- c(high_pass_freq, low_pass_freq) / nyquist
    w_trans <- c(high_pass_trans_freq, low_pass_trans_freq) / nyquist
    w_trans[is.na(w_trans)] <- suggested_trans_bandwidth

    w_stop <- clamp( w_pass + c(-1, 1) * abs(w_trans), min = 0, max = 1)

    # kaiserord
    bands <- c(w_stop[[1]], w_pass, w_stop[[2]])
    mag <- c(0, 1, 0)

    # firls
    mag2 <- c(0, 0, 1, 1, 0, 0)

  } else {
    ftype <- "stop"
    w_pass <- c(low_pass_freq, high_pass_freq) / nyquist
    auto_trans <- is.na(c(low_pass_trans_freq, high_pass_trans_freq))
    w_trans <- c(low_pass_trans_freq, high_pass_trans_freq) / nyquist
    w_trans[auto_trans] <- suggested_trans_bandwidth

    w_stop <- clamp( w_pass + c(1, -1) * abs(w_trans), min = 0, max = 1)

    # Guard: stopband edges must not cross. When auto-inferred transitions
    # are too wide (notch narrower than suggested_trans_bandwidth), silently
    # cap them so the filter remains constructible. Only error when the
    # user explicitly supplied transition values that are the cause.
    if (w_stop[[1]] >= w_stop[[2]]) {
      notch_width <- w_pass[[2]] - w_pass[[1]]
      fixed_total <- sum(abs(w_trans[!auto_trans]))
      if (!any(auto_trans) || fixed_total >= notch_width) {
        stopifnot2(FALSE, msg = sprintf(
          paste0(
            "Band-stop transition bandwidths overlap: notch is %.4g-%.4g Hz ",
            "(width %.4g Hz) but the combined specified transition bandwidth ",
            "exceeds it. Set each of `low_pass_trans_freq` and ",
            "`high_pass_trans_freq` to less than %.4g Hz."
          ),
          low_pass_freq, high_pass_freq,
          notch_width * nyquist,
          notch_width * nyquist / 2
        ))
      }
      n_auto <- sum(auto_trans)
      w_trans[auto_trans] <- (notch_width - fixed_total) * (1 - 1e-6) / n_auto
      w_stop <- clamp( w_pass + c(1, -1) * abs(w_trans), min = 0, max = 1)
    }

    # kaiserord
    bands <- c(w_pass[[1]], w_stop, w_pass[[2]])
    mag <- c(1, 0, 1)

    # firls
    mag2 <- c(1, 1, 0, 0, 1, 1)
  }



  r_stop <- stopband_attenuation
  r_pass <- -20 * log10(1 - 10 ^ (-r_stop / 20))

  # get filter
  # c("kaiser", "ls", "remez")
  # Wc is (w_pass + w_stop) / 2
  kaisprm <- gsignal::kaiserord(bands, mag, dev = 10 ^ (-r_stop / 20))
  if (!is.na(filter_order)) {
    kaisprm$n <- filter_order
  }
  if (kaisprm$n > max_nfilters) {
    # even order for stop and high to ensure unity gain is non-zero at nyquist
    if (ftype %in% c("stop", "high") && max_nfilters %% 2 == 1) {
      max_nfilters <- max_nfilters + 1
    }
    kaisprm$n <- max_nfilters
  }

  params <- list(
    type = c("fir", ftype),
    method = method,
    order = kaisprm$n,
    scale = scale,
    passband = w_pass,
    stopband = w_stop,
    sample_rate = sample_rate
  )

  switch(
    method,
    "kaiser" = {
      filter <- fir1(
        n = kaisprm$n, w = kaisprm$Wc, type = kaisprm$type,
        window = gsignal::kaiser(kaisprm$n + 1, kaisprm$beta),
        scale = FALSE  # outer block handles scaling uniformly across all methods
      )
      params$cutoff <- kaisprm$Wc
      params$beta <- kaisprm$beta
      params
    },
    "firls" = {
      w <- c(0, rep(bands, each = 2), 1)
      a <- rep(mag2, each = 2)
      a <- a[-c(1, length(a))]
      filter <- firls(N = kaisprm$n, freq = w, A = a)
      wind <- hamming(length(filter$b))
      filter$b <- filter$b * wind
      params$frequency <- c(0, bands, 1)
      params$amplitude <- mag2
    },
    "remez" = {
      if (kaisprm$n < 4) {
        kaisprm$n <- 4
      }
      b <- unclass(gsignal::remez(kaisprm$n, c(0, bands, 1), mag2))
      filter <- structure(
        class = "Arma",
        list(b = b, a = 1)
      )
      params$frequency <- c(0, bands, 1)
      params$amplitude <- mag2
    }
  )
  if ( scale ) {
    # Normalize filter to unity gain at a reference frequency f0.
    # Formula (matches MATLAB scalefilter / fir1 internals):
    #   b <- b / |H(e^{j*pi*f0})| = b / |sum_k b[k] * exp(-j*2*pi*k*f0/2)|
    #
    # Reference frequency f0 (in normalized [0,1] units, 1 = Nyquist):
    #   low-pass  -> f0 = 0   (DC gain = 1)
    #   high-pass -> f0 = 1   (Nyquist gain = 1)
    #   band-pass -> f0 = mean(kaisprm$Wc), the average of the two transition-band
    #               midpoints.  kaisprm$Wc[k] = (band_stop_edge[k] + band_pass_edge[k]) / 2,
    #               so mean(Wc) = (w_stop[1] + w_pass[1] + w_pass[2] + w_stop[2]) / 4.
    #               This is the convention used by MATLAB fir1 / scale_filter: f0 is taken
    #               from freq[3:4] of the internal frequency vector built by fir1, which are
    #               exactly the two edges of the first passband, i.e. Wc[1] and Wc[2].
    #
    #               FIX (2026-05): the original code used f0 = mean(w_pass), i.e. the
    #               midpoint of the passband edges only.  This differs from MATLAB's choice
    #               by (w_trans_high - w_trans_low) / 4: zero only when both transition
    #               bandwidths are equal.  Using mean(kaisprm$Wc) matches MATLAB exactly.
    #
    #               # OLD (pre-fix): f0 <- mean(w_pass)
    #
    #   band-stop -> f0 = 0   (DC gain = 1).
    #               MATLAB fir1 sets fband = TRUE for stop-band filters, which causes
    #               scale_filter to use b / sum(b), i.e. normalise at DC regardless of
    #               where the notch sits.
    #
    #               FIX (2026-05): the original code used a heuristic:
    #                 if (w_stop[1] + w_stop[2] > 1) f0 <- 0 else f0 <- 1
    #               The intent was to pick whichever passband (low or high) is larger,
    #               but this is inconsistent with MATLAB and can flip for borderline notch
    #               positions near Nyquist/2.  Always using f0 = 0 matches MATLAB.
    #
    #               # OLD (pre-fix):
    #               # if (w_stop[[1]] + w_stop[[2]] > 1) {
    #               #   # high-frequency band stop
    #               #   f0 <- 0
    #               # } else {
    #               #   f0 <- 1
    #               # }
    switch(
      ftype,
      "low"  = { f0 <- 0 },
      "high" = { f0 <- 1 },
      "pass" = {
        # Old code used mean(w_pass), i.e. midpoint of passband edges only, 
        # which differs from MATLAB's choice by (w_trans_high - w_trans_low) / 4
        # f0 <- mean(w_pass)

        # Updated: 2026-05
        # transition-band midpoints average: matches MATLAB fir1 scale_filter
        f0 <- mean(kaisprm$Wc)
      },
      {
        # if (w_stop[[1]] + w_stop[[2]] > 1) {
        #   # high-freqeuncy band stop
        #   f0 <- 0
        # } else {
        #   f0 <- 1
        # }

        # Updated: 2026-05
        # Always normalise at DC (f0 = 0) for band-stop filters,
        # band-stop: always DC, matching MATLAB fir1 (fband=TRUE -> b/sum(b))
        f0 <- 0
      }
    )
    L <- length(filter$b)
    filter$b <- filter$b / Mod(sum(
      exp(-1i * 2 * pi * seq(0, L - 1) * (f0 / 2)) * filter$b
    ))
  }

  filter$parameters <- params
  filter$checks <- check_filter(b = filter$b, a = filter$a, w = c(w_pass, w_stop),
                                r_expected = -rep(c(r_pass, r_stop), each = length(w_pass)), fs = sample_rate)


  class(filter) <- c("ravetools-design_filter_fir", "ravetools-design_filter",
                     "ravetools-printable", class(filter))

  filter
}

#' @export
`format.ravetools-design_filter` <- function(x, ...) {
  params <- x$parameters
  checks <- sprintf("  %s", format(x$checks)[-c(1, 2)])
  c(
    sprintf("<%s filter>", toupper(params$type[[1]])),
    sprintf("  Type : %s", params$type[[2]]),
    sprintf("  Method: %s", params$method),
    sprintf("  Order: %d", params$order),
    "  Magnitude:",
    checks
  )
}

# attr(design_filter_fir, "printable_message") <- c(
#   "Usage: ",
#   "",
#   deparse(as.call(c(list(quote(design_filter_fir)), formals(design_filter_fir)))),
#   "",
#   "Using band-pass filter as an example.",
#   "  The frequency response is (approximately) given by:",
#   "",
#   "     1-|               ----------",
#   "       |             /|         | \\",
#   "       |            / |         |  \\",
#   "       |           /  |         |   \\",
#   " |H|   |          /   |         |    \\",
#   "   Rpa-|---------/----+---------+-----\\",
#   "       |        /|    |         |     |\\           ",
#   "       |       / |    |         |     | \\          ",
#   "       |      /  |    |         |     |  \\         ",
#   "     0-|-----/   |    |         |     |   \\--------|",
#   "       0    Fs1 Fp1   Fh        Fl   Fp2  Fs2       Nyq",
#   "",
#   "  Where:",
#   "",
#   "      * Nyq:      (Nyquist) ",
#   "            = `sample_rate` / 2",
#   "      * Rpa:      (Transition attenuation in amplitude) ",
#   "            = 10 ^ { - `trans_attenuation` / 20 }",
#   "      * Fh:       (High-pass frequency) ",
#   "            = `high_pass_freq`",
#   "      * Fl:       (Low-pass frequency) ",
#   "            = `low_pass_freq`",
#   "      * Fp1, Fp2: (Transition frequency)",
#   "        Fp1 = Fh - `high_pass_trans_freq`",
#   "        Fp2 = Fl + `low_pass_trans_freq`",
#   "      * Fs1, Fs2: (Stop-band frequency, auto-derived)"
# )
#
# class(design_filter_fir) <- c("ravetools-printable", "function")

test_that("Extra local tests", {

  path <- "~/Dropbox (Personal)/projects/ravetools/adhoc/testextra"

  skip_if_not(dir.exists(path))

  fs <- list.files(path, full.names = TRUE, pattern = "\\.R$", ignore.case = TRUE)
  for (f in fs) {
    source(f, local = TRUE, chdir = TRUE)
  }

})

---
title: "Mesh geometry"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Mesh geometry}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 3.2,
  dpi = 96,
  out.width = "100%"
)
library(ravetools)
oldpar <- graphics::par(no.readonly = TRUE)
```

The `vcg_*` and `mris_*` families are `ravetools`' low-level geometry layer.
They are the building blocks that the surface pipelines are assembled from, and
they are usable directly. This article walks the whole family, grouped the way
the `README` table groups it: construct, measure, repair, smooth, query,
subset, and draw.

Every mesh these functions return carries the `ravetools_mesh3d` class, which
is an `rgl`-style `mesh3d` -- a `3 x n` vertex matrix `vb` and a `3 x m`
one-based triangle matrix `it` -- so `plot()` renders it in base `R` graphics
with no `rgl` dependency. `ensure_mesh3d` coerces the other common surface
formats (`fs.surface` from `freesurferformats`, `ieegio_surface` from `ieegio`,
`surf.asc`) into the same layout, so a `FreeSurfer` surface drops into any
function below.

## 1. Two primitives

The cheapest way to get a mesh is to make one. `vcg_sphere` returns a
subdivided `icosphere` of unit radius centered at the origin, and
`plane_geometry` returns a flat triangulated grid in the `z = 0` plane.

```{r primitives}
sphere <- vcg_sphere(sub_division = 4)
sphere

plane <- plane_geometry(width = 2.6, height = 2.6, shape = c(40, 40))
plane
```

`shape` counts vertices per side, so a `40 x 40` grid is `39 x 39` cells and
twice that many triangles. Tilt the plane and lift it, and it slices through
the sphere:

```{r tilt}
tilt <- new_matrix4()$make_rotation_y(25 * pi / 180)$to_array()
plane$vb <- (tilt %*% rbind(plane$vb[1:3, ], 1))[1:3, ] + c(0, 0, 0.35)
```

`new_matrix4` is one of the in-place geometry classes (`new_vector3`,
`new_matrix4`, `new_quaternion`); `to_array` hands back the plain `4x4` matrix.

Faces alone do not carry orientation. `vcg_update_normals` computes per-vertex
normals, which the smoothing, projection, and ray-casting routines all read:

```{r normals}
plane <- vcg_update_normals(plane, weight = "area")
dim(plane$normals)
```

## 2. Asking questions about geometry

Three functions answer "where is this thing relative to that thing", and they
trade accuracy against cost differently.

### Exact collision

`vcg_detect_collision` reports, for every element of `y`, whether it comes
within `radius` of `x`, together with the exact minimum distance. Either side
can be a point cloud, a chain of connected line segments, or a triangular mesh,
so one call covers all nine pairings. Both sides here are meshes, so answers
come back one per triangle of `y`:

```{r collision}
cut <- vcg_detect_collision(sphere, plane, radius = 0.02)
cut$summary$y$unit_type
sum(cut$hit_unit)
head(cut$representation)
```

`representation` has one row per hit: `unit` is the plane triangle, `index` is
the vertex of that triangle nearest the sphere, `distance` is the true minimum
distance, and `x_index` is the sphere face that was hit. Keeping those triangles
gives the intersection band as a mesh in its own right:

```{r band}
band <- plane
band$it <- plane$it[, which(cut$hit_unit), drop = FALSE]
band
```

The `radius` argument makes this a proximity test rather than a contact test.
Raising it thickens the band, because more triangles fall inside the tolerance:

```{r radius}
vapply(c(0.02, 0.1, 0.3), function(r) {
  sum(vcg_detect_collision(sphere, plane, radius = r)$hit_unit)
}, 0L)
```

Line segments are the case worth knowing about, because that is how diffusion
streamlines and electrode shafts arrive. Several chains share one matrix and
are delimited by rows of `NA`:

```{r streamlines}
streamlines <- rbind(
  cbind(seq(-3, 3, by = 0.5), 0, 0),
  c(NA, NA, NA),
  cbind(seq(-3, 3, by = 0.5), 5, 0)
)
res <- vcg_detect_collision(sphere, streamlines,
                            mode_y = "segments", radius = 0.1)
res$hit_unit
res$representation
```

The first chain runs through the sphere and the second passes well clear of it.
`test_level` controls how hard the scan works: `"element"` (the default)
measures every segment and reports the closest, `"unit"` stops at the first hit
inside each chain, and `"whole"` stops at the first hit anywhere and answers
only the yes-or-no question:

```{r test-level}
vcg_detect_collision(sphere, streamlines, mode_y = "segments",
                     radius = 0.1, test_level = "whole")$collide
```

Finally, `include_interior` decides whether geometry buried inside a closed `x`
counts as a collision even when it never approaches the surface. The center of
the sphere is one unit away from every point of it:

```{r interior}
center <- rbind(c(0, 0, 0))
vcg_detect_collision(sphere, center, radius = 0.1)$collide
vcg_detect_collision(sphere, center, radius = 0.1,
                     include_interior = TRUE)$collide
```

That test uses ray casting, so it needs `x` to be `watertight` with coherently
oriented faces -- see `vcg_fix_defects` below for repairing one that is not.

### Nearest neighbors

`vcg_kdtree_nearest` builds a `K-D` tree over `target` and, for each point of
`query`, returns the `k` closest target points and their distances. Unlike
`vcg_detect_collision` it measures to the nearest *vertex*, not to the surface,
so it is an approximation -- a cheaper one that is usually good enough, and the
right tool when the identity of the nearest vertex is what is wanted:

```{r kdtree}
kd <- vcg_kdtree_nearest(target = sphere, query = plane, k = 1)
str(kd)
range(kd$distance)
```

Asking for more than one neighbor widens both matrices to `k` columns, which is
how you get a local neighborhood to average over:

```{r kdtree-k}
kd3 <- vcg_kdtree_nearest(target = sphere, query = plane, k = 3)
head(kd3$index)
```

Both arguments accept either a mesh or a plain `n x 3` matrix, and 2D points
work as well as 3D.

### Ray casting

`vcg_raycaster` shoots a ray from each origin along each direction and reports
the first face it meets. Lift the plane clear of the sphere and drop a ray
straight down from every vertex:

```{r raycast}
above <- plane$vb
above[3, ] <- above[3, ] + 1.8
rays <- vcg_raycaster(sphere, ray_origin = above, ray_direction = c(0, 0, -1))
str(rays[c("has_intersection", "distance", "face_index")])
sum(rays$has_intersection)
```

Rays whose origin lies over the sphere's silhouette hit it; the rest report
`has_intersection = FALSE`, `NA` for `distance` and `face_index`, and an
`intersection` column that should be ignored. `both_sides = TRUE` also searches
backwards along the ray, and `max_distance` caps how far it travels.

The hits form a point cloud, which is a mesh with vertices and no faces:

```{r pierce}
pierce <- structure(
  list(vb = rays$intersection[, rays$has_intersection, drop = FALSE]),
  class = "mesh3d"
)
```

Putting the three answers side by side:

```{r query-figure, fig.height = 3.6}
col <- color_ramp_continuous(kd$distance[, 1],
                             cmap = c("#f2f2f2", "#7fa8c9", "#2c5f8a"))

graphics::par(mfrow = c(1, 2), mar = c(0.1, 0.1, 2.1, 0.1), cex.main = 0.95)
plot_mesh_polygon(
  list(sphere, plane, band),
  col = list("gray55", col, "#a33a3a"), alpha = c(0.5, 0.92, 1),
  eye = c(3.5, -4, 2.2), up = c(0, 0, 1), zoom = 1.15,
  shadow_color = "white", ambient_intensity = 0.55,
  main = "Distance; cut faces in red"
)
plot_mesh_polygon(
  list(sphere, pierce),
  col = list("gray72", "#a33a3a"), cex = 0.03,
  eye = c(3.5, -4, 2.2), up = c(0, 0, 1), zoom = 0.68,
  shadow_color = "white", ambient_intensity = 0.55,
  main = "Ray entry points"
)
```

## 3. Building a mesh from a volume

Real surfaces come from volumes. `vcg_isosurface` runs marching cubes over a
3D array and returns the triangulation, mapping `voxel` indices into `RAS`
through `vox_to_ras`:

```{r isosurface}
data("left_hippocampus_mask", package = "ravetools")
dim(left_hippocampus_mask)

raw_mesh <- vcg_isosurface(left_hippocampus_mask)
raw_mesh
```

`mesh_from_volume` wraps the same step together with re-sampling and smoothing,
which is convenient when the goal is a display surface rather than an exact
level set:

```{r mesh-from-volume}
smoothed <- mesh_from_volume(
  left_hippocampus_mask, output_format = "rgl", threshold = 0.5,
  remesh = TRUE, remesh_voxel_size = 1, smooth = TRUE, verbose = FALSE
)
smoothed
```

`mris_make_surfaces` is the fourth constructor. It deforms an existing surface
along the intensity gradient of a volume to produce paired white and `pial`
surfaces, the way a cortical reconstruction does.

## 4. Measuring and diagnosing

Before doing anything to a mesh it is worth asking what shape it is in.
`vcg_count_edge_defects` counts the two things that break most algorithms:

```{r defects}
vcg_count_edge_defects(raw_mesh)
```

Forty boundary edges means the marching-cubes output has holes, so it is not
closed. `vcg_mesh_volume` says so itself rather than quietly returning a
number that means nothing:

```{r measure}
vcg_mesh_volume(raw_mesh)
vcg_average_edge_length(raw_mesh)
vcg_max_edge_length(raw_mesh)
```

The two edge-length measures are the ones to check before any operation with a
length parameter, since they say what scale the mesh is sampled at.

`mris_curvature` returns the four standard curvature fields per vertex --
mean, `Gaussian`, and the two principal curvatures:

```{r curvature-fields}
curv <- mris_curvature(raw_mesh)
str(curv)
```

## 5. Repairing and re-meshing

`vcg_fix_defects` merges duplicate vertices, fills holes, and reorients faces,
and it reports what it did in an `info` attribute:

```{r fix}
mesh <- vcg_fix_defects(raw_mesh, verbose = FALSE)
info <- attr(mesh, "info")
info[c("boundary_edges_before", "boundary_edges_after",
       "holes_filled", "is_closed_manifold")]

# the same call that warned above, now on a closed surface
vcg_mesh_volume(mesh)
```

The mesh is a closed `manifold` now, so its volume is trustworthy and
`include_interior` collision tests will work on it. Center it for the plots that
follow:

```{r center}
mesh$vb[1:3, ] <- mesh$vb[1:3, ] - rowMeans(mesh$vb[1:3, ])
```

Four routines change the sampling. They differ in what they preserve:

```{r remesh}
uniform  <- vcg_uniform_remesh(mesh, voxel_size = 1, verbose = FALSE)
split    <- vcg_subdivision(mesh, method = "edge")
capped   <- vcg_subdivide_max_edge_length(mesh, max_edge_len = 0.8)
isotropic <- mris_remesh(mesh, target_edge_length = 1.5, verbose = FALSE)

data.frame(
  method = c("input", "vcg_uniform_remesh", "vcg_subdivision",
             "vcg_subdivide_max_edge_length", "mris_remesh"),
  vertices = c(ncol(mesh$vb), ncol(uniform$vb), ncol(split$vb),
               ncol(capped$vb), ncol(isotropic$vb)),
  avg_edge = round(vapply(list(mesh, uniform, split, capped, isotropic),
                          vcg_average_edge_length, 0), 3),
  max_edge = round(vapply(list(mesh, uniform, split, capped, isotropic),
                          vcg_max_edge_length, 0), 3)
)
```

`vcg_uniform_remesh` re-samples through a distance field on a `voxel` grid, so
it also repairs topology at the cost of exact geometry. `vcg_subdivision`
splits every edge, doubling resolution everywhere. `vcg_subdivide_max_edge_length`
splits only edges above a threshold, leaving well-sampled regions alone.
`mris_remesh` implements the isotropic re-meshing of `Botsch` and `Kobbelt`
(2003), driving every edge toward one target length, and is the one to reach for
when downstream code cares about triangle quality.

## 6. Smoothing and deforming

Two smoothers come from `vcglib` and one from the cortical-surface literature:

```{r smooth}
taubin <- vcg_smooth_explicit(mesh, type = "taubin", iteration = 10)
implicit <- vcg_smooth_implicit(mesh, lambda = 0.2, degree = 2)
fs_style <- mris_smooth(mesh, niterations = 20L)

vapply(list(mesh, taubin, implicit, fs_style), vcg_mesh_volume, 0)
```

`vcg_smooth_explicit` applies a per-vertex `Laplacian` step repeatedly;
`type = "taubin"` alternates a positive and a negative step so the surface does
not shrink. `vcg_smooth_implicit` solves for the smoothed positions in one
sparse solve, which is stable at much larger `lambda`. Both preserve the vertex
count; only the positions move.

`mris_inflate` goes further, flattening the folds while preserving total
surface area. It also returns the `sulc` depth field -- how far each vertex
traveled -- which is what makes an inflated surface readable:

```{r inflate}
inflated <- mris_inflate(fs_style, n_averages = 4L, niterations = 8L,
                         scale_brain = FALSE, verbose = FALSE)
names(inflated)
range(inflated$sulc)
```

`mris_sphere` continues the deformation all the way onto a sphere of
`target_radius`, which is the mapping that surface-based registration is built
on. How close it gets is worth checking, because the residual spread in vertex
radius is the honest measure of convergence:

```{r sphere-map}
spherical <- mris_sphere(fs_style, target_radius = 100, verbose = FALSE)
radius <- sqrt(colSums(spherical$vb[1:3, ]^2))
c(min = min(radius), max = max(radius), cv = stats::sd(radius) / mean(radius))
```

Under two percent here. Both `mris_inflate` and `mris_sphere` were designed for
cortical surfaces, where the input is large, smooth, and genuinely
sphere-like; a small closed structure such as this hippocampus maps only
approximately, and raising `niterations` past the default makes it worse rather
than better on such an input.

Mapped side by side with mean curvature, the first two stages show what
inflation preserves:

```{r inflate-figure, fig.height = 3.2}
curv <- mris_curvature(fs_style)
lim <- stats::quantile(abs(curv$mean), 0.95)
curv_col <- color_ramp_continuous(
  curv$mean, clim = c(-lim, lim),
  cmap = c("#2c5f8a", "#f2f2f2", "#a33a3a")
)

graphics::par(mfrow = c(1, 2), mar = c(0.1, 0.1, 2.1, 0.1))
plot_mesh_polygon(fs_style, col = curv_col, eye = c(0, 100, 30),
                  up = c(0, 0, 1), zoom = 1.1, main = "Mean curvature")
plot_mesh_polygon(inflated$mesh, col = curv_col, eye = c(0, 100, 30),
                  up = c(0, 0, 1), zoom = 1.1, main = "The same, inflated")
```

## 7. Subsetting and extracting

`vcg_subset_vertex` keeps the vertices a logical selector marks and the faces
whose corners all survive:

```{r subset}
selector <- mesh$vb[1, ] > 0
half <- vcg_subset_vertex(mesh, selector)
c(input = ncol(mesh$vb), kept = ncol(half$vb))
```

`vcg_mesh_patch` cuts a mesh along a closed loop of `waypoints`. They are
snapped to the nearest vertices, the loop between them is walked, and the result
is a two-element list: the enclosed patch and everything else.

```{r patch}
target <- vcg_uniform_remesh(vcg_sphere(), verbose = FALSE)
patches <- vcg_mesh_patch(target, waypoints = diag(1, 3))
vapply(patches, function(p) ncol(p$it), 0L)
```

```{r patch-figure, fig.height = 3.4}
graphics::par(mar = c(0.1, 0.1, 0.1, 0.1))
plot_mesh_polygon(patches, col = list("#a33a3a", "gray70"),
                  alpha = c(1, 0.55), eye = c(10, 10, 10), zoom = 1.2,
                  shadow_color = "white", ambient_intensity = 0.55)
```

`dijkstras_surface_distance` walks the mesh graph from a start vertex and
returns the `geodesic` distance to every other vertex; `surface_path` then reads
the shortest path to any target back out of that result. Note that it takes the
transposed matrices -- one row per vertex and one row per face:

```{r dijkstra}
dist <- dijkstras_surface_distance(
  positions = t(mesh$vb[1:3, ]),
  faces = t(mesh$it),
  start_node = 1,
  face_index_start = 1
)
path <- surface_path(dist, target_node = ncol(mesh$vb))
c(vertices_on_path = length(path$path), length = max(path$distance))
```

Because the walk follows edges, the answer depends on the sampling -- which is
the practical reason to run `mris_remesh` before measuring distances across a
surface.

## 8. Drawing

`plot_mesh_polygon` projects every triangle with an orthographic camera, shades
it by how directly it faces that camera, depth-sorts everything, and draws it in
a single `polygon` call. `plot_mesh_dotcloud` draws vertices as rim-lit dots
instead. `plot()` dispatches to whichever suits the mesh.

Both take a list of meshes and render them into one shared depth space, so
several surfaces compose correctly. For one mesh, `col` is a single color, a
character vector of one color per vertex, or any other vector, which is read as
a depth gradient; for a list of meshes, pass a list of those, one element per
mesh. `alpha` is one value per mesh. A mesh with no face matrix is drawn as one
small `vcg_sphere` per vertex, scaled by `cex` -- which is how the ray hits in
section 2 were rendered.

Two controls are worth knowing. `mesh_clipping` discards triangles whose normal
points along the camera ray, peeling the front cap off a closed surface; pair it
with `side = "both"` so the exposed back wall is drawn rather than culled in
turn. `clipping_plane` culls faces against arbitrary world-space planes instead
-- each a length-5 vector of a normal, a signed offset, and which half-space to
keep -- and `clipping_plane_enabled` exempts individual meshes from it, which is
how electrodes stay whole while the surface around them is cut away.

Scattering some probes inside the surface -- found with the same interior test
from section 2 -- shows why the exemption matters: cut the surface open and the
probes stay whole.

```{r clipping, fig.height = 3.4}
bbox <- apply(fs_style$vb[1:3, ], 1L, range)
set.seed(1)
candidates <- cbind(
  stats::runif(600, bbox[1, 1], bbox[2, 1]),
  stats::runif(600, bbox[1, 2], bbox[2, 2]),
  stats::runif(600, bbox[1, 3], bbox[2, 3])
)
inside <- vcg_detect_collision(fs_style, candidates,
                               include_interior = TRUE)$hit_unit %in% TRUE
probes <- structure(
  list(vb = t(candidates[inside, , drop = FALSE])),
  class = "mesh3d"
)
sum(inside)

eye <- c(0, 100, 30)
graphics::par(mfrow = c(1, 3), mar = c(0.1, 0.1, 2.1, 0.1), cex.main = 0.95)

plot_mesh_polygon(fs_style, col = "steelblue", eye = eye, up = c(0, 0, 1),
                  zoom = 1.1, main = "Whole surface")

plot_mesh_polygon(
  list(fs_style, probes),
  col = list("steelblue", "#a33a3a"), cex = 1.1,
  eye = eye, up = c(0, 0, 1), zoom = 1.1,
  clipping_plane = c(0, 1, 0, 0, -1),
  clipping_plane_enabled = c(TRUE, FALSE),
  main = "Cut open, probes exempt"
)

# `plot_mesh_dotcloud` has no `main`; add the title afterwards
plot_mesh_dotcloud(fs_style, col = "steelblue", eye = eye, up = c(0, 0, 1),
                   zoom = 1.1, cex = 0.45)
graphics::title(main = "plot_mesh_dotcloud")
```

If `rgl` is installed, `rgl_view` and `rgl_call` drive an interactive window
using the same meshes, and `rgl_plot_normals` draws the normal field:

```{r rgl, eval = FALSE}
rgl_view({
  rgl_call("shade3d", mesh, col = "steelblue")
  rgl_call("wire3d", mesh, col = "black")
})
```

```{r cleanup, include = FALSE}
graphics::par(oldpar)
```

## References

`Botsch`, M, and `Kobbelt`, L (2003). A `remeshing` approach to
`multiresolution` modeling. *Proceedings of the 2004
`Eurographics`/`ACM` `SIGGRAPH` Symposium on Geometry Processing*, 185-192.

`Fischl`, B, `Sereno`, MI, and Dale, AM (1999). Cortical surface-based analysis
II: inflation, flattening, and a surface-based coordinate system.
*`NeuroImage`*, 9(2), 195-207.

The `vcg_*` functions are built on
[`vcglib`](https://github.com/cnr-isti-vclab/vcglib) from the Visual Computing
Lab, `ISTI`.

---
title: "Global Time-Frequency Baseline Comparison"
author: "Zhengjia Wang"
date: "Date built: `r format(date())`"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## Objective

We want to sanity-check the global baseline results for time-frequency analysis. Our goal is to make sure the C++ implementation of function `ravetools::baseline_array` result in no difference (within floating error) than the naive R implementation. 

## Methods & Results

This document uses the `DemoSubject`. To start, let's clear the existing cache to make sure we start with a clean environment.

```{r, eval=FALSE}
raveio::clear_cached_files(quiet = TRUE)
```

To load power data in `RAVE`, we use the pre-built function `raveio::prepare_subject_power` to load electrode channel `14`, with `1` second before epoch onset and `2` seconds after onset (using default epoch).

```{r}
repo <- raveio::prepare_subject_power(
  subject = 'demo/DemoSubject',
  electrodes = 14,
  time_windows = c(-1, 2),
  verbose = FALSE
)
```

### Global Baseline via RAVE

`raveio::power_baseline` uses `ravetools::baseline_array` internally, with user-friendly parameters. Please use `?raveio::power_baseline` to check the documentation.

* `baseline_windows=c(-1,0)` means the baseline window is from `-1` to `0` seconds (`[-1,0]`, including time-points at both sides)
* `units=c("Frequency","Electrode")` means calculating baseline per frequency per electrode and across trials
* `method="percentage"` means calculating baseline using power percentage change

The resulting baseline data can be accessed via `repo$power$baselined`

*(Below is power percentage change)*

```{r, results='hide'}
raveio::power_baseline(repo, baseline_windows = c(-1,0), 
                       units = c("Frequency", "Electrode"),
                       method = "percentage")
power_percentage_change <- repo$power$baselined[]
```

*(Below is amplitude percentage change)*

```{r, results='hide'}
raveio::power_baseline(repo, baseline_windows = c(-1,0), 
                       units = c("Frequency", "Electrode"),
                       method = "sqrt_percentage")
amplitude_percentage_change <- repo$power$baselined[]
```

*(Below is decibel-unit change)*

```{r, results='hide'}
raveio::power_baseline(repo, baseline_windows = c(-1,0), 
                       units = c("Frequency", "Electrode"), 
                       method = "decibel")
decibel_change <- repo$power$baselined[]
```

*(Below is z-scored power)*

```{r, results='hide'}
raveio::power_baseline(repo, baseline_windows = c(-1,0), units = c("Frequency", "Electrode"), method = "zscore")
power_zscore <- repo$power$baselined[]
```

*(Below is z-scored amplitude)*

```{r, results='hide'}
raveio::power_baseline(repo, baseline_windows = c(-1,0), units = c("Frequency", "Electrode"), method = "sqrt_zscore")
amplitude_zscore <- repo$power$baselined[]
```

### Native implementation

For native implementation, let's get power data

```{r}
power <- repo$power$data_list$e_14
cat(
  "Margin names are: ",
  paste(names(dimnames(power)), collapse = ", "),
  "\nMargin dimensions are: ",
  paste(dim(power), collapse = " x ")
)
```

The variable `power` is a `Frequency x Time x Trial x Electrode` (4-mode) tensor array. 

The following script obtains the baseline part of the power data

```{r}
baseline_data <- subset(power, Time ~ Time >= -1 & Time <= 0)
cat(
  "Data in the baseline has margin dimensions: ",
  paste(dim(baseline_data), collapse = " x ")
)
```

The second margin dimension of `baseline_data` differs with `power`. This is because power has `r dim(power)[[2]]` time-points while baseline window only has `r dim(baseline_data)[[2]]`

To calculate **power percentage change**:

```{r}
# baseline per frequency
power_expected <- (power[] / rowMeans(baseline_data) - 1) * 100
```

The maximum relative difference (`power_percentage_change / power_expected - 1`) is:

```{r}
range(power_percentage_change / power_expected - 1)
```


Let's compare the rest baseline results:

**Amplitude percentage change**:

```{r}
amplitude_expected <- (sqrt(power[]) / rowMeans(sqrt(baseline_data)) - 1) * 100
range(amplitude_percentage_change / amplitude_expected - 1)
```

**Decibel change**:

```{r}
decibel_expected <- 10*log10(power[]) - rowMeans(10*log10(baseline_data))
range(decibel_change / decibel_expected - 1)
```

**Z-scored power**:

```{r}
power_zscore_expected <- (power[] - rowMeans(baseline_data)) / apply(baseline_data, 1, sd)
range(power_zscore / power_zscore_expected - 1)
```

**Z-scored amplitude**:

```{r}
amp_zscore_expected <- (sqrt(power[]) - rowMeans(sqrt(baseline_data))) / apply(sqrt(baseline_data), 1, sd)
range(amplitude_zscore / amp_zscore_expected - 1)
```

## Benchmark

Let's test the speed of running baseline using the entire 5 channels.

```{r}
repo <- raveio::prepare_subject_power(
  subject = 'demo/DemoSubject',
  electrodes = "13-16,24",
  time_windows = c(-1, 2),
  verbose = FALSE
)

# RAVE baseline on 5 electrode channels
system.time({
  raveio::power_baseline(repo, baseline_windows = c(-1,0), 
                       units = c("Frequency", "Electrode"),
                       method = "percentage")
})

# Naive implementation
system.time({
  lapply(repo$power$data_list, function(power) {
    baseline_data <- subset(power, Time ~ Time >= -1 & Time <= 0)
    power_expected <- (power[] / apply(baseline_data, 1, mean) - 1) * 100
    filearray::as_filearray(power_expected)
  })
})
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bpc.R
\name{plot.ravetools_bpc}
\alias{plot.ravetools_bpc}
\title{Plot Basis Profile Curve results}
\usage{
\method{plot}{ravetools_bpc}(x, ...)
}
\arguments{
\item{x}{a \code{ravetools_bpc} object.}

\item{...}{ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
S3 plot method for \code{ravetools_bpc} objects returned by
\code{\link{bpc}}. Draws a three-panel figure: the basis profile curves, the
subgroup-by-subgroup significance matrix \eqn{\Xi} ordered by assignment, and
the per-subgroup projection weights.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bpc.R
\name{bpc}
\alias{bpc}
\title{Basis Profile Curve (\verb{BPC}) identification}
\usage{
bpc(
  x,
  groups,
  time = NULL,
  time_window = NULL,
  n_bpc = NULL,
  initial_rank = NULL,
  zeta_threshold = 1,
  null_class = TRUE,
  nmf_max_iters = 10000,
  nmf_tol = c(1e-04, 1e-08),
  verbose = TRUE
)
}
\arguments{
\item{x}{numeric matrix of single-trial evoked voltages with shape
\code{time x trials} (rows are timepoints, columns are trials). All trials
across all stimulation groups are stacked column-wise.}

\item{groups}{vector of length \code{ncol(x)} (integer, character or factor)
giving the stimulation subgroup of each column of \code{x}. At least two
subgroups are required and each subgroup needs at least two trials.}

\item{time}{optional numeric vector of length \code{nrow(x)} giving the
stimulus-aligned time (in seconds) of each row of \code{x}. Only used to crop
to \code{time_window} and as the plotting axis; the method itself does not
require it.}

\item{time_window}{optional numeric \code{c(lo, hi)} analysis window (seconds);
requires \code{time}. An \code{NA} extends to the corresponding data edge.
\code{NULL} (the default) uses all rows.}

\item{n_bpc}{integer or \code{NULL}; if given, fixes the number of basis
curves and skips automatic rank selection.}

\item{initial_rank}{integer or \code{NULL}; the starting \verb{NMF} rank for
automatic selection. \code{NULL} uses \code{min(n - 1L, 10L)} where \eqn{n} is
the number of subgroups.}

\item{zeta_threshold}{numeric \eqn{> 0}; rank-selection cutoff on \eqn{\zeta}
(see \sQuote{Details}); smaller values give fewer curves. Defaults to \code{1},
matching the \verb{BPC} reference.}

\item{null_class}{logical; if \code{TRUE} (the default) subgroups whose
winner-take-all loading falls below \eqn{1/(2\sqrt{n})} are left unassigned
(\code{NA}); if \code{FALSE} every subgroup is forced into a curve.}

\item{nmf_max_iters, nmf_tol}{passed to \code{\link{naive_nmf}}.}

\item{verbose}{logical; whether to report progress.}
}
\value{
A named list of class \code{ravetools_bpc}:
\describe{
\item{\code{curves}}{Numeric matrix, time \eqn{\times} number of \verb{BPCs};
  column \eqn{q} is the basis profile curve \eqn{B_q(t)}, the first
  linear-kernel \verb{PCA} component of its member trials (unit-norm,
  sign-oriented to a positive mean projection).}
\item{\code{time}}{Numeric vector, the (windowed) time axis for \code{curves};
  the sample index when \code{time} was not supplied.}
\item{\code{group_labels}}{The unique stimulation subgroup labels, in the order
  used by \code{clusters} and \code{xi}.}
\item{\code{clusters}}{Integer vector, the \verb{BPC} index assigned to each
  subgroup; \code{NA} for subgroups left in the null class.}
\item{\code{excluded_groups}}{The labels of subgroups not represented by any
  \verb{BPC}.}
\item{\code{weights}}{A \code{data.frame} with one row per (\verb{BPC},
  subgroup) membership: \code{bpc}, \code{group}, \code{n_trials},
  \code{weight} (mean residual-normalized \eqn{\alpha}) and \code{p_value}.}
\item{\code{alpha}}{Numeric matrix, trials \eqn{\times} number of \verb{BPCs};
  the per-trial projection \eqn{\alpha} of every trial onto each basis curve.}
\item{\code{xi}}{Numeric matrix, the \eqn{n \times n} significance matrix
  \eqn{\Xi}.}
\item{\code{nmf}}{The \code{\link{naive_nmf}} result for the selected rank, with
  \code{H} (raw loadings) and \code{H0} (winner-take-all, thresholded).}
\item{\code{n_bpc}}{Integer, the number of basis curves found.}
\item{\code{time_window}}{The effective analysis window used (or \code{NULL}).}
}
}
\description{
Identifies a small set of canonical temporal response shapes - \emph{basis
profile curves} (\verb{BPCs}) - from the single-trial stimulation-evoked
responses recorded at one measurement electrode, where the trials are grouped
by stimulation site (or any other condition). Each stimulation group is
assigned to the \verb{BPC} that best explains its trials, and the projection
strength of every group is quantified. This is the across-stimulation-site
\verb{BPC} method (see \sQuote{References}); the related
\code{\link{crp_cluster}} applies the same idea across electrodes.
}
\details{
Let \eqn{V} be the (windowed) \eqn{T \times K} matrix of all single trials and
let \code{groups} map each of the \eqn{K} columns to one of \eqn{n} stimulation
subgroups.
\enumerate{
\item \strong{Window.} When \code{time} and \code{time_window} are supplied the
  rows of \code{x} are cropped to that window; otherwise all rows are used.
  Rows with non-finite values are dropped.
\item \strong{Internal projections.} Each trial is \eqn{L_2}-normalized into
  \eqn{V_0}, and \eqn{P = V_0^\top V} collects the projection of every native
  trial onto every normalized trial.
\item \strong{Significance matrix.} For each ordered pair of subgroups
  \eqn{(k, l)} the set of relevant entries of \eqn{P} is gathered (the
  off-diagonal of the within-group block when \eqn{k = l}, the whole cross
  block otherwise) and reduced to a one-sample t-statistic versus zero. These
  form the \eqn{n \times n} significance matrix \eqn{\Xi}.
\item \strong{Rank selection.} \eqn{\Xi} is made non-negative and rescaled, then
  factorized with \code{\link{naive_nmf}} at decreasing inner rank while the
  degeneracy score \eqn{\zeta} - the sum of the upper off-diagonal of the
  row-normalized \eqn{HH^\top} - exceeds \code{zeta_threshold}.
\item \strong{Assignment.} Each subgroup takes its winner-take-all \verb{BPC}
  over the normalized \verb{NMF} loadings; with \code{null_class}, loadings
  below \eqn{1/(2\sqrt{n})} are left unassigned (the \dQuote{null} class).
\item \strong{Basis curves.} Per \verb{BPC}, the first linear-kernel \verb{PCA}
  component of all trials in its member subgroups, sign-oriented to a positive
  mean projection.
\item \strong{Weights.} For each member subgroup the per-trial coefficient
  \eqn{\alpha} (projection onto the basis curve) is normalized by the residual
  magnitude; the mean is the projection weight and a one-sample t-test gives a
  significance value.
}
}
\examples{
# Three response shapes, several stimulation groups per shape.
\donttest{
set.seed(1)
n_time <- 300L
tt <- seq(-0.2, 1, length.out = n_time)
shapes <- list(
  exp(-((tt - 0.08) / 0.03)^2) - 0.5 * exp(-((tt - 0.18) / 0.04)^2),
  exp(-((tt - 0.38) / 0.03)^2) - 0.5 * exp(-((tt - 0.50) / 0.04)^2),
  exp(-((tt - 0.70) / 0.04)^2)
)

# 3 stimulation groups per shape, 12 trials each
V <- NULL
groups <- NULL
g <- 0L
for (s in seq_along(shapes)) {
  for (rep in seq_len(3L)) {
    g <- g + 1L
    trials <- outer(shapes[[s]], runif(12L, 0.5, 1.5)) +
      matrix(rnorm(n_time * 12L, sd = 0.2), n_time, 12L)
    V <- cbind(V, trials)
    groups <- c(groups, rep(g, 12L))
  }
}

res <- bpc(V, groups, time = tt, time_window = c(0, 1), verbose = TRUE)
res$n_bpc
res$clusters
plot(res)
}

}
\references{
The \verb{BPC} method is described in \doi{10.1371/journal.pcbi.1008710}, with
a reference Python implementation at
\url{https://github.com/MultimodalNeuroimagingLab/bpc_jupyter}.
}
\seealso{
\code{\link{crp}}, \code{\link{crp_cluster}}, \code{\link{naive_nmf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_kdtree_nearest}
\alias{vcg_kdtree_nearest}
\title{Find nearest \code{k} points}
\usage{
vcg_kdtree_nearest(target, query, k = 1, leaf_size = 16, max_depth = 64)
}
\arguments{
\item{target}{a matrix with \code{n} rows (number of points) and 2 or 3
columns, or a \code{mesh3d} object. This is the target point cloud where
nearest distances will be sought}

\item{query}{a matrix with \code{n} rows (number of points) and 2 or 3
columns, or a \code{mesh3d} object. This is the query point cloud where
for each point, the nearest \code{k} points in \code{target} will be sought.}

\item{k}{positive number of nearest neighbors to look for}

\item{leaf_size}{the suggested leaf size for the \code{K-D} tree; default is
\code{16}; larger leaf size will result in smaller depth}

\item{max_depth}{maximum depth of the \code{K-D} tree; default is \code{64}}
}
\value{
A list of two matrices: \code{index} is a matrix of indices of
\code{target} points, whose distances are close to the corresponding
\code{query} point. If no point in \code{target} is found, then \code{NA}
will be presented. Each \code{distance} is the corresponding distance
from the query point to the target point.
}
\description{
For each point in the query, find the nearest \code{k} points in target using
\code{K-D} tree.
}
\examples{

# Find nearest point in B with the smallest distance for each point in A

library(ravetools)

n <- 10
A <- matrix(rnorm(n * 2), nrow = n)
B <- matrix(rnorm(n * 4), nrow = n * 2)
result <- vcg_kdtree_nearest(
  target = B, query = A,
   k = 1
)

plot(
  rbind(A, B),
  pch = 20,
  col = c(rep("red", n), rep("black", n * 2)),
  xlab = "x",
  ylab = "y",
  main = "Black: target; Red: query"
)

nearest_points <- B[result$index, ]
arrows(A[, 1],
       A[, 2],
       nearest_points[, 1],
       nearest_points[, 2],
       col = "red",
       length = 0.1)

# ---- Sanity check ------------------------------------------------
nearest_index <- apply(A, 1, function(pt) {
  which.min(colSums((t(B) - pt) ^ 2))
})

result$index == nearest_index



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_peaks.R
\name{find_peaks}
\alias{find_peaks}
\title{Find peaks of a signal}
\usage{
find_peaks(x, min_val = NA, min_distance = 4, min_width = 2)
}
\arguments{
\item{x}{a numeric vector without missing values}

\item{min_val}{find peaks that are greater than this value}

\item{min_distance}{merge peaks that are less than \code{min_distance}
time-points away}

\item{min_width}{search radius (time-points) on whether the peak is
"local"; this is for seasonal oscillations.}
}
\value{
A list of peak index (1-based) and the corresponding value.
}
\description{
Find peaks of a signal
}
\examples{


# Basic example
x <- sin(seq(0, 10, 0.01)) + rnorm(1001) * 0.1

peaks <- find_peaks(x)

plot(x, type = 'l')
abline(v = peaks$index, col = 'red')

# merge peaks that are close
peaks <- find_peaks(x, min_distance = 400)

plot(x, type = 'l')
abline(v = peaks$index, col = 'red')

# with or without min_width
x <- c(0, 1, 0.5, 0.9, 0.2, 0.8, 0.2, 0.75, 0)

# without min_width
peaks <- find_peaks(x, min_width = 0)
plot(x, type = 'l')
abline(v = peaks$index, col = 'red')


# with min_width=2: t=4 is greater than t=6
peaks <- find_peaks(x, min_width = 2)
plot(x, type = 'l')
abline(v = peaks$index, col = 'red')


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwelch.R
\name{pwelch}
\alias{pwelch}
\alias{print.ravetools-pwelch}
\alias{plot.ravetools-pwelch}
\alias{mv_pwelch}
\title{Calculate 'Welch Periodogram'}
\usage{
pwelch(
  x,
  fs,
  window = 64,
  noverlap = window/2,
  nfft = "auto",
  window_family = hamming,
  col = "black",
  xlim = NULL,
  ylim = NULL,
  main = "Welch periodogram",
  plot = 0,
  log = c("xy", "", "x", "y"),
  ...
)

\method{print}{`ravetools-pwelch`}(x, ...)

\method{plot}{`ravetools-pwelch`}(
  x,
  log = c("xy", "x", "y", ""),
  se = FALSE,
  xticks,
  type = "l",
  add = FALSE,
  col = graphics::par("fg"),
  col.se = "orange",
  alpha.se = 0.5,
  lty = 1,
  lwd = 1,
  cex = 1,
  las = 1,
  main = "Welch periodogram",
  xlab,
  ylab,
  xlim = NULL,
  ylim = NULL,
  xaxs = "i",
  yaxs = "i",
  xline = 1.2 * cex,
  yline = 2 * cex,
  mar = c(2.6, 3.8, 2.1, 0.6) * (0.5 + cex/2),
  mgp = cex * c(2, 0.5, 0),
  tck = -0.02 * cex,
  grid = TRUE,
  ...
)

mv_pwelch(
  x,
  margin,
  fs,
  window = 64,
  noverlap = window/2,
  nfft = "auto",
  window_family = hamming
)
}
\arguments{
\item{x}{numerical vector or a row-major vector, signals.
If \code{x} is a matrix, then each row is a channel. For \code{plot}
function, \code{x} is the instance returned by \code{pwelch} function.}

\item{fs}{sample rate, average number of time points per second}

\item{window}{window length in time points, default size is \code{64}}

\item{noverlap}{overlap between two adjacent windows, measured in time
points; default is half of the \code{window}}

\item{nfft}{number of points in window function; default is automatically
determined from input data and window, scaled up to the nearest power of 2}

\item{window_family}{function generator for generating filter windows,
default is \code{\link{hamming}}. This can be any window function listed in
the filter window family, or any window generator function from package
\code{gsignal}. Default is \code{\link{hamming}}. For 'iEEG' users, both
\code{hamming} and \code{\link{blackmanharris}} are offered by 'EEG-lab';
while \code{blackmanharris} offers better attenuation than Hamming windows,
it also has lower spectral resolution. \code{hamming} has a 42.5 dB side-lobe
attenuation. This may mask spectral content below this value (relative
to the peak spectral content). Choosing different windows enables
you to make trade-off between resolution (e.g., using a rectangular
window) and side-lobe attenuation (e.g., using a \code{\link{hanning}}
window)}

\item{col, xlim, ylim, main, type, cex, las, xlab, ylab, lty, lwd, xaxs, yaxs, mar, mgp, tck}{parameters passed to \code{\link[graphics]{plot.default}}}

\item{plot}{integer, whether to plot the result or not; choices are \code{0}, no plot; \code{1} plot on a new canvas; \code{2} add to existing canvas}

\item{log}{indicates which axis should be \code{log10}-transformed, used by the plot function. For \code{'x'} axis, it's \code{log10}-transform; for \code{'y'} axis, it's \code{10log10}-transform (decibel unit). Choices are \code{"xy"}, \code{"x"}, \code{"y"}, and \code{""}.}

\item{...}{will be passed to \code{plot.pwelch} or ignored}

\item{se}{logical or a positive number indicating whether to plot standard
error of mean; default is false. If provided with a number, then a multiple
of standard error will be drawn. This option is only available when power
is in log-scale (decibel unit)}

\item{xticks}{ticks to show on frequency axis}

\item{add}{logical, whether the plot should be added to existing canvas}

\item{col.se, alpha.se}{controls the color and opacity of the standard error}

\item{xline, yline}{controls how close the axis labels to the corresponding axes}

\item{grid}{whether to draw rectangular grid lines to the plot; only
respected when \code{add=FALSE}; default is true}

\item{margin}{the margin in which \code{pwelch} should be applied to}
}
\value{
A list with class \code{'ravetools-pwelch'} that contains the
following items:
\describe{
\item{\code{freq}}{frequencies used to calculate the 'periodogram'}
\item{\code{spec}}{resulting spectral power for each frequency}
\item{\code{window}}{window function(in numerical vector) used}
\item{\code{noverlap}}{number of overlapping time-points between two adjacent windows}
\item{\code{nfft}}{number of basis functions}
\item{\code{fs}}{sample rate}
\item{\code{x_len}}{input signal length}
\item{\code{method}}{a character string \code{'Welch'}}
}
}
\description{
\code{pwelch} is for single signal trace only; \code{mv_pwelch}
is for multiple traces. Currently \code{mv_pwelch} is experimental and
should not be called directly.
}
\examples{

x <- rnorm(1000)
pwel <- pwelch(x, 100)
pwel

plot(pwel, log = "xy")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_fix_defects}
\alias{vcg_fix_defects}
\title{Detect and repair defects in a triangular surface mesh}
\usage{
vcg_fix_defects(
  mesh,
  merge_tolerance = NA,
  max_hole_size = 100L,
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.}

\item{merge_tolerance}{distance (in mesh units) below which vertices are
welded together; default is \code{NA}, in which case the tolerance is
derived automatically as \code{1e-4} times the mesh's average edge length.}

\item{max_hole_size}{maximum number of boundary edges of a hole that will
be triangulated (ear-cutting fill); holes larger than this threshold are
left untouched (and will be reported as remaining boundary edges in
\code{attr(..., "info")}).  Default is \code{100}.}

\item{verbose}{whether to print a short before/after diagnostic report;
default is \code{FALSE}.}
}
\value{
A repaired triangular mesh of class \code{'mesh3d'}, with an
additional attribute \code{"info"}, a named list reporting what was
found and changed: \code{boundary_edges_before/after},
\code{nonmanifold_edges_before/after}, \code{vertices_merged},
\code{merge_tolerance}, \code{holes_filled}, \code{is_oriented},
\code{is_orientable}, \code{normals_flipped_outward}, and
\code{is_closed_manifold} (\code{TRUE} when the repaired mesh is closed
and manifold, i.e. ready for \code{\link{mris_inflate}}).
}
\description{
Repairs common defects that prevent a mesh from being a closed, manifold,
genus-0 surface - a hard precondition of algorithms such as
\code{\link{mris_inflate}}.  Typical sources of such defects are surfaces
extracted from volumes via marching-cubes-style algorithms (e.g.
\code{\link{vcg_isosurface}}), which can leave behind small "cracks": isolated
boundary-edge loops bounding tiny holes that are not closed by simple
vertex-welding.
}
\details{
The repair pipeline applies, in order:
\enumerate{
  \item Remove degenerate and duplicate faces.
  \item Weld near-coincident vertices (closes cracks caused by duplicated
    vertices), using \code{merge_tolerance} or, by default, a distance
    derived from the mesh's average edge length.
  \item Triangulate ("ear-cut fill") any remaining small boundary loops --
    i.e. \emph{isolated edges} / genuine small holes that welding alone
    cannot close, up to \code{max_hole_size} edges.
  \item Remove unreferenced vertices.
  \item Re-orient all faces coherently (consistent winding order), and, if
    the result is a single watertight component, flip normals to point
    outward (this last step assumes the geometry is meant to be watertight).
}
}
\examples{
if (is_not_cran()) {

  mesh <- vcg_isosurface(left_hippocampus_mask)

  repaired <- vcg_fix_defects(mesh, verbose = TRUE)

  attr(repaired, "info")$is_closed_manifold

  # repaired mesh can now be inflated
  inflated <- mris_inflate(repaired, scale_brain = FALSE)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_max_edge_length}
\alias{vcg_max_edge_length}
\title{Maximum edge length of a triangular mesh}
\usage{
vcg_max_edge_length(mesh)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.}
}
\value{
A single numeric value: the maximum edge length, in mesh units.
}
\description{
Returns the length of the longest edge in the mesh.
}
\examples{
if (is_not_cran()) {

  sphere <- vcg_sphere()
  vcg_max_edge_length(sphere)

}

}
\seealso{
\code{\link{vcg_average_edge_length}},
  \code{\link{vcg_subdivide_max_edge_length}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_mesh_volume}
\alias{vcg_mesh_volume}
\title{Compute volume for manifold meshes}
\usage{
vcg_mesh_volume(mesh)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}}
}
\value{
The numeric volume of the mesh
}
\description{
Compute volume for manifold meshes
}
\examples{

# Initial mesh
mesh <- vcg_sphere()

vcg_mesh_volume(mesh)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_isosurface}
\alias{vcg_isosurface}
\title{Create surface mesh from 3D-array}
\usage{
vcg_isosurface(
  volume,
  threshold_lb = 0,
  threshold_ub = NA,
  vox_to_ras = diag(c(-1, -1, 1, 1))
)
}
\arguments{
\item{volume}{a volume or a mask volume}

\item{threshold_lb}{lower-bound threshold for creating the surface; default
is \code{0}}

\item{threshold_ub}{upper-bound threshold for creating the surface; default
is \code{NA} (no upper-bound)}

\item{vox_to_ras}{a \code{4x4} \code{'affine'} transform matrix indicating the
'voxel'-to-world transform.}
}
\value{
A triangular mesh of class \code{'mesh3d'}
}
\description{
Create surface from 3D-array using marching cubes algorithm
}
\examples{


if(is_not_cran()) {

library(ravetools)
data("left_hippocampus_mask")

mesh <- vcg_isosurface(left_hippocampus_mask)


rgl_view({

  rgl_call("mfrow3d", 1, 2)

  rgl_call("title3d", "Direct ISOSurface")
  rgl_call("shade3d", mesh, col = 2)

  rgl_call("next3d")
  rgl_call("title3d", "ISOSurface + Implicit Smooth")

  rgl_call("shade3d",
           vcg_smooth_implicit(mesh, degree = 2),
           col = 3)
})

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crp_cluster.R
\name{plot.ravetools_crp_cluster}
\alias{plot.ravetools_crp_cluster}
\title{Plot electrode clustering results}
\usage{
\method{plot}{ravetools_crp_cluster}(x, ...)
}
\arguments{
\item{x}{a \code{ravetools_crp_cluster} object.}

\item{...}{ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
S3 plot method for \code{ravetools_crp_cluster} objects returned
by \code{\link{crp_cluster}}. Draws a two-panel figure: the per-cluster basis
profile curves, and the electrode-by-electrode similarity matrix ordered by
cluster.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-fill-surface.R
\name{fill_surface}
\alias{fill_surface}
\title{Fill a volume cube based on water-tight surface}
\usage{
fill_surface(
  surface,
  inflate = 0,
  close_radius = NULL,
  resolution = 256L,
  IJK2RAS = NULL,
  preview = FALSE,
  preview_frame = 128
)
}
\arguments{
\item{surface}{a surface mesh; accepted classes include \code{'mesh3d'}
(from \code{rgl}), \code{'fs.surface'} (from \code{freesurferformats}),
\code{'ieegio_surface'} (from \pkg{ieegio}), or a bare list containing a
\code{vb} vertex matrix; see \code{\link{ensure_mesh3d}} for details and
for how \code{'ieegio_surface'} inputs are coerced}

\item{inflate}{amount of 'voxels' to inflate on the final result; must be
a non-negative integer. A zero \code{inflate} value means the resulting
volume is tightly close to the surface}

\item{close_radius}{radius (in 'voxels' along the finest 'IJK' axis; see
\code{IJK2RAS}) of the morphological closing (dilate-then-erode) operation
used to connect the embedded surface into a water-tight shell \emph{and} to
remove small topological handles/tunnels (for example a hole drilled
through an otherwise solid structure) before the surface is filled; default
is \code{NULL}, which derives a radius from the surface geometry (capped at
\code{15} 'voxels'). Handles/tunnels wider than roughly twice this radius
(in physical, \verb{RAS} space) will not be closed, so when the input
surface is known to contain larger topological defects, set this to a
larger number explicitly. Larger radii cost more time and memory (the
closing operates on a dense \code{resolution^3} volume) and - because the
operation is performed in a fixed-size volume - must leave enough margin
that the dilated shell does not reach the volume boundary; radii that are
too large for the working cube will be capped with a warning}

\item{resolution}{number of 'voxels' along each margin of the working
cube volume that the surface is embedded into; default is \code{256}.
Larger values increase the spatial precision of the fill and the closing
operation (at the cost of memory and computation time, which scale with
\code{resolution^3}); the value must leave enough room for the surface
(transformed into 'IJK' space via \code{IJK2RAS}) plus the closing margin
to fit inside the cube}

\item{IJK2RAS}{volume 'IJK' (zero-indexed coordinate index) to
\code{'tkrRAS'} transform, default is automatically determined
leave it `NULL` if you don't know how to set it. When the linear part of
\code{IJK2RAS} encodes voxel sizes that differ across axes (\code{pixdim},
i.e. the three 'IJK' axes do not advance by the same physical distance per
'voxel'), the dilate/erode operations underlying \code{close_radius} and
\code{inflate} automatically use a different number of 'voxels' along each
axis so that the resulting structuring element stays isotropic in physical
(\verb{RAS}) space}

\item{preview}{whether to preview the results; default is false}

\item{preview_frame}{integer from 1 to 256 the depth frame used to generate
preview.}
}
\value{
A list containing the filled volume and parameters used to generate
the volume
}
\description{
Create a cube volume (\code{256} 'voxels' on each margin), fill
in the 'voxels' that are inside of the surface.
}
\details{
This function creates a volume (256 on each margin) and fill in
the volume from a surface mesh. The surface vertex points will be embedded
into the volume first. These points may not be connected together, hence
for each 'voxel', a cube patch will be applied to grow the volume. Then,
the volume will be bucket-filled from a corner, forming a negated mask of
"outside-of-surface" area. The inverted bucket-filled volume is then shrunk
so the mask boundary tightly fits the surface
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

\donttest{

# takes > 5s to run example

# Generate a sphere
surface <- vcg_sphere()
surface$vb[1:3, ] <- surface$vb[1:3, ] * 50

fill_surface(surface, preview = TRUE)

}

}
\seealso{
\code{\link{ensure_mesh3d}}
}
\author{
Zhengjia Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter.R
\name{design_filter_iir}
\alias{design_filter_iir}
\title{Design an 'IIR' filter}
\usage{
design_filter_iir(
  method = c("butter", "cheby1", "cheby2", "ellip"),
  sample_rate,
  filter_order = NA,
  use_sos = TRUE,
  high_pass_freq = NA,
  high_pass_trans_freq = NA,
  low_pass_freq = NA,
  low_pass_trans_freq = NA,
  passband_ripple = 0.1,
  stopband_attenuation = 40
)
}
\arguments{
\item{method}{filter method name, choices are \code{"butter"},
\code{"cheby1"}, \code{"cheby2"}, and \code{"ellip"}}

\item{sample_rate}{sampling frequency}

\item{filter_order}{suggested filter order. When \code{use_sos=TRUE}
(default) this order is used as-is since 'SOS' form is numerically stable
at any order. When \code{use_sos=FALSE} and 'ARMA' form becomes numerically
unstable, the order will be automatically reduced with a warning;
leave \code{NA} (default) to let the function choose automatically.}

\item{use_sos}{logical; when \code{TRUE} (default), the filter is also
stored as a \code{Sos} (second-order sections) object in \code{filter$sos},
and no stability-based order correction is performed. When \code{FALSE},
the order is capped at the maximum stable 'ARMA' order (with a warning if
a specific \code{filter_order} was requested).}

\item{high_pass_freq}{high-pass frequency; default is \code{NA} (no high-pass
filter will be applied)}

\item{high_pass_trans_freq}{high-pass frequency band-width; default
is automatically inferred from filter type.}

\item{low_pass_freq}{low-pass frequency; default is \code{NA} (no
low-pass filter will be applied)}

\item{low_pass_trans_freq}{low-pass frequency band-width; default
is automatically inferred from filter type.}

\item{passband_ripple}{allowable pass-band ripple in decibel; default is
\code{0.1}}

\item{stopband_attenuation}{minimum stop-band attenuation (in decibel) at
transition frequency; default is \code{40} dB.}
}
\value{
A filter object with \code{$b}/\code{$a} 'ARMA' coefficients and
\code{$sos} second-order sections (a \code{gsignal} \code{Sos} object).
}
\description{
Design an 'IIR' filter
}
\examples{

sample_rate <- 500

my_diagnose <- function(
    filter, vlines = c(8, 12), cutoffs = c(-3, -6)) {
  diagnose_filter(
    b = filter$b,
    a = filter$a,
    fs = sample_rate,
    vlines = vlines,
    cutoffs = cutoffs
  )
}

# ---- Default using butterworth to generate 8-12 bandpass filter ----

# Butterworth filter with cut-off frequency
# 7 ~ 13 (default transition bandwidth is 1Hz) at -3 dB
filter <- design_filter_iir(
  method = "butter",
  low_pass_freq = 12,
  high_pass_freq = 8,
  sample_rate = 500
)

filter

my_diagnose(filter)

## explicit bandwidths and attenuation (sharper transition)

# Butterworth filter with cut-off frequency
# passband ripple is 0.5 dB (8-12 Hz)
# stopband attenuation is 40 dB (5-18 Hz)
filter <- design_filter_iir(
  method = "butter",
  low_pass_freq = 12, low_pass_trans_freq = 6,
  high_pass_freq = 8, high_pass_trans_freq = 3,
  sample_rate = 500,
  passband_ripple = 0.5,
  stopband_attenuation = 40
)

filter

my_diagnose(filter)

# ---- cheby1 --------------------------------

filter <- design_filter_iir(
  method = "cheby1",
  low_pass_freq = 12,
  high_pass_freq = 8,
  sample_rate = 500
)

my_diagnose(filter)

# ---- cheby2 --------------------------------

filter <- design_filter_iir(
  method = "cheby2",
  low_pass_freq = 12,
  high_pass_freq = 8,
  sample_rate = 500
)

my_diagnose(filter)

# ----- ellip ---------------------------------

filter <- design_filter_iir(
  method = "ellip",
  low_pass_freq = 12,
  high_pass_freq = 8,
  sample_rate = 500
)

my_diagnose(filter)




}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-utils.R
\name{check_filter}
\alias{check_filter}
\title{Check 'Arma' filter}
\usage{
check_filter(b, a, w = NULL, r_expected = NULL, fs = NULL)
}
\arguments{
\item{b}{moving average (\code{MA}) polynomial coefficients.}

\item{a}{auto-regressive (\code{AR}) polynomial coefficients.}

\item{w}{normalized frequency, ranging from 0 to 1, where 1 is 'Nyquist'}

\item{r_expected}{attenuation in decibel of each \code{w}}

\item{fs}{sample rate, used to infer the frequencies and formatting print
message, not used in calculation; leave it blank by default}
}
\value{
A list of power estimation and the reciprocal condition number
of the \code{AR} coefficients.
}
\description{
Check 'Arma' filter
}
\examples{


# create a butterworth filter with -3dB (half-power) at [1, 5] Hz
# and -60dB stop-band attenuation at [0.5, 6] Hz

sample_rate <- 20
nyquist <- sample_rate / 2

specs <- buttord(
  Wp = c(1, 5) / nyquist,
  Ws = c(0.5, 6) / nyquist,
  Rp = 3,
  Rs = 60
)
filter <- butter(specs)

# filter quality is poor because the AR-coefficients
# creates singular matrix with unstable inverse,
# this will cause `filtfilt` to fail
check_filter(
  b = filter$b, a = filter$a,

  # frequencies (normalized) where power is evaluated
  w = c(1, 5, 0.5, 6) / nyquist,

  # expected power
  r_expected = c(3, 3, 60, 60)

)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crp.R
\name{crp}
\alias{crp}
\title{Canonical Response Parameterization (\verb{CRP})}
\usage{
crp(
  x,
  time,
  t_start = 0.015,
  t_end = 1,
  remove_artifacts = TRUE,
  artifact_interval = c("full", "tR"),
  artifact_p_threshold = 1e-05,
  threshold_quantile = 0.98,
  time_step = 5L,
  detect_onset = FALSE,
  onset_search_start = NULL
)
}
\arguments{
\item{x}{numeric matrix of single-trial evoked voltages with shape
\code{time x trials} (i.e. rows are timepoints, columns are trials);
the matrix orientation matches the variable \code{V} / \code{data} in
the MATLAB reference. At least two trials are required.}

\item{time}{numeric vector of length \code{nrow(x)} giving the
stimulus-aligned time (in seconds) of each row of \code{x}; must be
monotonically increasing and span \code{[t_start, t_end]}.}

\item{t_start, t_end}{numeric scalars, post-stimulation start and end times
(in seconds) defining the analysis window. Defaults match the MATLAB
illustration (\code{0.015 s} to \code{1 s}).}

\item{remove_artifacts}{logical; if \code{TRUE} (the default), an initial
\verb{CRP} pass is run to identify outlier/artifact trials, which are then
dropped before the final pass. See \sQuote{Details}.}

\item{artifact_interval}{character, one of \code{"full"} (the default,
matching the active option in the MATLAB illustration) or \code{"tR"};
selects whether per-trial outlier statistics are computed on the
projection magnitudes for the full window or only at the response
duration \eqn{\tau_R}.}

\item{artifact_p_threshold}{numeric, p-value threshold below which a
trial is flagged as artifact (provided its mean projection is also
below the cohort mean); defaults to \code{1e-5}.}

\item{threshold_quantile}{numeric in \code{(0, 1)}; the fraction of the
peak mean projection magnitude used to derive the duration-uncertainty
bounds \code{tau_R_lower} and \code{tau_R_upper}. Defaults to
\code{0.98} as in the manuscript.}

\item{time_step}{integer, sampling step (in samples) used when sweeping
candidate response duration; defaults to \code{5L}, matching the
MATLAB \code{t_step}. Larger values are faster but smooth the
projection profile.}

\item{detect_onset}{logical; if \code{TRUE}, estimate the response
\emph{onset} \eqn{\tau_{onset}} via a reverse cumulative-projection scan
(see \sQuote{Details}). The forward canonical shape and per-trial weights
are left unchanged; only the \emph{reported} canonical shape \code{C} is
restricted to the active response support \eqn{[\tau_{onset}, \tau_R]}.
The onset may fall before \code{t_start} (down to \code{onset_search_start}).
Defaults to \code{FALSE}.}

\item{onset_search_start}{numeric scalar or \code{NULL}; the earliest time
(in seconds) the onset scan may reach when \code{detect_onset = TRUE}.
\code{NULL} (the default) uses the first loaded time point, so the scan can
look across the entire signal before \eqn{\tau_R}. Raise it to keep the scan
away from an early stimulation artifact. Clamped into
\code{[min(time), t_end]}.}
}
\value{
A named list with the following elements:
\describe{
\item{\code{parameters}}{A list of single-trial parameterizations
(\code{crp_parms} in MATLAB):
\describe{
  \item{\code{C}}{Numeric vector, the reported canonical response shape
    \eqn{C(t)}, taken as the slice of \code{C_full} over its active support
    (so it shares identical values with \code{C_full}). By default this is
    \eqn{[t_{start}, \tau_R]}, where it equals the first eigenvector of the
    linear kernel PCA on \code{V_tR} (oriented to the mean trace, unit-norm,
    length \eqn{T_R}). When \code{detect_onset = TRUE} it is restricted to
    \eqn{[\tau_{onset}, \tau_R]} (so its norm is then \eqn{\le 1}). The
    matching time axis is \code{params_times}.}
  \item{\code{C_full}}{Numeric vector spanning the \emph{entire} loaded
    time range (every row of \code{time}, including any baseline at
    \eqn{t < 0}). Obtained by applying the trial-space \code{loading} to the
    full retained data, so \code{C_full} coincides exactly with the
    untrimmed forward \code{C} on \eqn{[t_{start}, \tau_R]} and extrapolates
    the shape everywhere else. Time axis is \code{params_times_full}.
    Outside \eqn{[t_{start}, \tau_R]} cross-trial consistency is not
    optimized, so those portions are more variable.}
  \item{\code{loading}}{Numeric vector of length \eqn{K} (retained
    trials), the trial-space loading \eqn{g = V_{tR}^\top C = s_1 v_1}
    recovered from the forward decomposition. Applying it to a
    time \eqn{\times} trials matrix and dividing by \eqn{\|g\|^2}
    reconstructs the canonical shape over that time range; this is how
    \code{C_full} is formed.}
  \item{\code{al}}{Numeric vector of length \eqn{K} (number of trials),
    the per-trial alpha coefficient \eqn{\alpha_k = C^\top V_k}: scalar
    projection of trial \eqn{k} onto \eqn{C(t)}. Larger magnitude means
    the trial resembles the canonical shape more strongly; sign reflects
    polarity relative to \eqn{C}.}
  \item{\code{al_p}}{Numeric vector of length \eqn{K}, alpha-prime
    \eqn{\alpha_k / \sqrt{T_R}}: \code{al} rescaled to remove the
    duration dependence from the unit-norm convention on \eqn{C}.
    Expressed in \eqn{\mu V} and comparable across electrodes or
    conditions with different \eqn{\tau_R}.}
  \item{\code{ep}}{Numeric matrix of shape \eqn{T_R \times K}, the
    per-trial residual \eqn{\epsilon_k(t) = V_k(t) - \alpha_k C(t)}
    after the shared component is removed, computed over the forward window
    \eqn{[t_{start}, \tau_R]} against the untrimmed forward canonical shape.
    Access trial \eqn{k} via \code{ep[, k]}. (When
    \code{detect_onset = TRUE} the reported \code{C} may be shorter than
    \eqn{T_R}; \code{ep} always stays on the full forward window.)}
  \item{\code{epep_root}}{Numeric vector of length \eqn{K},
    \eqn{\|\epsilon_k\| = \sqrt{\epsilon_k^\top \epsilon_k}}: L2 norm
    of the residual per trial. Smaller values indicate the canonical
    shape describes that trial more faithfully.}
  \item{\code{Vsnr}}{Numeric vector of length \eqn{K}, per-trial
    signal-to-noise \eqn{\alpha_k / \|\epsilon_k\|}. Values \eqn{> 1}
    indicate the canonical component is larger than the residual.}
  \item{\code{expl_var}}{Numeric vector of length \eqn{K}, per-trial
    explained variance \eqn{1 - \|\epsilon_k\|^2 / \|V_k\|^2}: fraction
    of each trial's energy accounted for by \eqn{\alpha_k C(t)}.
    Ranges in \eqn{[0, 1]}.}
  \item{\code{tR}}{Numeric scalar, response duration \eqn{\tau_R} in
    seconds: the time at which mean cross-trial projection magnitude is
    maximized.}
  \item{\code{params_times}}{Numeric vector, time axis for the reported
    \code{C}; length matches \code{C} (\eqn{T_R}, or the onset-trimmed
    support when \code{detect_onset = TRUE}).}
  \item{\code{params_times_full}}{Numeric vector, time axis for
    \code{C_full}; equals the full loaded \code{time} (every row,
    including the baseline at \eqn{t < 0}).}
  \item{\code{V_tR}}{Numeric matrix \eqn{T_R \times K}, trial matrix
    truncated to \eqn{\tau_R} - the data actually decomposed. Together
    with \code{ep} and \code{avg_trace_tR} it always spans the full
    forward window \eqn{[t_{start}, \tau_R]}, independently of
    \code{detect_onset}.}
  \item{\code{avg_trace_tR}}{Numeric vector of length \eqn{T_R}, simple
    trial average truncated to \eqn{\tau_R}.}
}}
\item{\code{projections}}{A list of projection-stage outputs
(\code{crp_projs} in MATLAB):
\describe{
  \item{\code{proj_tpts}}{Numeric vector, candidate duration time points
    (seconds) at which projection magnitudes were evaluated.}
  \item{\code{S_all}}{Numeric matrix; rows are non-redundant off-diagonal
    trial-pair projections, columns correspond to \code{proj_tpts}.
    Units: \eqn{\mu V \cdot s^{1/2}}.}
  \item{\code{mean_proj_profile}}{Numeric vector, mean of \code{S_all}
    across trial pairs at each candidate duration, the profile whose
    maximum defines \eqn{\tau_R}.}
  \item{\code{var_proj_profile}}{Numeric vector, variance of
    \code{S_all} across trial pairs at each candidate duration.}
  \item{\code{tR_index}}{Integer, column index into \code{S_all} and
    \code{proj_tpts} corresponding to \eqn{\tau_R}.}
  \item{\code{tR_sample}}{Integer, row (sample) index of \eqn{\tau_R}
    within the windowed data; \code{V[seq_len(tR_sample), ]} is the
    truncated trial matrix \code{V_tR}.}
  \item{\code{avg_trace_input}}{Numeric vector, simple trial average
    over the full analysis window (not truncated to \eqn{\tau_R}).}
  \item{\code{stat_indices}}{Integer vector, row indices of \code{S_all}
    used for the significance t-tests, constructed so each trial-pair
    comparison appears at most once.}
  \item{\code{t_value_tR}, \code{p_value_tR}}{t-statistic and one-sided
    p-value (H1: mean projection \eqn{> 0}) at \eqn{\tau_R}. Primary
    extraction-significance test reported in the manuscript.}
  \item{\code{t_value_full}, \code{p_value_full}}{Same test at the full
    analysis-window duration.}
}}
\item{\code{bad_trials}}{Integer vector of column indices into the
original \code{x} flagged and removed as artifacts; \code{integer(0)}
when none removed or when \code{remove_artifacts = FALSE}.}
\item{\code{tau_R}}{Numeric scalar, estimated response duration
\eqn{\tau_R} in seconds (convenience copy of \code{parameters$tR}).}
\item{\code{tau_R_lower}, \code{tau_R_upper}}{Numeric scalars, lower
and upper threshold-crossing times (seconds) bracketing \eqn{\tau_R}
at the \code{threshold_quantile} fraction of the peak mean projection
magnitude.}
\item{\code{tau_onset}, \code{tau_onset_lower},
\code{tau_onset_upper}}{Numeric scalars, the estimated response onset
time (seconds) and its lower/upper threshold-crossing bounds, from the
reverse projection scan; all \code{NA} unless \code{detect_onset = TRUE}.}
\item{\code{onset}}{List with the reverse-scan profile (\code{onset_tpts}
and \code{mean_proj_profile}, mapped onto original time); \code{NULL}
unless \code{detect_onset = TRUE} (also \code{NULL} when the search window
had too few samples).}
\item{\code{t_start}, \code{t_end}}{The analysis window used.}
\item{\code{sample_rate}}{Numeric, sampling rate inferred from
\code{time}.}
}
}
\description{
Parameterizes single-trial evoked responses (e.g. cortico-cortical evoked
potentials, \verb{CCEPs}) using the Canonical Response Parameterization
method (see 'Citation'). The function estimates the response
duration \eqn{\tau_R}, the time after stimulus at which the evoked
response has its most consistent, shared structure across trials.
The estimator is obtained from the time course of cross-trial projection
magnitudes, extracts the canonical response shape \eqn{C(t)} via a linear
kernel-trick PCA on the trial matrix truncated at \eqn{\tau_R}, and
reports per-trial weights, residuals, signal-to-noise, explained variance
and extraction-significance statistics.

This is an R translation of \code{CRP_method.m} (and the surrounding
artifact-rejection / duration-uncertainty logic in
\code{CRP_illustration.m}) from the upstream MATLAB reference
implementation; see \sQuote{References}.
}
\details{
Briefly, the algorithm proceeds in three stages:
\enumerate{
\item For a sweep of candidate durations \eqn{k}, compute pairwise
L2-normalized cross-projection magnitudes between trials truncated to
\eqn{[0, k]}. The duration that maximizes the mean projection magnitude
is taken as the response duration \eqn{\tau_R}.
\item Apply linear kernel-trick PCA to the trial matrix truncated to
\eqn{\tau_R}; the first principal component is the canonical response
shape \eqn{C(t)}.
\item Project \eqn{C(t)} into each trial to obtain per-trial weights
\eqn{\alpha_k}; the residual \eqn{\epsilon_k = V_k - \alpha_k C}
summarizes trial-by-trial deviation from the canonical shape.
}
Significance is assessed by a one-sided t-test on the off-diagonal
projection magnitudes against zero, restricted to a non-overlapping
subset of comparison pairs to avoid double-counting.

When \code{remove_artifacts = TRUE}, the function performs an initial
\verb{CRP} pass and runs an unpaired t-test for each trial comparing the
projections it participates in against all other off-diagonal
projections. Trials with \code{p < artifact_p_threshold} \emph{and}
mean projection below the cohort mean are dropped, and \verb{CRP} is re-run.

When \code{detect_onset = TRUE}, a complementary reverse pass estimates the
response \emph{onset}. The retained trials between \code{onset_search_start}
and \eqn{\tau_R} are time-reversed, and the same cumulative cross-projection
profile is computed growing backward from \eqn{\tau_R}. The backward
duration that maximizes cross-trial consistency marks \eqn{\tau_{onset}},
the time at which trials begin to share structure - useful when the response
is delayed by an unknown latency. Because the scan can extend before
\code{t_start}, the onset may be earlier than the analysis window. This pass
only locates a time: it re-uses the forward loading, so the canonical shape
and per-trial weights are unchanged (the reported \code{C} is merely sliced
to \eqn{[\tau_{onset}, \tau_R]}).

Time points whose data are not finite (any \code{NA}, \code{NaN} or
\code{Inf} across trials) are dropped before analysis. \code{t_start},
\code{t_end} and \code{onset_search_start} are clamped into the available
time range rather than triggering an error.
}
\examples{
set.seed(42)

# Synthetic CCEP-like data: shared canonical shape with per-trial scaling
n_time   <- 500L
n_trials <- 20L
tt <- seq(-0.5, 1, length.out = n_time)
canonical <- exp(-((tt - 0.10) / 0.05)^2) -
             0.5 * exp(-((tt - 0.30) / 0.10)^2)
V <- (outer(canonical, runif(n_trials, 0.5, 1.5)) +
       matrix(rnorm(n_time * n_trials, sd = 0.3), n_time, n_trials)) * 2

res <- crp(V, tt)

op <- par(mfrow = c(1, 3), mar = c(4.5, 4, 3, 1))
on.exit({ par(op) })

# ---- Panel 1: all trials (full window) + mean + C(t) overlay ----------
parms <- res$parameters
matplot(tt, V, type = "l", lty = 1,
        col = "#80808060", xlab = "Time (s)",
        ylab = expression(mu * V),
        main = expression("Canonical shape " * C(t)))
# scale C(t) to the amplitude of the mean trace for overlay;
# C(t) ends at tau_R so the line is cut off there naturally
C_scaled <- parms$C * max(abs(rowMeans(V))) / max(abs(parms$C))
lines(parms$params_times, C_scaled, col = "#FFFF0080", lwd = 3)

# Mean
lines(tt, rowMeans(V), col = "black", lwd = 1)
legend("topright", c("mean", "C(t) scaled"),
       col = c("black", "#FFFF00"), lty = c(1, 2), lwd = 2,
       bty = "n", cex = 0.8)

# ---- Panel 2: per-trial alpha-prime weights -----------------------------
barplot(sort(parms$al_p), col = "steelblue", border = NA, las = 1,
        xlab = "Trial (sorted)",
        ylab = expression(alpha * "'" ~ (mu * V)),
        main = expression("Per-trial " * alpha * "' (alpha-prime)"))
abline(h = c(0, mean(parms$al_p)), lty = 2)

# ---- Panel 3: mean projection profile with tau_R bounds ----------------
proj <- res$projections
plot(proj$proj_tpts, proj$mean_proj_profile, type = "l", lwd = 2,
     xlab = "Candidate duration (s)",
     ylab = expression(bar(S) ~ (mu * V \%.\% s^{0.5})),
     main = expression("Projection profile & " * tau[R]),
     las = 1)
abline(v = c(res$tau_R_lower, res$tau_R, res$tau_R_upper),
       col = c("cyan3", "orange2", "red"),
       lty = c(2, 1, 2), lwd = 2)
legend("topright",
       legend = expression(tau[lb], tau[R], tau[ub]),
       col = c("cyan3", "orange2", "red"),
       lty = c(2, 1, 2), lwd = 2, bty = "n")

par(op)

# ---- Onset detection (reverse scan) on the same data -------------------
res_onset <- crp(V, tt, detect_onset = TRUE)
c(tau_onset = res_onset$tau_onset, tau_R = res_onset$tau_R)

}
\references{
The \verb{CRP} algorithm is described in \doi{10.1371/journal.pcbi.1011105},
with a reference MATLAB implementation at
\url{https://github.com/kaijmiller/crp_scripts}. See
\code{citation("ravetools")} for the full bibliographic entry.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-signals.R
\name{plot_signals}
\alias{plot_signals}
\title{Plot one or more signal traces in the same figure}
\usage{
plot_signals(
  signals,
  sample_rate = 1,
  col = graphics::par("fg"),
  space = 0.995,
  space_mode = c("quantile", "absolute"),
  start_time = 0,
  duration = NULL,
  compress = TRUE,
  channel_names = NULL,
  time_shift = 0,
  xlab = "Time (s)",
  ylab = "Electrode",
  lwd = 0.5,
  new_plot = TRUE,
  xlim = NULL,
  cex = 1,
  cex.lab = 1,
  mar = c(3.1, 2.1, 2.1, 0.8) * (0.25 + cex * 0.75) + 0.1,
  mgp = cex * c(2, 0.5, 0),
  xaxs = "r",
  yaxs = "i",
  xline = 1.5 * cex,
  yline = 1 * cex,
  tck = -0.005 * (3 + cex),
  ...
)
}
\arguments{
\item{signals}{numerical matrix with each row to be a signal trace and each
column contains the signal values at a time point}

\item{sample_rate}{sampling frequency}

\item{col}{signal color, can be vector of one or more}

\item{space}{vertical spacing among the traces; for values greater than 1,
the spacing is absolute; default is \code{0.995}; for values less equal to 1,
this is the percentile of the whole data. However, the quantile mode can be
manually turned off is \code{"absolute"} is required; see \code{space_mode}}

\item{space_mode}{mode of spacing, only used when \code{space} is less equal
to one; default is quantile}

\item{start_time}{the time to start drawing relative to the first column}

\item{duration}{duration of the signal to draw}

\item{compress}{whether to compress signals if the data is too large}

\item{channel_names}{\code{NULL} or a character vector of channel names}

\item{time_shift}{the actual start time of the signal. Unlike
\code{start_time}, this should be the actual physical time represented
by the first column}

\item{xlab, ylab, lwd, xlim, cex, cex.lab, mar, mgp, xaxs, yaxs, tck, ...}{plot
parameters; see \code{\link[base]{plot}} and \code{\link[graphics]{par}}}

\item{new_plot}{whether to draw a new plot; default is true}

\item{xline, yline}{the gap between axis and label}
}
\description{
Plot one or more signal traces in the same figure
}
\examples{


n <- 1000
base_signal <- c(rep(0, n/2), sin(seq(0,10,length.out = n/2))) * 10
signals <- rbind(rnorm(n) + base_signal,
                 rbinom(n, 10, 0.3) + base_signal,
                 rt(n, 5) + base_signal)
plot_signals(signals, sample_rate = 100)
plot_signals(signals, sample_rate = 100, start_time = 5)
plot_signals(signals, sample_rate = 100,
             start_time = 5, time_shift = 100)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_sphere}
\alias{vcg_sphere}
\title{Simple 3-dimensional sphere mesh}
\usage{
vcg_sphere(sub_division = 3L, normals = TRUE)
}
\arguments{
\item{sub_division}{density of vertex in the resulting mesh}

\item{normals}{whether the normal vectors should be calculated}
}
\value{
A \code{'mesh3d'} object
}
\description{
Simple 3-dimensional sphere mesh
}
\examples{

vcg_sphere()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image-mesh-from-volume.R
\name{mesh_from_volume}
\alias{mesh_from_volume}
\title{Generate 3D mesh surface from volume data}
\usage{
mesh_from_volume(
  volume,
  output_format = c("rgl", "freesurfer"),
  IJK2RAS = NULL,
  threshold = 0,
  verbose = TRUE,
  remesh = TRUE,
  remesh_voxel_size = 1,
  remesh_multisample = TRUE,
  remesh_automerge = TRUE,
  smooth = FALSE,
  smooth_lambda = 10,
  smooth_delta = 20,
  smooth_method = "surfPreserveLaplace"
)
}
\arguments{
\item{volume}{3-dimensional volume array}

\item{output_format}{resulting data format, choices are \code{'rgl'} and
\code{'freesurfer'}}

\item{IJK2RAS}{volume 'IJK' (zero-indexed coordinate index) to
\code{'tkrRAS'} transform, default is automatically determined}

\item{threshold}{threshold used to create volume mask; the surface will be
created to fit the mask boundaries}

\item{verbose}{whether to verbose the progress}

\item{remesh}{whether to re-sample the mesh using \code{\link{vcg_uniform_remesh}}}

\item{remesh_voxel_size, remesh_multisample, remesh_automerge}{see
arguments in \code{\link{vcg_uniform_remesh}}}

\item{smooth}{whether to smooth the mesh via \code{\link{vcg_smooth_explicit}}}

\item{smooth_lambda, smooth_delta, smooth_method}{see \code{\link{vcg_smooth_explicit}}}
}
\value{
A \code{'mesh3d'} surface if \code{output_format} is 'rgl', or
\code{'fs.surface'} surface otherwise.
}
\description{
This function is soft-deprecated. Please use
\code{\link{vcg_mesh_volume}}, \code{\link{vcg_uniform_remesh}}, and
\code{\link{vcg_smooth_explicit}} or \code{\link{vcg_smooth_implicit}}.
}
\examples{


volume <- array(0, dim = c(8,8,8))
volume[4:5, 4:5, 4:5] <- 1

graphics::image(x = volume[4,,])

# you can use rgl::wire3d(mesh) to visualize the mesh
mesh <- mesh_from_volume(volume, verbose = FALSE)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-Matrix4.R
\name{new_matrix4}
\alias{new_matrix4}
\alias{as_matrix4}
\title{Create a \code{Matrix4} instance for \code{'Affine'} transform}
\usage{
new_matrix4()

as_matrix4(m)
}
\arguments{
\item{m}{a matrix or a vector to be converted to the \code{Matrix4} instance;
\code{m} must be one of the followings: for matrices, the dimension must be
\code{4x4}, \code{3x4} (the last row will be \code{0 0 0 1}), or
\code{3x3} (linear transform); for vectors, the length must be
\code{16}, \code{12} (will append \code{0 0 0 1} internally),
\code{3} (translation), or \code{1} (scale).}
}
\value{
A \code{Matrix4} instance
}
\description{
Create a \code{Matrix4} instance for \code{'Affine'} transform
}
\seealso{
\code{\link{new_vector3}}, \code{\link{new_quaternion}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multitaper.R
\name{multitaper}
\alias{multitaper}
\alias{multitaper_config}
\title{Compute 'multitaper' spectral densities of time-series data}
\usage{
multitaper_config(
  data_length,
  fs,
  frequency_range = NULL,
  time_bandwidth = 5,
  num_tapers = NULL,
  window_params = c(5, 1),
  nfft = NA,
  detrend_opt = "linear"
)

multitaper(
  data,
  fs,
  frequency_range = NULL,
  time_bandwidth = 5,
  num_tapers = NULL,
  window_params = c(5, 1),
  nfft = NA,
  detrend_opt = "linear"
)
}
\arguments{
\item{data_length}{length of data}

\item{fs}{sampling frequency in 'Hz'}

\item{frequency_range}{frequency range to look at; length of two}

\item{time_bandwidth}{a number indicating time-half bandwidth product; i.e.
the window duration times the half bandwidth of main lobe; default is
\code{5}}

\item{num_tapers}{number of 'DPSS' tapers to use; default is \code{NULL} and
will be automatically computed from \code{floor(2*time_bandwidth - 1)}}

\item{window_params}{vector of two numbers; the first number is the
window size in seconds; the second number if the step size; default is
\code{c(5, 1)}}

\item{nfft}{'NFFT' size, positive; see 'Details'}

\item{detrend_opt}{how you want to remove the trend from data window; options
are \code{'linear'} (default), \code{'constant'}, and \code{'off'}}

\item{data}{numerical vector, signal traces}
}
\value{
\code{multitaper_config} returns a list of configuration parameters
for the filters; \code{multitaper} also returns the time, frequency and
corresponding spectral power.
}
\description{
Compute 'multitaper' spectral densities of time-series data
}
\details{
The original source code comes from 'Prerau' Lab (see 'Github'
repository \code{'multitaper_toolbox'} under user \code{'preraulab'}).
The results tend to agree with their 'Python' implementation with precision
on the order of at \code{1E-7} with standard deviation at most \code{1E-5}.
The original copy was licensed under a Creative Commons Attribution
'NC'-'SA' 4.0 International License
(\url{https://creativecommons.org/licenses/by-nc-sa/4.0/}).

This package (\code{'ravetools'}) redistributes the \code{multitaper}
function under minor modifications on \code{nfft}. In the original copy
there is no parameter to control the exact numbers of \code{nfft}, and
the \code{nfft} is always the power of 2. While choosing
\code{nfft} to be the power of 2 is always recommended, the modified code
allows other choices.
}
\examples{

\donttest{

# Takes long to run

time <- seq(0, 3, by = 0.001)
x <- sin(time * 20*pi) + exp(-time^2) * cos(time * 10*pi)

res <- multitaper(
  x, 1000, frequency_range = c(0,15),
  time_bandwidth=1.5,
  window_params=c(2,0.01)
)


image(
  x = res$time,
  y = res$frequency,
  z = 10 * log10(res$spec),
  xlab = "Time (s)",
  ylab = 'Frequency (Hz)',
  col = matlab_palette()
)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-diagnosis.R
\name{freqz2}
\alias{freqz2}
\title{Frequency response of digital filter}
\usage{
freqz2(b, a = 1, fs = 2 * pi, n = 512, whole = FALSE, ...)
}
\arguments{
\item{b}{the moving-average coefficients of an \code{ARMA} model}

\item{a}{the auto-regressive coefficients of an \code{ARMA} filter; default
is \code{1}}

\item{fs}{sampling frequency in \code{Hz}}

\item{n}{number of points at which to evaluate the frequency response;
default is \code{512}}

\item{whole}{whether to evaluate beyond \code{Nyquist} frequency; default
is false}

\item{...}{ignored}
}
\value{
A list of frequencies and corresponding responses in complex vector
}
\description{
Compute the z-plane frequency response of an \code{ARMA} model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg-collision.R
\name{vcg_detect_collision}
\alias{vcg_detect_collision}
\title{Detect collisions between two geometries}
\usage{
vcg_detect_collision(
  x,
  y,
  mode_x = c("auto", "points", "segments", "mesh"),
  mode_y = c("auto", "points", "segments", "mesh"),
  radius = 0,
  early_stop = FALSE,
  include_interior = FALSE
)
}
\arguments{
\item{x}{the geometry that gets indexed; a matrix with \code{n} rows and 2
or 3 columns, or a surface (see \code{\link{ensure_mesh3d}})}

\item{y}{the geometry that gets queried; results are aligned to this side.
Same accepted types as \code{x}}

\item{mode_x, mode_y}{how to interpret \code{x} and \code{y}: \code{'points'}
treats each row as an isolated point, \code{'segments'} chains consecutive
rows together (\code{p1-p2}, \code{p2-p3}, ...), and \code{'mesh'} uses
the triangular faces. The default \code{'auto'} resolves to \code{'mesh'}
for a surface that has faces and \code{'points'} otherwise, so a matrix of
connected line segments must ask for \code{'segments'} explicitly}

\item{radius}{distance tolerance; a collision is reported when the minimum
distance is at most \code{radius}. Defaults to \code{0}, meaning literal
contact; a positive tolerance is usually what is wanted}

\item{early_stop}{whether to stop as soon as a collision is found, scanning
independently within each group of \code{y} delimited by \code{NA} rows. A
matrix carrying no \code{NA} rows is a single group, so a point cloud stops
after one hit overall rather than per row; likewise \code{mode_y = 'mesh'}
stops at the first colliding face. Skipped elements report \code{hit = NA},
meaning \emph{not evaluated} rather than \emph{no collision}, and the
collision reported is the first in row order, not the closest. Use
\code{FALSE} (the default) whenever the per-element profile or a minimum
distance is wanted}

\item{include_interior}{whether geometry lying strictly inside a closed
\code{x} surface counts as a collision even when it never comes within
\code{radius} of the surface itself; requires \code{mode_x = 'mesh'}.
Default is \code{FALSE}}
}
\value{
A list of three vectors, each aligned one-to-one with the rows of
\code{y} (or with the faces of \code{y} when \code{mode_y = 'mesh'}):
\describe{
  \item{\code{hit}}{logical. \code{NA} marks an element with nothing to
    test: a separator row, or the final vertex of each group when
    \code{mode_y = 'segments'}. When \code{early_stop = TRUE}, \code{NA}
    also marks elements the scan never reached}
  \item{\code{distance}}{the exact minimum distance where \code{hit} is
    \code{TRUE}, and \code{NA} everywhere else. Distances beyond
    \code{radius} are never computed, so a \code{FALSE} element carries no
    distance}
  \item{\code{index}}{1-based index of the closest element of \code{x}
    within \code{radius}: a row when \code{mode_x} is \code{'points'}, the
    row where the segment starts when \code{'segments'}, or a face column
    when \code{'mesh'}. \code{NA} where there is no collision}
}
}
\description{
Reports, for every element of \code{y}, whether it comes within
\code{radius} of \code{x}, together with the exact minimum distance. Each
side may be a point cloud, a set of connected line segments (for example
diffusion streamlines or an electrode shaft), or a triangular mesh, so the
function covers all nine pairings with one call.

The test is exact: it produces neither false negatives nor false positives,
and \code{distance} is the true minimum distance rather than a
vertex-sampled approximation.
}
\details{
\code{x} is indexed once into a uniform spatial grid and \code{y} is
streamed against it, so put the larger or repeatedly-reused geometry in
\code{x} and the geometry you want per-element answers about in \code{y}.

Several chains share one matrix and are delimited by rows of \code{NA}; any
row that is not fully finite is treated as a separator. This lets a whole
bundle of streamlines be passed as a single matrix:

\preformatted{
  rbind(
    c(0, 0, 0), c(1, 0, 0), c(2, 0, 0),   # first streamline
    c(NA, NA, NA),                        # separator
    c(0, 5, 0), c(1, 5, 0)                # second streamline
  )
}

Those separators also define the unit that \code{early_stop} works on: it
stops once per group, so a bundle of streamlines yields at most one hit per
streamline.

When \code{mode_y = 'segments'}, row \code{i} of the result describes the
segment running from row \code{i} to row \code{i + 1}.

Both geometries must already share one coordinate space. Surface
coordinates, volume \verb{IJK} indices, and scanner coordinates are all
different spaces, and no transform is applied here.

\code{include_interior} relies on ray casting and therefore needs \code{x}
to be watertight with coherently oriented faces; see
\code{\link{vcg_fix_defects}} to repair a surface that is not.

Multi-threading follows \code{\link{ravetools_threads}}. The interior test
always runs single-threaded.
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

library(ravetools)

# A spherical region of interest
roi <- vcg_sphere()

# Two streamlines in one matrix, separated by an NA row: the first passes
# through the sphere, the second stays well clear of it
streamlines <- rbind(
  cbind(seq(-3, 3, by = 0.5), 0, 0),
  c(NA, NA, NA),
  cbind(seq(-3, 3, by = 0.5), 5, 0)
)

result <- vcg_detect_collision(
  x = roi,
  y = streamlines,
  mode_y = "segments",
  radius = 0.1
)

# Which segments touch the sphere, and how close do they get?
data.frame(
  hit = result$hit,
  distance = round(result$distance, 4)
)

# A point cloud against the same region, asking only for proximity
points <- rbind(c(0, 0, 0), c(1.05, 0, 0), c(10, 10, 10))
vcg_detect_collision(roi, points, radius = 0.1)$hit

# The centre of the sphere is far from its surface, so it only counts as a
# collision when the interior is included
vcg_detect_collision(roi, rbind(c(0, 0, 0)), radius = 0.1)$hit
vcg_detect_collision(roi, rbind(c(0, 0, 0)), radius = 0.1,
                     include_interior = TRUE)$hit

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_average_edge_length}
\alias{vcg_average_edge_length}
\title{Compute the average edge length of a triangular mesh}
\usage{
vcg_average_edge_length(mesh)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.}
}
\value{
A single numeric value: the average edge length, in mesh units.
}
\description{
Computes the average length of all face edges (each edge is counted once
per incident face, so edges shared by two faces are counted twice).  Useful
as a scale-aware reference length, e.g. to derive a vertex-merge tolerance
such as the one used internally by \code{\link{vcg_fix_defects}}.
}
\examples{
if (is_not_cran()) {

  sphere <- vcg_sphere()
  vcg_average_edge_length(sphere)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/baseline.R
\name{baseline_array}
\alias{baseline_array}
\alias{baseline_array.array}
\title{Calculate Contrasts of Arrays in Different Methods}
\usage{
baseline_array(x, along_dim, unit_dims = seq_along(dim(x))[-along_dim], ...)

\method{baseline_array}{array}(
  x,
  along_dim,
  unit_dims = seq_along(dim(x))[-along_dim],
  method = c("percentage", "sqrt_percentage", "decibel", "zscore", "sqrt_zscore",
    "db_zscore", "subtract_mean"),
  baseline_indexpoints = NULL,
  baseline_subarray = NULL,
  ...
)
}
\arguments{
\item{x}{array (tensor) to calculate contrast}

\item{along_dim}{integer range from 1 to the maximum dimension of \code{x}.
baseline along this dimension, this is usually the time dimension.}

\item{unit_dims}{integer vector, baseline unit: see Details.}

\item{...}{passed to other methods}

\item{method}{character, baseline method; one of \code{"percentage"},
\code{"sqrt_percentage"}, \code{"decibel"}, \code{"zscore"},
\code{"sqrt_zscore"}, \code{"db_zscore"}, or \code{"subtract_mean"}.
See Details.}

\item{baseline_indexpoints}{integer vector, which index points are counted
into baseline window? Each index ranges from 1 to \code{dim(x)[[along_dim]]}.
See Details.}

\item{baseline_subarray}{sub-arrays that should be used to calculate
baseline; default is \code{NULL} (automatically determined by
\code{baseline_indexpoints}).}
}
\value{
Contrast array with the same dimension as \code{x}.
}
\description{
Provides seven methods to baseline an array and calculate
contrast.
}
\details{
Consider a scenario where we want to baseline a bunch of signals recorded
from different locations. For each location, we record \code{n} sessions.
For each session, the signal is further decomposed into frequency-time
domain. In this case, we have the input \code{x} in the following form:
\deqn{session \times frequency \times time \times location}
Now we want to calibrate signals for each session, frequency and location
using the first 100 time points as baseline points, then the code will be
\code{baseline_array(x, along_dim=3, baseline_window=1:100, unit_dims=c(1,2,4))}
\code{along_dim=3} is dimension of time, in this case, it's the
third dimension of \code{x}. \code{baseline_indexpoints=1:100}, meaning
the first 100 time points are used to calculate baseline.
\code{unit_dims} defines the unit signal. Its value \code{c(1,2,4)}
means the unit signal is per session (first dimension), per frequency
(second) and per location (fourth).

In some other cases, we might want to calculate baseline across frequencies
then the unit signal is \eqn{frequency x time}, i.e. signals that share the
same session and location also share the same baseline. In this case,
we assign \code{unit_dims=c(1,4)}.

There are seven baseline methods. They fit for different types of data.
Denote \eqn{z} is a unit signal and \eqn{z_0} is its baseline slice. Then
these baseline methods are:

\describe{
\item{\code{"percentage"}}{
\deqn{
  \frac{z - \bar{z_{0}}}{\bar{z_{0}}} \times 100\%
}{
  (z / mean(z_0) - 1) x 100\%
}
}
\item{\code{"sqrt_percentage"}}{
\deqn{
  \frac{\sqrt{z} - \bar{\sqrt{z_{0}}}}{\bar{\sqrt{z_{0}}}} \times 100\%
}{
  (sqrt(z) / mean(sqrt(z_0)) - 1) x 100\%
}
}
\item{\code{"decibel"}}{
\deqn{
  10 \times ( \log_{10}(z) - \bar{\log_{10}(z_{0})} )
}{
  10 * ( log10(z) - mean( log10(z_0) ) )
}
}
\item{\code{"zscore"}}{
\deqn{
  \frac{z-\bar{z_{0}}}{sd(z_{0})}
}{
  (z - mean( z_0 )) / sd( z_0 )
}
}
\item{\code{"sqrt_zscore"}}{
\deqn{
  \frac{\sqrt{z}-\bar{\sqrt{z_{0}}}}{sd(\sqrt{z_{0}})}
}{
  (sqrt(z) - mean( sqrt(z_0) )) / sd( sqrt(z_0) )
}
}
\item{\code{"db_zscore"}}{
Z-score applied in the decibel (10log10) domain:
\deqn{
  \frac{10\log_{10}(z) - \bar{10\log_{10}(z_{0})}}{sd(10\log_{10}(z_{0}))}
}{
  ( log10(z) - mean( log10(z_0) ) ) / sd( log10(z_0) )
}
}
\item{\code{"subtract_mean"}}{
Simple mean subtraction with no scaling:
\deqn{
  z - \bar{z_{0}}
}{
  z - mean(z_0)
}
}


}
}
\examples{

# Set ncores = 2 to comply to CRAN policy. Please don't run this line
ravetools_threads(n_threads = 2L)


library(ravetools)
set.seed(1)

# Generate sample data
dims = c(10,20,30,2)
x = array(rnorm(prod(dims))^2, dims)

# Set baseline window to be arbitrary 10 timepoints
baseline_window = sample(30, 10)

# ----- baseline percentage change ------

# Using base functions
re1 <- aperm(apply(x, c(1,2,4), function(y) {
  m <- mean(y[baseline_window])
  (y/m - 1) * 100
}), c(2,3,1,4))

# Using ravetools
re2 <- baseline_array(x, 3, c(1,2,4),
                      baseline_indexpoints = baseline_window,
                      method = 'percentage')

# Check different, should be very tiny (double precisions)
range(re2 - re1)

\donttest{
# Check speed for large dataset, might take a while to profile

ravetools_threads(n_threads = -1)

dims <- c(200,20,300,2)
x <- array(rnorm(prod(dims))^2, dims)
# Set baseline window to be arbitrary 10 timepoints
baseline_window <- seq_len(100)
f1 <- function() {
  aperm(apply(x, c(1,2,4), function(y) {
    m <- mean(y[baseline_window])
    (y/m - 1) * 100
  }), c(2,3,1,4))
}
f2 <- function() {
  # equivalent as bl = x[,,baseline_window, ]
  #
  baseline_array(x, along_dim = 3,
                 baseline_indexpoints = baseline_window,
                 unit_dims = c(1,2,4), method = 'percentage')
}
range(f1() - f2())
microbenchmark::microbenchmark(f1(), f2(), times = 10L)

}



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raw_to_sexp.R
\name{raw-to-sexp}
\alias{raw-to-sexp}
\alias{raw_to_uint8}
\alias{raw_to_uint16}
\alias{raw_to_uint32}
\alias{raw_to_int8}
\alias{raw_to_int16}
\alias{raw_to_int32}
\alias{raw_to_int64}
\alias{raw_to_float}
\alias{raw_to_string}
\title{Convert raw vectors to R vectors}
\usage{
raw_to_uint8(x)

raw_to_uint16(x)

raw_to_uint32(x)

raw_to_int8(x)

raw_to_int16(x)

raw_to_int32(x)

raw_to_int64(x)

raw_to_float(x)

raw_to_string(x)
}
\arguments{
\item{x}{raw vector of bytes}
}
\value{
Numeric vectors, except for \code{raw_to_string}, which returns
a string.
}
\description{
Convert raw vectors to R vectors
}
\details{
For numeric conversions, the function names are straightforward.
For example,
\code{raw_to_uintN} converts raw vectors to unsigned integers, and
\code{raw_to_intN} converts raw vectors to signed integers. The number
\code{'N'} stands for the number of bits used to store the integer.
For example \code{raw_to_uint8} uses 8 bits (1 byte) to store an integer,
hence the value range is \code{0-255}.

The input data length must be multiple of the element size represented by
the underlying data. For example \code{uint16} integer uses 16 bites, and
one raw number uses 8 bits, hence two raw vectors can form one unsigned
integer-16. That is, \code{raw_to_uint16} requires the length of input
to be multiple of two. An easy calculation is: the length of \code{x} times
8, must be divided by \code{'N'} (see last paragraph for definition).

The returned data uses the closest available R native data type that can
fully represent the data. For example, R does not have single \code{float}
type, hence \code{raw_to_float} returns \code{double} type, which can
represent all possible values in \code{float}. For \code{raw_to_uint32},
the potential value range is \code{0 - (2^32-1)}. This exceeds the limit of
R integer type \code{(-2^31) - (2^31-1)}. Therefore, the returned values
will be real (double float) data type.

There is no native data type that can store integer-64 data in R, package
\code{bit64} provides \code{integer64} type, which will be used by
\code{raw_to_int64}. Currently there is no solution to convert raw to
unsigned integer-64 type.

\code{raw_to_string} converts raw to character string. This function respects
\code{null} character, hence is slightly different than the native
\code{\link{rawToChar}}, which translates raw byte-by-byte. If each
raw byte represents a valid character, then the above two functions returns
the same result. However, when the characters represented by raw bytes are
invalid, \code{raw_to_string} will stop parsing and returns only the valid
characters, while \code{\link{rawToChar}} will still try to parse, and
most likely to result in errors.
Please see Examples for comparisons.
}
\examples{

# 0x00, 0x7f, 0x80, 0xFF
x <- as.raw(c(0, 127, 128, 255))

raw_to_uint8(x)

# The first bit becomes the integer sign
# 128 -> -128, 255 -> -1
raw_to_int8(x)

## Comments based on little endian system

# 0x7f00 (32512), 0xFF80 (65408 unsigned, or -128 signed)
raw_to_uint16(x)
raw_to_int16(x)

# 0xFF807F00 (4286611200 unsigned, -8356096 signed)
raw_to_uint32(x)
raw_to_int32(x)

# ---------------------------- String ---------------------------

# ASCII case: all valid
x <- charToRaw("This is an ASCII string")

raw_to_string(x)
rawToChar(x)

x <- c(charToRaw("This is the end."),
       as.raw(0),
       charToRaw("*** is invalid"))

# rawToChar will raise error
raw_to_string(x)

# ---------------------------- Integer64 ------------------------
# Runs on little endian system
x <- as.raw(c(0x80, 0x00, 0x7f, 0x80, 0xFF, 0x50, 0x7f, 0x00))

# Calculate bitstring, which concaternates the followings
# 10000000 (0x80), 00000000 (0x00), 01111111 (0x7f), 10000000 (0x80),
# 11111111 (0xFF), 01010000 (0x50), 01111111 (0x7f), 00000000 (0x00)

if(.Platform$endian == "little") {
  bitstring <- paste0(
    "00000000011111110101000011111111",
    "10000000011111110000000010000000"
  )
} else {
  bitstring <- paste0(
    "00000001000000001111111000000001",
    "11111111000010101111111000000000"
  )
}

# This is expected value
bit64::as.integer64(structure(
  bitstring,
  class = "bitstring"
))

# This is actual value
raw_to_int64(x)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/band-pass.R
\name{band_pass}
\alias{band_pass}
\alias{band_pass1}
\alias{band_pass2}
\title{Band-pass signals}
\usage{
band_pass1(x, sample_rate, lb, ub, domain = 1, ...)

band_pass2(
  x,
  sample_rate,
  lb,
  ub,
  order,
  method = c("fir", "butter"),
  direction = c("both", "forward", "backward"),
  window = "hamming",
  ...
)
}
\arguments{
\item{x}{input signals, numeric vector or matrix. \code{x} must be
row-major if input is a matrix: each row is a channel, and each column is
a time-point.}

\item{sample_rate}{sampling frequency}

\item{lb}{lower frequency bound of the band-passing filter, must be positive}

\item{ub}{upper frequency bound of the band-passing filter, must be greater
than the lower bound and smaller than the half of sampling frequency}

\item{domain}{1 if \code{x} is in time-domain, or 0 if \code{x} is in
frequency domain}

\item{...}{ignored}

\item{order}{the order of the filter, must be positive integer and be
less than one-third of the sample rate}

\item{method}{filter type, choices are \code{'fir'} and \code{'butter'}}

\item{direction}{filter direction, choices are \code{'forward'},
\code{'backward'}, and \code{'both'} directions}

\item{window}{window type, can be a character, a function, or a vector.
For character, \code{window} is a function name in the
\code{signal} package, for example, \code{'hanning'}; for a function,
\code{window} takes one integer argument and returns a numeric vector
with length of that input; for vectors, \code{window} is a numeric vector
o length \code{order+1}.}
}
\value{
Filtered signals, vector if \code{x} is a vector, or matrix of
the same dimension as \code{x}
}
\description{
Band-pass signals
}
\examples{


t <- seq(0, 1, by = 0.0005)
x <- sin(t * 0.4 * pi) + sin(t * 4 * pi) + 2 * sin(t * 120 * pi)

oldpar <- par(mfrow = c(2, 2), mar = c(3.1, 2.1, 3.1, 0.1))
# ---- Using band_pass1 ------------------------------------------------

y1 <- band_pass1(x, 2000, 0.1, 1)
y2 <- band_pass1(x, 2000, 1, 5)
y3 <- band_pass1(x, 2000, 10, 80)

plot(t, x, type = 'l', xlab = "Time", ylab = "",
     main = "Mixture of 0.2, 2, and 60Hz")
lines(t, y1, col = 'red')
lines(t, y2, col = 'blue')
lines(t, y3, col = 'green')
legend(
  "topleft", c("Input", "Pass: 0.1-1Hz", "Pass 1-5Hz", "Pass 10-80Hz"),
  col = c(par("fg"), "red", "blue", "green"), lty = 1,
  cex = 0.6
)

# plot pwelch
pwelch(x, fs = 2000, window = 4000, noverlap = 2000, plot = 1)
pwelch(y1, fs = 2000, window = 4000, noverlap = 2000,
       plot = 2, col = "red")
pwelch(y2, fs = 2000, window = 4000, noverlap = 2000,
       plot = 2, col = "blue")
pwelch(y3, fs = 2000, window = 4000, noverlap = 2000,
       plot = 2, col = "green")


# ---- Using band_pass2 with FIR filters --------------------------------

order <- floor(2000 / 3)
z1 <- band_pass2(x, 2000, 0.1, 1, method = "fir", order = order)
z2 <- band_pass2(x, 2000, 1, 5, method = "fir", order = order)
z3 <- band_pass2(x, 2000, 10, 80, method = "fir", order = order)

plot(t, x, type = 'l', xlab = "Time", ylab = "",
     main = "Mixture of 0.2, 2, and 60Hz")
lines(t, z1, col = 'red')
lines(t, z2, col = 'blue')
lines(t, z3, col = 'green')
legend(
  "topleft", c("Input", "Pass: 0.1-1Hz", "Pass 1-5Hz", "Pass 10-80Hz"),
  col = c(par("fg"), "red", "blue", "green"), lty = 1,
  cex = 0.6
)

# plot pwelch
pwelch(x, fs = 2000, window = 4000, noverlap = 2000, plot = 1)
pwelch(z1, fs = 2000, window = 4000, noverlap = 2000,
       plot = 2, col = "red")
pwelch(z2, fs = 2000, window = 4000, noverlap = 2000,
       plot = 2, col = "blue")
pwelch(z3, fs = 2000, window = 4000, noverlap = 2000,
       plot = 2, col = "green")

# ---- Clean this demo --------------------------------------------------
par(oldpar)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crp.R
\name{plot.ravetools_crp}
\alias{plot.ravetools_crp}
\title{Plot \verb{CRP} results}
\usage{
\method{plot}{ravetools_crp}(x, ...)
}
\arguments{
\item{x}{an object of class \code{ravetools_crp} as returned by
\code{\link{crp}}.}

\item{...}{additional graphical parameters passed to \code{\link{par}}
(e.g. \code{mar}, \code{cex.axis}); currently unused beyond restoring
the previous \code{par} state on exit.}
}
\value{
Invisibly returns \code{x}.
}
\description{
S3 plot method for objects of class \code{ravetools_crp} returned by
\code{\link{crp}}.  Produces a three-panel figure:
\enumerate{
\item Single-trial traces over the entire loaded time range with the mean,
  the canonical shape \eqn{C(t)} on its active support (solid) and the
  full-range extension \eqn{C_{full}(t)} (dashed) overlaid; the analysis
  window edges and, when present, \eqn{\tau_{onset}} are marked.
\item Per-trial \eqn{\alpha'} weights sorted in ascending order.
\item Mean cross-trial projection profile with vertical lines marking
  \eqn{\tau_{lb}}, \eqn{\tau_R} and \eqn{\tau_{ub}}, plus the reverse
  onset profile and \eqn{\tau_{onset}} when \code{detect_onset} was used.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-Quaternion.R
\name{new_quaternion}
\alias{new_quaternion}
\alias{as_quaternion}
\title{Create a \code{Quaternion} instance to store '3D' rotation}
\usage{
new_quaternion(x = 0, y = 0, z = 0, w = 1)

as_quaternion(q)
}
\arguments{
\item{x, y, z, w}{numeric of length one}

\item{q}{R object to be converted to \code{Quaternion}}
}
\value{
A \code{Quaternion} instance
}
\description{
Create instances that mimic the \code{'three.js'} syntax.
}
\seealso{
\code{\link{new_vector3}}, \code{\link{new_matrix4}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{is_not_cran}
\alias{is_not_cran}
\title{Internal function}
\usage{
is_not_cran(if_interactive = TRUE, verbose = FALSE)
}
\arguments{
\item{if_interactive, verbose}{default is \code{TRUE}}
}
\value{
logical
}
\description{
Do not call this function directly
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter.R
\name{design_filter}
\alias{design_filter}
\title{Design a digital filter}
\usage{
design_filter(
  sample_rate,
  data = NULL,
  method = c("fir_kaiser", "firls", "fir_remez", "butter", "cheby1", "cheby2", "ellip"),
  high_pass_freq = NA,
  high_pass_trans_freq = NA,
  low_pass_freq = NA,
  low_pass_trans_freq = NA,
  passband_ripple = 0.1,
  stopband_attenuation = 40,
  filter_order = NA,
  use_sos = TRUE,
  ...,
  data_size = length(data)
)
}
\arguments{
\item{sample_rate}{data sample rate}

\item{data}{data to be filtered, can be optional (\code{NULL})}

\item{method}{filter method, options are \code{"fir"} (default),
\code{"butter"}, \code{"cheby1"}, \code{"cheby2"}, and \code{"ellip"}}

\item{high_pass_freq, low_pass_freq}{high-pass or low-pass frequency,
see \code{\link{design_filter_fir}} or \code{\link{design_filter_iir}}}

\item{high_pass_trans_freq, low_pass_trans_freq}{transition bandwidths,
see \code{\link{design_filter_fir}} or \code{\link{design_filter_iir}}}

\item{passband_ripple}{allowable pass-band ripple in decibel; default is
\code{0.1}}

\item{stopband_attenuation}{minimum stop-band attenuation (in decibel) at
transition frequency; default is \code{40} dB.}

\item{filter_order}{suggested filter order; for 'IIR' methods, see
\code{\link{design_filter_iir}} for details on how this interacts with
\code{use_sos}; for 'FIR' methods 'RAVE' may or may not adopt this
suggestion depending on the data and numerical feasibility}

\item{use_sos}{logical; passed to \code{\link{design_filter_iir}} for 'IIR'
methods (ignored for 'FIR'). When \code{TRUE} (default), filtering is done
via 'SOS' form using \code{gsignal::filtfilt}; when \code{FALSE}, the
ravetools 'ARMA' \code{filtfilt} is used.}

\item{...}{passed to filter generator functions}

\item{data_size}{used by 'FIR' filter design to determine maximum order,
ignored in 'IIR' filters; automatically derived from \code{data}}
}
\value{
If \code{data} is specified and non-empty, this function returns
filtered data via forward and backward \code{filtfilt}; if \code{data} is
\code{NULL}, then returns the generator function.
}
\description{
Provides 'FIR' and 'IIR' filter options; default is 'FIR', see also
\code{\link{design_filter_fir}}; for 'IIR' filters, see
\code{\link{design_filter_iir}}.
}
\examples{


sample_rate <- 200
t <- seq(0, 10, by = 1 / sample_rate)
x <- sin(t * 4 * pi) + sin(t * 20 * pi) +
  2 * sin(t * 120 * pi) + rnorm(length(t), sd = 0.4)

# ---- Using FIR ------------------------------------------------

# Low-pass filter
y1 <- design_filter(
  data = x,
  sample_rate = sample_rate,
  low_pass_freq = 3, low_pass_trans_freq = 0.5
)

# Band-pass cheby1 filter 8-12 Hz with custom transition
y2 <- design_filter(
  data = x,
  method = "cheby1",
  sample_rate = sample_rate,
  low_pass_freq = 12, low_pass_trans_freq = .25,
  high_pass_freq = 8, high_pass_trans_freq = .25
)

y3 <- design_filter(
  data = x,
  sample_rate = sample_rate,
  low_pass_freq = 80,
  high_pass_freq = 30
)

oldpar <- par(mfrow = c(2, 1),
              mar = c(3.1, 2.1, 3.1, 0.1))
plot(t, x, type = 'l', xlab = "Time", ylab = "",
     main = "Mixture of 2, 10, and 60Hz", xlim = c(0,1))
# lines(t, y, col = 'red')
lines(t, y3, col = 'green')
lines(t, y2, col = 'blue')
lines(t, y1, col = 'red')
legend(
  "topleft", c("Input", "Low: 3Hz", "Pass 8-12Hz", "Pass 30-80Hz"),
  col = c(par("fg"), "red", "blue", "green"), lty = 1,
  cex = 0.6
)

# plot pwelch
pwelch(x, fs = sample_rate, window = sample_rate * 2,
       noverlap = sample_rate, plot = 1, ylim = c(-100, 10))
pwelch(y1, fs = sample_rate, window = sample_rate * 2,
       noverlap = sample_rate, plot = 2, col = "red")
pwelch(y2, fs = sample_rate, window = sample_rate * 2,
       noverlap = sample_rate, plot = 2, col = "blue")
pwelch(y3, fs = sample_rate, window = sample_rate * 2,
       noverlap = sample_rate, plot = 2, col = "green")


# ---- Clean this demo --------------------------------------------------
par(oldpar)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-mesh-polygon.R
\name{plot_mesh_polygon}
\alias{plot_mesh_polygon}
\title{Render one or more meshes as flat-shaded triangles in base R}
\usage{
plot_mesh_polygon(
  mesh,
  eye = c(0, 0, 1000),
  lookat = c(0, 0, 0),
  up = c(0, 1, 0),
  col = c("white", "gray30"),
  cex = 1,
  add = FALSE,
  axes = FALSE,
  asp = 1,
  xlim = NULL,
  ylim = NULL,
  zoom = 1,
  xlab = "",
  ylab = "",
  main = "",
  side = c("front", "back", "both"),
  mesh_clipping = 1,
  sphere_subdivision = 1L,
  alpha = 1,
  shadow_color = NULL,
  light_intensity = 1,
  ambient_intensity = 0.2,
  clipping_plane = NULL,
  clipping_plane_enabled = TRUE,
  ...
)
}
\arguments{
\item{mesh}{a \code{'mesh3d'} object, or a list of \code{'mesh3d'} objects.
Meshes without a face matrix (\code{$it}) are rendered as sphere
instances (one \code{\link{vcg_sphere}} per vertex, radius \code{cex}).}

\item{eye}{numeric vector of length 3 - camera position in world space.}

\item{lookat}{numeric vector of length 3 - the world-space point the camera
is looking at.}

\item{up}{numeric vector of length 3 - world-space "up" direction; defaults
to \code{c(0, 1, 0)}.}

\item{col}{base color(s) per mesh.  Same forms as
\code{\link{plot_mesh_dotcloud}}: a single color, a depth-gradient vector,
a per-vertex character vector, or a list of any of these (one element per
mesh).  Default \code{c("white", "gray30")}.}

\item{cex}{radius of the substitute sphere used for point-cloud meshes
(world units).  Has no effect on meshes that already have faces.  Default
\code{1}.}

\item{add}{logical; if \code{TRUE} the faces are added to an existing plot
instead of opening a new one.  Default \code{FALSE}.}

\item{axes, asp, xlim, ylim, xlab, ylab, main}{passed to
\code{\link[graphics]{plot.default}} when \code{add = FALSE}.}

\item{zoom}{positive numeric magnification applied to the auto-computed
axis limits when \code{xlim} or \code{ylim} is \code{NULL}.  Values
greater than \code{1} zoom in, values in \code{(0, 1)} zoom out.
When \code{asp} is set, the zoom is plot-region aware (queried via
\code{par("pin")}): the axis that already fills the device after
\code{asp}-induced expansion is zoomed by exactly \code{zoom}, and
the other axis is zoomed less so the data fills more of the plot
region while preserving the requested \code{asp}.  Ignored for any
axis whose limit was supplied explicitly, and when \code{add = TRUE}.
Default \code{1}.}

\item{side}{which side of each triangle to render.  One of \code{"front"}
(default), \code{"back"}, or \code{"both"} (shows all triangles).}

\item{mesh_clipping}{numeric in \code{[0, 1]} controlling camera-facing
clipping: any triangle whose signed Lambert dot product
(\eqn{n \cdot z_{cam}}, i.e. positive for front-facing) is at or
above \code{mesh_clipping} is discarded.  A value of \code{1}
(default) keeps the full surface; \code{0.5} peels a 60-degree cap
off the front and reveals the back wall (whose absolute Lambert
shade is still large, so it renders brightly); \code{0} clips
every front-facing triangle and shows the interior only.
Back-facing triangles are never clipped by this rule.  Default
\code{1}.}

\item{sphere_subdivision}{integer subdivision level forwarded to
\code{\link{vcg_sphere}} when substituting point clouds.  Higher values
give smoother spheres at the cost of more triangles.  Default \code{1}
(80 triangles per sphere).}

\item{alpha}{numeric in \code{[0, 1]}; one value per mesh (recycled),
sets the transparency of each mesh (\code{1} = fully opaque,
\code{0} = fully transparent).  Faces belonging to a mesh with
\code{alpha = 0} are dropped entirely.  Because faces are drawn in
back-to-front (painter's) order, alpha blending across multiple
meshes follows that order.  Default \code{1}.}

\item{shadow_color}{color used for fully unlit (grazing/back) faces.
The Lambert shade linearly interpolates from \code{shadow_color}
(at \code{shade = 0}) toward the face color lit by a white light
(at \code{shade = 1}), so unlit faces fade to this color instead
of an implicit black background.  Default \code{par("fg")}, so the
shadow contrasts with the current device's canvas.}

\item{light_intensity}{non-negative scalar controlling the brightness
of the (white) light source: at \code{shade = 1} the face color is
multiplied by \code{light_intensity} (clamped to \code{[0, 1]}).
\code{1} (default) reproduces the face color exactly when fully
lit; smaller values darken the whole mesh, larger values saturate.}

\item{ambient_intensity}{scalar in \code{[0, 1]} acting as a lower
bound on the Lambert shade so grazing/back faces still receive some
light and never collapse fully to \code{shadow_color}.  An effective
shade of \code{max(shade, ambient_intensity)} is used in the
background-to-light blend.  Default \code{0.2}.}

\item{clipping_plane}{optional list of world-space clipping planes used to
hide parts of the scene.  Each plane is a numeric vector of length 5:
the first three entries are the plane normal \eqn{(n_x, n_y, n_z)}
(must be non-zero; normalized internally), the fourth is the signed
distance from the world origin to the plane along that normal (so the
plane equation is \eqn{n \cdot x = d}), and the fifth indicates which
half-space is kept: \code{1} keeps the front side (the side the normal
points to), \code{-1} keeps the back side, and \code{0} keeps whichever
side currently faces the camera (auto-flipped per call based on
\code{eye}).  Multiple planes are intersected.  Clipping is applied
per face using the face centroid (a face is kept only when its
centroid lies on the kept side of every plane).  A single length-5
numeric vector is also accepted as shorthand for a one-plane list.
Default \code{NULL} (no clipping).}

\item{clipping_plane_enabled}{logical vector, one entry per mesh
(recycled), controlling whether \code{clipping_plane} is applied to
that mesh.  \code{TRUE} (default) means the mesh participates in
clipping; \code{FALSE} exempts the mesh entirely (all of its faces
are kept regardless of the clipping planes).  Has no effect when
\code{clipping_plane} is \code{NULL}.}

\item{...}{additional graphical parameters forwarded to
\code{\link[graphics]{plot.default}} (new plot only).}
}
\value{
Invisibly returns a list with components \code{xlim} and
  \code{ylim} (the plot limits used).
}
\description{
Projects each triangular face onto a 2D plane using an orthographic camera,
shades it with a single color proportional to how directly it faces the
camera (Lambert), depth-sorts all faces across all meshes, and draws them
in a single \code{\link[graphics]{polygon}} call.

Meshes without faces (point clouds) are substituted by a small sphere
(\code{\link{vcg_sphere}}) centered at each point and scaled by
\code{cex}; they then participate in the same rendering pipeline as ordinary
faced meshes.

A camera-facing clipping pass discards triangles whose outward normal
points along the camera ray (signed \eqn{n \cdot z_{cam}} \eqn{> 1 - \mathrm{mesh\_clipping}}),
peeling the front cap off the surface so the back wall (and any
interior meshes) become visible. Set \code{mesh_clipping = 1} to disable
clipping. Point-cloud meshes (those rendered as substitute
\code{\link{vcg_sphere}} instances) are exempt from this clip so they
remain solid even when the enclosing surface is peeled.

Multiple meshes share a single depth space: all faces are projected,
sorted, and drawn together so the painter's algorithm works correctly
across meshes.
}
\details{
Limitations of the base-R polygon path (no \code{rgl}):
\itemize{
  \item Flat shading only - one color per triangle.  No per-vertex
    color interpolation.
  \item No depth buffer - faces are depth-sorted by centroid (painter's
    algorithm).  Interpenetrating triangles can render in the wrong
    order.
  \item Anti-aliasing seams can appear between adjacent triangles on
    raster devices; Cairo-based devices (\code{png(type = "cairo")},
    \code{svg()}, \code{pdf()}) produce cleaner output than the default
    quartz/X11 path.
}
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

mesh <- vcg_isosurface(left_hippocampus_mask)

# Surface alone
plot(
  mesh,
  eye    = c(150, 30, 0),
  lookat = c(0, 0, 0),
  up     = c(0, 0, 1),
  col    = "steelblue"
)

# Surface + electrode point cloud (rendered as small icospheres)
n_elec <- 20
electrodes <- structure(
  list(vb = matrix(rnorm(3 * n_elec, sd = 5), 3, n_elec) +
         rowMeans(mesh$vb)[1:3]),
  class = "mesh3d"
)
plot_mesh_polygon(
  mesh = list(mesh, electrodes),
  eye    = c(150, -30, 0),
  lookat = c(0, 0, 0),
  up     = c(0, 0, 1),
  alpha  = c(0.5, 1),
  col    = list("steelblue", "red"),
  cex    = 1.5
)

}
\seealso{
\code{\link{plot_mesh_dotcloud}}, \code{\link{vcg_isosurface}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_raycaster}
\alias{vcg_raycaster}
\title{Cast rays to intersect with mesh}
\usage{
vcg_raycaster(
  x,
  ray_origin,
  ray_direction,
  max_distance = Inf,
  both_sides = FALSE
)
}
\arguments{
\item{x}{surface mesh}

\item{ray_origin}{a matrix with 3 rows or a vector of length 3, the positions
of ray origin}

\item{ray_direction}{a matrix with 3 rows or a vector of length 3, the
direction of the ray, will be normalized to length 1}

\item{max_distance}{positive maximum distance to cast the normalized ray;
default is infinity. Any invalid distances (negative, zero, or \code{NA})
will be interpreted as unset.}

\item{both_sides}{whether to inverse the ray (search both positive and
negative ray directions); default is false}
}
\value{
A list of ray casting results: whether any intersection is found,
position and face normal of the intersection, distance of the ray, and the
index of the intersecting face (counted from 1)
}
\description{
Cast rays to intersect with mesh
}
\examples{

library(ravetools)
sphere <- vcg_sphere(normals = FALSE)
sphere$vb[1:3, ] <- sphere$vb[1:3, ] + c(10, 10, 10)
vcg_raycaster(
  x = sphere,
  ray_origin = array(c(0, 0, 0, 1, 0, 0), c(3, 2)),
  ray_direction = c(1, 1, 1)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mris-curvature.R
\name{mris_curvature}
\alias{mris_curvature}
\title{Estimate per-node curvature of a surface mesh}
\usage{
mris_curvature(mesh, verbose = FALSE)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'} (or coercible via
\code{\link{ensure_mesh3d}}). The mesh \strong{must} be closed, manifold,
and genus-0 (watertight, no boundary or non-manifold edges, single
connected component); \code{mris_curvature} raises an error if such defects
are detected}

\item{verbose}{logical; print progress messages. Default \code{FALSE}}
}
\value{
A named list of four numeric vectors, each of length
\code{ncol(mesh$vb)} (one entry per vertex, in vertex order):
\describe{
  \item{\code{mean}}{Mean curvature \eqn{H = (k_1 + k_2)/2}.}
  \item{\code{gaussian}}{Gaussian curvature \eqn{K = k_1 k_2}.}
  \item{\code{k1}}{First principal curvature (\eqn{k_1 \geq k_2}).}
  \item{\code{k2}}{Second principal curvature.}
}
}
\description{
Estimates, at every vertex of a closed triangular mesh, the local mean
curvature, Gaussian curvature, and the two principal curvatures, by fitting
an osculating quadratic surface to each vertex's neighborhood.
}
\details{
For each vertex, the function:
\enumerate{
  \item builds an orthonormal tangent frame \eqn{(e_1, e_2, n)}, where
    \eqn{n} is the vertex's (already-computed) unit normal;
  \item expresses each neighbor's offset from the vertex in this frame as
    tangential coordinates \eqn{(u, w)} and a height \eqn{h} above the
    tangent plane along \eqn{n};
  \item fits, by least squares over the vertex's 2-ring neighborhood, the
    osculating \verb{paraboloid} \eqn{h = a u^2 + b u w + c w^2};
  \item derives the mean curvature \eqn{H = a + c}, the Gaussian curvature
    \eqn{K = 4ac - b^2}, and the principal curvatures
    \eqn{k_{1,2} = H \pm \sqrt{\max(H^2 - K,\ 0)}}.
}
Because the tangent frame is orthonormal and the \verb{paraboloid} is fitted with
zero gradient at the vertex (by construction, since \eqn{n} is the fitted
normal direction), these are exactly the eigenvalues of the second
fundamental form, i.e. the principal curvatures.

The sign of the result follows the orientation of the per-vertex outward
normal: a locally convex ('\verb{gyrus}-like') patch, where neighbors lie
toward the surface's interior relative to the outward normal, has negative
mean curvature, while a locally concave ('\verb{sulcus}-like') patch has
positive mean curvature. For example, a sphere of radius \eqn{r} with outward-pointing
normals has uniform curvature \eqn{H = -1/r} and \eqn{K = 1/r^2} everywhere.

Vertices whose 2-ring neighborhood is degenerate (fewer than three
neighbors, or neighbor offsets that do not span the tangent plane, e.g.
nearly \verb{collinear}) are reported with all four curvature values set to zero.
}
\examples{


if (is_not_cran()) {

data("left_hippocampus_mask")
mesh <- vcg_isosurface(left_hippocampus_mask)

plot(mesh)

# Fix defects
mesh <- vcg_fix_defects(mesh, verbose = TRUE)

# Smooth
smoothed <- mris_smooth(mesh, verbose = TRUE)

res <- mris_curvature(smoothed)

range(res$k1)

col <- color_ramp_continuous(
  res$k1, clim = c(-1, 1), alpha = TRUE,
  cmap = c("black", "gray", "red"))

plot(smoothed, col = list(col),
     eye = c(-100, 100, 0), up = c(0, 0, 1))

}


}
\references{
Cortical surface-based analysis II: Inflation, flattening, and a
surface-based coordinate system. \emph{NeuroImage}, 9(2), 195-207 (1999).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-ants-io.R
\name{write_ants_warp}
\alias{write_ants_warp}
\alias{read_ants_warp}
\title{Read and write an \verb{'ANTs'} deformation (warp) field}
\usage{
write_ants_warp(
  field,
  file,
  vox2ras = attr(field, "vox2ras"),
  affine = NULL,
  direction = c("forward", "inverse")
)

read_ants_warp(file, recover_affine = FALSE)
}
\arguments{
\item{field}{a \code{(nx, ny, nz, 3)} array of \verb{RAS} displacements; the
\code{vox2ras} attribute is used when \code{vox2ras} is not supplied}

\item{file}{path to the warp \code{.nii} / \code{.nii.gz} file}

\item{vox2ras}{the fixed-grid \eqn{4\times 4} voxel-to-\verb{RAS} matrix}

\item{affine}{optional \eqn{4\times 4} \verb{RAS} \verb{affine} to embed (with
reduced precision) in the header; \code{NULL} (default) embeds nothing}

\item{direction}{\code{"forward"} or \code{"inverse"}; recorded as the
header direction flag when \code{affine} is embedded}

\item{recover_affine}{logical; if \code{TRUE}, attempt to decode a hidden
\verb{affine} from the header and attach it as the \code{"transform"} attribute
(with a \code{"direction"} attribute). Default \code{FALSE}}
}
\value{
\code{write_ants_warp} returns \code{file} invisibly;
\code{read_ants_warp} returns the \code{(nx, ny, nz, 3)} \verb{RAS} field with
a \code{"vox2ras"} attribute (and, if recovered, \code{"transform"} /
\code{"direction"} attributes).
}
\description{
\code{write_ants_warp} stores a dense displacement field (such as the
\code{forward_field} / \code{inverse_field} from
\code{\link{register_volume3d}}) as an \pkg{'ANTs'} 5-D warp \code{'NIfTI'}
(\code{intent_code} \code{1007}, \verb{LPS} vectors) readable by
\code{antsApplyTransforms}; \code{read_ants_warp} reads it back. Requires the
suggested \pkg{freesurferformats} package.

The field is stored on disk with displacement vectors in \verb{LPS} (the x
and y components are negated relative to the \verb{RAS} field) and an
\code{sform} equal to \code{vox2ras}, so the file describes the fixed grid's
physical space exactly as \pkg{'ANTs'} expects. Optionally the \verb{RAS}
\verb{affine} can be hidden in the header text fields
(\code{descrip}/\code{aux_file}); \pkg{'ANTs'} ignores those, and recovery on
read is opt-in.
}
\seealso{
\code{\link{register_volume3d}}, \code{\link{save_registration}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-notch.R
\name{notch_filter}
\alias{notch_filter}
\title{Apply 'Notch' filter}
\usage{
notch_filter(
  s,
  sample_rate,
  lb = c(59, 118, 178),
  ub = c(61, 122, 182),
  domain = 1
)
}
\arguments{
\item{s}{numerical vector if \code{domain=1} (voltage
signals), or complex vector if \code{domain=0}}

\item{sample_rate}{sample rate}

\item{lb}{filter lower bound of the frequencies to remove}

\item{ub}{filter upper bound of the frequencies to remove;
shares the same length as \code{lb}}

\item{domain}{\code{1} if the input signal is in the
time domain, \code{0} if it is in the frequency domain}
}
\value{
filtered signal in time domain (real numerical
vector)
}
\description{
Apply 'Notch' filter
}
\details{
Mainly used to remove electrical line frequencies
at 60, 120, and 180 \code{Hz}.
}
\examples{


time <- seq(0, 3, 0.005)
s <- sin(120 * pi * time) + rnorm(length(time))

# Welch periodogram shows a peak at 60Hz
pwelch(s, 200, plot = 1, log = "y")

# notch filter to remove 60Hz
s1 <- notch_filter(s, 200, lb = 59, ub = 61)
pwelch(s1, 200, plot = 2, log = "y", col = "red")


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_subset_vertex}
\alias{vcg_subset_vertex}
\title{Subset mesh by vertex}
\usage{
vcg_subset_vertex(x, selector)
}
\arguments{
\item{x}{surface mesh}

\item{selector}{logical vector (must not contain NA), and length must be
consistent with the number of vertices in \code{x}: which nodes are
to be kept}
}
\value{
A triangular mesh of class \code{'mesh3d'}, a subset of \code{x}
}
\description{
Subset mesh by vertex
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

sphere <- vcg_sphere()

nv <- ncol(sphere$vb)

selector <- seq_len(nv) > (nv / 2)

sub <- vcg_subset_vertex(sphere, selector)

if(is_not_cran()) {
  rgl_view({

    # subset sphere will be displayed in red
    rgl_call("shade3d", sub, col = 'red')

    # Original sphere will be displayed as wireframe
    rgl_call("wire3d", sphere, col = (2 - selector))

  })
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_subdivision}
\alias{vcg_subdivision}
\title{Sub-divide (up-sample) a triangular mesh}
\usage{
vcg_subdivision(mesh, method = c("edge", "barycenter"))
}
\arguments{
\item{mesh}{triangular mesh stored as object of class 'mesh3d'.}

\item{method}{either \code{'edge'} (default) to add new mid-point vertices to
edge, or \code{'barycenter'} to add new vertices at face \code{'Bary'}
centers.}
}
\value{
An object of class "mesh3d"
}
\description{
Up-sample a triangular mesh by adding a vertex at each edge or face center.
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

mesh <- plane_geometry()

# default
mesh_edge <- vcg_subdivision(mesh, "edge")

# barycenter
mesh_face <- vcg_subdivision(mesh, "barycenter")

if(is_not_cran()) {

  rgl_view({
    rgl_call("wire3d", mesh, col = 1)
    rgl_call("wire3d", mesh_edge, col = 2)
    rgl_call("wire3d", mesh_face, col = 3)
  })


}



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-filtfilt.R
\name{filtfilt}
\alias{filtfilt}
\title{Forward and reverse filter a one-dimensional signal}
\usage{
filtfilt(b, a = 1, x)
}
\arguments{
\item{b}{one-dimensional real numerical vector, the moving-average
coefficients of an \code{ARMA} filter; alternatively, a \code{Sos}
(second-order sections) object from \code{gsignal}, in which case
\code{a} is ignored and \code{gsignal::filtfilt} is used internally}

\item{a}{the auto-regressive (recursive) coefficients of an \code{ARMA}
filter; ignored when \code{b} is a \code{Sos} object}

\item{x}{numerical vector or matrix input (real value)}
}
\value{
The filtered signal, normally the same length as the input signal
   \code{x}.
}
\description{
The result has been tested against 'Matlab' \code{filtfilt}
   function. Currently this function only supports one filter at a time.
}
\examples{

t <- seq(0, 1, by = 0.01)
x <- sin(2 * pi * t * 2.3)
bf <- gsignal::butter(2, c(0.15, 0.3))

res <- filtfilt(bf$b, bf$a, x)

## Matlab (2022a) equivalent:
# t = [0:0.01:1];
# x = sin(2 * pi * t * 2.3);
# [b,a] = butter(2,[.15,.3]);
# res = filtfilt(b, a, x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mris-inflate.R
\name{mris_inflate}
\alias{mris_inflate}
\title{Inflate a cortical surface mesh}
\usage{
mris_inflate(
  mesh,
  n_averages = 16L,
  niterations = 10L,
  l_spring_norm = 1,
  l_dist = 0.1,
  momentum = 0.9,
  dt = 0.9,
  desired_rms = 0.015,
  scale_brain = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}, typically a white-matter
surface in millimeters. The mesh \strong{must} be closed, manifold, and
genus-0 (watertight, no boundary or non-manifold edges, single connected
component) - this is a hard precondition of the underlying algorithm. An
open or defective mesh (e.g. a raw surface extracted via
\code{\link{vcg_isosurface}}, which often contains small extraction
cracks) will silently inflate into a distorted, "sausage"-shaped result
rather than a smooth sphere-like surface, and \code{mris_inflate} will
raise an error if such defects are detected.}

\item{n_averages}{starting number of gradient-averaging passes (outer loop);
halved each level down to 0. Default \code{16}.}

\item{niterations}{number of inner iterations per averaging level.
Default \code{10}.}

\item{l_spring_norm}{normalized spring term coefficient. Default \code{1.0}.}

\item{l_dist}{distance-preservation coefficient (before per-level scaling).
Default \code{0.1}.}

\item{momentum}{momentum coefficient. Default \code{0.9}.}

\item{dt}{time step. Default \code{0.9}.}

\item{desired_rms}{target \verb{RMS} tangent-plane height for early stopping
of each inner loop. Default \code{0.015} (mm).}

\item{scale_brain}{logical; whether to scale the mesh to the canonical
surface area (110,000 \eqn{mm^2}) before inflation. Set \code{FALSE} only
for non-standard inputs. Default \code{TRUE}.}

\item{verbose}{logical; print per-iteration progress. Default \code{FALSE}.}
}
\value{
A named list:
\describe{
  \item{\code{mesh}}{Inflated surface as a \code{'mesh3d'} object with
    \code{vb}, \code{it}, and \code{normals}.}
  \item{\code{sulc}}{Numeric vector of the per-vertex depth values described
    above (zero-mean, in mesh units).}
}
}
\description{
Iteratively relaxes a closed cortical-surface mesh into a smoother, more
compact "inflated" shape, while tracking how far each vertex moves inward
along the surface normal as it goes. Returns both the inflated mesh and
this per-vertex depth map (\code{sulc}).
}
\details{
The implementation follows the inflation procedure described in the
literature (see \strong{References}):
\enumerate{
  \item Optionally rescale the mesh to a canonical surface area
    (110,000 \eqn{mm^2}, the normalization target used by the reference
    procedure).
  \item Outer loop: a neighborhood-averaging size \code{n_averages} is
    halved at each level (16, 8, ..., 0).
  \item Inner loop (\code{niterations} repetitions per level):
    \describe{
      \item{Distance term}{Restoring force pulling each vertex back towards
        its original distances to its neighbors, with coefficient
        \code{l_dist * sqrt(n_averages)}.}
      \item{Gradient averaging}{Smooth the distance-term gradient over
        \code{n_averages} neighborhood passes before adding the spring term.}
      \item{Normalized spring term}{Laplacian (neighbor-averaging) smoothing
        scaled by \code{sqrt(orig_area / current_area)}, with coefficient
        \code{l_spring_norm}; added \emph{after} gradient averaging.}
      \item{Momentum step}{Update vertex positions using momentum
        integration with a 1 mm per-step displacement cap.}
      \item{Depth accumulation}{Accumulate the normal-projected component
        of each step into the per-vertex depth map (\code{sulc}).}
    }
  \item Center the mesh, rescale it, and zero-mean the depth map (\code{sulc}).
}
}
\examples{
if (is_not_cran()) {


data("left_hippocampus_mask")
mesh <- vcg_isosurface(left_hippocampus_mask)

# Fix defects
mesh <- vcg_fix_defects(mesh, verbose = TRUE)

# Center the mesh
mesh$vb[1:3, ] <- mesh$vb[1:3, ] - rowMeans(mesh$vb[1:3, ])

# Inflate the surface while keeping the node distances
result <- mris_inflate(mesh, n_averages = 4L, niterations = 5L,
                       scale_brain = FALSE, verbose = TRUE)

# Visualize with the sulcal values
pal <- colorRampPalette(c("black", "gray", "red"))(128)
col <- pal[pmax(pmin(round(result$sulc * 10 + 64), 128), 1)]

oldpar <- par(mfrow = c(1, 2))
on.exit({ par(oldpar) })

plot(
  mesh, col = col,
  eye = c(0, 100, 0),
  up = c(1, 0, 0))

plot(
  result$mesh, col = col,
  eye = c(0, 100, 0),
  up = c(1, 0, 0))


}

}
\references{
Cortical surface-based analysis II: Inflation, flattening, and a
surface-based coordinate system. \emph{NeuroImage}, 9(2), 195-207 (1999).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-fir.R
\name{firls}
\alias{firls}
\title{Least-squares linear-phase \code{FIR} filter design}
\usage{
firls(N, freq, A, W = NULL, ftype = "", legacy = FALSE)
}
\arguments{
\item{N}{filter order, must be even (if odd, then will be increased by one)}

\item{freq}{vector of frequency points in the range from 0 to 1, where 1
corresponds to the \code{Nyquist} frequency.}

\item{A}{vector of the same length as \code{freq} containing the desired
amplitude at each of the points specified in \code{freq}.}

\item{W}{weighting function that contains one value for each band that
weights the mean squared error in that band. \code{W} must be half the
length of \code{freq}.}

\item{ftype}{transformer type; default is \code{""}; alternatively,
\code{'h'} or \code{'hilbert'} for 'Hilbert' transformer.}

\item{legacy}{whether to use the legacy implementations, which uses
\code{\link{qr.solve}} instead of faster \code{\link{chol}}}
}
\value{
The \code{FIR} filter coefficients with class \code{'Arma'}.
   The moving average coefficient is a vector of length \code{n+1}.
}
\description{
Produce a linear phase filter from the weighted mean squared such that error
   in the specified bands is minimized.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\docType{data}
\name{left_hippocampus_mask}
\alias{left_hippocampus_mask}
\title{Left 'Hippocampus' of 'N27-Collin' brain}
\format{
A three-mode integer mask array with values of \code{1} ('Hippocampus')
and \code{0} (other brain tissues)
}
\usage{
left_hippocampus_mask
}
\description{
Left 'Hippocampus' of 'N27-Collin' brain
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_subdivide_max_edge_length}
\alias{vcg_subdivide_max_edge_length}
\title{Selectively subdivide mesh edges that exceed a length threshold}
\usage{
vcg_subdivide_max_edge_length(mesh, max_edge_len, max_iter = NULL)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.}

\item{max_edge_len}{maximum allowed edge length (same units as mesh
coordinates).}

\item{max_iter}{maximum number of refinement passes. When \code{NULL}
(default), derived automatically from the current maximum edge length:
\code{ceiling(log2(current_max / max_edge_len)) + 1L}, the minimum number
of bisections needed in the worst case. The loop also terminates early once
no edge exceeds the threshold.}
}
\value{
An object of class \code{"mesh3d"} with all edges at most
  \code{max_edge_len} long (provided \code{max_iter} was sufficient).
}
\description{
Up-sample a triangular mesh by iteratively splitting only edges longer than
\code{max_edge_len}. Each long edge is split at its midpoint; the new vertex
is connected to the opposite corner of every adjacent face. Iteration stops
when no edge exceeds the threshold or \code{max_iter} passes are exhausted.

This is far cheaper than \code{\link{vcg_subdivision}} when most edges are
already short and only a small fraction need splitting.
}
\note{
The mesh must be manifold. Run \code{\link{vcg_fix_defects}} first if
  the mesh has boundary edges or non-manifold vertices.
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{
if (is_not_cran()) {

  sphere <- vcg_sphere()
  cur_max <- vcg_max_edge_length(sphere)
  sphere2 <- vcg_subdivide_max_edge_length(sphere, max_edge_len = cur_max * 0.4)
  vcg_max_edge_length(sphere2)  # should be <= cur_max * 0.4

}

}
\seealso{
\code{\link{vcg_max_edge_length}}, \code{\link{vcg_subdivision}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmf.R
\name{naive_nmf}
\alias{naive_nmf}
\title{A naive implementation of non-negative matrix factorization}
\usage{
naive_nmf(x, k, tol = c(1e-04, 1e-08), max_iters = 10000, verbose = TRUE)
}
\arguments{
\item{x}{a matrix, or can be converted into a matrix; all negative or missing
values will be treated as zero}

\item{k}{decomposition rank}

\item{tol}{stop criteria, a numeric of two; the first number is the
tolerance for root-mean-squared residuals, relative to the largest number in
\code{x}; the second number is the tolerance for weight differences; any
stopping criteria met will result in the stop of iteration}

\item{max_iters}{maximum iterations}

\item{verbose}{whether to report the progress; logical or a positive integer
(of step intervals)}
}
\value{
A list of weights (non-negative template matrix \code{W} and
non-negative \code{H}) and errors (root mean squared error of fitted,
matrix \code{W}, and \code{W} versus their previous iteration, respectively).
}
\description{
A pure-R vanilla implementation assuming inputs are non-negative matrices
without \code{NA}.
}
\examples{



x <- stats::toeplitz(.9 ^ (0:31))

nmf <- naive_nmf(x, k = 7, verbose = FALSE)

fitted <- nmf$W \%*\% nmf$H

oldpar <- par(mfrow = c(1, 2))
on.exit({ par(oldpar )})

image(x, zlim = c(0, 1), main = "Input")
image(fitted, zlim = c(0, 1),
      main = sprintf("Fitted with rank=\%d", nmf$k))




}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crp_cluster.R
\name{crp_cluster}
\alias{crp_cluster}
\title{Cluster electrodes by their canonical \verb{CRP} response shape}
\usage{
crp_cluster(
  crp_list,
  paired = TRUE,
  time_window = c(0, NA),
  n_clusters = NULL,
  initial_rank = NULL,
  zeta_threshold = 1,
  null_class = TRUE,
  nmf_max_iters = 10000,
  nmf_tol = c(1e-04, 1e-08),
  verbose = TRUE
)
}
\arguments{
\item{crp_list}{a named list of \code{ravetools_crp} objects (one per
electrode) from \code{\link{crp}}; list names are used as electrode labels.
Time axes need not match - the common overlapping domain is used (an error is
raised if the domains do not overlap). See \sQuote{Details}.}

\item{paired}{logical; \code{TRUE} (the default) builds the trial-level
\verb{BPC} t-statistic similarity and requires the single trials in
\code{.data$V}; \code{FALSE} uses the cosine similarity of the \code{C_full}
curves. See \sQuote{Details}.}

\item{time_window}{numeric \code{c(lo, hi)} analysis window, clipped into the
common domain; an \code{NA} extends to that domain edge. Defaults to
\code{c(0, NA)} (stimulus onset to the end of the domain).}

\item{n_clusters}{integer or \code{NULL}; if given, fixes the number of
clusters and skips the automatic selection (which otherwise uses
\code{zeta_threshold}).}

\item{initial_rank}{integer or \code{NULL}; the starting \verb{NMF} rank for
automatic selection. \code{NULL} uses \code{min(length(crp_list) - 1L, 10L)}.}

\item{zeta_threshold}{numeric \eqn{> 0}; rank-selection cutoff on \eqn{\zeta}
(see \sQuote{Details}); smaller values give fewer clusters. Defaults to
\code{1}, matching the \verb{BPC} reference.}

\item{null_class}{logical; if \code{TRUE} (the default) electrodes whose
winner-take-all loading falls below \eqn{1/(2\sqrt{N})} are left unassigned
(\code{NA}); if \code{FALSE} every electrode is forced into a cluster. See
\sQuote{Details}.}

\item{nmf_max_iters, nmf_tol}{passed to \code{\link{naive_nmf}}.}

\item{verbose}{logical; whether to report progress.}
}
\value{
A named list of class \code{ravetools_crp_cluster}:
\describe{
\item{\code{clusters}}{Integer vector, the cluster index assigned to each
  electrode (winner-take-all over the row-normalized \verb{NMF} loadings).
  When \code{null_class = TRUE}, electrodes whose top loading falls below
  \eqn{1/(2\sqrt{N})} are left unassigned (\code{NA}).}
\item{\code{basis_curves}}{Numeric matrix, time \eqn{\times} number of
  clusters; column \eqn{q} is the basis profile curve \eqn{B_q(t)} for
  cluster \eqn{q}, the first linear-kernel \verb{PCA} component of its member
  \code{C_full} curves, sign-oriented to the cluster mean.}
\item{\code{basis_times}}{Numeric vector, the time axis for
  \code{basis_curves} (the common overlapping time axis, restricted to
  \code{time_window}).}
\item{\code{similarity}}{Numeric matrix, the electrode-by-electrode
  similarity \eqn{\Xi} that was factorized (non-negative, scaled to a maximum
  of one).}
\item{\code{nmf}}{The \code{\link{naive_nmf}} result for the selected rank.}
\item{\code{n_clusters}}{Integer, the number of clusters found.}
\item{\code{domain}}{Numeric \code{c(lo, hi)}, the overlapping time domain
  shared by all electrodes.}
\item{\code{paired},\code{time_window}}{The settings used; \code{time_window}
  is the effective window after clipping into \code{domain}.}
}
}
\description{
Groups recording electrodes by the \emph{shape} of their canonical evoked
response. Each electrode is summarized by \code{\link{crp}} into one
amplitude-normalized canonical shape; an electrode-by-electrode similarity
matrix is factorized with \code{\link{naive_nmf}} to find clusters that share
a response shape, and a representative basis profile curve is extracted per
cluster. This applies the basis-profile-curve (\verb{BPC}) approach
\emph{across electrodes} rather than across stimulation sites; see
\sQuote{References}.
}
\details{
\enumerate{
\item \strong{Common window.} Electrodes may have different time axes
  (\code{\link{crp}} drops \code{NA} samples); the overlapping time domain is
  used, \code{time_window} is clipped into it, and each electrode is subset on
  the fly with no interpolation.
\item \strong{Similarity.} With \code{paired = TRUE} each entry is the
  one-sample t-statistic, across an electrode pair's common trials, of the
  per-trial \verb{CCEP} cross-projection of the \eqn{1/\alpha}-rescaled
  responses; negatives are zeroed, the matrix made symmetric and scaled to a
  maximum of one. With \code{paired = FALSE} it is the cosine cross-projection
  of the \code{C_full} curves.
\item \strong{Rank selection.} \code{\link{naive_nmf}} factorizes the
  similarity at rank \eqn{Q} (from \code{initial_rank}); each rank is re-run a
  few times and the lowest-error fit kept. \eqn{Q} is reduced while
  \eqn{\zeta} - the sum of the upper off-diagonal of the row-normalized
  \eqn{HH^\top} - exceeds \code{zeta_threshold}.
\item \strong{Assignment.} Each electrode takes its winner-take-all cluster
  over the normalized \verb{NMF} loadings; with \code{null_class}, loadings
  below \eqn{1/(2\sqrt{N})} are left unassigned.
\item \strong{Basis curves.} Per cluster, the first linear kernel \verb{PCA}
  component of its members' \code{C_full} curves (as in \code{\link{crp}}).
}
}
\examples{
# Four response shapes; shapes 3 and 4 differ only in amplitude, so they
# cluster together once shapes are amplitude-normalized.
\donttest{

n_time <- 300L
tt <- seq(-0.2, 1, length.out = n_time)
shapes <- list(
  exp(-((tt - 0.08) / 0.03)^2) - 0.5 * exp(-((tt - 0.18) / 0.04)^2),
  exp(-((tt - 0.38) / 0.03)^2) - 0.5 * exp(-((tt - 0.50) / 0.04)^2),
  exp(-((tt - 0.70) / 0.04)^2),
  exp(-((tt - 0.70) / 0.04)^2) * 2
)

# 4 electrodes per shape, each parameterized with crp()
crp_list <- list()
for (g in seq_along(shapes)) {
  for (e in seq_len(4L)) {
    V <- outer(shapes[[g]], runif(15L, 0.5, 1.5)) +
      matrix(rnorm(n_time * 15L, sd = 0.15), n_time, 15L)
    crp_list[[sprintf("elec_\%d_\%d", g, e)]] <-
      crp(V, tt, remove_artifacts = FALSE)
  }
}

res <- crp_cluster(crp_list, verbose = TRUE)
res$n_clusters
table(res$clusters)
plot(res)

}

}
\references{
The \verb{BPC} method is described in \doi{10.1371/journal.pcbi.1008710}; the
underlying \verb{CRP} method in \doi{10.1371/journal.pcbi.1011105}.
}
\seealso{
\code{\link{crp}}, \code{\link{naive_nmf}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/carla.R
\name{carla}
\alias{carla}
\title{Common Average Re-referencing by Least Anti-Correlation (CARLA)}
\usage{
carla(
  x,
  nboot = 100L,
  sensitive = FALSE,
  min_size = NULL,
  absolute_rank = FALSE,
  virtual_reference = FALSE
)
}
\arguments{
\item{x}{numeric array of shape \code{channels x time x trials}; if a
matrix is provided, it is treated as \code{channels x time} (single trial).
The signal must already be cropped to the responsive time window
(typically \code{0.01 s} to \code{0.3 s} post-stimulus).}

\item{nboot}{integer, number of bootstrapped trial resamplings used to
estimate the optimization statistic; defaults to \code{100}. Ignored when
a single trial is supplied.}

\item{sensitive}{logical; if \code{TRUE} (and more than one trial is
supplied), the more sensitive cutoff is used, corresponding to the channel
count just before the first statistically significant decrease in the mean
anti-correlation curve. The default \code{FALSE} returns the global maximum
of that curve.}

\item{min_size}{integer, minimum subset size considered when
\code{sensitive = TRUE}; defaults to
\code{max(2, ceiling(0.1 * nchan_good))}, where \code{nchan_good} is
the number of usable channels after the bad-channel mask has been
applied (see Details). Floored at 2 because subset size 1 is never
evaluated.}

\item{absolute_rank}{logical; if \code{FALSE} (the default), the
per-channel ranking statistic is the signed mean of the off-diagonal
trial-to-trial covariances, as in the original CARLA manuscript. This
assumes a phase-locked evoked response, so trial-pair covariances on
responsive channels are positive and a clean signed mean separates them
from non-responsive channels. Set to \code{TRUE} to use the mean of the
\strong{absolute} covariances instead, which is more robust to evoked
responses whose polarity flips across trials (e.g. alternating-polarity
stimulation, biphasic \verb{CCEPs} with jitter), at the cost of an upward
bias on the non-responsive floor. Ignored when only one trial is supplied
(the per-channel variance is used in that case).}

\item{virtual_reference}{logical; if \code{TRUE}, runs the modified
CARLA before the iterative subset
evaluation: rank channels once, designate the channel with the median
rank as a "virtual reference", subtract its signal from every channel,
then re-rank on the subtracted data. The virtual channel itself is
pinned to the very end of the new ordering so it is never picked into
the CAR until the final subset size. This is intended for data where
the recording reference is contaminated by stimulation artifact or
evoked activity: a contaminated reference makes every channel look
artificially similar and biases the covariance/variance ranking statistic,
but subtracting a mid-rank proxy of that contamination unbiases the
ranking so genuinely responsive channels rise to the top. Default
\code{FALSE} reproduces the original CARLA algorithm bit-for-bit.}
}
\value{
A list with the following elements:
\describe{
\item{\code{channels}}{integer vector, sorted indices (1-based) of the
channels chosen to construct the common average reference.}
\item{\code{car}}{numeric matrix of shape \code{time x trials} (or numeric
vector when only one trial is supplied) holding the common average signal
computed from \code{channels}. Subtract this from each channel of the
original signal to obtain the re-referenced data.}
\item{\code{order}}{integer vector, indices of the \strong{good}
channels sorted in increasing order of the ranking statistic. Bad
channels (zero variance / all-\code{NA}; see Details) are excluded.}
\item{\code{vars}}{numeric vector of length \code{nchan} containing the
per-channel ranking statistic (mean cross-trial covariance, or variance
for a single trial). Bad channels keep their raw value (zero or
\code{NA}) so callers can audit the mask.}
\item{\code{n_optimum}}{integer, the optimal subset size selected
(indexes into \code{order}).}
\item{\code{zmin_mean}}{numeric matrix of shape
\code{length(order) x nboot} (or a length-\code{length(order)} vector
for a single trial / \code{nboot = 1}) holding, for each subset size
and bootstrap, the mean Fisher z-transformed correlation of the most
globally anti-correlated unreferenced channel against the candidate CAR.
Row 1 is always \code{NA} (subset size 1 is not evaluated).}
\item{\code{bad_channels}}{integer vector of channel indices that were
excluded from the analysis because their ranking statistic was zero or
\code{NA} (flat / dead channels, plus the virtual channel itself when
\code{virtual_reference = TRUE}).}
\item{\code{virtual_channel}}{integer, the index (1-based) of the
channel used as the virtual reference when \code{virtual_reference =
TRUE}; \code{NA_integer_} otherwise.}
\item{\code{vars1}}{numeric vector of the first-pass ranking statistic
when \code{virtual_reference = TRUE} (the post-subtraction statistic is
returned in \code{vars}); \code{NULL} otherwise.}
}
}
\description{
Selects an optimal subset of channels to use as the common average reference
(CAR) for cortico-cortical evoked potential (\verb{CCEP}) data, following the
\verb{CARLA} (see 'Reference' and 'Citation'). Channels are ranked in
increasing order of their cross-trial covariance (or variance when only one
trial is available); subsets are then iteratively grown and the size that
yields the least anti-correlation between the candidate reference and the
remaining unreferenced channels is selected as optimal.
}
\details{
The function is a faithful port of the core \code{CARLA.m} routine; it does
not perform notch filtering, time-window cropping,
or grouping by stimulation site. Those steps belong to the surrounding
preprocess pipeline (see the example below).

For each candidate subset size \eqn{n = 2, \ldots, N}{n = 2, ..., N}, the
candidate reference is computed as the channel-wise mean of the \eqn{n}
lowest-ranked channels. Each of those \eqn{n} channels is then correlated,
in its unreferenced form, against every channel of the candidate
re-referenced subset. The resulting Pearson correlations are
Fisher z-transformed; the row corresponding to the most globally
anti-correlated channel is recorded as \code{zmin}. The optimal \eqn{n}
is the one that maximizes (i.e. makes least negative) the mean of
\code{zmin}.

\strong{Bad-channel mask.} Channels whose ranking statistic is exactly
zero or \code{NA} are flagged as "bad" and excluded both from the CAR
candidate pool and from the target-channel correlation rows. This
catches flat / dead channels (constant signal, no usable variance) and
, when \code{virtual_reference = TRUE}, automatically removes the virtual
channel itself, since subtracting it from itself yields a zero trace.
All references to channel indices in the returned \code{order},
\code{n_optimum}, and \code{zmin_mean} are with respect to the
\strong{good} channels only; \code{vars} is full-length so callers can
inspect the raw scores.
}
\examples{

# ---- Simulate a small CCEP-like dataset --------------------------------
# 16 channels, 12 trials, sampled at 1 kHz, 0.5 s peri-stimulus epoch.
# Channels 1:4 are "responsive" (carry an evoked potential); the rest are
# noise-only and should make up the optimal CAR.
srate <- 1000
tt <- seq(-0.1, 0.4 - 1 / srate, by = 1 / srate)   # time, seconds
nchan <- 16
ntrial <- 30
resp_ch <- 1:4
noise_ch <- 4:7

# Evoked potential template: damped sinusoid starting at t = 0
ep <- ifelse(tt >= 0,
             80 * exp(-tt / 0.05) * sin(2 * pi * 12 * tt),
             0)

# channels x time x trials
x_full <- array(rnorm(nchan * length(tt) * ntrial, sd = 5),
                dim = c(nchan, length(tt), ntrial))
for (ch in resp_ch) {
  for (k in seq_len(ntrial)) {
    x_full[ch, , k] <- x_full[ch, , k] + ep * runif(1, -0.8, 1.2)
  }
}
for (ch in noise_ch) {
  for (k in seq_len(ntrial)) {
    tmp <- x_full[ch, , k]
    x_full[ch, , k] <- tmp + sign(ch \%\% 2 - 0.5) * 5 *
      runif(length(tmp), 0.8, 1.2)
  }
}
# Add artifacts common to all channels and trials
artifacts <- 6 * sin(2 * pi * 60 * tt) + 7 * sin(2 * pi * 24 * tt)
x_full <- sweep(x_full, 2L, artifacts, "+")

# ---- 1. Notch filter line noise (per channel, per trial) ---------------
# The CARLA paper notch-filters before ranking; the re-reference itself
# is applied to the original (unfiltered) signal.
x_clean <- x_full
for (ch in seq_len(nchan)) {
  for (k in seq_len(ntrial)) {
    x_clean[ch, , k] <- notch_filter(
      x_full[ch, , k], sample_rate = srate,
      lb = c(59, 119, 179), ub = c(61, 121, 181)
    )
  }
}

# ---- 2. Crop to the responsive window (0.01 s to 0.3 s post-stim) ------
resp_idx <- which(tt > 0.0 & tt <= 0.3)
x_resp <- x_clean[, resp_idx, , drop = FALSE]

# ---- 3. Run CARLA to pick reference channels ---------------------------
fit <- carla(x_resp, sensitive = TRUE, absolute_rank = TRUE,
             virtual_reference = TRUE)
fit$channels        # selected reference channels (should exclude 1:4)
fit$n_optimum       # number of channels in the optimal CAR

# ---- 4. Re-reference the ORIGINAL (unfiltered) signal ------------------
# mean or median, your choice!
car_full <- apply(x_full[fit$channels, , , drop = FALSE], c(2, 3), mean)

# old-style: using all channels for CAR
car_old <- apply(x_full, c(2, 3), mean)

x_reref <- sweep(x_full, c(2, 3), car_full, "-")
x_compare <- sweep(x_full, c(2, 3), car_old, "-")

# ---- 5. Inspect: evoked potential is preserved on responsive channels --
op <- graphics::par(mfrow = c(2, 4), mar = c(4, 4, 2, 1))
ravetools::plot_signals(
  signals = x_full[, , 1],
  sample_rate = srate,
  main = "Trial 1 - (raw)")

ravetools::plot_signals(
  signals = x_clean[, , 1],
  sample_rate = srate,
  main = "Notch-filtered")

ravetools::plot_signals(
  signals = x_reref[, , 1],
  sample_rate = srate,
  main = sprintf("CARLA-ref (n=\%d)", length(fit$channels)))

ravetools::plot_signals(
  signals = x_compare[, , 1],
  sample_rate = srate,
  main = "Conventional CAR for comparison")

col <- adjustcolor(seq_len(nchan))
col[resp_ch] <- adjustcolor(col[resp_ch], alpha.f = 0.2)

graphics::matplot(tt, t(rowMeans(x_full, dims = 2L)),
                  type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
                  main = "Trial-averaged (raw", col = col)

graphics::matplot(tt, t(rowMeans(x_clean, dims = 2L)),
                  type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
                  main = "Notch-filtered", col = col)

graphics::matplot(tt, t(rowMeans(x_reref, dims = 2L)),
                  type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
                  main = "CARLA-referenced", col = col)

graphics::matplot(tt, t(rowMeans(x_compare, dims = 2L)),
                  type = "l", lty = 1, xlab = "Time (s)", ylab = "uV",
                  main = "conventional CAR-referenced", col = col)

graphics::par(op)

}
\references{
The CARLA algorithm (\code{virtual_reference = FALSE}) is described in
\doi{10.1016/j.jneumeth.2024.110153}; the modified CARLA precursor
(\code{virtual_reference = TRUE}) is described in
\doi{10.1016/j.jneumeth.2025.110461}, with a reference implementation
at \url{https://github.com/hharveygit/SPES_reference_contam}. See
\code{citation("ravetools")} for the full bibliographic entries of
both manuscripts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mris-smooth.R
\name{mris_smooth}
\alias{mris_smooth}
\title{Smooth a surface mesh}
\usage{
mris_smooth(
  mesh,
  niterations = 10L,
  npasses = 1L,
  rescale = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.  Must be closed,
manifold, and genus-0 (watertight, no boundary or non-manifold edges,
single connected component) - the same hard precondition as
\code{\link{mris_inflate}} (and for the same reason: an open or
defective mesh introduces a persistent directional bias into the
1-ring averaging).  \code{mris_smooth} will raise an error if such
defects are detected.}

\item{niterations}{number of include-self 1-ring averaging rounds applied
to the vertex positions in each pass.  Default \code{10}.}

\item{npasses}{number of outer smoothing passes.  Default \code{1}.}

\item{rescale}{logical; whether to rescale the surface back to its original
area after each pass.  Default \code{FALSE}.}

\item{verbose}{logical; print per-pass progress.  Default \code{FALSE}.}
}
\value{
The smoothed surface as a \code{'mesh3d'} object with \code{vb},
  \code{it}, and \code{normals}.
}
\description{
Smooths a closed triangular surface mesh by repeatedly replacing each
vertex with the average of itself and its immediate neighbors
(Laplacian/neighbor-averaging smoothing), optionally rescaling the surface
back to its original area afterwards.
}
\details{
For each of \code{npasses} passes, the algorithm:
\enumerate{
  \item Replaces each vertex position by the include-self mean of itself
    and its 1-ring (directly-connected) neighbors, repeated
    \code{niterations} times.
  \item Recomputes vertex normals and total surface area.
  \item Optionally rescales the surface back to its original area (only if
    \code{rescale = TRUE}).
}
}
\examples{

if (is_not_cran()) {

sphere <- vcg_sphere(sub_division = 4L)

# roughen the sphere slightly so smoothing has something to do
sphere$vb[1, ] <- sphere$vb[1, ] * (1 + 0.05 * rnorm(ncol(sphere$vb)))

# Fix defects
sphere <- vcg_fix_defects(sphere)

smoothed <- mris_smooth(sphere, niterations = 5L, verbose = TRUE)

plot_mesh_polygon(
  list(sphere, smoothed),
  alpha = c(0.3, 0.5),
  col = list("gray", "red"),
  main = "Gray: original; red: smoothed"
)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-decimate.R
\name{decimate}
\alias{decimate}
\title{Decimate with 'FIR' or 'IIR' filter}
\usage{
decimate(x, q, n = if (ftype == "iir") 8 else 30, ftype = "fir")
}
\arguments{
\item{x}{signal to be decimated}

\item{q}{integer factor to down-sample by}

\item{n}{filter order used in the down-sampling; default is \code{30}
if \code{ftype='fir'}, or \code{8} if \code{ftype='iir'}}

\item{ftype}{filter type, choices are \code{'fir'} (default) and
\code{'iir'}}
}
\value{
Decimated signal
}
\description{
Decimate with 'FIR' or 'IIR' filter
}
\details{
This function is migrated from \code{gsignal} package,
but with padding and indexing fixed. The results agree with 'Matlab'.
}
\examples{

x <- 1:100
y <- decimate(x, 2, ftype = "fir")
y

# compare with signal package
z <- gsignal::decimate(x, 2, ftype = "fir")

# Compare decimated results
plot(x, type = 'l')
points(seq(1,100, 2), y, col = "green")
points(seq(1,100, 2), z, col = "red")


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{parallel-options}
\alias{parallel-options}
\alias{detect_threads}
\alias{ravetools_threads}
\title{Set or get thread options}
\usage{
detect_threads()

ravetools_threads(n_threads = "auto", stack_size = "auto")
}
\arguments{
\item{n_threads}{number of threads to set}

\item{stack_size}{Stack size (in bytes) to use for worker threads. The
default used for \code{"auto"} is 2MB on 32-bit systems and 4MB on 64-bit
systems.}
}
\value{
\code{detect_threads} returns an integer of default threads that
is determined by the number of \code{CPU} cores; \code{ravetools_threads}
returns nothing.
}
\description{
Set or get thread options
}
\examples{

detect_threads()

ravetools_threads(n_threads = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapse.R
\name{collapse}
\alias{collapse}
\alias{collapse.array}
\title{Collapse array}
\usage{
collapse(x, keep, ...)

\method{collapse}{array}(
  x,
  keep,
  average = TRUE,
  transform = c("asis", "10log10", "square", "sqrt"),
  ...
)
}
\arguments{
\item{x}{A numeric multi-mode tensor (array), without \code{NA}}

\item{keep}{Which dimension to keep}

\item{...}{passed to other methods}

\item{average}{collapse to sum or mean}

\item{transform}{transform on the data before applying collapsing;
choices are \code{'asis'} (no change), \code{'10log10'} (used to calculate
decibel), \code{'square'} (sum-squared), \code{'sqrt'} (square-root and
collapse)}
}
\value{
a collapsed array with values to be mean or summation along
collapsing dimensions
}
\description{
Collapse array
}
\examples{

# Set ncores = 2 to comply to CRAN policy. Please don't run this line
ravetools_threads(n_threads = 2L)

# Example 1
x = matrix(1:16, 4)

# Keep the first dimension and calculate sums along the rest
collapse(x, keep = 1)
rowMeans(x)  # Should yield the same result

# Example 2
x = array(1:120, dim = c(2,3,4,5))
result = collapse(x, keep = c(3,2))
compare = apply(x, c(3,2), mean)
sum(abs(result - compare)) # The same, yield 0 or very small number (1e-10)


\donttest{

ravetools_threads(n_threads = -1)

# Example 3 (performance)

# Small data, no big difference
x = array(rnorm(240), dim = c(4,5,6,2))
microbenchmark::microbenchmark(
  result = collapse(x, keep = c(3,2)),
  compare = apply(x, c(3,2), mean),
  times = 1L, check = function(v) {
    max(abs(range(do.call('-', v)))) < 1e-10
  }
)

# large data big difference
x = array(rnorm(prod(300,200,105)), c(300,200,105,1))
microbenchmark::microbenchmark(
  result = collapse(x, keep = c(3,2)),
  compare = apply(x, c(3,2), mean),
  times = 1L , check = function(v) {
    max(abs(range(do.call('-', v)))) < 1e-10
  })

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\docType{data}
\name{stimulation_signal}
\alias{stimulation_signal}
\title{Sample stimulation recording}
\format{
A list of one-second signal trace and sample rate (30000)
}
\usage{
stimulation_signal
}
\description{
Sample stimulation recording
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast_cov.R
\name{fast_cov}
\alias{fast_cov}
\title{Calculate massive covariance matrix in parallel}
\usage{
fast_cov(x, y = NULL, col_x = NULL, col_y = NULL, df = NA)
}
\arguments{
\item{x}{a numeric vector, matrix or data frame; a matrix is highly
recommended to maximize the performance}

\item{y}{NULL (default) or a vector, matrix or data frame with compatible
dimensions to x; the default is equivalent to \code{y = x}}

\item{col_x}{integers indicating the subset indices (columns) of \code{x} to
calculate the covariance, or \code{NULL} to include all the columns; default
is \code{NULL}}

\item{col_y}{integers indicating the subset indices (columns) of \code{y} to
calculate the covariance, or \code{NULL} to include all the columns; default
is \code{NULL}}

\item{df}{a scalar indicating the degrees of freedom; default is
\code{nrow(x)-1}}
}
\value{
A covariance matrix of \code{x} and \code{y}. Note that there is no
\code{NA} handling. Any missing values will lead to \code{NA} in the
resulting covariance matrices.
}
\description{
Speed up covariance calculation for large matrices. The
default behavior is the same as \code{\link{cov}} (\code{'pearson'},
no \code{NA} handling).
}
\examples{

# Set ncores = 2 to comply to CRAN policy. Please don't run this line
ravetools_threads(n_threads = 2L)

x <- matrix(rnorm(400), nrow = 100)

# Call `cov(x)` to compare
fast_cov(x)

# Calculate covariance of subsets
fast_cov(x, col_x = 1, col_y = 1:2)

\donttest{

# Speed comparison, better to use multiple cores (4, 8, or more)
# to show the differences.

ravetools_threads(n_threads = -1)
x <- matrix(rnorm(100000), nrow = 1000)
microbenchmark::microbenchmark(
  fast_cov = {
    fast_cov(x, col_x = 1:50, col_y = 51:100)
  },
  cov = {
    cov(x[,1:50], x[,51:100])
  },
  unit = 'ms', times = 10
)

}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-fir.R
\name{fir1}
\alias{fir1}
\title{Window-based \code{FIR} filter design}
\usage{
fir1(
  n,
  w,
  type = c("low", "high", "stop", "pass", "DC-0", "DC-1"),
  window = hamming,
  scale = TRUE,
  hilbert = FALSE
)
}
\arguments{
\item{n}{filter order}

\item{w}{band edges, non-decreasing vector in the range 0 to 1, where 1 is
the \code{Nyquist} frequency. A scalar for high-pass or low-pass filters,
a vector pair for band-pass or band-stop, or a vector for an
alternating pass/stop filter.}

\item{type}{type of the filter, one of \code{"low"} for a low-pass filter,
\code{"high"} for a high-pass filter, \code{"stop"} for a stop-band
(band-reject) filter, \code{"pass"} for a pass-band filter, \code{"DC-0"}
for a band-pass as the first band of a multi-band filter, or \code{"DC-1"}
for a band-stop as the first band of a multi-band filter; default \code{"low"}}

\item{window}{smoothing window function or a numerical vector. The filter is
the same shape as the smoothing window. When \code{window} is a function,
\code{window(n+1)} will be called, otherwise the length of the window
vector needs to have length of \code{n+1}; default: \code{hamming}}

\item{scale}{whether to scale the filter; default is true}

\item{hilbert}{whether to use 'Hilbert' transformer; default is false}
}
\value{
The \code{FIR} filter coefficients with class \code{'Arma'}.
The moving average coefficient is a vector of length \code{n+1}.
}
\description{
Generate a \code{fir1} filter that is checked against \code{Matlab}
\code{fir1} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{internal_rave_function}
\alias{internal_rave_function}
\title{Get external function from 'RAVE'}
\usage{
internal_rave_function(name, pkg, inherit = TRUE, on_missing = NULL)
}
\arguments{
\item{name}{function or variable name}

\item{pkg}{'RAVE' package name}

\item{inherit}{passed to \code{\link{get0}}}

\item{on_missing}{default value to return of no function is found}
}
\value{
Function object if found, otherwise \code{on_missing}.
}
\description{
Internal function used for examples relative to 'RAVE' project and should
not be used directly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convolve.R
\name{grow_volume}
\alias{grow_volume}
\title{Grow volume mask}
\usage{
grow_volume(volume, x, y = x, z = x, threshold = 0.5)
}
\arguments{
\item{volume}{volume mask array, must be 3-dimensional array}

\item{x, y, z}{size of grow along each direction}

\item{threshold}{threshold after convolution}
}
\value{
A binary volume mask
}
\description{
Grow volume mask
}
\examples{


oldpar <- par(mfrow = c(2,3), mar = c(0.1,0.1,3.1,0.1))

mask <- array(0, c(21,21,21))
mask[11,11,11] <- 1
image(mask[11,,], asp = 1,
      main = "Original mask", axes = FALSE)
image(grow_volume(mask, 2)[11,,], asp = 1,
      main = "Dilated (size=2) mask", axes = FALSE)
image(grow_volume(mask, 5)[11,,], asp = 1,
      main = "Dilated (size=5) mask", axes = FALSE)

mask[11, sample(11,2), sample(11,2)] <- 1
image(mask[11,,], asp = 1,
      main = "Original mask", axes = FALSE)
image(grow_volume(mask, 2)[11,,], asp = 1,
      main = "Dilated (size=2) mask", axes = FALSE)
image(grow_volume(mask, 5)[11,,], asp = 1,
      main = "Dilated (size=5) mask", axes = FALSE)

par(oldpar)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_uniform_remesh}
\alias{vcg_uniform_remesh}
\title{Sample a surface mesh uniformly}
\usage{
vcg_uniform_remesh(
  x,
  voxel_size = NULL,
  offset = 0,
  discretize = FALSE,
  multi_sample = FALSE,
  absolute_distance = FALSE,
  merge_clost = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{x}{surface}

\item{voxel_size}{'voxel' size for space 'discretization'}

\item{offset}{offset position shift of the new surface from the input}

\item{discretize}{whether to use step function(\code{TRUE}) instead of
linear interpolation (\code{FALSE}) to calculate the position of the
intersected edge of the marching cube; default is \code{FALSE}}

\item{multi_sample}{whether to calculate multiple samples for more accurate
results (at the expense of more computing time) to remove artifacts; default
is \code{FALSE}}

\item{absolute_distance}{whether an unsigned distance field should be
computed. When set to \code{TRUE}, non-zero offsets is to be set, and
double-surfaces will be built around the original surface, like a sandwich.}

\item{merge_clost}{whether to merge close vertices; default is \code{TRUE}}

\item{verbose}{whether to verbose the progress; default is \code{TRUE}}
}
\value{
A triangular mesh of class \code{'mesh3d'}
}
\description{
Sample a surface mesh uniformly
}
\examples{

sphere <- vcg_sphere()
mesh <- vcg_uniform_remesh(sphere, voxel_size = 0.45)

if(is_not_cran()) {

rgl_view({

  rgl_call("mfrow3d", 1, 2)

  rgl_call("title3d", "Input")
  rgl_call("wire3d", sphere, col = 2)
  rgl_call("next3d")

  rgl_call("title3d", "Re-meshed to 0.1mm edge distance")
  rgl_call("wire3d", mesh, col = 3)
})

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image-resample3d.R
\name{resample_3d_volume}
\alias{resample_3d_volume}
\title{Sample '3D' volume in the world (anatomical \code{'RAS'}) space}
\usage{
resample_3d_volume(
  x,
  new_dim,
  vox2ras_old,
  vox2ras_new = vox2ras_old,
  na_fill = NA,
  interpolation = c("nearest", "trilinear", "bspline")
)
}
\arguments{
\item{x}{image (volume) to be sampled: \code{dim(x)} must have length of 3}

\item{new_dim}{target dimension, integers of length 3}

\item{vox2ras_old}{from volume index (column-row-slice) to \code{'RAS'}
(right-anterior-superior) transform: the volume index starts from 0 (C-style)
instead of 1 (R-style) to comply with \code{'NIfTI'} transform.}

\item{vox2ras_new}{the targeting transform from volume index to \code{'RAS'}}

\item{na_fill}{default numbers to fill if a pixel is out of bound; default is
\code{NA} or \code{as.raw(0)} if input \code{x} is raw type}

\item{interpolation}{interpolation method: \code{'nearest'} (default),
\code{'trilinear'}, or \code{'bspline'} (cubic \verb{Catmull-Rom}).
\verb{Trilinear} and B-spline are only applied when \code{x} is a
\code{double} volume; other storage types silently fall back to
nearest-neighbor.}
}
\value{
A newly sampled volume that aligns with \code{x} in the anatomical
\code{'RAS'} coordinate system. The underlying storage mode is the same as
\code{x}
}
\description{
Low-level implementation to sample a '3D' volume into a given orientation and
shape using nearest-neighbor, \verb{trilinear}, or cubic B-spline
interpolation.
}
\examples{


# up-sample and rotate image
x <- array(0, c(9, 9, 9))
x[4:6, 4:6, 4:6] <- 1
vox2ras <- matrix(nrow = 4, byrow = TRUE, c(
  0.7071, -0.7071, 0, 0,
  0.7071, 0.7071, 0, -5.5,
  0, 0, 1, -4,
  0, 0, 0, 1
))

new_vox2ras <- matrix(nrow = 4, byrow = TRUE, c(
  0, 0.5, 0, -4,
  0, 0, -0.5, 4,
  0.5, 0, 0, -4,
  0, 0, 0, 1
))

y <- resample_3d_volume(
  x,
  c(17, 17, 17),
  vox2ras_old = vox2ras,
  vox2ras_new = new_vox2ras,
  na_fill = 0
)


image(y[9,,])



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-registration-native.R
\name{apply_transform3d}
\alias{apply_transform3d}
\title{Apply a linear \verb{RAS} transform to resample a 3D volume}
\usage{
apply_transform3d(
  volume,
  vox2ras,
  transform,
  reference_dim = dim(volume),
  reference_vox2ras = vox2ras,
  interpolation = c("trilinear", "nearest", "bspline"),
  na_fill = 0
)
}
\arguments{
\item{volume}{moving 3D array to resample}

\item{vox2ras}{the moving volume's voxel-to-\verb{RAS} 4x4 transform}

\item{transform}{4x4 \verb{RAS}-to-\verb{RAS} transform (fixed to moving)}

\item{reference_dim}{output dimension (the fixed grid); defaults to
\code{dim(volume)}}

\item{reference_vox2ras}{the fixed grid's voxel-to-\verb{RAS} transform;
defaults to \code{vox2ras}}

\item{interpolation}{\code{'trilinear'} (default), \code{'nearest'}, or
\code{'bspline'} (cubic \verb{Catmull-Rom})}

\item{na_fill}{value for out-of-bounds voxels; default \code{0}}
}
\value{
The resampled volume on the reference grid, with a \code{'vox2ras'}
attribute equal to \code{reference_vox2ras}.
}
\description{
Warps a moving volume onto a reference grid given a 4x4
\verb{RAS}-to-\verb{RAS} transform (such as the \code{transform} returned
by \code{\link{register_volume3d}}). The transform maps reference (fixed)
\verb{RAS} coordinates to moving \verb{RAS} coordinates.
}
\seealso{
\code{\link{register_volume3d}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-fill-surface.R
\name{ensure_mesh3d}
\alias{ensure_mesh3d}
\title{Coerce a surface object to a \code{'mesh3d'} mesh}
\usage{
ensure_mesh3d(surface)
}
\arguments{
\item{surface}{a surface object. One of the following:
\describe{
  \item{\code{'mesh3d'}}{returned unchanged.}
  \item{\code{'fs.surface'}}{a surface in the format produced by
    \code{freesurferformats::read.fs.surface}. Vertices and faces are
    copied into a \code{'mesh3d'} list; zero-indexed faces are bumped
    by 1.}
  \item{\code{'ieegio_surface'}}{a surface object produced by
    \pkg{ieegio} (e.g. from \code{read_surface} or
    \code{volume_to_surface}). See the section below for important
    behavior changes.}
  \item{a bare \code{list}}{must contain a numeric \code{vb} matrix with
    3 or 4 rows; the list is reclassified as \code{'mesh3d'} with no
    other modification.}
}
Any other input triggers an error.}
}
\value{
An object of class \code{'mesh3d'} with at least the \code{vb}
(vertex) component and, when face information is available, an \code{it}
(triangle index) component.
}
\description{
Internal helper used throughout \pkg{ravetools} to accept a variety of
surface representations and return a list of class \code{'mesh3d'} (the
format used by the \code{rgl} package). Most mesh-consuming functions in
this package call \code{ensure_mesh3d} on their surface arguments, so the
coercion rules described here apply to all of them.
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

# mesh3d input is returned unchanged in shape
sphere <- vcg_sphere()
m <- ensure_mesh3d(sphere)
identical(m$vb, sphere$vb)

# A bare list with a `vb` slot is reclassified to mesh3d
bare <- list(vb = rbind(matrix(rnorm(30), nrow = 3), 1))
m <- ensure_mesh3d(bare)
inherits(m, "mesh3d")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesh-patch.R
\name{vcg_mesh_patch}
\alias{vcg_mesh_patch}
\title{Split a mesh into two patches along a geodesic boundary}
\usage{
vcg_mesh_patch(mesh, waypoints, seed_vertex = NULL, max_edge_length = NA)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.}

\item{waypoints}{numeric matrix with exactly 3 columns (\code{x}, \code{y},
\code{z}) and at least 3 rows; each row is a 3-D coordinate defining a
boundary corner. After any mesh subdivision, each coordinate is snapped to
the nearest vertex in the (possibly refined) mesh. Consecutive snapped
vertices are joined by geodesic paths to form the closed boundary loop.}

\item{seed_vertex}{integer (optional, 1-based). A vertex known to be inside
the desired first patch. When \code{NULL} (default) the smaller of the
two regions is returned first.}

\item{max_edge_length}{numeric (optional). When positive and finite, the mesh
is refined before patching so that no edge exceeds this length. Global
edge subdivision (\code{\link{vcg_subdivision}}) is applied repeatedly
until the average edge length falls below the threshold (fast), then
\code{\link{vcg_subdivide_max_edge_length}} handles any remaining outlier
edges. \verb{Waypoint} coordinates are snapped to vertices after
refinement, so finer vertices improve boundary accuracy.
Default \code{NA} (no refinement).}
}
\value{
A length-2 list of \code{mesh3d} objects. Each contains:
\describe{
  \item{\code{$orig_vertex}}{1-based integer vector: new vertex index \code{i}
    corresponds to column \code{orig_vertex[i]} of the original \code{mesh$vb}.}
}
The first element is the patch whose centroid is closest to the mean
\verb{waypoint} position; the second is the complementary connected patch.
On a multi-manifold mesh, disconnected components not adjacent to the
boundary loop appear in neither patch. When the boundary loop does not
divide the mesh (degenerate \verb{waypoints}), the second element is
\code{NULL}.
}
\description{
Connects a set of surface \verb{waypoints} with geodesic paths to form a
closed boundary loop, then splits the mesh into the two regions created by
that loop.

Consecutive \verb{waypoints} are joined by the geodesic (Dijkstra) shortest
path along the mesh surface; the last \verb{waypoint} connects back to the
first.
}
\note{
All \verb{waypoints} must lie on the same connected component of the mesh.
  The mesh should be manifold; run \code{\link{vcg_fix_defects}} first
  if needed.
}
\examples{

mesh <- vcg_sphere()
mesh <- vcg_uniform_remesh(mesh)

waypoints <- diag(1, 3)
patches <- vcg_mesh_patch(mesh, waypoints)

plot_mesh_polygon(
  patches,
  col = list("red", 'gray'),
  alpha = list(1, 0.5),
  eye = c(10, 10, 10)
)


}
\seealso{
\code{\link{dijkstras_surface_distance}}, \code{\link{surface_path}},
  \code{\link{vcg_fix_defects}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-fir-design.R
\name{design_filter_fir}
\alias{design_filter_fir}
\title{Design \code{'FIR'} filter using \code{\link{firls}}}
\usage{
design_filter_fir(
  sample_rate,
  filter_order = NA,
  data_size = NA,
  high_pass_freq = NA,
  high_pass_trans_freq = NA,
  low_pass_freq = NA,
  low_pass_trans_freq = NA,
  stopband_attenuation = 40,
  scale = TRUE,
  method = c("kaiser", "firls", "remez")
)
}
\arguments{
\item{sample_rate}{sampling frequency}

\item{filter_order}{filter order, leave \code{NA} (default) if undecided}

\item{data_size}{minimum length of data to apply the filter, used to
decide the maximum filter order. For 'FIR' filter, data length must be
greater than \code{3xfilter_order}}

\item{high_pass_freq}{high-pass frequency; default is \code{NA} (no
high-pass filter will be applied)}

\item{high_pass_trans_freq}{high-pass frequency band-width; default
is automatically inferred from data size.
Frequency \code{high_pass_freq - high_pass_trans_freq} is the corner
of the stop-band}

\item{low_pass_freq}{low-pass frequency; default is \code{NA} (no
low-pass filter will be applied)}

\item{low_pass_trans_freq}{low-pass frequency band-width; default
is automatically inferred from data size.
Frequency \code{low_pass_freq + low_pass_trans_freq} is the corner
of the stop-band}

\item{stopband_attenuation}{allowable power attenuation (in decibel) at
transition frequency; default is \code{40} dB.}

\item{scale}{whether to scale the filter for unity gain}

\item{method}{method to generate 'FIR' filter, default is using
\code{\link[gsignal]{kaiser}} estimate, other choices are
\code{\link{firls}} (with \code{hamming} window) and
\code{\link[gsignal]{remez}} design.}
}
\value{
'FIR' filter in 'Arma' form.
}
\description{
Design \code{'FIR'} filter using \code{\link{firls}}
}
\details{
Filter type is determined from \code{high_pass_freq} and
\code{low_pass_freq}. High-pass frequency is ignored if \code{high_pass_freq}
is \code{NA}, hence the filter is low-pass filter. When
\code{low_pass_freq} is \code{NA}, then
the filter is high-pass filter. When both \code{high_pass_freq} and
\code{low_pass_freq} are valid (positive, less than 'Nyquist'), then
the filter is a band-pass filter if band-pass is less than low-pass
frequency, otherwise the filter is band-stop.

Although the peak amplitudes are set at 1 by \code{low_pass_freq} and
\code{high_pass_freq}, the transition from peak amplitude to zero require
a transition, which is tricky but also important to set.
When 'FIR' filters have too steep transition boundaries, the filter tends to
have ripples in peak amplitude, introducing artifacts to the final signals.
When the filter is too flat, components from unwanted frequencies may also
get aliased into the filtered signals. Ideally, the transition bandwidth
cannot be too steep nor too flat. In this function, users may control
the transition frequency bandwidths via \code{low_pass_trans_freq} and
\code{high_pass_trans_freq}. The power at the end of transition is defined
by \code{stopband_attenuation}, with default value of \code{40} (i.e.
-40 dB, this number is automatically negated during the calculation).
By design, a low-pass 5 Hz filter with 1 Hz transition bandwidth results in
around -40 dB power at 6 Hz.
}
\examples{

# ---- Basic -----------------------------

sample_rate <- 500
data_size <- 1000

# low-pass at 5 Hz, with auto transition bandwidth
# from kaiser's method, with default stopband attenuation = 40 dB
filter <- design_filter_fir(
  low_pass_freq = 5,
  sample_rate = sample_rate,
  data_size = data_size
)

# Passband ripple is around 0.08 dB
# stopband attenuation is around 40 dB
print(filter)

diagnose_filter(
  filter$b, filter$a,
  fs = sample_rate,
  n = data_size,
  cutoffs = c(-3, -6, -40),
  vlines = 5
)

# ---- Advanced ---------------------------------------------

sample_rate <- 500
data_size <- 1000

# Rejecting 3-8 Hz, with transition bandwidth 0.5 Hz at both ends
# Using least-square (firls) to generate FIR filter
# Suggesting the filter order n=160
filter <- design_filter_fir(
  low_pass_freq = 3, low_pass_trans_freq = 0.5,
  high_pass_freq = 8, high_pass_trans_freq = 0.5,
  filter_order = 160,
  sample_rate = sample_rate,
  data_size = data_size,
  method = "firls"
)

#
print(filter)

diagnose_filter(
  filter$b, filter$a,
  fs = sample_rate,
  n = data_size,
  cutoffs = c(-1, -40),
  vlines = c(3, 8)
)



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gammatone.R
\name{gammatone_fast}
\alias{gammatone_fast}
\title{Apply gamma-tone filters to obtain auditory envelopes}
\usage{
gammatone_fast(
  x,
  sample_rate,
  center_frequencies,
  n_bands,
  use_hilbert = TRUE,
  downsample = NA,
  downsample_before_hilbert = FALSE
)
}
\arguments{
\item{x}{a numeric vector or matrix; if \code{x} is a matrix, it should
be column-major (each column is a sound track)}

\item{sample_rate}{sampling frequency}

\item{center_frequencies}{center frequencies at which the envelopes will
be derived; can be either a length of two defining the lower and
upper bound, and using \code{n_bands} to interpolate automatically, or
a length of multiple, with the frequencies specified explicitly}

\item{n_bands}{number of the center frequencies, can be missing if
\code{center_frequencies} is explicit and no interpolation is needed;
if specified, then the frequencies will be interpolated using
equivalent rectangular bandwidth rate (\code{'ERB'})}

\item{use_hilbert}{whether to apply 'Hilbert' transform; default is true,
which calculates the magnitude; set to false when only the filter is needed}

\item{downsample}{whether to down-sample the envelopes after the filters;
default is \code{NA} (no down-sample).}

\item{downsample_before_hilbert}{whether the down-sample happens before
or after the 'Hilbert' transform so speed up the computation if the signal
is too long; only used when \code{downsample} is greater than 1; default
is \code{FALSE}. Use with caution, especially when the voice center
frequency is close to the 'Nyquist' frequency. However, if used properly,
there will be significant performance boost on large signals with
high sampling rates}
}
\value{
A file-array object of filtered and potentially down-sampled
data; see 'Examples' on how to use this function.
}
\description{
Apply gamma-tone filters to obtain auditory envelopes
}
\examples{


fs <- 4000
time <- seq_len(8000) / fs
x <- sin(160 * pi * time) +
  sin(1000 * pi * time) * dnorm(time, mean = 1, sd = 0.1) +
  0.5 * rnorm(length(time))

# envelope
result <- gammatone_fast(
  x,
  sample_rate = fs,
  center_frequencies = c(20, 1000),
  n_bands = 128,

  # default downsample happens after hilbert
  downsample = 40
)


oldpar <- par(mfrow = c(2, 1))

plot(
  time,
  x,
  type = "l",
  xlab = "Time",
  ylab = "",
  main = "Original mixed 80Hz and 500Hz"
)

# only one channel
envelope <- subset(result, Channel ~ Channel == 1, drop = TRUE)
dnames <-  dimnames(envelope)
image(
  x = as.numeric(dnames$Time),
  y = as.numeric(dnames$Frequency),
  z = envelope,
  xlab = "Time",
  ylab = "Frequency",
  main = "Envelope from 20Hz to 1000Hz"
)



par(oldpar) # reset graphics state


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geometry-plane.R
\name{plane_geometry}
\alias{plane_geometry}
\title{Create a two-dimensional plane in three dimensional space}
\usage{
plane_geometry(width = 1, height = 1, shape = c(2, 2))
}
\arguments{
\item{width, height}{width and height of the plane, must not be \code{NA}}

\item{shape}{length of two to indicate the number of vertices along width
and height, default is only \code{c(2, 2)} (2 vertices each side, hence
one grid)}
}
\value{
A triangular mesh of class \code{'mesh3d'}
}
\description{
Create a two-dimensional plane in three dimensional space
}
\examples{

plane <- plane_geometry(5, 10, c(12, 22))

if(FALSE) {

  rgl_view({

    rgl_call("shade3d", plane, col = 3)
    rgl_call("wire3d", plane, col = 1)

  })

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-diagnosis.R
\name{diagnose_filter}
\alias{diagnose_filter}
\title{Diagnose digital filter}
\usage{
diagnose_filter(
  b,
  a,
  fs,
  n = 512,
  whole = FALSE,
  sample = stats::rnorm(n, mean = sample_signal(n), sd = 0.2),
  vlines = NULL,
  xlim = "auto",
  cutoffs = c(-3, -6, -12)
)
}
\arguments{
\item{b}{the moving-average coefficients of an \code{ARMA} model}

\item{a}{the auto-regressive coefficients of an \code{ARMA} filter; default
is \code{1}}

\item{fs}{sampling frequency in \code{Hz}}

\item{n}{number of points at which to evaluate the frequency response;
default is \code{512}}

\item{whole}{whether to evaluate beyond \code{Nyquist} frequency; default
is false}

\item{sample}{sample signal of length \code{n} for simulation}

\item{vlines}{additional vertical lines (frequencies) to plot}

\item{xlim}{frequency limit of frequency response plot; default is
\code{"auto"}, can be \code{"full"} or a numeric of length 2}

\item{cutoffs}{cutoff decibel powers to draw on the frequency plot, also used
to calculate the frequency limit when \code{xlim} is \code{"auto"}}
}
\value{
Nothing
}
\description{
Generate frequency response plot with sample-data simulation
}
\examples{



library(ravetools)

# sample rate
srate <- 500

# signal length
npts <- 1000

# band-pass
bpass <- c(1, 50)

# Nyquist
fn <- srate / 2
w <- bpass / fn

# ---- FIR filter ------------------------------------------------
order <- 160

# FIR1 is MA filter, a = 1
filter <- fir1(order, w, "pass")

diagnose_filter(
  b = filter$b, a = filter$a, n = npts,
  fs = srate, vlines = bpass
)

# ---- Butter filter --------------------------------------------
filter <- butter(3, w, "pass")

diagnose_filter(
  b = filter$b, a = filter$a, n = npts,
  fs = srate, vlines = bpass
)



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fftw.R
\name{fftw-internal}
\alias{fftw-internal}
\alias{fftw_r2c}
\alias{fftw_c2r}
\alias{fftw_c2c}
\alias{mvfftw_r2c}
\alias{mvfftw_c2c}
\alias{mvfftw_c2r}
\alias{fftw_r2c_2d}
\alias{fftw_c2c_2d}
\alias{fftw_r2c_3d}
\alias{fftw_c2c_3d}
\title{Low-level \pkg{FFTW3} wrappers}
\arguments{
\item{data}{Numeric (real) or complex input. For 2D/3D variants, a matrix
or 3-dimensional array. For \code{mvfftw_*} functions, a matrix whose
columns are transformed independently.}

\item{HermConj}{Integer \code{0} or \code{1}. When \code{1}, return the
full Hermitian-symmetric spectrum of length \code{N} (matching
\code{stats::fft}); when \code{0}, return only the non-redundant
one-sided half of length \code{floor(N/2) + 1}.}

\item{inverse}{Integer \code{0} (forward) or \code{1} (inverse). The
inverse transform is \strong{unnormalized}; divide by the number of
elements to invert a forward transform, mirroring
\code{stats::fft(., inverse = TRUE)}.}

\item{fftwplanopt}{Integer planner effort: \code{0} = \code{FFTW_ESTIMATE}
(default, fast planning), \code{1} = \code{FFTW_MEASURE},
\code{2} = \code{FFTW_PATIENT}, \code{3} = \code{FFTW_EXHAUSTIVE}.}

\item{retrows}{Integer; expected number of rows of the time-domain signal
for \code{mvfftw_c2r} when reconstructing from a one-sided spectrum.}

\item{ret}{Optional reusable output buffer of the correct type and
length; pass \code{NULL} (default) to let the function allocate one.
This input is for advanced users; leave \code{NULL} if you do not know
what you are d}
}
\value{
A complex (or real, for \code{*_c2r}) vector / matrix / array
  matching the corresponding base R transform up to floating-point error.
}
\description{
Thin R bindings around the \pkg{FFTW3} library. These are
\strong{low-level} routines exposed primarily for advanced users and other
packages that need maximum throughput. They perform minimal input checking
and follow \pkg{FFTW} conventions (e.g. unnormalized inverse transforms,
one-sided real-to-complex spectra). For most user code prefer
\code{\link[stats]{fft}}, \code{\link[stats]{mvfft}}, or higher-level
helpers in this package such as \code{\link{convolve}}, \code{\link{pwelch}},
\code{\link{multitaper}}, and the filtering utilities.

\strong{Warning:} the API is intentionally close to \pkg{FFTW}'s C interface
and may change between releases. Outputs match the corresponding base R
transforms up to floating-point round-off.
}
\details{
All functions preserve their \code{data} argument: the input buffer is
copied internally before planning when needed, so callers may safely reuse
\code{data} after the call. For multi-dimensional variants, axis ordering
follows R (column-major) conventions.
}
\examples{
set.seed(1)

## --- 1D real-to-complex --------------------------------------------------
x <- rnorm(16)
a <- ravetools::fftw_r2c(x, HermConj = 1)
b <- stats::fft(x)
all.equal(a, b)                                  # TRUE (within tol)

# one-sided spectrum (length floor(N/2)+1)
a_half <- ravetools::fftw_r2c(x, HermConj = 0)
all.equal(a_half, b[seq_len(length(x) \%/\% 2 + 1)])

## --- 1D complex-to-complex ----------------------------------------------
z <- complex(real = rnorm(16), imaginary = rnorm(16))
all.equal(ravetools::fftw_c2c(z, inverse = 0), stats::fft(z))
all.equal(ravetools::fftw_c2c(z, inverse = 1), stats::fft(z, inverse = TRUE))

## --- 1D complex-to-real (inverse of fftw_r2c) ---------------------------
# Using the full Hermitian spectrum:
xr <- ravetools::fftw_c2r(a, HermConj = 1) / length(x)
all.equal(xr, x)

## --- Multivariate (column-wise) ----------------------------------------
M <- matrix(rnorm(32), nrow = 8, ncol = 4)
all.equal(ravetools::mvfftw_r2c(M, HermConj = 1), stats::mvfft(M + 0i))

Mz <- matrix(complex(real = rnorm(32), imaginary = rnorm(32)),
             nrow = 8, ncol = 4)
all.equal(ravetools::mvfftw_c2c(Mz, inverse = 0), stats::mvfft(Mz))
all.equal(ravetools::mvfftw_c2c(Mz, inverse = 1),
          stats::mvfft(Mz, inverse = TRUE))

# one-sided -> back to real signal
Mh <- ravetools::mvfftw_r2c(M, HermConj = 0)
Mr <- ravetools::mvfftw_c2r(Mh, retrows = nrow(M)) / nrow(M)
all.equal(Mr, M)

## --- 2D ----------------------------------------------------------------
X2 <- matrix(rnorm(20), nrow = 5, ncol = 4)
all.equal(ravetools::fftw_r2c_2d(X2, HermConj = 1), stats::fft(X2 + 0i))

Z2 <- matrix(complex(real = rnorm(20), imaginary = rnorm(20)),
             nrow = 5, ncol = 4)
all.equal(ravetools::fftw_c2c_2d(Z2, inverse = 0), stats::fft(Z2))
all.equal(ravetools::fftw_c2c_2d(Z2, inverse = 1),
          stats::fft(Z2, inverse = TRUE))

## --- 3D ----------------------------------------------------------------
X3 <- array(rnorm(60), dim = c(5, 4, 3))
all.equal(ravetools::fftw_r2c_3d(X3, HermConj = 1), stats::fft(X3 + 0i))

Z3 <- array(complex(real = rnorm(60), imaginary = rnorm(60)),
            dim = c(5, 4, 3))
all.equal(ravetools::fftw_c2c_3d(Z3, inverse = 0), stats::fft(Z3))
all.equal(ravetools::fftw_c2c_3d(Z3, inverse = 1),
          stats::fft(Z3, inverse = TRUE))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/curve-catmull.R
\name{plot.ravetools_curve}
\alias{plot.ravetools_curve}
\title{Plot method for \code{ravetools_curve}}
\usage{
\method{plot}{ravetools_curve}(x, n = 200L, col = "steelblue", pch = 19L, cex = 1, use_rgl = TRUE, ...)
}
\arguments{
\item{x}{an object of class \code{ravetools_curve}.}

\item{n}{integer; number of sample points used to draw the smooth curve.
Default is \code{200L}.}

\item{col}{color for the spline curve line. Default \code{"steelblue"}.}

\item{pch, cex}{plotting character and scaling for the key control points
(base-R fallback only). Default \code{pch = 19}, \code{cex = 1}.}

\item{use_rgl}{logical; if \code{TRUE} (default) and the \code{rgl}
package is installed, an interactive 3D window is used. Set to
\code{FALSE} to force the base-R 2D projection panels.}

\item{...}{additional graphical parameters forwarded to the underlying
plot calls.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Plots a \code{ravetools_curve} object created by
\code{\link{catmull_rom_3d}}.  When the \code{rgl} package is available
and \code{use_rgl = TRUE} (default), an interactive 3D scene is opened.
Otherwise three 2-D projection panels (\emph{x-y}, \emph{x-z},
\emph{y-z}) are drawn using base \R{} graphics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-utils.R
\name{rcond_filter_ar}
\alias{rcond_filter_ar}
\title{Computer reciprocal condition number of an 'Arma' filter}
\usage{
rcond_filter_ar(a)
}
\arguments{
\item{a}{auto-regression coefficient, numerical vector; the first element
must not be zero}
}
\value{
Reciprocal condition number of matrix \code{z1}, used in
\code{\link{filtfilt}}. If the number is less than
\code{.Machine$double.eps}, then \code{\link{filtfilt}} will fail.
}
\description{
Test whether the filter is numerically stable for \code{\link{filtfilt}}.
}
\examples{


# Butterworth filter with low-pass at 0.1 Hz (order = 4)
filter <- butter(4, 0.1, "low")

# TRUE
rcond_filter_ar(filter$a) > .Machine$double.eps

diagnose_filter(filter$b, filter$a, 500)

# Bad filter (order is too high)
filter <- butter(50, 0.1, "low")

rcond_filter_ar(filter$a) > .Machine$double.eps

# filtfilt needs to inverse a singular matrix
diagnose_filter(filter$b, filter$a, 500)

}
\seealso{
\code{\link{check_filter}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detrend.R
\name{detrend}
\alias{detrend}
\title{Remove the trend for one or more signals}
\usage{
detrend(x, trend = c("constant", "linear"), break_points = NULL)
}
\arguments{
\item{x}{numerical or complex, a vector or a matrix}

\item{trend}{the trend of the signal; choices are \code{'constant'} and
\code{'linear'}}

\item{break_points}{integer vector, or \code{NULL}; only used when
\code{trend} is \code{'linear'} to remove piecewise linear trend; will
throw warnings if \code{trend} is \code{'constant'}}
}
\value{
The signals with trend removed in matrix form; the number of columns
is the number of signals, and number of rows is length of the signals
}
\description{
'Detrending' is often used before the signal power calculation.
}
\examples{

x <- rnorm(100, mean = 1) + c(
  seq(0, 5, length.out = 50),
  seq(5, 3, length.out = 50))
plot(x)

plot(detrend(x, 'constant'))
plot(detrend(x, 'linear'))
plot(detrend(x, 'linear', 50))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{matlab_palette}
\alias{matlab_palette}
\title{'Matlab' heat-map plot palette}
\usage{
matlab_palette()
}
\value{
vector of 64 colors
}
\description{
'Matlab' heat-map plot palette
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rgl.R
\name{rgl-call}
\alias{rgl-call}
\alias{rgl_call}
\alias{rgl_view}
\alias{rgl_plot_normals}
\title{Safe ways to call package \code{'rgl'} without requiring \code{'x11'}}
\usage{
rgl_call(FUN, ...)

rgl_view(expr, quoted = FALSE, env = parent.frame())

rgl_plot_normals(x, length = 1, lwd = 1, col = 1, ...)
}
\arguments{
\item{FUN}{\code{'rgl'} function name}

\item{...}{passed to \code{'rgl'} function}

\item{expr}{expression within which \code{'rgl'} functions are called}

\item{quoted}{whether \code{expr} is quoted}

\item{env}{environment in which \code{expr} is evaluated}

\item{x}{triangular \code{'mesh3d'} object}

\item{length, lwd, col}{normal vector length, size, and color}
}
\description{
Internally used for example show-cases. Please install package \code{'rgl'}
manually to use these functions.
}
\examples{


# Make sure the example does not run when compiling
# or check the package
if(FALSE) {

  volume <- array(0, dim = c(8,8,8))
  volume[4:5, 4:5, 4:5] <- 1
  mesh <- mesh_from_volume(volume, verbose = FALSE)

  rgl_view({

    rgl_call("shade3d", mesh, col = 3)
    rgl_plot_normals(mesh)

  })

}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-Vector3.R
\name{new_vector3}
\alias{new_vector3}
\alias{as_vector3}
\title{Create a \code{Vector3} instance to store '3D' points}
\usage{
new_vector3(x = 0, y = 0, z = 0)

as_vector3(v)
}
\arguments{
\item{x, y, z}{numeric, must have the same length, \code{'xyz'} positions}

\item{v}{R object to be converted to \code{Vector3} instance}
}
\value{
A \code{Vector3} instance
}
\description{
Create instances that mimic the \code{'three.js'} syntax.
}
\examples{

vec3 <- new_vector3(
  x = 1:9,
  y = 9:1,
  z = rep(c(1,2,3), 3)
)

vec3[]

# transform
m <- new_matrix4()

# rotation xy plane by 30 degrees
m$make_rotation_z(pi / 6)

vec3$apply_matrix4(m)

vec3[]

as_vector3(c(1,2,3))

}
\seealso{
\code{\link{new_matrix4}}, \code{\link{new_quaternion}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mris-make-surfaces.R
\name{mris_make_surfaces}
\alias{mris_make_surfaces}
\title{Localize white-matter and \verb{pial} surfaces from an intensity volume}
\usage{
mris_make_surfaces(
  mesh,
  volume,
  white_intensity,
  pial_intensity,
  IJK2RAS = NULL,
  max_thickness = 5,
  step_size = 0.4,
  n_averages = 4L,
  niterations = 10L,
  l_intensity = 1,
  l_spring = 0.5,
  momentum = 0.9,
  dt = 0.5,
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'} (or coercible via
\code{\link{ensure_mesh3d}}), in the same physical (\verb{RAS}) coordinate
space as \code{volume}. As with \code{\link{mris_inflate}}, the mesh
\strong{must} be closed, manifold, and genus-0 (watertight, no boundary or
non-manifold edges, single connected component); \code{mris_make_surfaces}
raises an error if such defects are detected}

\item{volume}{a 3-dimensional numeric array of image intensities (such as
a normalized \verb{T1} volume), co-registered with \code{mesh}}

\item{white_intensity}{target intensity for the white/gray-matter boundary;
the value the white-surface pass searches for along each vertex normal. For
a normalized \verb{T1} volume this is typically close to the midpoint
between the white-matter and gray-matter intensities}

\item{pial_intensity}{target intensity for the gray-matter/\verb{CSF}
boundary, analogous to \code{white_intensity} for the \verb{pial}-surface pass}

\item{IJK2RAS}{volume \verb{IJK} (zero-indexed voxel index) to surface
\verb{tkrRAS} transform (a \code{4x4} matrix); default is \code{NULL},
which assumes \code{volume} is a conformed volume (\verb{LIA} orientation,
\verb{1mm} isotropic 'voxels', centered at the volume midpoint, the same
convention \code{\link{fill_surface}} defaults to) and derives the
transform from \code{dim(volume)}; set this explicitly when \code{volume}
uses a different orientation or voxel size}

\item{max_thickness}{half-width, in \verb{mm}, of the per-vertex normal
search window the intensity-target term samples. Default \code{5}}

\item{step_size}{sampling step, in \verb{mm}, along the search window.
Default \code{0.4}}

\item{n_averages}{number of 1-ring gradient-averaging passes applied to the
intensity-target gradient before the smoothness term is added (mirrors
\code{\link{mris_inflate}}'s gradient-averaging step). Default \code{4}}

\item{niterations}{number of deformation iterations \emph{per pass}
(white, then \verb{pial}). Default \code{10}}

\item{l_intensity}{intensity-target term coefficient. Default \code{1.0}}

\item{l_spring}{smoothness (1-ring Laplacian spring) term coefficient.
Default \code{0.5}}

\item{momentum}{momentum coefficient. Default \code{0.9}}

\item{dt}{time step. Default \code{0.5}}

\item{verbose}{logical; print per-pass progress. Default \code{FALSE}}
}
\value{
A named list of two \code{'mesh3d'} surfaces (each with
\code{vb}, \code{it}, and \code{normals}):
\describe{
  \item{\code{white}}{Surface localized to \code{white_intensity}.}
  \item{\code{pial}}{Surface localized to \code{pial_intensity}, continuing
    from \code{white}.}
}
}
\description{
Starting from a single closed surface mesh (typically a smoothed estimate
of the white-matter boundary, in the same physical coordinate space as
\code{volume}) and a co-registered intensity volume (such as a normalized
\verb{T1} scan), iteratively deforms the surface in two passes to localize
the white/gray-matter and gray-matter/\verb{CSF} tissue-intensity
boundaries, producing a \code{white} and a \code{pial} surface.
}
\details{
The implementation keeps the two dominant ideas of the surface-placement
procedure described in the literature (see \strong{References}):
\enumerate{
  \item \strong{Intensity-target localization}: for each vertex, sample
    \code{volume} along the vertex's current normal at offsets spanning
    \eqn{\pm}\code{max_thickness} in steps of \code{step_size}, and pull
    the vertex toward the offset whose sampled intensity is closest to a
    single target value (\code{white_intensity} for the white-surface pass,
    \code{pial_intensity} for the \verb{pial}-surface pass).
  \item \strong{Smoothness}: a 1-ring Laplacian spring keeps the mesh
    regular while the per-vertex intensity term, which reacts independently
    to noisy image data, pulls vertices toward the tissue boundary.
}
Both terms are integrated with the same gradient-averaging and
momentum-integration machinery \code{\link{mris_inflate}} and
\code{\link{mris_sphere}} use: each inner iteration clears the gradient,
adds the intensity-target term, smooths it over \code{n_averages} passes of
1-ring averaging, adds the locally-acting smoothness term (which is not
itself smoothed),
then takes a momentum-integration step with a 1 \verb{mm} per-step
displacement cap, and refreshes the vertex normals.

The white-surface pass runs first, for \code{niterations} iterations; the
\verb{pial}-surface pass then continues from its result, with the momentum
velocity reset to rest, toward \code{pial_intensity} for another
\code{niterations} iterations.

This is a \emph{reduced} port: the literature's procedure is a
multi-resolution optimization over roughly seven weighted energy terms
(intensity, intensity gradient, smoothness, self-intersection repulsion,
curvature, and more), using per-vertex gray/white/\verb{CSF} intensity
statistics derived from a prior segmentation. Reproducing that faithfully
is out of scope for this package; \code{white_intensity} and
\code{pial_intensity} are supplied directly here instead, for example the
midpoints between the typical white-matter/gray-matter and
gray-matter/\verb{CSF} intensities of \code{volume}.
}
\examples{

if (is_not_cran()) {


data("left_hippocampus_mask")
n_vox <- length(left_hippocampus_mask)

volume <- left_hippocampus_mask + runif(n = n_vox, 0, 1)
vox2ras <- diag(1, 4)
mesh <- vcg_isosurface(volume, threshold_lb = 0.99)

plot(mesh)

# Fix defects
mesh <- vcg_fix_defects(mesh, verbose = TRUE, merge_tolerance = 1.75)


res <- mris_make_surfaces(
  mesh,
  volume,
  pial_intensity = 1.1,
  white_intensity = 1,
  IJK2RAS = vox2ras
)

plot(res$pial)


}


}
\references{
Cortical surface-based analysis I: Segmentation and surface
reconstruction. \emph{NeuroImage}, 9(2), 179-194 (1999).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dijkstras-path.R
\name{dijkstras-path}
\alias{dijkstras-path}
\alias{dijkstras_surface_distance}
\alias{surface_path}
\title{Calculate distances along a surface}
\usage{
dijkstras_surface_distance(
  positions,
  faces,
  start_node,
  face_index_start = NA,
  max_search_distance = NA,
  ...
)

surface_path(x, target_node)
}
\arguments{
\item{positions}{numeric matrix with no \code{NA} values. The number of row
is the total count of nodes (vertices), and the number of columns represent
the node dimension. Each row represents a node.}

\item{faces}{integer matrix with each row containing indices of nodes. For
graphs, \code{faces} is a matrix with two columns defining the connecting
edges; for '3D' mesh, \code{faces} is a three-column matrix defining the
face index of mesh triangles.}

\item{start_node}{integer, row index of \code{positions} on where to start
calculating the distances. This integer must be 1-indexed and cannot exceed
the total number of \code{positions} rows}

\item{face_index_start}{integer, the start of the nodes in \code{faces};
please specify this input explicitly if the first node is not contained
in \code{faces}.
Default is \code{NA} (determined by the minimal number in \code{faces}).
The reason to set this input is because some programs use \code{1} to
represent the first node, some start from \code{0}.}

\item{max_search_distance}{numeric, maximum distance to iterate;
default is \code{NA},
that is to iterate and search the whole mesh}

\item{...}{reserved for backward compatibility}

\item{x}{distance calculation results returned by
\code{dijkstras_surface_distance} function}

\item{target_node}{the target node number to reach (from the starting node);
\code{target_node} is always 1-indexed.}
}
\value{
\code{dijkstras_surface_distance} returns a list distance
table with the meta configurations. \code{surface_path} returns a data frame
of the node ID (from \code{start_node} to \code{target_node}) and cumulative
distance along the shortest path.
}
\description{
Calculate surface distances of graph or mesh using 'Dijkstra' method.
}
\examples{

# ---- Toy example --------------------

# Position is 2D, total 6 points
positions <- matrix(runif(6 * 2), ncol = 2)

# edges defines connected nodes
edges <- matrix(ncol = 2, byrow = TRUE, data = c(
  1,2,
  2,3,
  1,3,
  2,4,
  3,4,
  2,5,
  4,5,
  2,5,
  4,6,
  5,6
))

# calculate distances
ret <- dijkstras_surface_distance(
  start_node = 1,
  positions = positions,
  faces = edges,
  face_index_start = 1
)

# get shortest path from the first node to the last
path <- surface_path(ret, target_node = 6)

# plot the results
from_node <- path$path[-nrow(path)]
to_node <- path$path[-1]
plot(positions, pch = 16, axes = FALSE,
     xlab = "X", ylab = "Y", main = "Dijkstra's shortest path")
segments(
  x0 = positions[edges[,1],1], y0 = positions[edges[,1],2],
  x1 = positions[edges[,2],1], y1 = positions[edges[,2],2]
)

points(positions[path$path,], col = "steelblue", pch = 16)
arrows(
  x0 = positions[from_node,1], y0 = positions[from_node,2],
  x1 = positions[to_node,1], y1 = positions[to_node,2],
  col = "steelblue", lwd = 2, length = 0.1, lty = 2
)

points(positions[1,,drop=FALSE], pch = 16, col = "orangered")
points(positions[6,,drop=FALSE], pch = 16, col = "purple3")

# ---- Example with mesh ------------------------------------

\dontrun{

  # Please install the down-stream package `threeBrain`
  # and call library(threeBrain)
  # the following code set up the files

  read.fs.surface <- internal_rave_function(
    "read.fs.surface", "threeBrain")
  default_template_directory <- internal_rave_function(
    "default_template_directory", "threeBrain")
  surface_path <- file.path(default_template_directory(),
                            "N27", "surf", "lh.pial")
  if (!file.exists(surface_path)) {
    internal_rave_function(
      "download_N27", "threeBrain")()
  }

  # Example starts from here --->
  # Load the mesh
  mesh <- read.fs.surface(surface_path)

  # Calculate the path with maximum radius 100
  ret <- dijkstras_surface_distance(
    start_node = 1,
    positions = mesh$vertices,
    faces = mesh$faces,
    max_search_distance = 100,
    verbose = TRUE
  )

  # get shortest path from the first node to node 43144
  path <- surface_path(ret, target_node = 43144)

  # plot
  from_nodes <- path$path[-nrow(path)]
  to_nodes <- path$path[-1]
  # calculate colors
  pal <- colorRampPalette(
    colors = c("red", "orange", "orange3", "purple3", "purple4")
  )(1001)
  col <- pal[ceiling(
    path$distance / max(path$distance, na.rm = TRUE) * 1000
  ) + 1]
  oldpar <- par(mfrow = c(2, 2), mar = c(0, 0, 0, 0))
  for(xdim in c(1, 2, 3)) {
    if ( xdim < 3 ) {
      ydim <- xdim + 1
    } else {
      ydim <- 3
      xdim <- 1
    }
    plot(
      mesh$vertices[, xdim], mesh$vertices[, ydim],
      pch = ".", col = "#BEBEBE33", axes = FALSE,
      xlab = "P - A", ylab = "S - I", asp = 1
    )
    segments(
      x0 = mesh$vertices[from_nodes, xdim],
      y0 = mesh$vertices[from_nodes, ydim],
      x1 = mesh$vertices[to_nodes, xdim],
      y1 = mesh$vertices[to_nodes, ydim],
      col = col
    )
  }

  # plot distance map
  distances <- ret$paths$distance
  col <- pal[ceiling(distances / max(distances, na.rm = TRUE) * 1000) + 1]
  selection <- !is.na(distances)

  plot(
    mesh$vertices[, 2], mesh$vertices[, 3],
    pch = ".", col = "#BEBEBE33", axes = FALSE,
    xlab = "P - A", ylab = "S - I", asp = 1
  )
  points(
    mesh$vertices[selection, c(2, 3)],
    col = col[selection],
    pch = "."
  )

  # reset graphic state
  par(oldpar)

}






}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-mesh-dotcloud.R
\name{plot_mesh_dotcloud}
\alias{plot_mesh_dotcloud}
\title{Render one or more meshes as an orthographic dot cloud in base R}
\usage{
plot_mesh_dotcloud(
  mesh,
  eye = c(0, 0, 1000),
  lookat = c(0, 0, 0),
  up = c(0, 1, 0),
  col = c("white", "gray30"),
  pch = 16L,
  cex = 0.1,
  add = FALSE,
  axes = FALSE,
  asp = 1,
  xlim = NULL,
  ylim = NULL,
  zoom = 1,
  xlab = "",
  ylab = "",
  normal_weight = c("auto", "area", "angle"),
  side = c("front", "back", "both"),
  mesh_clipping = 0.7,
  alpha = 1,
  clipping_plane = NULL,
  clipping_plane_enabled = TRUE,
  ...
)
}
\arguments{
\item{mesh}{a \code{'mesh3d'} object, or a list of \code{'mesh3d'} objects.
Meshes without a face matrix (\code{$it}) are treated as point clouds.
If \code{mesh$normals} is absent and the mesh has faces, normals are
recomputed with \code{\link{vcg_update_normals}}.}

\item{eye}{numeric vector of length 3 - camera position in world space.}

\item{lookat}{numeric vector of length 3 - the world-space point the camera
is looking at.}

\item{up}{numeric vector of length 3 - a world-space vector indicating
which direction is "up" for the camera; defaults to \code{c(0, 1, 0)}.}

\item{col}{base color(s) for the dots.  Accepted forms:
\itemize{
  \item A single color string - applied to all meshes.
  \item A character vector of length 2 or more (not matching any vertex
    count) - used as a depth-ordered color gradient applied to all meshes.
  \item A character vector of length equal to the number of vertices in a
    single mesh - per-vertex colors for that mesh.
  \item A list of length equal to the number of meshes - each element is
    one of the above, applied to the corresponding mesh.
}
Default \code{"gray30"}.}

\item{pch}{point character; a scalar or vector/list with one value per mesh,
recycled as necessary.  Default \code{16L}.}

\item{cex}{point expansion factor; a scalar or vector/list with one value
per mesh, recycled as necessary.  Default \code{0.1}.}

\item{add}{logical; if \code{TRUE} the dots are added to an existing plot
instead of opening a new one.  Default \code{FALSE}.}

\item{axes}{logical; whether to draw axes on a new plot.  Ignored when
\code{add = TRUE}.  Default \code{FALSE}.}

\item{asp}{aspect ratio of the new plot; default \code{1} (equal scaling).
Ignored when \code{add = TRUE}.}

\item{xlim, ylim}{axis limits for the new plot; \code{NULL} (default) lets R
choose automatically from all meshes' projected vertices.  Ignored when
\code{add = TRUE}.}

\item{zoom}{positive numeric magnification applied to the auto-computed
axis limits when \code{xlim} or \code{ylim} is \code{NULL}.  Values
greater than \code{1} zoom in, values in \code{(0, 1)} zoom out.
When \code{asp} is set, the zoom is plot-region aware (queried via
\code{par("pin")}): the axis that already fills the device after
\code{asp}-induced expansion is zoomed by exactly \code{zoom}, and
the other axis is zoomed less so the data fills more of the plot
region while preserving the requested \code{asp}.  Ignored for any
axis whose limit was supplied explicitly, and when \code{add = TRUE}.
Default \code{1}.}

\item{xlab, ylab}{axis labels for the new plot.  Ignored when
\code{add = TRUE}.}

\item{normal_weight}{passed to \code{\link{vcg_update_normals}} when normals
must be recomputed; one of \code{"auto"} (area if no normals present,
otherwise skip), \code{"area"}, or \code{"angle"}.  Default \code{"auto"}.}

\item{side}{which side of meshed surfaces to render.  One of \code{"front"}
(default), \code{"back"}, or \code{"both"} (renders all vertices).  Point clouds are always
rendered regardless of this setting.}

\item{mesh_clipping}{numeric in \code{[0, 1]} controlling how much of
the surface is kept relative to the camera-facing direction.
Vertices whose rim-light weight (\eqn{1 - |n \cdot z_{cam}|}) is at
or below \code{1 - mesh_clipping} are dropped, leaving only the
silhouette/grazing band.  \code{mesh_clipping = 1} (no clipping)
keeps all vertices; smaller values peel away more of the
front/back-facing dots.  Surviving vertices are drawn opaque, and
their \code{cex} is multiplied by the rim-light weight so
grazing-edge dots appear larger than near-front-facing dots.
Drawing opaque points with size-modulated weight is dramatically
faster than R's true per-point transparency.  Default \code{0.7}.}

\item{alpha}{numeric in \code{[0, 1]}; one value per mesh (recycled),
sets the transparency of each mesh (\code{1} = fully opaque,
\code{0} = fully transparent).  Vertices belonging to a mesh with
\code{alpha = 0} are dropped entirely.  Default \code{1}.}

\item{clipping_plane}{optional list of world-space clipping planes used to
hide parts of the scene.  Each plane is a numeric vector of length 5:
the first three entries are the plane normal \eqn{(n_x, n_y, n_z)}
(must be non-zero; normalized internally), the fourth is the signed
distance from the world origin to the plane along that normal (so the
plane equation is \eqn{n \cdot x = d}), and the fifth indicates which
half-space is kept: \code{1} keeps the front side (the side the normal
points to), \code{-1} keeps the back side, and \code{0} keeps whichever
side currently faces the camera (auto-flipped per call based on
\code{eye}).  Multiple planes are intersected.  Clipping is applied
per vertex.  A single length-5 numeric vector is also accepted as
shorthand for a one-plane list.  Default \code{NULL} (no clipping).}

\item{clipping_plane_enabled}{logical vector, one entry per mesh
(recycled), controlling whether \code{clipping_plane} is applied to
that mesh.  \code{TRUE} (default) means the mesh participates in
clipping; \code{FALSE} exempts the mesh entirely (all of its vertices
are kept regardless of the clipping planes).  Has no effect when
\code{clipping_plane} is \code{NULL}.}

\item{...}{additional graphical parameters forwarded to
\code{\link[graphics]{plot.default}} (new plot) or
\code{\link[graphics]{points}} (when \code{add = TRUE}).}
}
\value{
Invisibly returns a list with components \code{xlim} and
  \code{ylim} (the plot limits used).
}
\description{
Projects mesh vertices onto a 2D plane using an orthographic camera defined
by an eye position, a look-at point, and an up direction, then draws the
projected dots with base-R \code{\link[graphics]{plot}}.  Each dot is
rendered opaque, but its \code{cex} is modulated by a rim-light weight
\eqn{1 - |n \cdot z_{cam}|}: front- and back-facing vertices shrink toward
zero size, while grazing-edge vertices keep their full size.  This
size-modulated trick gives a rim-light look without paying R's per-point
transparency-blending cost.

Multiple meshes share a single depth space: all vertices are projected
together and sorted globally by depth before a single \code{plot()} call,
so the painter's algorithm works correctly across meshes.  Meshes without
faces (point clouds) are rendered at full size and are not affected
by the \code{side} filter.
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

mesh <- vcg_isosurface(left_hippocampus_mask)

# Side view - rim-light shows the outline of the hippocampus
plot_mesh_dotcloud(
  mesh,
  eye = c(150, 0, 0),
  lookat = c(0, 0, 0),
  up = c(0, 0, 1),
  col = "steelblue",
  cex = 2
)

# Two meshes: surface + electrode point cloud
n_elec <- 20
electrodes <- structure(
  list(vb = matrix(rnorm(3 * n_elec, sd = 5), 3, n_elec) +
         rowMeans(mesh$vb)[1:3]),
  class = "mesh3d"
)
plot_mesh_dotcloud(
  mesh  = list(mesh, electrodes),
  eye    = c(150, 0, 0),
  lookat = c(0, 0, 0),
  up     = c(0, 0, 1),
  col    = list("steelblue", "red"),
  pch    = c(16L, 17L),
  cex    = c(2, 1.2)
)

}
\seealso{
\code{\link{vcg_update_normals}}, \code{\link{vcg_isosurface}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/curve-catmull.R
\name{catmull_rom_3d}
\alias{catmull_rom_3d}
\title{\verb{Catmull-Rom} 3D Spline Curve}
\usage{
catmull_rom_3d(
  points,
  curve_type = c("centripetal", "chordal", "uniform"),
  tension = 0.5,
  closed = FALSE
)
}
\arguments{
\item{points}{numeric matrix with at least 2 rows and exactly 3 columns
(\code{x}, \code{y}, \code{z}), giving the key (control) points through
which the curve passes.}

\item{curve_type}{character; One of
\code{"centripetal"} (default), \code{"chordal"}, or \code{"uniform"}.
\code{"centripetal"} uses \eqn{\alpha = 0.5} (square-root of chord
length), \code{"chordal"} uses \eqn{\alpha = 1} (chord length), and
\code{"uniform"} is the classic formulation controlled by
\code{tension}.}

\item{tension}{numeric scalar in \eqn{[0, 1]}; tangent scaling factor used
only when \code{curve_type = "uniform"}. At \code{0.5} (default) the
curve matches the standard \verb{Catmull-Rom} formulation.}

\item{closed}{logical; if \code{TRUE} the curve closes on itself by
connecting the last point back to the first. Default is \code{FALSE}.}
}
\value{
An object of class \code{"ravetools_curve"} (a list) with the
  following elements:
\describe{
  \item{\code{points}}{The input key-point matrix (\eqn{n \times 3}).}
  \item{\code{curve_type}}{Character, the parameterization type.}
  \item{\code{tension}}{Numeric, the tension value (relevant for
    \code{"uniform"} only).}
  \item{\code{closed}}{Logical, whether the curve is closed.}
  \item{\code{get_point}}{A \code{function(t)} that accepts a scalar
    \code{t} in \eqn{[0, 1]} and returns a named numeric vector
    on the curve.}
  \item{\code{get_points}}{A \code{function(n)} that returns an
    \eqn{n \times 3} matrix of \code{n} evenly spaced points along the
    curve, with column names \code{"x"}, \code{"y"}, \code{"z"}.}
  \item{\code{get_closest_t}}{A \code{function(query, coarse_n = 200L)}
    that, given a 3-element numeric vector \code{query} (\code{x}, \code{y},
    \code{z}), returns a list with elements \code{t} (the parameter value in
    \eqn{[0, 1]} of the nearest point), \code{point} (the closest point on
    the curve as a named numeric vector), and \code{distance} (Euclidean
    distance from \code{query} to the curve).  The search uses
    \code{coarse_n} uniform samples for an initial bracket followed by
    scalar optimization.}
  \item{\code{t_keypoints}}{Numeric vector of length \eqn{n} with the
    \code{t} parameter value where each key point lies on the curve.
    First element is always \code{0}, last is always \code{1}.}
  \item{\code{segment_lengths}}{Numeric vector of length \eqn{n-1} (open
    curve) or \eqn{n} (closed curve) containing the arc length of each
    spline segment, estimated by numerical integration.}
}
}
\description{
Creates a smooth \verb{Catmull-Rom} spline curve through a set of 3D key points.
}
\examples{

pts <- matrix(c(
  -33.0534, -10.6213, -21.8328,
  -34.7526, -25.5089, -14.5390,
  -41.2002, -10.4606, -22.0032,
  -46.4717, -10.3567, -22.1134,
  -51.7431, -10.2528, -22.2237,
  -57.0146, -10.1488, -22.3339,
  -62.2860, -10.0449, -22.4442,
  -67.5575,  -9.9410, -22.5544
), ncol = 3, byrow = TRUE)

curve <- catmull_rom_3d(pts)
print(curve)

# Sample 100 evenly spaced points along the curve
smooth <- curve$get_points(100)
head(smooth)

# Evaluate the curve at t = 0.5 (midpoint)
curve$get_point(0.5)

# get closest point on curve
curve$get_closest_t(c(-49, -10, -22))

plot(curve, use_rgl = FALSE)

}
\seealso{
\code{\link{print.ravetools_curve}},
  \code{\link{plot.ravetools_curve}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-registration-native.R
\name{register_volume3d}
\alias{register_volume3d}
\title{Native 3D volume registration (\code{'rigid'}, \code{'affine'}, or \code{'SyN'})}
\usage{
register_volume3d(
  source,
  target,
  source_vox2ras = NULL,
  target_vox2ras = NULL,
  source_mask = NULL,
  target_mask = NULL,
  source_points = NULL,
  target_points = NULL,
  points_weight = 0.5,
  weights = NULL,
  type = c("rigid", "affine", "syn", "syn_only"),
  metric = "mattes",
  shrink_factors = c(4, 2, 1),
  smoothing_sigmas = c(2, 1, 0),
  iterations = c(1000, 500, 250),
  sampling_rate = 0.2,
  interpolation = "trilinear",
  number_of_bins = 32L,
  seed = 1L,
  init_transform = NULL,
  syn_iterations = c(40, 20, 0),
  syn_sigma = 3,
  verbose = TRUE
)
}
\arguments{
\item{source}{the moving volume to be aligned, a 3D array (for example a
\code{'CT'}); the result transform maps \code{target} into this image's space.
May also be a \code{list} of already co-registered 3D arrays (multiple
modalities such as \code{'T1'}, \code{'T2'}, an atlas constraint); all source
channels must share the same dimensions and \code{vox2ras}}

\item{target}{the fixed/reference volume to align to, a 3D array (for example
a \code{'MRI'}), or a \code{list} of co-registered arrays matching
\code{source} channel-for-channel}

\item{source_vox2ras, target_vox2ras}{4x4 (or 3x4) matrices mapping the
0-indexed voxel coordinate (column-row-slice, \code{'C'}-style starting from
0, complying with \code{'NIfTI'}) to the \verb{RAS} coordinate system; if
\code{NULL}, the function looks for a \code{"vox2ras"} attribute on the array}

\item{source_mask, target_mask}{optional 3D mask arrays restricting where the
metric is evaluated; default \code{NULL} (no mask, evaluate everywhere). A
\code{target_mask} (on the target/fixed grid, same dimensions as
\code{target}) limits which target voxels drive the registration; a
\code{source_mask} (on the source/moving grid) drops samples that map outside
it. Non-zero voxels are included. Besides focusing the alignment on a region
of interest, masks speed things up by skipping background: the linear stage
samples only inside the mask, and the deformable stage skips warping voxels
outside the (dilated) mask. One mask per grid, shared across channels}

\item{source_points, target_points}{optional \code{N x 3} matrices of
corresponding landmark coordinates that add a surface/landmark term to the
\strong{deformable} (\code{"syn"}/\code{"syn_only"}) stage; default
\code{NULL} (no term). Row \code{i} of \code{target_points} (in the
target/fixed \verb{RAS}) and row \code{i} of \code{source_points} (in the
source/moving \verb{RAS}) must be the \emph{same} anatomical location, for
example corresponding cortical-surface vertices from a \code{'FreeSurfer'}
spherical registration. The term pulls the warp of each target point onto its
source correspondent, recovering cortical folding (\verb{gyrification})
that the intensity metric blurs over, while the image metric still drives
deep-brain \verb{subcortical} structures. Must be supplied together with
equal row counts. \strong{Points must be in the same \verb{RAS} frame as
\code{source_vox2ras}/\code{target_vox2ras}} (note \code{'FreeSurfer'}
surfaces use \verb{surface/tkr} \verb{RAS}, which differs from scanner
\verb{RAS} by \code{c_ras})}

\item{points_weight}{relative weight of the landmark term against the image
metric in the deformable stage; default \code{0.5}. Larger values follow the
landmarks more closely. The sparse landmark force is attenuated by
\code{syn_sigma} smoothing, so this typically needs tuning for a given point
count and spacing}

\item{weights}{optional numeric weights, one per source/target pair,
controlling each channel's contribution to the deformable cost; default is
equal weighting. Weights are normalized internally to sum to 1. Only the
deformable (\code{"syn"}/\code{"syn_only"}) stage is multivariate; the linear
stage always uses the first (primary) pair}

\item{type}{type of transform to estimate; one of \code{'rigid'} (6 degrees
of freedom), \code{'affine'} (12), \code{"syn"} (\verb{affine} followed
by a \verb{SDR} - symmetric \verb{diffeomorphic} deformation), or
\code{'syn_only'} (\verb{deformable} stage only — no \verb{affine} is
estimated; \code{init_transform} is used directly as the starting
\verb{affine}, useful when you already have a good linear alignment)}

\item{metric}{similarity metric: \code{"mattes"} (Mattes mutual information,
the default, best for cross-modal such as \code{'CT'}-\code{'MRI'}),
\code{"cc"} (normalized cross-correlation, for same-modality), or
\code{"meansquares"} (mean squared intensity difference). With multiple
channels, supply either a single metric (used for every pair) or a vector
with one metric per source/target pair}

\item{shrink_factors}{integer down-sampling factors, one per resolution level
(coarsest first); default \code{c(4, 2, 1)}}

\item{smoothing_sigmas}{Gaussian smoothing applied at each level, in voxels,
same length as \code{shrink_factors}; default \code{c(2, 1, 0)}}

\item{iterations}{maximum optimizer iterations per level; default
\code{c(1000, 500, 250)}}

\item{sampling_rate}{fraction of fixed voxels sampled to evaluate the metric
(speeds up large volumes); default \code{0.2}}

\item{interpolation}{output interpolation used when warping each modality
onto the target grid: \code{'trilinear'} (default), \code{'nearest'} (keeps
label or segmentation values intact), or \code{"bspline"} (cubic
\verb{Catmull-Rom}, higher quality, slower). Like \code{metric}, supply a
single value (applied to every channel) or one per source/target pair. This
affects only the returned warped images, never the optimization's internal
sampling (always \code{'trilinear'}, so every channel still produces
smooth gradients)}

\item{number_of_bins}{number of histogram bins for the \code{"mattes"}
metric; default \code{32}}

\item{seed}{random seed for the voxel sampler, for reproducibility}

\item{init_transform}{optional 4x4 initial \verb{RAS}-to-\verb{RAS}
transform (fixed to moving) to start from}

\item{syn_iterations, syn_sigma}{deformable stage controls (only used when
\code{type = "syn"} or \code{"syn_only"}): per-level iteration counts and the
Gaussian regularization sigma (in voxels) applied to the update field. Both
are recycled to the number of levels, so \code{syn_sigma} may be a vector to
vary the regularization per stage (e.g. \code{c(3, 3, 1)} to relax it at the
finest level for sharper detail)}

\item{verbose}{logical; if \code{TRUE} (default) print per-level and
per-iteration progress to the console, including the current stage, shrink
factor, smoothing sigma, cost metric, and step size (linear) or maximum
displacement (deformable); useful to monitor convergence on large volumes}
}
\value{
A list with:
\describe{
\item{\code{transform}}{the estimated 4x4 \verb{RAS}-to-\verb{RAS} linear
transform mapping \code{target} (fixed) coordinates to \code{source} (moving)
coordinates}
\item{\code{image}}{the (primary) \code{source} resampled onto the
\code{target} grid}
\item{\code{images}}{a list with every source channel resampled onto the
\code{target} grid using its own \code{interpolation}; \code{image} is the
first element. For single-image input this is a length-1 list}
\item{\code{forward_field},\code{inverse_field}}{(only for \code{"syn"}) the
deformation fields}
\item{\code{metric_trace}}{the metric value across optimizer iterations}
\item{\code{type},\code{metric}}{echoes of the inputs}
}
}
\description{
Self-contained image registration for 3D volumes, implemented purely in
\pkg{RcppEigen} (no other external registration library). It mirrors the
core behavior of \pkg{'ANTs'} \code{antsRegistration}: a multi-resolution,
physical-shift scaled gradient-descent optimizer driving a similarity metric,
working entirely in the anatomical \verb{RAS} (right-anterior-superior)
space. Each volume carries its own \code{vox2ras} (0-indexed voxel index to
anatomical \verb{RAS}) \eqn{4\times 4} transform, so volumes with
different sampling, orientation, or field of view are aligned correctly.
}
\examples{
\donttest{

# --- synthetic same-modality example -------------------------------------
nd <- c(50, 50, 50)
vox2ras <- diag(4); vox2ras[1:3, 4] <- -25
blob <- function(cx, cy, cz, s = 6) {
  g <- expand.grid(x = 0:(nd[1]-1), y = 0:(nd[2]-1), z = 0:(nd[3]-1))
  array(exp(-((g$x-cx)^2 + (g$y-cy)^2 + (g$z-cz)^2) / (2*s^2)), nd)
}
target <- blob(25, 25, 25)
source <- blob(28, 23, 26)            # shifted by a known (3, -2, 1) mm

res <- register_volume3d(
  source, target,
  source_vox2ras = vox2ras, target_vox2ras = vox2ras,
  type = "rigid", metric = "cc"
)
res$transform[1:3, 4]                 # ~ c(3, -2, 1)


# --- multimodal registration (several co-registered channels) ------------
# Two aligned modalities (e.g. a T1 and a T2) jointly drive the deformable
# stage. Each pair may use its own metric, and `weights` set their relative
# influence (normalized internally to sum to 1). All source channels must
# share a grid; likewise all target channels.
t1_target <- blob(25, 25, 25, s = 6)
t2_target <- blob(25, 25, 25, s = 9)  # same anatomy, different contrast
t1_source <- blob(27, 24, 26, s = 6)  # both channels share the same warp
t2_source <- blob(27, 24, 26, s = 9)

res_mm <- register_volume3d(
  source = list(t1_source, t2_source),
  target = list(t1_target, t2_target),
  source_vox2ras = vox2ras, target_vox2ras = vox2ras,
  weights = c(2, 1),                  # T1 counts twice as much as T2
  metric = c("mattes", "cc"),         # one metric per channel
  type = "syn"
)
res_mm$transform[1:3, 4]

}
}
\seealso{
\code{\link{apply_transform3d}}, \code{\link{resample_3d_volume}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convolve.R
\name{convolve}
\alias{convolve}
\alias{convolve_signal}
\alias{convolve_image}
\alias{convolve_volume}
\title{Convolution of \code{1D}, \code{2D}, \code{3D} data via \code{FFT}}
\usage{
convolve_signal(x, filter)

convolve_image(x, filter)

convolve_volume(x, filter)
}
\arguments{
\item{x}{one-dimensional signal vector, two-dimensional image, or
three-dimensional volume; numeric or complex}

\item{filter}{kernel with the same number of dimensions as \code{x}}
}
\value{
Convolution results with the same length and dimensions as \code{x}.
If \code{x} is complex, results will be complex, otherwise results will
be real numbers.
}
\description{
Use the 'Fast-Fourier' transform to compute the convolutions of two data
with zero padding. This function is mainly designed for image convolution.
For forward and backward convolution/filter, see \code{\link{filtfilt}}.
}
\details{
This implementation uses 'Fast-Fourier' transform to perform
\code{1D}, \code{2D}, or \code{3D} convolution. Compared to implementations
using original mathematical definition of convolution, this approach is
much faster, especially for image and volume convolutions.

The input \code{x} is zero-padded beyond edges. This is most common in image
or volume convolution, but less optimal for periodic one-dimensional signals.
Please use other implementations if non-zero padding is needed.

The convolution results might be different to the ground truth by a precision
error, usually at \code{1e-13} level, depending on the \code{'FFTW3'}
library precision and implementation.
}
\examples{


# ---- 1D convolution ------------------------------------
x <- cumsum(rnorm(100))
filter <- dnorm(-2:2)
# normalize
filter <- filter / sum(filter)
smoothed <- convolve_signal(x, filter)

plot(x, pch = 20)
lines(smoothed, col = 'red')

# ---- 2D convolution ------------------------------------
x <- array(0, c(100, 100))
x[
  floor(runif(10, min = 1, max = 100)),
  floor(runif(10, min = 1, max = 100))
] <- 1

# smooth
kernel <- outer(dnorm(-2:2), dnorm(-2:2), FUN = "*")
kernel <- kernel / sum(kernel)

y <- convolve_image(x, kernel)

oldpar <- par(mfrow = c(1,2))
image(x, asp = 1, axes = FALSE, main = "Origin")
image(y, asp = 1, axes = FALSE, main = "Smoothed")
par(oldpar)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shift_array.R
\name{shift_array}
\alias{shift_array}
\title{Shift array by index}
\usage{
shift_array(x, along_margin, unit_margin, shift_amount)
}
\arguments{
\item{x}{array, must have at least matrix}

\item{along_margin}{which index is to be shifted}

\item{unit_margin}{which dimension decides \code{shift_amount}}

\item{shift_amount}{shift amount along \code{along_margin}}
}
\value{
An array with same dimensions as the input \code{x}, but with
index shifted. The missing elements will be filled with \code{NA}.
}
\description{
Re-arrange arrays in parallel
}
\details{
A simple use-case for this function is to think of a matrix where each row
is a signal and columns stand for time. The objective is to align (time-lock)
each signal according to certain events. For each signal, we want to shift
the time points by certain amount.

In this case, the shift amount is defined by \code{shift_amount}, whose
length equals to number of signals. \code{along_margin=2} as we want to shift
time points (column, the second dimension) for each signal. \code{unit_margin=1}
because the shift amount is depend on the signal number.
}
\examples{

# Set ncores = 2 to comply to CRAN policy. Please don't run this line
ravetools_threads(n_threads = 2L)



x <- matrix(1:10, nrow = 2, byrow = TRUE)
z <- shift_array(x, 2, 1, c(1,2))

y <- NA * x
y[1,1:4] = x[1,2:5]
y[2,1:3] = x[2,3:5]

# Check if z ang y are the same
z - y

# array case
# x is Trial x Frequency x Time
x <- array(1:27, c(3,3,3))

# Shift time for each trial, amount is 1, -1, 0
shift_amount <- c(1,-1,0)
z <- shift_array(x, 3, 1, shift_amount)

oldpar <- par(mfrow = c(3, 2), mai = c(0.8, 0.6, 0.4, 0.1))
for( ii in 1:3 ) {
  image(t(x[ii, ,]), ylab = 'Frequency', xlab = 'Time',
        main = paste('Trial', ii))
  image(t(z[ii, ,]), ylab = 'Frequency', xlab = 'Time',
        main = paste('Shifted amount:', shift_amount[ii]))
}
par(oldpar)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexports.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{butter}
\alias{buttap}
\alias{buttord}
\alias{cheb1ap}
\alias{cheb1ord}
\alias{cheb2ap}
\alias{cheb2ord}
\alias{cheby1}
\alias{cheby2}
\alias{ellip}
\alias{ellipap}
\alias{ellipord}
\alias{resample}
\alias{hilbert}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{gsignal}{\code{\link[gsignal:buttap]{buttap()}}, \code{\link[gsignal:butter]{butter()}}, \code{\link[gsignal:buttord]{buttord()}}, \code{\link[gsignal:cheb1ap]{cheb1ap()}}, \code{\link[gsignal:cheb1ord]{cheb1ord()}}, \code{\link[gsignal:cheb2ap]{cheb2ap()}}, \code{\link[gsignal:cheb2ord]{cheb2ord()}}, \code{\link[gsignal:cheby1]{cheby1()}}, \code{\link[gsignal:cheby2]{cheby2()}}, \code{\link[gsignal:ellip]{ellip()}}, \code{\link[gsignal:ellipap]{ellipap()}}, \code{\link[gsignal:ellipord]{ellipord()}}, \code{\link[gsignal:hilbert]{hilbert()}}, \code{\link[gsignal:resample]{resample()}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-filtfilt.R
\name{filter_signal}
\alias{filter_signal}
\title{Filter one-dimensional signal}
\usage{
filter_signal(b, a, x, z)
}
\arguments{
\item{b}{one-dimensional real numerical vector, the moving-average
coefficients of an \code{ARMA} filter}

\item{a}{the auto-regressive (recursive) coefficients of an \code{ARMA} filter}

\item{x}{numerical vector input (real value)}

\item{z}{initial condition, must have length of \code{n-1}, where \code{n}
is the maximum of lengths of \code{a} and \code{b}; default is all zeros}
}
\value{
A list of two vectors: the first vector is the filtered signal;
the second vector is the final state of \code{z}
}
\description{
The function is written from the scratch. The result has been
compared against the 'Matlab' \code{filter} function with one-dimensional
real inputs. Other situations such as matrix \code{b} or multi-dimensional
\code{x} are not implemented. For double filters (forward-backward),
see \code{\link{filtfilt}}.
}
\examples{


t <- seq(0, 1, by = 0.01)
x <- sin(2 * pi * t * 2.3)
bf <- gsignal::butter(2, c(0.15, 0.3))

res <- filter_signal(bf$b, bf$a, x)
y <- res[[1]]
z <- res[[2]]

## Matlab (2022a) equivalent:
# t = [0:0.01:1];
# x = sin(2 * pi * t * 2.3);
# [b,a] = butter(2,[.15,.3]);
# [y,z] = filter(b, a, x)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/curve-catmull.R
\name{print.ravetools_curve}
\alias{print.ravetools_curve}
\title{Print method for \code{ravetools_curve}}
\usage{
\method{print}{ravetools_curve}(x, ...)
}
\arguments{
\item{x}{an object of class \code{ravetools_curve}.}

\item{...}{additional arguments passed to \code{\link{print.data.frame}}.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Prints a concise summary of a \code{ravetools_curve} object
returned by \code{\link{catmull_rom_3d}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-ants-io.R
\name{write_ants_transform}
\alias{write_ants_transform}
\alias{read_ants_transform}
\title{Read and write an \verb{'ITK'}/\verb{'ANTs'} \verb{affine} transform}
\usage{
write_ants_transform(transform, file)

read_ants_transform(file)
}
\arguments{
\item{transform}{a \eqn{4\times 4} (or \eqn{3\times 4}) \verb{RAS} \verb{affine}}

\item{file}{path to the \code{.mat} file}
}
\value{
\code{write_ants_transform} returns \code{file} invisibly;
\code{read_ants_transform} returns a \eqn{4\times 4} \verb{RAS} matrix.
}
\description{
\code{write_ants_transform} stores a \eqn{4\times 4} \verb{RAS}-to-\verb{RAS}
\verb{affine} (such as the \code{transform} returned by
\code{\link{register_volume3d}}) as an \pkg{'ITK'} \code{.mat} file (a
\code{'MATLAB'} level-5 binary, \verb{LPS} convention) compatible with
\code{antsApplyTransforms}. \code{read_ants_transform} reads such a file back
into a \eqn{4\times 4} \verb{RAS} matrix. The reader folds a non-zero
\verb{ITK} center of rotation into the translation, so transforms written by
\pkg{'ANTs'} are read correctly.
}
\examples{
tf <- tempfile(fileext = ".mat")
m <- diag(4); m[1:3, 4] <- c(3, -2, 1)
write_ants_transform(m, tf)
read_ants_transform(tf)
}
\seealso{
\code{\link{register_volume3d}}, \code{\link{save_registration}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-utils.R
\name{filter-window}
\alias{filter-window}
\alias{hanning}
\alias{hamming}
\alias{blackman}
\alias{blackmannuttall}
\alias{blackmanharris}
\alias{flattopwin}
\alias{bohmanwin}
\title{Filter window functions}
\usage{
hanning(n)

hamming(n)

blackman(n)

blackmannuttall(n)

blackmanharris(n)

flattopwin(n)

bohmanwin(n)
}
\arguments{
\item{n}{number of time-points in window}
}
\value{
A numeric vector of window with length \code{n}
}
\description{
Filter window functions
}
\examples{

hanning(10)
hamming(11)
blackmanharris(21)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-butter.R
\name{butter_max_order}
\alias{butter_max_order}
\title{'Butterworth' filter with maximum order}
\usage{
butter_max_order(
  w,
  type = c("low", "high", "pass", "stop"),
  r = 10 * log10(2),
  tol = .Machine$double.eps
)
}
\arguments{
\item{w}{scaled frequency ranging from 0 to 1, where 1 is 'Nyquist' frequency}

\item{type}{filter type}

\item{r}{decibel attenuation at frequency \code{w}, default is around
\code{3 dB} (half power)}

\item{tol}{tolerance of reciprocal condition number, default is
\code{.Machine$double.eps}.}
}
\value{
'Butterworth' filter in 'Arma' form.
}
\description{
Large filter order might not be optimal, but at lease this function
provides a feasible upper bound for the order such that the
filter has a stable \code{AR} component.
}
\examples{

# Find highest order (sharpest transition) of a band-pass filter
sample_rate <- 500
nyquist <- sample_rate / 2

type <- "pass"
w <- c(1, 50) / nyquist
Rs <- 6     # power attenuation at w

# max order filter
filter <- butter_max_order(w, "pass", Rs)

# -6 dB cutoff should be around 1 ~ 50 Hz
diagnose_filter(filter$b, filter$a, fs = sample_rate)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_smooth}
\alias{vcg_smooth}
\alias{vcg_smooth_implicit}
\alias{vcg_smooth_explicit}
\title{Implicitly smooth a triangular mesh}
\usage{
vcg_smooth_implicit(
  mesh,
  lambda = 0.2,
  use_mass_matrix = TRUE,
  fix_border = FALSE,
  use_cot_weight = FALSE,
  degree = 1L,
  laplacian_weight = 1
)

vcg_smooth_explicit(
  mesh,
  type = c("taubin", "laplace", "HClaplace", "fujiLaplace", "angWeight",
    "surfPreserveLaplace"),
  iteration = 10,
  lambda = 0.5,
  mu = -0.53,
  delta = 0.1
)
}
\arguments{
\item{mesh}{triangular mesh stored as object of class 'mesh3d'.}

\item{lambda}{In \code{vcg_smooth_implicit}, the amount of smoothness,
useful only if \code{use_mass_matrix} is \code{TRUE}; default is \code{0.2}.
In \code{vcg_smooth_explicit}, parameter for \code{'taubin'} smoothing.}

\item{use_mass_matrix}{logical: whether to use mass matrix to keep the mesh
close to its original position (weighted per area distributed on vertices);
default is \code{TRUE}}

\item{fix_border}{logical: whether to fix the border vertices of the mesh;
default is \code{FALSE}}

\item{use_cot_weight}{logical: whether to use cotangent weight; default is
\code{FALSE} (using uniform 'Laplacian')}

\item{degree}{integer: degrees of 'Laplacian'; default is \code{1}}

\item{laplacian_weight}{numeric: weight when \code{use_cot_weight} is \code{FALSE};
default is \code{1.0}}

\item{type}{method name of explicit smooth, choices are \code{'taubin'},
\code{'laplace'}, \code{'HClaplace'}, \code{'fujiLaplace'},
\code{'angWeight'}, \code{'surfPreserveLaplace'}.}

\item{iteration}{number of iterations}

\item{mu}{parameter for \code{'taubin'} explicit smoothing.}

\item{delta}{parameter for scale-dependent 'Laplacian' smoothing or
maximum allowed angle (in 'Radian') for deviation between surface preserving
'Laplacian'.}
}
\value{
An object of class "mesh3d" with:
\item{\code{vb}}{vertex coordinates}
\item{\code{normals}}{vertex normal vectors}
\item{\code{it}}{triangular face index}
}
\description{
Applies smoothing algorithms on a triangular mesh.
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

if(is_not_cran()) {

# Prepare mesh with no normals
data("left_hippocampus_mask")

# Grow 2mm on each direction to fill holes
volume <- grow_volume(left_hippocampus_mask, 2)

# Initial mesh
mesh <- vcg_isosurface(volume)

# Start: examples
rgl_view({
  rgl_call("mfrow3d", 2, 4)
  rgl_call("title3d", "Naive ISOSurface")
  rgl_call("shade3d", mesh, col = 2)

  rgl_call("next3d")
  rgl_call("title3d", "Implicit Smooth")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_implicit(mesh, degree = 2))

  rgl_call("next3d")
  rgl_call("title3d", "Explicit Smooth - taubin")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_explicit(mesh, "taubin"))

  rgl_call("next3d")
  rgl_call("title3d", "Explicit Smooth - laplace")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_explicit(mesh, "laplace"))

  rgl_call("next3d")
  rgl_call("title3d", "Explicit Smooth - angWeight")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_explicit(mesh, "angWeight"))

  rgl_call("next3d")
  rgl_call("title3d", "Explicit Smooth - HClaplace")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_explicit(mesh, "HClaplace"))

  rgl_call("next3d")
  rgl_call("title3d", "Explicit Smooth - fujiLaplace")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_explicit(mesh, "fujiLaplace"))

  rgl_call("next3d")
  rgl_call("title3d", "Explicit Smooth - surfPreserveLaplace")
  rgl_call("shade3d", col = 2,
           x = vcg_smooth_explicit(mesh, "surfPreserveLaplace"))
})

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-ants-io.R
\name{save_registration}
\alias{save_registration}
\alias{save_registration.ravetools_register_volume3d}
\alias{save_registration.default}
\alias{load_registration}
\title{Save or load a registration result in \verb{'ANTs'}-compatible files}
\usage{
save_registration(x, path, ...)

\method{save_registration}{ravetools_register_volume3d}(x, path, prefix = "registration", compress = TRUE, ...)

\method{save_registration}{default}(x, path, prefix = "registration", ...)

load_registration(file, recover_affine_from_header = FALSE)
}
\arguments{
\item{x}{a \code{ravetools_register_volume3d} object (or, for the default
method, a \eqn{4\times 4} matrix)}

\item{path}{output directory, or the manifest (\code{.dcf}) path}

\item{...}{passed to methods}

\item{prefix}{file-name prefix; defaults to \code{"registration"}}

\item{compress}{logical; write warp fields as \code{.nii.gz} (default) or
\code{.nii}}

\item{file}{a \code{.mat}, \code{.nii}/\code{.nii.gz}, or \code{.dcf} path}

\item{recover_affine_from_header}{logical; only used as a last resort when
loading. If \code{TRUE} and the \verb{affine} \code{.mat} is missing (or a warp
file is loaded directly), the \verb{affine} is recovered from the warp
header's hidden encoding. Off by default because that encoding is
reduced-precision}
}
\value{
\code{save_registration} returns the manifest path invisibly.
\code{load_registration} returns a \eqn{4\times 4} matrix, a field array
(with a \code{"vox2ras"} attribute), or a \code{ravetools_register_volume3d}
object, depending on the file type.
}
\description{
\code{save_registration} writes a \code{\link{register_volume3d}} result to
disk as \pkg{'ANTs'}-style files: an \verb{'ITK'} \verb{affine} \code{.mat}, the
forward/inverse warp \code{'NIfTI'}(s) when present, and a small
\verb{'DCF'} manifest (the same key-value format as \code{'DESCRIPTION'})
recording what each file is plus the registration parameters.
\code{load_registration} reads any of those back: a \code{.mat} returns a
\eqn{4\times 4} \verb{RAS} matrix, a warp \code{.nii}/\code{.nii.gz} returns
the field, and a \code{.dcf} manifest reassembles the full object.
}
\seealso{
\code{\link{register_volume3d}}, \code{\link{write_ants_transform}},
\code{\link{write_ants_warp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imaging-registration.R
\name{register_volume}
\alias{register_volume}
\title{Imaging registration using \code{'NiftyReg'}}
\usage{
register_volume(
  source,
  target,
  method = c("rigid", "affine", "nonlinear"),
  interpolation = c("cubic", "trilinear", "nearest"),
  threads = detect_threads(),
  symmetric = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{source}{source imaging data, or a \code{'nifti'} file path;
for example, 'CT'}

\item{target}{target imaging data to align to; for example, 'MRI'}

\item{method}{method of transformation, choices are \code{'rigid'},
\code{'affine'}, or \code{'nonlinear'}}

\item{interpolation}{how volumes should be interpolated, choices are
\code{'cubic'}, \code{'trilinear'}, or \code{'nearest'}}

\item{threads, symmetric, verbose, ...}{see \code{\link[RNiftyReg]{niftyreg}}}
}
\value{
See \code{\link[RNiftyReg]{niftyreg}}
}
\description{
Registers 'CT' to 'MRI', or 'MRI' to another 'MRI'
}
\examples{

\donttest{

source <- system.file("extdata", "epi_t2.nii.gz", package="RNiftyReg")
target <- system.file("extdata", "flash_t1.nii.gz", package="RNiftyReg")
aligned <- register_volume(source, target, verbose = FALSE)

source_img <- aligned$source[[1]]
target_img <- aligned$target
aligned_img <- aligned$image

oldpar <- par(mfrow = c(2, 2), mar = c(0.1, 0.1, 3.1, 0.1))

pal <- grDevices::grey.colors(256, alpha = 1)
image(source_img[,,30], asp = 1, axes = FALSE,
      col = pal, main = "Source image")
image(target_img[,,64], asp = 1, axes = FALSE,
      col = pal, main = "Target image")
image(aligned_img[,,64], asp = 1, axes = FALSE,
      col = pal, main = "Aligned image")

# bucket fill and calculate differences
aligned_img[is.nan(aligned_img) | aligned_img <= 1] <- 1
target_img[is.nan(target_img) | aligned_img <= 1] <- 1
diff <- abs(aligned_img / target_img - 1)
image(diff[,,64], asp = 1, axes = FALSE,
      col = pal, main = "Percentage Difference")

par(oldpar)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stimulation_pauses.R
\name{stimpulse_interpolate}
\alias{stimpulse_interpolate}
\alias{stimpulse_find}
\alias{stimpulse_extract}
\alias{stimpulse_align}
\title{Find and interpolate stimulation pulses}
\usage{
stimpulse_find(
  signal,
  sample_rate,
  pulse_duration,
  n_pulses = NA,
  threshold = NA
)

stimpulse_extract(
  signal,
  pulse_info,
  expand_timepoints = c(-10, 20),
  center = TRUE
)

stimpulse_align(signal, pulse_info, expand_timepoints = c(-10, 20))

stimpulse_interpolate(
  signal,
  sample_rate,
  pulse_info,
  max_offset = c(-2e-04, 5e-04)
)
}
\arguments{
\item{signal}{a channel signal trace}

\item{sample_rate}{sample rate}

\item{pulse_duration}{stimulation pulse duration in seconds}

\item{n_pulses}{suggested number of pulses}

\item{threshold}{suggested suggested threshold of responses to find
stimulation pulses}

\item{pulse_info}{a list containing number of pulses \code{n_pulse},
onset index (\code{onset_index}, first time-point is 1), offset index
(\code{offset_index}); this can be generated by \code{stimpulse_find};
see 'Examples' below}

\item{expand_timepoints}{point offsets allowed to align the pulses}

\item{center}{whether to center the pulses by median; default is true}

\item{max_offset}{maximum (edge) offsets in seconds to interpolate the
pulses; default is \code{-0.0002} seconds before stimulation onset
and \code{0.0005} seconds after the stimulation offset}
}
\value{
\code{stimpulse_find} and \code{stimpulse_align} returns the pulse
information (\code{pulse_info}) with the time-points of detected or corrected
stimulation on-set and off-set. The time-points are 1-indexed.
\code{stimpulse_extract} extract the signals around pulses;
\code{stimpulse_interpolate} returns interpolated signals.
}
\description{
Find and interpolate stimulation pulses
}
\examples{



data("stimulation_signal")

signal <- stimulation_signal$signal
sample_rate <- stimulation_signal$sample_rate

# each pulse is roughly <0.001 seconds
pulse_durations <- 0.001

# Initial pulses
pulse_info <- stimpulse_find(signal, sample_rate, pulse_durations)

# number of pulses detected
pulse_info$n_pulses

# extract responses -10 points before onset ~ 20 points after offset
expand_timepoints <- c(-20, 80)
pulses_snippets <- stimpulse_extract(
  signal = signal,
  pulse_info = pulse_info,
  expand_timepoints = expand_timepoints
)

# Visualize the pulses
snippet_time <- seq(
  expand_timepoints[[1]], by = 1,
  length.out = nrow(pulses_snippets)) / sample_rate * 1000
matplot(snippet_time, pulses_snippets, type = 'l', lty = 1, col = 'gray80',
        xlab = "Time (ms)", ylab = "uV", main = "Initial find")

# Align the pulses
pulse_info <- stimpulse_align(signal, pulse_info)

# Estimated pulse duration
estimated_duration <-
  (pulse_info$offset_index - pulse_info$onset_index + 1) / sample_rate

# reload aligned pulses
pulses_snippets <- stimpulse_extract(
  signal = signal,
  pulse_info = pulse_info,
  expand_timepoints = expand_timepoints
)
matplot(snippet_time, pulses_snippets, type = 'l', lty = 1, col = 'gray80',
        xlab = "Time (ms)", ylab = "uV", main = "Aligned pulses")
lines(snippet_time, rowMeans(pulses_snippets), col = 'red')


# Interpolate the pulses
interpolated <- stimpulse_interpolate(
  signal = signal,
  sample_rate = sample_rate,
  pulse_info = pulse_info,
  max_offset = c(-0.0003, 0.0005)
)


interp_snippets <- stimpulse_extract(
  signal = interpolated,
  pulse_info = pulse_info,
  expand_timepoints = expand_timepoints
)

oldpar <- par(mfrow = c(1, 2))
on.exit(par(oldpar))
matplot(snippet_time, pulses_snippets, type = 'l', lty = 1,
        col = 'gray80', xlab = "Time (ms)", ylab = "uV",
        main = "Stim pulses", ylim = c(-600, 400))
lines(snippet_time, rowMeans(pulses_snippets), col = 'red')
abline(v = max(estimated_duration) * 1000, lty = 2)

matplot(snippet_time, interp_snippets, type = 'l', lty = 1,
        col = 'gray80', xlab = "Time (ms)", ylab = "uV",
        main = "Interpolated 0.5 ms bandwidth")
lines(snippet_time, rowMeans(interp_snippets), col = 'red')
abline(v = max(estimated_duration) * 1000, lty = 2, col = "gray40")
abline(v = max(estimated_duration) * 1000 + 0.5, lty = 2)






}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mris-sphere.R
\name{mris_sphere}
\alias{mris_sphere}
\title{Project a surface onto a sphere and relax metric distortion}
\usage{
mris_sphere(
  mesh,
  target_radius = 100,
  n_averages = 64L,
  niterations = 25L,
  l_dist = 1,
  l_area = 1,
  momentum = 0.9,
  dt = 0.05,
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}, typically an inflated
surface (see \code{\link{mris_inflate}}).  Must be closed, manifold, and
genus-0 - the same hard precondition as \code{\link{mris_inflate}}.
\code{mris_sphere} will raise an error if such defects are detected.}

\item{target_radius}{radius of the target sphere.  Default \code{100}.}

\item{n_averages}{number of gradient-averaging passes applied to the
distance-term gradient each iteration (the same neighborhood-averaging
mechanism \code{\link{mris_inflate}} uses).  Default \code{64}; reduced
from the much larger averaging size used in production pipelines (tuned
for cortical surfaces with hundreds of thousands of vertices) to a size
more practical for typical \code{'mesh3d'} inputs.}

\item{niterations}{number of unfolding iterations.  Default \code{25}.}

\item{l_dist}{distance-preservation coefficient.  Default \code{1.0}.}

\item{l_area}{folded-face repulsion coefficient.  Default \code{1.0}.}

\item{momentum}{momentum coefficient.  Default \code{0.9}.}

\item{dt}{time step.  Default \code{0.05}.}

\item{verbose}{logical; print per-iteration progress.  Default \code{FALSE}.}
}
\value{
The projected and relaxed surface as a \code{'mesh3d'} object with
  \code{vb}, \code{it}, and \code{normals}.
}
\description{
Projects a closed triangular surface mesh radially onto a sphere and then
iteratively relaxes the metric distortion this introduces ("unfolding"), so
that geodesic distances and face orientations stay close to those of the
input surface - a step typically used to prepare an inflated cortical
surface for spherical registration.
}
\details{
\strong{This is a reduced procedure, not a complete reproduction of any
particular reference implementation.} The full unfolding procedure
described in the literature integrates seven weighted energy terms against
a separate reference surface through a multi-resolution, multi-thousand
line optimization pipeline, which is out of scope for this package. Instead,
this implementation keeps the two dominant terms and integrates them with
the same momentum-based machinery \code{\link{mris_inflate}} uses:
\describe{
  \item{Distance term (\code{l_dist})}{Restoring force pulling each vertex
    back towards the input mesh's own original distances to its
    neighbors.}
  \item{Area term (\code{l_area})}{Repulsive force that acts only on
    folded/negative-area faces, pushing their vertices apart - this is the
    actual "unfolding" mechanism.}
}
Both terms are integrated via momentum integration with gradient averaging
and a 1 mm per-step displacement cap, exactly as \code{\link{mris_inflate}}
does.

A further simplification: the reference procedure relaxes a freshly
spherical-projected surface against a separate, previously-loaded
white-matter reference surface for the distance term; since this function
takes a single mesh as input, the input mesh's own metric (captured before
projection) is used as the reference instead - the same convention
\code{\link{mris_inflate}} already uses, and the practical analogue (the
white-matter surface is normally what gets inflated to produce a typical
input to this kind of unfolding step).
}
\examples{
if (is_not_cran()) {

sphere <- vcg_sphere(sub_division = 3L)

# deform so there is metric distortion to relax
sphere$vb[1, ] <- sphere$vb[1, ] * (1 + 0.2 * rnorm(ncol(sphere$vb)))

# Fix defects
sphere <- vcg_fix_defects(sphere)

result <- mris_sphere(
  sphere,
  n_averages = 8L,
  niterations = 10L,
  target_radius = 1,
  verbose = TRUE
)

plot_mesh_polygon(list(sphere, result),
                  col = list("gray", "red"),
                  alpha = c(0.3, 0.9))



}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pwelch-diagnosis.R
\name{diagnose_channel}
\alias{diagnose_channel}
\title{Show channel signals with diagnostic plots}
\usage{
diagnose_channel(
  s1,
  s2 = NULL,
  sc = NULL,
  srate,
  name = "",
  try_compress = TRUE,
  max_freq = 300,
  window = ceiling(srate * 2),
  noverlap = window/2,
  std = 3,
  which = NULL,
  main = "Channel Inspection",
  col = c("black", "red"),
  cex = 1.2,
  cex.lab = 1,
  lwd = 0.5,
  plim = NULL,
  nclass = 100,
  start_time = 0,
  boundary = NULL,
  mar = c(3.1, 4.1, 2.1, 0.8) * (0.25 + cex * 0.75) + 0.1,
  mgp = cex * c(2, 0.5, 0),
  xaxs = "i",
  yaxs = "i",
  xline = 1.66 * cex,
  yline = 2.66 * cex,
  tck = -0.005 * (3 + cex),
  ...
)
}
\arguments{
\item{s1}{the main signal to draw}

\item{s2}{the comparing signal to draw; usually \code{s1} after some filters;
must be in the same sampling rate with \code{s1}; can be \code{NULL}}

\item{sc}{decimated \code{s1} to show if \code{srate} is too high; will
be automatically generated if \code{NULL}}

\item{srate}{sampling rate}

\item{name}{name of \code{s1}, or a vector of two names of \code{s1} and
\code{s2} if \code{s2} is provided}

\item{try_compress}{whether try to compress (decimate) \code{s1} if
\code{srate} is too high for performance concerns}

\item{max_freq}{the maximum frequency to display in 'Welch Periodograms'}

\item{window, noverlap}{see \code{\link{pwelch}}}

\item{std}{the standard deviation of the channel signals used to determine
\code{boundary}; default is plus-minus 3 standard deviation}

\item{which}{\code{NULL} or integer from 1 to 4; if \code{NULL}, all plots
will be displayed; otherwise only the subplot will be displayed}

\item{main}{the title of the signal plot}

\item{col}{colors of \code{s1} and \code{s2}}

\item{cex, lwd, mar, cex.lab, mgp, xaxs, yaxs, tck, ...}{graphical parameters; see
\code{\link[graphics]{par}}}

\item{plim}{the y-axis limit to draw in 'Welch Periodograms'}

\item{nclass}{number of classes to show in histogram
(\code{\link[graphics]{hist}})}

\item{start_time}{the starting time of channel (will only be used to draw
signals)}

\item{boundary}{a red boundary to show in channel plot; default is
to be automatically determined by \code{std}}

\item{xline, yline}{distance of axis labels towards ticks}
}
\value{
A list of boundary and y-axis limit used to draw the channel
}
\description{
The diagnostic plots include 'Welch Periodogram'
(\code{\link{pwelch}}) and histogram (\code{\link[graphics]{hist}})
}
\examples{
library(ravetools)

# Generate 20 second data at 2000 Hz
time <- seq(0, 20, by = 1 / 2000)
signal <- sin( 120 * pi * time) +
  sin(time * 20*pi) +
  exp(-time^2) *
  cos(time * 10*pi) +
  rnorm(length(time))

signal2 <- notch_filter(signal, 2000)

diagnose_channel(signal, signal2, srate = 2000,
                 name = c("Raw", "Filtered"), cex = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_count_edge_defects}
\alias{vcg_count_edge_defects}
\title{Count boundary and non-manifold edges of a triangular mesh}
\usage{
vcg_count_edge_defects(mesh)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'}.}
}
\value{
A named list with elements \code{boundary_edges} (number of
boundary edges), \code{nonmanifold_edges} (number of non-manifold edges),
and \code{is_closed_manifold} (\code{TRUE} when both counts are zero,
i.e. the mesh is closed and manifold and ready for
\code{\link{mris_inflate}}).
}
\description{
Detects topology defects that prevent a mesh from being a closed,
manifold, genus-0 surface, a hard precondition of algorithms such as
\code{\link{mris_inflate}}.  An edge is a \emph{boundary} edge when it is
referenced by exactly one face (i.e. it bounds a hole), and
\emph{non-manifold} when it is referenced by more than two faces.
}
\examples{
if (is_not_cran()) {

  sphere <- vcg_sphere()
  vcg_count_edge_defects(sphere)

  defective <- vcg_isosurface(left_hippocampus_mask)
  vcg_count_edge_defects(defective)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcg.R
\name{vcg_update_normals}
\alias{vcg_update_normals}
\title{Update vertex normal}
\usage{
vcg_update_normals(
  mesh,
  weight = c("area", "angle"),
  pointcloud = c(10, 0),
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh or a point-cloud (matrix of 3 columns)}

\item{weight}{method to compute per-vertex normal vectors: \code{"area"}
weighted average of surrounding face normal, or \code{"angle"} weighted
vertex normal vectors.}

\item{pointcloud}{integer vector of length 2: containing optional
parameters for normal calculation of point clouds; the first entry
specifies the number of neighboring points to consider; the second
entry specifies the amount of smoothing iterations to be performed.}

\item{verbose}{whether to verbose the progress}
}
\value{
A \code{'mesh3d'} object with normal vectors.
}
\description{
Update vertex normal
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{

if(is_not_cran()) {

# Prepare mesh with no normal
data("left_hippocampus_mask")
mesh <- vcg_isosurface(left_hippocampus_mask)
mesh$normals <- NULL

# Start: examples
new_mesh <- vcg_update_normals(mesh, weight = "angle",
                               pointcloud = c(10, 10))

rgl_view({
  rgl_call("mfrow3d", 1, 2)
  rgl_call("shade3d", mesh, col = 2)

  rgl_call("next3d")
  rgl_call("shade3d", new_mesh, col = 2)
})
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geometry-plane.R
\name{project_plane}
\alias{project_plane}
\title{Project plane to a surface}
\usage{
project_plane(
  target,
  width,
  height,
  shape,
  initial_positions,
  translate_first = TRUE,
  diagnostic = FALSE,
  n_iters = 5
)
}
\arguments{
\item{target}{target surface to be projected to, must be object that can
be converted to \code{'mesh3d'} (\code{'rgl'} surface), for example,
\code{'fs.surface'} (from \code{'freesurferformat'} package) or
\code{'ieegio_surface'} from \pkg{ieegio} package; see
\code{\link{ensure_mesh3d}} for the coercion rules.}

\item{width, height}{width and height of the plane in world space (for
\code{'ECoG'} grid, the unit is millimeter)}

\item{shape}{vector of two integers: the first element is the number of
vertices (or electrode contacts) along \code{'width'} direction; the second
element is the number of vertices along \code{'height'} direction. The total
number of vertices of the plane will be \code{prod(shape)}. Notice}

\item{initial_positions}{a \code{shape[[1]] x shape[[2]] x 3} array or a
\code{n x 3} matrix, where \code{n} is \code{prod(shape)}, the number of
vertices indicating the initial vertex positions of the plane}

\item{translate_first}{whether to translate the plane first if the
plane center is far from the surface; default is \code{FALSE}; set to
\code{TRUE} for a warm start}

\item{diagnostic}{whether to plot diagnostic figures showing the morphing
progress.}

\item{n_iters}{number of iterations; default is five}
}
\value{
The projected vertex locations, same order as \code{initial_positions}.
}
\description{
Project a two-dimensional plane (such as \code{'ECoG'} grid) to a
three-dimensional surface while preserving the order
}
\section{Coercing Surface Inputs}{

The surface objects are converted to \code{'mesh3d'} object before
applying further calculations.

When \code{surface} is a surface \pkg{ieegio} object, the returned
\code{mesh3d$vb} contains vertices that have been left-multiplied by
\code{surface$geometry$transforms[[1]]} (the first transform stored in the
geometry, typically the \code{ScannerAnat} or voxel-to-world transform).

\strong{Breaking change:} Earlier versions (before 0.2.6) of \pkg{ravetools}
returned the raw \code{surface$geometry$vertices} without applying any
transform, so downstream code often multiplied by
\code{surface$geometry$transforms[[1]]} (or an equivalent) manually before
working in world space. Such code will now \emph{double}
apply the transform and produce incorrect coordinates. If you previously
applied a transform from \code{surface$geometry$transforms} by hand after
calling a \pkg{ravetools} mesh function on an \code{'ieegio_surface'},
remove that manual step.

Surfaces with an empty or missing \code{geometry$transforms} list (for
example, surfaces produced by \pkg{ieegio}'s \code{volume_to_surface},
which stores an identity transform) are unaffected.

If \code{geometry$transforms} contains multiple transforms targeting
different coordinate spaces, only the first one is used. Callers that need
a specific target space should select and apply that transform themselves
before calling \pkg{ravetools} mesh functions.
}

\examples{


# Construct target surface

sphere <- vcg_sphere()
target <- structure(
  class = "mesh3d",
  list(
    vb = cbind(
      sphere$vb[1:3, ] - c(0.8, 0, 0),
      sphere$vb[1:3, ] + c(0.8, 0, 0)
    ),
    it = cbind(
      sphere$it[1:3, ],
      sphere$it[1:3, ] + ncol(sphere$vb)
    )
  )
)
n_surfverts <- ncol(target$vb)

plane <- plane_geometry(width = 3, height = 3, shape = c(30, 30))
plane$vb <- plane$vb[1:3, , drop = FALSE] + c(0, 0, 2)
n_contacts <- ncol(plane$vb)

# First plot
x <- t(cbind(target$vb, plane$vb))
colnames(x) <- c('x', 'y', 'z')
graphics::pairs(
  x = x, asp = 1,
  col = c(
    rep("black", n_surfverts),
    rep("green", n_contacts)
  ),
  pch = c(
    rep(46, n_surfverts),
    rep(20, n_contacts)
  )
)

projected <- project_plane(
  target = target, width = 3, height = 3, shape = c(30, 30),
  initial_positions = t(plane$vb),
  translate_first = TRUE, diagnostic = FALSE
)

y <- rbind(x, projected)
graphics::pairs(
  x = y, asp = 1,
  col = c(
    rep("black", ncol(target$vb)),
    rep("green", n_contacts),
    rep("red", n_contacts)
  ),
  pch = c(
    rep(46, n_surfverts),
    rep(1, n_contacts),
    rep(20, n_contacts)
  )
)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wavelet.R
\name{wavelet}
\alias{wavelet}
\alias{wavelet_kernels}
\alias{morlet_wavelet}
\alias{wavelet_cycles_suggest}
\title{'Morlet' wavelet transform (Discrete)}
\usage{
wavelet_kernels(freqs, srate, wave_num)

morlet_wavelet(
  data,
  freqs,
  srate,
  wave_num,
  precision = c("float", "double"),
  trend = c("constant", "linear", "none"),
  signature = NULL,
  segment_length = NULL,
  ...
)

wavelet_cycles_suggest(
  freqs,
  frequency_range = c(2, 200),
  cycle_range = c(3, 20)
)
}
\arguments{
\item{freqs}{frequency in which \code{data} will be projected on}

\item{srate}{sample rate, number of time points per second}

\item{wave_num}{desired number of cycles in wavelet kernels to
balance the precision in time and amplitude (control the
smoothness); positive integers are strongly suggested}

\item{data}{numerical vector such as analog voltage signals}

\item{precision}{the precision of computation; choices are
\code{'float'} (default) and \code{'double'}.}

\item{trend}{choices are \code{'constant'}: center the signal at zero;
\code{'linear'}: remove the linear trend; \code{'none'} do nothing}

\item{signature}{signature to calculate kernel path to save, internally used}

\item{segment_length}{optional positive integer; when provided, long signals
are processed in overlapping segments of this length (in samples) using
batched \code{\link{mvfftw_c2c}} convolutions instead of a single full-length
FFT. This dramatically reduces peak memory and FFT cost for long recordings
(e.g. multi-hour). Must be strictly greater than the longest wavelet kernel
length (the kernel at the lowest frequency); otherwise an error is raised.
Default is \code{NULL}, which uses the legacy single-shot path. When
\code{segment_length >= length(data)} the function silently falls back to
the legacy path. Results match the legacy path on the interior of the
signal up to floating-point error; the first/last
\code{ceiling(max_kernel_len/2)} samples may differ near the global
boundaries (both implementations have boundary artifacts there).}

\item{...}{further passed to \code{\link{detrend}};}

\item{frequency_range}{frequency range to calculate, default is 2 to 200}

\item{cycle_range}{number of cycles corresponding to \code{frequency_range}.
For default frequency range (2 - 200), the default \code{cycle_range} is
3 to 20. That is, 3 wavelet kernel cycles at 2 Hertz, and 20 cycles at 200
Hertz.}
}
\value{
\code{wavelet_kernels} returns wavelet kernels to be
used for wavelet function; \code{morlet_wavelet} returns a file-based array
if \code{precision} is \code{'float'}, or a list of real and imaginary
arrays if \code{precision} is \code{'double'}
}
\description{
Transform analog voltage signals with 'Morlet'
wavelets: complex wavelet kernels with \eqn{\pi/2} phase
differences.
}
\examples{

\donttest{

# generate sine waves
time <- seq(0, 3, by = 0.01)
x <- sin(time * 20*pi) + exp(-time^2) * cos(time * 10*pi)

plot(time, x, type = 'l')

# freq from 1 - 15 Hz; wavelet using float precision
freq <- seq(1, 15, 0.2)
coef <- morlet_wavelet(x, freq, 100, c(2,3))

# to get coefficients in complex number from 1-10 time points
coef[1:10, ]

# power
power <- Mod(coef[])^2

# Power peaks at 5Hz and 10Hz at early stages
# After 1.0 second, 5Hz component fade away
image(power, x = time, y = freq, ylab = "frequency")

# wavelet using double precision
coef2 <- morlet_wavelet(x, freq, 100, c(2,3), precision = "double")
power2 <- (coef2$real[])^2 + (coef2$imag[])^2

image(power2, x = time, y = freq, ylab = "frequency")

# The maximum relative change of power with different precisions
max(abs(power/power2 - 1))

# display kernels
freq <- seq(1, 15, 1)
kern <- wavelet_kernels(freq, 100, c(2,3))
print(kern)

plot(kern)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast_quantile.R
\name{fast_quantile}
\alias{fast_quantile}
\alias{fast_median}
\alias{fast_mvquantile}
\alias{fast_mvmedian}
\title{Compute quantiles}
\usage{
fast_quantile(x, prob = 0.5, na.rm = FALSE, ...)

fast_median(x, na.rm = FALSE, ...)

fast_mvquantile(x, prob = 0.5, na.rm = FALSE, ...)

fast_mvmedian(x, na.rm = FALSE, ...)
}
\arguments{
\item{x}{numerical-value vector for \code{fast_quantile} and
\code{fast_median}, and column-major matrix for \code{fast_mvquantile} and
\code{fast_mvmedian}}

\item{prob}{a probability with value from 0 to 1}

\item{na.rm}{logical; if true, any \code{NA} are removed from \code{x}
before the quantiles are computed}

\item{...}{reserved for future use}
}
\value{
\code{fast_quantile} and \code{fast_median} calculate univariate
quantiles (single-value return); \code{fast_mvquantile} and \code{fast_mvmedian}
calculate multivariate quantiles (for each column, result lengths equal to
the number of columns).
}
\description{
Compute quantiles
}
\examples{

fast_quantile(runif(1000), 0.1)
fast_median(1:100)

x <- matrix(rnorm(100), ncol = 2)
fast_mvquantile(x, 0.2)
fast_mvmedian(x)

# Compare speed for vectors (usually 30\% faster)
x <- rnorm(10000)
microbenchmark::microbenchmark(
  fast_median = fast_median(x),
  base_median = median(x),
  # bioc_median = Biobase::rowMedians(matrix(x, nrow = 1)),
  times = 100, unit = "milliseconds"
)

# Multivariate cases
# (5~7x faster than base R)
# (3~5x faster than Biobase rowMedians)
x <- matrix(rnorm(100000), ncol = 20)
microbenchmark::microbenchmark(
  fast_median = fast_mvmedian(x),
  base_median = apply(x, 2, median),
  # bioc_median = Biobase::rowMedians(t(x)),
  times = 10, unit = "milliseconds"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mris-remesh.R
\name{mris_remesh}
\alias{mris_remesh}
\title{Isotropic re-triangulation of a surface mesh}
\usage{
mris_remesh(
  mesh,
  target_edge_length = NULL,
  niterations = 5L,
  n_smooth = 2L,
  damping = 0.99,
  verbose = FALSE
)
}
\arguments{
\item{mesh}{triangular mesh of class \code{'mesh3d'} (or coercible via
\code{\link{ensure_mesh3d}}). The mesh \strong{must} be closed, manifold,
and genus-0 (watertight, no boundary or non-manifold edges, single
connected component); \code{mris_remesh} raises an error if such defects
are detected}

\item{target_edge_length}{numeric; desired uniform edge length in the same
units as \code{mesh$vb}. Default \code{NULL} uses the current average edge
length (quality improvement without size change)}

\item{niterations}{integer; number of split/collapse/smooth iterations.
Default \code{5}}

\item{n_smooth}{integer; number of tangential-smooth passes per iteration.
Default \code{2}}

\item{damping}{numeric; fraction of the tangential displacement applied at
each smooth pass (between 0 and 1). Default \code{0.99}}

\item{verbose}{logical; print per-iteration vertex and face counts.
Default \code{FALSE}}
}
\value{
A \code{'mesh3d'} object with \code{vb}, \code{it}, and
\code{normals}. Vertex and face counts differ from the input; the surface
geometry is preserved.
}
\description{
Re-triangulates a closed surface mesh so that all edges approach a uniform
target length, improving triangle regularity and quality without changing
the surface topology or shape. The output has a different vertex and face
count than the input but closely tracks the original surface geometry.
}
\details{
Each iteration applies three steps following the procedure described in the
\strong{References}:
\enumerate{
  \item \strong{Edge split}: every edge longer than
    \eqn{4/3 \times \code{target\_edge\_length}} is split at its midpoint.
    One-edge splits produce 2 sub-triangles; two-edge splits produce 3;
    three-edge splits produce 4 (the standard 1-to-4 uniform refinement).
  \item \strong{Edge collapse}: every edge shorter than
    \eqn{4/5 \times \code{target\_edge\_length}} is collapsed to its
    midpoint. A collapse is skipped if it would flip any surrounding
    face normal (manifold-safety check).
  \item \strong{Tangential smoothing}: \code{n_smooth} passes of
    uniform-weight 1-ring \verb{Laplacian} averaging, with each
    displacement projected onto the vertex tangent plane (the normal
    component is removed) before application. This improves vertex
    regularity while keeping vertices near the original surface.
}
After all iterations, vertex normals are refreshed.

Unlike \code{\link{vcg_uniform_remesh}}, which uses volumetric resampling
and may change topology, this function operates entirely on the mesh
surface and preserves the input genus and manifold structure.
}
\examples{

sphere <- vcg_sphere()
sphere

vcg_average_edge_length(sphere)
plot(sphere)

remeshed <- mris_remesh(
  sphere,
  target_edge_length = 0.3
)

plot(remeshed)



}
\references{
A \verb{remeshing} approach to \verb{multiresolution} modeling.
\emph{Proceedings of Shape \verb{Modelling} International}, 49-58 (2003).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{color_ramp_continuous}
\alias{color_ramp_continuous}
\title{Map continuous values to colors}
\usage{
color_ramp_continuous(
  values,
  clim = range(values, na.rm = TRUE),
  cmap = grDevices::hcl.colors(11),
  ...
)
}
\arguments{
\item{values}{numeric vector of values to map to colors}

\item{clim}{length-two numeric vector giving the value range to map from;
values outside \code{[clim[1], clim[2]]} are clamped to the nearer endpoint
before mapping. Default is \code{range(values, na.rm = TRUE)}}

\item{cmap}{the color ramp to map onto: either a vector of colors (passed
to \code{\link[grDevices]{colorRamp}} to build the ramp function) or a
function such as the one returned by \code{\link[grDevices]{colorRamp}}
that takes a numeric vector with elements in \code{[0, 1]} and returns an
\code{n x 3} (or \code{n x 4}, with alpha) matrix of color-channel values
in \code{[0, 255]}. Default is \code{grDevices::hcl.colors(11)}}

\item{...}{passed to \code{\link[grDevices]{colorRamp}} when \code{cmap} is
a vector of colors (for example, \code{alpha = TRUE} to build an
alpha-aware ramp)}
}
\value{
A character vector of \code{'#RRGGBB'} (or \code{'#RRGGBBAA'})
color strings, the same length as \code{values}.
}
\description{
Linearly maps a numeric vector onto a color ramp, clamping values outside
the given range to the range's endpoints.
}
\examples{

x <- rnorm(100)
col <- color_ramp_continuous(x)

plot(x, col = col, pch = 16)


# Change color palettes with vector of colors
col <- color_ramp_continuous(
  x, cmap = c("lightgreen", "white", "pink"))
plot(x, col = col, pch = 16)

# Using colorRamp
col <- color_ramp_continuous(
  x, cmap = colorRamp(c("black", "orangered", "orange")))
plot(x, col = col, pch = 16)

# Using color ramp palette `function(n) { ... }`
col <- color_ramp_continuous(
  x, cmap = hcl.colors, palette = "Blue-Red 3")
plot(x, col = col, pch = 16)

# Set range
col <- color_ramp_continuous(
  x, clim = c(0, 1),
  cmap = c("black", "orangered", "orange"))
plot(x, col = col, pch = 16)



}
